#! /bin/bash
#
# bdf_publisher
#
# Daemon that looks for finished BDF files under the WCBE-private BDF
# directory, closes them, and moves them to the non-private BDF
# directory.
#
if [ -f /etc/sysconfig/wcbe.conf ]; then
    . /etc/sysconfig/wcbe.conf
fi
if [ -f $WCBE_PATH/share/wcbe_common.sh ]; then
    . $WCBE_PATH/share/wcbe_common.sh
elif [ -f $DEFAULT_WCBE_PATH/share/wcbe_common.sh ]; then
    . $DEFAULT_WCBE_PATH/share/wcbe_common.sh
fi
set_environment

shopt -s nullglob

declare -r USE_XATTRS=yes

function usage {
    echo "Usage: $0"
}

# Close off a BDF file and atomically (if on same filesystem) move it
# to another directory.
#
# Arguments: BDF name (not UID), BDF creation directory path, BDF
# publication directory path
function publish_bdf {
    # close the file
    if [ $USE_XATTRS == yes ]; then
        # bdfclose will set number of integrations extended attribute
        bdfclose $2/$1
    else
        # bdfclose will output number of integrations
        NUM_INT=$( bdfclose $2/$1 | grep -E '^[[:digit:]]+' )
        if [ -z "$NUM_INT" ]; then
            NUM_INT=0
        fi
    fi
    if [ $? -ne 0 ]; then
        echo "Failed to close off $2/$1" >&2
        touch $2/FAILED_TO_CLOSE
        return 1
    fi
    # (hard) link the BDF out of private area
    ln $2/$1 $3
    if [ $? -ne 0 ]; then
        echo "Failed to link $2/$1 out of private directory" >&2
        touch $2/FAILED_TO_LINK
        return 1
    fi
    # remove the private BDF file creation directory
    rm -rf $2
    if [ $? -ne 0 ]; then
        echo "Failed to remove published BDF directory $2" >&2
        touch $2/FAILED_TO_REMOVE
        return 1
    fi
    # get BDF file metadata
    if [ $USE_XATTRS == yes ]; then
        BDFUID=$( getfattr --only-values -n user.edu.nrao.widar.uid \
                  $3/$1 2> /dev/null )
        SUBARRAY=$( getfattr --only-values -n user.edu.nrao.widar.subarray \
                    $3/$1 2> /dev/null )
        SCAN_NUM=$( getfattr --only-values -n user.edu.nrao.widar.scan \
                    $3/$1 2> /dev/null )
        SUBSCAN_NUM=$( getfattr --only-values -n user.edu.nrao.widar.subscan \
                       $3/$1 2> /dev/null )
        NUM_INT=$( getfattr --only-values -n user.edu.nrao.widar.integrations \
                   $3/$1 2> /dev/null )
    else
        BDFMETA=$( grep ^$1 $3/.private/.meta )
        if [ -n "$BDFMETA" ]; then
            BDFUID=$( echo "$BDFMETA" | cut -f 2 )
            SUBARRAY=$( echo "$BDFMETA" | cut -f 3 )
            SCAN_NUM=$( echo "$BDFMETA" | cut -f 4 )
            SUBSCAN_NUM=$( echo "$BDFMETA" | cut -f 5 )
        fi
        grep --invert-match ^$1 $3/.private/.meta >| $3/.private/.meta.next
        mv --force $3/.private/.meta.next $3/.private/.meta
    fi
    DATA_SIZE=$( ls -l $3/$1 | tr -s ' ' '\t' | cut -f 5 )
    # send metadata to MCAF via UDP
    if [[ ( -n $BDFUID ) && ( -n $SUBARRAY ) && ( -n $SCAN_NUM ) \
          && ( -n $SUBSCAN_NUM ) && ( -n $NUM_INT ) ]]; then
        cat >| /dev/udp/239.192.5.1/55000 2> /dev/null <<EOF
<BdfInfo>
  <uid>$BDFUID</uid>
  <subarrayId>$SUBARRAY</subarrayId>
  <scanNumber>$SCAN_NUM</scanNumber>
  <subscanNumber>$SUBSCAN_NUM</subscanNumber>
  <numIntegrations>$NUM_INT</numIntegrations>
  <dataSize>$DATA_SIZE</dataSize>
</BdfInfo>
EOF
        echo "Sent BDF message ($BDFUID, $SUBARRAY, $SCAN_NUM, $NUM_INT, $DATA_SIZE) to MCAF"
    else
        echo "Failed to retrieve all BDF file metadata for $1" >&2
    fi
    return 0
}

# First, we need the BDF directory
check_bdfdir
if [ $? -ne 0 ]; then
    echo "Failed to locate BDF root directory" >&2
    exit -1
fi

# Write a PID file for monitoring this process
# PIDFILE=./$0.pid
# function rm_pidfile {
#     if [ -f $PIDFILE ]; then
#         rm $PIDFILE
#     fi
# }
# trap rm_pidfile 0
# echo $$ > $PIDFILE

# time in seconds for a BDF file to be declared "stale"
declare -i STALE=$(( 30 * 60 ))

# Main loop.
SLEEP_INTERVAL=1s
while [ 1 ]; do
    sleep $SLEEP_INTERVAL
    # Search through the BDF directories in the private area
    for d in $WCBE_BDFDIR/.private/uid* ; do
        BDFNAME=${d#$WCBE_BDFDIR/.private/}
        BDFFILE=$d/$BDFNAME
        FAIL=""
        # Finding a BDF file in the directory is the usual case
        if [ -f $BDFFILE ]; then
            # Check for .stamp file
            STAMPFILE=$d/.stamp
            if [ -f $STAMPFILE ]; then
                # If BDF mtime is newer than .stamp file's mtime,
                # remove .stamp file
                if [ $BDFFILE -nt $STAMPFILE ]; then
                    rm $STAMPFILE
                    if [ $? -ne 0 ]; then
                        echo "FATAL: Failed to remove .stamp file from $d" >&2
                        exit -1
                    fi
                fi
            fi
            # If no lock files exist in directory, "publish" the BDF
            LOCKS=$( ls $d/{*.lck,.stamp} 2> /dev/null )
            if [ -z "$LOCKS" ]; then
                publish_bdf $BDFNAME $d $WCBE_BDFDIR
                FAIL=$?
            else
                # We try to clean up "stale" BDFs by copying them
                # before publishing to prevent errant writers from
                # messing with the file after publishing.
                NOW=$( date +%s )
                MTIME=$( stat --format=%Y $BDFFILE )
                if [ $USE_XATTRS == yes ]; then
                    START=$( getfattr --only-values \
                             -n user.edu.nrao.widar.start $BDFFILE \
                             2> /dev/null )
                else
                    START=$( grep $BDFNAME $WCBE_BDFDIR/.private/.meta \
                             | cut -f 6 )
                fi
                if [ -n "$START" ]; then
                    declare -i AGE=$(( $NOW - $START ))
                    declare -i UNCHG=$(( $NOW - $MTIME ))
                    if [ $UNCHG -gt $STALE -a $AGE -gt $STALE ]; then
                        cp $BDFFILE $BDFFILE.cpy \
                            && mv $BDFFILE $BDFFILE.stale \
                            && mv $BDFFILE.cpy $BDFFILE
                        if [ $? -eq 0 ]; then
                            publish_bdf $BDFNAME $d $WCBE_BDFDIR
                            FAIL=$?
                        else
                            echo "Failed to copy stale BDF file $BDFFILE" >&2
                            touch $d/FAILED_TO_COPY_STALE_BDF
                            FAIL=1
                        fi
                    fi
                fi
            fi
        else
            rm -rf $d
            if [ $? -ne 0 ]; then
                echo "Failed to remove empty directory $d" >&2
                touch $d/FAILED_TO_REMOVE
                FAIL=1
            fi
        fi
        # If publishing failed, move the directory to "limbo"
        if [[ ( -n $FAIL ) && ( $FAIL -ne 0 ) ]]; then
            # Create "limbo" as needed
            LIMBO=$WCBE_BDFDIR/.private/limbo
            if [ ! -d $LIMBO ]; then
                mkdir $LIMBO
                if [ $? -ne 0 ]; then
                    echo "FATAL: failed to create $LIMBO" >&2
                    exit -1
                fi
            fi
            mv $d $LIMBO
            if [ $? -ne 0 ]; then
                echo "Failed to move $d to $LIMBO: will delete" >&2
                rm -rf $d
                if [ $? -ne 0 ]; then
                    echo "FATAL: failed to remove $d" >&2
                fi
            fi
        fi
    done
done

# Note: One idea for scaling up this functionality is for a master
# script to spawn multiple instances of this script. Each of the child
# processes would then be responsible for monitoring a subset of uids.
# Mapping of uids to child processes could be done by taking hashes of
# the uids, and assigning a subset of hash values to each child
# process, for example by using a modulo operation on the hash value.
# Another possibility is for the directory/file creator
# (start_subscan) to add an extended attribute to the directory/file
# containing the uid hash value (or something similar), which would
# save continually recomputing the uid hash.
#
# Another possibility may be to have the BDFs created in various
# sub-directories (perhaps depending on a hash of the UIDs), with one
# bdf_publisher process responsible for one sub-directory. This would
# reduce the file metadata operations required of any single
# bdf_publisher process, and would avoid duplicate metadata operations
# among the processes in a set.
#
# Maybe there will never be very many BDFs being created
# simultaneously, in which case one bdf_publisher process might be
# sufficient.
