#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib.h>
#include <glib/gstdio.h>
#include <stdlib.h>
#include <errno.h>
#include <we_bdf.h>
#include <wcbe_config.h>

#define NSEC_PER_SEC 1000000000uLL
#define SEC_PER_DAY (24 * 60 * 60)
#define MJD_OFFSET (40587uLL * SEC_PER_DAY)

static sdmhdr_bincomp_t auto_data = {
    .axes = {[COBDF_TIME_AXIS] =                 FALSE,
             [COBDF_BASELINE_AXIS] =             FALSE,
             [COBDF_ANTENNA_AXIS] =              TRUE,
             [COBDF_BASEBAND_AXIS] =             TRUE,
             [COBDF_SPECTRAL_WINDOW_AXIS] =      TRUE,
             [COBDF_BIN_AXIS] =                  TRUE,
             [COBDF_ATM_PHASE_CORRECTION_AXIS] = FALSE,
             [COBDF_SPECTRAL_POINT_AXIS] =       TRUE,
             [COBDF_POLARIZATION_AXIS] =         TRUE}
};
static sdmhdr_bincomp_t cross_data = {
    .axes = {[COBDF_TIME_AXIS] =                 FALSE,
             [COBDF_BASELINE_AXIS] =             TRUE,
             [COBDF_ANTENNA_AXIS] =              FALSE,
             [COBDF_BASEBAND_AXIS] =             TRUE,
             [COBDF_SPECTRAL_WINDOW_AXIS] =      TRUE,
             [COBDF_BIN_AXIS] =                  TRUE,
             [COBDF_ATM_PHASE_CORRECTION_AXIS] = FALSE,
             [COBDF_SPECTRAL_POINT_AXIS] =       TRUE,
             [COBDF_POLARIZATION_AXIS] =         TRUE}
};

static sdmhdr_data_struct_t sdmhdr_data_struct = {
    .flags = NULL, .actual_times = NULL,
    .actual_durations = NULL, .zero_lags = NULL,
    .auto_data = NULL, .cross_data = NULL,
    .weights = NULL, .type = NULL
};

static sdmhdr_dataoid_t dataoid = {
    .title = "EVLA WIDAR correlator visibility data",
    .type = "locator",
    .href = NULL
};

static sdmhdr_execblock_t execblock = {
    .type = "simple",
    .href = NULL
};

static void parse_command_line(gint argc, gchar *argv[],
                               gchar **config_filename, gchar **uid,
                               gchar **project_path, gint64 *start);

gint
main(gint argc, gchar *argv[])
{
    gchar *config_filename = NULL;
    gchar *uid = NULL;
    gchar *project_path = NULL;
    gint64 start = -1;
    gint result = EXIT_FAILURE;
    gboolean need_auto = FALSE;
    gboolean need_cross = FALSE;

    // parse command line
    parse_command_line(argc, argv, &config_filename, &uid, &project_path,
                       &start);
    if (config_filename == NULL || uid == NULL || project_path == NULL
        || start == -1) {
        g_printerr("Usage: %s [CONFIG] [UID] [PRJ PATH] [START]\n",
                   g_path_get_basename(argv[0]));
        goto cleanup_and_return;
    }

    // parse config_file
    gchar *doc = NULL;
    GError *err = NULL;
    gboolean rc = g_file_get_contents(config_filename, &doc, NULL, &err);
    if (!rc) {
        g_printerr("Failed to read config file (%s) contents: %s\n",
                   config_filename, err->message);
        goto cleanup_and_return;
    }
    wcbe_config_t config;
    rc = wcbe_config_parse(doc, &config);
    g_free(doc);
    if (!rc) {
        g_printerr("Failed to parse config file (%s)\n", config_filename);
        goto cleanup_and_return;
    }

    if (config.output.subarrays != NULL) {
        gchar *tmpname = NULL;
        gint tmpfd = -1;
        FILE *tmpfile = NULL;

        // TODO: accept more than one subarray
        g_assert(config.output.subarrays->len == 1);

        wcbe_subarray_config_t *sa_config =
            &g_array_index(config.output.subarrays, wcbe_subarray_config_t, 0);
        guint64 integration = sa_config->integration * (guint64)1000; // ns

        // create basebands array for sdm hdr
        GPtrArray *bbs =
            g_ptr_array_sized_new(sa_config->basebands->len + 1);
        // loop over basebands in subarray config
        guint bb_idx;
        for (bb_idx = 0; bb_idx < sa_config->basebands->len; ++bb_idx) {
            wcbe_baseband_config_t *bb_config = &g_array_index(
                sa_config->basebands, wcbe_baseband_config_t, bb_idx);
            // create baseband instance for sdm hdr
            sdmhdr_bb_t *bb =
                g_malloc0(SIZEOF_SDMHDR_BB(bb_config->spectral_windows->len));
            bb->num_sw = bb_config->spectral_windows->len;
            // loop over spectral windows in baseband config
            guint sw_idx;
            for (sw_idx = 0; sw_idx < bb_config->spectral_windows->len;
                 ++sw_idx) {
                wcbe_spectral_window_config_t *sw_config = &g_array_index(
                    bb_config->spectral_windows, wcbe_spectral_window_config_t,
                    sw_idx);
                // fill in spectral window instance for sdm hdr
                sdmhdr_sw_t *sw = &(bb->sw[sw_idx]);
                sw->num_bin = sw_config->num_bins;
                sw->num_spectral_point = sw_config->num_channels;
                sw->sd_pol_products = sw_config->sdpp;
                sw->cross_pol_products = sw_config->xpp;
                // determine whether we're recording auto_data,
                // cross_data, or both
                if (sw->sd_pol_products != SDMHDR_APP_NONE
                    && sw->sd_pol_products != SDMHDR_APP_INVALID) {
                    need_auto = TRUE;
                }
                if (sw->cross_pol_products != SDMHDR_XPP_NONE
                    && sw->cross_pol_products != SDMHDR_XPP_INVALID) {
                    need_cross = TRUE;
                }
            }
            g_ptr_array_add(bbs, bb);
        }
        g_ptr_array_add(bbs, NULL);

        // add baseband array to sdm hdr
        sdmhdr_data_struct.basebands = (sdmhdr_bb_t **)bbs->pdata;

        // initialize sdm hdr APC array
        sdmhdr_data_struct.apc[0] = COBDF_AP_UNCORRECTED;
        sdmhdr_data_struct.apc[1] = COBDF_AP_INVALID;

        // initialize dataOID href for subscan
        dataoid.href = g_strdup(uid);

        // initialize exec block href
        execblock.href = g_strdup(uid); // FIXME

        // initialize top-level axis list
        CobdfAxisName dim[] = {COBDF_TIME_AXIS, COBDF_INVALID_AXIS};

        // select whether to record cross data, auto data or both
        if (need_cross) {
            sdmhdr_data_struct.cross_data = &cross_data;
        }
        if (need_auto) {
            sdmhdr_data_struct.auto_data = &auto_data;
        }

        // initialize the main header structure
        sdmhdr_main_hdr_t hdr = {
            .data_struct = &sdmhdr_data_struct,
            .num_antennas = sa_config->stations->len,
            .schema_version = 1,
            .revision = "bdf_1_0",
            .main_header_id = "sdmDataHeader",
            .byte_order = G_BYTE_ORDER,
            .time = ((start + MJD_OFFSET) * NSEC_PER_SEC) + integration / 2,
            .data_oid = &dataoid,
            .exec_block = &execblock,
            .resolution = COBDF_FULL_RESOLUTION,
            .dimensionality = dim,
            .project_path = g_strdup(project_path)
        };
        if (sdmhdr_data_struct.auto_data != NULL) {
            if (sdmhdr_data_struct.cross_data != NULL) {
                sdmhdr_data_struct.type = "CrossAndAutoDataFullResolution";
                hdr.mode = COBDF_CROSS_AND_AUTO;
                sdmhdr_set_cross_data_size(&hdr);
            }
            else {
                sdmhdr_data_struct.type = "AutoDataFullResolution";
                hdr.mode = COBDF_AUTO_ONLY;
            }
            sdmhdr_set_auto_data_size(&hdr);
        }
        else if (sdmhdr_data_struct.cross_data != NULL) {
            sdmhdr_data_struct.type = "CrossDataFullResolution";
            hdr.mode = COBDF_CROSS_ONLY;
            sdmhdr_set_cross_data_size(&hdr);
        }
        //sdmhdr_set_zero_lags_size(&hdr);

        // create a temporary BDF file
        err = NULL;
        tmpfd = g_file_open_tmp("XXXXXX.bdf", &tmpname, &err);
        if (err != NULL) {
            g_printerr("Failed to open a temporary file: %s\n", err->message);
            goto sa_cleanup;
        }
        gint rc = g_unlink(tmpname);
        if (rc == -1) {
            g_printerr("Failed to unlink temporary BDF file '%s': %s\n",
                       tmpname, g_strerror(errno));
        }

        tmpfile = fdopen(tmpfd, "w+");
        if (tmpfile == NULL) {
            g_printerr("Failed to open stream for temporary BDF file "
                       "descriptor: %s\n", g_strerror(errno));
            goto sa_cleanup;
        }

        // initialize the BDF
        we_bdf_t *bdf = we_bdf_new(tmpfile, &hdr);
        if (bdf == NULL) {
            g_printerr("Failed to create temporary BDF file\n");
            goto sa_cleanup;
        }
        we_bdf_free(bdf);

        // copy the BDF to stdout
        rc = fseeko(tmpfile, 0, SEEK_SET);
        if (rc != 0) {
            g_printerr("Failed to seek to start of temporary BDF file: %s\n",
                       g_strerror(errno));
            goto sa_cleanup;
        }
        flockfile(tmpfile);
        flockfile(stdout);
        gchar buf[4096];
        gsize nb = fread_unlocked(buf, 1, sizeof(buf), tmpfile);
        if (G_UNLIKELY(errno != 0)) {
            g_printerr("Read error on temporary BDF file: %s\n",
                       g_strerror(errno));
        }
        while (errno == 0 && nb > 0) {
            fwrite_unlocked(buf, 1, nb, stdout);
            if (G_UNLIKELY(errno != 0)) {
                g_printerr("Write error on stdout: %s\n", g_strerror(errno));
            }
            else {
                nb = fread_unlocked(buf, 1, sizeof(buf), tmpfile);
                if (G_UNLIKELY(errno != 0)) {
                    g_printerr("Read error on temporary BDF file: %s\n",
                               g_strerror(errno));
                }
            }
        }
        funlockfile(stdout);
        funlockfile(tmpfile);
        result = EXIT_SUCCESS;

    sa_cleanup:
        for (bb_idx = 0; bb_idx < bbs->len; ++bb_idx) {
            sdmhdr_bb_t *bb = (sdmhdr_bb_t *)g_ptr_array_index(bbs, bb_idx);
            g_free(bb);
        }
        g_ptr_array_free(bbs, TRUE);
        if (dataoid.href != NULL) {
            g_free(dataoid.href);
        }
        if (execblock.href != NULL) {
            g_free(execblock.href);
        }
        if (hdr.project_path != NULL) {
            g_free(hdr.project_path);
        }
        if (tmpname != NULL) {
            g_free(tmpname);
        }
        if (tmpfile != NULL) {
            fclose(tmpfile);
        }
        if (tmpfd != -1) {
            close(tmpfd);
        }
    }
    else {
        g_printerr("Configuration has no subarray definition, no BDF will "
                   "be created\n");
    }
    wcbe_config_free(&config);

 cleanup_and_return:
    if (config_filename != NULL) {
        g_free(config_filename);
    }
    if (uid != NULL) {
        g_free(uid);
    }
    if (project_path != NULL) {
        g_free(project_path);
    }

    return result;
}

static void
parse_command_line(gint argc, gchar *argv[], gchar **config_filename,
                   gchar **uid, gchar **project_path, gint64 *start)
{
    if (argc == 5) {
        *config_filename = g_strdup(argv[1]);
        *uid = g_strdup(argv[2]);
        *project_path = g_strdup(argv[3]);
        errno = 0;
        guint64 ustart = g_ascii_strtoull(argv[4], NULL, 0);
        if (errno == 0 && ustart <= G_MAXINT64) {
            *start = (gint64)ustart;
        }
        else {
            g_printerr("Failed to convert START argument\n");
        }
    }
}
