#! /usr/bin/env python
from wcbe import *
import os
import os.path
import stat
from time import strftime, time
from grp import getgrnam

class FifoError (Exception):
    """FIFO already exists.
    """
    pass

def create_fifo():
    """Create FIFO that receives config file paths. Fail if FIFO
    already exists (this is done to help avoid multiple instances of
    the config_listener from reading from the same FIFO).
    """
    if os.path.exists(os.environ['WCBE_CONFIG_FIFO']):
        raise FifoError
    os.mkfifo(os.environ['WCBE_CONFIG_FIFO'])
    grp = getgrnam(os.environ['WCBE_GROUP'])
    os.chown(os.environ['WCBE_CONFIG_FIFO'], -1, grp.gr_gid)
    os.chmod(os.environ['WCBE_CONFIG_FIFO'],
             stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IWGRP)
    return

def remove_fifo():
    """Remove FIFO that receives config file paths.
    """
    if os.path.exists(os.environ['WCBE_CONFIG_FIFO']):
        mode = os.stat(os.environ['WCBE_CONFIG_FIFO'])[stat.ST_MODE]
        if stat.S_ISFIFO(mode):
            os.unlink(os.environ['WCBE_CONFIG_FIFO'])
    return

class ConfigInstance (object):
    """Handy-dandy for maintaining state information about an instance
    during the configuration process.
    """
    def __init__(self, path, config_name):
        self.state = 'start'
        self.path = path
        self.name = os.path.basename(path)
        self.config_path = os.path.join(self.path, 'configurations',
                                        config_name)
        self.activate_path = os.path.join(self.path, 'activations',
                                          'now-%s' % config_name)
        self.active_link = os.path.join(self.path, 'active')
        return

def main(lfxset):
    create_fifo()
    while True:
        fifo = open(os.environ['WCBE_CONFIG_FIFO'])
        config_path = fifo.readline().strip()
        fifo.close()
        if os.path.isfile(config_path):
            instances = pipelinefs_glob(['*'], lfxset)
            config = open(config_path, 'r')
            config_name = strftime('%H_%M_%S')
            config_instances = [ConfigInstance(i, config_name)
                                for i in instances]
            stoptime = time() + 10.0
            while len(config_instances) > 0 and time() < stoptime:
                try:
                    i = config_instances.pop(0)
                    if i.state == 'start':
                        config.seek(0)
                        c = open(i.config_path, 'w')
                        c.writelines(config.readlines())
                        c.close()
                        i.state = 'configuring'
                    elif i.state == 'configuring':
                        if os.path.isdir(i.config_path):
                            i.state = 'configured'
                    if i.state == 'configured':
                        a = open(i.activate_path, 'w')
                        a.write('1')
                        a.close()
                        i.state = 'activating'
                    elif i.state == 'activating':
                        if os.path.islink(i.active_link) \
                               and (os.path.realpath(i.active_link)
                                    == i.config_path):
                            i.state = 'activated'
                    if i.state == 'activated':
                        os.rmdir(i.config_path)
                        i.state = 'done'
                        i = None
                except OSError:
                    pass
                except IOError, e:
                    print 'Failed to configure instance %s with %s' \
                          % (i.name, config_path)
                    i = None

                if i is not None:
                    config_instances.append(i)

            config.close()
            if len(config_instances) > 0:
                for i in config_instances:
                    print 'Configuration of instance %s with %s stalled' \
                          % (i.name, config_path)

if __name__ == '__main__':
    import sys
    set_environment('/etc/sysconfig/wcbe.conf')
    if not check_pipelinefs():
        print 'WCBE_PIPELINEFS environment variable (%s) does not ' \
              'refer to a valid pipelinefs file system.' \
              % os.environ['WCBE_PIPELINEFS']
        sys.exit(-1)
    if not check_bdfdir():
        print 'WCBE_BDFDIR environment variable (%s) does not refer ' \
              'to a valid directory for BDF file creation.' \
              % os.environ['WCBE_BDFDIR']
        sys.exit(-1)
    lfxset = get_lfxset()
    try:
        main(lfxset)
    except KeyboardInterrupt:
        remove_fifo()
    except:
        remove_fifo()
