// lagset_pipeline.c
//
// Lag set processing stage pipeline of WCBE.
//
/**
 * SECTION: lagset_pipeline
 * @short_description: Lag set processing stage pipeline of WCBE.
 * @stability: unstable
 *
 * Implementation of lag set processing stage of WCBE (on compute
 * nodes). Normally these processes are started and stopped by the
 * wcbe program automatically, as required by the WCBE configuration.
 * However, this program may also be used from the command line for
 * simulation or testing purposes.
 *
 * The pipelines in the lag set processing stage are dynamically
 * configured with any GStreamer pipeline elements (although, of
 * course, the elements should be able to process lag sets and
 * subsequent data types). Essentially, the lagset_pipeline program
 * simply takes a GStreamer pipeline specification, instantiates the
 * pipeline and wraps it in such a way as to be usable in the lag set
 * processing stage of WCBE. The only constraint on the pipelines that
 * may be created is that the first element in the pipeline must have
 * a %WCBE_LAGSET_MEDIA_TYPE capability on its sink pad. Note that a
 * #WcbeLagsetSrc element is implicitly added to the head of the
 * pipeline specification to provide the interface with the WCBE input
 * stage pipeline.
 *
 * <cmdsynopsis>
 *   <command>lagset_pipeline</command>
 *   <group choice='opt'>
 *     <group choice='req'>
 *       <arg choice='plain'>-i</arg>
 *       <arg choice='plain'>--input</arg>
 *     </group>
 *     <replaceable>NAME</replaceable>
 *   </group>
 *   <group choice='opt'>
 *     <group choice='req'>
 *       <arg choice='plain'>-o</arg>
 *       <arg choice='plain'>--output</arg>
 *     </group>
 *     <replaceable>NAME</replaceable>
 *   </group>
 *   <group choice='opt'>
 *     <group choice='opt'>
 *       <arg choice='plain'>-t</arg>
 *       <arg choice='plain'>--integration</arg>
 *     </group>
 *     <replaceable>INTERVAL (microseconds)</replaceable>
 *   </group>
 *   <group choice='opt'>
 *     <group choice='opt'>
 *       <arg choice='plain'>-c</arg>
 *       <arg choid='plain'>--config</arg>
 *     </group>
 *     <replaceable>FILE</replaceable>
 *   </group>
 *   <arg choice='plain'><replaceable>PIPELINE</replaceable></arg>
 * </cmdsynopsis>
 *
 * In the above command synopsis, <replaceable>NAME</replaceable> is
 * the lag set source (Unix) socket name, and
 * <replaceable>PIPELINE</replaceable> is the pipeline specification
 * in the syntax used by the <command>gst-launch</command> program.
 * The <replaceable>INTERVAL</replaceable> is used to specify the
 * integration time for the pipeline; because a single pipeline is
 * meant to process only lag sets for one subarray, there can be only
 * one integration time given. The "config" argument is used to name a
 * configuration file; this file name is optional, and the default is
 * to look for the file "config.xml" in the same directory in which
 * the input socket is found. The "config" file is nevertheless
 * optional, since there may be cases in which the pipeline does not
 * need that file.
 */
#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <gst/gst.h>

#include <term_source.h>
#include <enums.h>
#include <wcbe.h>
#include <wcbe_lagset_src.h>
#include <async_syslog.h>
#include <user_marshal.h>

/* Names of Unix sockets used by the program. TODO: These should be
 * defined elsewhere so that they are easier to keep consistent across
 * programs that use them.
 */
/* Name of socket on which lag sets are received. */
#define LAGSET_SRC_NAME "src"
/* Name of socket used to pass messages to/from the input stage pipeline. */
#define INPUT_STAGE_SOCK_NAME "issink"

/*
 * pipeline_t:
 * @pipeline: pointer to the GStreamer pipeline
 * @input_stage: full path to comm socket to input stage
 * @name: name of lagset pipeline
 * @sock: lag set input socket
 * @finished: %TRUE, if pipeline has received its last lag frame, but
 *            has not yet exited; %FALSE, otherwise
 *
 * Internal representation of pipeline created by the process.
 */
typedef struct pipeline {
    GstElement *pipeline;
    gchar *input_stage;
    gchar *pipeline_name;
    gchar *wcbe_name;
    gchar *output_dir;
    gint sock;
    gboolean finished;
    wcbe_config_t *config;
    guint integration;
    GSList *tmpfiles;
    GStaticMutex mtx;
} pipeline_t;

static void parse_options(gint argc, gchar *argv[], gchar **input,
                          guint *integration, gchar ***pipeline_desc,
                          gchar **config, gchar **output, gchar **pidsdir);
static void run_pipeline(pipeline_t *pipeline);
static gboolean bus_cb(GstBus *bus, GstMessage *msg, gpointer data);
static void forward_message_upstream(pipeline_t *pipeline,
                                     GstStructure *structure);
static void handle_stream_msg(pipeline_t *pipeline,
                              const GstStructure *structure);
static void element_list_query(pipeline_t *pipeline, GstStructure *structure);
static void property_list_query(pipeline_t *pipeline, GstStructure *structure);
static void property_query(pipeline_t *pipeline, GstStructure *structure);
static void property_command(pipeline_t *pipeline, GstStructure *structure);
static void pad_list_query(pipeline_t *pipeline, GstStructure *structure);
static void pad_caps_query(pipeline_t *pipeline, GstStructure *structure);
static void pad_peer_query(pipeline_t *pipeline, GstStructure *structure);
static void post_return_message(pipeline_t *pipeline, GstStructure *structure);
static gpointer get_config_cb(GstElement *element, gpointer data);
static gpointer get_wcbe_name_cb(GstElement *element, gpointer data);
static gpointer get_pipeline_name_cb(GstElement *element, gpointer data);
static guint get_be_integration_cb(GstElement *element, gpointer data);
static void register_tmpfile_cb(GstElement *element, gchar *filename,
                                gpointer data);
static void unregister_tmpfile_cb(GstElement *element, gchar *filename,
                                  gpointer data);
static gchar *get_bdfdir_cb(GstElement *element, gpointer data);

gint
main(gint argc, gchar *argv[])
{
    gchar *input = NULL;
    gchar **pipeline_desc = NULL;
    gchar *pidsdir = NULL;
    pipeline_t pipeline = {.pipeline = NULL, .sock = -1, .output_dir = NULL};
    gchar *configname = NULL;

    if (!g_thread_supported()) {
        g_thread_init(NULL);
    }
    g_static_mutex_init(&(pipeline.mtx));
    enums_register();

    async_syslog_init(500);
    async_syslog_set_default_log_handler();

    parse_options(argc, argv, &input, &(pipeline.integration), &pipeline_desc,
                  &configname, &(pipeline.output_dir), &pidsdir);

    // write pid file
    gchar *pidfile = NULL;
    if (pidsdir != NULL) {
        gchar *pid = g_strdup_printf("%u", getpid());
        pidfile = g_build_filename(pidsdir, pid, NULL);
        FILE *f = fopen(pidfile, "w");
        if (f != NULL) {
            gint rc = fclose(f);
            if (rc < 0) {
                g_warning("Failed to close file %s: %s", pidfile,
                          g_strerror(errno));
            }
        }
        else {
            g_critical("Failed to open pid file %s: %s", pidfile,
                       g_strerror(errno));
            return EXIT_FAILURE;
        }
        g_free(pid);
    }
    else {
        g_critical("--pids command line argument is required");
        return EXIT_FAILURE;
    }

    // set default value for 'output_dir'
    if (pipeline.output_dir == NULL) {
        pipeline.output_dir = g_get_current_dir();
    }

    gchar *confdir = g_path_get_dirname(input);

    // write config file in 'confdir'
    gboolean configname_default = FALSE;
    if (configname == NULL) {
        configname = g_build_filename(confdir, "config.xml", NULL);
        configname_default = TRUE;
    }
    pipeline.config = NULL;
    GError *err = NULL;
    gchar *doc = NULL;
    g_file_get_contents(configname, &doc, NULL, &err);
    if (err == NULL) {
        wcbe_config_t *config = g_new0(wcbe_config_t, 1);
        if (wcbe_config_parse(doc, config)) {
            pipeline.config = config;
            if (config->name == NULL) {
                config->name = g_path_get_basename(confdir);
            }
        }
        else {
            g_critical("Failed to parse contents of config file '%s'.",
                       configname);
        }
    }
    else {
        if (!configname_default || err->code != G_FILE_ERROR_NOENT) {
            g_critical("Failed to read contents of config file '%s': %s",
                       configname, err->message);
        }
    }
    if (doc != NULL) {
        g_free(doc);
    }

    gchar *pipedir = g_path_get_dirname(confdir);
    pipeline.wcbe_name = g_path_get_basename(pipedir);
    g_free(pipedir);
    g_free(confdir);

    GPtrArray *full_pipeline = g_ptr_array_new();
    g_ptr_array_add(full_pipeline, g_strdup("wcbe_lagset_src"));
    g_ptr_array_add(full_pipeline, g_strdup("name="LAGSET_SRC_NAME));
    gchar *insock = g_strdup_printf("input=%s", input);
    g_ptr_array_add(full_pipeline, insock);
    g_ptr_array_add(full_pipeline, g_strdup("!"));

    // Now add the rest of the pipeline description.
    gchar **component = pipeline_desc;
    while (component != NULL && *component != NULL) {
        g_ptr_array_add(full_pipeline, g_strdup(*component));
        ++component;
    }
    g_ptr_array_add(full_pipeline, NULL);
    g_free(pipeline_desc);

    err = NULL;
    pipeline.pipeline =
        gst_parse_launchv((const gchar **)full_pipeline->pdata, &err);
    if (err != NULL) {
        g_warning("Failed to create pipeline: %s", err->message);
        if (pipeline.pipeline != NULL) {
            gst_object_unref(GST_OBJECT(pipeline.pipeline));
            pipeline.pipeline = NULL;
        }
    }
    gchar **pdata = (gchar **)g_ptr_array_free(full_pipeline, FALSE);
    g_strfreev(pdata);

    if (pipeline.pipeline != NULL) {
        // TODO: not sure of the encoding the g_path... functions work
        // on; ignored for now, need to experiment
        gchar *basename = g_path_get_basename(input);
        gchar *confdir = g_path_get_dirname(input);
        gchar *confname = g_path_get_basename(confdir);
        gchar *bedir = g_path_get_dirname(confdir);
        pipeline.input_stage =
            g_build_filename(bedir, INPUT_STAGE_SOCK_NAME, NULL);
        pipeline.pipeline_name = g_build_filename(confname, basename, NULL);
        gst_element_set_name(pipeline.pipeline, basename);
        g_free(bedir);
        g_free(confname);
        g_free(confdir);
        g_free(basename);
        pipeline.sock = socket(PF_UNIX, SOCK_DGRAM, 0);
        if (pipeline.sock != -1) {
            gint rc = fcntl(pipeline.sock, F_SETFL, O_NONBLOCK);
            if (rc != 0) {
                GST_ERROR("Failed to set socket to non-blocking mode: %s",
                          g_strerror(errno));
            }
            gint bufsz;
            socklen_t len = sizeof(bufsz);
            rc = getsockopt(pipeline.sock, SOL_SOCKET, SO_RCVBUF, &bufsz,
                            &len);
            bufsz *= 10;
            rc = setsockopt(pipeline.sock, SOL_SOCKET, SO_RCVBUF, &bufsz,
                            sizeof(bufsz));
            rc = setsockopt(pipeline.sock, SOL_SOCKET, SO_SNDBUF, &bufsz,
                            sizeof(bufsz));
            rc = getsockopt(pipeline.sock, SOL_SOCKET, SO_RCVBUF, &bufsz,
                            &len);
            rc = getsockopt(pipeline.sock, SOL_SOCKET, SO_SNDBUF, &bufsz,
                            &len);
            run_pipeline(&pipeline);
            close(pipeline.sock);
        }
        else {
            g_printerr("Failed to open message passing socket: %s",
                       g_strerror(errno));
        }
        g_free(pipeline.input_stage);
        g_free(pipeline.pipeline_name);
        g_free(pipeline.wcbe_name);
        g_free(pipeline.output_dir);
    }
    if (pipeline.config != NULL) {
        wcbe_config_free(pipeline.config);
    }
    g_static_mutex_free(&(pipeline.mtx));
    gint rc = unlink(pidfile);
    if (rc < 0) {
        g_warning("Failed to unlink pid file %s: %s",
                  pidfile, g_strerror(errno));
    }
    g_free(pidfile);

    async_syslog_flush();
    async_syslog_quit();
    return EXIT_SUCCESS;
}

static void
parse_options(gint argc, gchar *argv[], gchar **input, guint *integration,
              gchar ***pipeline_desc, gchar **config, gchar **output,
              gchar **pidsdir)
{
    const GOptionEntry entries[] = {
        {"input", 'i', 0, G_OPTION_ARG_FILENAME, input,
         "Lag set source socket name", "NAME"},
        {"integration", 't', 0, G_OPTION_ARG_INT, integration,
         "Integration interval", "TIME (microseconds)"},
        {"config", 'c', 0, G_OPTION_ARG_STRING, config,
         "WCBE configuration file name", "FILE"},
        {"output", 'i', 0, G_OPTION_ARG_FILENAME, output,
         "Name of directory for output products", "NAME"},
        {"pids", 'p', 0, G_OPTION_ARG_FILENAME, pidsdir,
         "Directory for pid file", "NAME"},
        {NULL}
    };

    GError *error = NULL;
    GOptionContext *context = g_option_context_new(
        "PIPELINE - WIDAR correlator backend lag set pipeline");
    g_option_context_set_summary(
        context,
        "Lag set processing pipeline stage for compute "
        "node of WIDAR correlator backend system.\n"
        "PIPELINE argument should be given in the syntax "
        "used by gst-launch.");
    g_option_context_set_description(
        context,
        "Please report feature requests, bugs, etc. to\n"
        "mpokorny@nrao.edu.");
    g_option_context_set_help_enabled(context, TRUE);
    g_option_context_set_ignore_unknown_options(context, TRUE);
    g_option_context_add_main_entries(context, entries, NULL);
    g_option_context_add_group(context, gst_init_get_option_group());
    gboolean ctxt_parse =
        g_option_context_parse(context, &argc, &argv, &error);

    if (ctxt_parse) {
        GPtrArray *pipeline = g_ptr_array_new();
        guint i;
        for (i = 1; i < argc; ++i) {
            g_ptr_array_add(pipeline, argv[i]);
        }
        g_ptr_array_add(pipeline, NULL);
        *pipeline_desc = (gchar **)g_ptr_array_free(pipeline, FALSE);
    }

    if (!ctxt_parse) {
        if (error != NULL) {
            g_printerr("%s\n", error->message);
            g_error_free(error);
        }
        exit(-1);
    }
    g_option_context_free(context);
}

/*
 * run_pipeline:
 * @pipeline: pipeline instance
 *
 * Start the pipeline, and return only after it has exited. Posts
 * 'pipeline-state' messages to the pipeline message bus on startup
 * and shutdown.
 */
static void
run_pipeline(pipeline_t *pipeline)
{
    GstBus *bus = NULL;

    bus = gst_pipeline_get_bus(GST_PIPELINE(pipeline->pipeline));
    if (bus == NULL) {
        g_printerr("Failed to get pipeline message bus");
        goto cleanup_and_return;
    }
    gst_bus_add_watch(bus, bus_cb, pipeline);

    GstIterator *it = gst_bin_iterate_recurse(GST_BIN(pipeline->pipeline));
    gboolean done = FALSE;
    while (!done) {
        GstElement *element;
        switch (gst_iterator_next(it, (gpointer)&element)) {
          case GST_ITERATOR_OK: {
            if (object_has_signal(G_OBJECT(element), "get-config")) {
                g_signal_connect(element, "get-config",
                                 G_CALLBACK(get_config_cb), pipeline);
            }
            if (object_has_signal(G_OBJECT(element), "get-wcbe-name")) {
                g_signal_connect(element, "get-wcbe-name",
                                 G_CALLBACK(get_wcbe_name_cb), pipeline);
            }
            if (object_has_signal(G_OBJECT(element), "get-pipeline-name")) {
                g_signal_connect(element, "get-pipeline-name",
                                 G_CALLBACK(get_pipeline_name_cb), pipeline);
            }
            if (object_has_signal(G_OBJECT(element), "get-be-integration")) {
                g_signal_connect(element, "get-be-integration",
                                 G_CALLBACK(get_be_integration_cb), pipeline);
            }
            if (object_has_signal(G_OBJECT(element), "register-tmpfile")) {
                g_signal_connect(element, "register-tmpfile",
                                 G_CALLBACK(register_tmpfile_cb), pipeline);
            }
            if (object_has_signal(G_OBJECT(element), "unregister-tmpfile")) {
                g_signal_connect(element, "unregister-tmpfile",
                                 G_CALLBACK(unregister_tmpfile_cb), pipeline);
            }
            if (object_has_signal(G_OBJECT(element), "get-bdfdir")) {
                g_signal_connect(element, "get-bdfdir",
                                 G_CALLBACK(get_bdfdir_cb), pipeline);
            }
            gst_object_unref(GST_OBJECT(element));
            break;
          }
          case GST_ITERATOR_RESYNC:
            gst_iterator_resync(it);
            break;
          case GST_ITERATOR_ERROR:
          case GST_ITERATOR_DONE:
            done = TRUE;
            break;
        }
    }
    gst_iterator_free(it);

    GstStateChangeReturn ret =
        gst_element_set_state(pipeline->pipeline, GST_STATE_PLAYING);
    if (ret != GST_STATE_CHANGE_FAILURE) {
        GMainLoop *loop = g_main_loop_new(NULL, FALSE);

        // set up TermSource to catch termination signals
        pipeline->finished = FALSE;
        gboolean term_handler(gint *signum) {
            pipeline->finished = TRUE;
            return FALSE;
        }
        TermSource *term_source = term_source_new();
        term_source_set_callback(term_source, term_handler);
        term_source_attach(term_source, NULL);

        // On interrupt, try an orderly pipeline shutdown, then quit
        // main loop.
        gboolean done(gpointer data) {
            if (g_atomic_int_get(&(pipeline->finished))) {
                ret = gst_element_set_state(pipeline->pipeline,
                                            GST_STATE_NULL);
                if (ret == GST_STATE_CHANGE_ASYNC) {
                    gst_element_get_state(pipeline->pipeline, NULL, NULL,
                                          10 * GST_SECOND);
                }
                GstMessage *msg = gst_message_new_application(
                    NULL,
                    gst_structure_new(
                        "pipeline-state",
                        "value", PIPELINE_STATE_ENUM_TYPE,
                        PIPELINE_STATE_STOPPED,
                        NULL));
                if (!bus_cb(bus, msg, pipeline)) {
                    g_debug("Failed to post 'pipeline-state:value=stopped' "
                            "message to message bus");
                }
                gst_message_unref(msg);
                g_main_loop_quit(loop);
            }
            return TRUE;
        };
        g_timeout_add(100, done, NULL);

        GstMessage *msg = gst_message_new_application(
            NULL,
            gst_structure_new(
                "pipeline-state",
                "value", PIPELINE_STATE_ENUM_TYPE, PIPELINE_STATE_RUNNING,
                NULL));

        if (!bus_cb(bus, msg, pipeline)) {
            g_debug("Failed to post 'pipeline-state:value=running' message "
                    "to message bus");
        }
        gst_message_unref(msg);

        g_main_loop_run(loop);

        g_main_loop_unref(loop);

        g_debug("cleanup_tmpfiles");
        g_static_mutex_lock(&(pipeline->mtx));
        while (pipeline->tmpfiles != NULL) {
            GSList *tmpfile = pipeline->tmpfiles;
            gchar *filename = (gchar *)(tmpfile->data);
            gint rc = unlink(filename);
            if (G_UNLIKELY(rc == -1)) {
                g_info("Failed to unlink registered temporary file '%s': %s",
                       filename, g_strerror(errno));
            }
            g_free(filename);
            pipeline->tmpfiles =
                g_slist_delete_link(pipeline->tmpfiles, tmpfile);
        }
        g_static_mutex_unlock(&(pipeline->mtx));

    }
    else {
        g_warning("Failed to start pipeline");
        GstMessage *msg = gst_bus_poll(bus, GST_MESSAGE_ERROR, 0);
        if (msg != NULL) {
            GError *error = NULL;
            gst_message_parse_error(msg, &error, NULL);
            g_printerr("ERROR: %s\n", error->message);
            g_error_free(error);
            gst_message_unref(msg);
        }
    }

 cleanup_and_return:
    if (bus != NULL) {
        gst_object_unref(GST_OBJECT(bus));
    }
}

/*
 * bus_cb:
 * @bus: message bus
 * @msg: message
 * @data: pipeline instance that hosts the given message bus
 *
 * Message bus callback for pipeline. The following messages are
 * handled as indicated:
 * <informaltable>
 *   <tgroup>
 *     <thead>
 *       <row>
 *         <entry>Message</entry>
 *         <entry>Action</entry>
 *       </row>
 *     </thead>
 *     <tbody>
 *       <row>
 *         <entry>%GST_MESSAGE_ERROR</entry>
 *         <entry>Report error using g_printerr.</entry>
 *       </row>
 *       <row>
 *         <entry>%GST_MESSAGE_EOS</entry>
 *         <entry><para>
 *           Set flag (pipeline->finished) to trigger pipeline exit.
 *         </para></entry>
 *       </row>
 *       <row>
 *         <entry>%GST_MESSAGE_APPLICATION</entry>
 *         <entry><para>
 *           If the message name indicates that it came from the input
 *           stage, then handle the message in this pipeline.
 *           Otherwise, add a 'src' field to the structure contained
 *           in the message to indicate the current pipeline as the
 *           structure's source, and forward the structure upstream
 *           (to the input stage).
 *         </para></entry>
 *       </row>
 *     </tbody>
 *   </tgroup>
 * </informaltable>
 *
 * Returns: %TRUE
 *
 * @see_also: handle_stream_msg(), forward_message_upstream()
 */
static gboolean
bus_cb(GstBus *bus, GstMessage *msg, gpointer data)
{
    pipeline_t *pipeline = data;

    switch (GST_MESSAGE_TYPE(msg)) {
      case GST_MESSAGE_ERROR: {
        gchar *debug = NULL;
        GError *err = NULL;
        gst_message_parse_error(msg, &err, &debug);
        g_printerr("ERROR: %s\n", err->message);
        if (debug != NULL) {
            g_printerr("Debug details: %s\n", debug);
            g_free(debug);
        }
        g_error_free(err);
        break;
      }
      case GST_MESSAGE_EOS:
        g_atomic_int_set(&(pipeline->finished), TRUE);
        break;
      case GST_MESSAGE_APPLICATION: {
        const GstStructure *structure = gst_message_get_structure(msg);
        if (GST_MESSAGE_SRC(msg) != NULL) {
            gchar *src_name = gst_object_get_name(GST_MESSAGE_SRC(msg));
            if (strcmp(LAGSET_SRC_NAME, src_name) == 0
                && gst_structure_has_name(structure, STREAM_MSG_NAME)) {
                handle_stream_msg(pipeline, structure);
            }
            else {
                gchar *src = g_strjoin(G_DIR_SEPARATOR_S,
                                       pipeline->pipeline_name,
                                       src_name, NULL);
                GstStructure *new_structure = gst_structure_copy(structure);
                gst_structure_set(new_structure,
                                  "src", G_TYPE_STRING, src,
                                  NULL);
                forward_message_upstream(pipeline, new_structure);
                g_free(src);
            }
            g_free(src_name);
        }
        else {
            GstStructure *new_structure = gst_structure_copy(structure);
            gst_structure_set(new_structure,
                              "src", G_TYPE_STRING, pipeline->pipeline_name,
                              NULL);
            forward_message_upstream(pipeline, new_structure);
        }
        break;
      }
      default:
        break;
    }
    return TRUE;
}

/*
 * forward_message_upstream:
 * @pipeline: pipeline instance
 * @structure: message (#GstStructure, which will be freed by this
 *             function)
 *
 * Forward an arbitrary #GstStructure to the input stage. The
 * #GstStructure is serialized by calling gst_structure_to_string().
 */
static void
forward_message_upstream(pipeline_t *pipeline, GstStructure *structure)
{
    gchar *str = gst_structure_to_string(structure);
    guint str_len = strlen(str) + 1;
    struct sockaddr_un addr = {.sun_family = AF_UNIX};
    gint rc = g_strlcpy(addr.sun_path, pipeline->input_stage,
                        sizeof(addr.sun_path));
    g_assert(0 < rc && rc < sizeof(addr.sun_path));
    struct iovec iovec[] =
        {{.iov_base = &str_len, .iov_len = sizeof(str_len)},
         {.iov_base = str, .iov_len = str_len}};
    struct msghdr msghdr = {.msg_name = &addr,
                            .msg_namelen = sizeof(addr),
                            .msg_control = NULL, .msg_controllen = 0,
                            .msg_iov = iovec, .msg_iovlen = 2};
    gssize bytes_written = sendmsg(pipeline->sock, &msghdr, 0);
    if (bytes_written != iovec[0].iov_len + iovec[1].iov_len) {
        g_warning("Failed to write application message (%s) to socket: %s",
                  str, g_strerror(errno));
    }
    g_free(str);
    gst_structure_free(structure);
}

/*
 * handle_stream_msg:
 * @pipeline: pipeline instance
 * @structure: #GstStructure containing message
 *
 * Dispatch a message from the input stage to the appropriate message
 * handling function.
 */
static void
handle_stream_msg(pipeline_t *pipeline, const GstStructure *structure)
{
    if (gst_structure_has_field(structure, "msg")) {
        const gchar *msg = gst_structure_get_string(structure, "msg");
        GstStructure *msg_structure = gst_structure_from_string(msg, NULL);
        if (msg_structure != NULL) {
            const gchar *name = gst_structure_get_name(msg_structure);
            if (strcmp(name, ELEMENT_LIST_MSG_NAME) == 0) {
                element_list_query(pipeline, msg_structure);
            }
            else if (strcmp(name, PROPERTY_LIST_MSG_NAME) == 0) {
                property_list_query(pipeline, msg_structure);
            }
            else if (strcmp(name, GET_PROPERTY_MSG_NAME) == 0) {
                property_query(pipeline, msg_structure);
            }
            else if (strcmp(name, SET_PROPERTY_MSG_NAME) == 0) {
                property_command(pipeline, msg_structure);
            }
            else if (strcmp(name, PAD_LIST_MSG_NAME) == 0) {
                pad_list_query(pipeline, msg_structure);
            }
            else if (strcmp(name, PAD_CAPS_MSG_NAME) == 0) {
                pad_caps_query(pipeline, msg_structure);
            }
            else if (strcmp(name, PAD_PEER_MSG_NAME) == 0) {
                pad_peer_query(pipeline, msg_structure);
            }
            else {
                gst_structure_free(msg_structure);
            }
        }
    }
}

/*
 * element_list_query:
 * @pipeline: pipeline instance
 * @structure: structure with name %ELEMENT_LIST_MSG_NAME
 *
 * Handle an %ELEMENT_LIST_MSG_NAME message.
 */
static void
element_list_query(pipeline_t *pipeline, GstStructure *structure)
{
    // generate a tab-separated list of pipeline elements
    GString *names = g_string_new("");
    GstIterator *elements =
        gst_bin_iterate_recurse(GST_BIN(pipeline->pipeline));
    void element_name(gpointer data, gpointer user_data) {
        const gchar *fmt = ((names->len == 0) ? "%s" : "\t%s");
        const gchar *name = GST_ELEMENT_NAME(GST_ELEMENT(data));
        g_string_append_printf(names, fmt, name);
    }
    GstIteratorResult iter_result =
        gst_iterator_foreach(elements, element_name, NULL);
    gst_iterator_free(elements);
    g_assert(iter_result != GST_ITERATOR_OK);

    // create the return message, then post it on message bus
    if (iter_result == GST_ITERATOR_DONE) {
        gst_structure_set(structure, "list", G_TYPE_STRING, names->str, NULL);
    }
    else {
        gst_structure_set(structure, "list", G_TYPE_STRING, "", NULL);
    }
    post_return_message(pipeline, structure);
    g_string_free(names, TRUE);
}

/*
 * property_list_query:
 * @pipeline: pipeline instance
 * @structure: structure with name %PROPERTY_LIST_MSG_NAME
 *
 * Handle a %PROPERTY_LIST_MSG_NAME message.
 */
static void
property_list_query(pipeline_t *pipeline, GstStructure *structure)
{
    // get list of element properties
    const gchar *element_name =
        gst_structure_get_string(structure, "element");
    GString *properties = g_string_new("");
    GstElement *element = gst_bin_get_by_name(GST_BIN(pipeline->pipeline),
                                              element_name);
    if (element != NULL) {
        guint num_properties;
        GParamSpec **params = g_object_class_list_properties(
            G_OBJECT_GET_CLASS(G_OBJECT(element)),
            &num_properties);
#define PARAM_RW_STR(p) (((p)->flags & G_PARAM_READABLE)    \
                         ? (((p)->flags & G_PARAM_WRITABLE) \
                            ? "rw" : "r")                   \
                         : (((p)->flags & G_PARAM_WRITABLE) \
                            ? "w" : ""))
        guint i;
        for (i = 0; i < num_properties; ++i) {
            GParamSpec *param = params[i];
            const gchar *fmt =
                ((properties->len == 0) ? "%s<%s,%s>" : "\t%s<%s,%s>");
            g_string_append_printf(properties, fmt, param->name,
                                   g_type_name(param->value_type),
                                   PARAM_RW_STR(param));
        }
#undef PARAM_RW_STR
        g_free(params);
        gst_object_unref(GST_OBJECT(element));
    }
    // create the return message, and post it on message bus
    gst_structure_set(structure, "list", G_TYPE_STRING, properties->str, NULL);
    post_return_message(pipeline, structure);
    g_string_free(properties, TRUE);
}

/*
 * property_query:
 * @pipeline: pipeline instance
 * @structure: structure with name %GET_PROPERTY_MSG_NAME
 *
 * Handle a %GET_PROPERTY_MSG_NAME message.
 */
static void
property_query(pipeline_t *pipeline, GstStructure *structure)
{
    // get property value
    const gchar *element_property_name =
        gst_structure_get_string(structure, "property");
    gchar **name_parts =
        g_strsplit(element_property_name, G_DIR_SEPARATOR_S, 0);
    if (g_strv_length(name_parts) == 2) {
        GstElement *element = gst_bin_get_by_name(GST_BIN(pipeline->pipeline),
                                                  name_parts[0]);
        if (element != NULL) {
            // get the property's type, and set value field in return
            // message with value of that type
            GParamSpec *param = g_object_class_find_property(
                G_OBJECT_GET_CLASS(G_OBJECT(element)), name_parts[1]);
            if (param != NULL && ((param)->flags & G_PARAM_READABLE)) {
                switch (param->value_type) {
                  case G_TYPE_CHAR: {
                    gchar v;
                    g_object_get(G_OBJECT(element), name_parts[1], &v, NULL);
                    gst_structure_set(structure, "value", G_TYPE_CHAR, v,
                                      NULL);
                    break;
                  }
                  case G_TYPE_UCHAR: {
                    guchar v;
                    g_object_get(G_OBJECT(element), name_parts[1], &v, NULL);
                    gst_structure_set(structure, "value", G_TYPE_UCHAR, v,
                                      NULL);
                    break;
                  }
                  case G_TYPE_BOOLEAN: {
                    gboolean v;
                    g_object_get(G_OBJECT(element), name_parts[1], &v, NULL);
                    gst_structure_set(structure, "value", G_TYPE_BOOLEAN, v,
                                      NULL);
                    break;
                  }
                  case G_TYPE_INT: {
                    gint v;
                    g_object_get(G_OBJECT(element), name_parts[1], &v, NULL);
                    gst_structure_set(structure, "value", G_TYPE_INT, v,
                                      NULL);
                    break;
                  }
                  case G_TYPE_UINT: {
                    guint v;
                    g_object_get(G_OBJECT(element), name_parts[1], &v, NULL);
                    gst_structure_set(structure, "value", G_TYPE_UINT, v,
                                      NULL);
                    break;
                  }
                  case G_TYPE_LONG: {
                    glong v;
                    g_object_get(G_OBJECT(element), name_parts[1], &v, NULL);
                    gst_structure_set(structure, "value", G_TYPE_LONG, v,
                                      NULL);
                    break;
                  }
                  case G_TYPE_ULONG: {
                    gulong v;
                    g_object_get(G_OBJECT(element), name_parts[1], &v, NULL);
                    gst_structure_set(structure, "value", G_TYPE_ULONG, v,
                                      NULL);
                    break;
                  }
                  case G_TYPE_INT64: {
                    gint64 v;
                    g_object_get(G_OBJECT(element), name_parts[1], &v, NULL);
                    gst_structure_set(structure, "value", G_TYPE_INT64, v,
                                      NULL);
                    break;
                  }
                  case G_TYPE_UINT64: {
                    guint64 v;
                    g_object_get(G_OBJECT(element), name_parts[1], &v, NULL);
                    gst_structure_set(structure, "value", G_TYPE_UINT64, v,
                                      NULL);
                    break;
                  }
                  case G_TYPE_FLOAT: {
                    gfloat v;
                    g_object_get(G_OBJECT(element), name_parts[1], &v, NULL);
                    gst_structure_set(structure, "value", G_TYPE_FLOAT, v,
                                      NULL);
                    break;
                  }
                  case G_TYPE_DOUBLE: {
                    gdouble v;
                    g_object_get(G_OBJECT(element), name_parts[1], &v, NULL);
                    gst_structure_set(structure, "value", G_TYPE_DOUBLE, v,
                                      NULL);
                    break;
                  }
                  case G_TYPE_STRING: {
                    gchar *v;
                    g_object_get(G_OBJECT(element), name_parts[1], &v, NULL);
                    gst_structure_set(structure, "value", G_TYPE_STRING, v,
                                      NULL);
                    break;
                  }
                  default:
                    if (g_type_is_a(param->value_type, G_TYPE_ENUM)) {
                        gint v;
                        g_object_get(G_OBJECT(element), name_parts[1], &v,
                                     NULL);
                        gst_structure_set(structure, "value",
                                          param->value_type, v, NULL);
                    }
                    else if (g_type_is_a(param->value_type, G_TYPE_FLAGS)) {
                        guint v;
                        g_object_get(G_OBJECT(element), name_parts[1], &v,
                                     NULL);
                        gst_structure_set(structure, "value",
                                          param->value_type, v, NULL);
                    }
                    break;
                }
            }
            gst_object_unref(GST_OBJECT(element));
        }
    }
    // post the message on the message bus
    post_return_message(pipeline, structure);
}

/*
 * property_command:
 * @pipeline: pipeline instance
 * @structure: structure with name %SET_PROPERTY_MSG_NAME
 *
 * Handle a %SET_PROPERTY_MSG_NAME message.
 */
static void
property_command(pipeline_t *pipeline, GstStructure *structure)
{
    // set element property value
    const gchar *element_property_name =
        gst_structure_get_string(structure, "property");
    gchar **name_parts =
        g_strsplit(element_property_name, G_DIR_SEPARATOR_S, 0);
    if (g_strv_length(name_parts) == 2) {
        GstElement *element = gst_bin_get_by_name(GST_BIN(pipeline->pipeline),
                                                  name_parts[0]);
        if (element != NULL) {
            GParamSpec *param = g_object_class_find_property(
                G_OBJECT_GET_CLASS(G_OBJECT(element)), name_parts[1]);
            const GValue *value = gst_structure_get_value(structure, "value");
            if (param != NULL && ((param)->flags & G_PARAM_WRITABLE)
                && value != NULL) {
                // type conversion may occur in setting property value
                g_object_set_property(G_OBJECT(element), name_parts[1], value);
            }
            gst_object_unref(GST_OBJECT(element));
        }
    }
}

/*
 * pad_list_query:
 * @pipeline: pipeline instance
 * @structure: structure with name %PAD_LIST_MSG_NAME
 *
 * Handle a %PAD_LIST_MSG_NAME message.
 */
static void
pad_list_query(pipeline_t *pipeline, GstStructure *structure)
{
    // get list of element pads
    const gchar *element_name =
        gst_structure_get_string(structure, "element");
    GString *pads = g_string_new("");
    GstElement *element = gst_bin_get_by_name(GST_BIN(pipeline->pipeline),
                                              element_name);
    // generate comma-separated list of pad names
    if (element != NULL) {
        GstIterator *pads_iter = gst_element_iterate_pads(element);
        void pad_name(gpointer data, gpointer user_data) {
            GstPad *pad = GST_PAD(data);
            gchar *name = gst_pad_get_name(pad);
            const gchar *fmt = ((pads->len == 0) ? "%s" : "\t%s");
            g_string_append_printf(pads, fmt, name);
            g_free(name);
        }
        GstIteratorResult iter_result =
            gst_iterator_foreach(pads_iter, pad_name, NULL);
        gst_iterator_free(pads_iter);
        if (iter_result == GST_ITERATOR_ERROR) {
            g_string_truncate(pads, 0);
        }
        gst_object_unref(GST_OBJECT(element));
    }
    // create return message, and post it on the message bus
    gst_structure_set(structure, "list", G_TYPE_STRING, pads->str, NULL);
    post_return_message(pipeline, structure);
    g_string_free(pads, TRUE);
}

/*
 * pad_caps_query:
 * @pipeline: pipeline instance
 * @structure: structure with name %PAD_CAPS_MSG_NAME
 *
 * Handle a %PAD_CAPS_MSG_NAME message.
 */
static void
pad_caps_query(pipeline_t *pipeline, GstStructure *structure)
{
    // get pad caps
    gchar *str = NULL;
    const gchar *pad_name = gst_structure_get_string(structure, "pad");
    gchar **name_parts = g_strsplit(pad_name, G_DIR_SEPARATOR_S, 0);
    if (g_strv_length(name_parts) == 2) {
        GstElement *element =
            gst_bin_get_by_name(GST_BIN(pipeline->pipeline), name_parts[0]);
        if (element != NULL) {
            GstPad *pad = gst_element_get_static_pad(element, name_parts[1]);
            if (pad != NULL) {
                GstCaps *caps = gst_pad_get_negotiated_caps(pad);
                if (caps == NULL) {
                    caps = gst_pad_get_caps(pad);
                }
                if (caps != NULL) {
                    str = gst_caps_to_string(caps);
                    g_debug("pad caps '%s'", str);
                    gst_caps_unref(caps);
                }
                gst_object_unref(GST_OBJECT(pad));
            }
            gst_object_unref(GST_OBJECT(element));
        }
    }
    if (str == NULL) {
        str = g_strdup("");
    }
    // create return message, and post it on the message bus
    gst_structure_set(structure, "caps", G_TYPE_STRING, str, NULL);
    post_return_message(pipeline, structure);
    g_free(str);
}

/*
 * pad_peer_query:
 * @pipeline: pipeline instance
 * @structure: structure with name %PAD_PEER_MSG_NAME
 *
 * Handle a %PAD_PEER_MSG_NAME message.
 */
static void
pad_peer_query(pipeline_t *pipeline, GstStructure *structure)
{
    // get pad peer
    gchar *str = NULL;
    const gchar *pad_name = gst_structure_get_string(structure, "pad");
    gchar **name_parts = g_strsplit(pad_name, G_DIR_SEPARATOR_S, 0);
    if (g_strv_length(name_parts) == 2) {
        GstElement *element =
            gst_bin_get_by_name(GST_BIN(pipeline->pipeline), name_parts[0]);
        if (element != NULL) {
            GstPad *pad = gst_element_get_static_pad(element, name_parts[1]);
            if (pad != NULL && gst_pad_is_linked(pad)) {
                GstPad *peer = gst_pad_get_peer(pad);
                g_assert(peer != NULL);
                GstObject *parent = gst_pad_get_parent(peer);
                g_assert(parent != NULL);
                str = g_strdup_printf("%s/%s", GST_OBJECT_NAME(parent),
                                      GST_OBJECT_NAME(peer));
                gst_object_unref(parent);
                gst_object_unref(GST_OBJECT(peer));
                gst_object_unref(GST_OBJECT(pad));
            }
            gst_object_unref(GST_OBJECT(element));
        }
    }
    if (str == NULL) {
        str = g_strdup("");
    }
    // create return message and post it on message bus
    gst_structure_set(structure, "peer", G_TYPE_STRING, str, NULL);
    post_return_message(pipeline, structure);
    g_free(str);
}

/*
 * post_return_message:
 * @pipeline: pipeline instance
 * @structure: return message contents
 *
 * Create and application message with the given @structure, and post
 * it on the message bus.
 */
static void
post_return_message(pipeline_t *pipeline, GstStructure *structure)
{
    GstBus *bus = gst_pipeline_get_bus(GST_PIPELINE(pipeline->pipeline));
    GstMessage *retmsg = gst_message_new_application(NULL, structure);
    g_assert(retmsg != NULL);
    gst_bus_post(bus, retmsg);
    gst_object_unref(GST_OBJECT(bus));
}

static gpointer
get_config_cb(GstElement *element, gpointer data)
{
    pipeline_t *pipeline = data;
    g_debug("get_config_cb %s -> %p", GST_ELEMENT_NAME(element),
            pipeline->config);
    return pipeline->config;
}

static gpointer
get_pipeline_name_cb(GstElement *element, gpointer data)
{
    pipeline_t *pipeline = data;
    g_debug("get_pipeline_name_cb %s -> %p", GST_ELEMENT_NAME(element),
            pipeline->pipeline_name);
    return pipeline->pipeline_name;
}

static gpointer
get_wcbe_name_cb(GstElement *element, gpointer data)
{
    pipeline_t *pipeline = data;
    g_debug("get_wcbe_name_cb %s -> %p", GST_ELEMENT_NAME(element),
            pipeline->wcbe_name);
    return pipeline->wcbe_name;
}

static guint
get_be_integration_cb(GstElement *element, gpointer data)
{
    pipeline_t *pipeline = data;
    g_debug("get_be_integration_cb %s -> %u", GST_ELEMENT_NAME(element),
            pipeline->integration);
    return pipeline->integration;
}

static void
register_tmpfile_cb(GstElement *element, gchar *filename, gpointer data)
{
    g_debug("register_tmpfile_cb: %s, %s",
            GST_ELEMENT_NAME(element), filename);
    pipeline_t *pipeline = data;
    if (filename != NULL) {
        g_static_mutex_lock(&(pipeline->mtx));
        pipeline->tmpfiles =
            g_slist_prepend(pipeline->tmpfiles, g_strdup(filename));
        g_static_mutex_unlock(&(pipeline->mtx));
    }
}

static void
unregister_tmpfile_cb(GstElement *element, gchar *filename, gpointer data)
{
    g_debug("unregister_tmpfile_cb: %s, %s",
            GST_ELEMENT_NAME(element), filename);
    pipeline_t *pipeline = data;
    if (filename != NULL) {
        gint cmp_names(gconstpointer a, gconstpointer b) {
            const gchar *stra = a;
            const gchar *strb = b;
            return strcmp(stra, strb);
        }
        g_static_mutex_lock(&(pipeline->mtx));
        GSList *tmpfile =
            g_slist_find_custom(pipeline->tmpfiles, filename, cmp_names);
        if (tmpfile != NULL) {
            g_free((gchar *)(tmpfile->data));
            pipeline->tmpfiles =
                g_slist_delete_link(pipeline->tmpfiles, tmpfile);
        }
        g_static_mutex_unlock(&(pipeline->mtx));
    }
}

static gchar *
get_bdfdir_cb(GstElement *element, gpointer data)
{
    pipeline_t *pipeline = data;
    return pipeline->output_dir;
}
