#! /bin/bash
#
# start_subscan
#
# Send a "new subscan" event to WCBE pipeline instances for given
# configuration, start time, scan and subscan numbers.
#
if [ -f /etc/sysconfig/wcbe.conf ]; then
    . /etc/sysconfig/wcbe.conf
fi
if [ -f $WCBE_PATH/share/wcbe_common.sh ]; then
    . $WCBE_PATH/share/wcbe_common.sh
elif [ -f $DEFAULT_WCBE_PATH/share/wcbe_common.sh ]; then
    . $DEFAULT_WCBE_PATH/share/wcbe_common.sh
fi
set_environment

shopt -s nullglob

declare -r USE_XATTRS=yes

# print usage message
function usage {
    echo "Usage: $0 [-c config] [-s subarray] time scan [subscan]" >&2
}

# verify pipelinefs mount point
check_pipelinefs
if [ $? -ne 0 ]; then
    echo "Failed to locate pipelinefs mount point" >&2
    exit -1
fi

# verify BDF root directory
check_bdfdir
if [ $? -ne 0 ]; then
    echo "Failed to locate BDF root directory" >&2
    exit -1
fi

# get command line options
while [ 1 ]; do
    case $1 in
        -c)
            shift
            CONFIG=$1
            ;;
        -s)
            shift
            SUBARRAY=$1
            ;;
        -*)
            usage
            exit -1
            ;;
        *)
            break
            ;;
    esac
    shift
done

# check number of arguments
if [ $# -lt 2 -o $# -gt 3 ]; then
    usage
    exit -1
fi

# Get the array of instances to control
pipelinefs_glob '*'
if [ ${#INSTANCES} -eq 0 ]; then
    echo "No controllable pipeline instances were found."
    echo "Verify $WCBE_LFX_FILE for node list, and check in"
    echo "$WCBE_PIPELINEFS for instances."
    exit -1
fi

# When the config name is omitted, the config we use is the active
# config, whose name we get from the first instance in $INSTANCES.
if [ -z $CONFIG ]; then
    CFGDIR=$( readlink $( ls -1d $WCBE_PIPELINEFS/${INSTANCES[0]}/active ) )
    CONFIG=${CFGDIR#configurations/}
    if [ -z $CONFIG ]; then
        echo "Failed to find current configuration under $WCBE_PIPELINEFS" >&2
        exit -1
    fi
fi
CFGDIR=$( ls -1d $WCBE_PIPELINEFS/${INSTANCES[0]}/configurations/$CONFIG )
if [ -z $CFGDIR ]; then
    echo "Failed to find configuration $CONFIG under $WCBE_PIPELINEFS" >&2
    exit -1
fi

# when the subarray id is omitted, we use a stupid default value of
# "ptc"
if [ -z $SUBARRAY ]; then
    SUBARRAY=ptc
fi

# parse subscan start time
parse_start $1
if [ $? -ne 0 ]; then
    echo "Failed to determine start time from command line arguments" >&2
    exit -1
fi
shift

# subscan number default value is one
SCAN=$1
SUBSCAN=1
if [ $# -eq 2 ]; then
    SUBSCAN=$2
fi

# Generate a uid for the subscan using the current time in
# milliseconds since the (Unix) epoch.
TIME=$(date +%s.%N)
declare -i SEC=${TIME%%.[0-9]*}
FRC=${TIME##[0-9]*.}
declare -i NS=${FRC##0*}
MS=$(( 1000 * $SEC + $NS / 1000000 ))
BDFUID=uid://evla/bdf/$MS
BDF_FILENAME=$( echo $BDFUID | tr :/ _ )

# generate project path string
PROJECT_PATH=0/$SCAN/$SUBSCAN/

# make directory for BDF file creation
PRIV_BDFDIR=$WCBE_BDFDIR/.private/.$$-$BDF_FILENAME
mkdir --mode=2775 $PRIV_BDFDIR
if [ $? -ne 0 ]; then
    echo "Failed to create directory $PRIV_BDFDIR for BDF file creation" >&2
    exit -1
fi

function cleanup_priv_bdfdir {
    if [ -d $PRIV_BDFDIR ]; then
        rm -rf $PRIV_BDFDIR
    fi
}

# initialize the BDF
bdfinit $CFGDIR/specification $BDFUID $PROJECT_PATH $START \
    > $PRIV_BDFDIR/$BDF_FILENAME
if [ $? -ne 0 ]; then
    echo "Failed to initialize BDF file in $PRIV_BDFDIR/$BDF_FILENAME" >&2
    cleanup_priv_bdfdir
    exit -1
fi
chmod 0664 $PRIV_BDFDIR/$BDF_FILENAME
# record some BDF file metadata
if [ $USE_XATTRS == yes ]; then
    # record metadata using extended file system attributes
    setfattr -n user.edu.nrao.widar.uid -v $BDFUID \
        $PRIV_BDFDIR/$BDF_FILENAME
    setfattr -n user.edu.nrao.widar.subarray -v $SUBARRAY \
        $PRIV_BDFDIR/$BDF_FILENAME
    setfattr -n user.edu.nrao.widar.scan -v $SCAN \
        $PRIV_BDFDIR/$BDF_FILENAME
    setfattr -n user.edu.nrao.widar.subscan -v $SUBSCAN \
        $PRIV_BDFDIR/$BDF_FILENAME
    setfattr -n user.edu.nrao.widar.start -v $START \
        $PRIV_BDFDIR/$BDF_FILENAME
    setfattr -n user.edu.nrao.widar.integrations -v 0 \
        $PRIV_BDFDIR/$BDF_FILENAME
    setfattr -n user.edu.nrao.widar.config -v $CONFIG \
        $PRIV_BDFDIR/$BDF_FILENAME
else
    # record metadata using auxiliary file
    printf "$BDF_FILENAME\t$BDFUID\t$SUBARRAY\t$SCAN\t$SUBSCAN\t$START\t$CONFIG\n" \
        >> $WCBE_BDFDIR/.private/.meta
fi

# leave a BDF creation timestamp file (acts as an "in-use" signal)
touch $PRIV_BDFDIR/.stamp
chmod 0664 $PRIV_BDFDIR/.stamp
if [ $? -ne 0 ]; then
    echo "Failed to create timestamp file for BDF"
    cleanup_priv_bdfdir
    exit -1
fi

# unhide the BDF directory
NEW_PRIV_BDFDIR=$WCBE_BDFDIR/.private/$BDF_FILENAME
mv $PRIV_BDFDIR $NEW_PRIV_BDFDIR
if [ $? -ne 0 ]; then
    echo "Failed to unhide BDF creation directory $PRIV_BDFDIR" >&2
    cleanup_priv_bdfdir
    exit -1
fi

# generate "new_subscan" event string
NEWSUBSCAN_EVENT=0:$START:0-$CONFIG-$BDFUID

# send "new_subscan" event to controlled instances
for instance in ${INSTANCES[@]}; do
    echo $NEWSUBSCAN_EVENT > $WCBE_PIPELINEFS/$instance/new_subscan
done
