#! /bin/bash
#
# stop_subscan
#
# Send a "new subscan" event intended to stop a subscan to WCBE
# pipeline instances for given configuration and stop time.
#
if [ -f /etc/sysconfig/wcbe.conf ]; then
    . /etc/sysconfig/wcbe.conf
fi

if [ -f $WCBE_PATH/share/wcbe_common.sh ]; then
    . $WCBE_PATH/share/wcbe_common.sh
elif [ -f $DEFAULT_WCBE_PATH/share/wcbe_common.sh ]; then
    . $DEFAULT_WCBE_PATH/share/wcbe_common.sh
fi
set_environment

shopt -s nullglob

declare -r USE_XATTRS=yes

# print usage message
function usage {
    echo "Usage: $0 [-c config] [-s subarray] [time] " >&2
}

# verify pipelinefs mount point
check_pipelinefs
if [ $? -ne 0 ]; then
    echo "Failed to locate pipelinefs mount point" >&2
    exit -1
fi

# verify BDF root directory
check_bdfdir
if [ $? -ne 0 ]; then
    echo "Failed to locate BDF root directory" >&2
    exit -1
fi

# get command line options
while [ 1 ]; do
    case $1 in
        -c)
            shift
            CONFIG=$1
            ;;
        -s)
            shift
            SUBARRAY=$1
            ;;
        -*)
            usage
            exit -1
            ;;
        *)
            break
            ;;
    esac
    shift
done

# check number of arguments
if [ $# -gt 1 ]; then
    usage
    exit -1
fi

# parse subscan "start" time (i.e, stop time in this context)
parse_start $1
if [ $? -ne 0 ]; then
    echo "Failed to determine stop time from command line arguments" >&2
    exit -1
fi

# Get the array of instances to control
pipelinefs_glob '*'
if [ ${#INSTANCES} -eq 0 ]; then
    echo "No controllable pipeline instances were found."
    echo "Verify $WCBE_LFX_FILE for node list, and check in"
    echo "$WCBE_PIPELINEFS for instances."
    exit -1
fi

# get a configuration name
if [ -z $CONFIG ]; then
    # When the config name is omitted, the config we use is the active
    # config, whose name we get from the first node under
    # $WCBE_PIPELINEFS.
    CFGDIR=$( readlink $( ls -1d $WCBE_PIPELINEFS/${INSTANCES[0]}/active ) )
    CONFIG=${CFGDIR#configurations/}
    if [ -z $CONFIG ]; then
        echo "Failed to find current configuration under $WCBE_PIPELINEFS" >&2
        exit -1
    fi
else
    CFGDIR=$( ls -1 $WCBE_PIPELINEFS/${INSTANCES[0]}/configurations/$CONFIG )
    if [ -z $CFGDIR ]; then
        echo "Failed to find configuration $CONFIG under $WCBE_PIPELINEFS" >&2
        exit -1
    fi
fi

# First, we send a "new subscan" signal with a null configuration, so
# that the event occurs with the proper time. After this, we will
# cancel all subscans starting at a time later that this "new subscan"
# event.

# generate "new_subscan" event string
NEWSUBSCAN_EVENT=0:$START:0-$CONFIG-null

# send "new_subscan" event to controlled instances
for instance in ${INSTANCES[@]}; do
    echo $NEWSUBSCAN_EVENT > $WCBE_PIPELINEFS/$instance/new_subscan
done

# Gather up uids for all the currently known subscans under the
# configuration. The list of uids is sorted in reverse time order,
# which is the order in which subscans will be cancelled.
if [ $USE_XATTRS == yes ]; then
    TMPFILE=$( mktemp )
    for f in $WCBE_BDFDIR/.private/uid*/uid*; do
        if [[ $( getfattr --only-values -n user.edu.nrao.widar.config \
                 $f 2> /dev/null ) == $CONFIG \
              && ( -z $SUBARRAY \
                   || $( getfattr --only-values -n \
                         user.edu.nrao.widar.subarray \
                         $f 2> /dev/null ) == $SUBARRAY ) ]]; then
            TIME=$( getfattr --only-values -n user.edu.nrao.widar.start \
                    $f 2> /dev/null )
            BDFUID=$( getfattr --only-values -n user.edu.nrao.widar.uid \
                      $f 2> /dev/null )
            printf "$BDFUID\t$TIME\n" >> $TMPFILE
        fi
    done
    CANCEL_UIDS=$( sort --reverse --numeric --key=1 $TMPFILE )
    rm -f $TMPFILE
else
    IN_CONFIG=$( grep $SUBARRAY.*$CONFIG\$ $WCBE_BDFDIR/.private/.meta | \
                 sort --reverse --numeric --key=6 )
    CANCEL_UIDS=$( echo "$IN_CONFIG" | cut -f 2,6 )
fi
# now cancel each subscan in the list
echo CANCEL_UIDS: "$CANCEL_UIDS"
for instance in ${INSTANCES[@]}; do
    echo "$CANCEL_UIDS" | awk "{if (\$2 > $START) {print \$1; system(\"echo $CONFIG-\"\$1\" >| $WCBE_PIPELINEFS/$instance/cancel_subscan\")}}"
done
