// term_source.c
//
// Termination signal handler for glib.
//
/**
 * SECTION: term_source
 * @short_description: Termination signal handler for glib.
 * @stability: unstable
 * @include: term_source.h
 *
 * A #GSource event source that catches termination signals, and makes
 * them available in the main loop.
 */
#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <term_source.h>

#include <signal.h>
#include <time.h>

static gint last_signal = 0;
static struct timespec sig_timestamp = {.tv_sec = 0, .tv_nsec = 0};
static gint ref_count = 0;

static GStaticMutex mutex = G_STATIC_MUTEX_INIT;

static struct sigaction orig_sigterm;
static struct sigaction orig_sighup;
static struct sigaction orig_sigint;
static struct sigaction term_sigaction;

/**
 * term_handler:
 * @sig: signal
 *
 * Signal handler. Records the signal and time.
 */
static void
term_handler(gint sig)
{
    last_signal = sig;
    clock_gettime(CLOCK_MONOTONIC, &sig_timestamp);
}

/**
 * prepare:
 * @source: #TermSource event source
 * @timeout: returns poll timeout value (always 0)
 *
 * Check for a new signal that was caught.
 *
 * Returns: %TRUE, if a new signal was caught (by term_handler());
 *          %FALSE, otherwise.
 */
static gboolean
prepare(GSource *source, gint *timeout)
{
    gboolean result;
    TermSource *tsource = (TermSource *)source;
    if (sig_timestamp.tv_sec != tsource->sig_timestamp.tv_sec
        || sig_timestamp.tv_nsec != tsource->sig_timestamp.tv_nsec) {
        result = TRUE;
        *timeout = 0;
    }
    else {
        result = FALSE;
    }
    return result;
}

/**
 * check:
 * @source: #TermSource event source
 *
 * Check for a new signal that was caught.
 */
static gboolean
check(GSource *source)
{
    gint t;
    return prepare(source, &t);
}

/**
 * dispatch:
 * @source: #TermSource event source
 * @callback: callback function (may be NULL)
 * @user_data: unused
 *
 * Call callback with last caught signal.
 *
 * Returns: Result of callback; or %TRUE, if @callback is NULL.
 */
static gboolean
dispatch(GSource *source, GSourceFunc callback, gpointer user_data)
{
    gboolean result;
    TermSource *tsource = (TermSource *)source;
    tsource->sig_timestamp = sig_timestamp;
    if (callback != NULL) {
        gint sig = last_signal;
        result = callback(&sig);
    }
    else {
        result = TRUE;
    }
    return result;
}

/**
 * finalize:
 * @source: #TermSource event source
 *
 * Clean up module resources used by @source (but not @source itself,
 * use term_source_unref() for that).
 */
static void
finalize(GSource *source)
{
    g_static_mutex_lock(&mutex);
    --ref_count;
    if (ref_count == 0) {
        sigaction(SIGTERM, &orig_sigterm, NULL);
        sigaction(SIGHUP, &orig_sighup, NULL);
        sigaction(SIGINT, &orig_sigint, NULL);
    }
    g_static_mutex_unlock(&mutex);
}

static GSourceFuncs term_funcs = {
    .prepare = prepare,
    .check = check,
    .dispatch = dispatch,
    .finalize = finalize
};

/**
 * term_source_new:
 *
 * Create a new #TermSource instance. No need to ref it, but it should
 * be unref'ed after use.
 */
TermSource *
term_source_new(void)
{
    g_static_mutex_lock(&mutex);
    if (ref_count == 0) {
        term_sigaction.sa_handler = term_handler;
        sigaction(SIGTERM, &term_sigaction, &orig_sigterm);
        sigaction(SIGHUP, &term_sigaction, &orig_sighup);
        sigaction(SIGINT, &term_sigaction, &orig_sigint);
    }
    ++ref_count;
    g_static_mutex_unlock(&mutex);
    TermSource *result =
        (TermSource *)g_source_new(&term_funcs, sizeof(TermSource));
    result->sig_timestamp = sig_timestamp;
    return result;
}

/**
 * term_source_set_callback:
 * @source: #TermSource event source
 * @func: callback
 *
 * Register a callback for the event source.
 */
void
term_source_set_callback(TermSource *source, TermSourceFunc func)
{
    g_source_set_callback((GSource *)source, (GSourceFunc)func, NULL, NULL);
}
