#! /usr/bin/env python

import os
import select
import signal
import errno
import xml.sax
import xml.sax.saxutils
import tempfile

WCBE_PIPELINEFS = '/tmp/wcbe/nodes'
WCBE_CONFIGURATOR_FIFO = '/tmp/wcbe/configurator'

class ConfigContentHandler (xml.sax.saxutils.XMLGenerator):

    def __init__(self, configurator):
        xml.sax.saxutils.XMLGenerator.__init__(self, configurator.tmp)
        self.configurator = configurator
        self.top = None
        return

    def strip_id(self, attrs):
        if 'id' in attrs:
            id = attrs['id']
            newattrs = {}
            for (k, v) in attrs.items():
                if k != 'id':
                    newattrs[k] = v
            attrs = xml.sax.xmlreader.AttributesImpl(newattrs)
        else:
            id = None
        return (id, attrs)

    def startElement(self, name, attrs):
        if self.top is None:
            (id, attrs) = self.strip_id(attrs)
            self.top = [name, id, 1]
        elif self.top[0] == name:
            self.top[2] += 1
        xml.sax.saxutils.XMLGenerator.startElement(self, name, attrs)
        return

    def endElement(self, name):
        xml.sax.saxutils.XMLGenerator.endElement(self, name)
        if self.top[0] == name:
            self.top[2] -= 1
            if self.top[2] == 0:
                self.configurator.new_doc(self.top[0], self.top[1])
                self.top = None
        return

    def startElementNS(self, name, qname, attrs):
        if self.top is None:
            (id, attrs) = self.strip_id(attrs)
            self.top = [name, id, 1]
        elif self.top[0] == name:
            self.top[2] += 1
        xml.sax.saxutils.XMLGenerator.startElementNS(self, name, attrs)
        return

    def endElementNS(self, name, qname):
        xml.sax.saxutils.XMLGenerator.endElementNS(self, name, qname)
        if self.top[0] == name:
            self.top[2] -= 1
            if self.top[2] == 0:
                self.configurator.new_doc(self.top[0], self.top[1])
                self.top = None
        return

class ConfigErrorHandler (xml.sax.handler.ErrorHandler):

    def __init__(self):
        return

    def error(self, exc):
        raise exc
        return

    def fatalError(self, exc):
        raise exc
        return

    def warning(self, exc):
        return

class Configurator (object):

    top_level = ['foo', 'bar']

    def __init__(self, pipelinefs, fifo):
        super(Configurator, self).__init__()
        self.tmp = None
        self.fifo_path = fifo
        self.fifo = None
        self.reader = None
        return

    def __del__(self):
        if self.fifo is not None:
            self.fifo.close()
        return

    def make_parser(self):
        self.tmp = tempfile.TemporaryFile()
        self.reader = \
            xml.sax.make_parser(['xml.sax.xmlreader.IncrementalParser'])
        self.reader.setFeature(xml.sax.handler.feature_validation, 0)
        self.reader.setContentHandler(ConfigContentHandler(self))
        self.reader.setErrorHandler(ConfigErrorHandler())
        return

    def reset_parser(self):
        self.tmp = tempfile.TemporaryFile()
        self.reader.setContentHandler(ConfigContentHandler(self))
        self.reader.reset()
        return

    def open_fifo(self, poll):
        self.fifo = file(self.fifo_path, 'r')
        poll.register(self.fifo.fileno(), select.POLLIN | select.POLLERR
                      | select.POLLHUP)
        return

    def close_fifo(self, poll):
        poll.unregister(self.fifo.fileno())
        self.fifo.close()
        self.fifo = None
        return

    def new_doc(self, name, confid):
        self.tmp.flush()
        self.tmp.seek(0)
        print confid
        for line in self.tmp:
            print line,
        print
        self.reader = None
        return

    def run(self):
        poll = select.poll()
        done = False
        fifofd = None
        while not done:
            try:
                if self.fifo is None:
                    self.open_fifo(poll)
                    fifofd = self.fifo.fileno()
                events = poll.poll()
                for (fd, event) in events:
                    assert fd == fifofd
                    if event & select.POLLIN != 0:
                        buf = self.fifo.readline()
                        while buf != '':
                            self.process_input(buf)
                            buf = self.fifo.readline()
                    if event & select.POLLERR != 0:
                        self.close_fifo(poll)
                        done = True
                    if event & select.POLLHUP != 0:
                        self.close_fifo(poll)
            except KeyboardInterrupt:
                done = True
            except select.error, (errno.EINTR, str):
                done = True
        return

    def process_input(self, buf):
        if self.reader is None:
            self.make_parser()
        try:
            self.reader.feed(buf)
        except xml.sax.SAXParseException, e:
            self.reader = None
        return

if __name__ == '__main__':
    import os.path
    import sys

    pipelinefs_ok = False
    if 'WCBE_PIPELINEFS' in os.environ:
        WCBE_PIPELINEFS = os.environ['WCBE_PIPELINEFS']
    if os.path.ismount(WCBE_PIPELINEFS):
        pipelinefs = None
        try:
            pipelinefs = open(os.path.join(WCBE_PIPELINEFS, '00README'), 'r')
            readme = pipelinefs.readline().strip()
            if readme == 'pipelinefs root directory':
                pipelinefs_ok = True
        except IOError:
            pass
        if pipelinefs is not None:
            pipelinefs.close()
    if not pipelinefs_ok:
        print "'%s' is not a mount point for pipelinefs file system" \
              % WCBE_PIPELINEFS
        sys.exit(-1)

    def cleanup():
        try:
            os.unlink(WCBE_CONFIGURATOR_FIFO)
        except:
            pass
        return
    signal.signal(signal.SIGTERM, lambda x, y: cleanup())
    if 'WCBE_CONFIGURATOR_FIFO' in os.environ:
        WCBE_CONFIGURATOR_FIFO = os.environ['WCBE_CONFIGURATOR_FIFO']
    try:
        os.mkfifo(WCBE_CONFIGURATOR_FIFO)
    except OSError:
        print "Failed to create named pipe '%s'" % WCBE_CONFIGURATOR_FIFO
        sys.exit(-1)


    try:
        configurator = Configurator(WCBE_PIPELINEFS, WCBE_CONFIGURATOR_FIFO)
        configurator.run()
    finally:
        cleanup()
