#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netdb.h>
#include <sys/un.h>
#include <sys/wait.h>
#include <unistd.h>
#include <glib.h>
#include <glib/gstdio.h>
#include <gst/gst.h>

#include <term_source.h>
#include <enums.h>
#include <wcbe_private.h>
#include <wcbe_app.h>
#include <wcbe.h>
#include <wcbe_mc.h>
#include <wcbe_framedump.h>
#include <wcbe_config.h>
#include <wcbe_enums.h>
#include <async_syslog.h>
#include <idstring.h>
#include <wcbe_util.h>
#include <user_marshal.h>

#if CONTROL_SOCK_UDP
# include <netinet/in.h>
#elif CONTROL_SOCK_SCTP
# include <netinet/in.h>
# include <netinet/sctp.h>
#elif CONTROL_SOCK_TCP
# include <netinet/in.h>
# include <netinet/tcp.h>
#endif

#define CONFIG_TIMEOUT_MS 5000
#define LS_MSG_TIMEOUT_MS 4000
#define CLEANUP_TIMEOUT_SEC 1

#define INPUT_STAGE_SOCK_NAME "issink"

#if CONTROL_SOCK_UNIX
# define CONTROL_SOCK_NAME "ctlsock"
#endif

#define MAX_SPAWN_ATTEMPTS 3

typedef struct cb_data {
    gsize msgsize;
    request_cb_fn_t cb;
#if CONTROL_SOCK_UNIX
    struct sockaddr_un addr;
#else
    struct sockaddr_in addr;
#endif
    guint timeout;
    wcbe_msg_t *msg;
    pipeline_t *pipeline;
} cb_data_t;

static void rm_path(const gchar *path, const gchar *dir);
static void parse_options(gint argc, gchar *argv[], pipeline_t *pipeline,
                          gchar **conffile);
static gchar *parse_src(const gchar *value);
static gchar *parse_name(const gchar *value);
#if CONTROL_SOCK_UDP || CONTROL_SOCK_SCTP || CONTROL_SOCK_TCP
static gchar *parse_ctladdr(const gchar *value);
#endif
static gboolean parse_framedump(const gchar *value,
                                wcbe_dump_format_t *format);
static gboolean parse_flow(const gchar *value, wcbe_flow_t *flow);
static GIOChannel *open_unix_comm_channel(const gchar *sockdir,
                                          const gchar *name);
#if CONTROL_SOCK_UDP
static GIOChannel *open_udp_comm_channel(const gchar *ctladdr);
#elif CONTROL_SOCK_SCTP
static GIOChannel *open_sctp_comm_channel(const gchar *ctladdr);
#elif CONTROL_SOCK_TCP
static GIOChannel *open_tcp_comm_channel(const gchar *ctladdr);
#endif
static gint open_sock(const gchar *nodename, const gchar *servname,
                      gint family, gint socktype, gint protocol);
static GIOChannel *new_channel(gint fd);
static gboolean bus_cb(GstBus *bus, GstMessage *msg, gpointer data);
static void handle_pipeline_state(pipeline_t *pipeline,
                                  GstStructure *structure);
static gboolean handle_comm(GIOChannel *channel, GIOCondition condition,
                            gpointer data);
static gboolean handle_child_comm(GIOChannel *source, GIOCondition condition,
                                  gpointer data);
#if CONTROL_SOCK_TCP
static gboolean handle_connect(GIOChannel *channel, GIOCondition condition,
                               gpointer data);
#endif
static GstStateChangeReturn change_pipeline_state(GstElement *pipeline,
                                                  gboolean in_main_loop,
                                                  GstState state,
                                                  GstClockTime timeout);
static void destroy_configuration(gpointer data);
static gboolean spawn_ls_pipeline(const gchar *cfg_name,
                                  const wcbe_pipeline_config_t *pipeline,
                                  component_pipeline_t *cp,
                                  const gchar *bdfdir, guint integration);
static void create_ls_pipelines(pipeline_t *pipeline, configuration_t *config);
static void pipeline_exit_notify(GPid pid, gint status, gpointer data);
static gboolean pipeline_term(gpointer data);
static gboolean pipeline_kill(gpointer data);
static void free_component_pipeline(component_pipeline_t *lsp);
static void config_deactivated_handler(GstElement *element,
                                       activation_t *activation,
                                       gpointer data);
static void config_activated_handler(GstElement *element,
                                     gchar *name, gboolean result,
                                     gpointer data);
static void pipeline_eos_handler(GstElement *element, const gchar *lsp_name,
                                 gpointer data);
static gboolean cb_timeout(gpointer data);
static void insert_xml_ids(wcbe_config_t *config);
static gboolean cleanup_old_pipeline(gpointer data);
static gboolean activation_configuration_cleanup(gpointer data);
static void signal_destroy_config(configuration_t *config,
                                  GstPipeline* pipeline);
static gboolean signal_create_config(configuration_t *config,
                                     GstPipeline *pipeline);
static gboolean signal_add_activation(configuration_t *config,
                                      activation_t *activation,
                                      GstPipeline *pipeline);
static void signal_remove_activation(configuration_t *config,
                                     activation_t *activation,
                                     GstPipeline *pipeline);
static gboolean failure_accumulator(
    GSignalInvocationHint *ihint, GValue *return_accu,
    const GValue *handler_return, gpointer data);
static void fixup_octal_sequences(gchar *str);
static gboolean all_pipelines_finished(pipeline_t *pipeline);

static request_fn_t
request_handlers[WCBE_NUM_MSG_TYPES][WCBE_NUM_MSG_INTENTS] = {
    [WCBE_CONFIG_CREATE_MSG] =
    {[WCBE_COMMAND] = wcbe_mc_config_create_command,
     [WCBE_QUERY]   = NULL},

    [WCBE_ADD_ACTIVATION_MSG] =
    {[WCBE_COMMAND] = wcbe_mc_add_activation_command,
     [WCBE_QUERY]   = NULL},

    [WCBE_REMOVE_ACTIVATION_MSG] =
    {[WCBE_COMMAND] = wcbe_mc_remove_activation_command,
     [WCBE_QUERY]   = NULL},

    [WCBE_CONFIG_DESTROY_MSG] =
    {[WCBE_COMMAND] = wcbe_mc_config_destroy_command,
     [WCBE_QUERY]   = NULL},

    [WCBE_CONFIG_LIST_MSG] =
    {[WCBE_COMMAND] = NULL,
     [WCBE_QUERY]   = wcbe_mc_config_list_query},

    [WCBE_ACTIVATION_LIST_MSG] =
    {[WCBE_COMMAND] = NULL,
     [WCBE_QUERY]   = wcbe_mc_activation_list_query},

    [WCBE_SRC_MSG] =
    {[WCBE_COMMAND] = wcbe_mc_src_command,
     [WCBE_QUERY]   = wcbe_mc_src_query},

    [WCBE_TS0_MSG] =
    {[WCBE_COMMAND] = wcbe_mc_ts0_command,
     [WCBE_QUERY]   = wcbe_mc_ts0_query},

    [WCBE_FRAMEDUMP_MSG] =
    {[WCBE_COMMAND] = wcbe_mc_framedump_command,
     [WCBE_QUERY]   = wcbe_mc_framedump_query},

    [WCBE_FLOW_MSG] =
    {[WCBE_COMMAND] = wcbe_mc_flow_command,
     [WCBE_QUERY]   = wcbe_mc_flow_query},

    [WCBE_RATE_MSG] =
    {[WCBE_COMMAND] = wcbe_mc_rate_command,
     [WCBE_QUERY]   = wcbe_mc_rate_query},

    [WCBE_NAME_MSG] =
    {[WCBE_COMMAND] = NULL,
     [WCBE_QUERY]   = wcbe_mc_name_query},

    [WCBE_QUIT_MSG] =
    {[WCBE_COMMAND] = wcbe_mc_quit_command,
     [WCBE_QUERY]   = NULL},

    [WCBE_PIPELINE_LIST_MSG] =
    {[WCBE_COMMAND] = NULL,
     [WCBE_QUERY]   = wcbe_mc_pipeline_list_query},

    [WCBE_ELEMENT_LIST_MSG] =
    {[WCBE_COMMAND] = NULL,
     [WCBE_QUERY]   = wcbe_mc_element_list_query},

    [WCBE_PROPERTY_LIST_MSG] =
    {[WCBE_COMMAND] = NULL,
     [WCBE_QUERY]   = wcbe_mc_property_list_query},

    [WCBE_PROPERTY_MSG] =
    {[WCBE_COMMAND] = wcbe_mc_property_command,
     [WCBE_QUERY]   = wcbe_mc_property_query},

    [WCBE_PAD_LIST_MSG] =
    {[WCBE_COMMAND] = NULL,
     [WCBE_QUERY]   = wcbe_mc_pad_list_query},

    [WCBE_PAD_CAPS_MSG] =
    {[WCBE_COMMAND] = NULL,
     [WCBE_QUERY]   = wcbe_mc_pad_caps_query},

    [WCBE_PAD_PEER_MSG] =
    {[WCBE_COMMAND] = NULL,
     [WCBE_QUERY]   = wcbe_mc_pad_peer_query},

    [WCBE_CONFIG_DOC_MSG] =
    {[WCBE_COMMAND] = NULL,
     [WCBE_QUERY]   = wcbe_mc_config_doc_query},

    [WCBE_AUX_OUTPUT_DIR_MSG] =
    {[WCBE_COMMAND] = wcbe_mc_aux_output_dir_command,
     [WCBE_QUERY]   = wcbe_mc_aux_output_dir_query},

    [WCBE_RELEASE_MSG] =
    {[WCBE_COMMAND] = NULL,
     [WCBE_QUERY]   = NULL},

    [WCBE_NEW_SUBSCAN_MSG] =
    {[WCBE_COMMAND] = wcbe_mc_new_subscan_command,
     [WCBE_QUERY]   = NULL},

    [WCBE_CANCEL_SUBSCAN_MSG] =
    {[WCBE_COMMAND] = wcbe_mc_cancel_subscan_command,
     [WCBE_QUERY]   = NULL},

    [WCBE_BDF_OUTPUT_DIR_MSG] =
    {[WCBE_COMMAND] = wcbe_mc_bdf_output_dir_command,
     [WCBE_QUERY]   = wcbe_mc_bdf_output_dir_query},

    [WCBE_CONFIG_INTEGRATION_MSG] =
    {[WCBE_COMMAND] = NULL,
     [WCBE_QUERY]   = wcbe_mc_config_integration_query}
};

guint create_cfg_signal = 0;
guint destroy_cfg_signal = 0;
guint add_actvn_signal = 0;
guint remove_actvn_signal = 0;
guint new_subscan_signal = 0;
guint cancel_subscan_signal = 0;

gchar const input_stage_element_name[] = "input_stage";
gchar const src_element_name[] = "src";
gchar const framedump_element_name[] = "framedump";
gchar const decode_element_name[] = "decode";
gchar const sort_element_name[] = "sort";
gchar const blsdump_element_name[] = "blsdump";
gchar const sink_element_name[] = "sink";

gint
main(gint argc, char *argv[])
{
    pipeline_t pipeline =
        {.name = DEFAULT_PIPELINE_NAME,
         .src = DEFAULT_PIPELINE_SRC,
         .ts0 = DEFAULT_PIPELINE_TS0,
         .framedump = DEFAULT_PIPELINE_FRAMEDUMP,
         .flow = DEFAULT_PIPELINE_FLOW,
         .fps = DEFAULT_PIPELINE_FPS,
         .sockdir = DEFAULT_PIPELINE_SOCKDIR,
         .pidsdir = NULL,
         .auxdir = NULL,
         .bdfdir = NULL,
         .srcpath = NULL,
         .configless_pipelines = NULL,
         .msg_callbacks = DEFAULT_PIPELINE_MSG_CALLBACKS,
         .bus_watch = 0,
         .activations = NULL,
         .state = WCBE_STATE_INITIALIZING
#if CONTROL_SOCK_UDP || CONTROL_SOCK_SCTP || CONTROL_SOCK_TCP
         , .ctladdr = NULL
#endif
        };
    gchar *conffile = NULL;

    if (!g_thread_supported()) {
        g_thread_init(NULL);
    }
    enums_register();
    g_static_rec_mutex_init(&(pipeline.configs_mtx));

    async_syslog_init(1000);
    async_syslog_set_default_log_handler();

    pipeline.current.pipeline = NULL;
    pipeline.old.pipeline = NULL;
    pipeline.loop = g_main_loop_new(NULL, FALSE);
    pipeline.configs = g_hash_table_new_full(
        g_str_hash, g_str_equal, NULL, destroy_configuration);

    pipeline.msg_callbacks = g_hash_table_new(NULL, NULL);

    // Parse command line
    parse_options(argc, argv, &pipeline, &conffile);
    if (pipeline.sockdir != NULL) {
        pipeline.sockdir = g_strdup(pipeline.sockdir);
    }
    else {
        pipeline.sockdir = g_strdup(g_get_tmp_dir());
    }

    if (pipeline.auxdir == NULL) {
        pipeline.auxdir = g_get_current_dir();
    }

    if (pipeline.bdfdir == NULL) {
        pipeline.bdfdir = g_get_current_dir();
    }

    // Create socket directory
    if (pipeline.name == NULL) {
        guint i = 0;
        while (pipeline.name == NULL && i < 100) {
            gchar *trial = g_strdup_printf("cbe%u", i);
            gchar *dir = g_build_filename(pipeline.sockdir, trial, NULL);
            gint rc = mkdir(dir, S_IRWXU);
            if (rc == 0) {
                pipeline.name = trial;
                g_free(pipeline.sockdir);
                pipeline.sockdir = dir;
            }
            else {
                g_free(trial);
                g_free(dir);
            }
            ++i;
        }
    }
    else {
        GError *err = NULL;
        gchar *name =
            g_filename_from_utf8(pipeline.name, -1, NULL, NULL, &err);
        if (name != NULL) {
            gchar *dir = g_build_filename(pipeline.sockdir, name, NULL);
            gint rc = mkdir(dir, S_IRWXU);
            if (rc == 0) {
                g_free(pipeline.sockdir);
                pipeline.sockdir = dir;
            }
            else {
                g_free(pipeline.name);
                pipeline.name = NULL;
                g_free(dir);
            }
            g_free(name);
        }
        else {
            g_error("Failed to convert 'name' string to valid "
                    "directory name: %s", err->message);
        }
    }
    if (pipeline.name == NULL) {
        g_printerr("Failed to create socket directory for instance");
        exit(-1);
    }

    // create pids directory in pipeline.sockdir
    pipeline.pidsdir = g_build_filename(pipeline.sockdir, "pids", NULL);
    gint rc = mkdir(pipeline.pidsdir, S_IRWXU);
    if (rc == 0) {
        gchar *pid = g_strdup_printf("%u", getpid());
        gchar *wcbepid = g_build_filename(pipeline.pidsdir, pid, NULL);
        FILE *f = fopen(wcbepid, "w");
        if (f != NULL) {
            rc = fclose(f);
            if (rc != 0) {
                g_warning("Failed to close file %s: %s",
                          wcbepid, g_strerror(errno));
            }
        }
        else {
            g_warning("Failed to open file %s: %s",
                      wcbepid, g_strerror(errno));
        }
        g_free(wcbepid);
    }
    else {
        g_critical("Failed to create directory %s: %s",
                   pipeline.pidsdir, g_strerror(errno));
    }

    // file name for writing pipeline "src"
    pipeline.srcpath = g_build_filename(pipeline.sockdir, "src", NULL);

    // set up TermSource to catch termination signals
    gboolean finished = FALSE;
    gboolean term_handler(gint *signum) {
        finished = TRUE;
        return FALSE;
    }
    TermSource *term_source = term_source_new();
    term_source_set_callback(term_source, term_handler);
    term_source_attach(term_source, NULL);

    gboolean done(gpointer data) {
        if (finished && g_main_loop_is_running(pipeline.loop)) {
            g_main_loop_quit(pipeline.loop);
        }
        return TRUE;
    };
    g_timeout_add(100, done, NULL);

    // Set up comm channel
#if CONTROL_SOCK_TCP
    pipeline.ctlsock = NULL;
    pipeline.ctlsock_source = 0;
    pipeline.listensock = open_tcp_comm_channel(pipeline.ctladdr);
    rc = listen(g_io_channel_unix_get_fd(pipeline.listensock), 5);
    if (rc != 0) {
        g_error("Failed to open control socket: %s", g_strerror(errno));
    }
    g_io_add_watch(pipeline.listensock, G_IO_IN | G_IO_ERR, handle_connect,
                   &pipeline);
#else
# if CONTROL_SOCK_UNIX
    pipeline.ctlsock = open_unix_comm_channel(pipeline.sockdir, CONTROL_SOCK_NAME);
# elif CONTROL_SOCK_UDP
    pipeline.ctlsock = open_udp_comm_channel(pipeline.ctladdr);
# elif CONTROL_SOCK_SCTP
    pipeline.ctlsock = open_sctp_comm_channel(pipeline.ctladdr);
# endif
    if (pipeline.ctlsock == NULL) {
        g_error("Failed to open communication channel");
    }
    g_io_add_watch(pipeline.ctlsock, G_IO_IN | G_IO_ERR, handle_comm,
                   &pipeline);
#endif

    // Register signals that the pipeline emits.
    create_cfg_signal =
        g_signal_new("create-configuration",
                     GST_TYPE_PIPELINE, G_SIGNAL_RUN_LAST,
                     0, failure_accumulator, NULL,
                     g_cclosure_user_marshal_BOOLEAN__POINTER,
                     G_TYPE_BOOLEAN, 1, G_TYPE_POINTER);
    destroy_cfg_signal =
        g_signal_new("destroy-configuration",
                     GST_TYPE_PIPELINE, G_SIGNAL_RUN_FIRST,
                     0, NULL, NULL, g_cclosure_marshal_VOID__POINTER,
                     G_TYPE_NONE, 1, G_TYPE_POINTER);
    add_actvn_signal =
        g_signal_new("add-activation",
                     GST_TYPE_PIPELINE, G_SIGNAL_RUN_LAST,
                     0, failure_accumulator, NULL,
                     g_cclosure_user_marshal_BOOLEAN__POINTER,
                     G_TYPE_BOOLEAN, 1, G_TYPE_POINTER);
    remove_actvn_signal =
        g_signal_new("remove-activation",
                     GST_TYPE_PIPELINE, G_SIGNAL_RUN_FIRST,
                     0, NULL, NULL, g_cclosure_marshal_VOID__POINTER,
                     G_TYPE_NONE, 1, G_TYPE_POINTER);
    new_subscan_signal = g_signal_new(
        "new-subscan",
        GST_TYPE_PIPELINE, G_SIGNAL_RUN_FIRST,
        0, NULL, NULL,
        g_cclosure_user_marshal_VOID__STRING_STRING_UINT_UINT_UINT,
        G_TYPE_NONE, 5, G_TYPE_STRING, G_TYPE_STRING,
        G_TYPE_UINT, G_TYPE_UINT, G_TYPE_UINT);
    cancel_subscan_signal =
        g_signal_new("cancel-subscan",
                     GST_TYPE_PIPELINE, G_SIGNAL_RUN_FIRST,
                     0, NULL, NULL,
                     g_cclosure_user_marshal_VOID__STRING_STRING,
                     G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_STRING);

    // Create pipeline if possible
    if (pipeline.src != NULL) {
        wcbe_create_pipeline(&pipeline);
        if (pipeline.current.pipeline != NULL
            && pipeline.flow == WCBE_FLOW_ON) {
            wcbe_start_pipeline(&pipeline, FALSE, 0);
        }
    }

    // Set up initial configuration
    if (conffile != NULL) {
        GError *err = NULL;
        gchar *contents;
        gboolean rc = g_file_get_contents(conffile, &contents, NULL, &err);
        if (rc) {
            gsize desc_len = strlen(contents);
            gsize msgsize = sizeof(wcbe_config_create_msg_t) + desc_len;
            wcbe_config_create_msg_t *msg=  g_malloc(msgsize);
            wcbe_msg_t *cbemsg = &(msg->hdr);
            g_strlcpy(msg->name, "init", WCBE_MAX_NAME_LEN);
            msg->desc_len = desc_len;
            g_strlcpy(msg->desc, contents, desc_len);
            msg->hdr.type = WCBE_CONFIG_CREATE_MSG;
            wcbe_mc_config_create_command(&pipeline, &cbemsg, &msgsize);
            msg->hdr.type = WCBE_ADD_ACTIVATION_MSG;
            msgsize = sizeof(wcbe_add_activation_msg_t);
            wcbe_mc_add_activation_command(&pipeline, &cbemsg, &msgsize);
            g_free(contents);
        }
        else {
            g_warning("Failed to read contents of initial config file %s: %s",
                      conffile, err->message);
        }
    }

    g_timeout_add_seconds(
        CLEANUP_TIMEOUT_SEC, activation_configuration_cleanup, &pipeline);

    // Go
    if (!finished) {
        pipeline.state = WCBE_STATE_RUNNING;
        g_main_loop_run(pipeline.loop);
        g_debug("exited main loop");
    }

    // Clean up pipeline
    pipeline.state = WCBE_STATE_EXITING;
    wcbe_destroy_pipeline(&pipeline);
    cleanup_old_pipeline(&pipeline);
    void pkill(gpointer data, gpointer user_data) {
        component_pipeline_t *cp = data;
        if (cp->type == LAGSET_PIPELINE) {
            cp->state = PIPELINE_STATE_STOPPED;
            pipeline_kill(cp);
        }
    }
    g_slist_foreach(pipeline.configless_pipelines, pkill, NULL);
    g_hash_table_destroy(pipeline.msg_callbacks);
    g_hash_table_destroy(pipeline.configs);
    g_main_loop_unref(pipeline.loop);
    if (pipeline.ctlsock != NULL) {
        g_io_channel_unref(pipeline.ctlsock);
    }
#if CONTROL_SOCK_TCP
    if (pipeline.listensock != NULL) {
        g_io_channel_unref(pipeline.listensock);
    }
#endif
    rm_path(pipeline.sockdir, "");
    g_free(pipeline.sockdir);
    g_free(pipeline.pidsdir);
    g_free(pipeline.auxdir);
    g_free(pipeline.bdfdir);
    g_free(pipeline.srcpath);
#if CONTROL_SOCK_UDP || CONTROL_SOCK_SCTP || CONTROL_SOCK_TCP
    g_free(pipeline.ctladdr);
#endif
    g_static_rec_mutex_free(&(pipeline.configs_mtx));
    async_syslog_flush();
    async_syslog_quit();
    return 0;
}

static void
rm_path(const gchar *path, const gchar *dir)
{
    if (path != NULL && dir != NULL) {
        gchar *fullpath = g_build_filename(dir, path, NULL);
        if (g_file_test(fullpath, G_FILE_TEST_IS_DIR)) {
            GError *err = NULL;
            GDir *pdir = g_dir_open(fullpath, 0, &err);
            if (pdir != NULL) {
                const gchar *name = g_dir_read_name(pdir);
                while (name != NULL) {
                    rm_path(name, fullpath);
                    name = g_dir_read_name(pdir);
                }
                g_dir_close(pdir);
            }
            else {
                g_error("Failed to open directory '%s': %s",
                        fullpath, err->message);
            }
        }
        gint rc = g_remove(fullpath);
        if (rc != 0) {
            g_error("Failed to remove '%s': %s", fullpath,
                    g_strerror(errno));
        }
        g_free(fullpath);
    }
}

static void
parse_options(gint argc, gchar *argv[], pipeline_t *pipeline, gchar **conffile)
{
    gboolean src_cb(const gchar *option_name, const gchar *value,
                    gpointer data, GError **error) {
        pipeline->src = parse_src(value);
        return pipeline->src != NULL;
    }
    gboolean name_cb(const gchar *option_name, const gchar *value,
                     gpointer data, GError **error) {
        pipeline->name = parse_name(value);
        return pipeline->name != NULL;
    }
    gboolean framedump_cb(const gchar *option_name, const gchar *value,
                          gpointer data, GError **error) {
        gboolean result = parse_framedump(value, &(pipeline->framedump));
        if (!result) {
            g_set_error(error, G_OPTION_ERROR, G_OPTION_ERROR_FAILED,
                        "Failed to parse 'framedump' value; "
                        "must be unique prefix of 'binary' or 'ascii'.");
        }
        return result;
    }
    gboolean flow_cb(const char *option_name, const gchar *value,
                     gpointer data, GError **error) {
        gboolean result = parse_flow(value, &(pipeline->flow));
        if (!result) {
            g_set_error(error, G_OPTION_ERROR, G_OPTION_ERROR_FAILED,
                        "Failed to parse 'flow' value; "
                        "must be unique prefix of 'on' or 'off'.");
        }
        return result;
    }
#if CONTROL_SOCK_UDP || CONTROL_SOCK_SCTP || CONTROL_SOCK_TCP
    gboolean ctladdr_cb(const gchar *option_name, const gchar *value,
                        gpointer data, GError **error) {
        pipeline->ctladdr = parse_ctladdr(value);
        return pipeline->ctladdr != NULL;
    }
#endif
    gboolean ts0_cb(const gchar *option_name, const gchar *value,
                    gpointer data, GError **error) {
        pipeline->ts0 = wcbe_parse_ts0(value);
        return pipeline->ts0 != NULL;
    }

    const GOptionEntry entries[] = {
        {"name", 'n', G_OPTION_FLAG_OPTIONAL_ARG, G_OPTION_ARG_CALLBACK,
         name_cb, "WIDAR backend instance name", "NAME"},
        {"src", 's', G_OPTION_FLAG_OPTIONAL_ARG, G_OPTION_ARG_CALLBACK,
         src_cb, "WIDAR frame source URI", "URI"},
        {"timestamp0", 't', G_OPTION_FLAG_OPTIONAL_ARG, G_OPTION_ARG_CALLBACK,
         ts0_cb, "Time origin to use for regenerating timestamps in "
         "file-based lag frames ", "<epoch>:<seconds>:<clock cycles>"},
        {"rate", 'r', G_OPTION_FLAG_OPTIONAL_ARG, G_OPTION_ARG_INT,
         &(pipeline->fps), "File frame source frame rate (frames per second)",
         "RATE"},
        {"sockdir", 'k', G_OPTION_FLAG_OPTIONAL_ARG, G_OPTION_ARG_FILENAME,
         &(pipeline->sockdir), "Socket directory", "DIR"},
        {"framedump", 'd', G_OPTION_FLAG_OPTIONAL_ARG, G_OPTION_ARG_CALLBACK,
         framedump_cb, "Dump raw frames", "<none|binary|ascii>"},
        {"flow", 'f', G_OPTION_FLAG_OPTIONAL_ARG, G_OPTION_ARG_CALLBACK,
         flow_cb, "Data flow state", "<on|off>"},
        {"conf", 'c', G_OPTION_FLAG_OPTIONAL_ARG, G_OPTION_ARG_FILENAME,
         conffile, "Initial configuration file", "FILE"},
#if CONTROL_SOCK_UDP || CONTROL_SOCK_SCTP || CONTROL_SOCK_TCP
        {"ctladdr", 'a', 0, G_OPTION_ARG_CALLBACK,
         ctladdr_cb, "Control socket address", "IP:PORT"},
#endif
        {NULL}
    };

    GError *error = NULL;
    GOptionContext *context =
        g_option_context_new("- WIDAR correlator backend input stage");
    g_option_context_set_summary(
        context,
        "Input stage for compute nodes of WIDAR correlator "
        "backend system.");
    g_option_context_set_description(
        context,
        "Please report feature requests, bugs, etc. to\n"
        "mpokorny@nrao.edu.");
    g_option_context_set_help_enabled(context, TRUE);
    g_option_context_set_ignore_unknown_options(context, FALSE);
    g_option_context_add_main_entries(context, entries, NULL);
    g_option_context_add_group(context, gst_init_get_option_group());
    gboolean ctxt_parse =
        g_option_context_parse(context, &argc, &argv, &error);
    if (ctxt_parse) {
        if (pipeline->flow == WCBE_FLOW_ON && pipeline->src == NULL) {
            g_set_error(&error, G_OPTION_ERROR, G_OPTION_ERROR_FAILED,
                        "Invalid command line options: "
                        "'flow=on' requires a 'src' option value.");
            ctxt_parse = FALSE;
        }
    }
    if (!ctxt_parse) {
        if (error != NULL) {
            g_printerr("%s\n", error->message);
            g_error_free(error);
        }
        exit(-1);
    }
    g_option_context_free(context);
}

static gchar *
parse_src(const gchar *value)
{
    // TODO: validate 'value'
    return g_strdup(value);
}

#if CONTROL_SOCK_UDP || CONTROL_SOCK_SCTP || CONTROL_SOCK_TCP
static gchar *
parse_ctladdr(const gchar *value)
{
    // TODO: validate 'value'
    return g_strdup(value);
}
#endif

static gchar *
parse_name(const gchar *value)
{
    gchar *result = NULL;
    // We don't check for name conflicts here, since we're not ready
    // to open the named socket.
    if (strchr(value, G_DIR_SEPARATOR) == NULL) {
        result = g_strdup(value);
    }
    return result;
}

static gboolean
parse_framedump(const gchar *value, wcbe_dump_format_t *format)
{
    gboolean result;
    if (g_str_has_prefix("binary", value)) {
        *format = WCBE_DUMP_BINARY;
        result = TRUE;
    }
    else if (g_str_has_prefix("ascii", value)) {
        *format = WCBE_DUMP_ASCII;
        result = TRUE;
    }
    else if (g_str_has_prefix("none", value)) {
        *format = WCBE_DUMP_NONE;
        result = TRUE;
    }
    else {
        result = FALSE;
    }
    return result;
}

static gboolean
parse_flow(const gchar *value, wcbe_flow_t *flow)
{
    gboolean result = FALSE;
    if (value[0] == 'o') {
        if (g_str_has_prefix("on", value)) {
            *flow = WCBE_FLOW_ON;
            result = TRUE;
        }
        else if (g_str_has_prefix("off", value)) {
            *flow = WCBE_FLOW_OFF;
            result = TRUE;
        }
    }
    return result;
}

static GIOChannel *
open_unix_comm_channel(const gchar *sockdir, const gchar *name)
{
    GIOChannel *result = NULL;
    gchar *path = g_build_filename(sockdir, name, NULL);
    int sockfd = open_sock(NULL, path, PF_UNIX, SOCK_DGRAM, 0);
    if (sockfd != -1) {
        result = new_channel(sockfd);
    }
    g_free(path);
    return result;
}

#if CONTROL_SOCK_UDP
static GIOChannel *
open_udp_comm_channel(const gchar *ctladdr)
{
    GIOChannel *result = NULL;
    gchar **name_parts = g_strsplit(ctladdr, ":", 3);
    if (g_strv_length(name_parts) == 2) {
        int sockfd =
            open_sock(name_parts[0], name_parts[1], PF_INET, SOCK_DGRAM, 0);
        if (sockfd != -1) {
            result = new_channel(sockfd);
        }
    }
    else {
        g_warning("Failed to split control socket descriptor string '%s' "
                  "into address and port", ctladdr);
    }
    g_strfreev(name_parts);

    return result;
}
#elif CONTROL_SOCK_SCTP
static GIOChannel *
open_sctp_comm_channel(const gchar *ctladdr)
{
    GIOChannel *result = NULL;
    gchar **name_parts = g_strsplit(ctladdr, ":", 3);
    if (g_strv_length(name_parts) == 2) {
        int sockfd =
            open_sock(name_parts[0], name_parts[1], PF_INET, SOCK_SEQPACKET,
                      IPPROTO_SCTP);
        if (sockfd != -1) {
            int autoclose = 60;
            int rc = setsockopt(sockfd, SOL_SCTP, SCTP_AUTOCLOSE, &autoclose,
                                (socklen_t)sizeof(autoclose));
            if (rc != 0) {
                g_debug("Failed to set SCTP_AUTOCLOSE option: %s",
                        g_strerror(rc));
            }
            rc = listen(sockfd, 100);
            if (rc == 0) {
                result = new_channel(sockfd);
            }
            else {
                g_warning("Failed to listen for control socket "
                          "connections: %s",
                          g_strerror(rc));
            }
        }
    }
    else {
        g_warning("Failed to split control socket descriptor string '%s' "
                  "into address and port", ctladdr);
    }
    g_strfreev(name_parts);

    return result;
}
#elif CONTROL_SOCK_TCP
static GIOChannel *
open_tcp_comm_channel(const gchar *ctladdr)
{
    GIOChannel *result = NULL;
    gchar **name_parts = g_strsplit(ctladdr, ":", 3);
    if (g_strv_length(name_parts) == 2) {
        int sockfd =
            open_sock(name_parts[0], name_parts[1], PF_INET, SOCK_STREAM, 0);
        if (sockfd != -1) {
            result = new_channel(sockfd);
        }
    }
    else {
        g_warning("Failed to split control socket descriptor string '%s' "
                  "into address and port", ctladdr);
    }
    g_strfreev(name_parts);

    return result;
}
#endif

static gint
open_sock(const gchar *nodename, const gchar *servname,
          gint proto_family, gint socktype, gint protocol)
{
    gint result = 0;
    struct addrinfo *res = NULL;
    struct sockaddr *sockaddr = NULL;
    socklen_t socklen;
    // Initialize sockaddr
    switch (proto_family) {
      case PF_UNIX: {
        struct sockaddr_un *addr = g_new(struct sockaddr_un, 1);
        addr->sun_family = AF_UNIX;
        gint rc = g_strlcpy(addr->sun_path, servname, sizeof(addr->sun_path));
        if (rc >= sizeof(addr->sun_path)) {
            g_warning("Socket file name (%s) too long", servname);
            result = -1;
        }
        sockaddr = (struct sockaddr *)addr;
        socklen = sizeof(struct sockaddr_un);
        break;
      }
      case PF_INET: {
        struct addrinfo hints =
            {.ai_flags = AI_PASSIVE | AI_NUMERICHOST | AI_NUMERICSERV,
             .ai_family = AF_INET,
             .ai_socktype = 0,
             .ai_protocol = 0,
             // remainder are not input values but should be
             // set to zero for call to getaddrinfo
             .ai_addrlen = 0, .ai_addr = 0, .ai_canonname = 0, .ai_next = 0
            };
        gint rc = getaddrinfo(nodename, servname, &hints, &res);
        if (rc == 0) {
            sockaddr = res[0].ai_addr;
            socklen = res[0].ai_addrlen;
        }
        else {
            g_error("Failed 'getaddrinfo' call: %d", rc);
            result = -1;
        }
        break;
      }
      default:
        g_assert_not_reached();
        result = -1;
        break;
    }

    // Create and bind socket
    if (result != -1) {
        gint sockfd = socket(proto_family, socktype, protocol);
        if (sockfd != -1) {
            gint rc = bind(sockfd, sockaddr, socklen);
            if (rc != -1) {
                result = sockfd;
            }
            else {
                close(sockfd);
                g_warning("Failed to bind socket name: %s", g_strerror(errno));
            }
        }
        else {
            g_error("Failed to create socket: %s", g_strerror(errno));
        }
    }

    // Free memory resources
    if (res != NULL) {
        freeaddrinfo(res);
    }
    else if (sockaddr != NULL) {
        g_free(sockaddr);
    }

    return result;
}

static GIOChannel *
new_channel(gint fd)
{
    GIOChannel *result = g_io_channel_unix_new(fd);
    g_io_channel_set_close_on_unref(result, TRUE);
    GError *err = NULL;
    g_io_channel_set_encoding(result, NULL, &err);
    if (err != NULL) {
        g_error("Failed to set NULL encoding on communication channel: %s",
                err->message);
    }
    g_io_channel_set_buffered(result, FALSE);
    return result;
}

void
wcbe_create_pipeline(pipeline_t *pipeline)
{
    g_return_if_fail(pipeline->src != NULL);

    GstBus *bus = NULL;
    GstElement *src = NULL;
    GstElement *framedump = NULL;
    GstElement *decode = NULL;
    GstElement *sort = NULL;
    GstElement *lagsetdump = NULL;
    GstElement *sink = NULL;
    gboolean pipeline_ready = FALSE;

    // pipeline
    pipeline->current.pipeline = gst_pipeline_new(input_stage_element_name);
    if (pipeline->current.pipeline == NULL) {
        g_warning("Failed to create input stage pipeline");
        goto cleanup_and_return;
    }

    // bus
    bus = gst_pipeline_get_bus(GST_PIPELINE(pipeline->current.pipeline));
    if (bus == NULL) {
        g_warning("Failed to get pipeline message bus");
        goto cleanup_and_return;
    }
    pipeline->bus_watch = gst_bus_add_watch(bus, bus_cb, pipeline);

    // source element
    src = gst_element_factory_make("wcbe_src", src_element_name);
    if (src == NULL) {
        g_warning("Failed to create 'wcbe_src' element\n");
        goto cleanup_and_return;
    }
    if (!gst_uri_handler_set_uri(GST_URI_HANDLER(src), pipeline->src)) {
        g_warning("Failed to set pipeline source\n");
        g_free(pipeline->src);
        pipeline->src = DEFAULT_PIPELINE_SRC;
        goto cleanup_and_return;
    }
    if (pipeline->ts0 != NULL) {
        g_object_set(G_OBJECT(src), "timestamp-base", pipeline->ts0, NULL);
    }
    g_object_set(G_OBJECT(src), "fps", pipeline->fps, NULL);

    // framedump element
    framedump =
        gst_element_factory_make("wcbe_framedump", framedump_element_name);
    if (framedump == NULL) {
        g_warning("Failed to create 'wcbe_framedump' element\n");
        goto cleanup_and_return;
    }
    if (pipeline->framedump != WCBE_DUMP_NONE) {
        WcbeFramedumpFormat fdformat =
            ((pipeline->framedump == WCBE_DUMP_BINARY)
             ? WCBE_FRAMEDUMP_BINARY : WCBE_FRAMEDUMP_ASCII);
        g_object_set(G_OBJECT(framedump), "format", fdformat, NULL);
        g_object_set(G_OBJECT(framedump), "do-output", TRUE, NULL);
    }
    else {
        g_object_set(G_OBJECT(framedump), "do-output", FALSE, NULL);
    }
    g_object_set(G_OBJECT(framedump), "output-dir", pipeline->auxdir, NULL);

    // decode element
    decode = gst_element_factory_make("wcbe_decode", decode_element_name);
    if (decode == NULL) {
        g_warning("Failed to create 'wcbe_decode' element\n");
        goto cleanup_and_return;
    }

    // sort element
    sort = gst_element_factory_make("wcbe_sort", sort_element_name);
    if (sort == NULL) {
        g_warning("Failed to create 'wcbe_sort' element\n");
        goto cleanup_and_return;
    }
    g_signal_connect(G_OBJECT(sort), "config-deactivated",
                     G_CALLBACK(config_deactivated_handler), NULL);
    g_signal_connect(G_OBJECT(sort), "config-activated",
                     G_CALLBACK(config_activated_handler), NULL);

    // lagset dump element
    lagsetdump =
        gst_element_factory_make("wcbe_lagset_dump", blsdump_element_name);
    if (lagsetdump == NULL) {
        g_warning("Failed to create 'wcbe_lagset_dump' element\n");
        goto cleanup_and_return;
    }
    g_object_set(G_OBJECT(lagsetdump), "output-dir", pipeline->auxdir, NULL);

    // sink element
    sink = gst_element_factory_make("wcbe_lagset_sink", sink_element_name);
    if (sink == NULL) {
        g_warning("Failed to create 'wcbe_lagset_sink' element\n");
        goto cleanup_and_return;
    }
    g_object_set(G_OBJECT(sink), "sockdir", pipeline->sockdir, NULL);
    g_signal_connect(G_OBJECT(sink), "pipeline-eos",
                     G_CALLBACK(pipeline_eos_handler), pipeline);

    // assemble pipeline
    gst_bin_add_many(GST_BIN(pipeline->current.pipeline),
                     src, framedump, decode, sort, lagsetdump, sink, NULL);
    gboolean link_result = gst_element_link_many(
        src, framedump, decode, sort, lagsetdump, sink, NULL);
    src = framedump = decode = sort = lagsetdump = sink = NULL;
    if (!link_result) {
        g_warning("Failed to link pipeline elements\n");
        goto cleanup_and_return;
    }

    // create comm channel for child lagset pipelines
    pipeline->current.lsp_channel =
        open_unix_comm_channel(pipeline->sockdir, INPUT_STAGE_SOCK_NAME);
    if (pipeline->current.lsp_channel == NULL) {
        g_error("Failed to open communication channel for child "
                "lagset pipelines");
    }
    pipeline->current.lsp_sockfd =
        g_io_channel_unix_get_fd(pipeline->current.lsp_channel);
    pipeline->current.lsp_source =
        g_io_add_watch(pipeline->current.lsp_channel, G_IO_IN | G_IO_ERR,
                       handle_child_comm, pipeline);
    g_assert(pipeline->current.lsp_source != 0);

    component_pipeline_t *gst_pipeline = g_new0(component_pipeline_t, 1);
    gst_pipeline->type = INPUT_PIPELINE;
    gst_pipeline->state = PIPELINE_STATE_UNKNOWN;
    pipeline->configless_pipelines = g_slist_prepend(NULL, gst_pipeline);
    pipeline_ready = TRUE;

 cleanup_and_return:
    if (bus != NULL) {
        gst_object_unref(GST_OBJECT(bus));
    }
    if (sink != NULL) {
        gst_object_unref(GST_OBJECT(sink));
    }
    if (lagsetdump != NULL) {
        gst_object_unref(GST_OBJECT(lagsetdump));
    }
    if (sort != NULL) {
        gst_object_unref(GST_OBJECT(sort));
    }
    if (decode != NULL) {
        gst_object_unref(GST_OBJECT(decode));
    }
    if (framedump != NULL) {
        gst_object_unref(GST_OBJECT(framedump));
    }
    if (src != NULL) {
        gst_object_unref(GST_OBJECT(src));
    }
    if (!pipeline_ready && pipeline->current.pipeline != NULL) {
        gst_object_unref(GST_OBJECT(pipeline->current.pipeline));
        pipeline->current.pipeline = NULL;
    }
}

void
wcbe_destroy_pipeline(pipeline_t *pipeline)
{
    g_debug("wcbe_destroy_pipeline");
    pipeline->flow = DEFAULT_PIPELINE_FLOW;
    if (pipeline->src != DEFAULT_PIPELINE_SRC) {
        g_free(pipeline->src);
        pipeline->src = DEFAULT_PIPELINE_SRC;
    }
    if (pipeline->current.pipeline != NULL) {
        memcpy(&(pipeline->old), &(pipeline->current), sizeof(pipeline->current));
        pipeline->current.pipeline = NULL;

        void rm_config(gpointer key, gpointer value, gpointer user_data) {
            configuration_t *config = value;
            wcbe_destroy_config(pipeline, config);
        }
        g_static_rec_mutex_lock(&(pipeline->configs_mtx));
        g_hash_table_foreach(pipeline->configs, rm_config, NULL);
        g_static_rec_mutex_unlock(&(pipeline->configs_mtx));
        if (pipeline->old.pipeline != NULL) {
            GstState state;
            gst_element_get_state(pipeline->old.pipeline, &state, NULL, 0);
            if (state == GST_STATE_PLAYING) {
                gst_element_send_event(pipeline->old.pipeline,
                                       gst_event_new_eos());
            }
            // The following for paranoia's sake; it shouldn't be
            // necessary.
            g_timeout_add_seconds(5, cleanup_old_pipeline, pipeline);
        }
    }
}

static gboolean
cleanup_old_pipeline(gpointer data)
{
    pipeline_t *pipeline = data;
    g_debug("cleanup_old_pipeline, %u", pipeline->state);

    if (pipeline->old.pipeline != NULL) {
        g_debug("changing input pipeline state to NULL");
        change_pipeline_state(pipeline->old.pipeline,
                              g_main_loop_is_running(pipeline->loop),
                              GST_STATE_NULL, 1 * GST_SECOND);
    }

    if (pipeline->old.bus_watch > 0) {
        g_source_remove(pipeline->old.bus_watch);
        pipeline->old.bus_watch = 0;
    }
    if (pipeline->old.lsp_source > 0) {
        g_source_remove(pipeline->old.lsp_source);
        pipeline->old.lsp_source = 0;
    }
    if (pipeline->old.pipeline != NULL) {
        gst_object_unref(GST_OBJECT(pipeline->old.pipeline));
        pipeline->old.pipeline = NULL;
    }
    if (pipeline->old.lsp_channel != DEFAULT_PIPELINE_LSP_CHANNEL) {
        GError *err = NULL;
        g_io_channel_shutdown(pipeline->old.lsp_channel, TRUE, &err);
        if (err != NULL) {
            g_debug("Failed to shut down lagset pipelines IO channel: %s",
                    err->message);
        }
        g_io_channel_unref(pipeline->old.lsp_channel);
        pipeline->old.lsp_channel = DEFAULT_PIPELINE_LSP_CHANNEL;
        gchar *sockname =
            g_build_filename(pipeline->sockdir, INPUT_STAGE_SOCK_NAME, NULL);
        gint rc = g_remove(sockname);
        if (rc != 0) {
            g_warning("Failed to unlink socket directory entry '%s': %s",
                      sockname, g_strerror(errno));
        }
        g_free(sockname);
    }
    if (pipeline->state == WCBE_STATE_EXITING) {
        g_main_loop_quit(pipeline->loop);
    }
    return FALSE;
}

static gboolean
bus_cb(GstBus *bus, GstMessage *msg, gpointer data)
{
    pipeline_t *pipeline = data;

    switch (GST_MESSAGE_TYPE(msg)) {
      case GST_MESSAGE_ERROR: {
        gchar *debug = NULL;
        GError *err = NULL;
        gst_message_parse_error(msg, &err, &debug);
        g_warning("ERROR: %s\n", err->message);
        if (debug != NULL) {
            g_warning("Debug details: %s\n", debug);
            g_free(debug);
        }
        g_error_free(err);
        wcbe_destroy_pipeline(pipeline);
        break;
      }
      case GST_MESSAGE_APPLICATION: {
        const gchar *name = gst_structure_get_name(msg->structure);
        if (strcmp(name, "pipeline-state") == 0) {
            handle_pipeline_state(pipeline, msg->structure);
        }
        else if (strcmp(name, "deactivation") == 0) {
            gpointer ptr = NULL;
            gst_structure_get_pointer(msg->structure, "activation", &ptr);
            if (ptr != NULL) {
                ((activation_t *)ptr)->deactivated = TRUE;
            }
        }
        else if (strcmp(name, ELEMENT_LIST_MSG_NAME) == 0
                 || strcmp(name, PROPERTY_LIST_MSG_NAME) == 0
                 || strcmp(name, GET_PROPERTY_MSG_NAME) == 0
                 || strcmp(name, SET_PROPERTY_MSG_NAME) == 0
                 || strcmp(name, PAD_LIST_MSG_NAME) == 0
                 || strcmp(name, PAD_CAPS_MSG_NAME) == 0
                 || strcmp(name, PAD_PEER_MSG_NAME) == 0) {
            const gchar *id =
                gst_structure_get_string(msg->structure, "id");
            wcbe_msg_t *wmsg;
            sscanf(id, "%p", &wmsg);
            cb_data_t *cbdata =
                g_hash_table_lookup(pipeline->msg_callbacks, wmsg);
            if (cbdata != NULL) {
                g_source_remove(cbdata->timeout);
                g_hash_table_remove(pipeline->msg_callbacks, wmsg);
                cbdata->cb(pipeline, &wmsg, &(cbdata->msgsize),
                           msg->structure);
                struct iovec iovec[] =
                    {{.iov_base = wmsg, .iov_len = cbdata->msgsize}};
                struct msghdr msghdr = {
                    .msg_name = &(cbdata->addr),
                    .msg_namelen = sizeof(cbdata->addr),
                    .msg_control = NULL, .msg_controllen = 0,
                    .msg_iov = iovec, .msg_iovlen = 1};
                wmsg->direction = WCBE_RESPONSE;
                gssize num_bytes = sendmsg(
                    g_io_channel_unix_get_fd(pipeline->ctlsock), &msghdr, 0);
                if (num_bytes < (gssize)iovec[0].iov_len) {
                    g_warning("Input comm channel error on write: %s",
                              g_strerror(errno));
                }
                g_free(wmsg);
                g_free(cbdata);
            }
        }
        break;
      }
      case GST_MESSAGE_STATE_CHANGED: {
        gint newstate;
        if (gst_structure_get_enum(msg->structure, "new-state",
                                   GST_TYPE_STATE, &newstate)) {
	    if (strcmp(GST_OBJECT_NAME(GST_MESSAGE_SRC(msg)),
		       input_stage_element_name) == 0) {
	        if (newstate == GST_STATE_PLAYING) {
		    pipeline->flow = WCBE_FLOW_ON;
		}
		else {
		    pipeline->flow = WCBE_FLOW_OFF;
		}
	    }
            g_debug("Element '%s' changed state to '%s'",
                    GST_OBJECT_NAME(GST_MESSAGE_SRC(msg)),
                    gst_element_state_get_name(newstate));
        }
        else {
            g_warning("Failed to read 'newstate' field "
                      "from GST_MESSAGE_STATE_CHANGED message");
        }
        break;
      }
      case GST_MESSAGE_EOS: {
        g_debug("wcbe received EOS");
        component_pipeline_t *input_stage = wcbe_find_component_pipeline(
            pipeline->configless_pipelines, input_stage_pipeline_name);
        input_stage->destroy = TRUE;
        if (all_pipelines_finished(pipeline)) {
            pipeline->state = WCBE_STATE_EXITING;
            cleanup_old_pipeline(pipeline);
        }
        break;
      }
      default:
        break;
    }
    return TRUE;
}

component_pipeline_t *
wcbe_find_component_pipeline(GSList *list, const gchar *name)
{
    component_pipeline_t *result = NULL;
    while (result == NULL && list != NULL) {
        component_pipeline_t *cp = list->data;
        if (cp->type == LAGSET_PIPELINE) {
            if (strcmp(cp->lsp.name, name) == 0) {
                result = cp;
            }
        }
        else {
            if (strcmp(name, input_stage_pipeline_name) == 0) {
                result = cp;
            }
        }
        list = g_slist_next(list);
    }
    return result;
}

gchar *
wcbe_parse_ts0(const gchar *value)
{
    gchar *result = NULL;
    gchar **parts = g_strsplit(value, ":", 0);
    if (g_strv_length(parts) == 3) {
        char *endptr = NULL;
        errno = 0;
        gulong epoch = strtoul(parts[0], &endptr, 0);
        if (errno == 0 && *endptr == '\0'
            && epoch < (1uL << 3)) {
            gulong seconds G_GNUC_UNUSED = strtoul(parts[1], &endptr, 0);
            if (errno == 0 && *endptr == '\0'
#if SIZEOF_UNSIGNED_LONG > 4
                && seconds < (1uL << 32)
#endif
                ) {
                gulong cycles = strtoul(parts[2], &endptr, 0);
                if (errno == 0 && *endptr == '\0'
                    && cycles < WIDAR_CLOCK_RATE) {
                    result = g_strdup(value);
                }
            }
        }
    }
    g_strfreev(parts);
    return result;
}

static void
handle_pipeline_state(pipeline_t *pipeline, GstStructure *structure)
{
    const gchar *src = gst_structure_get_string(structure, "src");
    if (src != NULL) {
        gchar **src_parts = g_strsplit(src, G_DIR_SEPARATOR_S, 0);
        gchar *cfg_name = src_parts[0];
        if (cfg_name != NULL) {
            g_static_rec_mutex_lock(&(pipeline->configs_mtx));
            configuration_t *config =
                g_hash_table_lookup(pipeline->configs, cfg_name);
            g_static_rec_mutex_unlock(&(pipeline->configs_mtx));
            component_pipeline_t *cp = NULL;
            if (config != NULL) {
                cp = wcbe_find_component_pipeline(config->ls_pipelines, src);
            }
            if (cp == NULL) {
                cp = wcbe_find_component_pipeline(
                    pipeline->configless_pipelines, src);
            }
            if (cp != NULL) {
                ls_pipeline_t *lsp = &(cp->lsp);
                gint state_int;
                if (gst_structure_get_enum(structure,
                                           "value",
                                           PIPELINE_STATE_ENUM_TYPE,
                                           &state_int)) {
                    g_debug("got 'pipeline-state' message '%s'",
                            g_enum_get_value(
                                g_type_class_peek_static(
                                    PIPELINE_STATE_ENUM_TYPE),
                                state_int)->value_nick);
                    cp->state = state_int;
                    switch (cp->state) {
                      case PIPELINE_STATE_STOPPED:
                        if (lsp->exit_timeout != 0){
                            g_source_remove(lsp->exit_timeout);
                        }
                        lsp->exit_timeout =
                            g_timeout_add(5000, pipeline_kill, cp);
                        if (lsp->config != NULL
                            && lsp->config->state == CONFIG_STATE_READY) {
                            lsp->config->state = CONFIG_STATE_NOT_READY;
                        }
                        break;
                      case PIPELINE_STATE_RUNNING:
                        if (lsp->config != NULL) {
                            gboolean ready = TRUE;
                            GSList *item = lsp->config->ls_pipelines;
                            g_assert(item != NULL);
                            while (item != NULL && ready) {
                                ready = (((component_pipeline_t *)
                                          (item->data))->state
                                         == PIPELINE_STATE_RUNNING);
                                item = g_slist_next(item);
                            }
                            if (ready) {
                                lsp->config->state = CONFIG_STATE_READY;
                            }
                        }
                        break;
                      default:
                        g_assert_not_reached();
                        break;
                    }
                }
                else {
                    g_debug("Received 'pipeline-state' message without "
                            "'value' field");
                }
            }
            else {
                gint state_int;
                if (gst_structure_get_enum(structure,
                                           "value",
                                           PIPELINE_STATE_ENUM_TYPE,
                                           &state_int)) {
                    g_debug("Received 'pipeline-state:value=%s' message from "
                            "unexpected src '%s'",
                            g_enum_get_value(
                                g_type_class_peek_static(
                                    PIPELINE_STATE_ENUM_TYPE),
                                state_int)->value_nick, src);
                }
                else {
                    g_debug("Received 'pipeline-state' message from "
                            "unexpected src '%s'", src);
                }
            }
        }
        else {
            g_debug("Received 'pipeline-state' message with invalid "
                    "'src' value");
        }
        g_strfreev(src_parts);
    }
    else {
        g_debug("Received 'pipeline-state' message without 'src' field");
    }
}

/* handle_comm
 *
 * Read a datagram from the input comm socket.
 *
 * We are using a GIOChannel simply as an easy means to integrate the
 * socket polling into the main loop. We definitely do not read the
 * datagram via the GIOChannel because it simply reads bytes, not
 * datagrams. Reading bytes could be a problem if something unexpected
 * arrived on the socket---perhaps unlikely, but it's safer this way.
 */
static gboolean
handle_comm(GIOChannel *channel, GIOCondition condition, gpointer data)
{
    pipeline_t *pipeline = data;
    if (condition & G_IO_IN) {
        gssize num_bytes;
        wcbe_msg_t in_msg;
#if CONTROL_SOCK_UNIX
        struct sockaddr_un addr;
#else
        struct sockaddr_in addr;
#endif
        struct iovec iovec[] =
            {{.iov_base = &in_msg, .iov_len = sizeof(in_msg)}};
        struct msghdr msghdr = {
            .msg_name = &addr, .msg_namelen = sizeof(addr),
            .msg_control = NULL, .msg_controllen = 0,
            .msg_iov = iovec, .msg_iovlen = 1};
        gint sockfd = g_io_channel_unix_get_fd(channel);
        num_bytes = recvmsg(sockfd, &msghdr, MSG_PEEK);
        if (num_bytes == sizeof(in_msg)
            && in_msg.direction == WCBE_REQUEST
            && in_msg.type < WCBE_NUM_MSG_TYPES
            && in_msg.intent < WCBE_NUM_MSG_INTENTS) {
            wcbe_msg_t *msg = NULL;
            void read_fixed_size_msg(gssize size) {
                msg = g_malloc0(size);
                iovec[0].iov_base = msg;
                iovec[0].iov_len = size;
                num_bytes = recvmsg(sockfd, &msghdr, MSG_WAITALL);
                if (num_bytes != size) {
                    g_free(msg);
                    msg = NULL;
                }
            }
            switch (in_msg.type) {
              case WCBE_CONFIG_CREATE_MSG: {
                msg = g_malloc0(sizeof(wcbe_config_create_msg_t));
                iovec[0].iov_base = msg;
                iovec[0].iov_len = sizeof(wcbe_config_create_msg_t);
                num_bytes = recvmsg(sockfd, &msghdr, MSG_PEEK);
                if (num_bytes == sizeof(wcbe_config_create_msg_t)) {
                    wcbe_config_create_msg_t *ccmsg =
                        (wcbe_config_create_msg_t *)msg;
                    gsize msgsize =
                        sizeof(wcbe_config_create_msg_t) + ccmsg->desc_len;
                    msg = g_realloc(msg, msgsize);
                    iovec[0].iov_base = msg;
                    iovec[0].iov_len = msgsize;
                    num_bytes = recvmsg(sockfd, &msghdr, MSG_WAITALL);
                    if (num_bytes != msgsize) {
                        g_free(msg);
                        msg = NULL;
                    }
                }
                break;
              }
              case WCBE_ADD_ACTIVATION_MSG:
                read_fixed_size_msg(sizeof(wcbe_add_activation_msg_t));
                break;
              case WCBE_REMOVE_ACTIVATION_MSG:
                read_fixed_size_msg(sizeof(wcbe_remove_activation_msg_t));
                break;
              case WCBE_CONFIG_DESTROY_MSG:
                read_fixed_size_msg(sizeof(wcbe_config_destroy_msg_t));
                break;
              case WCBE_CONFIG_LIST_MSG:
                read_fixed_size_msg(sizeof(wcbe_config_list_msg_t));
                break;
              case WCBE_ACTIVATION_LIST_MSG:
                read_fixed_size_msg(sizeof(wcbe_activation_list_msg_t));
                break;
              case WCBE_SRC_MSG:
                read_fixed_size_msg(sizeof(wcbe_src_msg_t));
                break;
              case WCBE_TS0_MSG:
                read_fixed_size_msg(sizeof(wcbe_ts0_msg_t));
                break;
              case WCBE_FRAMEDUMP_MSG:
                read_fixed_size_msg(sizeof(wcbe_framedump_msg_t));
                break;
              case WCBE_FLOW_MSG:
                read_fixed_size_msg(sizeof(wcbe_flow_msg_t));
                break;
              case WCBE_RATE_MSG:
                read_fixed_size_msg(sizeof(wcbe_rate_msg_t));
                break;
              case WCBE_NAME_MSG:
                read_fixed_size_msg(sizeof(wcbe_name_msg_t));
                break;
              case WCBE_QUIT_MSG:
                read_fixed_size_msg(sizeof(wcbe_quit_msg_t));
                break;
              case WCBE_PIPELINE_LIST_MSG:
                read_fixed_size_msg(sizeof(wcbe_pipeline_list_msg_t));
                break;
              case WCBE_ELEMENT_LIST_MSG:
                read_fixed_size_msg(sizeof(wcbe_element_list_msg_t));
                break;
              case WCBE_PROPERTY_LIST_MSG:
                read_fixed_size_msg(sizeof(wcbe_property_list_msg_t));
                break;
              case WCBE_PAD_LIST_MSG:
                read_fixed_size_msg(sizeof(wcbe_pad_list_msg_t));
                break;
              case WCBE_PAD_CAPS_MSG:
                read_fixed_size_msg(sizeof(wcbe_pad_caps_msg_t));
                break;
              case WCBE_PAD_PEER_MSG:
                read_fixed_size_msg(sizeof(wcbe_pad_peer_msg_t));
                break;
              case WCBE_CONFIG_DOC_MSG: {
                msg = g_malloc0(sizeof(wcbe_config_doc_msg_t));
                iovec[0].iov_base = msg;
                iovec[0].iov_len = sizeof(wcbe_config_doc_msg_t);
                num_bytes = recvmsg(sockfd, &msghdr, MSG_PEEK);
                if (num_bytes == sizeof(wcbe_config_doc_msg_t)) {
                    wcbe_config_doc_msg_t *cdmsg =
                        (wcbe_config_doc_msg_t *)msg;
                    gsize msgsize =
                        sizeof(wcbe_config_doc_msg_t) + cdmsg->doc_len;
                    msg = g_realloc(msg, msgsize);
                    iovec[0].iov_base = msg;
                    iovec[0].iov_len = msgsize;
                    num_bytes = recvmsg(sockfd, &msghdr, MSG_WAITALL);
                    if (num_bytes != msgsize) {
                        g_free(msg);
                        msg = NULL;
                    }
                }
                break;
              }
              case WCBE_AUX_OUTPUT_DIR_MSG:
                read_fixed_size_msg(sizeof(wcbe_aux_output_dir_msg_t));
                break;
              case WCBE_RELEASE_MSG:
                read_fixed_size_msg(sizeof(wcbe_release_msg_t));
                break;
              case WCBE_NEW_SUBSCAN_MSG:
                read_fixed_size_msg(sizeof(wcbe_new_subscan_msg_t));
                break;
              case WCBE_CANCEL_SUBSCAN_MSG:
                read_fixed_size_msg(sizeof(wcbe_cancel_subscan_msg_t));
                break;
              case WCBE_BDF_OUTPUT_DIR_MSG:
                read_fixed_size_msg(sizeof(wcbe_bdf_output_dir_msg_t));
                break;
              case WCBE_CONFIG_INTEGRATION_MSG:
                read_fixed_size_msg(sizeof(wcbe_config_integration_msg_t));
                break;
              case WCBE_PROPERTY_MSG: {
                  msg = g_malloc0(sizeof(wcbe_property_msg_t));
                  iovec[0].iov_base = msg;
                  iovec[0].iov_len = sizeof(wcbe_property_msg_t);
                  num_bytes = recvmsg(sockfd, &msghdr, MSG_PEEK);
                  if (num_bytes == sizeof(wcbe_property_msg_t)) {
                      wcbe_property_msg_t *pmsg = (wcbe_property_msg_t *)msg;
                      gsize msgsize =
                          sizeof(wcbe_property_msg_t) + pmsg->value_len;
                      msg = g_realloc(msg, msgsize);
                      iovec[0].iov_base = msg;
                      iovec[0].iov_len = msgsize;
                      num_bytes = recvmsg(sockfd, &msghdr, MSG_WAITALL);
                      if (num_bytes != msgsize) {
                          g_free(msg);
                          msg = NULL;
                      }
                  }
                  break;
              }
              default:
                g_debug("Received unknown message of type %u and intent %u",
                        in_msg.type, in_msg.intent);
                read_fixed_size_msg(0);
                msg = NULL;
                break;
            }
            if (msg != NULL) {
                request_fn_t fn =
                    request_handlers[msg->type][msg->intent];
                if (fn != NULL) {
                    gsize msgsize = iovec[0].iov_len;
                    request_cb_fn_t cb = fn(pipeline, &msg, &msgsize);
                    if (cb == NULL) {
                        iovec[0].iov_base = msg;
                        iovec[0].iov_len = msgsize;
                        msg->direction = WCBE_RESPONSE;
                        num_bytes = sendmsg(sockfd, &msghdr, 0);
                        if (num_bytes < iovec[0].iov_len) {
                            g_warning("Input comm channel error on write: %s",
                                      g_strerror(errno));
                        }
                        g_free(msg);
                    }
                    else {
                        cb_data_t *cbdata = g_new(cb_data_t, 1);
                        cbdata->msgsize = msgsize;
                        cbdata->cb = cb;
                        cbdata->msg = msg;
                        cbdata->pipeline = pipeline;
                        memcpy(&(cbdata->addr), &addr, sizeof(addr));
                        cbdata->timeout = g_timeout_add(
                            LS_MSG_TIMEOUT_MS, cb_timeout, cbdata);
                        g_hash_table_insert(pipeline->msg_callbacks,
                                            msg, cbdata);
                    }
                }
                else {
                    if (msg->type == WCBE_RELEASE_MSG) {
                        g_free(msg);
                        goto shutdown_sock;
                    }
                    g_free(msg);
                }
            }
            else {
                g_info("Discarding message "
                       "of type %u and intent %u", in_msg.type, in_msg.intent);
            }
        }
        else if (num_bytes == 0) {
            goto shutdown_sock;
        }
        else if (num_bytes == -1) {
            if (errno != EAGAIN && errno != EWOULDBLOCK) {
                g_info("Error reading from control socket: %s",
                       g_strerror(errno));
                goto shutdown_sock;
            }
        }
        else {
            if (num_bytes != sizeof(in_msg)) {
                g_debug("Received partial control message...discarded");
            }
            if (in_msg.direction != WCBE_REQUEST) {
                g_debug("Received 'reply' control message...discarded");
            }
            if (in_msg.type >= WCBE_NUM_MSG_TYPES) {
                g_debug("Received unknown control message type %u...discarded",
                        in_msg.type);
            }
            if (in_msg.intent >= WCBE_NUM_MSG_INTENTS) {
                g_debug("Received unknown control message intent %u..."
                        "discarded", in_msg.intent);
            }
            recvmsg(sockfd, &msghdr, 0);
        }
    }
#if CONTROL_SOCK_TCP
    if (condition & G_IO_HUP) {
        goto shutdown_sock;
    }
#endif
    if (condition & G_IO_ERR) {
        goto shutdown_sock;
    }
    return TRUE;

shutdown_sock:
#if CONTROL_SOCK_TCP
    pipeline->ctlsock_source = 0;
#endif
    shutdown(g_io_channel_unix_get_fd(pipeline->ctlsock),
             SHUT_RDWR);
    g_io_channel_unref(pipeline->ctlsock); // closes file descriptor, too
    pipeline->ctlsock = NULL;
    return FALSE;
}

static gboolean
handle_child_comm(GIOChannel *source, GIOCondition condition, gpointer data)
{
    static gboolean initialized = FALSE;
    if (!initialized) {
        // Load enumeration types that might be passed back from the
        // lagset processing pipeline stage. TODO: implement a better
        // way to do this.
        if (WCBE_CORR_BITS_ENUM_TYPE == 0) {
            g_error("Failed to load WcbeCorrBits enumeration type.");
        }
        initialized = TRUE;
    }

    pipeline_t *pipeline = data;
    guint str_len;
    struct iovec iovec[2] =
        {{.iov_base = &str_len, .iov_len = sizeof(str_len)}};
    struct msghdr msghdr =
        {.msg_name = NULL, .msg_namelen = 0,
         .msg_control = NULL, .msg_controllen = 0,
         .msg_iov = iovec, .msg_iovlen = 1};
    gint fd = g_io_channel_unix_get_fd(source);
    gssize nbytes = recvmsg(fd, &msghdr, MSG_PEEK | MSG_TRUNC);
    if (G_LIKELY(nbytes > 0)) {
        iovec[1].iov_base = g_malloc(str_len);
        iovec[1].iov_len = str_len;
        msghdr.msg_iovlen = 2;
        nbytes = recvmsg(fd, &msghdr, 0);
        if (G_LIKELY(nbytes == iovec[0].iov_len + iovec[1].iov_len)) {
            ((gchar *)(iovec[1].iov_base))[iovec[1].iov_len - 1]
                = '\0';
            // The following is a workaround, since
            // gst_structure_from_string does not restore octal
            // sequences created by gst_structure_to_string in string
            // fields.
            fixup_octal_sequences((gchar *)iovec[1].iov_base);
            GstStructure *structure =
                gst_structure_from_string((gchar *)iovec[1].iov_base, NULL);
            if (structure != NULL) {
                GstElement *is_pipeline =
                    (pipeline->current.pipeline ? pipeline->current.pipeline
                     : pipeline->old.pipeline);
                g_assert(is_pipeline != NULL);
                GstMessage *msg = gst_message_new_application(
                    GST_OBJECT(is_pipeline), structure);
                gboolean res = gst_element_post_message(
                    GST_ELEMENT(is_pipeline), msg);
                if (!res) {
                    GST_DEBUG("Failed to post message to bus");
                }
            }
            else {
                GST_DEBUG("Failed to parse application message '%s'",
                          iovec[1].iov_base);
            }
        }
        else {
            GST_DEBUG("Failed to read expected number of bytes in "
                      "application message");
        }
        g_free(iovec[1].iov_base);
    }
    else {
        GST_DEBUG("Failed to read expected number of bytes in "
                  "application message header");
    }

    return TRUE;
}

#if CONTROL_SOCK_TCP
static gboolean
handle_connect(GIOChannel *channel, GIOCondition condition, gpointer data)
{
    gboolean result = TRUE;
    pipeline_t *pipeline = data;
    if (condition & G_IO_IN) {
        socklen_t socklen;
        gint fd =
            accept(g_io_channel_unix_get_fd(channel), NULL, &socklen);
        if (fd != -1) {
            if (pipeline->ctlsock == NULL) {
                // TODO: create channel with this fd, set flags, add
                // event handler for input, etc.
                pipeline->ctlsock = new_channel(fd);
                pipeline->ctlsock_source =
                    g_io_add_watch(pipeline->ctlsock,
                                   G_IO_IN | G_IO_HUP | G_IO_ERR,
                                   handle_comm, pipeline);
            }
            else {
                // only one ctlsock at a time, please
                close(fd);
            }
        }
        else {
            if (errno != EAGAIN && errno != EWOULDBLOCK) {
                g_warning("Failed to accept connection to control socket: %s",
                          g_strerror(errno));
            }
        }
    }
    if (condition & G_IO_ERR) {
        result = FALSE;
    }
    return result;
}
#endif

GstStateChangeReturn
wcbe_start_pipeline(pipeline_t *pipeline, gboolean in_main_loop,
                    GstClockTime timeout)
{
    g_assert(pipeline->current.pipeline != NULL);
    GstStateChangeReturn result =
        change_pipeline_state(pipeline->current.pipeline, in_main_loop,
                              GST_STATE_PLAYING, timeout);
    component_pipeline_t *cp = wcbe_find_component_pipeline(
        pipeline->configless_pipelines, input_stage_pipeline_name);
    cp->state = PIPELINE_STATE_RUNNING;
    return result;
}

GstStateChangeReturn
wcbe_stop_pipeline(pipeline_t *pipeline, gboolean in_main_loop,
                   GstClockTime timeout)
{
    g_assert(pipeline->current.pipeline != NULL);
    return change_pipeline_state(pipeline->current.pipeline, in_main_loop,
                                 GST_STATE_PAUSED, timeout);
}

static GstStateChangeReturn
change_pipeline_state(GstElement *pipeline, gboolean in_main_loop,
                      GstState state, GstClockTime timeout)
{
    g_return_val_if_fail(pipeline != NULL, GST_STATE_CHANGE_FAILURE);

    GstStateChangeReturn result;
    result = gst_element_set_state(pipeline, state);
    if (result == GST_STATE_CHANGE_FAILURE) {
        const gchar *st = (state == GST_STATE_PLAYING ? "start" : "stop");
        g_warning("Failed to %s pipeline", st);
        if (!in_main_loop) {
            GstBus *bus =
                gst_pipeline_get_bus(GST_PIPELINE(pipeline));
            GstMessage *msg = gst_bus_poll(bus, GST_MESSAGE_ERROR, 0);
            if (msg != NULL) {
                GError *error = NULL;
                gst_message_parse_error(msg, &error, NULL);
                g_warning("ERROR: %s\n", error->message);
                g_error_free(error);
                gst_message_unref(msg);
            }
            gst_object_unref(GST_OBJECT(bus));
        }
    }
    else if (timeout > 0 && result == GST_STATE_CHANGE_ASYNC) {
        result =
            gst_element_get_state(pipeline, NULL, NULL, timeout);
    }
    return result;
}

static void
destroy_configuration(gpointer data)
{
    configuration_t *config = data;
    gboolean in_main_loop = g_main_loop_is_running(config->pipeline->loop);

/*     config->destroy = TRUE; */
    config->state = CONFIG_STATE_FINALIZING;
    GSList *ls_pipelines = NULL;
    if (config->ls_pipelines != NULL) {
        GSList *list = config->ls_pipelines;
        while (list != NULL) {
            component_pipeline_t *cp = list->data;
            g_assert(cp->type == LAGSET_PIPELINE);
            if (cp->state != PIPELINE_STATE_DEFUNCT) {
                ls_pipelines = g_slist_prepend(ls_pipelines, cp);
            }
            else {
                free_component_pipeline(cp);
            }
            list = g_slist_next(list);
        }
        g_slist_free(config->ls_pipelines);
        config->ls_pipelines = NULL;

        void mark_destroyed(gpointer data, gpointer user_data) {
            component_pipeline_t *cp = data;
            cp->lsp.config = NULL;
            cp->destroy = TRUE;
        }
        g_slist_foreach(ls_pipelines, mark_destroyed, NULL);

        config->pipeline->configless_pipelines = g_slist_concat(
            config->pipeline->configless_pipelines, ls_pipelines);
    }

    // destroy config in input pipeline
    if (config->pipeline->current.pipeline != NULL) {
        signal_destroy_config(
            config, GST_PIPELINE(config->pipeline->current.pipeline));
    }
    if (config->pipeline->old.pipeline != NULL) {
        signal_destroy_config(
            config, GST_PIPELINE(config->pipeline->old.pipeline));
    }

    // schedule a process kill signal to every pipeline in
    // ls_pipelines
    void schedule_kill(gpointer data, gpointer user_data) {
        component_pipeline_t *cp = data;
        if (in_main_loop) {
            cp->lsp.exit_timeout = g_timeout_add(5000, pipeline_term, cp);
        }
        else {
            pipeline_kill(cp);
        }
    }
    g_slist_foreach(ls_pipelines, schedule_kill, NULL);
    wcbe_cleanup_config(config);
}

static void
create_ls_pipelines(pipeline_t *pipeline, configuration_t *config)
{
    GArray *pipeline_configs = config->wconfig.pipelines.pipelines;
    if (pipeline_configs != NULL
        && config->wconfig.output.subarrays != NULL) {
        guint i;
        for (i = 0; i < pipeline_configs->len; ++i) {
            const wcbe_pipeline_config_t *pcfg =
                &g_array_index(pipeline_configs, wcbe_pipeline_config_t, i);
            component_pipeline_t *cp = g_new0(component_pipeline_t, 1);
            cp->state = PIPELINE_STATE_UNKNOWN;
            cp->type = LAGSET_PIPELINE;
            cp->destroy = FALSE;
            ls_pipeline_t *lsp = &(cp->lsp);
            lsp->pipeline = pipeline;
            lsp->config = config;
            const wcbe_subarray_config_t *sa = &g_array_index(
                config->wconfig.output.subarrays, wcbe_subarray_config_t,
                pcfg->subarray_index);
            if (spawn_ls_pipeline(config->wconfig.name, pcfg, cp,
                                  pipeline->bdfdir, sa->integration)) {
                config->ls_pipelines =
                    g_slist_prepend(config->ls_pipelines, cp);
            }
            else {
                g_free(cp);
            }
        }
    }
}

static gboolean
spawn_ls_pipeline(const gchar *cfg_name,
                  const wcbe_pipeline_config_t *pipeline,
                  component_pipeline_t *cp, const gchar *bdfdir,
                  guint integration)
{
    g_return_val_if_fail(cp->type == LAGSET_PIPELINE, FALSE);
    ls_pipeline_t *lsp = &(cp->lsp);
    gboolean result = FALSE;
    gboolean free_lsp = TRUE;
    lsp->name = g_build_filename(cfg_name, pipeline->id, NULL);
    lsp->sockaddr.sun_family = AF_UNIX;
    gint ret = g_snprintf(lsp->sockaddr.sun_path,
                          sizeof(lsp->sockaddr.sun_path),
                          "%s/%s", lsp->pipeline->sockdir, lsp->name);
    g_assert(ret <= sizeof(lsp->sockaddr.sun_path));
    gchar *cmd = g_strdup_printf(
        "lagset_pipeline --input %s --pids %s --output %s --integration %u %s",
        lsp->sockaddr.sun_path, lsp->pipeline->pidsdir, bdfdir, integration,
        pipeline->spec.gst_spec);
    lsp->argv = NULL;
    gint argc;
    GError *err = NULL;
    gboolean rc = g_shell_parse_argv(cmd, &argc, &(lsp->argv), &err);
    if (rc) {
        rc = g_spawn_async(NULL, lsp->argv, NULL,
                           G_SPAWN_DO_NOT_REAP_CHILD | G_SPAWN_SEARCH_PATH,
                           NULL, NULL, &(lsp->pid), &err);
        if (rc) {
            g_child_watch_add(lsp->pid, pipeline_exit_notify, cp);
            lsp->spawn_count++;
            result = TRUE;
            free_lsp = FALSE;
        }
        else {
            g_debug("Failed to spawn lagset pipeline process: %s",
                    err->message);
        }
    }
    else {
        g_info("Failed to parse lagset pipeline command line '%s': %s",
               cmd, err->message);
    }
    if (free_lsp && lsp->argv != NULL) {
        g_strfreev(lsp->argv);
    }
    g_free(cmd);

    if (free_lsp) {
        if (lsp->name != NULL) {
            g_free(lsp->name);
        }
    }
    return result;
}

static void
pipeline_exit_notify(GPid pid, gint status, gpointer data)
{
    component_pipeline_t *cp = data;
    g_return_if_fail(cp->type == LAGSET_PIPELINE);
    ls_pipeline_t *lsp = &(cp->lsp);
    gboolean respawned = FALSE;

    g_spawn_close_pid(pid);

    if (lsp->exit_timeout != 0){
        g_source_remove(lsp->exit_timeout);
        lsp->exit_timeout = 0;
    }

    if (!cp->destroy) {
        if (WIFEXITED(status)) {
            g_info("Pipeline process exited unexpectedly with status %u",
                   WEXITSTATUS(status));
        }
        else if (WIFSIGNALED(status)) {
            g_info("Pipeline process exited with signal %u",
                   WTERMSIG(status));
        }
        if (lsp->spawn_count < MAX_SPAWN_ATTEMPTS) {
            GError *err = NULL;
            gint rc = g_spawn_async(NULL, lsp->argv, NULL,
                                    G_SPAWN_DO_NOT_REAP_CHILD
                                    | G_SPAWN_SEARCH_PATH,
                                    NULL, NULL, &(lsp->pid), &err);
            if (rc) {
                g_child_watch_add(lsp->pid, pipeline_exit_notify, lsp);
                lsp->spawn_count++;
                cp->state = PIPELINE_STATE_UNKNOWN;
                respawned = TRUE;
            }
            else {
                g_debug("Failed to spawn lagset pipeline process: %s",
                        err->message);
            }
        }
        else {
            g_info("Pipeline program '%s' failed %u times...will not respawn",
                   lsp->argv[0], lsp->spawn_count);
        }
    }

    if (!respawned) {
        configuration_t *config = lsp->config;
        cp->state = PIPELINE_STATE_DEFUNCT;
        GSList **list = NULL;
        gchar *list_adjective = NULL;
        if (config == NULL) {
            list = &(lsp->pipeline->configless_pipelines);
            list_adjective = "old";
        }
        else {
            config->state = CONFIG_STATE_NOT_READY;
/*             config->destroy = TRUE; */
            list = &(config->ls_pipelines);
            list_adjective = "config";
        }
        if (list != NULL) {
            component_pipeline_t *cp =
                wcbe_find_component_pipeline(*list, lsp->name);
            if (cp != NULL) {
                *list = g_slist_remove(*list, cp);
                if (all_pipelines_finished(lsp->pipeline)) {
                    lsp->pipeline->state = WCBE_STATE_EXITING;
                    g_debug("process exit triggered cleanup");
                    cleanup_old_pipeline(lsp->pipeline);
                }
                free_component_pipeline(cp);
            }
            else {
                g_warning("Failed to find defunct pipeline '%s' in '%s' list",
                          lsp->name, list_adjective);
            }
        }
    }
}

static gboolean
pipeline_term(gpointer data)
{
    component_pipeline_t *cp = data;
    g_return_val_if_fail(cp->type == LAGSET_PIPELINE, FALSE);
    cp->destroy = TRUE;
    gint rc = kill(cp->lsp.pid, SIGTERM);
    if (rc != 0 && errno != ESRCH) {
        // ignore errno == ESRCH since the process could have exited
        // between the timeout and the 'kill' invocation
        g_warning("Failed to quit lagset pipeline process %d: %s",
                  cp->lsp.pid, g_strerror(errno));
    }
    cp->lsp.exit_timeout = g_timeout_add(5000, pipeline_kill, cp);
    return FALSE;
}

static gboolean
pipeline_kill(gpointer data)
{
    component_pipeline_t *cp = data;
    g_return_val_if_fail(cp->type == LAGSET_PIPELINE, FALSE);
    cp->destroy = TRUE;
    gint rc = kill(cp->lsp.pid, SIGKILL);
    if (rc != 0 && errno != ESRCH) {
        // ignore errno == ESRCH since the process could have exited
        // between the timeout and the 'kill' invocation
        g_warning("Failed to kill lagset pipeline process %d: %s",
                  cp->lsp.pid, g_strerror(errno));
    }
    return FALSE;
}

void
wcbe_cleanup_config(configuration_t *config)
{
    // remove config directory
    rm_path(config->wconfig.dir, "");

    g_free(config->wconfig.name);
    config->wconfig.name = NULL;
    g_free(config->wconfig.dir);
    config->wconfig.dir = NULL;
    wcbe_config_free(&(config->wconfig));
    g_free(config);
}

gboolean
wcbe_create_config(pipeline_t *pipeline, configuration_t *config)
{
    g_assert(pipeline->current.pipeline != NULL);

    gboolean result;

    insert_xml_ids(&(config->wconfig));
    gchar *doc = NULL;
    wcbe_config_doc(&(config->wconfig), &doc);
    if (doc != NULL) {
        gchar *filename =
            g_build_filename(config->wconfig.dir, "config.xml", NULL);
        FILE *f = fopen(filename, "w");
        if (f != NULL) {
            fputs(doc, f);
            fclose(f);
        }
        else {
            g_critical("Failed to open file '%s': config file will not be "
                       "written to config directory.", filename);
        }
        g_free(filename);
        g_free(doc);
    }
    else {
        g_critical("Failed to create config doc string: config file will "
                   "not be written in config directory.");
    }

    result =
        signal_create_config(config, GST_PIPELINE(pipeline->current.pipeline));
    if (result) {
        config->state = CONFIG_STATE_INITIALIZING;
        create_ls_pipelines(pipeline, config);
        g_static_rec_mutex_lock(&(pipeline->configs_mtx));
        g_hash_table_insert(pipeline->configs, config->wconfig.name, config);
        g_static_rec_mutex_unlock(&(pipeline->configs_mtx));
    }
    return result;
}

void
wcbe_add_activation(pipeline_t *pipeline, configuration_t *config,
                    const lf_widar_time_t *time, gboolean now)
{
    g_return_if_fail(!config->destroy);

    g_assert(pipeline->current.pipeline != NULL);

    activation_t *activation = g_new(activation_t, 1);
    memcpy(&(activation->time), time, sizeof(lf_widar_time_t));
    activation->config = (wcbe_config_t *)config;
    activation->in_use = FALSE;
    activation->deactivated = FALSE;
    activation->now = now;
    config->num_activations++;
    if (signal_add_activation(config, activation,
                              GST_PIPELINE(pipeline->current.pipeline))) {
        pipeline->activations =
            g_slist_prepend(pipeline->activations, activation);
    }
}

void
wcbe_remove_activation(pipeline_t *pipeline, configuration_t *config,
                       const lf_widar_time_t *time)
{
    g_assert(pipeline->current.pipeline != NULL);

    activation_t activation = {.config = (wcbe_config_t *)config,
                               .deactivated = FALSE};
    memcpy(&(activation.time), time, sizeof(lf_widar_time_t));
    gint cmp_act(gconstpointer a, gconstpointer b) {
        const activation_t *act_a = (activation_t *)a;
        const activation_t *act_b = (activation_t *)b;
        if (act_a->config == act_b->config
            && memcmp(&(act_a->time), &(act_b->time),
                      sizeof(lf_widar_time_t)) == 0) {
            return 0;
        }
        return 1;
    }
    GSList *act =
        g_slist_find_custom(pipeline->activations, &activation, cmp_act);
    if (act != NULL) {
        signal_remove_activation(config, act->data,
                                 GST_PIPELINE(pipeline->current.pipeline));
    }
}

void
wcbe_destroy_config(pipeline_t *pipeline, configuration_t *config)
{
    config->destroy = TRUE;
    void signal_rm(gpointer data, gpointer user_data G_GNUC_UNUSED) {
        activation_t *act = data;
        if (act->config == (wcbe_config_t *)config && !act->in_use
            && !act->deactivated) {
            if (pipeline->current.pipeline != NULL) {
                signal_remove_activation(
                    config, act, GST_PIPELINE(pipeline->current.pipeline));
            }
            if (pipeline->old.pipeline != NULL) {
                signal_remove_activation(
                    config, act, GST_PIPELINE(pipeline->old.pipeline));
            }
        }
    }
    g_slist_foreach(pipeline->activations, signal_rm, NULL);
}

static void
free_component_pipeline(component_pipeline_t *cp)
{
    if (cp->type == LAGSET_PIPELINE) {
        gchar *pid = g_strdup_printf("%u", cp->lsp.pid);
        gchar *pidfile =
            g_build_filename(cp->lsp.pipeline->pidsdir, pid, NULL);
        gint rc = unlink(pidfile);
        if (rc != 0 && errno != ENOENT) {
            g_warning("Failed to unlink lagset_pipeline pid file %s: %s",
                      pidfile, g_strerror(errno));
        }
        g_free(pidfile);
        g_free(pid);
        g_strfreev(cp->lsp.argv);
        g_free(cp->lsp.name);
    }
    g_free(cp);
}

static void
pipeline_eos_handler(GstElement *element, const gchar *lsp_name, gpointer data)
{
    pipeline_t *pipeline = data;
    gchar **parts = g_strsplit(lsp_name, G_DIR_SEPARATOR_S, 2);
    g_static_rec_mutex_lock(&(pipeline->configs_mtx));
    configuration_t *config = g_hash_table_lookup(pipeline->configs, parts[0]);
    g_static_rec_mutex_unlock(&(pipeline->configs_mtx));
    GSList *slist = ((config != NULL) ? config->ls_pipelines
                     : pipeline->configless_pipelines);
    component_pipeline_t *cp = wcbe_find_component_pipeline(slist, lsp_name);
    if (cp != NULL) {
        cp->destroy = TRUE;
    }
    g_strfreev(parts);
}

static void
config_deactivated_handler(GstElement *element, activation_t *activation,
                           gpointer data)
{
    activation->deactivated = TRUE;
    g_info("Deactivated config %s", activation->config->name);
}

static void
config_activated_handler(GstElement *element,
                         gchar *name, gboolean result,
                         gpointer data)
{
    const gchar *qual;
    if (result) {
        qual = "Successfully";
    }
    else {
        qual = "Unsuccessfully";
    }
    g_info("%s activated configuration %s", qual, name);
}

static gboolean
cb_timeout(gpointer data)
{
    if (data != NULL) {
        cb_data_t *cbdata = data;
        if (g_hash_table_remove(cbdata->pipeline->msg_callbacks,
                                cbdata->msg)) {
            struct iovec iovec[] =
                {{.iov_base = cbdata->msg, .iov_len = cbdata->msgsize}};
            struct msghdr msghdr = {
                .msg_name = &(cbdata->addr),
                .msg_namelen = sizeof(cbdata->addr),
                .msg_control = NULL, .msg_controllen = 0,
                .msg_iov = iovec, .msg_iovlen = 1};
            cbdata->msg->direction = WCBE_RESPONSE;
            gssize num_bytes =
                sendmsg(g_io_channel_unix_get_fd(cbdata->pipeline->ctlsock),
                        &msghdr, 0);
            if (num_bytes < (gssize)iovec[0].iov_len) {
                g_warning("Input comm channel error on write: %s",
                          g_strerror(errno));
            }
        }
        g_free(cbdata->msg);
        g_free(cbdata);
    }
    return FALSE;
}

static void
insert_xml_ids(wcbe_config_t *config)
{
    g_return_if_fail(config != NULL);

    // start with product ids
    GPtrArray *products = config->input.products;
    if (products != NULL) {
        IdStore *prd4ids = idstring_new_store("prd4");
        IdStore *prd7ids = idstring_new_store("prd7");
        guint i;
        // first pass, reserve already present ids
        for (i = 0; i < products->len; ++i) {
            wcbe_product_config_t *prd = g_ptr_array_index(products, i);
            if (prd->nbits == 4) {
                if (prd->id != NULL) {
                    idstring_reserve_id(prd4ids, prd->id);
                }
            }
            else if (prd->nbits == 7) {
                if (prd->id != NULL) {
                    idstring_reserve_id(prd7ids, prd->id);
                }
            }
        }
        // second pass, generate missing ids
        for (i = 0; i < products->len; ++i) {
            wcbe_product_config_t *prd = g_ptr_array_index(products, i);
            if (prd->nbits == 4) {
                if (prd->id == NULL) {
                    prd->id = g_strdup(idstring_new_id(prd4ids));
                }
            }
            else if (prd->nbits == 7) {
                if (prd->id == NULL) {
                    prd->id = g_strdup(idstring_new_id(prd7ids));
                }
            }
        }
        idstring_destroy_store(prd7ids);
        idstring_destroy_store(prd4ids);
    }

    // now do pipeline ids
    GArray *pipelines = config->pipelines.pipelines;
    if (pipelines != NULL) {
        IdStore *plnids = idstring_new_store("pln");
        // first pass, reserve already present ids
        guint i;
        for (i = 0; i < pipelines->len; ++i) {
            wcbe_pipeline_config_t *pln =
                &g_array_index(pipelines, wcbe_pipeline_config_t, i);
            if (pln->id != NULL) {
                idstring_reserve_id(plnids, pln->id);
            }
        }
        // second pass, generate missing ids
        for (i = 0; i < pipelines->len; ++i) {
            wcbe_pipeline_config_t *pln =
                &g_array_index(pipelines, wcbe_pipeline_config_t, i);
            if (pln->id == NULL) {
                pln->id = g_strdup(idstring_new_id(plnids));
            }
        }
        idstring_destroy_store(plnids);
    }
}

static gboolean
activation_configuration_cleanup(gpointer data)
{
    g_return_val_if_fail(data != NULL, FALSE);

    gboolean result = TRUE;
    pipeline_t *pipeline = (pipeline_t *)data;
    static enum {ACTIVATIONS_PHASE=0, CONFIGURATIONS_PHASE} phase =
        ACTIVATIONS_PHASE;
    static guint finalization_countdown = MAX(10 / CLEANUP_TIMEOUT_SEC, 4);

    switch (phase) {
      case ACTIVATIONS_PHASE: {
        GSList *link = pipeline->activations;
        GSList *prev = NULL;
        while (link != NULL) {
            GSList *next = link->next;
            activation_t *act = (activation_t *)(link->data);
            if (act->deactivated) {
                ((configuration_t *)(act->config))->num_activations--;
                g_slist_free1(link);
                g_free(act);
                if (prev != NULL) {
                    prev->next = next;
                }
                else {
                    pipeline->activations = next;
                }
            }
            else {
                prev = link;
            }
            link = next;
        }
        phase = CONFIGURATIONS_PHASE;
        break;
      }
      case CONFIGURATIONS_PHASE: {
        gboolean destroy_test(gpointer key, gpointer value,
                              gpointer user_data) {
            configuration_t *config = (configuration_t *)value;
            gboolean result =
                (config->destroy
                 && (config->ls_pipelines == NULL
                     || ((config->num_activations == 0)
                         && (config->state == CONFIG_STATE_READY
                             || config->state == CONFIG_STATE_NOT_READY))));
            if (result) {
                g_debug("removing config %s", config->wconfig.name);
            }
            return result;
        }
        g_static_rec_mutex_lock(&(pipeline->configs_mtx));
        g_hash_table_foreach_remove(pipeline->configs, destroy_test, NULL);
        g_static_rec_mutex_unlock(&(pipeline->configs_mtx));
        phase = ACTIVATIONS_PHASE;
        break;
      }
      default:
        g_assert_not_reached();
        break;
    }

    if (pipeline->state == WCBE_STATE_FINALIZING) {
        g_static_rec_mutex_lock(&(pipeline->configs_mtx));
        if (--finalization_countdown == 0
            || g_hash_table_size(pipeline->configs) == 0) {
            pipeline->state = WCBE_STATE_EXITING;
            result = FALSE;
        }
        g_static_rec_mutex_unlock(&(pipeline->configs_mtx));
    }

    return result;
}

static void
signal_destroy_config(configuration_t *config, GstPipeline *pipeline)
{
    g_return_if_fail(config != NULL);
    g_return_if_fail(pipeline != NULL);

    g_debug("signalling 'destroy-configuration'");
    g_signal_emit(pipeline, destroy_cfg_signal, 0, &(config->wconfig));
}

static gboolean
signal_create_config(configuration_t *config, GstPipeline *pipeline)
{
    g_return_val_if_fail(config != NULL, FALSE);
    g_return_val_if_fail(pipeline != NULL, FALSE);

    gboolean result;
    g_signal_emit(pipeline, create_cfg_signal, 0, &(config->wconfig), &result);
    return !result;
}

static gboolean
signal_add_activation(configuration_t *config, activation_t *activation,
                      GstPipeline *pipeline)
{
    g_return_val_if_fail(config != NULL, FALSE);
    g_return_val_if_fail(pipeline != NULL, FALSE);

    g_debug("signalling 'add-activation'");
    gboolean result;
    g_signal_emit(pipeline, add_actvn_signal, 0, activation, &result);
    return !result;
}

static void
signal_remove_activation(configuration_t *config, activation_t *activation,
                         GstPipeline *pipeline)
{
    g_return_if_fail(config != NULL);
    g_return_if_fail(pipeline != NULL);

    g_debug("signalling 'remove-activation'");
    g_signal_emit(pipeline, remove_actvn_signal, 0, activation);
}

static gboolean
failure_accumulator(GSignalInvocationHint *ihint, GValue *return_accu,
                    const GValue *handler_return, gpointer data)
{
    // It would be simpler to have an "and" accumulator on the return
    // values, but since the return_accu value is initialized to 0, I
    // don't see any other way around it.
    g_value_set_boolean(return_accu,
                        g_value_get_boolean(return_accu)
                        || !g_value_get_boolean(handler_return));
    return TRUE;
}

static void
fixup_octal_sequences(gchar *str)
{
    guint odigit = 0;
    gchar octal = 0;
    gchar *in = str;
    gchar *out = str;
    while (*in != '\0') {
        switch (odigit) {
          case 0:
            if (*in == '\\') {
                odigit = 1;
            }
            else {
                *out++ = *in;
            }
            break;
          case 1:
            if (*in == '0') {
                odigit = 2;
                octal = 0;
            }
            else {
                *out++ = *(in - 1);
                *out++ = *in;
                odigit = 0;
            }
            break;
          case 2:
            if ('0' <= *in && *in < '8') {
                odigit = 3;
                octal = *in - '0';
            }
            else {
                *out++ = *(in - 2);
                *out++ = *(in - 1);
                *out++ = *in;
                odigit = 0;
            }
            break;
          case 3:
            if ('0' <= *in && *in < '8') {
                octal = 8 * octal + *in - '0';
                *out++ = octal;
                odigit = 0;
            }
            else {
                *out++ = *(in - 3);
                *out++ = *(in - 2);
                *out++ = *(in - 1);
                *out++ = *in;
                odigit = 0;
            }
            break;
          default:
            g_assert_not_reached();
            break;
        }
        ++in;
    }
    *out = *in;
}

static gboolean
all_pipelines_finished(pipeline_t *pipeline)
{
    gboolean result = (pipeline->old.pipeline != NULL);
    GSList *plist = pipeline->configless_pipelines;
    while (result && plist != NULL) {
        component_pipeline_t *cp = plist->data;
        result = cp->destroy;
        plist = g_slist_next(plist);
    }
    void chk_config(gpointer key, gpointer value, gpointer user_data) {
        configuration_t *config = value;
        result = result && g_slist_length(config->ls_pipelines) == 0;
    }
    g_static_rec_mutex_lock(&(pipeline->configs_mtx));
    g_hash_table_foreach(pipeline->configs, chk_config, NULL);
    g_static_rec_mutex_unlock(&(pipeline->configs_mtx));
    return result;
}
