#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/un.h>

#include <wcbe_mc.h>
#include <wcbe_framedump.h>
#include <wcbe_config.h>
#include <wcbe_util.h>

static request_cb_fn_t ls_message(
    pipeline_t *pipeline, const gchar *config_name,
    const gchar *ls_pipeline_name, const wcbe_msg_t *msg,
    GstStructure *structure, request_cb_fn_t cb);
static gboolean forward_message(
    pipeline_t *pipeline, const gchar *config_name,
    const gchar *ls_pipeline_name, GstStructure *structure);
static configuration_t *find_config(pipeline_t *pipeline, const gchar *name);
static void element_list_query(pipeline_t *pipeline, wcbe_msg_t **msg,
                               gsize *msgsize, GstStructure *structure);
static void property_list_query(pipeline_t *pipeline, wcbe_msg_t **msg,
                                gsize *msgsize, GstStructure *structure);
static void property_query(pipeline_t *pipeline, wcbe_msg_t **msg,
                           gsize *msgsize, GstStructure *structure);
static void pad_list_query(pipeline_t *pipeline, wcbe_msg_t **msg,
                           gsize *msgsize, GstStructure *structure);
static void pad_caps_query(pipeline_t *pipeline, wcbe_msg_t **msg,
                           gsize *msgsize, GstStructure *structure);
static void pad_peer_query(pipeline_t *pipeline, wcbe_msg_t **msg,
                           gsize *msgsize, GstStructure *structure);

gchar const input_stage_pipeline_name[] = "INPUT_STAGE";

request_cb_fn_t
wcbe_mc_config_create_command(pipeline_t *pipeline, wcbe_msg_t **msg,
                              gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_CONFIG_CREATE_MSG, NULL);

    wcbe_config_create_msg_t *create_msg = (wcbe_config_create_msg_t *)*msg;
    configuration_t *config = find_config(pipeline, create_msg->name);
    if (config == NULL) {
        config = g_new0(configuration_t, 1);
        gboolean doc_ok =
            wcbe_config_parse(create_msg->desc, &(config->wconfig));
        g_debug("'%s' parse result: %d", create_msg->name, doc_ok);
        if (doc_ok) {
            config->wconfig.name = g_strdup(create_msg->name);
            config->wconfig.dir = g_build_filename(
                pipeline->sockdir, config->wconfig.name, NULL);
            config->ls_pipelines = NULL;
            config->destroy = FALSE;
            config->pipeline = pipeline;
            config->num_activations = 0;
            gint rc = mkdir(config->wconfig.dir, S_IRWXU);
            if (rc == 0) {
                gboolean free_config = TRUE;
                if (pipeline->current.pipeline != NULL) {
                    free_config = !wcbe_create_config(pipeline, config);
                }
                if (free_config) {
                    wcbe_cleanup_config(config);
                }
            }
            else {
                g_info("Failed to create configuration directory '%s': %s",
                       config->wconfig.dir, g_strerror(errno));
            }
        }
        else {
            g_info("Failed to parse configuration '%s'", create_msg->name);
        }
    }
    else {
        g_info("Rejected configuration '%s': duplicate name",
               create_msg->name);
    }
    return NULL;
}

request_cb_fn_t
wcbe_mc_add_activation_command(pipeline_t *pipeline, wcbe_msg_t **msg,
                               gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_ADD_ACTIVATION_MSG, NULL);

    wcbe_add_activation_msg_t *activation_msg =
        (wcbe_add_activation_msg_t *)*msg;
    configuration_t *config = find_config(pipeline, activation_msg->config);
    if (config != NULL) {
        if (!config->destroy) {
            if (pipeline->current.pipeline != NULL) {
                lf_widar_time_t time = activation_msg->time;
                wcbe_add_activation(pipeline, config, &time,
                                    activation_msg->now);
            }
            else {
                g_debug("Failed to add activation for config '%s': "
                        "input stage pipeline does not exist",
                        activation_msg->config);
            }
        }
        else {
            g_info("Cannot add activation for config '%s': "
                   "configuration scheduled to be destroyed",
                   activation_msg->config);
        }
    }
    else {
        g_info("Cannot add activation for config '%s': "
               "configuration does not exist", activation_msg->config);
    }
    return NULL;
}

request_cb_fn_t
wcbe_mc_remove_activation_command(pipeline_t *pipeline, wcbe_msg_t **msg,
                                  gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_REMOVE_ACTIVATION_MSG, NULL);

    wcbe_remove_activation_msg_t *activation_msg =
        (wcbe_remove_activation_msg_t *)*msg;
    configuration_t *config = find_config(pipeline, activation_msg->config);
    if (config != NULL) {
        if (pipeline->current.pipeline != NULL) {
            lf_widar_time_t time = activation_msg->time;
            wcbe_remove_activation(pipeline, config, &time);
        }
        else {
            g_debug("Failed to remove activation for config '%s': "
                    "input stage pipeline does not exist",
                    activation_msg->config);
        }
    }
    else {
        g_info("Cannot remove activation for config '%s': "
               "configuration does not exist", activation_msg->config);
    }
    return NULL;
}

request_cb_fn_t
wcbe_mc_activation_list_query(pipeline_t *pipeline, wcbe_msg_t **msg,
                              gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_ACTIVATION_LIST_MSG, NULL);

    wcbe_activation_list_msg_t **list_msg = (wcbe_activation_list_msg_t **)msg;
    GString *activations = g_string_new("");
    // activations
    void activation_name(gpointer data, gpointer user_data) {
        activation_t *act = data;
        gchar left, right;
        if (act->deactivated) {
            left = '[';
            right = ']';
        }
        else if (act->in_use) {
            left = '<';
            right = '>';
        }
        else {
            left = '(';
            right = ')';
        }
        const gchar *fmt =
            ((activations->len == 0) ? "%c%s:%d:%u:%u%c": "\t%c%s:%d:%u:%u%c");
        g_string_append_printf(activations, fmt, left, act->config->name,
                               act->time.epoch, act->time.seconds,
                               act->time.clock_cycles, right);
    }
    g_slist_foreach(pipeline->activations, activation_name, NULL);

    (*list_msg)->list_len = activations->len + 1;
    *msgsize = sizeof(wcbe_activation_list_msg_t) + (*list_msg)->list_len;
    *list_msg = g_realloc(*list_msg, *msgsize);
    g_strlcpy((*list_msg)->list, activations->str, (*list_msg)->list_len);
    g_string_free(activations, TRUE);
    return NULL;
}

request_cb_fn_t
wcbe_mc_config_destroy_command(pipeline_t *pipeline, wcbe_msg_t **msg,
                               gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_CONFIG_DESTROY_MSG, NULL);

    wcbe_config_destroy_msg_t *destroy_msg = (wcbe_config_destroy_msg_t *)*msg;
    configuration_t *config = find_config(pipeline, destroy_msg->name);
    if (config != NULL) {
        if (pipeline->current.pipeline != NULL) {
            wcbe_destroy_config(pipeline, config);
        }
        else {
            g_debug("Cannot destroy config '%s': "
                    "input stage pipeline does not exist",
                    destroy_msg->name);
        }
    }
    else {
        g_debug("Cannot destroy config '%s': config does not exist",
                destroy_msg->name);
    }
    return NULL;
}

request_cb_fn_t
wcbe_mc_config_list_query(pipeline_t *pipeline, wcbe_msg_t **msg,
                          gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_CONFIG_LIST_MSG, NULL);

    wcbe_config_list_msg_t **config_list_msg = (wcbe_config_list_msg_t **)msg;
    GString *configs = g_string_new("");
    // configs in active pipeline
    void config_name(gpointer key, gpointer value, gpointer user_data) {
        const gchar *name = key;
        const configuration_t *cfg = value;
        const gchar *fmt;
        if (cfg->destroy) {
            fmt = ((configs->len == 0) ? "[%s]" : "\t[%s]");
        }
        else {
            fmt = ((configs->len == 0) ? "(%s)" : "\t(%s)");
        }
        g_string_append_printf(configs, fmt, name);
    }
    g_static_rec_mutex_lock(&(pipeline->configs_mtx));
    g_hash_table_foreach(pipeline->configs, config_name, NULL);
    g_static_rec_mutex_unlock(&(pipeline->configs_mtx));

    (*config_list_msg)->list_len = configs->len + 1;
    *msgsize = sizeof(wcbe_config_list_msg_t) + (*config_list_msg)->list_len;
    *config_list_msg = g_realloc(*config_list_msg, *msgsize);
    g_strlcpy((*config_list_msg)->list, configs->str,
              (*config_list_msg)->list_len);
    g_string_free(configs, TRUE);
    return NULL;
}

request_cb_fn_t
wcbe_mc_src_command(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_SRC_MSG, NULL);

    wcbe_src_msg_t *src_msg = (wcbe_src_msg_t *)*msg;
    src_msg->name[sizeof(src_msg->name) - 1] = '\0';

    if (pipeline->src == NULL
        || strcmp(src_msg->name, pipeline->src) != 0) {
        wcbe_flow_t flow = pipeline->flow;
        wcbe_destroy_pipeline(pipeline);
        pipeline->src = g_strdup(src_msg->name);
        unlink(pipeline->srcpath);
        if (pipeline->src[0] != '\0') {
            FILE *f = fopen(pipeline->srcpath, "w");
            if (f != NULL) {
                gint rc = fputs(pipeline->src, f);
                if (rc == EOF) {
                    g_warning("Failed to write '%s' to file %s: %s",
                              pipeline->src, pipeline->srcpath,
                              g_strerror(errno));
                }
                rc = fclose(f);
                if (rc != 0) {
                    g_warning("Failed to close file %s: %s",
                              pipeline->srcpath, g_strerror(errno));
                }
            }
            else {
                g_warning("Failed to open file %s: %s", pipeline->srcpath,
                          g_strerror(errno));
            }
            wcbe_create_pipeline(pipeline);
            if (pipeline->current.pipeline != NULL && flow == WCBE_FLOW_ON) {
                wcbe_start_pipeline(pipeline, TRUE, 0);
            }
        }
    }
    return NULL;
}

request_cb_fn_t
wcbe_mc_src_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_SRC_MSG, NULL);

    wcbe_src_msg_t *src_msg = (wcbe_src_msg_t *)*msg;
    const gchar *name = (pipeline->src ? pipeline->src : "");
    g_strlcpy(src_msg->name, name, sizeof(src_msg->name));
    return NULL;
}

request_cb_fn_t
wcbe_mc_ts0_command(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_TS0_MSG, NULL);

    wcbe_ts0_msg_t *ts0_msg = (wcbe_ts0_msg_t *)*msg;
    if (pipeline->ts0 == NULL
        || strcmp(ts0_msg->ts0, pipeline->ts0) != 0) {
        if (pipeline->ts0 != NULL) {
            g_free(pipeline->ts0);
        }
        gchar *newbase = NULL;
        if (pipeline->current.pipeline != NULL) {
            GstElement *src = gst_bin_get_by_name(
                GST_BIN(pipeline->current.pipeline), "src");
            if (src != NULL) {
                g_object_set(G_OBJECT(src), "timestamp-base", ts0_msg->ts0,
                             NULL);
                g_object_get(G_OBJECT(src), "timestamp-base", &newbase, NULL);
                gst_object_unref(GST_OBJECT(src));
            }
            else {
                g_debug("Failed to get 'src' element in input stage");
            }
        }
        else {
            newbase = wcbe_parse_ts0(ts0_msg->ts0);
        }
        pipeline->ts0 = newbase;
    }
    return NULL;
}

request_cb_fn_t
wcbe_mc_ts0_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_TS0_MSG, NULL);

    wcbe_ts0_msg_t *ts0_msg = (wcbe_ts0_msg_t *)*msg;
    if (pipeline->ts0 != NULL) {
        g_strlcpy(ts0_msg->ts0, pipeline->ts0, WCBE_MAX_NAME_LEN);
    }
    else {
        ts0_msg->ts0[0] = '\0';
    }
    return NULL;
}

request_cb_fn_t
wcbe_mc_framedump_command(pipeline_t *pipeline, wcbe_msg_t **msg,
                          gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_FRAMEDUMP_MSG, NULL);

    wcbe_framedump_msg_t *framedump_msg = (wcbe_framedump_msg_t *)*msg;
    if (framedump_msg->format != pipeline->framedump) {
        if (pipeline->current.pipeline != NULL) {
            GstElement *framedump = gst_bin_get_by_name(
                GST_BIN(pipeline->current.pipeline), "framedump");
            g_assert(framedump != NULL);
            if (framedump_msg->format != WCBE_DUMP_NONE) {
                WcbeFramedumpFormat fdformat =
                    ((framedump_msg->format == WCBE_DUMP_BINARY)
                     ? WCBE_FRAMEDUMP_BINARY
                     : WCBE_FRAMEDUMP_ASCII);
                g_object_set(G_OBJECT(framedump), "do-output", FALSE, NULL);
                g_object_set(G_OBJECT(framedump), "format", fdformat, NULL);
                g_object_set(G_OBJECT(framedump), "do-output", TRUE, NULL);
            }
            else {
                g_object_set(G_OBJECT(framedump), "do-output", FALSE, NULL);
            }
            gst_object_unref(GST_OBJECT(framedump));
        }
        pipeline->framedump = framedump_msg->format;
    }
    return NULL;
}

request_cb_fn_t
wcbe_mc_framedump_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_FRAMEDUMP_MSG, NULL);

    wcbe_framedump_msg_t *framedump_msg = (wcbe_framedump_msg_t *)*msg;
    framedump_msg->format = pipeline->framedump;
    return NULL;
}

request_cb_fn_t
wcbe_mc_flow_command(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_FLOW_MSG, NULL);

    wcbe_flow_msg_t *flow_msg = (wcbe_flow_msg_t *)*msg;
    if (flow_msg->state != pipeline->flow) {
        if (flow_msg->state == WCBE_FLOW_ON) {
            if (pipeline->current.pipeline != NULL) {
                wcbe_start_pipeline(pipeline, TRUE, 0);
            }
        }
        else { // flow_msg->state == WCBE_FLOW_OFF
            if (pipeline->current.pipeline != NULL) {
                wcbe_stop_pipeline(pipeline, TRUE, 0);
            }
        }
        // start/stop_pipeline update pipeline->flow automatically
    }
    return NULL;
}

request_cb_fn_t
wcbe_mc_flow_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_FLOW_MSG, NULL);

    wcbe_flow_msg_t *flow_msg = (wcbe_flow_msg_t *)*msg;
    flow_msg->state = pipeline->flow;
    return NULL;
}

request_cb_fn_t
wcbe_mc_rate_command(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_RATE_MSG, NULL);

    wcbe_rate_msg_t *rate_msg = (wcbe_rate_msg_t *)*msg;
    if (rate_msg->fps != pipeline->fps) {
        if (pipeline->current.pipeline != NULL) {
            GstElement *src = gst_bin_get_by_name(
                GST_BIN(pipeline->current.pipeline), "src");
            g_assert(src != NULL);
            g_object_set(G_OBJECT(src), "fps", rate_msg->fps, NULL);
            gst_object_unref(GST_OBJECT(src));
        }
        pipeline->fps = rate_msg->fps;
    }
    return NULL;
}

request_cb_fn_t
wcbe_mc_rate_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_RATE_MSG, NULL);

    wcbe_rate_msg_t *rate_msg = (wcbe_rate_msg_t *)*msg;
    rate_msg->fps = pipeline->fps;
    return NULL;
}

request_cb_fn_t
wcbe_mc_name_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_NAME_MSG, NULL);

    wcbe_name_msg_t *name_msg = (wcbe_name_msg_t *)*msg;
    g_assert(pipeline->name != NULL);
    g_strlcpy(name_msg->value, pipeline->name, sizeof(name_msg->value));
    return NULL;
}

request_cb_fn_t
wcbe_mc_quit_command(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_debug("wcbe_mc_quit_command");
    g_return_val_if_fail((*msg)->type == WCBE_QUIT_MSG, NULL);
    wcbe_destroy_pipeline(pipeline);
    pipeline->state = WCBE_STATE_FINALIZING;
    return NULL;
}

request_cb_fn_t
wcbe_mc_pipeline_list_query(pipeline_t *pipeline, wcbe_msg_t **msg,
                            gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_PIPELINE_LIST_MSG, NULL);

    wcbe_pipeline_list_msg_t **p_msg = (wcbe_pipeline_list_msg_t **)msg;
    GString *pipelines = g_string_new("");
    g_static_rec_mutex_lock(&(pipeline->configs_mtx));
    configuration_t *config =
        g_hash_table_lookup(pipeline->configs, (*p_msg)->configuration);
    g_static_rec_mutex_unlock(&(pipeline->configs_mtx));
    if (config != NULL) {
        g_string_assign(pipelines, (gchar *)input_stage_pipeline_name);
        void pipeline_name(gpointer data, gpointer user_data) {
            ls_pipeline_t *p = &(((component_pipeline_t *)data)->lsp);
            gchar *pname = strchr(p->name, G_DIR_SEPARATOR);
            g_assert(pname != NULL);
            pname += 1;
            g_string_append_printf(pipelines, "\t%s", pname);
        }
        g_slist_foreach(config->ls_pipelines, pipeline_name, NULL);
    }
    (*p_msg)->list_len = pipelines->len + 1;
    *msgsize = sizeof(wcbe_pipeline_list_msg_t) + (*p_msg)->list_len;
    *p_msg = g_realloc(*p_msg, *msgsize);
    g_strlcpy((*p_msg)->list, pipelines->str, (*p_msg)->list_len);
    g_string_free(pipelines, TRUE);
    return NULL;
}

request_cb_fn_t
wcbe_mc_element_list_query(pipeline_t *pipeline, wcbe_msg_t **msg,
                           gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_ELEMENT_LIST_MSG, NULL);

    request_cb_fn_t result = NULL;
    wcbe_element_list_msg_t **e_msg = (wcbe_element_list_msg_t **)msg;
    gchar **name_parts = g_strsplit((*e_msg)->pipeline, G_DIR_SEPARATOR_S, 3);
    if (g_strv_length(name_parts) == 2) {
        const gchar *config_name = name_parts[0];
        const gchar *pipeline_name = name_parts[1];
        if (strcmp(pipeline_name, input_stage_pipeline_name) == 0) {
            // elements in input stage pipeline
            if (pipeline->current.pipeline != NULL) {
                GString *elements = g_string_new("");
                void element_name(gpointer data, gpointer user_data) {
                    const gchar *name = GST_ELEMENT_NAME(GST_ELEMENT(data));
                    const gchar *fmt = ((elements->len == 0) ? "%s" : "\t%s");
                    g_string_append_printf(elements, fmt, name);
                }
                GstIterator *children = gst_bin_iterate_recurse(
                    GST_BIN(pipeline->current.pipeline));
                GstIteratorResult iter_result =
                    gst_iterator_foreach(children, element_name, NULL);
                gst_iterator_free(children);
                if (iter_result == GST_ITERATOR_ERROR) {
                    g_string_truncate(elements, 0);
                }
                (*e_msg)->list_len = elements->len + 1;
                *msgsize =
                    sizeof(wcbe_element_list_msg_t) + (*e_msg)->list_len;
                *e_msg = g_realloc(*e_msg, *msgsize);
                g_strlcpy((*e_msg)->list, elements->str, (*e_msg)->list_len);
                g_string_free(elements, TRUE);
            }
        }

        else {
            // elements (perhaps) in a lag processing stage pipeline
            GstStructure *structure =
                gst_structure_empty_new(ELEMENT_LIST_MSG_NAME);
            g_assert(structure != NULL);
            result = ls_message(pipeline, config_name, pipeline_name, *msg,
                                structure, element_list_query);
        }
    }
    g_strfreev(name_parts);
    return result;
}

request_cb_fn_t
wcbe_mc_property_list_query(pipeline_t *pipeline, wcbe_msg_t **msg,
                            gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_PROPERTY_LIST_MSG, NULL);

    request_cb_fn_t result = NULL;
    wcbe_property_list_msg_t **p_msg = (wcbe_property_list_msg_t **)msg;
    gchar **name_parts = g_strsplit((*p_msg)->element, G_DIR_SEPARATOR_S, 4);
    if (g_strv_length(name_parts) == 3) {
        const gchar *config_name = name_parts[0];
        const gchar *pipeline_name = name_parts[1];
        const gchar *element_name = name_parts[2];
        if (strcmp(pipeline_name, input_stage_pipeline_name) == 0) {
            // input stage element
            if (pipeline->current.pipeline != NULL) {
                GString *properties = g_string_new("");
                GstElement *element = gst_bin_get_by_name(
                    GST_BIN(pipeline->current.pipeline), element_name);
                if (element != NULL) {
                    guint num_properties;
                    GParamSpec **params = g_object_class_list_properties(
                        G_OBJECT_GET_CLASS(G_OBJECT(element)),
                        &num_properties);
#define PARAM_RW_STR(p) (((p)->flags & G_PARAM_READABLE)    \
                         ? (((p)->flags & G_PARAM_WRITABLE) \
                            ? "rw" : "r")                   \
                         : (((p)->flags & G_PARAM_WRITABLE) \
                            ? "w" : ""))
                    guint i;
                    for (i = 0; i < num_properties; ++i) {
                        GParamSpec *param = params[i];
                        const gchar *fmt =
                            ((properties->len == 0)
                             ? "%s<%s,%s>" : "\t%s<%s,%s>");
                        g_string_append_printf(properties, fmt, param->name,
                                               g_type_name(param->value_type),
                                               PARAM_RW_STR(param));
                    }
#undef PARAM_RW_STR
                    g_free(params);
                    gst_object_unref(GST_OBJECT(element));
                }
                (*p_msg)->list_len = properties->len + 1;
                *msgsize =
                    sizeof(wcbe_property_list_msg_t) + (*p_msg)->list_len;
                *p_msg = g_realloc(*p_msg, *msgsize);
                g_strlcpy((*p_msg)->list, properties->str, (*p_msg)->list_len);
                g_string_free(properties, TRUE);
            }
        }
        else {
            // element (perhaps) in a lag processing stage pipeline
            GstStructure *structure =
                gst_structure_new(PROPERTY_LIST_MSG_NAME,
                                  "element", G_TYPE_STRING, element_name,
                                  NULL);
            g_assert(structure != NULL);
            result = ls_message(pipeline, config_name, pipeline_name, *msg,
                                structure, property_list_query);
        }
    }
    g_strfreev(name_parts);

    return result;
}

request_cb_fn_t
wcbe_mc_property_command(pipeline_t *pipeline, wcbe_msg_t **msg,
                         gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_PROPERTY_MSG, NULL);

    wcbe_property_msg_t *p_msg = (wcbe_property_msg_t *)*msg;
    gchar **name_parts = g_strsplit(p_msg->property, G_DIR_SEPARATOR_S, 5);
    if (g_strv_length(name_parts) == 4) {
        const gchar *config_name = name_parts[0];
        const gchar *pipeline_name = name_parts[1];
        const gchar *element_name = name_parts[2];
        const gchar *property_name = name_parts[3];
        if (strcmp(pipeline_name, input_stage_pipeline_name) == 0) {
            // input stage element property
            if (pipeline->current.pipeline != NULL) {
                GstElement *element = gst_bin_get_by_name(
                    GST_BIN(pipeline->current.pipeline), element_name);
                if (element != NULL) {
                    GParamSpec *param = g_object_class_find_property(
                        G_OBJECT_GET_CLASS(G_OBJECT(element)), property_name);
                    GType valtype = g_type_from_name(p_msg->type);
                    if (param != NULL && ((param)->flags & G_PARAM_WRITABLE)
                        && valtype != 0) {
                        switch (valtype) {
                          case G_TYPE_CHAR:
                            g_object_set(G_OBJECT(element), property_name,
                                         *((gint8 *)(p_msg->value)), NULL);
                            break;
                          case G_TYPE_UCHAR:
                            g_object_set(G_OBJECT(element), property_name,
                                         *((guchar *)(p_msg->value)), NULL);
                            break;
                          case G_TYPE_BOOLEAN:
                            g_object_set(G_OBJECT(element), property_name,
                                         *((gint *)(p_msg->value)), NULL);
                            break;
                          case G_TYPE_INT:
                            g_object_set(G_OBJECT(element), property_name,
                                         *((gint *)(p_msg->value)), NULL);
                            break;
                          case G_TYPE_UINT:
                            g_object_set(G_OBJECT(element), property_name,
                                         *((guint *)(p_msg->value)), NULL);
                            break;
                          case G_TYPE_LONG:
                            g_object_set(G_OBJECT(element), property_name,
                                         *((glong *)(p_msg->value)), NULL);
                            break;
                          case G_TYPE_ULONG:
                            g_object_set(G_OBJECT(element), property_name,
                                         *((gulong *)(p_msg->value)), NULL);
                            break;
                          case G_TYPE_INT64:
                            g_object_set(G_OBJECT(element), property_name,
                                         *((gint64 *)(p_msg->value)), NULL);
                            break;
                          case G_TYPE_UINT64:
                            g_object_set(G_OBJECT(element), property_name,
                                         *((guint64 *)(p_msg->value)), NULL);
                            break;
                          case G_TYPE_FLOAT:
                            g_object_set(G_OBJECT(element), property_name,
                                         *((gfloat *)(p_msg->value)), NULL);
                            break;
                          case G_TYPE_DOUBLE:
                            g_object_set(G_OBJECT(element), property_name,
                                         *((gdouble *)(p_msg->value)), NULL);
                            break;
                          case G_TYPE_STRING:
                            g_object_set(G_OBJECT(element), property_name,
                                         p_msg->value, NULL);
                            break;
                          default:
                            if (g_type_is_a(valtype, G_TYPE_ENUM)) {
                                g_object_set(G_OBJECT(element), property_name,
                                             *((gint *)(p_msg->value)), NULL);
                            }
                            else if (g_type_is_a(valtype, G_TYPE_FLAGS)) {
                                g_object_set(G_OBJECT(element), property_name,
                                             *((guint *)(p_msg->value)), NULL);
                            }
                            break;
                        }
                    }
                    gst_object_unref(GST_OBJECT(element));
                }
            }
        }
        else {
            // property of element (perhaps) in lag processing stage
            gchar *propname = g_strjoin(
                G_DIR_SEPARATOR_S, element_name, property_name, NULL);
            GstStructure *structure =
                gst_structure_new(SET_PROPERTY_MSG_NAME,
                                  "property", G_TYPE_STRING, propname,
                                  NULL);
            g_assert(structure != NULL);
            GType valtype = g_type_from_name(p_msg->type);
            switch (valtype) {
              case G_TYPE_CHAR:
                gst_structure_set(structure, "value", G_TYPE_CHAR,
                                  *(gint8 *)(p_msg->value), NULL);
                break;
              case G_TYPE_UCHAR:
                gst_structure_set(structure, "value", G_TYPE_UCHAR,
                                  *(guchar *)(p_msg->value), NULL);
                break;
              case G_TYPE_BOOLEAN:
                gst_structure_set(structure, "value", G_TYPE_BOOLEAN,
                                  *(gboolean *)(p_msg->value), NULL);
                break;
              case G_TYPE_INT:
                gst_structure_set(structure, "value", G_TYPE_INT,
                                  *((gint *)(p_msg->value)), NULL);
                break;
              case G_TYPE_UINT:
                gst_structure_set(structure, "value", G_TYPE_UINT,
                                  *((guint *)(p_msg->value)), NULL);
                break;
              case G_TYPE_LONG:
                gst_structure_set(structure, "value", G_TYPE_LONG,
                                  *((glong *)(p_msg->value)), NULL);
                break;
              case G_TYPE_ULONG:
                gst_structure_set(structure, "value", G_TYPE_ULONG,
                                  *((gulong *)(p_msg->value)), NULL);
                break;
              case G_TYPE_INT64:
                gst_structure_set(structure, "value", G_TYPE_INT64,
                                  *((gint64 *)(p_msg->value)), NULL);
                break;
              case G_TYPE_UINT64:
                gst_structure_set(structure, "value", G_TYPE_UINT64,
                                  *((guint64 *)(p_msg->value)), NULL);
                break;
              case G_TYPE_FLOAT:
                gst_structure_set(structure, "value", G_TYPE_FLOAT,
                                  *((gfloat *)(p_msg->value)), NULL);
                break;
              case G_TYPE_DOUBLE:
                gst_structure_set(structure, "value", G_TYPE_DOUBLE,
                                  *((gdouble *)(p_msg->value)), NULL);
                break;
              case G_TYPE_STRING:
                gst_structure_set(structure, "value", G_TYPE_STRING,
                                  p_msg->value, NULL);
                break;
              default:
                if (g_type_is_a(valtype, G_TYPE_ENUM)) {
                    gst_structure_set(structure, "value", valtype,
                                      *((gint *)(p_msg->value)), NULL);
                }
                else if (g_type_is_a(valtype, G_TYPE_FLAGS)) {
                    gst_structure_set(structure, "value", valtype,
                                      *((guint *)(p_msg->value)), NULL);
                }
                break;
            }
            ls_message(pipeline, config_name, pipeline_name, *msg, structure,
                       NULL);
            g_free(propname);
        }
    }
    g_strfreev(name_parts);
    return NULL;
}

request_cb_fn_t
wcbe_mc_property_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_PROPERTY_MSG, NULL);

    request_cb_fn_t result = NULL;
    wcbe_property_msg_t **p_msg = (wcbe_property_msg_t **)msg;
    (*p_msg)->value_len = 0;
    gchar **name_parts = g_strsplit((*p_msg)->property, G_DIR_SEPARATOR_S, 5);
    if (g_strv_length(name_parts) == 4) {
        const gchar *config_name = name_parts[0];
        const gchar *pipeline_name = name_parts[1];
        const gchar *element_name = name_parts[2];
        const gchar *property_name = name_parts[3];
        if (strcmp(pipeline_name, input_stage_pipeline_name) == 0) {
            // input stage element property
            if (pipeline->current.pipeline != NULL) {
                GstElement *element = gst_bin_get_by_name(
                    GST_BIN(pipeline->current.pipeline), element_name);
                if (element != NULL) {
                    GParamSpec *param = g_object_class_find_property(
                        G_OBJECT_GET_CLASS(G_OBJECT(element)), property_name);
                    if (param != NULL && ((param)->flags & G_PARAM_READABLE)) {
                        void prep_msg(GType type, gsize value_len) {
                            *msgsize = sizeof(wcbe_property_msg_t) + value_len;
                            *p_msg = g_realloc(*p_msg, *msgsize);
                            (*p_msg)->value_len = value_len;
                            g_strlcpy((*p_msg)->type, g_type_name(type),
                                      sizeof((*p_msg)->type));
                        }
                        void get_simple(void) {
                            g_object_get(G_OBJECT(element), property_name,
                                         (*p_msg)->value, NULL);
                        }
                        switch (param->value_type) {
                          case G_TYPE_CHAR:
                            // gint8 is the correct type according to
                            // GObject docs
                            prep_msg(G_TYPE_CHAR, sizeof(gint8));
                            get_simple();
                            break;
                          case G_TYPE_UCHAR:
                            prep_msg(G_TYPE_UCHAR, sizeof(guchar));
                            get_simple();
                            break;
                          case G_TYPE_BOOLEAN:
                            prep_msg(G_TYPE_BOOLEAN, sizeof(gboolean));
                            get_simple();
                            break;
                          case G_TYPE_INT:
                            prep_msg(G_TYPE_INT, sizeof(gint));
                            get_simple();
                            break;
                          case G_TYPE_UINT:
                            prep_msg(G_TYPE_UINT, sizeof(guint));
                            get_simple();
                            break;
                          case G_TYPE_LONG:
                            prep_msg(G_TYPE_LONG, sizeof(glong));
                            get_simple();
                            break;
                          case G_TYPE_ULONG:
                            prep_msg(G_TYPE_ULONG, sizeof(gulong));
                            get_simple();
                            break;
                          case G_TYPE_INT64:
                            prep_msg(G_TYPE_INT64, sizeof(gint64));
                            get_simple();
                            break;
                          case G_TYPE_UINT64:
                            prep_msg(G_TYPE_UINT64, sizeof(guint64));
                            get_simple();
                            break;
                          case G_TYPE_FLOAT:
                            prep_msg(G_TYPE_FLOAT, sizeof(gfloat));
                            get_simple();
                            break;
                          case G_TYPE_DOUBLE:
                            prep_msg(G_TYPE_DOUBLE, sizeof(gdouble));
                            get_simple();
                            break;
                          case G_TYPE_STRING: {
                              gchar *str;
                              g_object_get(G_OBJECT(element), property_name,
                                           &str, NULL);
                              guint len = strlen(str) + 1;
                              prep_msg(G_TYPE_STRING, len);
                              g_strlcpy((*p_msg)->value, str, len);
                              break;
                          }
                          default:
                            if (g_type_is_a(param->value_type, G_TYPE_ENUM)) {
                                prep_msg(param->value_type, sizeof(gint));
                                get_simple();
                            }
                            else if (g_type_is_a(param->value_type,
                                                 G_TYPE_FLAGS)) {
                                prep_msg(param->value_type, sizeof(guint));
                                get_simple();
                            }
                            break;
                        }
                    }
                    gst_object_unref(GST_OBJECT(element));
                }
            }
        }
        else {
            // property of element (perhaps) in lag processing stage
            gchar *propname = g_strjoin(
                G_DIR_SEPARATOR_S, element_name, property_name, NULL);
            GstStructure *structure =
                gst_structure_new(GET_PROPERTY_MSG_NAME,
                                  "property", G_TYPE_STRING, propname,
                                  NULL);
            g_assert(structure != NULL);
            result = ls_message(pipeline, config_name, pipeline_name, *msg,
                                structure, property_query);
            g_free(propname);
        }
    }
    g_strfreev(name_parts);
    return result;
}

request_cb_fn_t
wcbe_mc_pad_list_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_PAD_LIST_MSG, NULL);

    request_cb_fn_t result = NULL;
    wcbe_pad_list_msg_t **p_msg = (wcbe_pad_list_msg_t **)msg;
    gchar **name_parts = g_strsplit((*p_msg)->element, G_DIR_SEPARATOR_S, 4);
    if (g_strv_length(name_parts) == 3) {
        const gchar *config_name = name_parts[0];
        const gchar *pipeline_name = name_parts[1];
        const gchar *element_name = name_parts[2];
        if (strcmp(pipeline_name, input_stage_pipeline_name) == 0) {
            // input stage element pad
            if (pipeline->current.pipeline != NULL) {
                GString *pads = g_string_new("");
                GstElement *element = gst_bin_get_by_name(
                    GST_BIN(pipeline->current.pipeline), element_name);
                if (element != NULL) {
                    GstIterator *pads_iter = gst_element_iterate_pads(element);
                    void pad_name(gpointer data, gpointer user_data) {
                        GstPad *pad = GST_PAD(data);
                        gchar *name = gst_pad_get_name(pad);
                        const gchar *fmt = ((pads->len == 0) ? "%s" : "\t%s");
                        g_string_append_printf(pads, fmt, name);
                        g_free(name);
                    }
                    GstIteratorResult iter_result =
                        gst_iterator_foreach(pads_iter, pad_name, NULL);
                    gst_iterator_free(pads_iter);
                    if (iter_result == GST_ITERATOR_ERROR) {
                        g_string_truncate(pads, 0);
                    }
                    gst_object_unref(GST_OBJECT(element));
                }
                (*p_msg)->list_len = pads->len + 1;
                *msgsize = sizeof(wcbe_pad_list_msg_t) + (*p_msg)->list_len;
                *p_msg = g_realloc(*p_msg, *msgsize);
                g_strlcpy((*p_msg)->list, pads->str, (*p_msg)->list_len);
                g_string_free(pads, TRUE);
            }
        }
        else {
            GstStructure *structure =
                gst_structure_new(PAD_LIST_MSG_NAME,
                                  "element", G_TYPE_STRING, element_name,
                                  NULL);
            g_assert(structure != NULL);
            result = ls_message(pipeline, config_name, pipeline_name, *msg,
                                structure, pad_list_query);
        }
    }
    g_strfreev(name_parts);

    return result;
}

request_cb_fn_t
wcbe_mc_pad_caps_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_PAD_CAPS_MSG, NULL);

    request_cb_fn_t result = NULL;
    wcbe_pad_caps_msg_t *p_msg = (wcbe_pad_caps_msg_t *)*msg;
    p_msg->caps[0] = '\0';
    gchar **name_parts = g_strsplit(p_msg->pad, G_DIR_SEPARATOR_S, 5);
    if (g_strv_length(name_parts) == 4) {
        const gchar *config_name = name_parts[0];
        const gchar *pipeline_name = name_parts[1];
        const gchar *element_name = name_parts[2];
        const gchar *pad_name = name_parts[3];
        if (strcmp(pipeline_name, input_stage_pipeline_name) == 0) {
            // input stage element property
            if (pipeline->current.pipeline != NULL) {
                GstElement *element = gst_bin_get_by_name(
                    GST_BIN(pipeline->current.pipeline), element_name);
                if (element != NULL) {
                    GstPad *pad =
                        gst_element_get_static_pad(element, pad_name);
                    if (pad != NULL) {
                        GstCaps *caps = gst_pad_get_negotiated_caps(pad);
                        if (caps == NULL) {
                            caps = gst_pad_get_caps(pad);
                        }
                        if (caps != NULL) {
                            gchar *str = gst_caps_to_string(caps);
                            g_strlcpy(p_msg->caps, str, sizeof(p_msg->caps));
                            g_free(str);
                            gst_caps_unref(caps);
                        }
                        gst_object_unref(GST_OBJECT(pad));
                    }
                    gst_object_unref(GST_OBJECT(element));
                }
            }
        }
        else {
        // element (perhaps) in lag processing stage
        gchar *pname =
            g_strjoin(G_DIR_SEPARATOR_S, element_name, pad_name, NULL);
        GstStructure *structure =
            gst_structure_new(PAD_CAPS_MSG_NAME,
                              "pad", G_TYPE_STRING, pname,
                              NULL);
        g_assert(structure != NULL);
        result = ls_message(pipeline, config_name, pipeline_name, *msg,
                            structure, pad_caps_query);
        g_free(pname);
        }
    }
    g_strfreev(name_parts);
    return result;
}

request_cb_fn_t
wcbe_mc_pad_peer_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_PAD_PEER_MSG, NULL);

    request_cb_fn_t result = NULL;
    wcbe_pad_peer_msg_t *p_msg = (wcbe_pad_peer_msg_t *)*msg;
    p_msg->peer[0] = '\0';
    gchar **name_parts = g_strsplit(p_msg->pad, G_DIR_SEPARATOR_S, 5);
    if (g_strv_length(name_parts) == 4) {
        const gchar *config_name = name_parts[0];
        const gchar *pipeline_name = name_parts[1];
        const gchar *element_name = name_parts[2];
        const gchar *pad_name = name_parts[3];
        if (strcmp(pipeline_name, input_stage_pipeline_name) == 0) {
            // input stage element property
            if (pipeline->current.pipeline != NULL) {
                GstElement *element = gst_bin_get_by_name(
                    GST_BIN(pipeline->current.pipeline), element_name);
                if (element != NULL) {
                    GstPad *pad =
                        gst_element_get_static_pad(element, pad_name);
                    if (pad != NULL)
                        if (gst_pad_is_linked(pad)) {
                            GstPad *peer = gst_pad_get_peer(pad);
                            g_assert(peer != NULL);
                            GstObject *parent = gst_pad_get_parent(peer);
                            g_assert(parent != NULL);
                            g_snprintf(p_msg->peer, sizeof(p_msg->peer),
                                       "%s/%s", GST_OBJECT_NAME(parent),
                                       GST_OBJECT_NAME(peer));
                            gst_object_unref(parent);
                            gst_object_unref(GST_OBJECT(peer));
                        }
                        gst_object_unref(GST_OBJECT(pad));
                    }
                gst_object_unref(GST_OBJECT(element));
            }
        }
        else {
            // element (perhaps) in lag processing stage
            gchar *pname =
                g_strjoin(G_DIR_SEPARATOR_S, element_name, pad_name, NULL);
            GstStructure *structure =
                gst_structure_new(PAD_PEER_MSG_NAME,
                                  "pad", G_TYPE_STRING, pname,
                                  NULL);
            g_assert(structure != NULL);
            result = ls_message(pipeline, config_name, pipeline_name, *msg,
                                structure, pad_peer_query);
            g_free(pname);
        }
    }
    g_strfreev(name_parts);
    return result;
}

request_cb_fn_t
wcbe_mc_aux_output_dir_command(pipeline_t *pipeline, wcbe_msg_t **msg,
                               gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_AUX_OUTPUT_DIR_MSG, NULL);
    wcbe_aux_output_dir_msg_t *a_msg = (wcbe_aux_output_dir_msg_t *)*msg;
    if (pipeline->auxdir != NULL) {
        g_free(pipeline->auxdir);
    }
    pipeline->auxdir = g_strdup(a_msg->name);
    if (pipeline->current.pipeline != NULL) {
        GstElement *framedump = gst_bin_get_by_name(
            GST_BIN(pipeline->current.pipeline), framedump_element_name);
        if (framedump != NULL) {
            g_object_set(G_OBJECT(framedump), "output-dir", pipeline->auxdir,
                         NULL);
            gst_object_unref(GST_OBJECT(framedump));
        }
        else {
            g_debug("Failed to get '%s' element from input stage",
                    framedump_element_name);
        }
        GstElement *blsdump = gst_bin_get_by_name(
            GST_BIN(pipeline->current.pipeline), blsdump_element_name);
        if (blsdump != NULL) {
            g_object_set(G_OBJECT(blsdump), "output-dir", pipeline->auxdir,
                         NULL);
            gst_object_unref(GST_OBJECT(blsdump));
        }
        else {
            g_debug("Failed to get '%s' element from input stage",
                    blsdump_element_name);
        }
    }
    return NULL;
}

request_cb_fn_t
wcbe_mc_aux_output_dir_query(pipeline_t *pipeline, wcbe_msg_t **msg,
                             gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_AUX_OUTPUT_DIR_MSG, NULL);
    wcbe_aux_output_dir_msg_t *a_msg = (wcbe_aux_output_dir_msg_t *)*msg;
    g_strlcpy(a_msg->name, pipeline->auxdir, sizeof(a_msg->name));
    return NULL;
}

static request_cb_fn_t
ls_message(pipeline_t *pipeline, const gchar *config_name,
           const gchar *ls_pipeline_name, const wcbe_msg_t *msg,
           GstStructure *structure, request_cb_fn_t cb)
{
    request_cb_fn_t result;
    // elements (perhaps) in a lag processing stage pipeline
    if (cb != NULL) {
        gchar *id = g_strdup_printf("%p", msg);
        gst_structure_set(structure,
                          "id", G_TYPE_STRING, id,
                          NULL);
        g_free(id);
    }
    if (forward_message(pipeline, config_name, ls_pipeline_name, structure)) {
        result = cb;
    }
    else {
        gst_structure_free(structure);
        result = NULL;
    }
    return result;
}

static gboolean
forward_message(pipeline_t *pipeline, const gchar *config_name,
                const gchar *ls_pipeline_name, GstStructure *structure)
{
    gboolean result = FALSE;
    g_static_rec_mutex_lock(&(pipeline->configs_mtx));
    configuration_t *config =
        g_hash_table_lookup(pipeline->configs, config_name);
    g_static_rec_mutex_unlock(&(pipeline->configs_mtx));
    if (config != NULL) {
        gchar *full_pipeline_name =
            g_strjoin(G_DIR_SEPARATOR_S, config_name, ls_pipeline_name, NULL);
        component_pipeline_t *cp =
            wcbe_find_component_pipeline(config->ls_pipelines,
                                         full_pipeline_name);
        if (cp != NULL) {
            gchar *msg = gst_structure_to_string(structure);
            wcbe_stream_msg_t stream_msg;
            stream_msg.object.type = WCBE_STREAM_MSG;
            stream_msg.structure_len = strlen(msg) + 1;
            struct iovec iovec[2] =
                {{.iov_base = &stream_msg, .iov_len = sizeof(stream_msg)},
                 {.iov_base = msg, .iov_len = stream_msg.structure_len}};
            struct msghdr msghdr =
                {.msg_name = &(cp->lsp.sockaddr),
                 .msg_namelen = sizeof(cp->lsp.sockaddr),
                 .msg_control = NULL, .msg_controllen = 0,
                 .msg_iov = iovec, .msg_iovlen = G_N_ELEMENTS(iovec)};
            errno = 0;
            gssize bytes_written =
                sendmsg(pipeline->current.lsp_sockfd, &msghdr, 0);
            if (G_LIKELY(bytes_written
                         == sizeof(stream_msg) + stream_msg.structure_len)) {
                result = TRUE;
            }
            else {
                const gchar *errstr;
                if (errno != 0) {
                    errstr = g_strerror(errno);
                }
                else {
                    errstr = "'errno' not set";
                }
                GST_WARNING("Failed to write stream object to socket %s: %s",
                            ((struct sockaddr_un *)msghdr.msg_name)->sun_path,
                            errstr);
            }
            g_free(msg);
        }
        g_free(full_pipeline_name);
    }
    return result;
}


static configuration_t *
find_config(pipeline_t *pipeline, const gchar *name)
{
    g_static_rec_mutex_lock(&(pipeline->configs_mtx));
    configuration_t *result = g_hash_table_lookup(pipeline->configs, name);
    g_static_rec_mutex_unlock(&(pipeline->configs_mtx));
    return result;
}

static void
element_list_query(pipeline_t *pipeline, wcbe_msg_t **msg,
                   gsize *msgsize, GstStructure *structure)
{
    g_return_if_fail((*msg)->type == WCBE_ELEMENT_LIST_MSG);

    wcbe_element_list_msg_t **e_msg = (wcbe_element_list_msg_t **)msg;
    const gchar *list = gst_structure_get_string(structure, "list");
    (*e_msg)->list_len = strlen(list) + 1;
    *msgsize = sizeof(wcbe_element_list_msg_t) + (*e_msg)->list_len;
    *e_msg = g_realloc(*e_msg, *msgsize);
    g_strlcpy((*e_msg)->list, list, (*e_msg)->list_len);
}

static void
property_list_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize,
                    GstStructure *structure)
{
    g_return_if_fail((*msg)->type == WCBE_PROPERTY_LIST_MSG);

    wcbe_property_list_msg_t **p_msg = (wcbe_property_list_msg_t **)msg;
    const gchar *list = gst_structure_get_string(structure, "list");
    (*p_msg)->list_len = strlen(list) + 1;
    *msgsize = sizeof(wcbe_property_list_msg_t) + (*p_msg)->list_len;
    *p_msg = g_realloc(*p_msg, *msgsize);
    g_strlcpy((*p_msg)->list, list, (*p_msg)->list_len);
}

static void
property_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize,
               GstStructure *structure)
{
    g_return_if_fail((*msg)->type == WCBE_PROPERTY_MSG);

    wcbe_property_msg_t **p_msg = (wcbe_property_msg_t **)msg;
    const GValue *value = gst_structure_get_value(structure, "value");
    if (value != NULL) {
        void prep_msg(GType type, gsize value_len) {
            *msgsize = sizeof(wcbe_property_msg_t) + value_len;
            *p_msg = g_realloc(*p_msg, *msgsize);
            (*p_msg)->value_len = value_len;
            g_strlcpy((*p_msg)->type, g_type_name(type),
                      sizeof((*p_msg)->type));
        }
        void copy_val(gconstpointer v) {
            memcpy((*p_msg)->value, v, (*p_msg)->value_len);
        }
        switch (G_VALUE_TYPE(value)) {
          case G_TYPE_CHAR: {
              // gint8 is the correct type according to GObject docs
              prep_msg(G_TYPE_CHAR, sizeof(gint8));
              gchar v = g_value_get_char(value);
              copy_val(&v);
              break;
          }
          case G_TYPE_UCHAR: {
              prep_msg(G_TYPE_UCHAR, sizeof(guchar));
              guchar v = g_value_get_uchar(value);
              copy_val(&v);
              break;
          }
          case G_TYPE_BOOLEAN: {
              prep_msg(G_TYPE_BOOLEAN, sizeof(gboolean));
              gboolean v = g_value_get_boolean(value);
              copy_val(&v);
              break;
          }
          case G_TYPE_INT: {
              prep_msg(G_TYPE_INT, sizeof(gint));
              gint v = g_value_get_int(value);
              copy_val(&v);
              break;
          }
          case G_TYPE_UINT: {
              prep_msg(G_TYPE_UINT, sizeof(guint));
              guint v = g_value_get_uint(value);
              copy_val(&v);
              break;
          }
          case G_TYPE_LONG: {
              prep_msg(G_TYPE_LONG, sizeof(glong));
              glong v = g_value_get_long(value);
              copy_val(&v);
              break;
          }
          case G_TYPE_ULONG: {
              prep_msg(G_TYPE_ULONG, sizeof(gulong));
              gulong v = g_value_get_ulong(value);
              copy_val(&v);
              break;
          }
          case G_TYPE_INT64: {
              prep_msg(G_TYPE_INT64, sizeof(gint64));
              gint64 v = g_value_get_int64(value);
              copy_val(&v);
              break;
          }
          case G_TYPE_UINT64: {
              prep_msg(G_TYPE_UINT64, sizeof(guint64));
              guint64 v = g_value_get_uint64(value);
              copy_val(&v);
              break;
          }
          case G_TYPE_FLOAT: {
              prep_msg(G_TYPE_FLOAT, sizeof(gfloat));
              gfloat v = g_value_get_float(value);
              copy_val(&v);
              break;
          }
          case G_TYPE_DOUBLE: {
              prep_msg(G_TYPE_DOUBLE, sizeof(gdouble));
              gdouble v = g_value_get_double(value);
              copy_val(&v);
              break;
          }
          case G_TYPE_STRING: {
              const gchar *str = g_value_get_string(value);
              guint len = strlen(str) + 1;
              prep_msg(G_TYPE_STRING, len);
              copy_val(str);
              break;
          }
          default:
            if (g_type_is_a(G_VALUE_TYPE(value), G_TYPE_ENUM)) {
                prep_msg(G_VALUE_TYPE(value), sizeof(gint));
                gint v = g_value_get_enum(value);
                copy_val(&v);
            }
            else if (g_type_is_a(G_VALUE_TYPE(value), G_TYPE_FLAGS)) {
                prep_msg(G_VALUE_TYPE(value), sizeof(guint));
                guint v = g_value_get_flags(value);
                copy_val(&v);
            }
            break;
        }
    }
    else {
        (*p_msg)->value_len = 0;
    }
}

static void
pad_list_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize,
               GstStructure *structure)
{
    g_return_if_fail((*msg)->type == WCBE_PAD_LIST_MSG);

    wcbe_pad_list_msg_t **p_msg = (wcbe_pad_list_msg_t **)msg;
    const gchar *list = gst_structure_get_string(structure, "list");
    (*p_msg)->list_len = strlen(list) + 1;
    *msgsize = sizeof(wcbe_pad_list_msg_t) + (*p_msg)->list_len;
    *p_msg = g_realloc(*p_msg, *msgsize);
    g_strlcpy((*p_msg)->list, list, (*p_msg)->list_len);
}

static void
pad_caps_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize,
               GstStructure *structure)
{
    g_return_if_fail((*msg)->type == WCBE_PAD_CAPS_MSG);

    wcbe_pad_caps_msg_t *p_msg = (wcbe_pad_caps_msg_t *)*msg;
    const gchar *caps = gst_structure_get_string(structure, "caps");
    g_strlcpy(p_msg->caps, caps, sizeof(p_msg->caps));
}

static void
pad_peer_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize,
               GstStructure *structure)
{
    g_return_if_fail((*msg)->type == WCBE_PAD_PEER_MSG);

    wcbe_pad_peer_msg_t *p_msg = (wcbe_pad_peer_msg_t *)*msg;
    const gchar *peer = gst_structure_get_string(structure, "peer");
    g_strlcpy(p_msg->peer, peer, sizeof(p_msg->peer));
}

request_cb_fn_t
wcbe_mc_config_doc_query(pipeline_t *pipeline, wcbe_msg_t **msg,
                         gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_CONFIG_DOC_MSG, NULL);

    wcbe_config_doc_msg_t **doc_msg = (wcbe_config_doc_msg_t **)msg;
    configuration_t *config = find_config(pipeline, (*doc_msg)->name);
    gchar *doc = NULL;
    (*doc_msg)->doc_len = 0;
    if (config != NULL) {
        wcbe_config_doc(&(config->wconfig), &doc);
        if (doc != NULL) {
            (*doc_msg)->doc_len = strlen(doc) + 1;
        }
    }
    *msgsize = sizeof(wcbe_config_doc_msg_t) + (*doc_msg)->doc_len;
    if ((*doc_msg)->doc_len > 0) {
        *doc_msg = g_realloc(*doc_msg, *msgsize);
        g_strlcpy((*doc_msg)->doc, doc, (*doc_msg)->doc_len);
    }
    if (doc != NULL) {
        g_free(doc);
    }
    return NULL;
}

request_cb_fn_t
wcbe_mc_new_subscan_command(pipeline_t *pipeline, wcbe_msg_t **msg,
                            gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_NEW_SUBSCAN_MSG, NULL);
    wcbe_new_subscan_msg_t *s_msg = (wcbe_new_subscan_msg_t *)*msg;
    if (pipeline->current.pipeline != NULL) {
        gchar *uid = s_msg->uid;
        if (strcmp(s_msg->uid, "null") == 0 || strlen(s_msg->uid) == 0) {
            uid = "";
        }
        g_signal_emit(GST_PIPELINE(pipeline->current.pipeline),
                      new_subscan_signal, 0,
                      s_msg->config, uid,
                      s_msg->time0.epoch, s_msg->time0.seconds,
                      s_msg->time0.clock_cycles);
    }
    return NULL;
}

request_cb_fn_t
wcbe_mc_cancel_subscan_command(pipeline_t *pipeline, wcbe_msg_t **msg,
                               gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_CANCEL_SUBSCAN_MSG, NULL);
    wcbe_cancel_subscan_msg_t *s_msg = (wcbe_cancel_subscan_msg_t *)*msg;
    if (pipeline->current.pipeline != NULL) {
        g_signal_emit(GST_PIPELINE(pipeline->current.pipeline),
                      cancel_subscan_signal, 0,
                      s_msg->config, s_msg->uid);
    }
    return NULL;
}

request_cb_fn_t
wcbe_mc_bdf_output_dir_command(pipeline_t *pipeline, wcbe_msg_t **msg,
                               gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_BDF_OUTPUT_DIR_MSG, NULL);
    wcbe_bdf_output_dir_msg_t *b_msg = (wcbe_bdf_output_dir_msg_t *)*msg;
    struct stat statbuf;
    gint rc = stat(b_msg->name, &statbuf);
    if (rc == 0) {
        if (S_ISDIR(statbuf.st_mode)) {
            if (pipeline->bdfdir != NULL) {
                g_free(pipeline->bdfdir);
            }
            pipeline->bdfdir = g_strdup(b_msg->name);
        }
        else {
            g_info("BDF output directory path name is not a directory");
        }
    }
    else {
        g_info("Failed to 'stat' BDF directory '%s': %s",
               b_msg->name, g_strerror(errno));
    }
    return NULL;
}

request_cb_fn_t
wcbe_mc_bdf_output_dir_query(pipeline_t *pipeline, wcbe_msg_t **msg,
                             gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_BDF_OUTPUT_DIR_MSG, NULL);
    wcbe_bdf_output_dir_msg_t *b_msg = (wcbe_bdf_output_dir_msg_t *)*msg;
    g_strlcpy(b_msg->name, pipeline->bdfdir, sizeof(b_msg->name));
    return NULL;
}

request_cb_fn_t
wcbe_mc_config_integration_query(pipeline_t *pipeline, wcbe_msg_t **msg,
                                 gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_CONFIG_INTEGRATION_MSG, NULL);
    wcbe_config_integration_msg_t *i_msg =
        (wcbe_config_integration_msg_t *)*msg;
    configuration_t *config = find_config(pipeline, i_msg->name);
    i_msg->usec = 0;
    if (config != NULL && config->wconfig.output.subarrays != NULL) {
        g_assert(config->wconfig.output.subarrays->len <= 1);
        if (config->wconfig.output.subarrays->len == 1) {
            wcbe_subarray_config_t *sa = &g_array_index(
                config->wconfig.output.subarrays, wcbe_subarray_config_t, 0);
            i_msg->usec = sa->integration;
        }
    }
    return NULL;
}
