#! /usr/bin/env python
#
# wcbetool
#
# Utility for controlling and monitoring wcbe pipelines under an
# instance of pipelinefs.
#
import os
import os.path
from sys import stderr
from time import time, sleep, strftime
from glob import glob
from socket import gethostbyname, gethostbyaddr
from subprocess import Popen, PIPE, STDOUT
import stat
import math
from wcbe import *

def parse_args(usage, description, modes):
    """
    Parse command line arguments; expanding mode argument prefix
    values.
    """
    from optparse import OptionParser
    parser = OptionParser(usage=usage, description=description)
    parser.disable_interspersed_args()
    (options, args) = parser.parse_args()
    if len(args) == 0 or not any([m.startswith(args[0]) for m in modes]):
        print usage
        sys.exit(-1)
    mode_prefix = args.pop(0)
    modes = reduce(lambda acc, m: acc + [m] if m.startswith(mode_prefix)
                   else acc, modes, [])
    if len(modes) > 1:
        print 'Mode argument (%s) matches more than one option' \
              % mode_prefix, modes
        print usage
        sys.exit(-1)
    return (options, args, modes[0])

def info(patterns, lfxset):
    """
    Get wcbe pipeline instance information. The information displayed
    by this function is relatively static for a pipeline instance,
    more related to configuration than current operational state.
    """
    if len(patterns) == 0:
        patterns = ['*']
    lines = []
    num_id_parts = int(os.environ['WCBE_NODE_ID_PARTS'])
    for path in pipelinefs_glob(patterns, lfxset):
        vals = {}
        lines.append(vals)
        (vals['dir'], vals['node'], unused, vals['nic'], vals['proc']) = \
            instance_id(path, num_id_parts)
        try:
            vals['name'] = read_pipeline_file(os.path.join(path, 'name'))
            vals['bdfdir'] = read_pipeline_file(os.path.join(path, 'bdfdir'))
            vals['auxdir'] = read_pipeline_file(os.path.join(path, 'auxdir'))
            src = read_pipeline_file(os.path.join(path, 'src'))
            if ':' in src:
                src = src.split(':')
                vals['ip'] = src[1][src[1].rfind('/') + 1:]
                vals['port'] = int(src[2])
                # TODO: handle stderr of subprocceses
                subproc = Popen('ssh %s@%s "/sbin/ifconfig | grep -C 1 %s '
                                '| head -1"'
                                % (os.environ['WCBE_USER'], vals['node'],
                                   vals['ip']),
                                shell=True, stdout=PIPE, stderr=PIPE)
                (out, err) = subproc.communicate()
                vals['mac'] = out.split()[-1]
            else:
                vals['ip'] = '-'
                vals['port'] = 0
                vals['mac'] = '-'
        except IOError:
            lines.pop()

    heads = {'dir': 'ID', 'node': 'Node', 'nic': 'NIC', 'ip': 'IP',
             'mac': 'MAC', 'proc': 'Proc', 'port': 'Port', 'name': 'Name',
             'bdfdir': 'BDF dir', 'auxdir': 'BLF/BLS dir'}
    field_lengths = {}
    for (k, v) in heads.items():
        field_lengths[k] = len(v) + 1
    for line in lines:
        for (k, v) in line.items():
            vlen = len(str(v)) + 1
            if vlen > field_lengths[k]:
                field_lengths[k] = vlen
    fmt = '#(dir)-%(dir)ds#(node)-%(node)ds#(nic)-%(nic)dd' \
          '#(ip)-%(ip)ds#(mac)-%(mac)ds' \
          '#(port)-%(port)dd#(proc)-%(proc)dd#(name)-%(name)ds' \
          '#(bdfdir)-%(bdfdir)ds#(auxdir)-%(auxdir)ds' % field_lengths
    fmt = fmt.replace('#', '%')
    hdrfmt = '#(dir)-%(dir)ds#(node)-%(node)ds#(nic)-%(nic)ds' \
             '#(ip)-%(ip)ds#(mac)-%(mac)ds' \
             '#(port)-%(port)ds#(proc)-%(proc)ds#(name)-%(name)ds' \
             '#(bdfdir)-%(bdfdir)ds#(auxdir)-%(auxdir)ds' % field_lengths
    hdrfmt = hdrfmt.replace('#', '%')
    print hdrfmt % heads
    def cmp(n0, n1):
        node0 = n0['node']
        node1 = n1['node']
        if node0 < node1:
            result = -1
        elif node0 > node1:
            result = 1
        else:
            nic0 = n0['nic']
            nic1 = n1['nic']
            if nic0 < nic1:
                result = -1
            elif nic0 > nic1:
                result = 1
            else:
                proc0 = n0['proc']
                proc1 = n1['proc']
                if proc0 < proc1:
                    result = -1
                elif proc0 > proc1:
                    result = 1
                else:
                    result = 0
        return result
    lines.sort(cmp=cmp)
    for line in lines:
        print fmt % line

def status(patterns, lfxset):
    """
    Get wcbe pipeline instance status information. The information
    displayed by this function is relatively dynamic for a pipeline
    instance. It is intended to reflect current operational status
    rather than configuration information.
    """
    if len(patterns) == 0:
        patterns = ['*']
    lines = []
    num_id_parts = int(os.environ['WCBE_NODE_ID_PARTS'])
    for path in pipelinefs_glob(patterns, lfxset):
        vals = {}
        lines.append(vals)
        (vals['id'], unused, vals['node'], vals['nic'], vals['proc']) = \
            instance_id(path, num_id_parts)
        active = os.path.join(path, 'active')
        if os.path.islink(active):
            vals['active'] = os.path.basename(os.path.realpath(active))
            try:
                vals['rate'] = float(read_pipeline_file(
                    os.path.join(path, 'active', 'pipelines', 'INPUT_STAGE',
                                 'src', 'properties', 'actual-rate')))
            except IOError:
                vals['rate'] = 0.0
        else:
            vals['active'] = '-'
            vals['rate'] = 0.0

        configs = [os.path.basename(c) for c
                   in glob(os.path.join(path, 'configurations', '*'))]
        vals['configs'] = ','.join([c for c in configs if c != '00README'])

        try:
            if read_pipeline_file(
                os.path.join(path, 'configurations', 'default', 'pipelines',
                             'INPUT_STAGE', 'framedump', 'properties',
                             'do-output')) == '1':
                vals['blf'] = 'y'
            else:
                vals['blf'] = 'n'
        except IOError:
            vals['blf'] = '-'

        try:
            if read_pipeline_file(
                os.path.join(path, 'configurations', 'default', 'pipelines',
                             'INPUT_STAGE', 'blsdump', 'properties',
                             'do-output')) == '1':
                vals['bls'] = 'y'
            else:
                vals['bls'] = 'n'
        except IOError:
            vals['bls'] = '-'

    heads = {'id': 'ID', 'rate': 'Rate', 'blf': 'BLF', 'bls': 'BLS',
             'active': 'Active', 'configs':'Configs'}
    field_fmts = {'rate': '%-.0f'}
    field_lengths = {}
    for (k, v) in heads.items():
        field_lengths[k] = len(v) + 1
    for line in lines:
        for (k, v) in line.items():
            if k in field_lengths:
                if k in field_fmts:
                    strv = field_fmts[k] % v
                else:
                    strv = str(v)
                vlen = len(strv) + 1
                if vlen > field_lengths[k]:
                    field_lengths[k] = vlen
    fmt = '#(id)-%(id)ds#(rate)-%(rate)d.0f#(blf)-%(blf)ds#(bls)-%(bls)ds' \
          '#(active)-%(active)ds#(configs)-%(configs)ds' \
          % field_lengths
    fmt = fmt.replace('#', '%')
    hdrfmt = '#(id)-%(id)ds#(rate)-%(rate)ds#(blf)-%(blf)ds#(bls)-%(bls)ds' \
             '#(active)-%(active)ds#(configs)-%(configs)ds' \
             % field_lengths
    hdrfmt = hdrfmt.replace('#', '%')
    print hdrfmt % heads
    def cmp(n0, n1):
        node0 = n0['node']
        node1 = n1['node']
        if node0 < node1:
            result = -1
        elif node0 > node1:
            result = 1
        else:
            nic0 = n0['nic']
            nic1 = n1['nic']
            if nic0 < nic1:
                result = -1
            elif nic0 > nic1:
                result = 1
            else:
                proc0 = n0['proc']
                proc1 = n1['proc']
                if proc0 < proc1:
                    result = -1
                elif proc0 > proc1:
                    result = 1
                else:
                    result = 0
        return result
    lines.sort(cmp=cmp)
    for line in lines:
        print fmt % line

class PipelineProc (object):

    def __init__(self, path=None, node=None, nic=None, proc=None, input=None):
        self.path = path
        self.node = node
        self.nic = nic
        self.proc = proc
        self.input = input
        self.state = 'spawning'
        self.last_path = ''
        self.last_str = ''
        return

def up(patterns, lfxset):
    """
    Start up some wcbe pipeline instances.
    """
    # Default configuration
    default_configuration = """\
<?xml version='1.0' encoding='utf-8'?>
<config xmlns='http://evla.nrao.edu/wcbe'>
    <inputConfig/>
    <outputConfig/>
    <pipelineConfig/>
</config>"""

    # A few convenience functions follow.
    def has_wildcard(str):
        return '*' in str or '?' in str or '[' in str

    # The next three are used to simplify the state machine code for
    # initializing the pipeline instances.
    def writef(pl, path, str, next_state):
        "Write 'str' to 'path'"
        f = open(path, 'w')
        f.write(str)
        f.close()
        pl.state = next_state
        pl.last_path = path
        pl.last_str = str
        return

    def readf(pl, next_state):
        "Read back last write to file, transition state if verified"
        if os.path.isfile(pl.last_path):
            f = open(pl.last_path, 'r')
            line = f.readline().strip()
            f.close()
            if line == pl.last_str:
                pl.state = next_state
        return

    def checkd(pl, next_state):
        "Check last directory creation, transition state if verified"
        if os.path.isdir(pl.last_path):
            pl.state = next_state
        return

    if len(patterns) == 0:
        patterns = ['*']
    num_id_parts = int(os.environ['WCBE_NODE_ID_PARTS'])
    # Check given patterns for validity
    for p in patterns:
        parts = p.split('-')
        if len(parts) == num_id_parts + 2:
            proc_part = parts[num_id_parts + 1]
            if has_wildcard(proc_part):
                print >> stderr, \
                      ('Invalid pattern (%s): no wildcard characters ' \
                       'allowed in third (process id) part of patterns.' \
                       % p)
                sys.exit(-1)
        if len(parts) > num_id_parts + 2:
            print >> stderr, \
                  ('Invalid pattern (%s): number of dash-separated parts ' \
                   'is limited to three.' % p)

    # Patterns in 'patterns' need to be expanded against lfxset to the
    # matching node and interface numbers.
    expanded_patterns = name_glob(patterns, lfxset, True)
    paths = [os.path.join(os.environ['WCBE_PIPELINEFS'], p)
             for p in expanded_patterns]

    # First pass, winnow existing paths from 'paths'
    new_paths = []
    for path in paths:
        do_winnow = False
        basename = os.path.basename(path)
        if os.path.isdir(path):
            if os.path.isfile(os.path.join(path, 'name')):
                print '%s is already up' % basename
            else:
                print '%s appears in %s, but is unresponsive' \
                      % (basename, os.environ['WCBE_PIPELINEFS'])
        else:
            new_paths.append(path)

    # Second pass, spawn the pipelines
    input_port_base = int(os.environ['WCBE_LAG_FRAME_INPUT_PORT'])
    mc_port_base = int(os.environ['WCBE_MC_PORT'])
    max_interfaces = int(os.environ['WCBE_MAX_LAG_FRAME_INTERFACES'])
    spawned = []
    for path in new_paths:
        (id, node, unused, nic, proc) = instance_id(path, num_id_parts)
        print 'Spawning pipeline %s' % id
        try:
            f = open(path, 'w')
            f.write('%s:%d' % \
                    (node, mc_port_base + proc * max_interfaces + nic - 1))
            f.close()
            pipeline = PipelineProc(path=path, node=node, nic=nic, proc=proc)
            spawned.append(pipeline)
        except IOError, e:
            print 'Failed to spawn pipeline %s: %s' % (basename, e)

    # Do IP address lookups for data interfaces on spawned pipelines
    for pipeline in spawned:
        name = '%s-%d' % (pipeline.node, pipeline.nic)
        pipeline.input = gethostbyname(name)
        pipeline.last_path = pipeline.path

    # Third pass, initialize pipelines as they come up.
    # TODO: improve the timeout mechanism
    stoptime = time() + 10.0
    while len(spawned) > 0 and time() < stoptime:
        pipeline = spawned.pop(0)
        try:
            if pipeline.state == 'spawning':
                checkd(pipeline, 'spawned')
            if pipeline.state == 'spawned':
                path = os.path.join(pipeline.path, 'src')
                str = 'widar://%s:%d' \
                      % (pipeline.input, input_port_base + pipeline.proc)
                writef(pipeline, path, str, 'srcing')
            elif pipeline.state == 'srcing':
                readf(pipeline, 'srced')
            if pipeline.state == 'srced':
                path = os.path.join(pipeline.path, 'configurations', 'default')
                str = default_configuration.strip()
                writef(pipeline, path, str, 'configuring')
            elif pipeline.state == 'configuring':
                checkd(pipeline, 'configured')
            if pipeline.state == 'configured':
                path = os.path.join(pipeline.path, 'flow')
                str = 'on'
                writef(pipeline, path, str, 'flowing')
            elif pipeline.state == 'flowing':
                readf(pipeline, 'flowed')
            if pipeline.state == 'flowed':
                path = os.path.join(pipeline.path, 'bdfdir')
                str = os.path.join(os.environ['WCBE_BDFDIR'], '.private')
                writef(pipeline, path, str, 'bdfdiring')
            elif pipeline.state == 'bdfdiring':
                readf(pipeline, 'bdfdired')
            if pipeline.state == 'bdfdired':
                # TODO: set up auxdir
                path = os.path.join(pipeline.path, 'activations',
                                    'now-default')
                str = '1'
                writef(pipeline, path, str, 'activating')
            elif pipeline.state == 'activating':
                dirents = glob(os.path.join(pipeline.path,
                                            'activations', '*-default'))
                for d in dirents:
                    if os.path.isdir(d):
                        print 'Spawned and initialized pipeline %s' \
                              % os.path.basename(pipeline.path)
                        pipeline = None
                        break
        except IOError, e:
            pipeline.state = 'failed'
            print 'Failed to write "%s" to %s: %s' % (str, path, e)
            pipeline = None
        if pipeline is not None:
            spawned.append(pipeline)

    for p in spawned:
        print 'Pipeline %s stalled in state %s' \
              % (os.path.basename(p.path), p.state)
    return

def down(patterns, lfxset):
    """
    Shut down some wcbe pipeline instances.
    """
    if len(patterns) == 0:
        patterns = ['*']
    # Remove the directories from pipelinefs
    removed = []
    for path in pipelinefs_glob(patterns, lfxset):
        basename = os.path.basename(path)
        try:
            f = open(os.path.join(path, 'name'), 'r')
            localname = f.readline().strip()
            f.close()
            print 'Shutting down', basename
            os.rmdir(path)
            removed.append((path, localname))
        except:
            print 'Error removing directory %s...pipeline %s may not ' \
                  'shut down properly' \
                  % (path, basename)

    # In order to avoid returning before the remote process has been
    # killed, we poll the remote machines.
    # TODO: improve the timeout mechanism
    stoptime = time() + 15.0
    num_id_parts = int(os.environ['WCBE_NODE_ID_PARTS'])
    while len(removed) > 0 and time () < stoptime:
        instance = removed.pop(0)
        (id, node, unused, unused, unused) = \
             instance_id(instance[0], num_id_parts)
        subproc = Popen('ssh %s@%s ls /tmp/%s'
                        % (os.environ['WCBE_USER'], node, instance[1]),
                        shell=True, bufsize=0, stdout=PIPE, stderr=PIPE)
        (out, err) = subproc.communicate()
        if err.find('No such file or directory') != -1:
            instance = None
            print id, 'has shut down'
        if instance is not None:
            removed.append(instance)
    for r in removed:
        basename = os.path.basename(r[0])
        print basename, 'may not have shut down cleanly'
    return

def zombies(patterns, lfxset):
    """
    Find pipeline processes on cbe slave nodes that are not listed
    in pipelinefs, and give the user an option to clean them up.
    """
    if len(patterns) == 0:
        patterns = ['*']
    num_id_parts = int(os.environ['WCBE_NODE_ID_PARTS'])
    nodes = list(set(['-'.join(p.split('-')[:num_id_parts])
                      for p in name_glob(patterns, lfxset, True)]))
    nodes.sort()
    nodedirs = [os.path.basename(d) for d in pipelinefs_glob(patterns, lfxset)]
    prefix_len = len(os.environ['WCBE_NODE_PREFIX'])
    port_base = int(os.environ['WCBE_LAG_FRAME_INPUT_PORT'])

    def cmp(d0, d1):
        d0 = [int(p) for p in d0[0].split('-')]
        d1 = [int(p) for p in d1[0].split('-')]
        if d0[0] < d1[0]:
            ret = -1
        elif d0[0] > d1[0]:
            ret = 1
        else:
            if d0[1] < d1[1]:
                ret = -1
            elif d0[1] > d1[1]:
                ret = 1
            else:
                if d0[2] < d1[2]:
                    ret = -1
                elif d0[2] > d1[2]:
                    ret = 1
                else:
                    ret = 0
        return ret

    def yn(prompt, default='y'):
        valid = 'yYnN'
        ret = raw_input(prompt)
        if ret == '':
            ret = default
        while ret not in valid:
            print 'Invalid input...try again'
            ret = raw_input(prompt)
            if ret == '':
                ret = default
        return ret

    for node in nodes:
        subproc = Popen('ssh %s@%s'\
                        % (os.environ['WCBE_USER'],
                           os.environ['WCBE_NODE_PREFIX'] + node),
                        shell=True, bufsize=0, stdin=PIPE, stdout=PIPE,
                        stderr=PIPE)
        # TODO: The following will miss instances that have not set a
        # 'src' value. Very rare occurrence, but might want to fix
        # this.
        (out, err) = subproc.communicate(
            'for p in /tmp/cbe*/src; do echo -n "$p "; cat $p; echo; done')
        zombies = []
        for line in out.splitlines():
            if line.startswith('/tmp/cbe'):
                lsparts = line.split()
                if len(lsparts) == 2:
                    srcparts = lsparts[1].split(':')
                    if srcparts[0] == 'widar':
                        # skip over '//' in srcparts[1] for IP address
                        hnames = gethostbyaddr(srcparts[1][2:])
                        hostnames = [hnames[0]] + hnames[1]
                        hostname = None
                        for h in hostnames:
                            if h.startswith(os.environ['WCBE_NODE_PREFIX']):
                                hostname = h
                                break
                        if hostname is not None:
                            dot = hostname.find('.')
                            if dot != -1:
                                hostname = hostname[:dot]
                            instance = int(srcparts[2]) - port_base
                            host_suffix = hostname[prefix_len:]
                            dirname = '%s-%d' % (host_suffix, instance)
                            if dirname not in nodedirs:
                                zombies.append((dirname, hostname,
                                                os.path.dirname(lsparts[0])))
        zombies.sort(cmp=cmp)
        for z in zombies:
            print 'Zombie: %s on %s' % z[:2]
#             reconnect = yn('Attempt to reconnect [yN]? ', 'n')
#             if reconnect == 'y' or reconnect == 'Y':
#                 print 'FIXME!'
            clean = yn('Clean up [Yn]? ')
            if clean == 'y' or clean == 'Y':
                subproc = Popen('ssh %s@%s'\
                                % (os.environ['WCBE_USER'],
                                   os.environ['WCBE_NODE_PREFIX'] + node),
                                shell=True, bufsize=0, stdin=PIPE, stdout=PIPE,
                                stderr=PIPE)
                cmd = 'for p in `ls %s`; do ps x | awk "{if (\\$1 == $p) ' \
                      '{if (match(\\$5, \\"^(wcbe|lagset_pipeline)$\\") ' \
                      '!= 0) {system(\\"kill -s KILL \\" \\$1)}}}"; done; ' \
                      'rm -rf %s' \
                      % (os.path.join(z[2], 'pids'), z[2])
                (out, err) = subproc.communicate(cmd)
    return

def auxout(patterns, lfxset, filetype, state):
    """
    Start and stop BLF or BLS file output by wcbe pipeline
    instances.
    """
    if len(patterns) == 0:
        patterns = ['*']
    controls = []
    if state == 'on':
        istate = '1'
    else:
        istate = '0'
    if filetype == 'blf':
        element = 'framedump'
    else:
        element = 'blsdump'
    # First pass: determine which instances to control, and set their
    # auxdir paths.
    pwd = os.getcwd()
    for path in pipelinefs_glob(patterns, lfxset):
        (dir_id, node, node_id, nic, proc) = instance_id(path)
        do_output = os.path.join(
            path, 'active', 'pipelines', 'INPUT_STAGE', element,
            'properties', 'do-output')
        if istate == '1':
            auxdir = os.environ['WCBE_AUXDIR_PATTERN']\
                     .replace('@PWD@', pwd)\
                     .replace('@WCBE_NODE_PREFIX@',
                              os.environ['WCBE_NODE_PREFIX'])\
                     .replace('@NODE_ID@', node_id)\
                     .replace('@LFX_INDEX@', str(nic))\
                     .replace('@PROC_INDEX@', str(proc))
            subproc = Popen('ssh %s@%s' % (os.environ['WCBE_USER'], node),
                            shell=True, bufsize=0, stdin=PIPE, stdout=PIPE,
                            stderr=PIPE)
            cmd = 'echo "access test" > %s/.$$.tmp && ' \
                  'cat %s/.$$.tmp && rm %s/.$$.tmp' \
                  % (auxdir, auxdir, auxdir)
            (out, err) = subproc.communicate(cmd)
            if out.strip() == 'access test':
                pth = os.path.join(path, 'auxdir')
                try:
                    f = open(pth, 'w')
                    f.write(auxdir)
                    f.close()
                    controls.append((do_output, dir_id))
                except:
                    print '%s: Failed to write %s directory path to %s...' \
                          'will not start %s file output' \
                          % (dir_id, filetype, pth, filetype)
            else:
                print '%s: Unable to confirm read/write access by %s user ' \
                      'to directory %s...will not start %s ' \
                      'file output' % \
                      (dir_id, os.environ['WCBE_USER'], auxdir, filetype)
        else:
            controls.append((do_output, dir_id))
    # Second pass: write control message to instances selected in
    # first pass.
    for control in controls:
        try:
            f = open(control[0], 'w')
            f.write(istate)
            f.close()
            print '%s: %s %s file output' \
                  % (control[1], 'Started' if istate == '1' else 'Stopped',
                     filetype)
        except:
            print '%s: Failed to set %s output state...will not %s ' \
                  '%s file output' \
                  % (control[1], filetype,
                     'start' if istate == '1' else 'stop', filetype)
    return

def activate(args, lfxset):
    """
    Activate a configuration immediately. Does not wait for
    configuration activation to occur before returning.

    Requires a configuration name argument, and, optionally, the
    number of seconds to wait before activating the configuration. The
    number of seconds to wait must be prefixed with a plus sign.

    Examples:

    * To activate configuration 'foo' immediately (on all pipelines):
      wcbetool activate foo

    * To activate configuration 'foo' in approximately five seconds (on
      all pipelines):
      wcbetool activate foo +5
    """
    if len(args) == 0:
        print 'Nothing to activate: activate mode requires a ' \
              'configuration name argument'
        sys.exit(-1)
    config = args.pop(0)
    configuration_name = os.path.join('configurations', config)
    if len(args) > 0 and len(pipelinefs_glob([args[0]], lfxset)) == 0:
        t = args.pop(0)
        if t[0] == '+':
            t0 = math.ceil(time() + float(t[1:]))
            activation_time = '0:%u:0' % int(t0)
            res = 'activate at %u' % int(t0)
        else:
            print 'Absolute activation times are not yet implemented'
            sys.exit(-1)
    else:
        activation_time = 'now'
        res = 'activated'
    activation_name = os.path.join(
        'activations', '%s-%s' % (activation_time, config))
    if len(args) == 0:
        args = ['*']
    paths = []
    for path in pipelinefs_glob(args, lfxset):
        if os.path.isdir(os.path.join(path, configuration_name)):
            paths.append(os.path.join(path, activation_name))
    for path in paths:
        f = open(path, 'w')
        f.write('a')
        f.close()
    for path in paths:
        print '%s\t%s %s' % (path.split('/')[-3], res, config)
    return

class ConfigInstance (object):
    """Handy-dandy for maintaining state information about an instance
    during the configuration process.
    """
    def __init__(self, path, config_name, is_transient):
        self.state = 'configuring'
        self.path = path
        self.name = os.path.basename(path)
        self.config_path = os.path.join(self.path, 'configurations',
                                        config_name)
        self.activate_path = os.path.join(self.path, 'activations',
                                          'now-%s' % config_name)
        self.active_link = os.path.join(self.path, 'active')
        self.is_transient = is_transient
        self.timeout = time() + float(os.environ['WCBE_CONFIGURATION_TIMEOUT'])
        return

    def update(self):
        result = 'continue'
        initial_state = self.state
        if self.state == 'configuring':
            if os.path.isdir(self.config_path):
                self.state = 'configured'
        if self.state == 'configured':
            if not self.is_transient:
                result = 'done'
            else:
                f = open(self.activate_path, 'w')
                f.write('a')
                f.close()
                self.state = 'activating'
        elif self.state == 'activating':
            try:
                if os.path.islink(self.active_link) \
                       and (os.path.realpath(self.active_link)
                            == self.config_path):
                    result = 'done'
                    os.rmdir(self.config_path)
            except OSError:
                pass
        if self.state == initial_state and result == 'continue':
            if time() >= self.timeout:
                result = 'timeout'
        return result

def configure(args, lfxset):
    """
    Configure pipeline instances.

    The args must include a configuration file path. Optionally, a
    configuration name may also be present. The existence of a
    configuration name affects the functionality as follows.

    * When the configuration name is provided, a persistent
      configuration is created. The pipelines will be given the
      configuration, but that configuration will not be activated.

    * When the configuration name is not provided, a transient
      configuration is created. The configuration name is generated
      automatically based on the time, the configuration is activated,
      and then deleted (so that it will disappear when a new
      activation occurs.)

    Examples:

    * To create a transient configuration:
      wcbetool configure /path/to/config/file

    * To create a persistent configuration named 'foo':
      wcbetool configure /path/to/config/file foo
    """
    if len(args) == 0:
        print 'No configuration file path: configure mode requires a ' \
              'configuration file path argument'
        sys.exit(-1)
    config_filename = args.pop(0)
    try:
        config = open(config_filename, 'r')
    except IOError:
        print 'Missing or invalid configuration file path argument (%s)' \
              % config_filename
        sys.exit(-1)
    if len(args) > 0 and len(pipelinefs_glob([args[0]], lfxset)) == 0:
        config_name = args.pop(0)
        transient_config = False
    else:
        config_name = strftime('%H_%M_%S')
        transient_config = True
    if len(args) == 0:
        args = ['*']
    configs = []
    for path in pipelinefs_glob(args, lfxset):
        config.seek(0)
        fname = os.path.join(path, 'configurations', config_name)
        try:
            f = open(fname, 'w')
            f.writelines(config.readlines())
            f.close()
        except OSError:
            pass
        except IOError:
            pass
        else:
            configs.append(ConfigInstance(path, config_name, transient_config))
    config.close()
    while len(configs) > 0:
        for c in configs:
            rc = c.update()
            res = None
            if rc == 'timeout':
                res = 'timed out'
            elif rc == 'failed':
                res = 'failed'
            elif rc == 'done':
                if transient_config:
                    res = 'configured and activated'
                else:
                    res = 'configured'
            if res is not None:
                print '%s\t%s %s' % (c.name, res, config_name)
                configs.remove(c)
    return

def deconfigure(args, lfxset):
    """
    Destroy configuration. Does not wait for the configuration to
    disappear from the pipelines' configuration lists.
    """
    if len(args) == 0:
        print 'No configuration name: deconfigure mode requires a ' \
              'configuration name argument'
        sys.exit(-1)
    config_name = args.pop(0)
    config_path = os.path.join('configurations', config_name)
    if len(args) == 0:
        args = ['*']
    paths = []
    for path in pipelinefs_glob(args, lfxset):
        config = os.path.join(path, config_path)
        if os.path.isdir(config):
            try:
                os.rmdir(config)
                active = os.path.join(path, 'active')
                try:
                    if os.path.realpath(active) == config:
                        res = 'pending destroy'
                    else:
                        res = 'destroyed'
                except OSError:
                    res = 'destroyed'
                print '%s\t%s %s' % (os.path.basename(path), res, config_name)
            except OSError:
                print '%s\tfailed to destroy %s' \
                      % (os.path.basename(path), config_name)
    return

if __name__ == '__main__':
    import sys
    set_environment('/etc/sysconfig/wcbe.conf')
    if not check_pipelinefs():
        print 'WCBE_PIPELINEFS environment variable (%s) does not ' \
              'refer to a valid pipelinefs file system.' \
              % os.environ['WCBE_PIPELINEFS']
        sys.exit(-1)
    lfxset = get_lfxset()
    modes =['info', 'up', 'down', 'status', 'zombies', 'blf-on', 'blf-off',
            'bls-on', 'bls-off', 'activate', 'configure', 'deconfigure']
    modes.sort()
    usage = 'usage: %s <%s> [mode args] instance*' \
            % (os.path.basename(sys.argv[0]), ' | '.join(modes))
    description = 'Tool for managing wcbe pipeline instances. The ' \
                  '"instance*" argument is a sequence of special regular ' \
                  'expressions that select pipeline instances on which to ' \
                  'act. No "instance" argument is required; by default ' \
                  'all instances listed in %s will be acted upon. Valid ' \
                  'regular expressions for "instance" consist of a ' \
                  'sequence of dash separated parts: the first %s to a node ' \
                  'id, the next to a data interface index, and the last, to ' \
                  'a process index. The three parts may contain shell-style ' \
                  'regular expressions, restricted to "*", "?", and "[]" ' \
                  'special characters. With fewer than three parts, a ' \
                  'regular expression is treated as if the missing parts ' \
                  'were "*" characters.' \
                  'For help on a mode, use "-h" after the mode argument.' \
                  % (os.environ['WCBE_LFX_FILE'],
                     'part refers' \
                     if int(os.environ['WCBE_NODE_ID_PARTS']) == 1 \
                     else '%d parts refer' \
                     % (os.environ['WCBE_NODE_ID_PARTS']))

    (options, args, mode) = parse_args(usage, description, modes)
    fn = None
    fn_args = ()
    if mode == 'info':
        fn = info
        fn_args = (args, lfxset)
    elif mode == 'up':
        fn = up
        fn_args = (args, lfxset)
    elif mode == 'down':
        fn = down
        fn_args = (args, lfxset)
    elif mode == 'status':
        fn = status
        fn_args = (args, lfxset)
    elif mode == 'zombies':
        fn = zombies
        fn_args = (args, lfxset)
    elif mode == 'blf-on':
        fn = auxout
        fn_args = (args, lfxset, 'blf', 'on')
    elif mode == 'blf-off':
        fn = auxout
        fn_args = (args, lfxset, 'blf', 'off')
    elif mode == 'bls-on':
        fn = auxout
        fn_args = (args, lfxset, 'bls', 'on')
    elif mode == 'bls-off':
        fn = auxout
        fn_args = (args, lfxset, 'bls', 'off')
    elif mode == 'activate':
        fn = activate
        fn_args = (args, lfxset)
    elif mode == 'configure':
        fn = configure
        fn_args = (args, lfxset)
    elif mode == 'deconfigure':
        fn = deconfigure
        fn_args = (args, lfxset)
    if fn is not None:
        if len(args) > 0 and args[0] == '-h':
            doc = '\n'.join(fn.__doc__.splitlines()[1:-1])
            print doc
            sys.exit(-1)
        else:
            print '*** wcbe pipeline instances under %s ***' \
                  % os.environ['WCBE_PIPELINEFS']
            fn(*fn_args)
