// async_syslog.c
//
// Asynchronous interface to syslog.
//
// Some of the code in this module that implements a GLib log handler
// has been copied and modified from the GLib source (gmessages.c).
// The copyright statement on that code is the following:
//
/* GLIB - Library of useful routines for C programming
 * Copyright (C) 1995-1997  Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
/**
 * SECTION: async_syslog
 * @short_description: Asynchronous syslog facility
 * @stability: unstable
 * @include: async_syslog.h
 *
 * Asynchronous interface to <function>syslog</function>, together
 * with GLib-style log handler function that uses the asynchronous
 * <function>syslog</function> functionality.
 *
 * As for the design, this subsystem uses a helper thread to actually
 * make the <function>syslog</function> call. Log messages are passed
 * to the helper thread using a #log_queue_t, which imposes a limit on
 * the number of queued messages. While the queue is filled to
 * capacity, messages logged through this subsystem will not be passed
 * to the helper thread, but will end up going to
 * <systemitem>stdout</systemitem> or <systemitem>stderr</systemitem>
 * instead.
 */
#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib.h>
#include <syslog.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>

#include <async_syslog.h>
#include <log_queue.h>

/**
 * ASYNC_SYSLOG_FACILITY:
 *
 * The <function>syslog</function> facility used by this subsystem.
 */
#define ASYNC_SYSLOG_FACILITY LOG_LOCAL5

// The following preprocessor symbols/macros are from GLib.
/* For a radix of 8 we need at most 3 output bytes for 1 input
 * byte. Additionally we might need up to 2 output bytes for the
 * radix prefix and 1 byte for the trailing NULL.
 */
#define FORMAT_UNSIGNED_BUFSIZE ((GLIB_SIZEOF_LONG * 3) + 3)
#define STRING_BUFFER_SIZE (FORMAT_UNSIGNED_BUFSIZE + 32)
#define ALERT_LEVELS (G_LOG_LEVEL_ERROR | G_LOG_LEVEL_CRITICAL | \
                      G_LOG_LEVEL_WARNING)
#define CHAR_IS_SAFE(wc) (\
    !((wc < 0x20 && wc != '\t' && wc != '\n' && wc != '\r') || \
      (wc == 0x7f) || \
      (wc >= 0x80 && wc < 0xa0)))

static gpointer logger(gpointer data);
static void g_messages_prefixed_init(void);
static int mklevel_prefix(gchar level_prefix[STRING_BUFFER_SIZE],
                          GLogLevelFlags log_level);
static void format_unsigned(gchar *buf, gulong num, guint radix);
static void escape_string(GString *string);
static gchar *strdup_convert(const gchar *string, const gchar *charset);
static void write_string(gint fd, const gchar *str);
static void log_fallback_handler(const gchar *log_domain,
                                 GLogLevelFlags log_level,
                                 const gchar *message, gpointer reserved);

static GStaticMutex subsys_ctl_mtx = G_STATIC_MUTEX_INIT;
static guint initialized = 0;
static GThread *logger_thread = NULL;
static log_queue_t *log_queue = NULL;
static GLogLevelFlags g_log_msg_prefix = G_LOG_LEVEL_ERROR | \
       G_LOG_LEVEL_WARNING | G_LOG_LEVEL_CRITICAL | G_LOG_LEVEL_DEBUG;
static GMutex *flush_mtx = NULL;
static gboolean is_flushing = FALSE;
static GCond *flushed_cond = NULL;

/**
 * async_syslog_init:
 * @backlog: maximum number of queued messages
 *
 * Initialize the asynchronous syslog subsystem. Every thread that
 * uses the subsystem should call this function exactly once.
 *
 * @see_also: async_syslog_quit()
 */
void
async_syslog_init(guint backlog)
{
    g_static_mutex_lock(&subsys_ctl_mtx);
    if (g_atomic_int_get((gint *)&initialized) == 0) {
        initialized = 1; // atomic set not needed since logger thread
                         // not running
        if (!g_thread_supported()) {
            g_thread_init(NULL);
        }
        flush_mtx = g_mutex_new();
        log_queue = log_queue_new(backlog);
        is_flushing = FALSE;
        flushed_cond = g_cond_new();
        GError *err = NULL;
        logger_thread = g_thread_create(logger, log_queue, TRUE, &err);
        if (logger_thread != NULL) {
            openlog(NULL, LOG_CONS | LOG_PID, ASYNC_SYSLOG_FACILITY);
        }
        else {
            g_warning("Failed to create async_syslog logger thread: %s",
                      err->message);
            initialized = 0;
            log_queue_unref(log_queue);
            g_mutex_free(flush_mtx);
        }
    }
    else {
        g_atomic_int_inc((gint *)&initialized);
    }
    g_static_mutex_unlock(&subsys_ctl_mtx);
}

/**
 * async_syslog_quit:
 *
 * Release resources used by the asynchronous syslog subsystem. Every
 * thread that has called async_syslog_init() should call this function
 * to signal that the thread is finished using the subsystem.
 *
 * @see_also: async_syslog_init(), async_syslog_flush()
 */
void
async_syslog_quit(void)
{
    g_static_mutex_lock(&subsys_ctl_mtx);
    if (g_atomic_int_dec_and_test((gint *)&initialized)) {
        log_queue_unref(log_queue);
        g_thread_join(logger_thread);
        g_mutex_free(flush_mtx);
        flush_mtx = NULL;
        logger_thread = NULL;
        log_queue = NULL;
        closelog();
    }
    g_static_mutex_unlock(&subsys_ctl_mtx);
}

/**
 * async_syslog_flush:
 *
 * Flush all queued log messages. Flushing involves processing all the
 * messages in the queue completely. It often may be useful to call
 * this function before calling the async_syslog_quit() function.
 *
 * @see_also: async_syslog_quit()
 */
void
async_syslog_flush(void)
{
    g_static_mutex_lock(&subsys_ctl_mtx);
    log_queue_drop(log_queue);
    g_mutex_lock(flush_mtx);
    is_flushing = TRUE;
    while (is_flushing) {
        g_cond_wait(flushed_cond, flush_mtx);
    }
    g_mutex_unlock(flush_mtx);
    log_queue_accept(log_queue);
    g_static_mutex_unlock(&subsys_ctl_mtx);
}

/**
 * async_syslog:
 * @priority: message priority
 * @message: <function>printf</function>-style format string
 * @Varargs: message arguments
 *
 * Like <function>syslog</function>, but asynchronous.
 *
 * Note: this function doesn't support the '&percnt;m' conversion
 * specification as supported by <function>syslog</function>. In fact,
 * any '&percnt;m' that appears in the message after normal
 * <function>printf</function> expansion will be wrongly interpolated
 * when the <function>syslog</function> function is later called. It
 * might be useful to handle the '&percnt;m' conversion in this
 * function before the normal <function>syslog</function> gets ahold
 * of the string, but I'm too lazy at the moment to do it properly.
 */
void
async_syslog(int priority, const char *message, ...)
{
    va_list ap;
    va_start(ap, message);
    gchar *msg = g_strdup_vprintf(message, ap);
    va_end(ap);
    if (!log_queue_push(log_queue, priority, msg)) {
        GLogLevelFlags log_level = 0;
        if (priority & LOG_ERR) {
            log_level = G_LOG_LEVEL_ERROR;
        }
        else if (priority & LOG_CRIT) {
            log_level = G_LOG_LEVEL_CRITICAL;
        }
        else if (priority & LOG_WARNING) {
            log_level = G_LOG_LEVEL_WARNING;
        }
        else if (priority & LOG_NOTICE) {
            log_level = G_LOG_LEVEL_MESSAGE;
        }
        else if (priority & LOG_INFO) {
            log_level = G_LOG_LEVEL_INFO;
        }
        else if (priority & LOG_DEBUG) {
            log_level = G_LOG_LEVEL_DEBUG;
        }
        // log_fallback_handler will add some Glib-ish stuff to the
        // message, but we can live with it
        log_fallback_handler(NULL, log_level, msg, NULL);
        g_free(msg);
    }
}

/**
 * async_syslog_handler:
 * @log_domain: log domain
 * @log_level: log level
 * @message: log message
 * @reserved: unused
 *
 * A log handler, as can be used by GLib logging functions.
 *
 * This code, and supporting code, is basically from GLib gmessages.c
 * source file. Some simplifications have been made to the GLib code
 * for this module's purposes.
 *
 * @see_also: async_syslog_set_log_handler(),
 *            async_syslog_set_default_log_handler()
 *
 */
void
async_syslog_handler(const gchar *log_domain, GLogLevelFlags log_level,
                     const gchar *message, gpointer reserved)
{
    gboolean is_fatal = (log_level & G_LOG_FLAG_FATAL) != 0;
    gchar level_prefix[STRING_BUFFER_SIZE], *string;
    GString *gstring;

    if (log_level & G_LOG_FLAG_RECURSION || g_thread_self() == logger_thread) {
        log_fallback_handler(log_domain, log_level, message, reserved);
        return;
    }

    g_messages_prefixed_init();

    mklevel_prefix(level_prefix, log_level);

    gstring = g_string_new(NULL);
    if (log_level & ALERT_LEVELS) {
        g_string_append(gstring, "\n");
    }
    if (!log_domain) {
        g_string_append(gstring, "** ");
    }

#if LOG_EXPLICIT_PROC_NAME_AND_ID
    if ((g_log_msg_prefix & log_level) == log_level) {
        const gchar *prg_name = g_get_prgname();

        if (!prg_name) {
            g_string_append_printf(gstring, "(process:%lu): ",
                                   (gulong)getpid ());
        }
        else {
            g_string_append_printf(gstring, "(%s:%lu): ",
                                   prg_name, (gulong)getpid ());
        }
    }
#endif

    if (log_domain) {
        g_string_append(gstring, log_domain);
        g_string_append_c(gstring, '-');
    }
    g_string_append(gstring, level_prefix);

    g_string_append(gstring, ": ");
    if (!message) {
        g_string_append(gstring, "(NULL) message");
    }
    else {
        GString *msg;
        const gchar *charset;

        msg = g_string_new(message);
        escape_string(msg);

        if (g_get_charset(&charset)) {
            g_string_append(gstring, msg->str); /* charset is UTF-8 already */
        }
        else {
            string = strdup_convert(msg->str, charset);
            g_string_append(gstring, string);
            g_free(string);
        }

        g_string_free(msg, TRUE);
    }
    if (is_fatal) {
        g_string_append(gstring, "\naborting...\n");
    }
    else {
        g_string_append(gstring, "\n");
    }

    string = g_string_free(gstring, FALSE);

    gint priority = 0;
    if (log_level & G_LOG_LEVEL_ERROR) {
        priority = LOG_ERR;
    }
    else if (log_level & G_LOG_LEVEL_CRITICAL) {
        priority = LOG_CRIT;
    }
    else if (log_level & G_LOG_LEVEL_WARNING) {
        priority = LOG_WARNING;
    }
    else if (log_level & G_LOG_LEVEL_MESSAGE) {
        priority = LOG_NOTICE;
    }
    else if (log_level & G_LOG_LEVEL_INFO) {
        priority = LOG_INFO;
    }
    else if (log_level & G_LOG_LEVEL_DEBUG) {
        priority = LOG_DEBUG;
    }
    if (!log_queue_push(log_queue, priority, string)) {
        log_fallback_handler(log_domain, log_level, message, NULL);
        g_free(string);
    }
    if (is_fatal) {
        async_syslog_flush();
    }
}

/**
 * async_syslog_set_log_handler:
 * @log_domain: log domain
 * @log_levels: log levels
 *
 * Set the GLib log handler for the given log domain and log levels to
 * async_syslog_handler().
 *
 * Returns: the id of the new handler
 *
 * @see_also: async_syslog_handler()
 */
guint
async_syslog_set_log_handler(const gchar *log_domain,
                             GLogLevelFlags log_levels)
{
    return g_log_set_handler(log_domain, log_levels,
                             async_syslog_handler, NULL);
}

/**
 * async_syslog_set_default_log_handler:
 *
 * Set the GLib default log handler to async_syslog_handler().
 *
 * Returns: the previous default log handler
 *
 * @see_also: async_syslog_handler()
 */
GLogFunc
async_syslog_set_default_log_handler(void)
{
    return g_log_set_default_handler(async_syslog_handler, NULL);
}

/*
 * logger
 *
 * Main function for thread that makes the actual calls to syslog. It
 * reads messages and the dropped message count from the global
 * 'log_queue' queue, and writes appropriate messages to syslog.
 * Periodically it checks whether to return (and exit from the
 * thread), but it does need to empty the queue before making that
 * check.
 */
static gpointer
logger(gpointer data)
{
    log_queue_t *log_queue = data;
    log_queue_ref(log_queue);

    while (g_atomic_int_get((gint *)&initialized) != 0) {
        GTimeVal end_wait;
        g_get_current_time(&end_wait);
        g_time_val_add(&end_wait, 0.5 * G_USEC_PER_SEC);
        log_queue_elem_t *elem = log_queue_timed_pop(log_queue, &end_wait);
        while (elem != NULL) {
            if (elem->lost > 0) {
                syslog(LOG_WARNING,
                       "async_syslog: dropped %u messages", elem->lost);
            }
            if (elem->msg != NULL) {
                syslog(elem->priority, elem->msg);
            }
            log_queue_elem_free(elem);
            elem = log_queue_try_pop(log_queue);
        }
        g_mutex_lock(flush_mtx);
        if (is_flushing) {
            is_flushing = FALSE;
            g_cond_signal(flushed_cond);
        }
        g_mutex_unlock(flush_mtx);
    }
    log_queue_unref(log_queue);
    return NULL;
}

static void
g_messages_prefixed_init(void)
{
    static gboolean initialized = FALSE;

    if (!initialized) {
        const gchar *val;

        initialized = TRUE;
        val = g_getenv("G_MESSAGES_PREFIXED");

        if (val) {
            const GDebugKey keys[] = {
                {"error", G_LOG_LEVEL_ERROR},
                {"critical", G_LOG_LEVEL_CRITICAL},
                {"warning", G_LOG_LEVEL_WARNING},
                {"message", G_LOG_LEVEL_MESSAGE},
                {"info", G_LOG_LEVEL_INFO},
                {"debug", G_LOG_LEVEL_DEBUG}
            };

            g_log_msg_prefix =
                g_parse_debug_string(val, keys, G_N_ELEMENTS (keys));
        }
    }
}

static gint
mklevel_prefix(gchar level_prefix[STRING_BUFFER_SIZE],
               GLogLevelFlags log_level)
{
    gboolean to_stdout = TRUE;

    /* we may not call _any_ GLib functions here */
    switch (log_level & G_LOG_LEVEL_MASK) {
      case G_LOG_LEVEL_ERROR:
        strcpy(level_prefix, "ERROR");
        to_stdout = FALSE;
        break;
      case G_LOG_LEVEL_CRITICAL:
        strcpy(level_prefix, "CRITICAL");
        to_stdout = FALSE;
        break;
      case G_LOG_LEVEL_WARNING:
        strcpy(level_prefix, "WARNING");
        to_stdout = FALSE;
        break;
      case G_LOG_LEVEL_MESSAGE:
        strcpy(level_prefix, "Message");
        to_stdout = FALSE;
        break;
      case G_LOG_LEVEL_INFO:
        strcpy(level_prefix, "INFO");
        break;
      case G_LOG_LEVEL_DEBUG:
        strcpy(level_prefix, "DEBUG");
        break;
      default:
        if (log_level) {
            strcpy(level_prefix, "LOG-");
            format_unsigned(
                level_prefix + 4, log_level & G_LOG_LEVEL_MASK, 16);
        }
        else {
            strcpy(level_prefix, "LOG");
        }
        break;
    }
    if (log_level & G_LOG_FLAG_RECURSION) {
        strcat(level_prefix, " (recursed)");
    }
    if (log_level & ALERT_LEVELS) {
        strcat(level_prefix, " **");
    }
    return to_stdout ? STDOUT_FILENO : STDERR_FILENO;
}

static void
format_unsigned(gchar *buf, gulong num, guint radix)
{
    gulong tmp;
    gchar c;
    gint i, n;

    /* we may not call _any_ GLib functions here (or macros like
     * g_return_if_fail()) */

    if (radix != 8 && radix != 10 && radix != 16) {
        *buf = '\000';
        return;
    }

    if (!num) {
        *buf++ = '0';
        *buf = '\000';
        return;
    }

    if (radix == 16) {
        *buf++ = '0';
        *buf++ = 'x';
    }
    else if (radix == 8) {
        *buf++ = '0';
    }

    n = 0;
    tmp = num;
    while (tmp) {
        tmp /= radix;
        n++;
    }

    i = n;

    /* Again we can't use g_assert; actually this check should _never_ fail. */
    if (n > FORMAT_UNSIGNED_BUFSIZE - 3) {
        *buf = '\000';
        return;
    }

    while (num) {
        i--;
        c = (num % radix);
        if (c < 10) {
            buf[i] = c + '0';
        }
        else {
            buf[i] = c + 'a' - 10;
        }
        num /= radix;
    }

    buf[n] = '\000';
}

static void
escape_string(GString *string)
{
    const char *p = string->str;
    gunichar wc;

    while (p < string->str + string->len) {
        gboolean safe;

        wc = g_utf8_get_char_validated(p, -1);
        if (wc == (gunichar)-1 || wc == (gunichar)-2) {
            gchar *tmp;
            guint pos;

            pos = p - string->str;

            /* Emit invalid UTF-8 as hex escapes
             */
            tmp = g_strdup_printf("\\x%02x", (guint)(guchar)*p);
            g_string_erase(string, pos, 1);
            g_string_insert(string, pos, tmp);

            p = string->str + (pos + 4); /* Skip over escape sequence */

            g_free (tmp);
            continue;
        }
        if (wc == '\r') {
            safe = *(p + 1) == '\n';
        }
        else {
            safe = CHAR_IS_SAFE (wc);
        }

        if (!safe) {
            gchar *tmp;
            guint pos;

            pos = p - string->str;

            /* Largest char we escape is 0x0a, so we don't have to worry
             * about 8-digit \Uxxxxyyyy
             */
            tmp = g_strdup_printf("\\u%04x", wc);
            g_string_erase(string, pos, g_utf8_next_char (p) - p);
            g_string_insert(string, pos, tmp);
            g_free(tmp);

            p = string->str + (pos + 6); /* Skip over escape sequence */
        }
        else {
            p = g_utf8_next_char(p);
        }
    }
}

static gchar*
strdup_convert(const gchar *string, const gchar *charset)
{
    if (!g_utf8_validate(string, -1, NULL)) {
        GString *gstring = g_string_new("[Invalid UTF-8] ");
        guchar *p;

        for (p = (guchar *)string; *p; p++) {
            if (CHAR_IS_SAFE(*p)
                && !(*p == '\r' && *(p + 1) != '\n')
                && *p < 0x80) {
                g_string_append_c(gstring, *p);
            }
            else {
                g_string_append_printf(gstring, "\\x%02x", (guint)(guchar)*p);
            }
        }

        return g_string_free(gstring, FALSE);
    }
    else {
        GError *err = NULL;

        gchar *result = g_convert_with_fallback(
            string, -1, charset, "UTF-8", "?", NULL, NULL, &err);
        if (result) {
            return result;
        }
        else {
            /* Not thread-safe, but doesn't matter if we print the
             * warning twice
             */
            static gboolean warned = FALSE;
            if (!warned) {
                warned = TRUE;
                fputs("async_syslog: Cannot convert message: ", stderr);
                fputs(err->message, stderr);
                fputs("\n", stderr);
            }
            g_error_free(err);

            return g_strdup(string);
        }
    }
}

static void
write_string(gint fd, const gchar *str)
{
    write(fd, str, strlen(str));
}

static void
log_fallback_handler(const gchar *log_domain, GLogLevelFlags log_level,
                     const gchar *message, gpointer reserved)
{
    gchar level_prefix[STRING_BUFFER_SIZE];
    gboolean is_fatal = (log_level & G_LOG_FLAG_FATAL) != 0;

    /* we can not call _any_ GLib functions in this fallback handler,
     * which is why we skip UTF-8 conversion, etc.
     * since we either recursed or ran out of memory, we're in a pretty
     * pathologic situation anyways, what we can do is giving the
     * the process ID unconditionally however.
     */

    gint fd = mklevel_prefix(level_prefix, log_level);
    if (!message) {
        message = "(NULL) message";
    }

    if (log_domain) {
        write_string(fd, "\n");
    }
    else {
        write_string(fd, "\n** ");
    }

    if (log_domain) {
        write_string(fd, log_domain);
        write_string(fd, "-");
    }
    write_string(fd, level_prefix);
    write_string(fd, ": ");
    write_string(fd, message);
    if (is_fatal) {
        write_string(fd, "\naborting...\n");
    }
    else {
        write_string(fd, "\n");
    }
}
