// idstring.c
/**
 * SECTION: idstring
 * @short_description: ID string generator
 * @stability: unstable
 * @include: idstring.h
 *
 * Manager for unique W3C XML Schema ID type values. #IdStore instance
 * is used to allocate and free ID values with any given prefix. Of
 * course, values will only be unique within the set of IDs that is
 * known to the #IdStore instance.
 */
#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <idstring.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <wcbe_util.h>

/**
 * struct _IdStore:
 * @prefix: ID prefix
 */
struct _IdStore {
    gchar *prefix;
    guint prefix_len;
    GHashTable *ids;
    guint16 v_min;
};

/**
 * idstring_new_store:
 * @prefix: prefix of generated ID values; %NULL indicates that no
 *          prefix should be used
 *
 * Create an #IdStore instance that can be used to generate ID values
 * with a given prefix.
 *
 * Returns: new #IdStore instance
 */
IdStore *
idstring_new_store(const gchar *prefix)
{
    IdStore *result = g_new(IdStore, 1);
    if (prefix != NULL) {
        result->prefix = g_strdup_printf("%s.", prefix);
    }
    else {
        result->prefix = g_strdup("");
    }
    result->prefix_len = strlen(result->prefix);
    result->ids = g_hash_table_new_full(NULL, NULL, NULL, g_free);
    result->v_min = 0;
    return result;
}

/**
 * idstring_reserve_ids:
 * @store: pointer to #IdStore instance
 * @id: id string to reserve
 *
 * Reserve given ID to prevent the ID generator from duplicating any
 * of those values. This function should be called immediately after
 * creating the #IdStore instance @store as many times as required;
 * this function does nothing if any IDs have already been generated
 * by @store.
 */
void
idstring_reserve_id(IdStore *store, const gchar *id)
{
    g_return_if_fail(store != NULL);
    g_return_if_fail(id != NULL);

    if (g_hash_table_size(store->ids) == 0) {
        if (g_str_has_prefix(id, store->prefix)) {
            // try to convert ID suffix from ASCII-encoded hex digit
            // string to integer
            gulong v = strtoul((char *)id + store->prefix_len, NULL, 16);
            if (errno == 0 && v <= G_MAXUINT16) {
                guint v16 = v;
                // since generated IDs exclusively use uppercase
                // letters for hex digits, convert the integer found
                // in the doc ID to an uppercase hex string, and
                // compare that with the doc ID suffix
                gchar *vstr = g_strdup_printf("%04X", v16);
                if (strcmp((char *)id + store->prefix_len, vstr) == 0) {
                    // reserve the doc ID number with a NULL pointer
                    // for the hash table value (g_free called with
                    // NULL is OK when the entry is removed)
                    g_hash_table_insert(store->ids, GUINT_TO_POINTER(v16),
                                        NULL);
                }
            }
        }
    }
    else {
        g_info("IDs have already been assigned - "
               "unable to reserve document IDs");
    }
}

/**
 * idstring_new_id:
 * @store: pointer to #IdStore instance
 *
 * Generate a unique ID string. The returned string should not be
 * modified, but it may be copied. The format of generated IDs is the
 * assigned prefix followed by an eight character ASCII-encoded 16 bit
 * number.
 *
 * Returns: ID string
 */
const gchar *
idstring_new_id(IdStore *store)
{
    g_return_val_if_fail(store != NULL, NULL);

    guint16 v = store->v_min;
    gchar *str = g_hash_table_lookup(store->ids, GSIZE_TO_POINTER(v));
    while (str != NULL) {
        str = g_hash_table_lookup(store->ids, GSIZE_TO_POINTER(++v));
    };
    gchar *result = g_strdup_printf("%s%04X", store->prefix, v);
    g_hash_table_insert(store->ids, GSIZE_TO_POINTER(v), result);
    store->v_min = ++v;
    return result;
}

/**
 * idstring_destroy_store:
 * @store: pointer to #IdStore instance
 *
 * Free the resources used by an #IdStore instance.
 */
void
idstring_destroy_store(IdStore *store)
{
    g_return_if_fail(store != NULL);

    g_hash_table_destroy(store->ids);
    g_free(store->prefix);
    g_free(store);
}
