// log_queue.c
//
// Asynchronous queue with size limit.
//
/**
 * SECTION: log_queue
 * @short_description: Asynchronous queue with size limit
 * @stability: unstable
 * @include: log_queue.h
 *
 * Akin to #GAsyncQueue, but with a limited number of messages allowed
 * in the queue. The primary intent of the design is to support the
 * async_syslog module.
 */
#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <log_queue.h>

/**
 * log_queue_t:
 * @q: #GQueue
 * @refcount: reference count
 * @size: maximum number of elements allowed in queue
 * @lost: number of messages that were not able to be pushed onto the
 *        queue since last pushed message
 * @drop_msgs: flag controlling whether any messages may be pushed onto queue
 * @mtx: mutex to protect access to all structure members except @lost
 * @empty_cond: #GCond variable to signal empty queue
 * @is_empty: flag indicating whether queue is empty
 *
 * Log queue structure. All fields are private.
 */
struct _log_queue {
    /*<private>*/
    GQueue *q;
    guint refcount;
    gint size;
    guint lost;
    gboolean drop_msgs;
    GMutex *mtx;
    GCond *empty_cond;
    GCond *nonempty_cond;
    gboolean is_empty;
};

/**
 * log_queue_new:
 * @size: maximum size of queue
 *
 * Create and initialize a new #log_queue_t instance with a given queue
 * size limit.
 *
 * Returns: #log_queue_t instance
 */
log_queue_t *
log_queue_new(guint size)
{
    g_return_val_if_fail(size <= G_MAXINT, NULL);
    log_queue_t *result = g_new0(log_queue_t, 1);
    result->size = size;
    result->q = g_queue_new();
    result->refcount = 1;
    result->drop_msgs = FALSE;
    result->mtx = g_mutex_new();
    result->empty_cond = g_cond_new();
    result->nonempty_cond = g_cond_new();
    result->lost = 0;
    result->is_empty = TRUE;
    return result;
}

/**
 * log_queue_ref:
 * @queue: pointer to #log_queue_t instance
 *
 * Increase the reference count of a #log_queue_t instance. Every
 * thread except the one that called log_queue_new() should call this
 * function before using the log queue instance.
 *
 * Returns: @queue
 */
log_queue_t *
log_queue_ref(log_queue_t *queue)
{
    g_return_val_if_fail(queue != NULL, NULL);
    g_mutex_lock(queue->mtx);
    queue->refcount++;
    g_mutex_unlock(queue->mtx);
    return queue;
}

/**
 * log_queue_unref:
 * @queue: pointer to #log_queue_t instance
 *
 * Decrease the reference count of a #log_queue_t instance. Every
 * thread should call this function after it's done using the log
 * queue instance.
 */
void
log_queue_unref(log_queue_t *queue)
{
    g_return_if_fail(queue != NULL);
    g_mutex_lock(queue->mtx);
    queue->refcount--;
    if (queue->refcount == 0) {
        // init 'elem' to any non-NULL pointer value
        log_queue_elem_t *elem = GINT_TO_POINTER(1);
        while (elem != NULL) {
            elem = g_queue_pop_head(queue->q);
            if (elem != NULL) {
                log_queue_elem_free(elem);
            }
        }
        g_queue_free(queue->q);
        g_cond_free(queue->empty_cond);
        g_cond_free(queue->nonempty_cond);
        GMutex *mtx = queue->mtx;
        queue->mtx = NULL;
        g_mutex_unlock(mtx);
        g_mutex_free(mtx);
        g_free(queue);
    }
    else {
        g_mutex_unlock(queue->mtx);
    }
}

/**
 * log_queue_drop:
 * @queue: pointer to #log_queue_t instance
 *
 * Set log queue so that it will drop every message pushed onto it.
 *
 * @see_also: log_queue_accept()
 */
void
log_queue_drop(log_queue_t *queue)
{
    g_return_if_fail(queue != NULL);
    g_mutex_lock(queue->mtx);
    queue->drop_msgs = TRUE;
    g_mutex_unlock(queue->mtx);
}

/**
 * log_queue_accept:
 * @queue: pointer to #log_queue_t instance
 *
 * Set log queue so that it will accept messages pushed onto it.
 *
 * @see_also: log_queue_drop()
 */
void
log_queue_accept(log_queue_t *queue)
{
    g_return_if_fail(queue != NULL);
    g_mutex_lock(queue->mtx);
    queue->drop_msgs = FALSE;
    g_mutex_unlock(queue->mtx);
}

/**
 * log_queue_wait_until_empty:
 * @queue: pointer to #log_queue_t instance
 *
 * Wait until (and return only after) log queue is empty.
 */
void
log_queue_wait_until_empty(log_queue_t *queue)
{
    g_return_if_fail(queue != NULL);
    g_mutex_lock(queue->mtx);
    while (!queue->is_empty) {
        g_cond_wait(queue->empty_cond, queue->mtx);
    }
    g_mutex_unlock(queue->mtx);
}

/**
 * log_queue_push:
 * @queue: pointer to #log_queue_t instance
 * @priority: <function>syslog</function>-style message priority
 * @msg: log message; string must be dynamically allocated,
 *       and may only be freed by the caller if this function returns
 *       %FALSE
 *
 * Push a message onto the log queue.
 *
 * Returns: %TRUE, if message was successfully pushed onto queue;
 *          %FALSE, otherwise.
 */
gboolean
log_queue_push(log_queue_t *queue, gint priority, gchar *msg)
{
    g_return_val_if_fail(queue != NULL, FALSE);
    g_return_val_if_fail(msg != NULL, FALSE);

    gboolean result;
    g_mutex_lock(queue->mtx);
    if (!queue->drop_msgs && g_queue_get_length(queue->q) < queue->size) {
        log_queue_elem_t *elem = g_slice_new(log_queue_elem_t);
        elem->msg = msg;
        elem->priority = priority;
        elem->lost = queue->lost;
        queue->lost = 0;
        g_queue_push_tail(queue->q, elem);
        if (queue->is_empty) {
            queue->is_empty = FALSE;
            g_cond_signal(queue->nonempty_cond);
        }
        result = TRUE;
    }
    else {
        queue->lost++;
        if (queue->is_empty) {
            queue->is_empty = FALSE;
            g_cond_signal(queue->nonempty_cond);
        }
        result = FALSE;
    }
    g_mutex_unlock(queue->mtx);
    return result;
}

/**
 * log_queue_pop:
 * @queue: pointer to #log_queue_t instance
 *
 * Pop a message from the log queue. Function will block until a
 * message is available (an available message being either a log
 * message, a lost message count or both).
 *
 * Returns: #log_queue_elem_t from top of queue
 */
log_queue_elem_t *
log_queue_pop(log_queue_t *queue)
{
    g_return_val_if_fail(queue != NULL, NULL);

    g_mutex_lock(queue->mtx);
    log_queue_elem_t *result = g_queue_pop_head(queue->q);
    if (result == NULL) {
        if (queue->lost > 0) {
            result = g_slice_new(log_queue_elem_t);
            result->msg = NULL;
            result->lost = queue->lost;
            queue->lost = 0;
            // queue can not have been empty (block guarded by lost > 0), so
            // we can signal empty_cond now
            g_assert(!queue->is_empty);
            queue->is_empty = TRUE;
            g_cond_signal(queue->empty_cond);
        }
        else {
            // queue must be empty at this point
            g_assert(queue->is_empty);
            while (queue->is_empty) {
                g_cond_wait(queue->nonempty_cond, queue->mtx);
            }
            // queue must be non-empty at this point
            result = g_queue_pop_head(queue->q);
            g_assert(result != NULL);
            queue->is_empty =
                (g_queue_get_length(queue->q) == 0 && queue->lost == 0);
            if (queue->is_empty) {
                g_cond_signal(queue->empty_cond);
            }
        }
    }
    else {
        // queue can not have been empty
        g_assert(!queue->is_empty);
        if (g_queue_get_length(queue->q) == 0 && queue->lost == 0) {
            queue->is_empty = TRUE;
            g_cond_signal(queue->empty_cond);
        }
    }
    g_mutex_unlock(queue->mtx);
    return result;
}

/**
 * log_queue_try_pop:
 * @queue: pointer to #log_queue_t instance
 *
 * Try to pop a message from the log queue. Function returns a
 * message if available (an available message being either a log
 * message, a lost message count or both), otherwise %NULL.
 *
 * Returns: #log_queue_elem_t from top of queue if a message is available;
 *          otherwise, %NULL
 */
log_queue_elem_t *
log_queue_try_pop(log_queue_t *queue)
{
    g_return_val_if_fail(queue != NULL, NULL);

    g_mutex_lock(queue->mtx);
    log_queue_elem_t *result = g_queue_pop_head(queue->q);
    if (result == NULL) {
        if (queue->lost > 0) {
            result = g_slice_new(log_queue_elem_t);
            result->msg = NULL;
            result->lost = queue->lost;
            queue->lost = 0;
            // queue can not have been empty (block guarded by lost >
            // 0), so we can signal empty_cond now
            g_assert(!queue->is_empty);
            queue->is_empty = TRUE;
            g_cond_signal(queue->empty_cond);
        }
    }
    else {
        // queue can not have been empty
        g_assert(!queue->is_empty);
        if (g_queue_get_length(queue->q) == 0 && queue->lost == 0) {
            queue->is_empty = TRUE;
            g_cond_signal(queue->empty_cond);
        }
    }
    g_mutex_unlock(queue->mtx);
    return result;
}

/**
 * log_queue_timed_pop:
 * @queue: pointer to #log_queue_t instance
 * @end_time: latest time to wait for a message
 *
 * Try to pop a message from the log queue, blocking until, at latest,
 * a given time. Function returns a message if available (an available
 * message being either a log message, a lost message count or both),
 * otherwise %NULL.
 *
 * Returns: #log_queue_elem_t from top of queue if a message was
 *          available before the time @end_time;
 *          otherwise, %NULL
 */
log_queue_elem_t *
log_queue_timed_pop(log_queue_t *queue, GTimeVal *end_time)
{
    g_return_val_if_fail(queue != NULL, NULL);

    g_mutex_lock(queue->mtx);
    log_queue_elem_t *result = g_queue_pop_head(queue->q);
    if (result == NULL) {
        if (queue->lost > 0) {
            result = g_slice_new(log_queue_elem_t);
            result->msg = NULL;
            result->lost = queue->lost;
            queue->lost = 0;
            // queue can not have been empty (block guarded by lost >
            // 0), so we can signal empty_cond now
            g_assert(!queue->is_empty);
            queue->is_empty = TRUE;
            g_cond_signal(queue->empty_cond);
        }
        else {
            // queue must be empty at this point
            g_assert(queue->is_empty);
            gboolean no_timeout = TRUE;
            while (no_timeout && queue->is_empty) {
                no_timeout = g_cond_timed_wait(
                    queue->nonempty_cond, queue->mtx, end_time);
            }
            if (!queue->is_empty) {
                result = g_queue_pop_head(queue->q);
                g_assert(result != NULL);
                queue->is_empty =
                    (g_queue_get_length(queue->q) == 0 && queue->lost == 0);
                if (queue->is_empty) {
                    g_cond_signal(queue->empty_cond);
                }
            }
        }
    }
    else {
        // queue can not have been empty
        g_assert(!queue->is_empty);
        if (g_queue_get_length(queue->q) == 0 && queue->lost == 0) {
            queue->is_empty = TRUE;
            g_cond_signal(queue->empty_cond);
        }
    }
    g_mutex_unlock(queue->mtx);
    return result;
}

/**
 * log_queue_elem_free:
 * @elem: pointer to #log_queue_elem_t
 *
 * Free resources used by a #log_queue_elem_t instance.
 */
void
log_queue_elem_free(log_queue_elem_t *elem)
{
    g_return_if_fail(elem != NULL);
    if (elem->msg != NULL) {
        g_free(elem->msg);
    }
    g_slice_free(log_queue_elem_t, elem);
}
