#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <stddef.h>
#include <stdio.h>

#include <mdarray.h>

#if !GLIB_CHECK_VERSION(2, 6, 0)
# define G_GSIZE_FORMAT "u"
#endif

/* mdarray_count
 *
 * Number of elements needed for an 'mdarray_t' of the given
 * specification.
 */
guint64
mdarray_count(const mdarray_t *mdarray)
{
    g_return_val_if_fail(mdarray != NULL, 0);

    guint64 num_elements = 1;
    guint *dim = (guint *)(mdarray->dimensions->data);
    while (*dim != COORD_VECTOR_END) {
        num_elements *= *dim++;
    }
    return num_elements;
}

/* mdarray_new
 *
 * Create an 'mdarray_t'.
 *
 * *Arguments*
 *
 * dimensions - sizes of the array axes (COORD_VECTOR_END-terminated)
 *
 * *Return value*
 *
 * pointer to newly allocated 'mdarray_t'
 *
 */
mdarray_t *
mdarray_new(const guint *dimensions, gsize element_size)
{
    g_return_val_if_fail(dimensions != NULL, NULL);
    g_return_val_if_fail(element_size > 0, NULL);

    mdarray_t *result = g_new0(mdarray_t, 1);
    result->dimensions = g_array_new(FALSE, FALSE, sizeof(guint));
    while (*dimensions != COORD_VECTOR_END) {
        g_array_append_vals(result->dimensions, dimensions, 1);
        ++dimensions;
    }
    g_array_append_vals(result->dimensions, dimensions, 1);
    result->element_size = element_size;
    return result;
}

void
mdarray_assign_memory(mdarray_t *mdarray, gpointer buffer, off_t offset)
{
    g_return_if_fail(mdarray != NULL);
    g_return_if_fail(buffer != NULL);
    g_return_if_fail(offset >= 0);

    mdarray->stream = NULL;
    mdarray->buffer = buffer;
    mdarray->offset = offset;
}

void
mdarray_assign_stream(mdarray_t *mdarray, FILE *stream, off_t offset)
{
    g_return_if_fail(mdarray != NULL);
    g_return_if_fail(stream != NULL);
    g_return_if_fail(offset >= 0);

    mdarray->buffer = NULL;
    mdarray->stream = stream;
    mdarray->offset = offset;
}

void
mdarray_unassign(mdarray_t *mdarray)
{
    g_return_if_fail(mdarray != NULL);
    mdarray->buffer = NULL;
    mdarray->stream = NULL;
    mdarray->offset = 0;
}

/**
 * mdarray_foreach:
 * @mdarray: mdarray
 * @fn: function to apply
 * @user_data: pointer passed to @fn for every element
 *
 * Apply a function to every element of the array. Loop over elements
 * can be stopped at any point by returning #TRUE from @fn.
 */
void
mdarray_foreach(mdarray_t *mdarray, mdarray_fn_t fn, gpointer user_data)
{
    g_return_if_fail(mdarray != NULL);
    g_return_if_fail(fn != NULL);

    guint max_dimension = mdarray->dimensions->len - 2;
    guint coords[max_dimension + 2];
    memset(coords, 0, (max_dimension + 1) * sizeof(guint));
    coords[max_dimension + 1] = COORD_VECTOR_END;
    gpointer mem_value = NULL;
    gboolean memory_fn(guint *c, gpointer udata) {
        gboolean result = fn(c, mem_value, udata);
        mem_value += mdarray->element_size;
        return result;
    }
    gboolean stream_fn(guint *c, gpointer udata) {
        gchar value[mdarray->element_size];
        gboolean result = FALSE;
        if (fread_unlocked(value, mdarray->element_size, 1,
                           mdarray->stream) == 1) {
            result = fn(c, value, udata);
            if (fseeko(mdarray->stream, -mdarray->element_size, SEEK_CUR)
                == 0) {
                if (fwrite_unlocked(value, mdarray->element_size, 1,
                                    mdarray->stream) != 1) {
                    result = FALSE;
                    g_warning("Failed to fwrite mdarray element: %s",
                              g_strerror(errno));
                }
            }
            else {
                result = FALSE;
                g_warning("Failed to fseeko to mdarray element: %s",
                          g_strerror(errno));
            }
        }
        else {
            g_warning("Failed to fread mdarray element: %s",
                      g_strerror(errno));
        }
        return result;
    }
    gboolean (*mdfn)(guint *, gpointer);
    gboolean stop_loop = FALSE;
    if (mdarray->buffer != NULL) {
        mdfn = memory_fn;
        mem_value = mdarray->buffer + mdarray->offset;
    }
    else {
        g_assert(mdarray->stream != NULL);
        mdfn = stream_fn;
        flockfile(mdarray->stream);
        if (fseeko(mdarray->stream, mdarray->offset, SEEK_SET) != 0) {
            g_warning("Failed to fseeko to beginning of mdarray array: %s",
                      g_strerror(errno));
            stop_loop = TRUE;
        }
    }
    guint max_count = mdarray_count(mdarray);
    guint count;
    for (count = 0; !stop_loop && count < max_count; ++count) {
        stop_loop = mdfn(coords, user_data);
        coords[max_dimension]++;
        guint dim = max_dimension;
        while ((dim > 0)
               && G_UNLIKELY(
                   coords[dim]
                   >= g_array_index(mdarray->dimensions, guint, dim))) {
            coords[dim] = 0;
            coords[--dim]++;
        }
    }
    if (mdarray->stream != NULL) {
        fflush_unlocked(mdarray->stream);
        funlockfile(mdarray->stream);
    }
}

/* mdarray_free
 *
 * Free the memory used by an 'mdarray_t' instance.
 */
void
mdarray_free(mdarray_t *mdarray)
{
    g_return_if_fail(mdarray != NULL);

    if (mdarray->dimensions != NULL) {
        g_array_free(mdarray->dimensions, TRUE);
    }
    g_free(mdarray);
}

/**
 * mdarray_index_offset:
 * @mdarray: mdarray
 * @coordinates: array coordinates (COORD_VECTOR_END-terminated vector)
 *
 * @returns: the offset to the element at a given index of an 'mdarray_t'
 *           instance; -1, if coordinates are out of bounds
 */
off_t
mdarray_index_offset(const mdarray_t *mdarray, const guint *coordinates)
{
    g_return_val_if_fail(mdarray != NULL, -1);
    g_return_val_if_fail(coordinates != NULL, -1);

    off_t result = 0;
    guint *dim = &g_array_index(mdarray->dimensions, guint, 0);
    while (*dim != COORD_VECTOR_END && *coordinates != COORD_VECTOR_END
           && result != -1) {
        if (G_LIKELY(*coordinates < *dim)) {
            result *= *dim++;
            result += *coordinates++;
        }
        else {
            result = -1;
        }
    }
    if (G_LIKELY(*coordinates == COORD_VECTOR_END && result >= 0)) {
        while (*dim != COORD_VECTOR_END) {
            result *= *dim++;
        }
        result = result * mdarray->element_size;
    }
    else {
        result = -1;
    }
    return result;
}

/* mdarray_reduced
 *
 * Create a subarray of an 'mdarray_t' with a set of fixed leading
 * indices. Note that this is not a general subarray! Also, be aware
 * that the returned mdarray_t references the element buffer of the
 * parent mdarray_t; however, no reference counting is done at this
 * level, so you should not free the parent array before the returned
 * subarray.
 *
 * 'fixed' is a COORD_VECTOR_END-terminated array of coordinates
 */
mdarray_t *
mdarray_reduced(const mdarray_t *mdarray, const guint *fixed,
                off_t *element_offset)
{
    mdarray_t *result;
    off_t off = mdarray_index_offset(mdarray, fixed);
    if (off >= 0) {
        guint num_fixed = 0;
        while (fixed[num_fixed] != COORD_VECTOR_END) {
            ++num_fixed;
        }
        guint *dimensions =
            &g_array_index(mdarray->dimensions, guint, num_fixed);
        result = mdarray_new(dimensions, mdarray->element_size);
        if (element_offset != NULL) {
            *element_offset = off;
        }
    }
    else {
        result = NULL;
    }
    return result;
}
