/* sdmhdr.c
 *
 * Routines for reading and writing 'sdmDataHeader' elements of
 * the correlator output binary data format.
 *
 * The parser is meant to be very liberal with respect to what it will
 * accept in an XML document; missing elements are acceptable, and
 * undefined elements (and attributes) are simply ignored. The
 * returned sdmhdr_t instance, however, has enough information to
 * indicate missing elements, but not the existence of undefined
 * elements and attributes. The caller may inspect the returned data
 * structure to determine whether the document lacks required
 * elements. Alternatively, the caller may validate the document
 * beforehand.
 */
#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib.h>
#include <glib/gprintf.h>
#include <libxml/xmlreader.h>
#include <libxml/xmlwriter.h>
#include <sys/mman.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <sdmhdr.h>
#include <wcbe_util.h>

#if !GLIB_CHECK_VERSION(2, 6, 0)
# define G_GSIZE_FORMAT "u"
#endif

#define XML_PARSE_OPTIONS (XML_PARSE_NOENT | XML_PARSE_NOBLANKS \
                           | XML_PARSE_NONET | XML_PARSE_NOCDATA)

typedef gint (*element_parser_fn_t)(xmlTextReaderPtr reader,
                                    gpointer header,
                                    gpointer user_data);

typedef gint (*attribute_parser_fn_t)(gchar *str, gpointer val);

typedef struct element {
    const gchar *nsuri;
    const gchar *name;
    element_parser_fn_t parser;
    gpointer user_data;
} element_t;

typedef struct attribute {
    const gchar *nsuri;
    const gchar *name;
    attribute_parser_fn_t parser;
    gpointer val;
} attribute_t;

typedef gint (*hdr_write_fn_t)(xmlTextWriterPtr writer, gconstpointer header);

/* XLink namespace URI */
static const gchar const xlink_uri[] = "http://www.w3.org/1999/xlink";
static const gchar const xlink_prefix[] = "xl";
/* XMLSchema namespace URI */
static const gchar const xsi_uri[] =
    "http://www.w3.org/2001/XMLSchema-instance";
static const gchar const xsi_prefix[] = "xsi";
/* SDM version URI */
static const gchar const xvers_uri[] = "http://TBD/XVERSION";
static const gchar const xvers_prefix[] = "xv";
/* SDM BDF URI */
static const gchar const bdf_uri[] = "http://TBD/XSDM/sdmbin";

/* Private function prototypes */
static sdmhdr_main_hdr_t *start_main_hdr_read(xmlTextReaderPtr reader);
static sdmhdr_subset_hdr_t *start_subset_hdr_read(xmlTextReaderPtr reader);

static gint parse_attributes(xmlTextReaderPtr reader,
                             const attribute_t attributes[]);
static gint parse_element(xmlTextReaderPtr reader, gpointer header,
                          const element_t children[],
                          const attribute_t attributes[],
                          gchar **text);

static gint parse_byteOrder(gchar *str, gpointer val);
static gint parse_dataOID(xmlTextReaderPtr reader, gpointer header,
                          gpointer user_data);
static gint parse_dimensionality(xmlTextReaderPtr reader, gpointer header,
                                 gpointer user_data);
static gint parse_axes(gchar *str, gpointer val);
static gint parse_execBlock(xmlTextReaderPtr reader, gpointer header,
                            gpointer user_data);
static gint parse_numAntenna(xmlTextReaderPtr reader, gpointer header,
                             gpointer user_data);
static gint parse_correlationMode(xmlTextReaderPtr reader, gpointer header,
                                  gpointer user_data);
static gint parse_spectralResolution(xmlTextReaderPtr reader, gpointer header,
                                     gpointer user_data);
static gint parse_main_bincomp(xmlTextReaderPtr reader, gpointer header,
                               gpointer user_data);
static gint parse_weights(xmlTextReaderPtr reader, gpointer header,
                          gpointer user_data);

static gint parse_main_dataStruct(xmlTextReaderPtr reader, gpointer header,
                                  gpointer user_data);
static gint parse_apc(gchar *str, gpointer val);
static gint parse_baseband(xmlTextReaderPtr reader, gpointer header,
                           gpointer data);
static gint parse_spectralWindow(xmlTextReaderPtr reader,
                                 gpointer header, gpointer data);
static void normalize_list_whitespace(gchar *str);
static gint parse_sdPolProducts(gchar *str, gpointer val);
static gint parse_crossPolProducts(gchar *str, gpointer val);

static gint parse_string(gchar *str, gpointer val);
static gint parse_uint(gchar *str, gpointer val);
static gint parse_size(gchar *str, gpointer val);

static gint parse_uint64_content(xmlTextReaderPtr reader, gpointer header,
                                 gpointer user_data);
static gint parse_schedulePeriodTime(xmlTextReaderPtr reader, gpointer header,
                                     gpointer user_data);
static gint parse_subset_dataStruct(xmlTextReaderPtr reader, gpointer header,
                                    gpointer user_data);
static gint parse_subset_bincomp(xmlTextReaderPtr reader, gpointer header,
                                 gpointer user_data);
static gint parse_subset_crossData(xmlTextReaderPtr reader, gpointer header,
                                   gpointer user_data);
static gint parse_primitive_data_type(gchar *str, gpointer val);

static xmlBufferPtr sdmhdr_hdr_write(gconstpointer header,
                                     hdr_write_fn_t write_fn, gboolean indent);

static xmlBufferPtr sdmhdr_main_hdr_write(const sdmhdr_main_hdr_t *header,
                                          gboolean indent);
static xmlBufferPtr sdmhdr_subset_hdr_write(const sdmhdr_subset_hdr_t *header,
                                            gboolean indent);
static gint start_main_hdr_write(xmlTextWriterPtr writer,
                                 const sdmhdr_main_hdr_t *header);
static gint start_subset_hdr_write(xmlTextWriterPtr writer,
                                   const sdmhdr_subset_hdr_t *header);
static gsize write_memory(const xmlBufferPtr xml, gchar **buff);
static gsize write_fd(const xmlBufferPtr xml, gint fd);
static gint write_startTime(xmlTextWriterPtr writer,
                            const sdmhdr_main_hdr_t *header);
static gint write_dataOID(xmlTextWriterPtr writer,
                          const sdmhdr_main_hdr_t *header);
static gint write_dimensionality(xmlTextWriterPtr writer,
                                 const sdmhdr_main_hdr_t *header);
static gint write_execBlock(xmlTextWriterPtr writer,
                            const sdmhdr_main_hdr_t *header);
static gint write_numAntenna(xmlTextWriterPtr writer,
                             const sdmhdr_main_hdr_t *header);
static gint write_correlationMode(xmlTextWriterPtr writer,
                                  const sdmhdr_main_hdr_t *header);
static gint write_spectralResolution(xmlTextWriterPtr writer,
                                     const sdmhdr_main_hdr_t *header);
static gint write_main_dataStruct(xmlTextWriterPtr writer,
                                  const sdmhdr_main_hdr_t *header);
static gint write_main_bincomp(xmlTextWriterPtr writer, const gchar *name,
                               const sdmhdr_bincomp_t *tree,
                               gboolean end_element);
static gint write_weights(xmlTextWriterPtr writer,
                          const sdmhdr_weights_t *weights);
#ifndef USE_ALMA_5_1_0_5_XML_WORKAROUNDS
static gint write_baseband(xmlTextWriterPtr writer,
                           const sdmhdr_bb_t *baseband);
#else
static gint write_baseband(xmlTextWriterPtr writer,
                           const sdmhdr_bb_t *baseband, guint i);
#endif
static gint write_spectralWindow(xmlTextWriterPtr writer,
                                 const sdmhdr_sw_t *sw);
static gint write_schedulePeriodTime(xmlTextWriterPtr writer,
                                     const sdmhdr_subset_hdr_t *header);
static gint write_uint64_element(xmlTextWriterPtr writer, const gchar *element,
                                 guint64 value);
static gint write_subset_dataStruct(xmlTextWriterPtr writer,
                                    const sdmhdr_subset_hdr_t *header);
static gint write_subset_bincomp(xmlTextWriterPtr writer,
                                 const gchar *element, const gchar *href);
static gint write_subset_crossData(xmlTextWriterPtr writer,
                                   const sdmhdr_subset_hdr_t *header);

static void set_bincomp_size(const sdmhdr_main_hdr_t *header,
                             sdmhdr_bincomp_t *btree,
                             guint par_mult, guint cross_mult);
static gsize num_elements(const sdmhdr_main_hdr_t *header,
                          const gboolean *axes,
                          guint parallel_pp_mult, guint cross_pp_mult);
static sdmhdr_bb_t *maximal_baseband(const sdmhdr_bb_t **basebands,
                                     guint parallel_mult, guint cross_mult);
static void maximal_spectral_window(const sdmhdr_bb_t *baseband,
                                    sdmhdr_sw_t *max_sw,
                                    guint parallel_mult, guint cross_mult);
static guint num_cross_primitives(sdmhdr_xpp_t xpp, guint parallel_mult,
                                  guint cross_mult);
static guint num_auto_primitives(sdmhdr_app_t app, guint parallel_mult,
                                 guint cross_mult);
static void *mmap_xml_element(gint fd, off_t off, const gchar *name,
                              gsize *len, gchar **element);
static gint positioned_fd(FILE *file);

const gchar const *sdmhdr_encoding = "utf-8";

/**
 * sdmhdr_main_hdr_read_memory:
 * @buff: pointer to XML doc string
 * @bufflen: length of XML doc string
 *
 * Create sdmhdr_main_hdr_t structure from string. Parsed element is
 * sdmDataHeader.
 *
 * @returns: sdmhdr_main_hdr_t instance
 */
sdmhdr_main_hdr_t *
sdmhdr_main_hdr_read_memory(const gchar *buff, gsize bufflen)
{
    sdmhdr_main_hdr_t *result = NULL;
    xmlTextReaderPtr reader =
        xmlReaderForMemory(buff, bufflen, "", NULL, XML_PARSE_OPTIONS);
    if (reader != NULL) {
        result = start_main_hdr_read(reader);
        xmlFreeTextReader(reader);
    }
    return result;
}

/**
 * sdmhdr_main_hdr_read_stream:
 * @fd: stream from which to read XML doc
 *
 * Create sdmhdr_main_hdr_t structure from file contents. Parsed
 * element is sdmDataHeader.
 *
 * Reading data from the stream begins from the current file offset,
 * and ends wherever the 'sdmDataHeader' element ends (or immediately
 * after an error, whatever that means).
 *
 * @returns: sdmhdr_main_hdr_t instance
 */
sdmhdr_main_hdr_t *
sdmhdr_main_hdr_read_stream(FILE *file)
{
    sdmhdr_main_hdr_t *result = NULL;
    gint fd = positioned_fd(file);
    if (fd != -1) {
        gchar *element;
        gsize bufflen;
        void *buff = mmap_xml_element(fd, ftello(file), "sdmDataHeader",
                                      &bufflen, &element);
        if (buff != NULL) {
            xmlTextReaderPtr reader = xmlReaderForMemory(
                element, strlen(element), NULL, NULL, XML_PARSE_OPTIONS);
            if (reader != NULL) {
                result = start_main_hdr_read(reader);
                xmlFreeTextReader(reader);
            }
            gint rc = munmap(buff, bufflen);
            if (rc == -1) {
                g_critical("Failed to munmap 'sdmDataHeader' element: %s",
                           g_strerror(errno));
            }
        }
        else {
            g_message("Failed to find and memory map 'sdmDataHeader' "
                      "element");
        }
    }
    return result;
}

sdmhdr_main_hdr_t *
sdmhdr_main_hdr_copy(const sdmhdr_main_hdr_t *header)
{
    g_return_val_if_fail(header != NULL, NULL);

    sdmhdr_main_hdr_t *result = g_memdup(header, sizeof(sdmhdr_main_hdr_t));
    result->revision = g_strdup(header->revision);
    result->main_header_id = g_strdup(header->main_header_id);
    result->project_path = g_strdup(header->project_path);

    if (header->data_oid != NULL) {
        result->data_oid = g_new0(sdmhdr_dataoid_t, 1);
        result->data_oid->href = g_strdup(header->data_oid->href);
        result->data_oid->title = g_strdup(header->data_oid->title);
        result->data_oid->type = g_strdup(header->data_oid->type);
    }

    if (header->exec_block != NULL) {
        result->exec_block = g_new0(sdmhdr_execblock_t, 1);
        result->exec_block->href = g_strdup(header->exec_block->href);
        result->exec_block->type = g_strdup(header->exec_block->type);
    }

    if (header->dimensionality != NULL) {
        guint n = 1; // start with 1 to account for COBDF_INVALID_AXIS entry
        CobdfAxisName *name = header->dimensionality;
        while (*name++ != COBDF_INVALID_AXIS) {
            ++n;
        }
        result->dimensionality =
            g_memdup(header->dimensionality, n * sizeof(*name));
    }

    if (header->data_struct != NULL) {
        result->data_struct = g_new0(sdmhdr_data_struct_t, 1);
        result->data_struct->type = g_strdup(header->data_struct->type);
        memcpy(&(result->data_struct->apc), &(header->data_struct->apc),
               sizeof(result->data_struct->apc));

        void dup_bincomp(glong offset) {
            sdmhdr_bincomp_t *header_bc = G_STRUCT_MEMBER(
                sdmhdr_bincomp_t *, header->data_struct, offset);
            if (header_bc != NULL) {
                sdmhdr_bincomp_t **result_bc =
                    G_STRUCT_MEMBER_P(result->data_struct, offset);
                *result_bc = g_memdup(header_bc, sizeof(sdmhdr_bincomp_t));
            }
        }
#define DUP_BINCOMP(name) \
    dup_bincomp(G_STRUCT_OFFSET(sdmhdr_data_struct_t, name))
        DUP_BINCOMP(flags);
        DUP_BINCOMP(actual_times);
        DUP_BINCOMP(actual_durations);
        DUP_BINCOMP(zero_lags);
        DUP_BINCOMP(auto_data);
        DUP_BINCOMP(cross_data);
#undef DUP_BINCOMP

        if (header->data_struct->weights != NULL) {
            result->data_struct->weights =
                g_memdup(header->data_struct->weights,
                         SIZEOF_SDMHDR_WEIGHTS(
                             header->data_struct->weights->num_weights));
        }

        if (header->data_struct->basebands != NULL) {
            guint n = num_basebands(header->data_struct->basebands);
            result->data_struct->basebands = g_new0(sdmhdr_bb_t *, n + 1);
            while (n-- > 0) {
                result->data_struct->basebands[n] =
                    g_memdup(header->data_struct->basebands[n],
                             SIZEOF_SDMHDR_BB(
                                 header->data_struct->basebands[n]->num_sw));
            }
        }
    }
    return result;
}

/**
 * sdmhdr_main_hdr_write_memory:
 * @header: main data header
 * @buff: where XML document string buffer pointer should be written
 *
 * Write an XML document string with one 'sdmDataHeader' element from
 * an sdmhdr_main_hdr_t instance.
 *
 * @returns: Size of allocated buffer for XML document.
 */
gsize
sdmhdr_main_hdr_write_memory(const sdmhdr_main_hdr_t *header, gchar **buff,
                             gboolean indent)
{
    return write_memory(sdmhdr_main_hdr_write(header, indent), buff);
}

/**
 * sdmhdr_main_hdr_write_stream:
 * @header: main data header
 * @file: BDF stream
 *
 * Write to a stream an XML document with one 'sdmDataHeader' element
 * from an sdmhdr_main_hdr_t instance.
 *
 * @returns: Number of characters written to the stream.
 */
gsize
sdmhdr_main_hdr_write_stream(const sdmhdr_main_hdr_t *header, FILE *file,
                             gboolean indent)
{
    gsize result = -1;
    gint fd = positioned_fd(file);
    if (fd != -1) {
        result = write_fd(sdmhdr_main_hdr_write(header, indent), fd);
    }
    return result;
}

static gsize
write_memory(const xmlBufferPtr xml, gchar **buff)
{
    gsize result = 0;
    *buff = NULL;
    if (xml != NULL) {
        result = xml->use + 1;
        *buff = g_malloc(result);
        memcpy(*buff, xml->content, result);
        xmlBufferFree(xml);
    }
    return result;
}

static gsize
write_fd(const xmlBufferPtr xml, gint fd)
{
    gsize result = 0;
    if (xml != NULL) {
        gssize nc;
        do {
            nc = write(fd, xml->content + result, xml->use - result);
            if (nc >= 0) {
                result += nc;
            }
        } while (nc >= 0 && result < xml->use);
        xmlBufferFree(xml);
    }
    return result;
}

static xmlBufferPtr
sdmhdr_main_hdr_write(const sdmhdr_main_hdr_t *header, gboolean indent)
{
    return sdmhdr_hdr_write(header, (hdr_write_fn_t)start_main_hdr_write,
                            indent);
}

static xmlBufferPtr
sdmhdr_subset_hdr_write(const sdmhdr_subset_hdr_t *header, gboolean indent)
{
    return sdmhdr_hdr_write(header, (hdr_write_fn_t)start_subset_hdr_write,
                            indent);
}

static xmlBufferPtr
sdmhdr_hdr_write(gconstpointer header,
                 gint (*hdr_write_fn)(xmlTextWriterPtr, gconstpointer),
                 gboolean indent)
{
    xmlBufferPtr result = xmlBufferCreate();
    if (result != NULL) {
        xmlTextWriterPtr writer = xmlNewTextWriterMemory(result, 0);
        if (writer != NULL) {
            gint rc = xmlTextWriterSetIndent(writer, (indent ? 1 : 0));
            if (rc != 0) {
                g_info("Failed to set 'indent' value of xmlTextWriterPtr");
            }
            rc = xmlTextWriterStartDocument(writer, NULL, sdmhdr_encoding,
                                            "yes");
            if (rc >= 0) {
                int rc = hdr_write_fn(writer, header);
                if (rc >= 0) {
                    rc = xmlTextWriterEndDocument(writer);
                    if (rc >= 0) {
                        rc = xmlTextWriterFlush(writer);
                    }
                }
            }
            xmlFreeTextWriter(writer);
            if (rc < 0) {
                xmlBufferFree(result);
                result = NULL;
            }
        }
    }
    return result;
}

/* sdmhdr_main_hdr_free:
 * @header: sdmhdr_main_hdr_t instance
 *
 * Free resources used by an sdmhdr_main_hdr_t instance.
 */
void
sdmhdr_main_hdr_free(sdmhdr_main_hdr_t *header)
{
    if (header != NULL) {
        if (header->data_oid != NULL) {
            if (header->data_oid->href != NULL) {
                g_free(header->data_oid->href);
                g_free(header->data_oid->title);
                g_free(header->data_oid->type);
            }
            g_free(header->data_oid);
        }
        if (header->data_struct != NULL) {
            sdmhdr_data_struct_t *ds = header->data_struct;
            if (ds->flags != NULL) {
                g_free(ds->flags);
            }
            if (ds->actual_times != NULL) {
                g_free(ds->actual_times);
            }
            if (ds->actual_durations != NULL) {
                g_free(ds->actual_durations);
            }
            if (ds->zero_lags != NULL) {
                g_free(ds->zero_lags);
            }
            if (ds->cross_data != NULL) {
                g_free(ds->cross_data);
            }
            if (ds->auto_data != NULL) {
                g_free(ds->auto_data);
            }
            if (ds->weights != NULL) {
                g_free(ds->weights);
            }
            if (ds->basebands != NULL) {
                sdmhdr_bb_t **bb = ds->basebands;
                while (*bb != NULL) {
                    guint i;
                    for (i = 0; i < (*bb)->num_sw; ++i) {
                        if ((*bb)->sw[i].id != NULL) {
                            g_free((*bb)->sw[i].id);
                        }
                    }
                    g_free(*bb);
                    ++bb;
                }
                g_free(ds->basebands);
            }
            if (ds->type != NULL) {
                g_free(ds->type);
            }
            g_free(ds);
        }
        if (header->exec_block != NULL) {
            if (header->exec_block->href != NULL) {
                g_free(header->exec_block->href);
            }
            if (header->exec_block->type != NULL) {
                g_free(header->exec_block->type);
            }
            g_free(header->exec_block);
        }
        if (header->revision != NULL) {
            g_free(header->revision);
        }
        if (header->main_header_id != NULL) {
            g_free(header->main_header_id);
        }
        if (header->project_path != NULL) {
            g_free(header->project_path);
        }
        if (header->dimensionality != NULL) {
            g_free(header->dimensionality);
        }
        g_free(header);
    }
}

sdmhdr_subset_hdr_t *
sdmhdr_subset_hdr_read_memory(const gchar *buff, gsize bufflen)
{
    sdmhdr_subset_hdr_t *result = NULL;
    xmlTextReaderPtr reader =
        xmlReaderForMemory(buff, bufflen, "", NULL, XML_PARSE_OPTIONS);
    if (reader != NULL) {
        result = start_subset_hdr_read(reader);
        xmlFreeTextReader(reader);
    }
    return result;
}

sdmhdr_subset_hdr_t *
sdmhdr_subset_hdr_read_stream(FILE *file)
{
    sdmhdr_subset_hdr_t *result = NULL;
    gint fd = positioned_fd(file);
    if (fd != -1) {
        gchar *element;
        gsize bufflen;
        off_t pos = ftello(file);
        void *buff = mmap_xml_element(
            fd, pos, "sdmDataSubsetHeader", &bufflen, &element);
        if (buff != NULL) {
            xmlTextReaderPtr reader = xmlReaderForMemory(
                element, strlen(element), NULL, NULL, XML_PARSE_OPTIONS);
            if (reader != NULL) {
                result = start_subset_hdr_read(reader);
                xmlFreeTextReader(reader);
            }
            gint rc = munmap(buff, bufflen);
            if (rc == -1) {
                g_critical("Failed to munmap 'sdmDataSubsetHeader' "
                           "element: %s", g_strerror(errno));
            }
        }
        else {
            g_message("Failed to find and memory map '"
                      "sdmDataSubsetHeader' element");
        }
        fseeko(file, pos, SEEK_SET);
    }
    return result;
}

sdmhdr_subset_hdr_t *
sdmhdr_subset_hdr_copy(const sdmhdr_subset_hdr_t *header)
{
    sdmhdr_subset_hdr_t *result =
        g_memdup(header, sizeof(sdmhdr_subset_hdr_t));
    void dup_string(glong offset) {
        gchar *header_str = G_STRUCT_MEMBER(gchar *, header, offset);
        if (header_str != NULL) {
            gchar **result_str = G_STRUCT_MEMBER_P(result, offset);
            *result_str = g_strdup(header_str);
        }
    }
#define DUP_STRING(name) dup_string(G_STRUCT_OFFSET(sdmhdr_subset_hdr_t, name))
    DUP_STRING(type);
    DUP_STRING(project_path);
    DUP_STRING(data_struct_ref);
    DUP_STRING(flags_href);
    DUP_STRING(actual_times_href);
    DUP_STRING(actual_durations_href);
    DUP_STRING(weights_href);
    DUP_STRING(zero_lags_href);
    DUP_STRING(cross_data_href);
    DUP_STRING(auto_data_href);
#undef DUP_STRING
    return result;
}

gsize
sdmhdr_subset_hdr_write_memory(const sdmhdr_subset_hdr_t *header,
                               gchar **buff, gboolean indent)
{
    return write_memory(sdmhdr_subset_hdr_write(header, indent), buff);
}

gsize
sdmhdr_subset_hdr_write_stream(const sdmhdr_subset_hdr_t *header, FILE *file,
                               gboolean indent)
{
    gsize result = -1;
    gint fd = positioned_fd(file);
    if (fd != -1) {
        result = write_fd(sdmhdr_subset_hdr_write(header, indent), fd);
    }
    return result;
}

void
sdmhdr_subset_hdr_free(sdmhdr_subset_hdr_t *header)
{
    if (header != NULL) {
        if (header->type != NULL) {
            g_free(header->type);
        }
        if (header->project_path != NULL) {
            g_free(header->project_path);
        }
        if (header->data_struct_ref != NULL) {
            g_free(header->data_struct_ref);
        }
        if (header->flags_href != NULL) {
            g_free(header->flags_href);
        }
        if (header->actual_times_href != NULL) {
            g_free(header->actual_times_href);
        }
        if (header->actual_durations_href != NULL) {
            g_free(header->actual_durations_href);
        }
        if (header->weights_href != NULL) {
            g_free(header->weights_href);
        }
        if (header->zero_lags_href != NULL) {
            g_free(header->zero_lags_href);
        }
        if (header->cross_data_href != NULL) {
            g_free(header->cross_data_href);
        }
        if (header->auto_data_href != NULL) {
            g_free(header->auto_data_href);
        }
        g_free(header);
    }
}

void
sdmhdr_set_flags_size(sdmhdr_main_hdr_t *header)
{
    g_return_if_fail(header != NULL);
    g_return_if_fail(header->data_struct != NULL);
    g_return_if_fail(header->data_struct->flags != NULL);

    set_bincomp_size(header, header->data_struct->flags, 1, 1);
}

void
sdmhdr_set_actual_times_size(sdmhdr_main_hdr_t *header)
{
    g_return_if_fail(header != NULL);
    g_return_if_fail(header->data_struct != NULL);
    g_return_if_fail(header->data_struct->actual_times != NULL);

    set_bincomp_size(header, header->data_struct->actual_times, 1, 1);
}

void
sdmhdr_set_actual_durations_size(sdmhdr_main_hdr_t *header)
{
    g_return_if_fail(header != NULL);
    g_return_if_fail(header->data_struct != NULL);
    g_return_if_fail(header->data_struct->actual_durations != NULL);

    set_bincomp_size(header, header->data_struct->actual_durations, 1, 1);
}

void
sdmhdr_set_weights_size(sdmhdr_main_hdr_t *header)
{
    g_return_if_fail(header != NULL);
    g_return_if_fail(header->data_struct != NULL);
    g_return_if_fail(header->data_struct->weights != NULL);

    set_bincomp_size(header, &(header->data_struct->weights->common), 1, 1);
}

void
sdmhdr_set_zero_lags_size(sdmhdr_main_hdr_t *header)
{
    g_return_if_fail(header != NULL);
    g_return_if_fail(header->data_struct != NULL);
    g_return_if_fail(header->data_struct->zero_lags != NULL);
    g_return_if_fail(
        !header->data_struct->zero_lags->axes[COBDF_BASELINE_AXIS]);

    set_bincomp_size(header, header->data_struct->zero_lags, 1,
                     0 /* cross-pol products aren't present */);
}

void
sdmhdr_set_cross_data_size(sdmhdr_main_hdr_t *header)
{
    g_return_if_fail(header != NULL);
    g_return_if_fail(header->data_struct != NULL);
    g_return_if_fail(header->data_struct->cross_data != NULL);
    g_return_if_fail(
        !header->data_struct->cross_data->axes[COBDF_ANTENNA_AXIS]);

    set_bincomp_size(header, header->data_struct->cross_data, 2, 2);
}

void
sdmhdr_set_auto_data_size(sdmhdr_main_hdr_t *header)
{
    g_return_if_fail(header != NULL);
    g_return_if_fail(header->data_struct != NULL);
    g_return_if_fail(header->data_struct->auto_data != NULL);
    g_return_if_fail(
        !header->data_struct->auto_data->axes[COBDF_BASELINE_AXIS]);

    set_bincomp_size(header, header->data_struct->auto_data, 1, 2);
}

/**
 * sdmhdr_set_subset_hrefs:
 * @main: main header
 * @subset: subset header
 *
 * Initialize the "href" fields in @subset, that have a corresponding
 * non-NULL #sdmhdr_bincomp_t element in @main's 'data_struct' field,
 * with the standard names. Fields already initialized are left as-is.
 * The hrefs are created using the 'project_path' field of @subset as
 * prefix.
 */
void
sdmhdr_set_subset_hrefs(const sdmhdr_main_hdr_t *main,
                        sdmhdr_subset_hdr_t *subset)
{
    g_return_if_fail(main != NULL);
    g_return_if_fail(main->data_struct != NULL);
    g_return_if_fail(subset != NULL);
    g_return_if_fail(subset->project_path != NULL);

#define INIT_HREF(field, name) do { \
    if (main->data_struct->field != NULL \
        && subset->field ## _href == NULL) { \
        subset->field ## _href = \
            g_build_path(G_DIR_SEPARATOR_S, subset->project_path, \
                         name ".bin", NULL); \
    } \
} while (0)

    INIT_HREF(flags, "flags");
    INIT_HREF(actual_times, "actualTimes");
    INIT_HREF(actual_durations, "actualDurations");
    INIT_HREF(weights, "weights");
    INIT_HREF(zero_lags, "zeroLags");
    INIT_HREF(cross_data, "crossData");
    INIT_HREF(auto_data, "autoData");
}

/**
 * start_main_hdr_read:
 * @reader: xmlTextReaderPtr instance for sdmDataHeader element
 *
 * Start document parsing, looking for the 'sdmDataHeader' element.
 *
 * @returns: sdmhdr_main_hdr_t instance, which should be freed using
 *           sdmhdr_main_hdr_free()
 */
static sdmhdr_main_hdr_t *
start_main_hdr_read(xmlTextReaderPtr reader)
{
    sdmhdr_main_hdr_t *result = g_new0(sdmhdr_main_hdr_t, 1);
    gint rc = xmlTextReaderRead(reader);
    if (rc == 1) {
        switch (xmlTextReaderNodeType(reader)) {
          case XML_READER_TYPE_ELEMENT: {
            const xmlChar *name = xmlTextReaderConstName(reader);
            const xmlChar *nsuri = xmlTextReaderConstNamespaceUri(reader);
            if (name != NULL
                && xmlStrEqual(BAD_CAST "sdmDataHeader", name)
                && nsuri != NULL
                && xmlStrEqual(BAD_CAST bdf_uri, nsuri)) {
                attribute_t attributes[] = {
                    {.name = "schemaVersion", .nsuri = xvers_uri,
                     .parser = parse_uint,
                     .val = &(result->schema_version)},
                    {.name = "revision", .nsuri = xvers_uri,
                     .parser = parse_string, .val = &(result->revision)},
                    {.name = "mainHeaderId", .nsuri = NULL,
                     .parser = parse_string,
                     .val = &(result->main_header_id)},
                    {.name = "byteOrder", .nsuri = NULL,
                     .parser = parse_byteOrder, .val = &(result->byte_order)},
                    {.name = "projectPath", .nsuri = NULL,
                     .parser = parse_string,
                     .val = &(result->project_path)},
                    {.name = NULL}
                };
                element_t children[] = {
                    {.name = "startTime", .nsuri = bdf_uri,
                     .parser = parse_uint64_content,
                     .user_data = &(result->time)},
                    {.name = "dataOID", .nsuri = bdf_uri,
                     .parser = parse_dataOID},
                    {.name = "dimensionality", .nsuri = bdf_uri,
                     .parser = parse_dimensionality},
                    {.name = "execBlock", .nsuri = bdf_uri,
                     .parser = parse_execBlock},
                    {.name = "numAntenna", .nsuri = bdf_uri,
                     .parser = parse_numAntenna},
                    {.name = "correlationMode", .nsuri = bdf_uri,
                     .parser = parse_correlationMode},
                    {.name = "spectralResolution", .nsuri = bdf_uri,
                     .parser = parse_spectralResolution},
                    {.name = "dataStruct", .nsuri = bdf_uri,
                     .parser = parse_main_dataStruct},
                    {.name = NULL}
                };
                rc = parse_element(reader, result, children, attributes, NULL);
            }
            break;
          }
          default:
            g_debug("Failed to find 'sdmDataHeader' element");
            break;
        }
    }
    if (rc != 1) {
        sdmhdr_main_hdr_free(result);
        result = NULL;
    }

    return result;
}

static sdmhdr_subset_hdr_t *
start_subset_hdr_read(xmlTextReaderPtr reader)
{
    sdmhdr_subset_hdr_t *result = g_new0(sdmhdr_subset_hdr_t, 1);
    gint rc = xmlTextReaderRead(reader);
    if (rc == 1) {
        switch (xmlTextReaderNodeType(reader)) {
          case XML_READER_TYPE_ELEMENT: {
            const xmlChar *name = xmlTextReaderConstName(reader);
            const xmlChar *nsuri = xmlTextReaderConstNamespaceUri(reader);
            if (name != NULL
                && xmlStrEqual(BAD_CAST "sdmDataSubsetHeader", name)
                && nsuri != NULL
                && xmlStrEqual(BAD_CAST bdf_uri, nsuri)) {
                attribute_t attributes[] = {
                    {.name = "type", .nsuri = xsi_uri,
                     .parser = parse_string, .val = &(result->type)},
                    {.name = "projectPath", .nsuri = NULL,
                     .parser = parse_string, .val = &(result->project_path)},
                    {.name = NULL}
                };
                element_t children[] = {
                    {.name = "schedulePeriodTime", .nsuri = bdf_uri,
                     .parser = parse_schedulePeriodTime},
                    {.name = "dataStruct", .nsuri = bdf_uri,
                     .parser = parse_subset_dataStruct},
                    {.name = "flags", .nsuri = bdf_uri,
                     .parser = parse_subset_bincomp,
                     .user_data = &(result->flags_href)},
                    {.name = "actualTimes", .nsuri = bdf_uri,
                     .parser = parse_subset_bincomp,
                     .user_data = &(result->actual_times_href)},
                    {.name = "actualDurations", .nsuri = bdf_uri,
                     .parser = parse_subset_bincomp,
                     .user_data = &(result->actual_durations_href)},
                    {.name = "weights", .nsuri = bdf_uri,
                     .parser = parse_subset_bincomp,
                     .user_data = &(result->weights_href)},
                    {.name = "zeroLags", .nsuri = bdf_uri,
                     .parser = parse_subset_bincomp,
                     .user_data = &(result->zero_lags_href)},
                    {.name = "autoData", .nsuri = bdf_uri,
                     .parser = parse_subset_bincomp,
                     .user_data = &(result->auto_data_href)},
                    {.name = "crossData", .nsuri = bdf_uri,
                     .parser = parse_subset_crossData},
                    {.name = NULL}
                };
                rc = parse_element(reader, result, children, attributes, NULL);
            }
            break;
          }
          default:
            g_debug("Failed to find 'sdmDataHeader' element");
            break;
        }
    }
    if (rc != 1) {
        sdmhdr_subset_hdr_free(result);
        result = NULL;
    }

    return result;
}

static gint
start_main_hdr_write(xmlTextWriterPtr writer, const sdmhdr_main_hdr_t *header)
{
    gint result = -1;
    if (xmlTextWriterStartElementNS(writer, NULL, BAD_CAST "sdmDataHeader",
                                    BAD_CAST bdf_uri) >= 0
        && xmlTextWriterWriteAttributeNS(writer, BAD_CAST "xmlns",
                                         BAD_CAST xsi_prefix, NULL,
                                         BAD_CAST xsi_uri) >= 0
        && xmlTextWriterWriteAttributeNS(writer, BAD_CAST "xmlns",
                                         BAD_CAST xlink_prefix, NULL,
                                         BAD_CAST xlink_uri) >= 0
        && xmlTextWriterWriteAttributeNS(writer, BAD_CAST "xmlns",
                                         BAD_CAST xvers_prefix, NULL,
                                         BAD_CAST xvers_uri) >= 0
        && xmlTextWriterWriteFormatAttributeNS(
            writer, BAD_CAST xsi_prefix, BAD_CAST "schemaLocation", NULL,
            "http://TBD/XSDM/sdmbin %u/sdmDataObject.xsd",
            header->schema_version) >= 0
        && xmlTextWriterWriteFormatAttributeNS(
            writer, BAD_CAST xvers_prefix, BAD_CAST "schemaVersion", NULL,
            "%u", header->schema_version) >= 0
        && xmlTextWriterWriteAttributeNS(writer, BAD_CAST xvers_prefix,
                                         BAD_CAST "revision", NULL,
                                         BAD_CAST header->revision) >= 0
        && xmlTextWriterWriteAttribute(writer, BAD_CAST "mainHeaderId",
                                       BAD_CAST header->main_header_id) >= 0
        && xmlTextWriterWriteAttribute(
            writer, BAD_CAST "byteOrder", BAD_CAST g_enum_get_value(
                g_type_class_peek_static(COBDF_BYTE_ORDER_ENUM_TYPE),
                header->byte_order)->value_nick) >= 0
        && xmlTextWriterWriteAttribute(writer, BAD_CAST "projectPath",
                                       BAD_CAST header->project_path) >= 0
        && write_startTime(writer, header) >= 0
        && write_dataOID(writer, header) >= 0
        && write_dimensionality(writer, header) >= 0
        && write_execBlock(writer, header) >= 0
        && write_numAntenna(writer, header) >= 0
        && write_correlationMode(writer, header) >= 0
        && write_spectralResolution(writer, header) >= 0
        && write_main_dataStruct(writer, header) >= 0) {
        result = 0;
    }
    return result;
}

static gint
start_subset_hdr_write(xmlTextWriterPtr writer,
                       const sdmhdr_subset_hdr_t *header)
{
    gint result = -1;
    const gchar *bintype = NULL;
    if (header->cross_data_href != NULL) {
        if (header->auto_data_href != NULL) {
            bintype = "BinaryCrossAndAutoDataFullResolution";
        }
        else {
            bintype = "BinaryCrossDataFullResolution";
        }
    }
    else if (header->auto_data_href != NULL) {
        bintype = "BinaryAutoDataFullResolution";
    }

    if (xmlTextWriterStartElementNS(writer, NULL,
                                    BAD_CAST "sdmDataSubsetHeader",
                                    BAD_CAST bdf_uri) >= 0
        && xmlTextWriterWriteAttributeNS(writer, BAD_CAST "xmlns",
                                         BAD_CAST xsi_prefix, NULL,
                                         BAD_CAST xsi_uri) >= 0
        && xmlTextWriterWriteAttributeNS(writer, BAD_CAST "xmlns",
                                         BAD_CAST xlink_prefix, NULL,
                                         BAD_CAST xlink_uri) >= 0
        && (bintype == NULL
            || xmlTextWriterWriteAttributeNS(
                writer, BAD_CAST xsi_prefix, BAD_CAST "type", NULL,
                BAD_CAST bintype) >= 0)
        && xmlTextWriterWriteAttribute(writer, BAD_CAST "projectPath",
                                       BAD_CAST header->project_path) >= 0
        && write_schedulePeriodTime(writer, header) >= 0
        && write_subset_dataStruct(writer, header) >= 0
        && write_subset_bincomp(writer, "flags", header->flags_href) >= 0
        && write_subset_bincomp(writer, "actualTimes",
                                header->actual_times_href) >= 0
        && write_subset_bincomp(writer, "actualDurations",
                                header->actual_durations_href) >= 0
        && write_subset_bincomp(writer, "weights", header->weights_href) >= 0
        && (header->cross_data_href == NULL
            || write_subset_crossData(writer, header) >= 0)
        && write_subset_bincomp(writer, "autoData",
                                header->auto_data_href) >= 0
        && write_subset_bincomp(writer, "zeroLags",
                                header->zero_lags_href) >= 0) {
        result = 0;
    }
    return result;
}

static gint
parse_element(xmlTextReaderPtr reader, gpointer header,
              const element_t children[], const attribute_t attributes[],
              gchar **text)
{
    gint result = -1;
    const element_t *scan_children(const xmlChar *name, const xmlChar *ns) {
        const element_t *ch = children;
        element_parser_fn_t parser = NULL;
        if (children != NULL) {
            while (parser == NULL && ch->name != NULL) {
                if (xmlStrEqual(BAD_CAST ch->name, name)
                    && xmlStrEqual(BAD_CAST ch->nsuri, ns)) {
                    parser = ch->parser;
                }
                else {
                    ++ch;
                }
            }
        }
        return ((parser == NULL) ? NULL : ch);
    }
    const xmlChar *element_name = xmlTextReaderConstName(reader);
    const xmlChar *element_nsuri = xmlTextReaderConstNamespaceUri(reader);
    g_assert(element_nsuri != NULL);
    gint rc = parse_attributes(reader, attributes);
    if (rc == 0) {
        result = 1;
        if (!xmlTextReaderIsEmptyElement(reader)) {
            gint rc = xmlTextReaderRead(reader);
            while (result == 1 && rc == 1) {
                const xmlChar *name = xmlTextReaderConstName(reader);
                const xmlChar *nsuri = xmlTextReaderConstNamespaceUri(reader);
                switch (xmlTextReaderNodeType(reader)) {
                  case XML_READER_TYPE_ELEMENT: {
                    g_assert(nsuri != NULL);
                    const element_t *ch = scan_children(name, nsuri);
                    if (ch != NULL) {
                        rc = ch->parser(reader, header, ch->user_data);
                    }
                    else {
                        result = -1;
                    }
                    break;
                  }
                  case XML_READER_TYPE_END_ELEMENT: {
                    g_assert(nsuri != NULL);
                    if (xmlStrEqual(element_name, name)
                        && nsuri != NULL
                        && xmlStrEqual(nsuri, BAD_CAST element_nsuri)) {
                        rc = 0;
                    }
                    else {
                        result = -1;
                    }
                    break;
                  }
                  case XML_READER_TYPE_TEXT:
                    if (text != NULL) {
                        xmlChar *val = xmlTextReaderValue(reader);
                        *text = g_strdup((gchar *)val);
                        xmlFree(val);
                    }
                    else {
                        result = -1;
                    }
                    break;
                  default:
                    break;
                }
                if (result != -1 && rc == 1) {
                    rc = xmlTextReaderRead(reader);
                    if (rc == 0) {
                        result = 0;
                    }
                }
            }
        }
    }
    return result;
}

static gint
parse_attributes(xmlTextReaderPtr reader, const attribute_t attributes[])
{
    gint result = 0;
    if (attributes != NULL) {
        const attribute_t *attr = attributes;
        while (result == 0 && attr->name != NULL) {
            xmlChar *val_str = xmlTextReaderGetAttributeNs(
                reader, BAD_CAST attr->name, BAD_CAST attr->nsuri);
            if (val_str != NULL) {
                result = attr->parser((gchar *)val_str, attr->val);
                xmlFree(val_str);
            }
            ++attr;
        }
    }
    return result;
}

static gint
parse_dataOID(xmlTextReaderPtr reader, gpointer header, gpointer user_data)
{
    sdmhdr_main_hdr_t *hdr = header;
    hdr->data_oid = g_new0(sdmhdr_dataoid_t, 1);
    attribute_t attributes[] = {
        {.nsuri = xlink_uri, .name = "href",
         .parser = parse_string, .val = &(hdr->data_oid->href)},
        {.nsuri = xlink_uri, .name = "title",
         .parser = parse_string, .val = &(hdr->data_oid->title)},
        {.nsuri = xlink_uri, .name = "type",
         .parser = parse_string, .val = &(hdr->data_oid->type)},
        {.name = NULL}
    };
    return parse_element(reader, hdr, NULL, attributes, NULL);
}

static gint
parse_dimensionality(xmlTextReaderPtr reader, gpointer header,
                     gpointer user_data)
{
    sdmhdr_main_hdr_t *hdr = header;
    GArray *dim = g_array_new(FALSE, FALSE, sizeof(CobdfAxisName));
    attribute_t attributes[] = {
        {.nsuri = NULL, .name = "axes", .parser = parse_axes,
         .val = dim},
        {.name = NULL}
    };
    gchar *len_str = NULL;
    gint result = parse_element(reader, hdr, NULL, attributes, &len_str);
    CobdfAxisName invalid_axis = COBDF_INVALID_AXIS;
    g_array_append_val(dim, invalid_axis);
    if (result == 1 && len_str != NULL) {
        errno = 0;
        gulong len = strtoul(len_str, NULL, 10);
        if ((errno == 0) && (len <= G_MAXUINT)
            && ((guint)len + 1 == dim->len)) {
            hdr->dimensionality = (CobdfAxisName *)g_array_free(dim, FALSE);
            dim = NULL;
        }
        else {
            result = -1;
        }
    }
    else {
        result = -1;
    }
    if (len_str != NULL) {
        g_free(len_str);
    }
    if (dim != NULL) {
        g_array_free(dim, TRUE);
        hdr->dimensionality = NULL;
    }
    return result;
}

static gint
parse_execBlock(xmlTextReaderPtr reader, gpointer header, gpointer user_data)
{
    sdmhdr_main_hdr_t *hdr = header;
    hdr->exec_block = g_new0(sdmhdr_execblock_t, 1);
    attribute_t attributes[] = {
        {.nsuri = xlink_uri, .name = "href", .parser = parse_string,
         .val = &(hdr->exec_block->href)},
        {.nsuri = xlink_uri, .name = "type", .parser = parse_string,
         .val = &(hdr->exec_block->type)},
        {.name = NULL}
    };
    return parse_element(reader, hdr, NULL, attributes, NULL);
}

static gint
parse_numAntenna(xmlTextReaderPtr reader, gpointer header, gpointer user_data)
{
    sdmhdr_main_hdr_t *hdr = header;
    gchar *str = NULL;
    gint result = parse_element(reader, header, NULL, NULL, &str);
    if (result == 1 && str != NULL) {
        errno = 0;
        gulong val = strtoul(str, NULL, 10);
        if (errno == 0 && val <= G_MAXUINT) {
            hdr->num_antennas = (guint)val;
        }
        else {
            result = -1;
        }
    }
    else {
        result = -1;
    }
    if (str != NULL) {
        g_free(str);
    }
    return result;
}

static gint
parse_correlationMode(xmlTextReaderPtr reader, gpointer header,
                      gpointer user_data)
{
    sdmhdr_main_hdr_t *hdr = header;
    gchar *str = NULL;
    gint result = parse_element(reader, header, NULL, NULL, &str);
    if (result == 1 && str != NULL) {
        GEnumValue *val = g_enum_get_value_by_nick(
            g_type_class_peek_static(COBDF_CORRELATION_MODE_ENUM_TYPE),
            g_strstrip(str));
        if (val != NULL) {
            hdr->mode = val->value;
        }
        else {
            result = -1;
        }
    }
    else {
        result = -1;
    }
    if (str != NULL) {
        g_free(str);
    }
    return result;
}

static gint
parse_spectralResolution(xmlTextReaderPtr reader, gpointer header,
                         gpointer user_data)
{
    sdmhdr_main_hdr_t *hdr = header;
    gchar *str = NULL;
    gint result = parse_element(reader, header, NULL, NULL, &str);
    if (result == 1 && str != NULL) {
        GEnumValue *val = g_enum_get_value_by_nick(
            g_type_class_peek_static(COBDF_SPECTRAL_RESOLUTION_TYPE_ENUM_TYPE),
            g_strstrip(str));
        if (val != NULL) {
            hdr->resolution = val->value;
        }
        else {
            // TODO: this may not be the correct default case, but the
            // spectralResolution element is destined to be removed
            // very soon, so I'll ignore the problem
            hdr->resolution = COBDF_RESOLUTION_UNKNOWN;
        }
    }
    else {
        result = -1;
    }
    if (str != NULL) {
        g_free(str);
    }
    return result;
}

static gint
parse_main_dataStruct(xmlTextReaderPtr reader, gpointer header,
                      gpointer user_data)
{
    gint result = -1;
    sdmhdr_main_hdr_t *hdr = header;
    if (hdr->data_struct == NULL) {
        sdmhdr_data_struct_t *ds = g_new0(sdmhdr_data_struct_t, 1);
        hdr->data_struct = ds;
        GPtrArray *bb = g_ptr_array_new();
        attribute_t attributes[] = {
            {.name = "type", .nsuri = xsi_uri, .parser = parse_string,
             .val = &(ds->type)},
            {.name = "apc", .nsuri = NULL, .parser = parse_apc,
             .val = ds->apc},
            {.name = NULL}
        };
        element_t children[] = {
            {.name = "baseband", .nsuri = bdf_uri, .parser = parse_baseband,
             .user_data = bb},
            {.name = "flags", .nsuri = bdf_uri, .parser = parse_main_bincomp,
             .user_data = &(ds->flags)},
            {.name = "actualTimes", .nsuri = bdf_uri,
             .parser = parse_main_bincomp, .user_data = &(ds->actual_times)},
            {.name = "actualDurations", .nsuri = bdf_uri,
             .parser = parse_main_bincomp,
             .user_data = &(ds->actual_durations)},
            {.name = "zeroLags", .nsuri = bdf_uri,
             .parser = parse_main_bincomp, .user_data = &(ds->zero_lags)},
            {.name = "autoData", .nsuri = bdf_uri,
             .parser = parse_main_bincomp, .user_data = &(ds->auto_data)},
            {.name = "crossData", .nsuri = bdf_uri,
             .parser = parse_main_bincomp, .user_data = &(ds->cross_data)},
            {.name = "weights", .nsuri = bdf_uri, .parser = parse_weights},
            {.name = NULL}
        };
        result = parse_element(reader, header, children, attributes, NULL);
        if (result == 1) {
            g_ptr_array_add(bb, NULL);
            ds->basebands = (sdmhdr_bb_t **)g_ptr_array_free(bb, FALSE);
        }
        else {
            g_ptr_array_free(bb, TRUE);
        }
    }
    return result;
}

static gint
parse_apc(gchar *str, gpointer val)
{
    gint result = -1;
    sdmhdr_data_struct_t ds G_GNUC_UNUSED;
    CobdfAtmPhaseCorrection *apc = val;

    gchar **apcstrs = g_strsplit(g_strstrip(str), " ", 0);
    guint num_apc = g_strv_length(apcstrs);
    if (num_apc <= G_N_ELEMENTS(ds.apc)) {
        result = 0;
        guint i;
        for (i = 0; result == 0 && i < num_apc; ++i) {
            GEnumValue *enumval = g_enum_get_value_by_nick(
                g_type_class_peek_static(COBDF_ATM_PHASE_CORRECTION_ENUM_TYPE),
                apcstrs[i]);
            if (enumval != NULL) {
                apc[i] = enumval->value;
            }
            else {
                result = -1;
            }
        }
        if (result == 0) {
            while (i < G_N_ELEMENTS(ds.apc)) {
                apc[i++] = COBDF_AP_INVALID;
            }
        }
    }
    g_strfreev(apcstrs);

    return result;
}

static gint
parse_main_bincomp(xmlTextReaderPtr reader, gpointer header,
                   gpointer user_data)
{
    sdmhdr_bincomp_t **tbl = user_data;
    GArray *axes = g_array_new(FALSE, FALSE, sizeof(CobdfAxisName));
    gsize size;
    attribute_t attributes[] = {
        {.name = "size", .nsuri = NULL, .parser = parse_size,
         .val = &size},
        {.name = "axes", .nsuri = NULL, .parser = parse_axes,
         .val = axes},
        {.name = NULL}
    };
    gint result = parse_element(reader, header, NULL, attributes, NULL);
    if (result == 1) {
        *tbl = g_new0(sdmhdr_bincomp_t, 1);
        (*tbl)->size = size;
        guint i;
        for (i = 0; i < axes->len; ++i) {
            (*tbl)->axes[g_array_index(axes, CobdfAxisName, i)] = TRUE;
        }
    }
    g_array_free(axes, TRUE);
    return result;
}

static gint
parse_axes(gchar *str, gpointer val)
{
    gint result = 0;
    GArray *array = val;
    gchar **axes = g_strsplit(g_strstrip(str), " ", 0);
    gchar **axis = axes;
    GEnumClass *axis_names =
        g_type_class_peek_static(COBDF_AXIS_NAME_ENUM_TYPE);
    while (result == 0 && *axis != NULL) {
        GEnumValue *axis_name = g_enum_get_value_by_nick(axis_names, *axis);
        if (axis_name != NULL) {
            CobdfAxisName axis_name_val = axis_name->value;
            if ((array->len == 0)
                || (g_array_index(array, CobdfAxisName, array->len - 1)
                    > axis_name_val)) {
                g_array_append_val(array, axis_name->value);
            }
            else {
                result = -1;
            }
        }
        else {
            result = -1;
        }
        ++axis;
    }
    g_strfreev(axes);
    return result;
}

static gint
parse_size(gchar *str, gpointer val)
{
    gint result;
    gsize *size = val;
    guint64 n = g_ascii_strtoull(str, NULL, 10);
    if (n <= G_MAXSIZE) {
        *size = (gsize)n;
        result = 0;
    }
    else {
        result = -1;
    }
    return result;
}

static gint
parse_weights(xmlTextReaderPtr reader, gpointer header, gpointer user_data)
{
    sdmhdr_main_hdr_t *hdr = header;
    g_return_val_if_fail(hdr->data_struct != NULL, -1);

    GArray *axes = g_array_new(FALSE, FALSE, sizeof(CobdfAxisName));
    gsize size;
    attribute_t attributes[] = {
        {.name = "size", .nsuri = NULL, .parser = parse_size,
         .val = &size},
        {.name = "axes", .nsuri = NULL, .parser = parse_axes,
         .val = axes},
        {.name = NULL}
    };
    gchar *weights_str = NULL;
    gint result =
        parse_element(reader, header, NULL, attributes, &weights_str);
    if (result == 1 && weights_str != NULL) {
        g_strstrip(weights_str);
        gchar **weights = g_strsplit_set(weights_str, " \f\n\r\t\v", 0);
        sdmhdr_weights_t *w =
            g_malloc0(SIZEOF_SDMHDR_WEIGHTS(g_strv_length(weights)));
        gchar **weight = weights;
        guint i = 0;
        while (result == 1 && *weight != NULL) {
            gdouble wgt = g_ascii_strtod(*weight, NULL);
            if (errno == 0 && wgt <= G_MAXFLOAT) {
                w->weights[i++] = (gfloat)wgt;
                ++weight;
            }
            else {
                result = -1;
            }
        }
        w->num_weights = i;
        if (result == 1) {
            w->common.size = size;
            for (i = 0; i < axes->len; ++i) {
                w->common.axes[g_array_index(axes, CobdfAxisName, i)] = TRUE;
            }
            hdr->data_struct->weights = w;
        }
        else {
            g_free(w);
        }
        g_strfreev(weights);
    }
    g_array_free(axes, TRUE);
    if (weights_str != NULL) {
        g_free(weights_str);
    }
    return result;
}

static gint
parse_uint(gchar *str, gpointer val)
{
    guint result;
    guint *uint = val;
    errno = 0;
    guint64 n = g_ascii_strtoull(str, NULL, 10);
    if (errno == 0 && n <= G_MAXUINT) {
        *uint = (guint)n;
        result = 0;
    }
    else {
        result = -1;
    }
    return result;
}

static gint
parse_string(gchar *str, gpointer val)
{
    gchar **cpy = val;
    *cpy = g_strdup(str);
    return 0;
}

static gint
parse_baseband(xmlTextReaderPtr reader, gpointer header, gpointer user_data)
{
/*     gchar *name = NULL; */
/*     attribute_t attributes[] = { */
/*         {.name = "name", .nsuri = NULL, .parser = parse_string, */
/*          .val = &name}, */
/*         {.name = NULL} */
/*     }; */
    GArray *sws = g_array_new(FALSE, FALSE, sizeof(sdmhdr_sw_t));
    element_t children[] = {
        {.name = "spectralWindow", .nsuri = bdf_uri,
         .parser = parse_spectralWindow, .user_data = sws},
        {.name = NULL}
    };
    gint result = parse_element(reader, header, children, NULL, NULL);
    if (result == 1) {
        sdmhdr_bb_t *bb = g_malloc(SIZEOF_SDMHDR_BB(sws->len));
        bb->num_sw = sws->len;
        memcpy(bb->sw, sws->data, sws->len * sizeof(sdmhdr_sw_t));
        GPtrArray *basebands = user_data;
        g_ptr_array_add(basebands, bb);
    }
    g_array_free(sws, TRUE);
    return result;
}

static gint
parse_spectralWindow(xmlTextReaderPtr reader, gpointer header,
                     gpointer user_data)
{
    sdmhdr_sw_t sw = {
        .id = NULL,
        .sd_pol_products = SDMHDR_APP_NONE,
        .cross_pol_products = SDMHDR_XPP_NONE,
        .num_bin = 0,
        .num_spectral_point = 0
    };
    // intentionally, I'm not reading the 'scaleFactor', and
    // 'sideband' attributes
    attribute_t attributes[] = {
        {.name = "id", .nsuri = NULL, .parser = parse_string,
         .val = &(sw.id)},
        {.name = "sdPolProducts", .nsuri = NULL,
         .parser = parse_sdPolProducts, .val = &(sw.sd_pol_products)},
        {.name = "crossPolProducts", .nsuri = NULL,
         .parser = parse_crossPolProducts, .val = &(sw.cross_pol_products)},
        {.name = "numBin", .nsuri = NULL, .parser = parse_uint,
         .val = &(sw.num_bin)},
        {.name = "numSpectralPoint", .nsuri = NULL, .parser = parse_uint,
         .val = &(sw.num_spectral_point)},
        {.name = NULL}
    };
    gint result = parse_element(reader, header, NULL, attributes, NULL);
    if (result == 1 && sw.sd_pol_products != SDMHDR_APP_INVALID &&
        sw.cross_pol_products != SDMHDR_XPP_INVALID) {
        GArray *sws = user_data;
        g_array_append_val(sws, sw);
    }
    else {
        result = -1;
    }
    return result;
}

/**
 * normalize_list_whitespace:
 * @str: string of space-separated tokens
 *
 * Strip leading and trailing space characters from @str, and reduce
 * each run of multiple space characters between tokens to a single
 * space character.
 */
static void
normalize_list_whitespace(gchar *str)
{
    guint len = strlen(str) + 1;
    gchar *c = str;
    while (*c == ' ') {
        ++c;
        --len;
    }
    if (c != str) {
        memmove(str, c, len);
    }
    while (*c != '\0') {
        if (*c++ == ' ') {
            gchar *d = c;
            while (*d == ' ') {
                ++d;
                --len;
            }
            if (d != c) {
                memmove(c, d, len);
            }
        }
        --len;
    }
    if (*(c - 1) == ' ') {
        *(c - 1) = '\0';
    }
}

static gint
parse_sdPolProducts(gchar *str, gpointer val)
{
    gint result = 0;
    sdmhdr_app_t *app = val;
    normalize_list_whitespace(str);
    if (strcmp(str, "RR") == 0) {
        *app = SDMHDR_APP_RR;
    }
    else if (strcmp(str, "LL") == 0) {
        *app = SDMHDR_APP_LL;
    }
    else if (strcmp(str, "RR LL") == 0) {
        *app = SDMHDR_APP_RR_LL;
    }
    else if (strcmp(str, "RR RL LL") == 0) {
        *app = SDMHDR_APP_RR_RL_LL;
    }
    else {
        *app = SDMHDR_APP_INVALID;
        result = -1;
    }
    return result;
}

static gint
parse_crossPolProducts(gchar *str, gpointer val)
{
    gint result = 0;
    sdmhdr_xpp_t *xpp = val;
    normalize_list_whitespace(str);
    if (strcmp(str, "RR") == 0) {
        *xpp = SDMHDR_XPP_RR;
    }
    else if (strcmp(str, "LL") == 0) {
        *xpp = SDMHDR_XPP_LL;
    }
    else if (strcmp(str, "RR LL") == 0) {
        *xpp = SDMHDR_XPP_RR_LL;
    }
    else if (strcmp(str, "RR RL LR LL") == 0) {
        *xpp = SDMHDR_XPP_RR_RL_LR_LL;
    }
    else {
        *xpp = SDMHDR_XPP_INVALID;
        result = -1;
    }
    return result;
}

static gint
parse_byteOrder(gchar *str, gpointer val)
{
    gint result;
    CobdfByteOrder *byteorder = val;
    GEnumValue *enumval = g_enum_get_value_by_nick(
        g_type_class_peek_static(COBDF_BYTE_ORDER_ENUM_TYPE),
        g_strstrip(str));
    if (enumval != NULL) {
        *byteorder = enumval->value;
        result = 0;
    }
    else {
        result = -1;
    }
    return result;
}

static gint
parse_uint64_content(xmlTextReaderPtr reader, gpointer header,
                     gpointer user_data)
{
    guint64 *uint64 = user_data;
    gchar *str = NULL;
    gint result = parse_element(reader, header, NULL, NULL, &str);
    if (result == 1 && str != NULL) {
        errno = 0;
        *uint64 = strtoull(str, NULL, 10);
        if (errno != 0) {
            result = -1;
        }
    }
    else {
        result = -1;
    }
    if (str != NULL) {
        g_free(str);
    }
    return result;
}

static gint
parse_schedulePeriodTime(xmlTextReaderPtr reader, gpointer header,
                         gpointer user_data)
{
    sdmhdr_subset_hdr_t *hdr = header;
    element_t children[] = {
        {.name = "time", .nsuri = bdf_uri, .parser = parse_uint64_content,
         .user_data = &(hdr->time)},
        {.name = "interval", .nsuri = bdf_uri, .parser = parse_uint64_content,
         .user_data = &(hdr->interval)},
        {.name = NULL}
    };
    return parse_element(reader, header, children, NULL, NULL);
}

static gint
parse_subset_dataStruct(xmlTextReaderPtr reader, gpointer header,
                        gpointer user_data)
{
    sdmhdr_subset_hdr_t *hdr = header;
    attribute_t attributes[] = {
        {.name = "ref", .nsuri = NULL, .parser = parse_string,
         .val = &(hdr->data_struct_ref)},
        {.name = NULL}
    };
    return parse_element(reader, header, NULL, attributes, NULL);
}

static gint
parse_subset_bincomp(xmlTextReaderPtr reader, gpointer header,
                     gpointer user_data)
{
    attribute_t attributes[] = {
        {.name = "href", .nsuri = xlink_uri, .parser = parse_string,
         .val = user_data},
        {.name = NULL}
    };
    return parse_element(reader, header, NULL, attributes, NULL);
}

static gint
parse_subset_crossData(xmlTextReaderPtr reader, gpointer header,
                       gpointer user_data)
{
    sdmhdr_subset_hdr_t *hdr = header;
    attribute_t attributes[] = {
        {.name = "href", .nsuri = xlink_uri, .parser = parse_string,
         .val = &(hdr->cross_data_href)},
        {.name = "type", .nsuri = NULL, .parser = parse_primitive_data_type,
         .val = &(hdr->cross_data_type)},
        {.name = NULL}
    };
    return parse_element(reader, header, NULL, attributes, NULL);
}

static gint
parse_primitive_data_type(gchar *str, gpointer val)
{
    gint result;
    CobdfPrimitiveDataType *type = val;
    GEnumValue *enumval = g_enum_get_value_by_nick(
        g_type_class_peek_static(COBDF_PRIMITIVE_DATA_TYPE_ENUM_TYPE),
        g_strstrip(str));
    if (enumval != NULL) {
        *type = enumval->value;
        result = 0;
    }
    else {
        result = -1;
    }
    return result;
}

static gint
write_startTime(xmlTextWriterPtr writer, const sdmhdr_main_hdr_t *header)
{
    return xmlTextWriterWriteFormatElement(
        writer, BAD_CAST "startTime", "%"G_GUINT64_FORMAT, header->time);
}

static gint
write_dataOID(xmlTextWriterPtr writer, const sdmhdr_main_hdr_t *header)
{
    gint result = 0;
    if (header->data_oid != NULL) {
        sdmhdr_dataoid_t *oid = header->data_oid;
        if (xmlTextWriterStartElement(writer, BAD_CAST "dataOID") < 0
            || xmlTextWriterWriteAttributeNS(writer, BAD_CAST xlink_prefix,
                                             BAD_CAST "type", NULL,
                                             BAD_CAST oid->type) < 0
            || xmlTextWriterWriteAttributeNS(writer, BAD_CAST xlink_prefix,
                                             BAD_CAST "href", NULL,
                                             BAD_CAST oid->href) < 0
            || xmlTextWriterWriteAttributeNS(writer, BAD_CAST xlink_prefix,
                                             BAD_CAST "title", NULL,
                                             BAD_CAST oid->title) < 0
            || xmlTextWriterEndElement(writer) < 0) {
            result = -1;
        }
    }
    return result;
}

static gint
write_dimensionality(xmlTextWriterPtr writer, const sdmhdr_main_hdr_t *header)
{
    gint result = 0;
    if (header->dimensionality != NULL) {
        GPtrArray *dims = g_ptr_array_new();
        CobdfAxisName *axis = header->dimensionality;
        while (*axis != COBDF_INVALID_AXIS) {
            g_ptr_array_add(
                dims,
                (gchar *)g_enum_get_value(
                    g_type_class_peek_static(COBDF_AXIS_NAME_ENUM_TYPE),
                    *axis)->value_nick);
            ++axis;
        }
        g_ptr_array_add(dims, NULL);
        gchar *axes = g_strjoinv(" ", (gchar **)dims->pdata);
        if (xmlTextWriterStartElement(writer, BAD_CAST "dimensionality") < 0
            || xmlTextWriterWriteAttribute(writer, BAD_CAST "axes",
                                           BAD_CAST axes) < 0
            || xmlTextWriterWriteFormatString(writer, "%u", dims->len - 1) < 0
            || xmlTextWriterEndElement(writer) < 0) {
            result = -1;
        }
        g_free(axes);
        g_ptr_array_free(dims, TRUE);
    }
    return result;
}

static gint
write_execBlock(xmlTextWriterPtr writer, const sdmhdr_main_hdr_t *header)
{
    gint result = 0;
    if (header->exec_block != NULL) {
        sdmhdr_execblock_t *eblk = header->exec_block;
        if (xmlTextWriterStartElement(writer, BAD_CAST "execBlock") < 0
            || xmlTextWriterWriteAttributeNS(writer, BAD_CAST xlink_prefix,
                                             BAD_CAST "href", NULL,
                                             BAD_CAST eblk->href) < 0
            || xmlTextWriterWriteAttributeNS(writer, BAD_CAST xlink_prefix,
                                             BAD_CAST "type", NULL,
                                             BAD_CAST eblk->type) < 0
            || xmlTextWriterEndElement(writer) < 0) {
            result = -1;
        }
    }
    return result;
}

static gint
write_numAntenna(xmlTextWriterPtr writer, const sdmhdr_main_hdr_t *header)
{
    return xmlTextWriterWriteFormatElement(writer, BAD_CAST "numAntenna",
                                           "%u", header->num_antennas);
}

static gint
write_correlationMode(xmlTextWriterPtr writer, const sdmhdr_main_hdr_t *header)
{
    return xmlTextWriterWriteElement(
        writer, BAD_CAST "correlationMode",
        BAD_CAST g_enum_get_value(
            g_type_class_peek_static(COBDF_CORRELATION_MODE_ENUM_TYPE),
            header->mode)->value_nick);
}

static gint
write_spectralResolution(xmlTextWriterPtr writer,
                         const sdmhdr_main_hdr_t *header)
{
    return xmlTextWriterWriteElement(
        writer, BAD_CAST "spectralResolution",
        BAD_CAST g_enum_get_value(
            g_type_class_peek_static(COBDF_SPECTRAL_RESOLUTION_TYPE_ENUM_TYPE),
            header->resolution)->value_nick);
}

static gint
write_main_dataStruct(xmlTextWriterPtr writer, const sdmhdr_main_hdr_t *header)
{
    gint result = 0;
    if (header->data_struct != NULL) {
        sdmhdr_data_struct_t *ds = header->data_struct;
        GPtrArray *apcs = g_ptr_array_new();
        guint i;
        for (i = 0;
             ds->apc[i] != COBDF_AP_INVALID && i < G_N_ELEMENTS(ds->apc);
             ++i) {
            const gchar *str = g_enum_get_value(
                g_type_class_peek_static(COBDF_ATM_PHASE_CORRECTION_ENUM_TYPE),
                ds->apc[i])->value_nick;
            g_ptr_array_add(apcs, (gchar *)str);
        }
        g_ptr_array_add(apcs, NULL);
        gchar *apcval = g_strjoinv(" ", (gchar **)apcs->pdata);
        g_ptr_array_free(apcs, TRUE);
        if (xmlTextWriterStartElement(writer, BAD_CAST "dataStruct") >= 0
            && xmlTextWriterWriteAttributeNS(writer, BAD_CAST xsi_prefix,
                                             BAD_CAST "type", NULL,
                                             BAD_CAST ds->type) >= 0
            && xmlTextWriterWriteAttribute(
                writer, BAD_CAST "apc", BAD_CAST apcval) >= 0) {
            sdmhdr_bb_t **bb = ds->basebands;
#ifdef USE_ALMA_5_1_0_5_XML_WORKAROUNDS
            guint bb_idx = 1;
#endif
            while (result >= 0 && *bb != NULL) {
#ifndef USE_ALMA_5_1_0_5_XML_WORKAROUNDS
                result = write_baseband(writer, *bb);
#else
                result = write_baseband(writer, *bb, bb_idx++);
#endif
                ++bb;
            }
            if (result >= 0) {
                if (write_main_bincomp(writer, "flags", ds->flags, TRUE) < 0
                    || write_main_bincomp(writer, "actualTimes",
                                           ds->actual_times, TRUE) < 0
                    || write_main_bincomp(writer, "actualDurations",
                                           ds->actual_durations, TRUE) < 0
                    || write_weights(writer, ds->weights) < 0
                    || write_main_bincomp(writer, "crossData",
                                           ds->cross_data, TRUE) < 0
                    || write_main_bincomp(writer, "autoData",
                                           ds->auto_data, TRUE) < 0
                    || write_main_bincomp(writer, "zeroLags",
                                           ds->zero_lags, TRUE) < 0
                    || xmlTextWriterEndElement(writer) < 0) {
                    result = -1;
                }
            }
        }
        else {
            result = -1;
        }
        g_free(apcval);
    }
    return result;
}

static gint
write_main_bincomp(xmlTextWriterPtr writer, const gchar *name,
                   const sdmhdr_bincomp_t *bincomp, gboolean end_element)
{
    gint result = 0;
    if (bincomp != NULL) {
        if (xmlTextWriterStartElement(writer, BAD_CAST name) >= 0
            && xmlTextWriterWriteFormatAttribute(writer, BAD_CAST "size",
                                                 "%"G_GSIZE_FORMAT,
                                                 bincomp->size) >= 0) {
            GPtrArray *axes = g_ptr_array_new();
            gint i;
            for (i = COBDF_NUM_AXES - 1; i >= 0; --i) {
                if (bincomp->axes[i]) {
                    g_ptr_array_add(
                        axes,
                        (gchar *)g_enum_get_value(
                            g_type_class_peek_static(
                                COBDF_AXIS_NAME_ENUM_TYPE), i)->value_nick);
                }
            }
            g_ptr_array_add(axes, NULL);
            gchar *list = g_strjoinv(" ", (gchar **)axes->pdata);
            if (xmlTextWriterWriteAttribute(writer, BAD_CAST "axes",
                                            BAD_CAST list) >= 0) {
                if (end_element && xmlTextWriterEndElement(writer) < 0) {
                    result = -1;
                }
            }
            else {
                result = -1;
            }
            g_free(list);
            g_ptr_array_free(axes, TRUE);
        }
        else {
            result = -1;
        }
    }
    return result;
}

static gint
write_weights(xmlTextWriterPtr writer, const sdmhdr_weights_t *weights)
{
    gint result = 0;
    if (weights != NULL) {
        if (write_main_bincomp(writer, "weights", &(weights->common),
                               FALSE) >= 0) {
            if (weights->num_weights > 0) {
                GString *str = g_string_new("");
                g_string_printf(str, "%f", weights->weights[0]);
                guint i;
                for (i = 1; i < weights->num_weights; ++i) {
                    g_string_append_printf(str, " %f", weights->weights[i]);
                }
                if (xmlTextWriterWriteString(writer, BAD_CAST str->str) < 0
                    || xmlTextWriterEndElement(writer) < 0) {
                    result = -1;
                }
                g_string_free(str, TRUE);
            }
        }
        else {
            result = -1;
        }
    }
    return result;
}

static gint
write_baseband(xmlTextWriterPtr writer, const sdmhdr_bb_t *baseband
#ifdef USE_ALMA_5_1_0_5_XML_WORKAROUNDS
               , guint idx
#endif
    )
{
    gint result = 0;
    if (baseband != NULL) {
        if (xmlTextWriterStartElement(writer, BAD_CAST "baseband") >= 0
#ifdef USE_ALMA_5_1_0_5_XML_WORKAROUNDS
            && xmlTextWriterWriteFormatAttribute(writer, BAD_CAST "name",
                                                 "BB_%u", idx) >= 0
#endif
            ) {
            guint i;
            for (i = 0; result >= 0 && i < baseband->num_sw; ++i) {
                const sdmhdr_sw_t *sw = &(baseband->sw[i]);
                result = write_spectralWindow(writer, sw);
            }
            if (result >= 0) {
                result = xmlTextWriterEndElement(writer);
            }
        }
        else {
            result = -1;
        }
    }
    return result;
}

static gint
write_spectralWindow(xmlTextWriterPtr writer, const sdmhdr_sw_t *sw)
{
    static const gchar *app_vals[SDMHDR_APP_NUM_VALS] =
        {[SDMHDR_APP_RR] = "RR",
         [SDMHDR_APP_LL] = "LL",
         [SDMHDR_APP_RR_LL] = "RR LL",
         [SDMHDR_APP_RR_RL_LL] = "RR RL LL"
        };
    static const gchar *xpp_vals[SDMHDR_XPP_NUM_VALS] =
        {[SDMHDR_XPP_RR] = "RR",
         [SDMHDR_XPP_LL] = "LL",
         [SDMHDR_XPP_RR_LL] = "RR LL",
         [SDMHDR_XPP_RR_RL_LR_LL] = "RR RL LR LL"
        };
    gint result = 0;
    if (sw != NULL) {
        if (xmlTextWriterStartElement(writer, BAD_CAST "spectralWindow")
            >= 0) {
            if (sw->id != NULL) {
                result = xmlTextWriterWriteAttribute(writer, BAD_CAST "id",
                                                     BAD_CAST sw->id);
            }
            if (result >= 0) {
                if (sw->sd_pol_products == SDMHDR_APP_INVALID
                    || (sw->sd_pol_products != SDMHDR_APP_NONE
                        && xmlTextWriterWriteAttribute(
                            writer, BAD_CAST "sdPolProducts",
                            BAD_CAST app_vals[sw->sd_pol_products]) < 0)
                    || sw->cross_pol_products == SDMHDR_XPP_INVALID
                    || (sw->cross_pol_products != SDMHDR_XPP_NONE
                        && xmlTextWriterWriteAttribute(
                            writer, BAD_CAST "crossPolProducts",
                            BAD_CAST xpp_vals[sw->cross_pol_products]) < 0)
                    || xmlTextWriterWriteFormatAttribute(
                        writer, BAD_CAST "numSpectralPoint",
                        "%u", sw->num_spectral_point) < 0
                    || xmlTextWriterWriteFormatAttribute(
                        writer, BAD_CAST "numBin", "%u", sw->num_bin) < 0
#ifdef USE_BDF_V1_0_SCHEMA_WORKAROUNDS
                    || xmlTextWriterWriteFormatAttribute(
                        writer, BAD_CAST "scaleFactor", "%f", 1.0f) < 0
#endif
                    || xmlTextWriterWriteAttribute(
                        writer, BAD_CAST "sideband",
                        BAD_CAST g_enum_get_value(
                            g_type_class_peek_static(
                                COBDF_NET_SIDEBAND_ENUM_TYPE),
                            COBDF_NOSB)->value_nick) < 0
                    || xmlTextWriterEndElement(writer) < 0) {
                    result = -1;
                }
            }
        }
        else {
            result = -1;
        }
    }
    return result;
}

static gint
write_schedulePeriodTime(xmlTextWriterPtr writer,
                         const sdmhdr_subset_hdr_t *header)
{
    gint result = -1;
    if (xmlTextWriterStartElement(writer, BAD_CAST "schedulePeriodTime") >= 0
        && write_uint64_element(writer, "time", header->time) >= 0
        && write_uint64_element(writer, "interval", header->interval) >= 0
        && xmlTextWriterEndElement(writer) >= 0) {
        result = 0;
    }
    return result;
}

static gint
write_uint64_element(xmlTextWriterPtr writer, const gchar *element,
                     guint64 value)
{
    return xmlTextWriterWriteFormatElement(writer, BAD_CAST element,
                                           "%"G_GUINT64_FORMAT, value);
}

static gint
write_subset_dataStruct(xmlTextWriterPtr writer,
                        const sdmhdr_subset_hdr_t *header)
{
    gint result = -1;
    if (xmlTextWriterStartElement(writer, BAD_CAST "dataStruct") >= 0
        && xmlTextWriterWriteAttribute(writer, BAD_CAST "ref",
                                       BAD_CAST header->data_struct_ref) >= 0
        && xmlTextWriterEndElement(writer) >= 0) {
        result = 0;
    }
    return result;
}

static gint
write_subset_bincomp(xmlTextWriterPtr writer, const gchar *element,
                     const gchar *href)
{
    gint result = -1;
    if (href != NULL) {
        if (xmlTextWriterStartElement(writer, BAD_CAST element) >= 0
            && xmlTextWriterWriteAttributeNS(writer, BAD_CAST xlink_prefix,
                                             BAD_CAST "href", NULL,
                                             BAD_CAST href) >= 0
            && xmlTextWriterEndElement(writer) >= 0) {
            result = 0;
        }
    }
    else {
        result = 0;
    }

    return result;
}

static gint
write_subset_crossData(xmlTextWriterPtr writer,
                       const sdmhdr_subset_hdr_t *header)
{
    gint result = -1;
    if (xmlTextWriterStartElement(writer, BAD_CAST "crossData") >= 0
        && xmlTextWriterWriteAttributeNS(writer, BAD_CAST xlink_prefix,
                                         BAD_CAST "href", NULL,
                                         BAD_CAST header->cross_data_href) >= 0
        && xmlTextWriterWriteAttribute(
            writer, BAD_CAST "type",
            BAD_CAST g_enum_get_value(
                g_type_class_peek_static(COBDF_PRIMITIVE_DATA_TYPE_ENUM_TYPE),
                header->cross_data_type)->value_nick) >= 0
        && xmlTextWriterEndElement(writer) >= 0) {
        result = 0;
    }
    return result;
}

static void
set_bincomp_size(const sdmhdr_main_hdr_t *header, sdmhdr_bincomp_t *bincomp,
                 guint par_mult, guint cross_mult)
{
    bincomp->size = num_elements(header, bincomp->axes, par_mult, cross_mult);
}

static gsize
num_elements(const sdmhdr_main_hdr_t *header, const gboolean *axes,
             guint parallel_pp_mult, guint cross_pp_mult)
{
    g_return_val_if_fail(header != NULL, 0);
    g_return_val_if_fail(header->data_struct != NULL, 0);
    g_return_val_if_fail(header->data_struct->basebands != NULL, 0);

    gsize result = 0;

#define AXIS_LEN(a, tlen, flen) (axes[a] ? (tlen) : (flen))

    guint num_antennas = AXIS_LEN(COBDF_ANTENNA_AXIS, header->num_antennas, 0);
    guint num_baselines =
        AXIS_LEN(COBDF_BASELINE_AXIS,
                 (header->num_antennas * (header->num_antennas - 1)) / 2, 0);
    // To handle the unlikely case that neither the BAL nor ANT axis
    // exists, we arbitrarily do the calculation in a baseline index
    // context by setting 'num_baselines' to one.
    if (!axes[COBDF_BASELINE_AXIS] && !axes[COBDF_ANTENNA_AXIS]) {
        num_baselines = 1;
    }

    // loop through basebands and spectral windows to add up number of
    // table elements, taking into account omitted table axes
    g_return_val_if_fail(num_basebands(header->data_struct->basebands) > 0, 0);
    sdmhdr_bb_t **real_basebands = header->data_struct->basebands;
    sdmhdr_bb_t *max_bb[1] = {NULL};
    sdmhdr_bb_t **basebands;
    guint bb_max;
    if (axes[COBDF_BASEBAND_AXIS]) {
        bb_max = num_basebands(real_basebands);
        basebands = real_basebands;
    }
    else {
        bb_max = 1;
        basebands = max_bb;
        max_bb[0] = maximal_baseband((const sdmhdr_bb_t **)real_basebands,
                                     parallel_pp_mult, cross_pp_mult);
    }
    guint bb_idx;
    for (bb_idx = 0; bb_idx < bb_max; ++bb_idx) {
        const sdmhdr_bb_t *real_bb = basebands[bb_idx];
        g_return_val_if_fail(real_bb->num_sw > 0, 0);
        sdmhdr_bb_t *max_bb = NULL; // this hides the outer def of max_bb
        const sdmhdr_bb_t *bb;
        guint sw_max;
        if (axes[COBDF_SPECTRAL_WINDOW_AXIS]) {
            sw_max = real_bb->num_sw;
            bb = real_bb;
        }
        else {
            sw_max = 1;
            max_bb = g_malloc0(SIZEOF_SDMHDR_BB(1));
            max_bb->sw[0].cross_pol_products = SDMHDR_XPP_INVALID;
            max_bb->sw[0].sd_pol_products = SDMHDR_APP_INVALID;
            maximal_spectral_window(real_bb, &(max_bb->sw[0]),
                                    parallel_pp_mult, cross_pp_mult);
            bb = max_bb;
        }
        guint sw_idx;
        for (sw_idx = 0; sw_idx < sw_max; ++sw_idx) {
            const sdmhdr_sw_t *sw = &(bb->sw[sw_idx]);
            g_return_val_if_fail(sw->num_bin > 0, 0);
            g_return_val_if_fail(sw->num_spectral_point > 0, 0);
            gsize sw_elements = AXIS_LEN(COBDF_BIN_AXIS, sw->num_bin, 1)
                * AXIS_LEN(COBDF_SPECTRAL_POINT_AXIS, sw->num_spectral_point,
                           1);
            gsize per_antenna = sw_elements;
            gsize per_baseline = sw_elements;
            if (axes[COBDF_POLARIZATION_AXIS]) {
                per_antenna *= num_auto_primitives(
                    sw->sd_pol_products, parallel_pp_mult, cross_pp_mult);
                per_baseline *= num_cross_primitives(
                    sw->cross_pol_products, parallel_pp_mult, cross_pp_mult);
            }
            else {
                per_antenna *= MAX(parallel_pp_mult, cross_pp_mult);
                per_baseline *= MAX(parallel_pp_mult, cross_pp_mult);
            }
            result +=
                per_antenna * num_antennas + per_baseline * num_baselines;
        }
        if (max_bb != NULL) {
            g_free(max_bb);
        }
    }
    if (max_bb[0] != NULL) {
        g_free(max_bb[0]);
    }
    guint apc_mult;
    if (axes[COBDF_ATM_PHASE_CORRECTION_AXIS]) {
        apc_mult = 0;
        while (apc_mult < G_N_ELEMENTS(header->data_struct->apc)
               && header->data_struct->apc[apc_mult] != COBDF_AP_INVALID) {
            ++apc_mult;
        }
    }
    else {
        apc_mult = 1;
    }
    return result * apc_mult;
#undef AXIS_LEN
}

static sdmhdr_bb_t *
maximal_baseband(const sdmhdr_bb_t **basebands, guint parallel_mult,
                 guint cross_mult)
{
    g_return_val_if_fail(basebands != NULL, NULL);

    guint max_num_sw = 0;
    const sdmhdr_bb_t **bbs = basebands;
    while (*bbs != NULL) {
        if ((*bbs)->num_sw > max_num_sw) {
            max_num_sw = (*bbs)->num_sw;
        }
        ++bbs;
    }
    sdmhdr_bb_t *result = g_malloc0(SIZEOF_SDMHDR_BB(max_num_sw));
    result->num_sw = max_num_sw;
    sdmhdr_sw_t *max_sw = result->sw;
    guint i;
    for (i = 0; i < max_num_sw; ++i) {
        max_sw[i].cross_pol_products = SDMHDR_XPP_INVALID;
        max_sw[i].sd_pol_products = SDMHDR_APP_INVALID;
    }
    bbs = basebands;
    while (*bbs != NULL) {
        for (i = 0; i < (*bbs)->num_sw; ++i) {
            const sdmhdr_sw_t *sw = &((*bbs)->sw[i]);
            if (sw->num_bin > max_sw[i].num_bin) {
                max_sw[i].num_bin = sw->num_bin;
            }
            if (sw->num_spectral_point > max_sw[i].num_spectral_point) {
                max_sw[i].num_spectral_point = sw->num_spectral_point;
            }
            if (num_cross_primitives(sw->cross_pol_products, parallel_mult,
                                     cross_mult)
                > num_cross_primitives(max_sw[i].cross_pol_products,
                                       parallel_mult, cross_mult)) {
                max_sw[i].cross_pol_products = sw->cross_pol_products;
            }
            if (num_auto_primitives(sw->sd_pol_products, parallel_mult,
                                    cross_mult)
                > num_auto_primitives(max_sw[i].sd_pol_products,
                                      parallel_mult, cross_mult)) {
                max_sw[i].sd_pol_products = sw->sd_pol_products;
            }
        }
        ++bbs;
    }
    return result;
}

static void
maximal_spectral_window(const sdmhdr_bb_t *baseband, sdmhdr_sw_t *max_sw,
                        guint parallel_mult, guint cross_mult)
{
    g_return_if_fail(baseband != NULL);
    g_return_if_fail(max_sw != NULL);

    guint i;
    for (i = 0; i < baseband->num_sw; ++i) {
        const sdmhdr_sw_t *sw = &(baseband->sw[i]);
        if (sw->num_bin > max_sw->num_bin) {
            max_sw->num_bin = sw->num_bin;
        }
        if (sw->num_spectral_point > max_sw->num_spectral_point) {
            max_sw->num_spectral_point = sw->num_spectral_point;
        }
        if (num_cross_primitives(sw->cross_pol_products, parallel_mult,
                                 cross_mult)
            > num_cross_primitives(max_sw->cross_pol_products, parallel_mult,
                                   cross_mult)) {
            max_sw->cross_pol_products = sw->cross_pol_products;
        }
        if (num_auto_primitives(sw->sd_pol_products, parallel_mult,
                                cross_mult)
            > num_auto_primitives(max_sw->sd_pol_products, parallel_mult,
                                  cross_mult)) {
            max_sw->sd_pol_products = sw->sd_pol_products;
        }
    }
}

static guint
num_cross_primitives(sdmhdr_xpp_t xpp, guint parallel_mult, guint cross_mult)
{
    guint result;

    switch (xpp) {
      case SDMHDR_XPP_RR:
      case SDMHDR_XPP_LL:
        result = parallel_mult;
        break;
      case SDMHDR_XPP_RR_LL:
        result = 2 * parallel_mult;
        break;
      case SDMHDR_XPP_RR_RL_LR_LL:
        result = 2 * (parallel_mult + cross_mult);
        break;
      case SDMHDR_XPP_INVALID:
      case SDMHDR_XPP_NONE:
        result = 0;
        break;
      default:
        g_assert_not_reached();
        result = 0;
        break;
    }
    return result;
}

static guint
num_auto_primitives(sdmhdr_app_t app, guint parallel_mult, guint cross_mult)
{
    guint result;

    switch (app) {
      case SDMHDR_APP_RR:
      case SDMHDR_APP_LL:
        result = parallel_mult;
        break;
      case SDMHDR_APP_RR_LL:
        result = 2 * parallel_mult;
        break;
      case SDMHDR_APP_RR_RL_LL:
        result = 2 * parallel_mult + cross_mult;
        break;
      case SDMHDR_APP_INVALID:
      case SDMHDR_APP_NONE:
        result = 0;
        break;
      default:
        g_assert_not_reached();
        result = 0;
        break;
    }
    return result;
}

static void *
mmap_xml_element(gint fd, off_t off, const gchar *name, gsize *len,
                 gchar **element)
{
    *len = 0;
    *element = NULL;
    glong pagesize = sysconf(_SC_PAGESIZE);
    off_t aligned_off = (off / pagesize) * pagesize;
    // We allocate twice the page size to allow the opening element to
    // span a page boundary, but this will only work if the length of
    // the element name doesn't exceed one page.
    pagesize *= 2;
    gchar *result =
        mmap(NULL, pagesize, PROT_READ | PROT_WRITE, MAP_PRIVATE, fd,
             aligned_off);
    if (result != MAP_FAILED) {
        off -= aligned_off;
        gchar *start = g_strdup_printf("<%s", name);
        gchar *end = g_strdup_printf("</%s>", name);
        *element = g_strstr_len(result + off, pagesize - off, start);
        if (*element != NULL) {
            *len = pagesize;
            ptrdiff_t start_offset = *element - result;
            guint endlen = strlen(end);
            ptrdiff_t search_offset = -endlen;
            gchar *strend =
                g_strstr_len(*element, pagesize - start_offset, end);
            while (result != NULL && strend == NULL) {
                search_offset += pagesize;
                result =
                    mremap(result, *len, *len + pagesize, MREMAP_MAYMOVE);
                if (result != MAP_FAILED) {
                    *len += pagesize;
                    *element = result + start_offset;
                    strend = g_strstr_len(result + search_offset,
                                          pagesize + endlen, end);
                }
                else {
                    result = NULL;
                    *element = NULL;
                    *len = 0;
                    strend = NULL;
                }
            }
            if (strend != NULL) {
                *(strend + strlen(end)) = '\0';
            }
        }
        else {
            munmap(result, pagesize);
            result = NULL;
        }
        g_free(end);
        g_free(start);
    }
    else {
        result = NULL;
    }
    return result;
}

static gint
positioned_fd(FILE *file)
{
    gint result = -1;
    gint fd = fileno(file);
    if (fd != -1) {
        errno = 0;
        off_t begin = ftello(file);
        if (begin != -1) {
            off_t off_rc = lseek(fd, begin, SEEK_SET);
            if (off_rc != -1) {
                result = fd;
            }
            else {
                g_critical("failed to set file descriptor position: %s",
                           g_strerror(errno));
            }
        }
        else if (errno == ESPIPE) {
            result = fd;
        }
        else {
            g_critical("Failed to get file position from stream: %s",
                       g_strerror(errno));
        }
    }
    else {
        g_critical("Failed to get file descriptor from stream: %s",
                   g_strerror(errno));
    }
    return result;
}
