#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <mdarray.h>
#include <tree.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/mman.h>
#include <unistd.h>

typedef enum {
    STORE_HEAP,
    STORE_STREAM,
    STORE_MMAP,
    STORE_NONE
} store_type_t;

static void normalize(tree_t *tree);
static GNode *scan_node(GScanner *scanner);
static gint scan_num_children(GScanner *scanner);
static GPtrArray *scan_children(GScanner *scanner);
static glong scan_multiplier(GScanner *scanner);
static gpointer copy_mdarray(gconstpointer src, gpointer data);
static void assign_memory(tree_t *tree, gpointer buffer, off_t offset);
static void unassign_memory(tree_t *tree);
static void assign_stream(tree_t *tree, off_t offset);
static branch_t *tree_branch_with_offset(
    const tree_t *tree, const guint *coordinates, off_t offset);
static gpointer tree_memory(const tree_t *tree);
static FILE *tree_stream(const tree_t *tree);
static store_type_t tree_branch_store_type(const branch_t *branch);
static store_type_t tree_store_type(const tree_t *tree);

#define NODE_ARRAY(node) (G_NODE_IS_ROOT(node) ? \
    ((root_data_t *)((node)->data))->mdarray : (node)->data)

/* tree_t *tree_new(const gchar *spec)
 *
 * Create a new tree with the given structure. Note that the tree does
 * not have any storage allocated (or assigned) by this function; for
 * that, use 'tree_alloc', 'tree_mmap_pars' or 'tree_stream_pars'
 * after calling this function. Trees are constrained to have a
 * constant height.
 *
 * The tree structure specification string is optional. Passing a NULL
 * pointer will create only the root node of the tree. A user may add
 * to the tree using the glib 'g_node' functions. The syntax of the
 * string is as follows.
 *
 * node = "{" node-contents "}" ;
 * node-contents = [ leaves | branches ] ;
 * leaves = integer ;
 * branches = integer "," child-list ;
 * child-list = "[" rep-node { "," rep-node } "]" ;
 * rep-node = [ integer ] node ;
 * integer = pos-digit { digit } ;
 * pos-digit = "1" | "2" | "3" | "4" | "5" | "6" | "7" | "8" | "9" ;
 * digit = "0" | pos-digit ;
 *
 * Some notes:
 * - The 'rep-node' indicates a sequence of identical nodes.
 * - 'branches' indicates, first, the number of child nodes, and,
 *   second, a list of nodes. The list is traversed cyclically to
 *   generate the child nodes ('rep-node' expansion occurs before list
 *   traversal).
 * - 'leaves' is equivalent to a node with "empty" child nodes (that
 *   is, the children are leaf nodes), e.g., {N} <=> {N, [{}]}
 * - An empty ('empty') node is a leaf node. It is simpler to
 *   indicate leaf nodes through the alternate 'leaves'
 *   construction.
 * - Whitespace characters are ignored.
 *
 * Example:
 * {3, [{4, [{2, [{2}]}, {1, [{3}]}, 2{2, [{2}]}]}, 2{2, [{3,
 * [{1}, {2}, {3}]}, {2, [{4}, {2}]}]}]}
 * specifies the following tree
 *
               +--------------------------+-----------------------+
               |                          |                       |
      +-----+--+--+-------+         +-----+-----+           +-----+-----+
      |     |     |       |         |           |           |           |
    +-+     +     +-+     +-+   +---+---+       +---+   +---+---+       +---+
    | |     |     | |     | |   |   |   |       |   |   |   |   |       |   |
  +-+ +-+ +-+-+ +-+ +-+ +-+ +-+ + +-+ +-+-+ +-+-+-+ +-+ + +-+ +-+-+ +-+-+-+ +-+
  | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | |
  @ @ @ @ @ @ @ @ @ @ @ @ @ @ @ @ @ @ @ @ @ @ @ @ @ @ @ @ @ @ @ @ @ @ @ @ @ @ @
 *
 *
 * Arguments:
 *   spec - tree structure specification string, may be NULL
 *
 * Return value:
 *   pointer to the newly allocated tree
 */
tree_t *
tree_new(const gchar *spec, gsize element_size)
{
    g_return_val_if_fail(element_size > 0, NULL);

    tree_t *result;
    if (spec != NULL) {
        GScanner *scanner = g_scanner_new(NULL);
        scanner->input_name = "tree_spec";
        scanner->config->scan_identifier = FALSE;
        scanner->config->scan_float = FALSE;
        g_scanner_input_text(scanner, spec, strlen(spec));
        result = scan_node(scanner);
        gboolean set_size(GNode *node, gpointer unused G_GNUC_UNUSED) {
            mdarray_t *mdarray = node->data;
            mdarray->element_size = element_size;
            return FALSE;
        }
        g_node_traverse(result, G_IN_ORDER, G_TRAVERSE_LEAVES, -1,
                        set_size, NULL);
        root_data_t *rdata = g_new0(root_data_t, 1);
        rdata->mdarray = result->data;
        rdata->element_size = element_size;
        rdata->mmap_fd = -1;
        rdata->stream = NULL;
        result->data = rdata;
        g_scanner_destroy(scanner);
    }
    else {
        // FIXME: doesn't work (the root node data won't be correct)
        result = g_new0(tree_t, 1);
        g_error("unimplemented");
    }
    return result;
}

/* gint tree_height(tree_t *tree)
 *
 * The height of the tree.
 *
 * Arguments:
 *   tree - pointer to tree
 *
 * Return values:
 *   -1, if tree height is not equal on all branches
 *   tree height, otherwise
 */
gint
tree_height(const tree_t *tree)
{
    gint result = 0;
    if (tree != NULL) {
        gboolean is_variable = FALSE;
        gboolean check_depth(GNode *node, gpointer unused G_GNUC_UNUSED) {
            guint depth = g_node_depth(node) - 1;
            mdarray_t *mdarray = NODE_ARRAY(node);
            if (node->data != NULL) {
                depth += mdarray->dimensions->len - 1;
            }
            if (result != 0) {
                if (depth != result) {
                    is_variable = TRUE;
                }
            }
            else {
                result = depth;
            }
            return is_variable;
        }
        g_node_traverse((tree_t *)tree, G_IN_ORDER, G_TRAVERSE_LEAVES, -1,
                        check_depth, NULL);
        if (is_variable) {
            result = -1;
        }
    }
    return result;
}

/* guint64 tree_count(tree_t *tree)
 *
 * Number of leaf nodes in tree.
 */
guint64
tree_count(const tree_t *tree)
{
    guint64 result = 0;
    if (tree != NULL) {
        gboolean count(GNode *node, gpointer unused G_GNUC_UNUSED) {
            mdarray_t *mdarray = NODE_ARRAY(node);
            if (mdarray != NULL) {
                result += mdarray_count(mdarray);
            }
            else {
                ++result;
            }
            return FALSE;
        }
        g_node_traverse((tree_t *)tree, G_IN_ORDER, G_TRAVERSE_LEAVES, -1,
                        count, NULL);
    }
    return result;
}

/* gint tree_alloc(tree_t *tree)
 *
 * Allocate memory on heap for a tree. Should only be applied to the
 * root node of a tree.
 *
 * *Return value*
 *   0, on success
 *   -1, otherwise
 */
gint
tree_alloc(tree_t *tree)
{
    g_return_val_if_fail(tree != NULL, -1);
    g_return_val_if_fail(G_NODE_IS_ROOT(tree), -1);
    g_return_val_if_fail(tree_height(tree) > 0, -1);

    gint result = -1;
    guint64 size = tree_size(tree);
    if (size <= G_MAXULONG) {
        gpointer buffer = g_try_malloc((gulong)size);
        if (buffer != NULL) {
            normalize(tree);
            assign_memory(tree, buffer, 0);
            result = 0;
        }
        else {
            g_warning("Failed to allocate %"G_GUINT64_FORMAT
                      " bytes of memory for tree: %s",
                      size, strerror(errno));
        }
    }
    else {
        g_message("Unable to allocate %"G_GUINT64_FORMAT" bytes of memory",
                  size);
    }
    return result;
}

/* void tree_mmap_pars(tree_t *tree, int fd, int prot, off_t offset)
 *
 * Assign parameters to tree for memory mapping file contents as tree
 * element storage.
 */
void
tree_mmap_pars(tree_t *tree, int fd, int prot, off_t offset)
{
    g_return_if_fail(tree != NULL);
    g_return_if_fail(G_NODE_IS_ROOT(tree));
    g_return_if_fail(tree_height(tree) > 0);
    g_return_if_fail(fd >= 0);

    root_data_t *rdata = tree->data;
    rdata->mmap_fd = fd;
    rdata->mmap_prot = prot;
    rdata->data_offset = offset;
    normalize(tree);
}

/* void tree_stream_pars(tree_t *tree, FILE *stream, off_t offset)
 *
 * Assign parameters to tree for using a stream as tree element
 * storage.
 */
void
tree_stream_pars(tree_t *tree, FILE *stream, off_t offset)
{
    g_return_if_fail(tree != NULL);
    g_return_if_fail(G_NODE_IS_ROOT(tree));
    g_return_if_fail(tree_height(tree) > 0);
    g_return_if_fail(stream != NULL);

    root_data_t *rdata = tree->data;
    rdata->stream = stream;
    rdata->data_offset = offset;
    normalize(tree);
    assign_stream(tree, offset);
}


/* gpointer tree_free(tree_t *tree, gboolean free_buffer)
 *
 * Free resources used by tree. This function should only be applied
 * to the root node of a tree. Will free memory allocated by
 * 'tree_alloc'.
 *
 * Arguments:
 *   tree - pointer to tree
 */
void
tree_free(tree_t *tree)
{
    g_return_if_fail(tree != NULL);
    g_return_if_fail(G_NODE_IS_ROOT(tree));

    root_data_t *root_data = tree->data;
    if (root_data->mmap_fd == -1 && tree_stream(tree) == NULL) {
        gpointer origin = tree_memory(tree);
        if (origin != NULL) {
            g_free(origin);
        }
    }
    gboolean free_array(GNode *node, gpointer unused G_GNUC_UNUSED) {
        mdarray_t *mdarray = NODE_ARRAY(node);
        if (mdarray != NULL) {
            mdarray_free(mdarray);
        }
        return FALSE;
    }
    g_node_traverse(tree, G_IN_ORDER, G_TRAVERSE_LEAVES, -1, free_array, NULL);
    g_node_destroy(tree);
    g_free(root_data);
}

/* branch_t *tree_branch(tree_t *tree, guint *fixed_coordinates)
 *
 * Select a branch of a tree.
 *
 * Arguments:
 *   tree - pointer to tree
 *   fixed_coordinates - index of desired branch (COORD_VECTOR_END-terminated)
 *
 * Return values:
 *   NULL, iff branch index is out of bounds;
 *   branch pointer, otherwise
 */
branch_t *
tree_branch(const tree_t *tree, const guint *coordinates)
{
    root_data_t *rdata = g_node_get_root((tree_t *)tree)->data;
    return tree_branch_with_offset(tree, coordinates, rdata->data_offset);
}

/**
 * tree_branch_root:
 * @tree: tree
 *
 * Create a 'branch_t' at the root of a tree.
 *
 * @returns: branch at root of tree
 */
branch_t *
tree_branch_root(const tree_t *tree)
{
    g_return_val_if_fail(tree != NULL, NULL);

    guint origin[] = {COORD_VECTOR_END};
    return tree_branch(tree, origin);
}

/* guint64 tree_branch_count(const branch_t *branch)
 *
 * Number of leaf nodes in tree branch.
 */
guint64
tree_branch_count(const branch_t *branch)
{
    guint64 result = 0;
    switch (branch->type) {
      case TREE:
        result = tree_count(branch->tree);
        break;
      case ARRAY:
        result = mdarray_count(branch->array);
        break;
      default:
        g_assert_not_reached();
        break;
    }
    return result;
}

/**
 * tree_branch_height:
 * @branch: tree branch
 *
 * @return: The height of the tree branch; or -1, if the tree branch
 * does not have a constant height.
 */
gint
tree_branch_height(const branch_t *branch)
{
    gint result = 0;
    if (branch != NULL) {
        switch (branch->type) {
          case TREE:
            result = tree_height(branch->tree);
            break;
          case ARRAY:
            result = branch->array->dimensions->len - 1;
            break;
          default:
            g_assert_not_reached();
            break;
        }
    }
    return result;
}

/**
 * tree_branch_clear:
 * @branch: tree branch
 *
 * Clear the memory used by the elements of a tree branch.
 */
void
tree_branch_clear(branch_t *branch)
{
    g_return_if_fail(branch != NULL);

    if (tree_branch_memory(branch) != NULL) {
        memset(tree_branch_memory(branch) + branch->data_offset, 0,
               tree_branch_size(branch));
    }
    else {
        FILE *stream = tree_branch_stream(branch);
        if (stream != NULL) {
            flockfile(stream);
            if (fseeko(stream, branch->data_offset, SEEK_SET)
                == 0) {
                glong pagesize = sysconf(_SC_PAGESIZE);
                gsize num_pages = tree_branch_size(branch) / pagesize;
                gsize partial_page = tree_branch_size(branch) % pagesize;
                gchar page[pagesize];
                memset(page, 0, pagesize);
                gint rc = 1;
                while (rc == 1 && num_pages-- > 0) {
                    rc = fwrite_unlocked(page, pagesize, 1, stream);
                }
                if (rc == 1 && partial_page > 0) {
                    rc = fwrite_unlocked(page, partial_page, 1, stream);
                }
                if (rc != 1) {
                    g_warning("Failed to clear all tree branch elements: %s",
                              g_strerror(errno));
                }
            }
            else {
                g_warning("Failed to fseeko tree branch file position to "
                          "origin: %s", g_strerror(errno));
            }
            funlockfile(stream);
        }
    }
}

/**
 * tree_branch_element_size:
 * @branch: tree branch
 *
 * @returns: size of data elements in tree branch
 */
gsize
tree_branch_element_size(const branch_t *branch)
{
    gsize result;
    switch (branch->type) {
      case TREE:
        result = ((root_data_t *)(g_node_get_root(branch->tree)->data))
            ->element_size;
        break;
      case ARRAY:
        result = branch->array->element_size;
        break;
      default:
        g_assert_not_reached();
        result = 0;
        break;
    }
    return result;
}

/**
 * tree_branch_foreach:
 * @branch: tree branch
 * @fn: function to apply to each tree branch element
 * @user_data: pointer to arbitrary data to be passed to @fn
 *
 * Loop over the elements stored in a tree branch (in order), and
 * apply a function to each element. The loop can be halted at any
 * point by returning #TRUE from @fn.
 */
void
tree_branch_foreach(branch_t *branch, tree_fn_t fn, gpointer user_data)
{
    guint height = tree_branch_height(branch);
    guint coords[height + 1];
    memset(coords, 0, sizeof(guint) * height);
    coords[height] = COORD_VECTOR_END;
    gsize element_size = tree_branch_element_size(branch);

    FILE *stream = tree_branch_stream(branch);
    gpointer memory_value = tree_branch_memory(branch);

    gboolean memory_fn(guint *c, gpointer udata) {
        gboolean result = fn(c, memory_value, udata);
        memory_value += element_size;
        return result;
    }
    gboolean stream_fn(guint *c, gpointer udata) {
        gchar value[element_size];
        gboolean result = FALSE;
        if (fread_unlocked(value, element_size, 1, stream) == 1) {
            result = fn(c, value, udata);
            if (fseeko(stream, -element_size, SEEK_CUR) == 0) {
                if (fwrite_unlocked(value, element_size, 1, stream) != 1) {
                    result = FALSE;
                    g_warning("Failed to fwrite mdarray element: %s",
                              g_strerror(errno));
                }
            }
            else {
                result = FALSE;
                g_warning("Failed to fseeko to mdarray element: %s",
                          g_strerror(errno));
            }
        }
        else {
            g_warning("Failed to fread mdarray element: %s",
                      g_strerror(errno));
        }
        return result;
    }

    gboolean (*eltfn)(guint *, gpointer) = NULL;
    gint prev_depth = -10;
    gint initial_depth = -1;
    gboolean tfn(GNode *node, gpointer udata) {
        gboolean result = FALSE;
        gint depth = g_node_depth(node) - initial_depth - 1;
        guint d1 = depth + 1;
        if (depth <= prev_depth) {
            coords[depth]++;
            memset(coords + d1, 0, sizeof(guint) * (height - d1));
        }
        if (G_UNLIKELY(d1 == height)) {
            result = eltfn(coords, udata);
        }
        else if (G_NODE_IS_LEAF(node)) {
            mdarray_t *mdarray = NODE_ARRAY(node);
            guint tidx[height + 1];
            memcpy(tidx, coords, height * sizeof(guint));
            tidx[height] = COORD_VECTOR_END;
            gboolean mfn(guint *idx, gpointer value, gpointer u2data) {
                gboolean mresult;
                guint i = d1;
                guint *tix = &(tidx[i]);
                guint *cix = &(coords[i]);
                while (i++ < height) {
                    *(tix++) = *(cix++) + *(idx++);
                }
                mresult = fn(tidx, value, u2data);
                memory_value = value + element_size;
                return mresult;
            }
            mdarray_foreach(mdarray, mfn, udata);
            memcpy(coords + d1, tidx + d1, (height - d1) * sizeof(guint));
            depth = height - 1;
        }
        prev_depth = depth;
        return result;
    }

    switch (branch->type) {
      case TREE: {
        if (stream == NULL) {
            if (memory_value != NULL) {
                memory_value += branch->data_offset;
                eltfn = memory_fn;
            }
        }
        else {
            flockfile(stream);
            if (fseeko(stream, branch->data_offset, SEEK_SET) == 0) {
                eltfn = stream_fn;
            }
            else {
                g_warning("Failed to fseeko to beginning of mdarray "
                          "array: %s", g_strerror(errno));
            }
        }
        if (eltfn != NULL) {
            initial_depth = g_node_depth(branch->tree);
            g_node_traverse(branch->tree, G_PRE_ORDER, G_TRAVERSE_ALL, -1,
                            tfn, user_data);
        }
        if (stream != NULL) {
            fflush_unlocked(stream);
            funlockfile(stream);
        }
        break;
      }
      case ARRAY:
        mdarray_foreach(branch->array, fn, user_data);
        break;
      default:
        g_assert_not_reached();
        break;
    }
}

/* branch_t *tree_branch_branch(branch_t *branch, guint *fixed_coordinates)
 *
 * Select a branch of a branch.
 *
 * Arguments:
 *   branch - pointer to branch
 *   fixed_coordinates - index of desired branch (COORD_VECTOR_END-terminated)
 *
 * Return values:
 *   NULL, iff branch index is out of bounds;
 *   branch pointer, otherwise
 */
branch_t *
tree_branch_branch(const branch_t *branch, const guint *coordinates)
{
    g_return_val_if_fail(branch != NULL, NULL);

    branch_t *result = NULL;
    switch (branch->type) {
      case TREE:
        result = tree_branch_with_offset(
            branch->tree, coordinates, branch->data_offset);
        break;
      case ARRAY: {
        off_t offset;
        result = g_new0(branch_t, 1);
        result->type = ARRAY;
        result->array = mdarray_reduced(branch->array, coordinates, &offset);
        result->mmap_fd = -1;
        result->data_offset = branch->data_offset + offset;
        switch (tree_branch_store_type(branch)) {
          case STORE_HEAP:
            mdarray_assign_memory(result->array, branch->array->buffer,
                                  result->data_offset);
            break;
          case STORE_MMAP: {
            long pagesize = sysconf(_SC_PAGESIZE);
            result->mmap_off = (result->data_offset / pagesize) * pagesize;
            result->mmap_data_offset = result->data_offset % pagesize;
            result->mmap_len =
                mdarray_size(result->array) + result->mmap_data_offset;
            result->mmap_fd = branch->mmap_fd;
            result->mmap_prot = branch->mmap_prot;
            result->mmap_ptr = MAP_FAILED;
            break;
          }
          case STORE_STREAM:
            mdarray_assign_stream(result->array, tree_branch_stream(branch),
                                  result->data_offset);
            break;
          default:
            break;
        }
        break;
      }
      default:
        g_assert_not_reached();
        break;
    }
    g_assert(tree_branch_store_type(branch) == tree_branch_store_type(result));
    return result;
}

/* gint tree_branch_mmap(branch_t *branch)
 *
 * Memory map a tree branch. Memory mapping parameters must have been
 * previously provided to the tree using 'tree_mmap_pars'. Note that
 * only 'branch_t's can be mmapped, 'tree_t's cannot (but this is not
 * a real restriction because you can create a 'branch_t' from the
 * root node of a 'tree_t').
 */
gint
tree_branch_mmap(branch_t *branch)
{
    g_return_val_if_fail(branch != NULL, -1);
    g_return_val_if_fail(branch->mmap_fd != -1, -1);
    g_return_val_if_fail(branch->mmap_ptr == MAP_FAILED, -1);

    gint result = -1;
    if (branch->mmap_len <= G_MAXSIZE) {
        branch->mmap_ptr = mmap(NULL, branch->mmap_len, branch->mmap_prot,
                                MAP_SHARED, branch->mmap_fd, branch->mmap_off);
        if (branch->mmap_ptr != MAP_FAILED) {
            switch (branch->type) {
              case TREE:
                assign_memory(
                    branch->tree, branch->mmap_ptr, branch->mmap_data_offset);
                break;
              case ARRAY:
                mdarray_assign_memory(
                    branch->array, branch->mmap_ptr, branch->mmap_data_offset);
                break;
              default:
                g_assert_not_reached();
                break;
            }
            result = 0;
        }
        else {
            g_warning("Failed to mmap tree branch: %s", strerror(errno));
        }
    }
    else {
        g_message("Unable to mmap tree branch of %"G_GSIZE_FORMAT
                  " bytes", branch->mmap_len);
    }
    return result;
}

/* void tree_branch_munmap(branch_t *branch)
 *
 * Unmap a tree branch from memory.
 */
void
tree_branch_munmap(branch_t *branch)
{
    g_return_if_fail(branch != NULL);
    g_return_if_fail(branch->mmap_fd != -1);
    g_return_if_fail(branch->mmap_ptr != MAP_FAILED);

    switch (branch->type){
      case TREE:
        unassign_memory(branch->tree);
        break;
      case ARRAY:
        mdarray_unassign(branch->array);
        break;
      default:
        g_assert_not_reached();
        break;
    }
    gint rc = munmap(branch->mmap_ptr, branch->mmap_len);
    if (rc == 0) {
        branch->mmap_ptr = MAP_FAILED;
    }
    else {
        g_warning("Failed to munmap tree branch memory: %s", strerror(errno));
    }
}

/* void tree_branch_free(branch_t *branch)
 *
 * Free resources used by a branch.
 */
void
tree_branch_free(branch_t *branch)
{
    g_return_if_fail(branch != NULL);

    if (branch->mmap_fd != -1 && branch->mmap_ptr != MAP_FAILED) {
        tree_branch_munmap(branch);
    }
    switch (branch->type) {
      case ARRAY:
        mdarray_free(branch->array);
        break;
      default:
        // Nothing to do for a TREE, since it just references a
        // node in the original tree.
        break;
    }
    g_free(branch);
}

gpointer
tree_branch_memory(const branch_t *branch)
{
    gpointer *result;
    switch (branch->type) {
      case TREE:
        result = tree_memory(branch->tree);
        break;
      case ARRAY:
        result = branch->array->buffer;
        break;
      default:
        g_assert_not_reached();
        result = NULL;
        break;
    }
    return result;
}

FILE *
tree_branch_stream(const branch_t *branch)
{
    FILE *result;
    switch (branch->type) {
      case TREE:
        result = tree_stream(branch->tree);
        break;
      case ARRAY:
        result = branch->array->stream;
        break;
      default:
        g_assert_not_reached();
        result = NULL;
        break;
    }
    return result;
}

off_t
tree_branch_index_offset(const branch_t *branch, const guint *coords)
{
    g_return_val_if_fail(branch != NULL, -1);
    g_return_val_if_fail(coords != NULL, -1);

    off_t result = -1;
    switch (branch->type) {
      case TREE:
        result = 0;//FIXME
        break;
      case ARRAY:
        result = mdarray_index_offset(branch->array, coords);
        break;
      default:
        g_assert_not_reached();
        break;
    }
    return result;
}

/* static void normalize(tree_t *tree)
 *
 * Merge tree branches into nodes containing arrays where possible.
 * Called before instantiation.
 */
static void
normalize(tree_t *tree)
{
    g_assert(G_NODE_IS_ROOT(tree));
    root_data_t *rdata = tree->data;
    gboolean normalize_node(GNode *node, gpointer unused G_GNUC_UNUSED) {
        GNode *child = g_node_first_child(node);
        if (child != NULL) {
            gboolean equal_dims = TRUE;
            GArray *dims = NULL;
            // First child becomes reference against which other
            // children are compared.
            if (G_NODE_IS_LEAF(child)) { // only leaves may be merged
                if (child->data != NULL) {
                    dims = ((mdarray_t *)(child->data))->dimensions;
                }
            }
            child = g_node_next_sibling(child);
            // Compare other child node arrays with the first child
            // node's. If first child node doesn't have an array, the
            // other children shouldn't have arrays either if they are
            // to be merged.
            while (equal_dims && child != NULL) {
                mdarray_t *mdarray = child->data;
                equal_dims =
                    (G_NODE_IS_LEAF(child)
                     && ((mdarray == NULL && dims == NULL)
                         || (mdarray != NULL && dims != NULL
                             && mdarray->dimensions->len == dims->len
                             && (memcmp(mdarray->dimensions->data, dims->data,
                                        dims->len * sizeof(guint)) == 0))));
                child = g_node_next_sibling(child);
            }
            guint num_children[] = {g_node_n_children(node), COORD_VECTOR_END};
            if (num_children[0] > 1 && equal_dims) {
                // Child node arrays are equivalent, so merge them.
                // Increase rank of this node's array, and append
                // child node array dimensions to this node's array
                // dimensions.
                mdarray_t *new_md;
                if (dims == NULL) {
                    new_md =
                        mdarray_new(num_children, rdata->element_size);
                }
                else {
                    g_array_prepend_val(dims, num_children[0]);
                    new_md =
                        mdarray_new((guint *)dims->data, rdata->element_size);
                }
                if (G_NODE_IS_ROOT(node)) {
                    ((root_data_t *)(node->data))->mdarray = new_md;
                }
                else {
                    g_assert(node->data == NULL);
                    node->data = new_md;
                }
                // Free merged child nodes.
                child = g_node_first_child(node);
                while (child != NULL) {
                    if (child->data != NULL) {
                        mdarray_free(child->data);
                    }
                    GNode *next = g_node_next_sibling(child);
                    g_node_destroy(child);
                    child = next;
                }
                g_assert(g_node_first_child(node) == NULL);
            }
        }
        return FALSE;
    }

    g_node_traverse(tree, G_POST_ORDER, G_TRAVERSE_ALL, -1,
                    normalize_node, NULL);
}

/* static GNode *scan_node(GScanner *scanner)
 *
 * Scan a node element. Returns a tree node.
 */
static GNode *
scan_node(GScanner *scanner)
{
    GNode *result = NULL;
    gint num_children[] = {0, COORD_VECTOR_END};
    switch (g_scanner_peek_next_token(scanner)) {
      case G_TOKEN_LEFT_CURLY:
        g_scanner_get_next_token(scanner);
        num_children[0] = scan_num_children(scanner);
        if (num_children[0] > 0) {
            result = g_node_new(NULL);
            GPtrArray *proto_children = scan_children(scanner);
            // The first two cases below are optimizations, but they
            // can reduce memory usage significantly in some cases.
            // Effectively, we're partially "normalizing" the tree, or
            // merging nodes with data arrays, on the fly.
            if (proto_children->len == 0) {
                result->data = mdarray_new((guint *)num_children, 1);
            }
            else {
                mdarray_t *chmd =
                    (mdarray_t *)
                    (((tree_t *)g_ptr_array_index(proto_children, 0))->data);
                if (chmd != NULL && proto_children->len == 1) {
                    g_array_prepend_val(chmd->dimensions, num_children[0]);
                    result->data =
                        mdarray_new((guint *)(chmd->dimensions->data), 1);
                    mdarray_free(chmd);
                    g_node_destroy(g_ptr_array_index(proto_children, 0));
                }
                else {
                    // Go through child list cyclically, using and copying
                    // as many list items as necessary to generate
                    // 'num_children' children.
                    guint i;
                    for (i = 0; i < num_children[0]; ++i) {
                        GNode *child;
                        if (i < proto_children->len) {
                            // First time through the list, don't make
                            // copies.
                            child = g_ptr_array_index(proto_children, i);
                        }
                        else {
                            // Reusing list items requires copying them.
                            child = g_node_copy_deep(
                                g_ptr_array_index(proto_children,
                                                  i % proto_children->len),
                                copy_mdarray, NULL);
                        }
                        g_node_append(result, child);
                    }
                    // Free any unused list items.
                    for (; i < proto_children->len; ++i) {
                        g_node_destroy(g_ptr_array_index(proto_children, i));
                    }
                }
            }
            g_ptr_array_free(proto_children, TRUE);
            if (g_scanner_get_next_token(scanner) != G_TOKEN_RIGHT_CURLY) {
                g_scanner_unexp_token(scanner, G_TOKEN_RIGHT_CURLY,
                                      NULL, NULL, NULL, NULL, TRUE);
            }
        }
        else if (num_children[0] == 0) {
            result = g_node_new(NULL);
        }
        else {
            g_scanner_warn(scanner, "Number of children must be at least 0");
        }
        break;

      default:
        g_scanner_get_next_token(scanner);
        g_scanner_unexp_token(scanner, G_TOKEN_LEFT_CURLY,
                              NULL, NULL, NULL, NULL, TRUE);
        break;
    }
    return result;
}

/* static gint scan_num_children(GScanner *scanner)
 *
 * Scan the number of children. Returns the number.
 */
static gint
scan_num_children(GScanner *scanner)
{
    glong result = 0;
    switch (g_scanner_peek_next_token(scanner)) {
      case G_TOKEN_INT:
        g_scanner_get_next_token(scanner);
        result = scanner->value.v_int;
        break;
      case G_TOKEN_RIGHT_CURLY:
        break;
      default:
        g_scanner_get_next_token(scanner);
        g_scanner_unexp_token(scanner, G_TOKEN_INT,
                              NULL, NULL, NULL, NULL, TRUE);
        break;
    }
    return result;
}

/* static GPtrArray *scan_children(GScanner *scanner)
 *
 * Scan node child list. Returns the nodes in a GPtrArray, after
 * "multiplier" expansion.
 */
static GPtrArray *
scan_children(GScanner *scanner)
{
    GPtrArray *result = g_ptr_array_new();
    GNode *node = NULL;
    switch (g_scanner_peek_next_token(scanner)) {
      case G_TOKEN_COMMA:
        g_scanner_get_next_token(scanner);
        if (g_scanner_get_next_token(scanner) == G_TOKEN_LEFT_BRACE) {
            // Loop to extract list items.
            do {
                glong mult = scan_multiplier(scanner);
                // Expand nodes by making copies.
                if (mult > 0) {
                    node = scan_node(scanner);
                    if (node != NULL) {
                        g_ptr_array_add(result, node);
                        while (--mult > 0) {
                            node = g_node_copy_deep(node, copy_mdarray, NULL);
                            g_ptr_array_add(result, node);
                        }
                    }
                }
                else {
                    g_scanner_warn(scanner, "Multiplier must be at least one");
                }
            } while (node != NULL
                     && g_scanner_get_next_token(scanner) == G_TOKEN_COMMA);
            if (node != NULL && scanner->token != G_TOKEN_RIGHT_BRACE) {
                g_scanner_unexp_token(scanner, G_TOKEN_RIGHT_BRACE,
                                      NULL, NULL, NULL, NULL, TRUE);
                guint i;
                for (i = 0; i < result->len; ++i) {
                    g_node_destroy(g_ptr_array_index(result, i));
                }
                g_ptr_array_set_size(result, 0);
            }
        }
        else {
            g_scanner_unexp_token(scanner, G_TOKEN_LEFT_BRACE,
                                  NULL, NULL, NULL, NULL, TRUE);
        }
        break;

      case G_TOKEN_RIGHT_CURLY:
        break;

      default:
        g_scanner_unexp_token(scanner, G_TOKEN_COMMA,
                              NULL, NULL, NULL, NULL,
                              TRUE);
        break;
    }
    return result;
}

/* static glong scan_multiplier(GScanner *scanner)
 *
 * Scan a node multiplier. Return the multiplier.
 */
static glong
scan_multiplier(GScanner *scanner)
{
    glong result = 1;
    if (g_scanner_peek_next_token(scanner) == G_TOKEN_INT) {
        g_scanner_get_next_token(scanner);
        result = scanner->value.v_int;
    }
    return result;
}

/* static gpointer copy_mdarray(gconstpointer src, gpointer data)
 *
 * Copy an "unassigned" mdarray_t element. Should be used when
 * copying tree nodes (before tree memory is allocated).
 */
static gpointer
copy_mdarray(gconstpointer src, gpointer data G_GNUC_UNUSED)
{
    const mdarray_t *mdarray = src;
    mdarray_t *result;
    if (mdarray != NULL && mdarray->dimensions != NULL) {
        result = mdarray_new((guint *)(mdarray->dimensions->data), 1);
    }
    else {
        result = NULL;
    }
    return result;
}

/* static void assign_memory(tree_t *tree, gpointer buffer, off_t offset)
 *
 * Attach memory to store data elements in leaf nodes of a tree.
 *
 * Arguments:
 *   tree - pointer to tree
 *   buffer - pointer to start of memory block to use for storage
 */
static void
assign_memory(tree_t *tree, gpointer buffer, off_t offset)
{
    g_return_if_fail(tree != NULL);
    g_return_if_fail(tree_height(tree) > 0);
    g_return_if_fail(buffer != NULL);

    gboolean array_assign(GNode *node, gpointer unused G_GNUC_UNUSED) {
        mdarray_t *mdarray = NODE_ARRAY(node);
        mdarray_assign_memory(mdarray, buffer, offset);
        offset += mdarray_size(mdarray);
        return FALSE;
    }
    g_node_traverse(tree, G_IN_ORDER, G_TRAVERSE_LEAVES, -1, array_assign,
                    NULL);
}

static void
unassign_memory(tree_t *tree)
{
    g_return_if_fail(tree != NULL);
    gboolean array_unassign(GNode *node, gpointer unused G_GNUC_UNUSED) {
        mdarray_t *mdarray = NODE_ARRAY(node);
        mdarray_unassign(mdarray);
        return FALSE;
    }
    g_node_traverse(tree, G_IN_ORDER, G_TRAVERSE_LEAVES, -1, array_unassign,
                    NULL);
}

static void
assign_stream(tree_t *tree, off_t offset)
{
    g_return_if_fail(tree != NULL);
    g_return_if_fail(tree_height(tree) > 0);

    FILE *stream = tree_stream(tree);
    gboolean array_assign(GNode *node, gpointer unused G_GNUC_UNUSED) {
        mdarray_t *mdarray = NODE_ARRAY(node);
        mdarray_assign_stream(mdarray, stream, offset);
        offset += mdarray_size(mdarray);
        return FALSE;
    }
    g_node_traverse(tree, G_IN_ORDER, G_TRAVERSE_LEAVES, -1, array_assign,
                    NULL);
}

static branch_t *
tree_branch_with_offset(const tree_t *tree, const guint *coordinates,
                        off_t offset)
{
    g_return_val_if_fail(tree != NULL, NULL);
    g_return_val_if_fail(tree_store_type(tree) != STORE_NONE, NULL);

    tree_t *root = g_node_get_root((tree_t *)tree);
    root_data_t *rdata = root->data;
    branch_t *result = NULL;
    gsize mmap_len = 0;
    gboolean coords_ok = TRUE;
    while (coords_ok && *coordinates != COORD_VECTOR_END) {
        mdarray_t *mdarray = NODE_ARRAY(tree);
        if (mdarray == NULL) {
            // scan node children to indexed child, adding to 'offset'
            // along the way
            tree = g_node_first_child(tree);
            guint j;
            for (j = 0; j < *coordinates && tree != NULL; ++j) {
                offset += tree_size(tree);
                tree = g_node_next_sibling(tree);
            }
            if (G_UNLIKELY(tree == NULL)) {
                g_warning("Index out of bounds");
                coords_ok = FALSE;
            }
        }
        else {
            // arrived at an array, use 'mdarray_reduced' to get
            // sub-array
            off_t array_offset;
            mdarray_t *array =
                mdarray_reduced(mdarray, coordinates, &array_offset);
            if (array != NULL) {
                offset += array_offset;
                if (mdarray->buffer != NULL) {
                    if (tree_store_type(tree) == STORE_HEAP) {
                        mdarray_assign_memory(array, mdarray->buffer, offset);
                    }
                }
                else {
                    mdarray_assign_stream(array, mdarray->stream, offset);
                }
                result = g_new0(branch_t, 1);
                result->type = ARRAY;
                result->array = array;
                mmap_len = mdarray_size(array);
                // skip the remainder of the loop
                while (*(coordinates + 1) != COORD_VECTOR_END) {
                    ++coordinates;
                }
            }
            else {
                g_warning("Index out of bounds");
                coords_ok = FALSE;
            }
        }
        ++coordinates;
    }
    if (coords_ok) {
        if (result == NULL) {
            // result is a tree
            result = g_new0(branch_t, 1);
            result->type = TREE;
            result->tree = (tree_t *)tree;
            mmap_len = tree_size(tree);
        }
        switch (tree_store_type(tree)) {
          case STORE_MMAP: {
            long pagesize = sysconf(_SC_PAGESIZE);
            result->mmap_off = (offset / pagesize) * pagesize;
            result->mmap_data_offset = offset % pagesize;
            result->mmap_len = mmap_len + result->mmap_data_offset;
            result->data_offset = offset;
            result->mmap_fd = rdata->mmap_fd;
            result->mmap_prot = rdata->mmap_prot;
            result->mmap_ptr = MAP_FAILED;
            break;
          }
          case STORE_HEAP:
          case STORE_STREAM:
            result->mmap_fd = -1;
            result->data_offset = offset;
            break;
          default:
            g_assert_not_reached();
            break;
        }
    }
    g_assert(tree_store_type(tree) == tree_branch_store_type(result));
    return result;
}

/** tree_memory:
 *  @tree: tree
 *
 * @returns: pointer to the origin of the memory segment backing the tree;
 *           NULL if tree is stream-backed.
 */
static gpointer
tree_memory(const tree_t *tree)
{
    g_return_val_if_fail(tree != NULL, NULL);
    gpointer result = NULL;
    gboolean get_buffer(tree_t *node, gpointer unused G_GNUC_UNUSED) {
        mdarray_t *mdarray = NODE_ARRAY(node);
        result = mdarray->buffer;
        return TRUE;
    }
    g_node_traverse((tree_t *)tree, G_IN_ORDER, G_TRAVERSE_LEAVES, -1,
                    get_buffer, NULL);
    return result;
}

/** tree_stream:
 *  @tree: tree
 *
 * @returns: stream associated with the tree element storage;
 *           NULL if tree is memory-backed.
 */
static FILE *
tree_stream(const tree_t *tree)
{
    g_return_val_if_fail(tree != NULL, NULL);
    return ((root_data_t *)(g_node_get_root((tree_t *)tree)->data))->stream;
}

static store_type_t
tree_branch_store_type(const branch_t *branch)
{
    store_type_t result;
    if (branch->mmap_fd != -1) {
        result = STORE_MMAP;
    }
    else if (tree_branch_stream(branch) != NULL) {
        result = STORE_STREAM;
    }
    else if (tree_branch_memory(branch) != NULL) {
        result = STORE_HEAP;
    }
    else {
        result = STORE_NONE;
    }
    return result;
}

static store_type_t
tree_store_type(const tree_t *tree)
{
    store_type_t result;
    if (((root_data_t *)(g_node_get_root((tree_t *)tree)->data))->mmap_fd
        != -1) {
        result = STORE_MMAP;
    }
    else if (tree_stream(tree) != NULL) {
        result = STORE_STREAM;
    }
    else if (tree_memory(tree) != NULL) {
        result = STORE_HEAP;
    }
    else {
        result = STORE_NONE;
    }
    return result;
}
