#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <libxml/xmlreader.h>
#include <libxml/xmlwriter.h>

#include <wcbe_config.h>
#include <wcbe_util.h>

typedef gint (*element_parser_fn_t)(xmlTextReaderPtr reader,
                                    wcbe_config_t *config,
                                    gpointer user_data);

typedef gint (*attribute_parser_fn_t)(gchar *str, gpointer val);

typedef struct element {
    const gchar *name;
    element_parser_fn_t parser;
    gpointer user_data;
} element_t;

typedef struct attribute {
    const gchar *name;
    attribute_parser_fn_t parser;
    gpointer val;
} attribute_t;

static gint parse_element(xmlTextReaderPtr reader, wcbe_config_t *config,
                          const element_t children[],
                          const attribute_t attributes[]);
static gint parse_attributes(xmlTextReaderPtr reader,
                             const attribute_t attributes[]);
static gint parse_string(gchar *str, gpointer val);
static gint parse_uint(gchar *str, gpointer val);
static gint parse_char(gchar *str, gpointer val);
static void normalize_list_whitespace(gchar *str);
static gint parse_sdmhdr_app(gchar *str, gpointer val);
static gint parse_sdmhdr_xpp(gchar *str, gpointer val);
static gint parse_input_config(xmlTextReaderPtr reader, wcbe_config_t *config,
                               gpointer user_data);
static gint parse_strm(xmlTextReaderPtr reader, wcbe_config_t *config,
                       gpointer user_data);
static gint parse_prd4(xmlTextReaderPtr reader, wcbe_config_t *config,
                       gpointer user_data);
static gint parse_prd7(xmlTextReaderPtr reader, wcbe_config_t *config,
                       gpointer user_data);
static void set_stream_ref(GArray *streams, const gchar *strm_id,
                           wcbe_stream_config_t **strmp);
static gint parse_sPrd(xmlTextReaderPtr reader, wcbe_config_t *config,
                       gpointer user_data);
static gint parse_cIntn(xmlTextReaderPtr reader, wcbe_config_t *config,
                        gpointer user_data);
static gint parse_lcSeg(xmlTextReaderPtr reader, wcbe_config_t *config,
                        gpointer user_data);
static gint parse_inp(xmlTextReaderPtr reader, wcbe_config_t *config,
                      gpointer user_data);
static gint parse_output_config(xmlTextReaderPtr reader, wcbe_config_t *config,
                                gpointer user_data);
static gint parse_subarray(xmlTextReaderPtr reader, wcbe_config_t *config,
                           gpointer user_data);
static gint parse_bb1(xmlTextReaderPtr reader, wcbe_config_t *config,
                      gpointer user_data);
static gint parse_bb2(xmlTextReaderPtr reader, wcbe_config_t *config,
                      gpointer user_data);
static gint parse_sw(xmlTextReaderPtr reader, wcbe_config_t *config,
                     gpointer user_data);
static gint parse_pipeline_config(xmlTextReaderPtr reader,
                                  wcbe_config_t *config,
                                  gpointer user_data);
static gint parse_pln(xmlTextReaderPtr reader, wcbe_config_t *config,
                      gpointer user_data);
static gint parse_spec(xmlTextReaderPtr reader, wcbe_config_t *config,
                       gpointer user_data);
static gint parse_gstSpec(xmlTextReaderPtr reader, wcbe_config_t *config,
                          gpointer user_data);
static gint parse_swmap(xmlTextReaderPtr reader, wcbe_config_t *config,
                        gpointer user_data);
static gint write_config(xmlTextWriterPtr writer, const wcbe_config_t *config);
static gint write_input_config(xmlTextWriterPtr writer,
                               const wcbe_input_config_t *input) ;
static gint write_strm(xmlTextWriterPtr writer,
                       const wcbe_stream_config_t *stream);
static gint write_prd(xmlTextWriterPtr writer,
                      const wcbe_product_config_t *prd);
static gint write_lag_chains(xmlTextWriterPtr writer,
                             const gchar *container, GArray *segs);
static gint write_cIntn(xmlTextWriterPtr writer,
                        const wcbe_integration_config_t *integration);
static gint write_lcSeg(xmlTextWriterPtr writer,
                        const wcbe_lag_chain_config_t *lc);
static gint write_output_config(xmlTextWriterPtr writer,
                                const wcbe_output_config_t *output);
static gint write_subarray(xmlTextWriterPtr writer,
                           const wcbe_subarray_config_t *sa);
static gint write_baseband(xmlTextWriterPtr writer,
                           const wcbe_baseband_config_t *bb);
static gint write_sw(xmlTextWriterPtr writer,
                     const wcbe_spectral_window_config_t *sw);
static gint write_pipeline_config(xmlTextWriterPtr writer,
                                  const wcbe_pipelines_config_t *pconf);
static gint write_pln(xmlTextWriterPtr writer,
                      const wcbe_pipeline_config_t *pln);
static gint write_swmap(xmlTextWriterPtr writer,
                        const wcbe_swmap_config_t *swmap);

static gchar const config_ns_uri[] = "http://evla.nrao.edu/wcbe";

#define ARRAY_APPEND_NEW(a, type) ({\
    guint len = (a)->len; \
    g_array_set_size(a, len + 1); \
    &g_array_index(a, type, len);})

gboolean
wcbe_config_parse(const gchar *doc, wcbe_config_t *config)
{
    gboolean result = FALSE;
    xmlTextReaderPtr reader =
        xmlReaderForMemory(doc, strlen(doc), "", NULL,
                           XML_PARSE_RECOVER | XML_PARSE_NOBLANKS
                           | XML_PARSE_NONET | XML_PARSE_NSCLEAN);
    if (reader != NULL) {
        gint rc = xmlTextReaderRead(reader);
        if (rc == 1) {
            rc = -1;
            switch (xmlTextReaderNodeType(reader)) {
              case XML_READER_TYPE_ELEMENT: {
                const xmlChar *name = xmlTextReaderConstName(reader);
                const xmlChar *nsuri = xmlTextReaderConstNamespaceUri(reader);
                if (xmlStrEqual(BAD_CAST "config", name)
                    && nsuri != NULL
                    && xmlStrEqual(nsuri, BAD_CAST config_ns_uri)) {
                    element_t children[] = {
                        {.name = "inputConfig",
                         .parser = parse_input_config},
                        {.name = "outputConfig",
                         .parser = parse_output_config},
                        {.name = "pipelineConfig",
                         .parser = parse_pipeline_config},
                        {.name = NULL}
                    };
                    memset(config, 0, sizeof(*config));
                    rc = parse_element(reader, config, children, NULL);
                }
                break;
              }
              default:
                break;
            }
        }
        xmlFreeTextReader(reader);
        if (rc != -1) {
            result = TRUE;
        }
        else {
            wcbe_config_free(config);
        }
    }
    return result;
}

void
wcbe_config_free(wcbe_config_t *config)
{
    if (config != NULL) {
        if (config->name != NULL) {
            g_free(config->name);
        }
        if (config->dir != NULL) {
            g_free(config->dir);
        }
        // input
        if (config->input.streams != NULL) {
            guint i;
            for (i = 0; i < config->input.streams->len; ++i) {
                wcbe_stream_config_t *stream =
                    &g_array_index(config->input.streams,
                                   wcbe_stream_config_t, i);
                if (stream->id != NULL) {
                    g_free(stream->id);
                }
            }
            g_array_free(config->input.streams, TRUE);
        }
        if (config->input.products != NULL) {
            guint i;
            for (i = 0; i < config->input.products->len; ++i) {
                wcbe_product_config_t *prd =
                    g_ptr_array_index(config->input.products, i);
                if (prd->id != NULL) {
                    g_free(prd->id);
                }
                if (prd->nbits == WCBE_4BITS) {
                    wcbe_product4_config_t *prd4 =
                        (wcbe_product4_config_t *)prd;
                    if (prd4->lag_chains != NULL) {
                        g_array_free(prd4->lag_chains, TRUE);
                    }
                }
                else if (prd->nbits == WCBE_7BITS) {
                    wcbe_product7_config_t *prd7 =
                        (wcbe_product7_config_t *)prd;
                    if (prd7->lag_chains_ll != NULL) {
                        g_array_free(prd7->lag_chains_ll, TRUE);
                    }
                    if (prd7->lag_chains_lm != NULL) {
                        g_array_free(prd7->lag_chains_lm, TRUE);
                    }
                    if (prd7->lag_chains_ml != NULL) {
                        g_array_free(prd7->lag_chains_ml, TRUE);
                    }
                    if (prd7->lag_chains_mm != NULL) {
                        g_array_free(prd7->lag_chains_mm, TRUE);
                    }
                }
                g_free(prd);
            }
            g_ptr_array_free(config->input.products, TRUE);
        }
        // output
        if (config->output.subarrays != NULL) {
            guint i;
            for (i = 0; i < config->output.subarrays->len; ++i) {
                wcbe_subarray_config_t *sa =
                    &g_array_index(config->output.subarrays,
                                   wcbe_subarray_config_t, i);
                if (sa->stations != NULL) {
                    g_array_free(sa->stations, TRUE);
                }
                if (sa->basebands != NULL) {
                    guint j;
                    for (j = 0; j < sa->basebands->len; ++j) {
                        wcbe_baseband_config_t *bb =
                            &g_array_index(sa->basebands,
                                           wcbe_baseband_config_t, j);
                        if (bb->spectral_windows != NULL) {
                            g_array_free(bb->spectral_windows, TRUE);
                        }
                    }
                    g_array_free(sa->basebands, TRUE);
                }
            }
            g_array_free(config->output.subarrays, TRUE);
        }
        // pipelines
        if (config->pipelines.pipelines != NULL) {
            guint i;
            for (i = 0; i < config->pipelines.pipelines->len; ++i) {
                wcbe_pipeline_config_t *pipeline = &g_array_index(
                    config->pipelines.pipelines, wcbe_pipeline_config_t, i);
                if (pipeline->id != NULL) {
                    g_free(pipeline->id);
                }
                if (pipeline->spec.gst_spec != NULL) {
                    g_free(pipeline->spec.gst_spec);
                }
                if (pipeline->swmaps != NULL) {
                    guint j;
                    for (j = 0; j < pipeline->swmaps->len; ++j) {
                        wcbe_swmap_config_t *swmap = &g_array_index(
                            pipeline->swmaps, wcbe_swmap_config_t, j);
                        if (swmap->subband_indices != NULL) {
                            g_array_free(swmap->subband_indices, TRUE);
                        }
                    }
                    g_array_free(pipeline->swmaps, TRUE);
                }
            }
            g_array_free(config->pipelines.pipelines, TRUE);
        }
    }
}

void
wcbe_config_doc(const wcbe_config_t *config, gchar **buf)
{
    g_return_if_fail(config != NULL);

    *buf = NULL;

    xmlBufferPtr xbuf = xmlBufferCreate();
    if (xbuf != NULL) {
        xmlTextWriterPtr writer = xmlNewTextWriterMemory(xbuf, 0);
        xmlTextWriterSetIndent(writer, 1);
        if (writer != NULL) {
            int rc = xmlTextWriterStartDocument(writer, NULL, "utf-8", "yes");
            if (rc >= 0) {
                gboolean doc_written = FALSE;
                rc = write_config(writer, config);
                if (rc >= 0) {
                    doc_written = TRUE;
                }
                else {
                    g_info("Failed to write configuration document");
                }
                rc = xmlTextWriterEndDocument(writer);
                if (rc >= 0) {
                    if (doc_written) {
                        xmlTextWriterFlush(writer);
                        *buf = g_malloc(xbuf->use + 1);
                        memcpy(*buf, xbuf->content, xbuf->use + 1);
                    }
                }
                else {
                    g_info("Failed to end configuration document");
                }
            }
            else {
                g_info("Failed to start configuration document");
            }
            xmlFreeTextWriter(writer);
        }
        else {
            g_info("Failed to create xmlTextWriter for configuration "
                   "document");
        }
        xmlBufferFree(xbuf);
    }
    else {
        g_info("Failed to allocate xmlBuffer for configuration document ");
    }
}

static gint
parse_element(xmlTextReaderPtr reader, wcbe_config_t *config,
              const element_t children[], const attribute_t attributes[])
{
    gint result = -1;
    const element_t *scan_children(const xmlChar *name) {
        const element_t *ch = children;
        element_parser_fn_t parser = NULL;
        if (children != NULL) {
            while (parser == NULL && ch->name != NULL) {
                if (xmlStrEqual(BAD_CAST ch->name, name)) {
                    parser = ch->parser;
                }
                else {
                    ++ch;
                }
            }
        }
        return ((parser == NULL) ? NULL : ch);
    }
    const xmlChar *element_name = xmlTextReaderConstName(reader);
    gint rc = parse_attributes(reader, attributes);
    if (rc == 0) {
        result = 1;
        if (!xmlTextReaderIsEmptyElement(reader)) {
            gint rc = xmlTextReaderRead(reader);
            while (result == 1 && rc == 1) {
                const xmlChar *name = xmlTextReaderConstName(reader);
                const xmlChar *nsuri = xmlTextReaderConstNamespaceUri(reader);
                switch (xmlTextReaderNodeType(reader)) {
                  case XML_READER_TYPE_ELEMENT: {
                    if (nsuri != NULL
                        && xmlStrEqual(nsuri, BAD_CAST config_ns_uri)) {
                        const element_t *ch = scan_children(name);
                        if (ch != NULL) {
                            rc = ch->parser(reader, config, ch->user_data);
                        }
                        else {
                            result = -1;
                        }
                    }
                    else {
                        result = -1;
                    }
                    break;
                  }
                  case XML_READER_TYPE_END_ELEMENT: {
                    if (xmlStrEqual(element_name, name)
                        && nsuri != NULL
                        && xmlStrEqual(nsuri, BAD_CAST config_ns_uri)) {
                        rc = 0;
                    }
                    else {
                        result = -1;
                    }
                    break;
                  }
                  default:
                    break;
                }
                if (result != -1 && rc == 1) {
                    rc = xmlTextReaderRead(reader);
                    if (rc == 0) {
                        result = 0;
                    }
                }
            }
            if (rc == -1) {
                result = -1;
            }
        }
    }
    return result;
}

static gint
parse_attributes(xmlTextReaderPtr reader, const attribute_t attributes[])
{
    gint result = 0;
    if (attributes != NULL) {
        const attribute_t *attr = attributes;
        while (result == 0 && attr->name != NULL) {
            xmlChar *val_str =
                xmlTextReaderGetAttribute(reader, BAD_CAST attr->name);
            if (val_str != NULL) {
                result = attr->parser((gchar *)val_str, attr->val);
                xmlFree(val_str);
            }
            ++attr;
        }
    }
    return result;
}

static gint
parse_string(gchar *str, gpointer val)
{
    gchar **cpy = val;
    *cpy = g_strdup(str);
    return 0;
}

static gint
parse_uint(gchar *str, gpointer val)
{
    gint result = 0;
    guint *uintval = val;
    errno = 0;
    gulong ulongval = strtoul(str, NULL, 10);
    if (ulongval > G_MAXUINT) {
        result = -1;
    }
    else {
        *uintval = (guint)ulongval;
        result = errno;
    }
    return result;
}

static gint
parse_char(gchar *str, gpointer val)
{
    gchar *chr = val;
    *chr = str[0];
    return 0;
}

static gint
parse_uint_list(gchar *str, gpointer val)
{
    gint result = 0;
    gchar **strvals = g_strsplit(str, " ", 0);
    gchar **strval0 = strvals;
    guint num_vals = 0;
    while (*strval0++ != NULL) {
        ++num_vals;
    }
    GArray *uints = val;
    g_array_set_size(uints, num_vals);
    guint i = 0;
    strval0 = strvals;
    while (result == 0 && *strval0 != NULL) {
        result =
            parse_uint(*strval0, &g_array_index(uints, guint, i++));
        ++strval0;
    }
    g_strfreev(strvals);
    return result;
}

/**
 * normalize_list_whitespace:
 * @str: string of space-separated tokens
 *
 * Strip leading and trailing space characters from @str, and reduce
 * each run of multiple space characters between tokens to a single
 * space character.
 */
static void
normalize_list_whitespace(gchar *str)
{
    guint len = strlen(str) + 1;
    gchar *c = str;
    while (*c == ' ') {
        ++c;
        --len;
    }
    if (c != str) {
        memmove(str, c, len);
    }
    while (*c != '\0') {
        if (*c++ == ' ') {
            gchar *d = c;
            while (*d == ' ') {
                ++d;
                --len;
            }
            if (d != c) {
                memmove(c, d, len);
            }
        }
        --len;
    }
    if (*(c - 1) == ' ') {
        *(c - 1) = '\0';
    }
}

static gint
parse_sdmhdr_app(gchar *str, gpointer val)
{
    gint result = 0;
    sdmhdr_app_t *app = val;
    normalize_list_whitespace(str);
    if (strcmp(str, "RR") == 0) {
        *app = SDMHDR_APP_RR;
    }
    else if (strcmp(str, "LL") == 0) {
        *app = SDMHDR_APP_LL;
    }
    else if (strcmp(str, "RR LL") == 0) {
        *app = SDMHDR_APP_RR_LL;
    }
    else if (strcmp(str, "RR RL LL") == 0) {
        *app = SDMHDR_APP_RR_RL_LL;
    }
    else {
        *app = SDMHDR_APP_INVALID;
        result = -1;
    }
    return result;
}

static gint
parse_sdmhdr_xpp(gchar *str, gpointer val)
{
    gint result = 0;
    sdmhdr_xpp_t *xpp = val;
    normalize_list_whitespace(str);
    if (strcmp(str, "RR") == 0) {
        *xpp = SDMHDR_XPP_RR;
    }
    else if (strcmp(str, "LL") == 0) {
        *xpp = SDMHDR_XPP_LL;
    }
    else if (strcmp(str, "RR LL") == 0) {
        *xpp = SDMHDR_XPP_RR_LL;
    }
    else if (strcmp(str, "RR RL LR LL") == 0) {
        *xpp = SDMHDR_XPP_RR_RL_LR_LL;
    }
    else {
        *xpp = SDMHDR_XPP_INVALID;
        result = -1;
    }
    return result;
}

static gint
parse_input_config(xmlTextReaderPtr reader, wcbe_config_t *config,
                   gpointer user_data G_GNUC_UNUSED)
{
    element_t children[] = {
        {.name = "strm", .parser = parse_strm},
        {.name = "prd4", .parser = parse_prd4},
        {.name = "prd7", .parser = parse_prd7},
        {.name = NULL}
    };
    return parse_element(reader, config, children, NULL);
}

static gint
parse_strm(xmlTextReaderPtr reader, wcbe_config_t *config,
           gpointer user_data G_GNUC_UNUSED)
{
    if (config->input.streams == NULL) {
        config->input.streams =
            g_array_new(FALSE, TRUE, sizeof(wcbe_stream_config_t));
    }
    wcbe_stream_config_t *strm =
        ARRAY_APPEND_NEW(config->input.streams, wcbe_stream_config_t);
    attribute_t attributes[] = {
        {.name = "id", .parser = parse_string, .val = &(strm->id)},
        {.name = "stn", .parser = parse_uint, .val = &(strm->station_id)},
        {.name = "bb", .parser = parse_uint, .val = &(strm->baseband_id)},
        {.name = "sb", .parser = parse_uint, .val = &(strm->subband_id)},
        {.name = NULL}
    };
    return parse_element(reader, config, NULL, attributes);
}

static gint
parse_prd4(xmlTextReaderPtr reader, wcbe_config_t *config,
           gpointer user_data G_GNUC_UNUSED)
{
    if (config->input.products == NULL) {
        config->input.products = g_ptr_array_new();
    }
    wcbe_product4_config_t *prd4 = g_new0(wcbe_product4_config_t, 1);
    prd4->common.nbits = WCBE_4BITS;
    prd4->lag_chains =
        g_array_new(FALSE, TRUE, sizeof(wcbe_lag_chain_config_t));
    g_ptr_array_add(config->input.products, prd4);
    gchar *strmA, *strmB;
    attribute_t attributes[] = {
        {.name = "id", .parser = parse_string,
         .val = &(prd4->common.id)},
        {.name = "nLg", .parser = parse_uint,
         .val = &(prd4->common.num_lags)},
        {.name = "nBn", .parser = parse_uint,
         .val = &(prd4->common.num_bins)},
        {.name = "strmA", .parser = parse_string,
         .val = &strmA},
        {.name = "strmB", .parser = parse_string,
         .val = &strmB},
        {.name = "sbIdx", .parser = parse_uint,
         .val = &(prd4->common.subband_index)},
        {.name = NULL}
    };
    element_t children[] = {
        {.name = "cIntn", .parser = parse_cIntn,
         .user_data = prd4},
        {.name = "lcSeg", .parser = parse_lcSeg,
         .user_data = prd4->lag_chains},
        {.name = NULL}
    };
    gint result = parse_element(reader, config, children, attributes);
    if (result == 1) {
        guint i;
        for (i = 0; i < prd4->lag_chains->len; ++i) {
            wcbe_lag_chain_config_t *lc =
                &g_array_index(prd4->lag_chains, wcbe_lag_chain_config_t, i);
            lc->subprod = 0;
        }
        g_assert(strmA != NULL);
        set_stream_ref(config->input.streams, strmA, &(prd4->common.streamA));
        g_free(strmA);
        g_assert(strmB != NULL);
        set_stream_ref(config->input.streams, strmB, &(prd4->common.streamB));
        g_free(strmB);
    }
    return result;
}

static gint
parse_prd7(xmlTextReaderPtr reader, wcbe_config_t *config,
           gpointer user_data G_GNUC_UNUSED)
{
    if (config->input.products == NULL) {
        config->input.products = g_ptr_array_new();
    }
    wcbe_product7_config_t *prd7 = g_new0(wcbe_product7_config_t, 1);
    prd7->common.nbits = WCBE_7BITS;
    prd7->lag_chains_ll =
        g_array_new(FALSE, TRUE, sizeof(wcbe_lag_chain_config_t));
    prd7->lag_chains_lm =
        g_array_new(FALSE, TRUE, sizeof(wcbe_lag_chain_config_t));
    prd7->lag_chains_ml =
        g_array_new(FALSE, TRUE, sizeof(wcbe_lag_chain_config_t));
    prd7->lag_chains_mm =
        g_array_new(FALSE, TRUE, sizeof(wcbe_lag_chain_config_t));
    g_ptr_array_add(config->input.products, prd7);
    gchar *strmA, *strmB;
    attribute_t attributes[] = {
        {.name = "id", .parser = parse_string,
         .val = &(prd7->common.id)},
        {.name = "nLg", .parser = parse_uint,
         .val = &(prd7->common.num_lags)},
        {.name = "nBn", .parser = parse_uint,
         .val = &(prd7->common.num_bins)},
        {.name = "strmA", .parser = parse_string,
         .val = &strmA},
        {.name = "strmB", .parser = parse_string,
         .val = &strmB},
        {.name = "sbIdx", .parser = parse_uint,
         .val = &(prd7->common.subband_index)},
        {.name = NULL}
    };
    element_t children[] = {
        {.name = "cIntn", .parser = parse_cIntn,
         .user_data = prd7},
        {.name = "sPrdLL", .parser = parse_sPrd,
         .user_data = prd7->lag_chains_ll},
        {.name = "sPrdLM", .parser = parse_sPrd,
         .user_data = prd7->lag_chains_lm},
        {.name = "sPrdML", .parser = parse_sPrd,
         .user_data = prd7->lag_chains_ml},
        {.name = "sPrdMM", .parser = parse_sPrd,
         .user_data = prd7->lag_chains_mm},
        {.name = NULL}
    };
    gint result = parse_element(reader, config, children, attributes);
    if (result == 1) {
        guint i;
        for (i = 0; i < prd7->lag_chains_ll->len; ++i) {
            wcbe_lag_chain_config_t *lc = &g_array_index(
                prd7->lag_chains_ll, wcbe_lag_chain_config_t, i);
            lc->subprod = WCBE_7BIT_LL;
        }
        for (i = 0; i < prd7->lag_chains_lm->len; ++i) {
            wcbe_lag_chain_config_t *lc = &g_array_index(
                prd7->lag_chains_lm, wcbe_lag_chain_config_t, i);
            lc->subprod = WCBE_7BIT_LM;
        }
        for (i = 0; i < prd7->lag_chains_ml->len; ++i) {
            wcbe_lag_chain_config_t *lc = &g_array_index(
                prd7->lag_chains_ml, wcbe_lag_chain_config_t, i);
            lc->subprod = WCBE_7BIT_ML;
        }
        for (i = 0; i < prd7->lag_chains_mm->len; ++i) {
            wcbe_lag_chain_config_t *lc = &g_array_index(
                prd7->lag_chains_mm, wcbe_lag_chain_config_t, i);
            lc->subprod = WCBE_7BIT_MM;
        }
        g_assert(strmA != NULL);
        set_stream_ref(config->input.streams, strmA, &(prd7->common.streamA));
        g_free(strmA);
        g_assert(strmB != NULL);
        set_stream_ref(config->input.streams, strmB, &(prd7->common.streamB));
        g_free(strmB);
    }
    return result;
}

static void
set_stream_ref(GArray *streams, const gchar *strm_id,
               wcbe_stream_config_t **strmp)
{
    *strmp = NULL;
    guint i;
    for (i = 0; *strmp == NULL && i < streams->len; ++i) {
        wcbe_stream_config_t *s =
            &g_array_index(streams, wcbe_stream_config_t, i);
        if (strcmp(strm_id, s->id) == 0) {
            *strmp = s;
        }
    }
}

static gint
parse_sPrd(xmlTextReaderPtr reader, wcbe_config_t *config, gpointer user_data)
{
    element_t children[] = {
        {.name = "lcSeg", .parser = parse_lcSeg, .user_data = user_data},
        {.name = NULL}
    };
    return parse_element(reader, config, children, NULL);
}

static gint
parse_cIntn(xmlTextReaderPtr reader, wcbe_config_t *config, gpointer user_data)
{
    wcbe_product_config_t *prd = user_data;
    attribute_t attributes[] = {
        {.name = "minHW", .parser = parse_uint,
         .val = &(prd->integration.min_hw)},
        {.name = "hw", .parser = parse_uint,
         .val = &(prd->integration.hw)},
        {.name = "lta", .parser = parse_uint,
         .val = &(prd->integration.lta)},
        {.name = NULL}
    };
    return parse_element(reader, config, NULL, attributes);
}

static gint
parse_lcSeg(xmlTextReaderPtr reader, wcbe_config_t *config, gpointer user_data)
{
    GArray *lag_chains = user_data;
    wcbe_lag_chain_config_t *lc =
        ARRAY_APPEND_NEW(lag_chains, wcbe_lag_chain_config_t);
    guint board_id;
    attribute_t attributes[] = {
        {.name = "bdId", .parser = parse_uint, .val = &board_id},
        {.name = "x", .parser = parse_uint, .val = &(lc->ccX) },
        {.name = "y", .parser = parse_uint, .val = &(lc->ccY)},
        {.name = "ccc0", .parser = parse_uint , .val = &(lc->ccc0)},
        {.name = "cccN", .parser = parse_uint, .val = &(lc->cccN)},
        {.name = "rec", .parser = parse_uint, .val = &(lc->recirc)},
        {.name = NULL}
    };
    element_t children[] = {
        {.name = "xInp", .parser = parse_inp, .user_data = &(lc->inpX)},
        {.name = "yInp", .parser = parse_inp, .user_data = &(lc->inpY)},
        {.name = NULL}
    };
    gint result = parse_element(reader, config, children, attributes);
    lc->board_id = (guint16)board_id;
    return result;
}

static gint
parse_inp(xmlTextReaderPtr reader, wcbe_config_t *config, gpointer user_data)
{
    wcbe_input_stream_config_t *inp = user_data;
    attribute_t attributes[] = {
        {.name = "strm", .parser = parse_char, .val = &(inp->stream)},
        {.name = "inpBlk", .parser = parse_uint, .val = &(inp->lag_block)},
        {.name = NULL}
    };
    return parse_element(reader, config, NULL, attributes);
}

static gint
parse_output_config(xmlTextReaderPtr reader, wcbe_config_t *config,
                    gpointer user_data G_GNUC_UNUSED)
{
    element_t child_elements[] = {
        {.name = "subarray", .parser = parse_subarray},
        {.name = NULL}
    };
    return parse_element(reader, config, child_elements, NULL);
}

static gint
parse_subarray(xmlTextReaderPtr reader, wcbe_config_t *config,
               gpointer user_data G_GNUC_UNUSED)
{
    if (config->output.subarrays == NULL) {
        config->output.subarrays =
            g_array_new(FALSE, TRUE, sizeof(wcbe_subarray_config_t));
    }
    wcbe_subarray_config_t *subarray =
        ARRAY_APPEND_NEW(config->output.subarrays, wcbe_subarray_config_t);
    subarray->stations = g_array_new(FALSE, TRUE, sizeof(guint));
    attribute_t attributes[] = {
        {.name = "stns", .parser = parse_uint_list,
         .val = subarray->stations},
        {.name = "nAPC", .parser = parse_uint,
         .val = &(subarray->num_apc)},
        {.name = "intn", .parser = parse_uint,
         .val = &(subarray->integration)},
        {.name = NULL}
    };
    element_t child_elements[] = {
        {.name = "bb1", .parser = parse_bb1, .user_data = subarray},
        {.name = "bb2", .parser = parse_bb2, .user_data = subarray},
        {.name = NULL},
    };
    return parse_element(reader, config, child_elements, attributes);
}

static gint
parse_bb1(xmlTextReaderPtr reader, wcbe_config_t *config, gpointer user_data)
{
    wcbe_subarray_config_t *subarray = user_data;
    if (subarray->basebands == NULL) {
        subarray->basebands =
            g_array_new(FALSE, TRUE, sizeof(wcbe_baseband_config_t));
    }
    wcbe_baseband_config_t *baseband =
        ARRAY_APPEND_NEW(subarray->basebands, wcbe_baseband_config_t);
    baseband->num_polarizations = 1;
    baseband->baseband_ids[0] = -1;
    baseband->baseband_ids[1] = -1;
    GArray *baseband_ids = g_array_new(FALSE, FALSE, sizeof(guint));
    attribute_t attributes[] = {
        {.name = "bbs", .parser = parse_uint_list, .val = baseband_ids},
        {.name = NULL}
    };
    element_t child_elements[] = {
        {.name= "sw", .parser = parse_sw, .user_data = baseband},
        {.name = NULL}
    };
    gint result = parse_element(reader, config, child_elements, attributes);
    if (result == 1) {
        if (baseband_ids->len > 0) {
            baseband->baseband_ids[0] =
                g_array_index(baseband_ids, guint, 0);
            if (baseband_ids->len > 1) {
                baseband->baseband_ids[1] =
                    g_array_index(baseband_ids, guint, 1);
            }
            // It should be invalid to have polarization products
            // other than "RR" in these elements, but that restriction
            // is not (yet) in the schema, so we enforce it here.
            guint i;
            for (i = 0; i < baseband->spectral_windows->len; ++i) {
                wcbe_spectral_window_config_t *sw =
                    &g_array_index(baseband->spectral_windows,
                                   wcbe_spectral_window_config_t, i);
                if (sw->sdpp != SDMHDR_APP_INVALID
                    && sw->sdpp != SDMHDR_APP_NONE) {
                    sw->sdpp = SDMHDR_APP_RR;
                }
                if (sw->xpp != SDMHDR_XPP_INVALID
                    && sw->xpp != SDMHDR_XPP_NONE) {
                    sw->xpp = SDMHDR_XPP_RR;
                }
            }
        }
        else {
            result = -1;
        }
    }
    g_array_free(baseband_ids, TRUE);
    return result;
}

static gint
parse_bb2(xmlTextReaderPtr reader, wcbe_config_t *config, gpointer user_data)
{
    wcbe_subarray_config_t *subarray = user_data;
    if (subarray->basebands == NULL) {
        subarray->basebands =
            g_array_new(FALSE, TRUE, sizeof(wcbe_baseband_config_t));
    }
    wcbe_baseband_config_t *baseband =
        ARRAY_APPEND_NEW(subarray->basebands, wcbe_baseband_config_t);
    baseband->num_polarizations = 2;
    attribute_t attributes[] = {
        {.name = "bbL", .parser = parse_uint,
         .val = &(baseband->baseband_ids[0])},
        {.name = "bbR", .parser = parse_uint,
         .val = &(baseband->baseband_ids[1])},
        {.name = NULL}
    };
    element_t child_elements[] = {
        {.name= "sw", .parser = parse_sw, .user_data = baseband},
        {.name = NULL}
    };
    return parse_element(reader, config, child_elements, attributes);
}

static gint
parse_sw(xmlTextReaderPtr reader, wcbe_config_t *config, gpointer user_data)
{
    wcbe_baseband_config_t *baseband = user_data;
    if (baseband->spectral_windows == NULL) {
        baseband->spectral_windows =
            g_array_new(FALSE, TRUE, sizeof(wcbe_spectral_window_config_t));
    }
    wcbe_spectral_window_config_t *sw =
        ARRAY_APPEND_NEW(baseband->spectral_windows,
                         wcbe_spectral_window_config_t);
    sw->sdpp = SDMHDR_APP_NONE;
    sw->xpp = SDMHDR_XPP_NONE;
    attribute_t attributes[] = {
        {.name = "nSb", .parser = parse_uint,
         .val = &(sw->num_subbands)},
        {.name = "nBn", .parser = parse_uint,
         .val = &(sw->num_bins)},
        {.name = "nCh", .parser = parse_uint,
         .val = &(sw->num_channels)},
        {.name = "sdpp", .parser = parse_sdmhdr_app,
         .val = &(sw->sdpp)},
        {.name = "xpp", .parser = parse_sdmhdr_xpp,
         .val = &(sw->xpp)},
        {.name = NULL}
    };
    gint result = parse_element(reader, config, NULL, attributes);
    if (sw->xpp == SDMHDR_XPP_INVALID || sw->sdpp == SDMHDR_APP_INVALID) {
        result = -1;
    }
    return result;
}

static gint
parse_pipeline_config(xmlTextReaderPtr reader, wcbe_config_t *config,
                      gpointer user_data G_GNUC_UNUSED)
{
    element_t child_elements[] = {
        {.name = "pln", .parser = parse_pln},
        {.name = NULL}
    };
    return parse_element(reader, config, child_elements, NULL);
}

static gint
parse_pln(xmlTextReaderPtr reader, wcbe_config_t *config,
          gpointer user_data G_GNUC_UNUSED)
{
    if (config->pipelines.pipelines == NULL) {
        config->pipelines.pipelines =
            g_array_new(FALSE, TRUE, sizeof(wcbe_pipeline_config_t));
    }
    wcbe_pipeline_config_t *pipeline =
        ARRAY_APPEND_NEW(config->pipelines.pipelines, wcbe_pipeline_config_t);
    attribute_t attributes[] = {
        {.name = "id", .parser = parse_string,
         .val = &(pipeline->id)},
        {.name = "sa", .parser = parse_uint,
         .val = &(pipeline->subarray_index)},
        {.name = NULL}
    };
    element_t child_elements[] = {
        {.name = "spec", .parser = parse_spec, .user_data = pipeline},
        {.name = "swmap", .parser = parse_swmap, .user_data = pipeline},
        {.name = NULL}
    };
    return parse_element(reader, config, child_elements, attributes);
}

static gint
parse_spec(xmlTextReaderPtr reader, wcbe_config_t *config, gpointer user_data)
{
    element_t child_elements[] = {
        {.name = "gstSpec", .parser = parse_gstSpec, .user_data = user_data},
        {.name = NULL}
    };
    return parse_element(reader, config, child_elements, NULL);
}

static gint
parse_gstSpec(xmlTextReaderPtr reader, wcbe_config_t *config,
              gpointer user_data)
{
    wcbe_pipeline_config_t *pipeline = user_data;
    gint result = xmlTextReaderRead(reader);
    if (result == 1 && xmlTextReaderNodeType(reader) == XML_READER_TYPE_TEXT) {
        xmlChar *text = xmlTextReaderValue(reader);
        pipeline->spec.gst_spec = g_strdup((gchar *)text);
        xmlFree(text);
        gchar *ch = pipeline->spec.gst_spec;
        while (*ch != '\0') {
            if (*ch == '\n') {
                *ch = ' ';
            }
            ++ch;
        }
        g_strstrip(pipeline->spec.gst_spec);
        result = xmlTextReaderRead(reader);
        if (result == 1) {
            const xmlChar *name = xmlTextReaderConstName(reader);
            const xmlChar *nsuri = xmlTextReaderConstNamespaceUri(reader);
            if (xmlTextReaderNodeType(reader) != XML_READER_TYPE_END_ELEMENT
                || !xmlStrEqual(name, BAD_CAST "gstSpec")
                || nsuri == NULL
                || !xmlStrEqual(nsuri, BAD_CAST config_ns_uri)) {
                result = -1;
            }
        }
    }
    else {
        result = -1;
    }
    return result;
}

static gint
parse_swmap(xmlTextReaderPtr reader, wcbe_config_t *config, gpointer user_data)
{
    wcbe_pipeline_config_t *pipeline = user_data;
    if (pipeline->swmaps == NULL) {
        pipeline->swmaps =
            g_array_new(FALSE, TRUE, sizeof(wcbe_swmap_config_t));
    }
    wcbe_swmap_config_t *swmap =
        ARRAY_APPEND_NEW(pipeline->swmaps, wcbe_swmap_config_t);
    swmap->subband_indices = g_array_new(FALSE, TRUE, sizeof(guint));
    attribute_t attributes[] = {
        {.name = "bb", .parser = parse_uint,
         .val = &(swmap->baseband_index)},
        {.name = "sw", .parser = parse_uint,
         .val = &(swmap->spectral_window_index)},
        {.name = "sbs", .parser = parse_uint_list,
         .val = swmap->subband_indices},
        {.name = NULL}
    };
    return parse_element(reader, config, NULL, attributes);
}

static gint
write_config(xmlTextWriterPtr writer, const wcbe_config_t *config)
{
    gint result = -1;

    if (xmlTextWriterStartElementNS(writer, NULL, BAD_CAST "config",
                                    BAD_CAST config_ns_uri) >= 0
        && write_input_config(writer, &(config->input)) >= 0
        && write_output_config(writer, &(config->output)) >= 0
        && write_pipeline_config(writer, &(config->pipelines)) >= 0
        && xmlTextWriterEndElement(writer) >= 0) {
        result = 0;
    }
    return result;
}

static gint
write_input_config(xmlTextWriterPtr writer, const wcbe_input_config_t *input)
{
    gint result = -1;

    if (xmlTextWriterStartElement(writer, BAD_CAST "inputConfig") >= 0) {
        result = 0;
        guint i;
        if (input->streams != NULL) {
            for (i = 0; result >= 0 && i < input->streams->len; ++i) {
                result = write_strm(
                    writer,
                    &g_array_index(input->streams, wcbe_stream_config_t, i));
            }
        }
        if (input->products != NULL) {
            for (i = 0; result >= 0 && i < input->products->len; ++i) {
                result = write_prd(writer,
                                   g_ptr_array_index(input->products, i));
            }
        }
        if (result >= 0) {
            result = xmlTextWriterEndElement(writer);
        }
    }
    return result;
}

static gint
write_strm(xmlTextWriterPtr writer, const wcbe_stream_config_t *stream)
{
    gint result = -1;

    if (xmlTextWriterStartElement(writer, BAD_CAST "strm") >= 0
        && xmlTextWriterWriteAttribute(
            writer, BAD_CAST "id", BAD_CAST stream->id) >= 0
        && xmlTextWriterWriteFormatAttribute(
            writer, BAD_CAST "stn", "%u", stream->station_id) >= 0
        && xmlTextWriterWriteFormatAttribute(
            writer, BAD_CAST "bb", "%u", stream->baseband_id) >= 0
        && xmlTextWriterWriteFormatAttribute(
            writer, BAD_CAST "sb", "%u", stream->subband_id) >= 0
        && xmlTextWriterEndElement(writer) >= 0) {
        result = 0;
    }
    return result;
}

static gint
write_prd(xmlTextWriterPtr writer, const wcbe_product_config_t *prd)
{
    gint result = -1;

    const gchar *element_name = ((prd->nbits == WCBE_4BITS) ? "prd4" : "prd7");
    if (xmlTextWriterStartElement(writer, BAD_CAST element_name) >= 0
        && xmlTextWriterWriteAttribute(
            writer, BAD_CAST "id", BAD_CAST prd->id) >= 0
        && xmlTextWriterWriteAttribute(
            writer, BAD_CAST "strmA", BAD_CAST prd->streamA->id) >= 0
        && xmlTextWriterWriteAttribute(
            writer, BAD_CAST "strmB", BAD_CAST prd->streamB->id) >= 0
        && xmlTextWriterWriteFormatAttribute(
            writer, BAD_CAST "nLg", "%u", prd->num_lags) >= 0
        && xmlTextWriterWriteFormatAttribute(
            writer, BAD_CAST "nBn", "%u", prd->num_bins) >= 0
        && xmlTextWriterWriteFormatAttribute(
            writer, BAD_CAST "sbIdx", "%u", prd->subband_index) >= 0
        && write_cIntn(writer, &(prd->integration)) >= 0) {
        if (prd->nbits == WCBE_4BITS) {
            const wcbe_product4_config_t *prd4 =
                (const wcbe_product4_config_t *)prd;
            result = write_lag_chains(writer, NULL, prd4->lag_chains);
        }
        else {
            const wcbe_product7_config_t *prd7 =
                (const wcbe_product7_config_t *)prd;
            if (write_lag_chains(writer, "sPrdLL",
                                 prd7->lag_chains_ll) >= 0
                && write_lag_chains(writer, "sPrdLM",
                                    prd7->lag_chains_lm) >= 0
                && write_lag_chains(writer, "sPrdML",
                                    prd7->lag_chains_ml) >= 0
                && write_lag_chains(writer, "sPrdMM",
                                    prd7->lag_chains_mm) >= 0) {
                result = 0;
            }
        }
        if (result >= 0) {
            result = xmlTextWriterEndElement(writer);
        }
    }
    return result;
}

static gint
write_lag_chains(xmlTextWriterPtr writer, const gchar *container, GArray *segs)
{
    gint result = 0;

    if (container != NULL) {
        result = xmlTextWriterStartElement(writer, BAD_CAST container);
    }
    guint i;
    for (i = 0; result >= 0 && i < segs->len; ++i) {
        result = write_lcSeg(
            writer,
            &g_array_index(segs, wcbe_lag_chain_config_t, i));
    }
    if (container != NULL && result >= 0) {
        result = xmlTextWriterEndElement(writer);
    }
    return result;
}

static gint
write_cIntn(xmlTextWriterPtr writer,
            const wcbe_integration_config_t *integration)
{
    gint result = -1;

    if (xmlTextWriterStartElement(writer, BAD_CAST "cIntn") >= 0
        && xmlTextWriterWriteFormatAttribute(
            writer, BAD_CAST "minHW", "%u", integration->min_hw) >= 0
        && xmlTextWriterWriteFormatAttribute(
            writer, BAD_CAST "hw", "%u", integration->hw) >= 0
        && xmlTextWriterWriteFormatAttribute(
            writer, BAD_CAST "lta", "%u", integration->lta) >= 0
        && xmlTextWriterEndElement(writer) >= 0) {
        result = 0;
    }
    return result;
}

static gint
write_lcSeg(xmlTextWriterPtr writer, const wcbe_lag_chain_config_t *lc)
{
    gint result = -1;

    if (xmlTextWriterStartElement(writer, BAD_CAST "lcSeg") >= 0
        && xmlTextWriterWriteFormatAttribute(
            writer, BAD_CAST "bdId", "%u", lc->board_id) >= 0
        && xmlTextWriterWriteFormatAttribute(
            writer, BAD_CAST "x", "%u", lc->ccX) >= 0
        && xmlTextWriterWriteFormatAttribute(
            writer, BAD_CAST "y", "%u", lc->ccY) >= 0
        && xmlTextWriterWriteFormatAttribute(
            writer, BAD_CAST "ccc0", "%u", lc->ccc0) >= 0
        && xmlTextWriterWriteFormatAttribute(
            writer, BAD_CAST "cccN", "%u", lc->cccN) >= 0
        && xmlTextWriterWriteFormatAttribute(
            writer, BAD_CAST "rec", "%u", lc->recirc) >= 0

        && xmlTextWriterStartElement(writer, BAD_CAST "xInp") >= 0
        && xmlTextWriterWriteAttribute(
            writer, BAD_CAST "strm",
            BAD_CAST ((lc->inpX.stream == 'A') ? "A" : "B")) >= 0
        && xmlTextWriterWriteFormatAttribute(
            writer, BAD_CAST "inpBlk", "%u", lc->inpX.lag_block) >= 0
        && xmlTextWriterEndElement(writer) >= 0

        && xmlTextWriterStartElement(writer, BAD_CAST "yInp") >= 0
        && xmlTextWriterWriteAttribute(
            writer, BAD_CAST "strm",
            BAD_CAST ((lc->inpY.stream == 'A') ? "A" : "B")) >= 0
        && xmlTextWriterWriteFormatAttribute(
            writer, BAD_CAST "inpBlk", "%u", lc->inpY.lag_block) >= 0
        && xmlTextWriterEndElement(writer) >= 0

        && xmlTextWriterEndElement(writer) >= 0) {
        result = 0;
    }
    return result;
}

static gint
write_output_config(xmlTextWriterPtr writer,
                    const wcbe_output_config_t *output)
{
    gint result;

    result = xmlTextWriterStartElement(writer, BAD_CAST "outputConfig");
    guint i;
    if (output->subarrays != NULL) {
        for (i = 0; result >= 0 && i < output->subarrays->len; ++i) {
            result = write_subarray(
                writer,
                &g_array_index(output->subarrays, wcbe_subarray_config_t, i));
        }
    }
    if (result >= 0) {
        result = xmlTextWriterEndElement(writer);
    }
    return result;
}

static gint
write_subarray(xmlTextWriterPtr writer, const wcbe_subarray_config_t *sa)
{
    gint result = -1;

    if (xmlTextWriterStartElement(writer, BAD_CAST "subarray") >= 0) {
        GString *stations = g_string_new("");
        guint i;
        for (i = 0; i < sa->stations->len; ++i) {
            const gchar *fmt = ((i == 0) ? "%u" : " %u");
            g_string_append_printf(stations, fmt,
                                   g_array_index(sa->stations, guint, i));
        }
        result = xmlTextWriterWriteAttribute(
            writer, BAD_CAST "stns", BAD_CAST stations->str);
        g_string_free(stations, TRUE);
        if (result >= 0
            && xmlTextWriterWriteFormatAttribute(
                writer, BAD_CAST "nAPC", "%u", sa->num_apc) >= 0
            && xmlTextWriterWriteFormatAttribute(
                writer, BAD_CAST "intn", "%u", sa->integration) >= 0) {
            for (i = 0; result >= 0 && i < sa->basebands->len; ++i) {
                result = write_baseband(
                    writer,
                    &g_array_index(sa->basebands, wcbe_baseband_config_t, i));
            }
            result = xmlTextWriterEndElement(writer);
        }
    }
    return result;
}

static gint
write_baseband(xmlTextWriterPtr writer, const wcbe_baseband_config_t *bb)
{
    gint result = -1;

    if (bb->num_polarizations == 1) {
        if (xmlTextWriterStartElement(writer, BAD_CAST "bb1") >= 0) {
            if (bb->baseband_ids[0] != -1) {
                if (bb->baseband_ids[1] != -1) {
                    if (xmlTextWriterWriteFormatAttribute(
                            writer, BAD_CAST "bbs", "%u %u",
                            bb->baseband_ids[0], bb->baseband_ids[1]) >= 0) {
                        result = 0;
                    }
                }
                else {
                    if (xmlTextWriterWriteFormatAttribute(
                            writer, BAD_CAST "bbs", "%u",
                            bb->baseband_ids[0]) >= 0) {
                        result = 0;
                    }
                }
            }
        }
    }
    else {
        if (xmlTextWriterStartElement(writer, BAD_CAST "bb2") >= 0
            && xmlTextWriterWriteFormatAttribute(
                writer, BAD_CAST "bbL", "%u", bb->baseband_ids[0]) >= 0
            && xmlTextWriterWriteFormatAttribute(
                writer, BAD_CAST "bbR", "%u", bb->baseband_ids[1]) >= 0) {
            result = 0;
        }
    }
    guint i;
    for (i = 0; result >= 0 && i < bb->spectral_windows->len; ++i) {
        result = write_sw(
            writer,
            &g_array_index(bb->spectral_windows,
                           wcbe_spectral_window_config_t, i));
    }
    if (result >= 0) {
        result = xmlTextWriterEndElement(writer);
    }
    return result;
}

static gint
write_sw(xmlTextWriterPtr writer, const wcbe_spectral_window_config_t *sw)
{
    gint result = -1;

    static const gchar *sdpp[SDMHDR_APP_NUM_VALS] = {
        [SDMHDR_APP_RR] = "RR",
        [SDMHDR_APP_LL] = "LL",
        [SDMHDR_APP_RR_LL] = "RR LL",
        [SDMHDR_APP_RR_RL_LL] = "RR RL LL"
    };
    static const gchar *xpp[SDMHDR_XPP_NUM_VALS] = {
        [SDMHDR_XPP_RR] = "RR",
        [SDMHDR_XPP_LL] = "LL",
        [SDMHDR_XPP_RR_LL] = "RR LL",
        [SDMHDR_XPP_RR_RL_LR_LL] = "RR RL LR LL"
    };

    if (xmlTextWriterStartElement(writer, BAD_CAST "sw") >= 0
        && xmlTextWriterWriteFormatAttribute(
            writer, BAD_CAST "nSb", "%u", sw->num_subbands) >= 0
        && xmlTextWriterWriteFormatAttribute(
            writer, BAD_CAST "nBn", "%u", sw->num_bins) >= 0
        && xmlTextWriterWriteFormatAttribute(
            writer, BAD_CAST "nCh", "%u", sw->num_channels) >= 0
        && sw->sdpp != SDMHDR_APP_INVALID
        && (sw->sdpp == SDMHDR_APP_NONE
            || xmlTextWriterWriteAttribute(
                writer, BAD_CAST "sdpp", BAD_CAST sdpp[sw->sdpp]) >= 0)
        && sw->xpp != SDMHDR_XPP_INVALID
        && (sw->xpp == SDMHDR_XPP_NONE
            || xmlTextWriterWriteAttribute(
                writer, BAD_CAST "xpp", BAD_CAST xpp[sw->xpp]) >= 0)
        && xmlTextWriterEndElement(writer) >= 0) {
        result = 0;
    }
    return result;
}

static gint
write_pipeline_config(xmlTextWriterPtr writer,
                      const wcbe_pipelines_config_t *pconf)
{
    gint result = xmlTextWriterStartElement(writer, BAD_CAST "pipelineConfig");
    guint i;
    if (pconf->pipelines != NULL) {
        for (i = 0; result >= 0 && i < pconf->pipelines->len; ++i) {
            result = write_pln(
                writer,
                &g_array_index(pconf->pipelines, wcbe_pipeline_config_t, i));
        }
    }
    if (result >= 0) {
        result = xmlTextWriterEndElement(writer);
    }
    return result;
}

static gint
write_pln(xmlTextWriterPtr writer, const wcbe_pipeline_config_t *pln)
{
    gint result = -1;

    if (xmlTextWriterStartElement(writer, BAD_CAST "pln") >= 0
        && xmlTextWriterWriteAttribute(
            writer, BAD_CAST "id", BAD_CAST pln->id) >= 0
        && xmlTextWriterWriteFormatAttribute(
            writer, BAD_CAST "sa", "%u", pln->subarray_index) >= 0
        && xmlTextWriterStartElement(writer, BAD_CAST "spec") >= 0
        && xmlTextWriterWriteElement(writer, BAD_CAST "gstSpec",
                                     BAD_CAST pln->spec.gst_spec) >= 0
        && xmlTextWriterEndElement(writer) >= 0) {
        result = 0;
        guint i;
        for (i = 0; result >= 0 && i < pln->swmaps->len; ++i) {
            result = write_swmap(
                writer,
                &g_array_index(pln->swmaps, wcbe_swmap_config_t, i));
        }
        if (result == 0) {
            result = xmlTextWriterEndElement(writer);
        }
    }
    return result;
}

static gint
write_swmap(xmlTextWriterPtr writer, const wcbe_swmap_config_t *swmap)
{
    gint result = -1;

    if (xmlTextWriterStartElement(writer, BAD_CAST "swmap") >= 0
        && xmlTextWriterWriteFormatAttribute(
            writer, BAD_CAST "bb", "%u", swmap->baseband_index) >= 0
        && xmlTextWriterWriteFormatAttribute(
            writer, BAD_CAST "sw", "%u", swmap->spectral_window_index) >= 0) {
        GString *subbands = g_string_new("");
        guint i;
        for (i = 0; i < swmap->subband_indices->len; ++i) {
            const gchar *fmt = ((i == 0) ? "%u" : " %u");
            g_string_append_printf(
                subbands, fmt,
                g_array_index(swmap->subband_indices, guint, i));
        }
        result =
            xmlTextWriterWriteAttribute(writer, BAD_CAST "sbs",
                                        BAD_CAST subbands->str);
        g_string_free(subbands, TRUE);
        if (result >= 0) {
            result = xmlTextWriterEndElement(writer);
        }
    }
    return result;
}
