/*
 * wcbe_lag_frame.c
 *
 * Lag frame utility functions.
 *
 */
#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <netinet/in.h>
#include <glib.h>

#include <wcbe_lag_frame.h>
#include <wcbe.h>

#if SIZEOF_UINT32_T != SIZEOF_LF_DATA_FRAME_WORD_T
# error "lag_frame module implementation doesn't support data frame word size"
#endif

static int decode_nlags(uint8_t nlags) ATTRIBUTE((__const__));
static int to_cc_lag_frame(lf_lag_frame_t *lag_frame, guint8 num_lags)
    ATTRIBUTE((__nonnull__));
static int to_lta_lag_frame(lf_lag_frame_t *lag_frame, guint8 num_lags)
    ATTRIBUTE((__nonnull__));

/* lf_to_cc_lag_frame
 *
 * Convert a correlator chip data frame to a lf_lag_frame_t. Data
 * frame is expected to be in network byte order, result is in host
 * byte order. Conversion is done in place!
 *
 * *Arguments*
 *
 * lag_frame : pointer to lf_cc_lag_frame_t structure, with a 'lags'
 * array of at least 'max_lags' elements
 *
 * *Return values*
 *
 * LF_SUCCESS : 'data_frame' successfully converted to 'lag_frame',
 * with verified checksum
 *
 * LF_E_CHECKSUM : checksum of 'data_frame' could not be verified
 *
 * LF_E_START_SYNC : start sync word not present
 *
 * LF_E_END_SYNC : end sync word not present
 *
 * LF_E_FTYPE : CMMD field value not equal to LF_CC_DATA_FRAME_FTYPE
 *
 * *Side effects consequent to the following return values*
 *
 * LF_SUCCESS : 'lag_frame' is filled with values converted from
 * 'data_frame'
 *
 * LF_E_CHECKSUM : 'lag_frame' is filled with values converted from
 * 'data_frame'
 *
 * LF_E_START_SYNC, LF_E_END_SYNC, LF_E_FTYPE : no side-effects
 */
int
lf_to_cc_lag_frame(lf_lag_frame_t *lag_frame)
{
    int result;
    uint8_t num_lags;

    /* check start sync word */
    if (G_UNLIKELY(lag_frame->words[0] != g_htonl(LF_START_SYNC_WORD))) {
        result = LF_E_START_SYNC;
        goto done;
    }

    /* get number of lags */
    num_lags = LF_BITFIELD_EXTRACT(g_ntohl(lag_frame->words[1]), 10, 9);

    /* check end sync word */
    if (G_UNLIKELY(lag_frame->words[LF_DATA_FRAME_NUM_WORDS(num_lags) - 2]
                   != g_htonl(LF_END_SYNC_WORD))) {
        result = LF_E_END_SYNC;
        goto done;
    }

    /* check station ids */
    lf_data_frame_word_t word = g_ntohl(lag_frame->words[2]);
    if (G_UNLIKELY(LF_BITFIELD_EXTRACT(word, 8, 0) == 0
                   || LF_BITFIELD_EXTRACT(word, 8, 16) == 0)) {
        result = LF_E_STATION_ZERO;
        goto done;
    }

    result = to_cc_lag_frame(lag_frame, num_lags);

 done:
    return result;
}

/* lf_to_lta_lag_frame
 *
 * Convert an LTA data frame to a lf_lag_frame_t. Data frame is
 * expected to be in network byte order, result is in host byte order.
 * Conversion is done in place!
 *
 * *Arguments*
 *
 * lag_frame : pointer to lf_lta_lag_frame_t structure, with a 'lags'
 * array of at least 'max_lags' elements
 *
 * *Return values*
 *
 * LF_SUCCESS : 'data_frame' successfully converted to 'lag_frame',
 * with verified checksum
 *
 * LF_E_CHECKSUM : checksum of 'data_frame' could not be verified
 *
 * LF_E_START_SYNC : start sync word not present
 *
 * LF_E_END_SYNC : end sync word not present
 *
 * LF_E_FTYPE : FType field value not equal to LF_LTA_DATA_FRAME_FTYPE
 *
 * LF_E_NLAGS : 'nlags' field value is invalid
 *
 * *Side effects consequent to the following return values*
 *
 * LF_SUCCESS : 'lag_frame' is filled with values converted from
 * 'data_frame'
 *
 * LF_E_CHECKSUM : 'lag_frame' is filled with values converted from
 * 'data_frame'
 *
 * LF_E_START_SYNC, LF_E_END_SYNC, LF_E_FTYPE, LF_E_NLAGS : no
 * side-effects
 */
int
lf_to_lta_lag_frame(lf_lag_frame_t *lag_frame)
{
    int result;
    int nlags;
    uint8_t num_lags;

    /* check start sync word */
    if (G_UNLIKELY(lag_frame->words[0] != g_htonl(LF_START_SYNC_WORD))) {
        result = LF_E_START_SYNC;
        goto done;
    }

    /* get nlags and convert to number of lags */
    lf_data_frame_word_t word = g_ntohl(lag_frame->words[1]);
    nlags = decode_nlags(LF_BITFIELD_EXTRACT(word, 3, 13));
    if (G_UNLIKELY(nlags < 0)) {
        result = LF_E_NLAGS;
        goto done;
    }
    num_lags = nlags;

    /* check end sync word */
    if (G_UNLIKELY(lag_frame->words[LF_DATA_FRAME_NUM_WORDS(num_lags) - 2]
                   != g_htonl(LF_END_SYNC_WORD))) {
        result = LF_E_END_SYNC;
        goto done;
    }

    /* check station ids */
    word = g_ntohl(lag_frame->words[2]);
    if (G_UNLIKELY(LF_BITFIELD_EXTRACT(word, 8, 0) == 0
                   || LF_BITFIELD_EXTRACT(word, 8, 16) == 0)) {
        result = LF_E_STATION_ZERO;
        goto done;
    }

    result = to_lta_lag_frame(lag_frame, num_lags);

 done:
    return result;
}

/* decode_num_lags
 *
 * Decode the 'nlags' field of a data frame into a number of lags.
 *
 * *Arguments*
 *
 * nlags - value of nlags field
 *
 * *Return values*
 *
 * -1 : an error, invalid 'nlags' value
 *
 * any value except -1 : the decoded integer value given by the
 * 'nlags' value of a data frame (_not_ a correlator chip data frame)
 */
static int
decode_nlags(uint8_t nlags)
{
    int result;

    switch (nlags) {
    case 0:
        result = 128;
        break;
    case 1:
        result = 64;
        break;
    default:
        result = -1;
        break;
    }
    return result;
}

/* to_cc_lag_frame
 *
 * Convert a correlator chip data frame to a lf_cc_lag_frame_t. Data
 * frame is expected to be in network byte-order. Sync words in the
 * data frame are not checked.
 *
 * *Arguments*
 *
 * lag_frame : pointer to lf_cc_lag_frame_t structure
 *
 * *Return values*
 *
 * LF_SUCCESS : 'data_frame' successfully converted to 'lag_frame',
 * with verified checksum
 *
 * LF_E_CHECKSUM : checksum of 'data_frame' could not be verified
 *
 * LF_E_FTYPE : CMMD field value not equal to LF_CC_DATA_FRAME_FTYPE
 *
 * *Side effects consequent to the following return values*
 *
 * LF_SUCCESS : 'lag_frame' is filled with values converted from
 * 'data_frame'
 *
 * LF_E_CHECKSUM : 'lag_frame' is filled with values converted from
 * 'data_frame'
 *
 * LF_E_FTYPE : no side-effects
 */
static int
to_cc_lag_frame(lf_lag_frame_t *lag_frame, guint8 num_lags)
{
    int result;

    /* check CMMD value */
    if (G_UNLIKELY(LF_BITFIELD_EXTRACT(g_ntohl(lag_frame->words[1]), 3, 0)
                   != LF_CC_DATA_FRAME_FTYPE)) {
        result = LF_E_FTYPE;
        goto done;
    }

    lag_frame->key[3] = 0;

    /*
     * Convert fields to host byte order, and compute checksum.
     */
    lf_data_frame_word_t checksum = 0;
    lf_data_frame_word_t *df = &(lag_frame->words[0]);
    guint i;
    for (i = LF_DATA_FRAME_NUM_WORDS(num_lags) - 1; i > 0; --i) {
        *df = g_ntohl(*df);
        checksum += *df++;
    }
    if (checksum != g_ntohl(*df)) {
        result = LF_E_CHECKSUM;
        goto done;
    }

    /* word 1 */
    lag_frame->key[0] = lag_frame->words[1];

    /* word 2 */
    lag_frame->key[1] = lag_frame->words[2];

    /* word 3 */
    lag_frame->key[2] = lag_frame->words[3];

    /* word 4 */
    lag_frame->timestamp.seconds = lag_frame->words[4];

    /* word 5 */
    lf_data_frame_word_t word = lag_frame->words[5];
    lag_frame->timestamp.clock_cycles = LF_BITFIELD_EXTRACT(word, 29, 0);
    lag_frame->timestamp.epoch = LF_BITFIELD_EXTRACT(word, 3, 29);
    if (lag_frame->timestamp.clock_cycles >= WIDAR_CLOCK_RATE) {
        lag_frame->timestamp.seconds++;
        lag_frame->timestamp.clock_cycles -= WIDAR_CLOCK_RATE;
    }

    lf_lag_t *l0 G_GNUC_UNUSED = (lf_lag_t *)&LF_LAG_FRAME_LAG0(lag_frame);
    // Just to make certain that the format of GCC complex types
    // agrees with what I'm expecting...
    g_assert(((void *)&LF_LAG_REAL(*l0) == (void *)l0)
             && ((void *)&LF_LAG_IMAG(*l0)
                 == ((void *)l0 + sizeof(lf_data_frame_word_t))));

    result = LF_SUCCESS;

 done:
    return result;
}

/* to_lta_lag_frame
 *
 * Convert an LTA data frame to a lf_lta_lag_frame_t. Data frame is
 * expected to be in network byte-order. Sync words in the data frame
 * are not checked.
 *
 * *Arguments*
 *
 * lag_frame : pointer to lf_lta_lag_frame_t structure
 *
 * *Return values*
 *
 * LF_SUCCESS : 'data_frame' successfully converted to 'lag_frame',
 * with verified checksum
 *
 * LF_E_CHECKSUM : checksum of 'data_frame' could not be verified
 *
 * LF_E_FTYPE : FType field value not equal to LF_LTA_DATA_FRAME_FTYPE
 *
 * *Side effects consequent to the following return values*
 *
 * LF_SUCCESS : 'lag_frame' is filled with values converted from
 * 'data_frame'
 *
 * LF_E_CHECKSUM : 'lag_frame' is filled with values converted from
 * 'data_frame'
 *
 * LF_E_FTYPE, LF_E_NLAGS: no side-effects
 */
static int
to_lta_lag_frame(lf_lag_frame_t *lag_frame, guint8 num_lags)
{
    int result;

    /* check FType value */
    if (G_UNLIKELY(LF_BITFIELD_EXTRACT(g_ntohl(lag_frame->words[1]), 3, 0)
                   != LF_LTA_DATA_FRAME_FTYPE)) {
        result = LF_E_FTYPE;
        goto done;
    }

    /*
     * Convert fields to host byte order, and compute checksum.
     */
    lf_data_frame_word_t checksum = 0;
    lf_data_frame_word_t *df = &(lag_frame->words[0]);
    guint i;
    for (i = LF_DATA_FRAME_NUM_WORDS(num_lags) - 1; i > 0; --i) {
        *df = g_ntohl(*df);
        checksum += *df++;
    }
    if (checksum != g_ntohl(*df)) {
        result = LF_E_CHECKSUM;
        goto done;
    }

    /* word 1 */
    lag_frame->key[0] = lag_frame->words[1];

    /* word 2 */
    lag_frame->key[1] = lag_frame->words[2];

    /* word 3 */
    lag_frame->key[2] = lag_frame->words[3];

    /* word 4 */
    lag_frame->timestamp.seconds = lag_frame->words[4];

    /* word 5 */
    lf_data_frame_word_t word = lag_frame->words[5];
    lag_frame->timestamp.clock_cycles = LF_BITFIELD_EXTRACT(word, 29, 0);
    lag_frame->timestamp.epoch = LF_BITFIELD_EXTRACT(word, 3, 29);
    if (lag_frame->timestamp.clock_cycles >= WIDAR_CLOCK_RATE) {
        lag_frame->timestamp.seconds++;
        lag_frame->timestamp.clock_cycles -= WIDAR_CLOCK_RATE;
    }

    /* word 8 */
    lag_frame->key[3] = lag_frame->words[8];

    lf_lag_t *l0 G_GNUC_UNUSED = (lf_lag_t *)&LF_LAG_FRAME_LAG0(lag_frame);
    // Just to make certain that the format of GCC complex types
    // agrees with what I'm expecting...
    g_assert(((void *)&LF_LAG_REAL(*l0) == (void *)l0)
             && ((void *)&LF_LAG_IMAG(*l0)
                 == ((void *)l0 + sizeof(lf_data_frame_word_t))));

    result = LF_SUCCESS;

 done:
    return result;
}

/* lf_strerror
 *
 * Returns string corresponding to given error code from this module.
 * The string should only be used until the next call to this function
 * (which therefore is not thread-safe).
 */
const char *
lf_strerror(int errnum)
{
    const char *result;
    static const char success[] = "success";
    static const char too_small[] = "buffer too small";
    static const char checksum[] = "checksum verification failed";
    static const char start_sync[] = "invalid lag frame start sync word";
    static const char end_sync[] = "invalid lag frame end sync word";
    static const char ftype[] = "invalid lag frame FTYPE value";
    static const char nlags[] = "invalid lag frame NLAGS value";
    static const char stn_zero[] = "invalid station id value (0)";
    static const char unknown[] = "unknown error (%d)";
    static char *str = NULL;

    switch (errnum) {
    case LF_SUCCESS:
        result = success;
        break;
    case LF_E_TOO_SMALL:
        result = too_small;
        break;
    case LF_E_CHECKSUM:
        result = checksum;
        break;
    case LF_E_START_SYNC:
        result = start_sync;
        break;
    case LF_E_END_SYNC:
        result = end_sync;
        break;
    case LF_E_FTYPE:
        result = ftype;
        break;
    case LF_E_NLAGS:
        result = nlags;
        break;
    case LF_E_STATION_ZERO:
        result = stn_zero;
        break;
    default:
        if (str != NULL) {
            g_free(str);
        }
        str = g_strdup_printf(unknown, errnum);
        result = str;
        break;
    }
    return result;
}
