#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <wcbe_util.h>
#include <wcbe.h>
#include <wcbe_lagset.h>

#include <stdarg.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

static gsize wcbe_vect_alloc_size_va(guint num_blocks, va_list ap);
static gsize wcbe_vect_build_size_va(wcbe_vect_block_t *block, va_list ap);
static void wcbe_vect_init_from_pars_va(wcbe_vect_t *buff, guint num_blocks,
                                        va_list ap);
static void wcbe_vect_init_from_blocks_va(
    wcbe_vect_t *buff, wcbe_vect_block_t *block, va_list ap);

#ifdef __va_copy
# define VA_COPY(dst, src) __va_copy(dst, src)
#else
# define VA_COPY(dst, src) dst = src
#endif
gboolean
gst_structure_get_pointer(const GstStructure *structure,
                          const gchar *fieldname, gpointer *value)
{
    g_return_val_if_fail(structure != NULL, FALSE);
    g_return_val_if_fail(fieldname != NULL, FALSE);
    g_return_val_if_fail(value != NULL, FALSE);

    gboolean result = FALSE;

    *value = NULL;
    if (gst_structure_has_field_typed(structure, fieldname, G_TYPE_POINTER)) {
        const GValue *gvalue = gst_structure_get_value(structure, fieldname);
        *value = g_value_get_pointer(gvalue);
        result = TRUE;
    }
    return result;
}

gboolean
gst_structure_get_uint64(const GstStructure *structure,
                         const gchar *fieldname, guint64 *value)
{
    g_return_val_if_fail(structure != NULL, FALSE);
    g_return_val_if_fail(fieldname != NULL, FALSE);
    g_return_val_if_fail(value != NULL, FALSE);

    gboolean result = FALSE;

    *value = 0;
    if (gst_structure_has_field_typed(structure, fieldname, G_TYPE_UINT64)) {
        const GValue *gvalue = gst_structure_get_value(structure, fieldname);
        *value = g_value_get_uint64(gvalue);
        result = TRUE;
    }
    return result;
}

gboolean
gst_structure_get_gtype(const GstStructure *structure,
                        const gchar *fieldname, GType *value)
{
    g_return_val_if_fail(structure != NULL, FALSE);
    g_return_val_if_fail(fieldname != NULL, FALSE);
    g_return_val_if_fail(value != NULL, FALSE);

    gboolean result = FALSE;

    *value = 0;
    if (gst_structure_has_field_typed(structure, fieldname, G_TYPE_GTYPE)) {
        const GValue *gvalue = gst_structure_get_value(structure, fieldname);
        *value = g_value_get_gtype(gvalue);
        result = TRUE;
    }
    return result;
}

gboolean
object_has_signal(GObject *object, const gchar *signal)
{
    gboolean result = FALSE;
    guint n_ids;
    guint *ids = g_signal_list_ids(G_OBJECT_TYPE(object), &n_ids);
    while (!result && n_ids-- > 0) {
        result = (strcmp(g_signal_name(ids[n_ids]), signal) == 0);
    }
    g_free(ids);
    return result;
}

/**
 * aligner_t:
 * @base: one byte filler
 * @offset: aligned member
 *
 * Helper structure for determining GCC's "maximum useful alignment"
 * on the target machine.
 */
typedef struct aligner_t {
    guchar base;
    guchar offset ATTRIBUTE((aligned));
} aligner_t;
#define MAX_ALIGNMENT (offsetof(aligner_t, offset))
#define SIZEOF_WCBE_VECT_BLOCK_T(len, eltsz) ({\
    gsize _datasz = len * eltsz; \
    (MAX_ALIGNMENT * ((_datasz / MAX_ALIGNMENT) + \
                      ((_datasz % MAX_ALIGNMENT) ? 1: 0)) \
     + offsetof(wcbe_vect_block_t, data)); \
})

/**
 * wcbe_vect_block_index_block:
 * @vect: array of #wcbe_vect_block_t
 * @i: block index
 *
 * Get pointer to given block of an array of #wcbe_vect_block_t elements.
 */
wcbe_vect_block_t *
wcbe_vect_index_block(wcbe_vect_t *vect, guint i)
{
    g_return_val_if_fail(vect != NULL, NULL);
    g_return_val_if_fail(i < vect->num_blocks, NULL);

    wcbe_vect_block_t *result = &(vect->blocks);
    while (i-- > 0) {
        result = (void *)result
            + SIZEOF_WCBE_VECT_BLOCK_T(result->length, result->element_size);
    }
    return result;
}

/**
 * wcbe_vect_alloc:
 * @num_blocks: number of #wcbe_vect_block_t blocks to allocate
 * @Varargs: sequence of "length", and "element size" values for the
 *           blocks, one pair for each block, each value has type #guint.
 *
 * Allocate a vector of #wcbe_vect_block_t elements.
 *
 * @returns: properly aligned vector of #wcbe_vect_block_t elements of
 *           the given parameters; %NULL, if allocation failed. Caller
 *           should free returned vector using free().
 */
wcbe_vect_t *
wcbe_vect_alloc(guint num_blocks, ...)
{
    wcbe_vect_t *result = NULL;
    va_list ap;

    // calculate buffer size for vector
    va_start(ap, num_blocks);
    gsize total_size = wcbe_vect_alloc_size_va(num_blocks, ap);
    va_end(ap);

    // allocate memory for vector, initialize vector
    // header, and initialize block headers
    void *mem;
    gint rc = posix_memalign(&mem, MAX_ALIGNMENT, total_size);
    if (rc == 0) {
        result = mem;
        va_start(ap, num_blocks);
        wcbe_vect_init_from_pars_va(result, num_blocks, ap);
        va_end(ap);
    }
    else {
        g_critical("Failed to allocate memory of wcbe_vect_t: %s",
                   g_strerror(rc));
    }

    return result;
}

/**
 * wcbe_vect_build:
 * @block: first block
 * @Varargs: %NULL-terminated list of additional blocks
 *
 * Build a #wcbe_vect_t vector from a list of blocks. Blocks are
 * copied to the newly-allocated vector.
 *
 * @returns: vector built of given blocks, in order. Caller should
 * free the result using free().
 */
wcbe_vect_t *
wcbe_vect_build(wcbe_vect_block_t *block, ...)
{
    wcbe_vect_t *result = NULL;
    va_list ap;

    // count number of blocks in arg list
    guint num_blocks = 0;
    if (block != NULL) {
        num_blocks++;
        va_start(ap, block);
        wcbe_vect_block_t *vb = va_arg(ap, wcbe_vect_block_t *);
        while (vb != NULL) {
            num_blocks++;
            vb = va_arg(ap, wcbe_vect_block_t *);
        }
        va_end(ap);
    }
    // calculate buffer size for vector
    va_start(ap, block);
    gsize total_size = wcbe_vect_build_size_va(block, ap);
    va_end(ap);

    // second pass: allocate memory for vector, initialize vector
    // header, and copy blocks into new vector (in order)
    void *mem;
    gint rc = posix_memalign(&mem, MAX_ALIGNMENT, total_size);
    if (rc == 0) {
        result = mem;
        va_start(ap, block);
        wcbe_vect_init_from_blocks_va(result, block, ap);
        va_end(ap);
    }
    else {
        g_critical("Failed to allocate memory of wcbe_vect_t: %s",
                   g_strerror(rc));
    }

    return result;
}

/**
 * wcbe_vect_alloc_size:
 * @num_blocks: number of vector blocks
 * @Varargs: sequence of "length", and "element size" values for the
 *           blocks, one pair for each block, each value has type #guint.
 *
 * Compute the memory buffer size required for a #wcbe_vect_t vector
 * given the parameters for each block in the vector.
 *
 * @returns: number of bytes required for a vector with blocks having
 *           the given parameters.
 */
gsize
wcbe_vect_alloc_size(guint num_blocks, ...)
{
    va_list ap;
    va_start(ap, num_blocks);
    gsize result = wcbe_vect_alloc_size_va(num_blocks, ap);
    va_end(ap);
    return result;
}

/**
 * wcbe_vect_build_size:
 * @block: first vector block
 * @Varargs: %NULL-terminated list of additional blocks
 *
 * Compute the memory buffer size required for a #wcbe_vect_t vector
 * comprised of the given blocks.
 *
 * @returns: number of bytes required for a vector comprised of copies
 *           of the given blocks.
 */
gsize
wcbe_vect_build_size(wcbe_vect_block_t *block, ...)
{
    va_list ap;
    va_start(ap, block);
    gsize result = wcbe_vect_build_size_va(block, ap);
    va_end(ap);
    return result;
}

void
wcbe_vect_init_from_pars(wcbe_vect_t *buff, guint num_blocks, ...)
{
    g_return_if_fail(buff != NULL);
    g_return_if_fail(wcbe_vect_check_alignment(buff, FALSE));

    va_list ap;
    va_start(ap, num_blocks);
    wcbe_vect_init_from_pars_va(buff, num_blocks, ap);
    va_end(ap);
}

void
wcbe_vect_init_from_blocks(wcbe_vect_t *buff, wcbe_vect_block_t *block, ...)
{
    g_return_if_fail(buff != NULL);
    g_return_if_fail(wcbe_vect_check_alignment(buff, FALSE));

    va_list ap;
    va_start(ap, block);
    wcbe_vect_init_from_blocks_va(buff, block, ap);
    va_end(ap);
}

/**
 * wcbe_vect_check_alignment:
 * @buff: pointer to buffer allocated for #wcbe_vect_t instance
 * @log_msg: %TRUE, if a log message should be generated if the buffer
 *           fails the alignment test
 *
 * Check whether a buffer has proper alignment for a #wcbe_vect_t
 * instance.
 *
 * @returns: %TRUE, if alignment is correct; %FALSE, otherwise
 */
gboolean
wcbe_vect_check_alignment(const void *buff, gboolean log_msg)
{
    gboolean result;
    if (((gsize)buff % MAX_ALIGNMENT) == 0) {
        result = TRUE;
    }
    else {
        result = FALSE;
        if (log_msg) {
            g_critical("Buffer is mis-aligned for wcbe_vect_t instance.");
        }
    }
    return result;
}

/**
 * wcbe_vect_copy:
 * @vect: vector
 *
 * Copy a #wcbe_vect_t instance.
 *
 * @returns: copy of @vect. Caller should free the result using g_free().
 */
wcbe_vect_t *
wcbe_vect_copy(wcbe_vect_t *vect)
{
    g_return_val_if_fail(vect != NULL, NULL);

    wcbe_vect_t *result = NULL;
    void *mem;
    gint rc = posix_memalign(&mem, MAX_ALIGNMENT, vect->total_size);
    if (rc == 0) {
        result = mem;
        memcpy(result, vect, vect->total_size);
    }
    else {
        g_critical("Failed to allocate memory for wcbe_vect_t (size %zu): %s",
                   vect->total_size, g_strerror(rc));
    }
    return result;
}

GstFlowReturn
wcbe_vect_pad_buffer_alloc(GstPad *pad, guint64 offset,
                           guint size, GstCaps *caps, GstBuffer **buf)
{
    GstFlowReturn result = GST_FLOW_ERROR;

    if (GST_CAPS_IS_SIMPLE(caps)) {
        GstStructure *s = gst_caps_get_structure(caps, 0);
        if (gst_structure_has_name(s, WCBE_VECTOR_MEDIA_TYPE)
            || gst_structure_has_name(s, WCBE_LAGSET_MEDIA_TYPE)) {
            void *mem;
            gint rc = posix_memalign(&mem, MAX_ALIGNMENT, size);
            if (rc == 0) {
                *buf = gst_buffer_new();
                GST_BUFFER_SIZE(*buf) = size;
                GST_BUFFER_OFFSET(*buf) = offset;
                GST_BUFFER_MALLOCDATA(*buf) = mem;
                GST_BUFFER_DATA(*buf) = GST_BUFFER_MALLOCDATA(*buf);
                gst_buffer_set_caps(*buf, caps);
                result = GST_FLOW_OK;
            }
            else {
                g_critical("Failed to allocate memory for wcbe_vect_t "
                           "(size %u): %s", size, g_strerror(rc));
            }
        }
        else {
            g_debug("wcbe_vect_pad_buffer_alloc: invalid requested buffer "
                    "caps '%s'", gst_structure_get_name(s));
        }
    }
    else {
        g_debug("wcbe_vect_pad_buffer_alloc: caps not simple");
    }
    return result;
}

/**
 * wcbe_vect_clear_all:
 * @vect: vector
 *
 * Clear all data in vector; metadata is retained.
 */
void
wcbe_vect_clear_all(wcbe_vect_t *vect)
{
    g_return_if_fail(vect != NULL);

    guint i;
    for (i = 0; i < vect->num_blocks; ++i) {
        wcbe_vect_block_t *vb = wcbe_vect_index_block(vect, i);
        wcbe_vect_block_clear(vb);
    }
}

/**
 * wcbe_vect_block_clear:
 * @block: vector block
 *
 * Clear all data in vector block; metadata is retained.
 */
void
wcbe_vect_block_clear(wcbe_vect_block_t *block)
{
    g_return_if_fail(block != NULL);
    memset(block->data, 0, block->length * block->element_size);
}

static gsize
wcbe_vect_alloc_size_va(guint num_blocks, va_list ap)
{
    gsize result = offsetof(wcbe_vect_t, blocks);
    if (num_blocks > 0) {
        guint i;
        for (i = 0; i < num_blocks; ++i) {
            guint length = va_arg(ap, guint);
            guint element_size = va_arg(ap, guint);
            result += SIZEOF_WCBE_VECT_BLOCK_T(length, element_size);
        }
    }
    return result;
}

static gsize
wcbe_vect_build_size_va(wcbe_vect_block_t *block, va_list ap)
{
    gsize result = offsetof(wcbe_vect_t, blocks);
    if (block != NULL) {
        result += SIZEOF_WCBE_VECT_BLOCK_T(block->length, block->element_size);
        wcbe_vect_block_t *vb = va_arg(ap, wcbe_vect_block_t *);
        while (vb != NULL) {
            result +=
                SIZEOF_WCBE_VECT_BLOCK_T(vb->length, vb->element_size);
            vb = va_arg(ap, wcbe_vect_block_t *);
        }
    }
    return result;
}

static void
wcbe_vect_init_from_pars_va(wcbe_vect_t *buff, guint num_blocks, va_list ap)
{
    g_return_if_fail(buff != NULL);
    g_return_if_fail(wcbe_vect_check_alignment(buff, FALSE));

    // calculate buffer size for vector
    va_list ap2;
    VA_COPY(ap2, ap);
    gsize total_size = wcbe_vect_alloc_size_va(num_blocks, ap2);
    va_end(ap2);

    // initialize vector header, and initialize block headers
    buff->num_blocks = num_blocks;
    buff->total_size = total_size;
    wcbe_vect_block_t *vb = &(buff->blocks);
    guint i;
    for (i = 0; i < num_blocks; ++i) {
        vb->length = va_arg(ap, guint);
        vb->element_size = va_arg(ap, guint);
        vb = (void *)vb
            + SIZEOF_WCBE_VECT_BLOCK_T(vb->length, vb->element_size);
    }
}

static void
wcbe_vect_init_from_blocks_va(wcbe_vect_t *buff, wcbe_vect_block_t *block,
                              va_list ap)
{
    g_return_if_fail(buff != NULL);
    g_return_if_fail(wcbe_vect_check_alignment(buff, FALSE));

    va_list ap2;
    VA_COPY(ap2, ap);
    // count number of blocks in arg list
    guint num_blocks = 0;
    if (block != NULL) {
        num_blocks++;
        wcbe_vect_block_t *vb = va_arg(ap2, wcbe_vect_block_t *);
        while (vb != NULL) {
            num_blocks++;
            vb = va_arg(ap2, wcbe_vect_block_t *);
        }
    }
    va_end(ap2);

    VA_COPY(ap2, ap);
    // calculate buffer size for vector
    gsize total_size = wcbe_vect_build_size_va(block, ap2);
    va_end(ap2);

    // initialize vector header, and copy blocks into new vector (in
    // order)
    buff->num_blocks = num_blocks;
    buff->total_size = total_size;
    if (block != NULL) {
        wcbe_vect_block_t *newblk = &(buff->blocks);
        void cpyblk(const wcbe_vect_block_t *src) {
            gsize blksz =
                SIZEOF_WCBE_VECT_BLOCK_T(src->length, src->element_size);
            memcpy(newblk, src, blksz);
            newblk = (void *)newblk + blksz;
        }
        cpyblk(block);
        wcbe_vect_block_t *vb = va_arg(ap, wcbe_vect_block_t *);
        while (vb != NULL) {
            cpyblk(vb);
            vb = va_arg(ap, wcbe_vect_block_t *);
        }
    }
}

gsize
wcbe_vect_alignment(void)
{
    return MAX_ALIGNMENT;
}
