#ifndef WCBE_UTIL_H_
#define WCBE_UTIL_H_

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib.h>
#include <gst/gst.h>

G_BEGIN_DECLS

#ifdef G_HAVE_ISO_VARARGS
#define g_info(...)    g_log (G_LOG_DOMAIN,       \
                              G_LOG_LEVEL_INFO,   \
                              __VA_ARGS__)
#elif defined(G_HAVE_GNUC_VARARGS)
#define g_info(format...)      g_log (G_LOG_DOMAIN,         \
                                      G_LOG_LEVEL_INFO,     \
                                      format)
#else   /* no varargs macros */
static void
g_info (const gchar *format, ...)
{
  va_list args;
  va_start (args, format);
  g_logv (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, format, args);
  va_end (args);
}
#endif  /* !__GNUC__ */

extern gboolean gst_structure_get_pointer(const GstStructure *structure,
                                          const gchar *fieldname,
                                          gpointer *value);
extern gboolean gst_structure_get_uint64(const GstStructure *structure,
                                         const gchar *fieldname,
                                         guint64 *value);
extern gboolean gst_structure_get_gtype(const GstStructure *structure,
                                         const gchar *fieldname,
                                         GType *value);
extern gboolean object_has_signal(GObject *object, const gchar *signal);

/**
 * _wcbe_vect_block_t:
 * @length: length of vector (actual number)
 * @element_size: size of vector elements (in bytes)
 *
 * A block of elements of a single data type in a vector of multiple
 * data types.
 *
 * @see_also: #wcbe_vect_t structure
 */
typedef struct wcbe_vect_block_t {
    /*< private >*/
    guint length;
    guint element_size;
    guint8 data[] ATTRIBUTE((aligned));
} wcbe_vect_block_t ATTRIBUTE((aligned));

/**
 * wcbe_vect_t:
 * @num_blocks: number of #wcbe_vect_block_t elements in vector
 *
 * A single, multi-datatype vector organized in blocks of single
 * datatypes. Within wcbe, we use a general data structure that
 * consists of a vector and any number of associated metadata vectors;
 * for example, lags (data) together with weights and timestamps
 * (metadata).
 *
 * For efficiency, and ease of use within GStreamer, this
 * data structure is implemented as a single, contiguous block of
 * memory. The organization is as follows:
 *
 * +===========================+
 * |      vector header        |
 * +===========================+ - - - -
 * |        header_0           |   ^
 * +---------------------------+   |
 * |       element_0_0         |   |
 * +---------------------------+   |
 * |       element_0_1         | Block 0
 * +---------------------------+   |
 * |         ...               |   |
 * +---------------------------+   |
 * |       element_0_N_0       |   v
 * +===========================+ - - - -
 * |        header_1           |   ^
 * +---------------------------+   |
 * |       element_1_0         |   |
 * +---------------------------+   |
 * |       element_1_1         | Block 1
 * +---------------------------+   |
 * |         ...               |   |
 * +---------------------------+   |
 * |       element_1_N_1       |   v
 * +===========================+ - - - -
 * |        ...                |
 * +===========================+ - - - -
 * |        header_M           |   ^
 * +---------------------------+   |
 * |       element_M_0         |   |
 * +---------------------------+   |
 * |       element_M_1         | Block M
 * +---------------------------+   |
 * |        ...                |   |
 * +---------------------------+   |
 * |       element_M_N_M       |   v
 * +===========================+ - - - -
 *
 * The size of every element in each block is fixed, although between
 * blocks there are no constraints on element size. The beginning of
 * each block is aligned on the maximum useful alignment boundary for
 * the target machine for which you compile the module. For
 * dynamically allocated memory, use the allocation functions in this
 * module to ensure the expected alignment.
 */
typedef struct {
    guint num_blocks;
    gsize total_size;
    wcbe_vect_block_t blocks ATTRIBUTE((aligned));
} wcbe_vect_t;

extern wcbe_vect_block_t *wcbe_vect_index_block(wcbe_vect_t *v, guint i);
#define wcbe_vect_block_index_element(blk, type, idx) \
    (((type *)((blk)->data))[idx])
#define wcbe_vect_index(v, blk, type, idx) \
    wcbe_vect_block_index_element(wcbe_vect_index_block(v, blk), type, idx)
extern wcbe_vect_t *wcbe_vect_alloc(guint num_blocks, ...);
extern wcbe_vect_t *wcbe_vect_build(wcbe_vect_block_t *block, ...);
extern gsize wcbe_vect_alloc_size(guint num_blocks, ...);
extern gsize wcbe_vect_build_size(wcbe_vect_block_t *block, ...);
extern void wcbe_vect_init_from_pars(wcbe_vect_t *buff, guint num_blocks, ...);
extern void wcbe_vect_init_from_blocks(wcbe_vect_t *buff,
                                       wcbe_vect_block_t *block, ...);
extern gboolean wcbe_vect_check_alignment(const void *buff, gboolean log_msg);
extern wcbe_vect_t *wcbe_vect_copy(wcbe_vect_t *vect);
extern GstFlowReturn wcbe_vect_pad_buffer_alloc(
    GstPad *pad, guint64 offset, guint size, GstCaps *caps, GstBuffer **buf);
extern void wcbe_vect_clear_all(wcbe_vect_t *vect);
extern void wcbe_vect_block_clear(wcbe_vect_block_t *block);
extern gsize wcbe_vect_alignment(void);

G_END_DECLS

#endif /* UTIL_H_ */
