#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <we_bdf.h>
#include <sdmhdr.h>
#include <wcbe_util.h>

#include <string.h>
#include <strings.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>

typedef struct subset_metadata {
    sdmhdr_subset_hdr_t *header;
    off_t header_offset;
    off_t flags_offset;
    off_t actual_times_offset;
    off_t actual_durations_offset;
    off_t weights_offset;
    off_t zero_lags_offset;
    off_t cross_data_offset;
    off_t auto_data_offset;
} subset_metadata_t;

static sdmhdr_main_hdr_t *get_main_hdr(const we_bdf_t *we_bdf);
static tree_t *get_bincomp(const we_bdf_t *we_bdf,
                           const subset_metadata_t *subset,
                           CobdfBincomp bincomp);
static gssize write_empty_bincomp(we_bdf_t *we_bdf, const gchar *bdy,
                                  const gchar *bincomp_ref, gsize bincomp_size,
                                  gint padding, off_t *binary_offset);
static GHashTable *read_mime_headers(FILE *file);
static gint next_part(FILE *file, const gchar *bdy);
static gint check_boundary(FILE *file, const gchar *bdy);
static gchar *get_boundary(GHashTable *headers);
static void slurp_blank_lines(FILE *file, gboolean strictly_empty, guint max);
static const subset_metadata_t *get_subset_metadata(we_bdf_t *we_bdf,
                                                    const gchar *project_path);
static const gchar *index_next_subset(we_bdf_t *we_bdf);
static guint num_cross_primitives(sdmhdr_xpp_t xpp, CobdfBincomp bincomp);
static guint num_auto_primitives(sdmhdr_app_t app, CobdfBincomp bincomp);
static sdmhdr_bb_t *maximal_baseband(const sdmhdr_bb_t **basebands,
                                     CobdfBincomp bincomp);
static void maximal_spectral_window(const sdmhdr_bb_t *baseband,
                                    sdmhdr_sw_t *max_sw, CobdfBincomp bincomp);
static guint node_multiplier(CobdfBincomp bincomp);
static gchar *spp_nodes(const gboolean *axes, const sdmhdr_main_hdr_t *hdr,
                        CobdfBincomp bincomp, CobdfAxisName bal_or_ant,
                        const sdmhdr_sw_t *spectral_window);
static gchar *apc_nodes(const gboolean *axes, const sdmhdr_main_hdr_t *hdr,
                        CobdfBincomp bincomp, CobdfAxisName bal_or_ant,
                        const sdmhdr_sw_t *spectral_window);
static gchar *bin_nodes(const gboolean *axes, const sdmhdr_main_hdr_t *hdr,
                        CobdfBincomp bincomp, CobdfAxisName bal_or_ant,
                        const sdmhdr_sw_t *spectral_window);
static gchar *sw_nodes(const gboolean *axes, const sdmhdr_main_hdr_t *hdr,
                       CobdfBincomp bincomp, CobdfAxisName bal_or_ant,
                       const sdmhdr_bb_t *baseband, guint num_sw);
static gchar *baseband_nodes(const gboolean *axes,
                             const sdmhdr_main_hdr_t *hdr,
                             CobdfBincomp bincomp, CobdfAxisName bal_or_ant,
                             guint num_basebands);
static gchar *balant_nodes(const gboolean *axes, const sdmhdr_main_hdr_t *hdr,
                           CobdfBincomp bincomp, CobdfAxisName bal_or_ant);
static subset_metadata_t *alloc_subset_metadata(
    const we_bdf_t *we_bdf, off_t offset, const gchar *bdy,
    sdmhdr_subset_hdr_t *subset);
static void destroy_subset_metadata(gpointer data);
static guint weights_nbits(sdmhdr_weights_t *weights);
static void crlfify(gchar **str);

static const gchar uid_prefix[] = "uid://";

#ifndef USE_ALMA_5_1_0_5_MIME_WORKAROUNDS
# define CRLF "\r\n"
# define CRLF_LEN 2
# define CONTENT_TYPE "Content-type"
# define CONTENT_LOCATION "Content-location"
# define CONTENT_DESCRIPTION "Content-description"
#else
# define CRLF "\n"
# define CRLF_LEN 1
# define CONTENT_TYPE "Content-Type"
# define CONTENT_LOCATION "Content-Location"
# define CONTENT_DESCRIPTION "Content-Description"
const gchar boundary1[] = "MIME_boundary-1";
const gchar boundary2[] = "MIME_boundary-2";
#endif

#define BDY_OK 0
#define BDY_EOF -1
#define BDY_NOMATCH -2
#define BDY_END -3

#define CHECKSUM_ALGORITHM G_CHECKSUM_SHA256

#define WRITE_INDENTED_HEADERS TRUE

we_bdf_t *
we_bdf_new(FILE *file, const sdmhdr_main_hdr_t *header)
{
    g_return_val_if_fail(header != NULL, NULL);
    g_return_val_if_fail(header->data_oid != NULL, NULL);
    g_return_val_if_fail(header->data_oid->href != NULL, NULL);

    // in case a file read was done before this, we set file position
    off_t origin_offset = ftello(file);
    if (origin_offset == -1) {
        g_critical("Failed to get position of BDF file: %s",
                   g_strerror(errno));
        return NULL;
    }
    gint rc = fseeko(file, origin_offset, SEEK_SET);
    if (rc != 0) {
        g_critical("Failed to set position of BDF file: %s",
                   g_strerror(errno));
        return NULL;
    }

    const gchar mime_hdr[] =
        "MIME-Version: 1.0"CRLF
        CONTENT_TYPE": multipart/mixed; boundary=%s"CRLF
        CONTENT_DESCRIPTION": EVLA/CORRELATOR/WIDAR/FULL_RESOLUTION"CRLF
        CONTENT_LOCATION": http://evla.nrao.edu/wcbe/XSDM/%s"CRLF
        ""CRLF
        "--%s"CRLF
        CONTENT_TYPE": text/xml; charset=%s"CRLF
        CONTENT_LOCATION": sdmDataHeader.xml"CRLF
        ""CRLF
        "%s";
    we_bdf_t *result = g_new0(we_bdf_t, 1);
    result->file = file;
    result->origin_offset = origin_offset;
    result->main_hdr = sdmhdr_main_hdr_copy(header);
    const gchar *oid = header->data_oid->href;
    if (g_str_has_prefix(oid, uid_prefix)) {
        oid += strlen(uid_prefix);
    }
    gchar *hdrstr = NULL;
    sdmhdr_main_hdr_write_memory(result->main_hdr, &hdrstr,
                                 WRITE_INDENTED_HEADERS);
    crlfify(&hdrstr);
    gsize len_hdrstr = strlen(hdrstr);
    if (hdrstr[len_hdrstr - 1] == '\n') {
        hdrstr[len_hdrstr - CRLF_LEN] = '\0';
    }
#ifndef USE_ALMA_5_1_0_5_MIME_WORKAROUNDS
    result->main_boundary =
        g_compute_checksum_for_string(CHECKSUM_ALGORITHM, hdrstr, -1);
#else
    result->main_boundary = g_strdup(boundary1);
#endif
    rc = fprintf(file, mime_hdr, result->main_boundary, oid,
                 result->main_boundary, sdmhdr_encoding, hdrstr);
    g_free(hdrstr);
    if (rc >= 0) {
        result->highest_read_offset = ftello(result->file);
        result->index =
            g_hash_table_new_full(g_str_hash, g_str_equal, g_free,
                                  destroy_subset_metadata);
        g_static_rec_mutex_init(&(result->index_mutex));
    }
    else {
        g_critical("Failed to write MIME file header: %s", g_strerror(errno));
        g_free(result);
        result = NULL;
    }
    return result;
}

we_bdf_t *
we_bdf_read(FILE *file)
{
    gboolean we_bdf_ok = FALSE;
    flockfile(file);
    fflush_unlocked(file);
    rewind(file);
    we_bdf_t *result = g_new0(we_bdf_t, 1);
    result->file = file;
    GHashTable *mime_hdrs = read_mime_headers(result->file);
    if (mime_hdrs != NULL) {
        result->main_boundary = get_boundary(mime_hdrs);
        if (result->main_boundary != NULL) {
            if (next_part(file, result->main_boundary) == 0) {
                result->index = g_hash_table_new_full(
                    g_str_hash, g_str_equal, g_free, destroy_subset_metadata);
                g_assert(result->index != NULL);
                g_static_rec_mutex_init(&(result->index_mutex));
                GHashTable *mime_hdrs2 = read_mime_headers(file);
                slurp_blank_lines(file, FALSE, 0);
                result->origin_offset = ftello(result->file);
                result->main_hdr = get_main_hdr(result);
                if (result->main_hdr != NULL) {
                    result->highest_read_offset = ftello(result->file);
                    if (result->highest_read_offset != -1) {
                        we_bdf_ok = TRUE;
                    }
                    else {
                        g_critical("Failed to 'ftello' BDF stream: %s",
                                   g_strerror(errno));
                    }
                }
                else {
                    g_info("Failed to parse sdmDataHeader element in BDF "
                           "stream");
                }
                if (mime_hdrs2 != NULL) {
                    g_hash_table_destroy(mime_hdrs2);
                }
            }
            else {
                g_info("Failed to find BDF stream main header");
            }
        }
        else {
            g_info("Failed to read top-level MIME boundary string from "
                   "BDF stream");
        }
        g_hash_table_destroy(mime_hdrs);
    }
    else {
        g_info("Failed to read MIME headers from BDF stream");
    }
    funlockfile(file);
    if (!we_bdf_ok) {
        we_bdf_free(result);
        result = NULL;
    }
    return result;
}

gchar **
we_bdf_project_paths(we_bdf_t *we_bdf)
{
    gchar **result;

    // scan remainder of file for subsets
    gint rc = fseeko(we_bdf->file, we_bdf->highest_read_offset, SEEK_SET);
    if (rc != -1) {
        while (index_next_subset(we_bdf) != NULL) {
            we_bdf->highest_read_offset = ftello(we_bdf->file);
        }
        // copy all subset project paths
        GPtrArray *paths = g_ptr_array_new();
        void gather_path(gpointer key, gpointer value G_GNUC_UNUSED,
                         gpointer reserved G_GNUC_UNUSED) {
            gchar *prjpth = g_strdup(key);
            g_ptr_array_add(paths, prjpth);
        }
        g_static_rec_mutex_lock((GStaticRecMutex *)&(we_bdf->index_mutex));
        g_hash_table_foreach(we_bdf->index, gather_path, NULL);
        g_static_rec_mutex_unlock((GStaticRecMutex *)&(we_bdf->index_mutex));
        g_ptr_array_add(paths, NULL);
        result = (gchar **)g_ptr_array_free(paths, FALSE);

        // sort the project path names by increasing file offset
        gint sort_by_offset(gchar **path0, gchar **path1) {
            subset_metadata_t *md0 =
                g_hash_table_lookup(we_bdf->index, *path0);
            subset_metadata_t *md1 =
                g_hash_table_lookup(we_bdf->index, *path1);
            gint result;
            if (md0->header_offset < md1->header_offset) {
                result = -1;
            }
            else if (md0->header_offset > md1->header_offset) {
                result = 1;
            }
            else {
                result = 0;
            }
            return result;
        }
        //qsort(result, g_strv_length(result), sizeof(gpointer), sort_by_offset);
    }
    else {
        g_warning("Failed to seek file position: %s", g_strerror(errno));
        result = NULL;
    }
    return result;
}

/**
 * we_bdf_main_hdr:
 * @we_bdf: initialized #we_bdf_t instance
 *
 * Read the main header of a BDF file.
 *
 * @returns: #sdmhdr_main_hdr_t instance based on main header of BDF
 * file; must be freed by caller with sdmhdr_main_hdr_free().
 */
sdmhdr_main_hdr_t *
we_bdf_main_hdr(we_bdf_t *we_bdf)
{
    g_return_val_if_fail(we_bdf != NULL, NULL);

    return sdmhdr_main_hdr_copy(we_bdf->main_hdr);
}

static sdmhdr_main_hdr_t *
get_main_hdr(const we_bdf_t *we_bdf)
{
    sdmhdr_main_hdr_t *result = NULL;
    // set file position to sdmDataHeader element, looking only within
    // first 1024 bytes of the file
    static gchar buff[1024];
    flockfile(we_bdf->file);
    gint rc = fseeko(we_bdf->file, we_bdf->origin_offset, SEEK_SET);
    if (rc != 0) {
        g_critical("Failed to set file position to origin of BDF file: %s",
                   g_strerror(errno));
        goto cleanup_and_return;
    }
    gsize nchars = fread_unlocked(buff, 1, sizeof(buff), we_bdf->file);
    if (nchars != 0) {
        gchar *hdr = g_strstr_len(buff, nchars, "<sdmDataHeader");
        if (hdr != NULL) {
            // sdmDataHeader element found (at least quite likely),
            // read the header using sdmhdr_main_hdr_read_streamd
            rc = fseeko(we_bdf->file, (hdr - buff) + we_bdf->origin_offset,
                        SEEK_SET);
            if (rc != 0) {
                g_critical("Failed to set file position of BDF file: %s",
                           g_strerror(errno));
                goto cleanup_and_return;
            }
            result = sdmhdr_main_hdr_read_stream(we_bdf->file);
        }
        else {
            g_message("Failed to find sdmDataHeader in BDF stream within "
                      "first %"G_GSIZE_FORMAT" bytes", sizeof(buff));
        }
    }
    else {
        g_message("Failed to read anything from BDF stream");
    }
 cleanup_and_return:
    funlockfile(we_bdf->file);

    return result;
}

/**
 * we_bdf_subset_hdr:
 * @we_bdf: initialized #we_bdf_t instance
 * @project_path: projectPath string of requested data subset
 *
 * Read the subset data header of a given data subset of a BDF file.
 *
 * @returns: #sdmhdr_subset_hdr_t instance based on subset header of
 * BDF file; must be freed by caller with sdmhdr_subset_hdr_free().
 */
sdmhdr_subset_hdr_t *
we_bdf_subset_hdr(we_bdf_t *we_bdf, const gchar *project_path)
{
    g_return_val_if_fail(we_bdf != NULL, NULL);
    g_return_val_if_fail(we_bdf->file != NULL, NULL);
    g_return_val_if_fail(project_path != NULL, NULL);

    sdmhdr_subset_hdr_t *result = NULL;
    const subset_metadata_t *metadata =
        get_subset_metadata(we_bdf, project_path);
    if (metadata != NULL) {
        result = sdmhdr_subset_hdr_copy(metadata->header);
    }
    return result;
}

tree_t *
we_bdf_bincomp(we_bdf_t *we_bdf, const gchar *project_path,
                CobdfBincomp bincomp)
{
    g_return_val_if_fail(we_bdf != NULL, NULL);
    g_return_val_if_fail(project_path != NULL, NULL);

    tree_t *result = NULL;

    // first find the appropriate MIME part's content-location value,
    // and calculate the size of the binary component in bytes
    const subset_metadata_t *metadata =
        get_subset_metadata(we_bdf, project_path);
    if (metadata != NULL) {
        sdmhdr_data_struct_t *ds = we_bdf->main_hdr->data_struct;
        off_t offset = -1;
        switch (bincomp) {
          case COBDF_ACTUAL_TIMES_BINCOMP:
            if (ds->actual_times != NULL) {
                offset = metadata->actual_times_offset;
            }
            break;
          case COBDF_ACTUAL_DURATIONS_BINCOMP:
            if (ds->actual_durations != NULL) {
                offset = metadata->actual_durations_offset;
            }
            break;
          case COBDF_ZERO_LAGS_BINCOMP:
            if (ds->zero_lags != NULL) {
                offset = metadata->zero_lags_offset;
            }
            break;
          case COBDF_FLAGS_BINCOMP:
            if (ds->flags != NULL) {
                offset = metadata->flags_offset;
            }
            break;
          case COBDF_CROSS_DATA_BINCOMP:
            if (ds->cross_data != NULL) {
                offset = metadata->cross_data_offset;
            }
            break;
          case COBDF_AUTO_DATA_BINCOMP:
            if (ds->auto_data != NULL) {
                offset = metadata->auto_data_offset;
            }
            break;
          case COBDF_WEIGHTS_BINCOMP:
            if (ds->weights != NULL) {
                offset = metadata->weights_offset;
            }
            break;
          default:
            g_assert_not_reached();
            break;
        }

        // now find the MIME part, and create a tree_t using the
        // binary
        if (offset >= 0) {
            flockfile(we_bdf->file);
            gint rc = fseeko(we_bdf->file, offset, SEEK_SET);
            if (rc != -1) {
                result = get_bincomp(we_bdf, metadata, bincomp);
            }
            else {
                g_critical("Failed to set BDF stream position for "
                           "previously located data subset '%s'",
                           project_path);
            }
            funlockfile(we_bdf->file);
        }
    }
    return result;
}

/**
 * get_bincomp:
 * @we_bdf: we_bdf instance
 * @subset: subset metadata instance
 * @bincomp: binary component type
 *
 * Create a tree_t structure for the binary component at the current
 * we_bdf stream position.
 */
static tree_t *
get_bincomp(const we_bdf_t *we_bdf, const subset_metadata_t *subset,
            CobdfBincomp bincomp)
{
    g_return_val_if_fail(we_bdf != NULL, NULL);
    g_return_val_if_fail(we_bdf->main_hdr != NULL, NULL);
    g_return_val_if_fail(we_bdf->main_hdr->data_struct != NULL, NULL);
    g_return_val_if_fail(bincomp < COBDF_NUM_BINCOMPS, NULL);

    tree_t *result = NULL;
    sdmhdr_bincomp_t *bc = NULL;
    gsize element_size;
    off_t offset = -1;
    sdmhdr_data_struct_t *ds = we_bdf->main_hdr->data_struct;
    // select a sdmhdr_bincomp_t element from ds, find the (primitive)
    // element size, and get the stream offset for the binary
    // component
    switch (bincomp) {
      case COBDF_ACTUAL_TIMES_BINCOMP:
        bc = ds->actual_times;
        element_size = sizeof(cobdf_actual_times_t);
        offset = subset->actual_times_offset;
        break;
      case COBDF_ACTUAL_DURATIONS_BINCOMP:
        bc = ds->actual_durations;
        element_size = sizeof(cobdf_actual_durations_t);
        offset = subset->actual_durations_offset;
        break;
      case COBDF_FLAGS_BINCOMP:
        bc = ds->flags;
        element_size = sizeof(cobdf_flags_t);
        offset = subset->flags_offset;
        break;
      case COBDF_ZERO_LAGS_BINCOMP:
        bc = ds->zero_lags;
        element_size = sizeof(cobdf_zero_lags_t);
        offset = subset->zero_lags_offset;
        break;
      case COBDF_CROSS_DATA_BINCOMP:
        switch (subset->header->cross_data_type) {
          case COBDF_INT16:
            element_size = 2;
            break;
#ifdef USE_ALMA_5_1_0_5_XML_WORKAROUNDS
          case COBDF_INT:
          case COBDF_FLOAT:
#endif
          case COBDF_INT32:
          case COBDF_FLOAT32:
            element_size = 4;
            break;
          default:
            element_size = 0;
            break;
        }
        if (element_size != 0) {
            bc = ds->cross_data;
            offset = subset->cross_data_offset;
        }
        break;
      case COBDF_AUTO_DATA_BINCOMP:
        bc = ds->auto_data;
        element_size = sizeof(cobdf_auto_data_real_t);
        offset = subset->auto_data_offset;
        break;
      case COBDF_WEIGHTS_BINCOMP:
        bc = &(ds->weights->common);
        element_size = weights_nbits(ds->weights);
        if (element_size % 8 == 0) {
            element_size /= 8;
        }
        else {
            g_warning("Number of bits to encode weights is unsupported: "
                      "not a multiple of eight.");
            element_size = 0;
        }
        offset = subset->weights_offset;
        break;
      default:
        g_assert_not_reached();
        break;
    }

    if (bc->axes[COBDF_TIME_AXIS]) {
        g_warning("TIM axis in binary components is unsupported: "
                  "axis ignored");
    }

    if (bc != NULL && element_size > 0) {
        // build the tree specification string required by the tree
        // module
        const sdmhdr_main_hdr_t *hdr = we_bdf->main_hdr;
        gchar *spec = NULL;
        gchar *baseline_nodes = NULL;
        gchar *antenna_nodes = NULL;
        guint num_baselines =
            (hdr->num_antennas * (hdr->num_antennas - 1)) / 2;
        if (bc->axes[COBDF_BASELINE_AXIS]) {
            baseline_nodes =
                balant_nodes(bc->axes, hdr, bincomp, COBDF_BASELINE_AXIS);
        }
        if (bc->axes[COBDF_ANTENNA_AXIS]) {
            antenna_nodes =
                balant_nodes(bc->axes, hdr, bincomp, COBDF_ANTENNA_AXIS);
        }
        if (bc->axes[COBDF_BASELINE_AXIS]
            && bc->axes[COBDF_ANTENNA_AXIS]) {
            if (baseline_nodes != NULL && antenna_nodes != NULL) {
                spec = g_strdup_printf("{%u,[%u%s,%u%s]}",
                                       num_baselines + hdr->num_antennas,
                                       num_baselines, baseline_nodes,
                                       hdr->num_antennas, antenna_nodes);
            }
            else if (baseline_nodes == NULL && antenna_nodes == NULL) {
                spec = g_strdup_printf("{%u}",
                                       num_baselines + hdr->num_antennas);
            }
            else {
                g_warning("Inconsistent tree structures under BAL and "
                          "ANT indices");
            }
        }
        else if (bc->axes[COBDF_BASELINE_AXIS]
                 && !bc->axes[COBDF_ANTENNA_AXIS]) {
            if (baseline_nodes != NULL) {
                spec = g_strdup_printf("{%u,[%s]}", num_baselines,
                                       baseline_nodes);
            }
            else {
                spec = g_strdup_printf("{%u}", num_baselines);
            }
        }
        else if (!bc->axes[COBDF_BASELINE_AXIS]
                 && bc->axes[COBDF_ANTENNA_AXIS]) {
            if (antenna_nodes != NULL) {
                spec = g_strdup_printf("{%u,[%s]}", hdr->num_antennas,
                                       antenna_nodes);
            }
            else {
                spec = g_strdup_printf("{%u}", hdr->num_antennas);
            }
        }
        else {
            // Without at least one of the BAL or ANT axes, the binary
            // components aren't well defined since there is in some
            // cases a dependence in the structure on whether a
            // baseline or antenna index is in effect (the dependence
            // is in the POL axis). Here we just arbitrarily assume
            // that only baselines are being indexed in terms of the
            // effect on the POL axis size.
            baseline_nodes =
                balant_nodes(bc->axes, hdr, bincomp, COBDF_BASELINE_AXIS);
            if (baseline_nodes != NULL) {
                spec = g_strdup(baseline_nodes);
            }
            else {
                spec = g_strdup("{1}");
            }
        }
        if (baseline_nodes != NULL) {
            g_free(baseline_nodes);
        }
        if (antenna_nodes != NULL) {
            g_free(antenna_nodes);
        }

        if (spec != NULL) {
            // create the tree, pointing to the binary component in
            // the stream
            result = tree_new(spec, element_size);
            if (tree_count(result) == bc->size) {
#if MMAP_BDF_BINARIES
                int fd = fileno(we_bdf->file);
                if (fd >= 0) {
                    tree_mmap_pars(result, fd, PROT_READ | PROT_WRITE, offset);
                }
                else {
                    g_critical("Failed to get file descriptor for stream: %s",
                               g_strerror(errno));
                    tree_free(result);
                    result = NULL;
                }
#else
                tree_stream_pars(result, we_bdf->file, offset);
#endif
            }
            else {
                g_warning("Declared table size not equal to calculated size");
                tree_free(result);
                result = NULL;
            }
            g_free(spec);
        }
    }

    return result;
}

void
we_bdf_add_subset_seq(we_bdf_t *we_bdf,
                      const sdmhdr_subset_hdr_t *subset_template,
                      guint last_index)
{
    g_return_if_fail(we_bdf != NULL);
    g_return_if_fail(subset_template != NULL);

    // scan remainder of file for subsets
    gint rc = fseeko(we_bdf->file, we_bdf->highest_read_offset, SEEK_SET);
    if (rc != -1) {
        while (index_next_subset(we_bdf) != NULL) {
            we_bdf->highest_read_offset = ftello(we_bdf->file);
        }
    }
    else {
        g_warning("Failed to seek BDF file pointer: %s", g_strerror(errno));
    }

    // find project path and file offset to last subset
    gchar *end_path = NULL;
    off_t end_off = 0;
    void find_end(gpointer key, gpointer value, gpointer user_data) {
        gchar *project_path = key;
        subset_metadata_t *subset = value;
        if (subset->header_offset > end_off) {
            end_off = subset->header_offset;
            end_path = project_path;
        }
    }
    g_static_rec_mutex_lock(&(we_bdf->index_mutex));
    g_hash_table_foreach(we_bdf->index, find_end, NULL);
    g_static_rec_mutex_unlock(&(we_bdf->index_mutex));
    g_debug("found through %s", end_path);

    // get index of last subset
    guint next_subset = 0;
    if (end_path != NULL) {
        gchar *chr = end_path + strlen(end_path) - 1;
        while (*chr == '/' && chr != end_path) {
            --chr;
        }
        while (*chr != '/' && chr != end_path) {
            --chr;
        }
        ++chr;
        guint64 last_subset = g_ascii_strtoull(chr, NULL, 0);
        next_subset = (gint)last_subset + 1;
    }

    if (last_index >= next_subset) {
        // copy some stuff from the subset_template
        sdmhdr_subset_hdr_t subset;
        subset.cross_data_type = subset_template->cross_data_type;
        subset.type = g_strdup(subset_template->type);
        subset.data_struct_ref = g_strdup(subset_template->data_struct_ref);
        subset.interval = subset_template->interval;
        subset.project_path = NULL;
        subset.flags_href = NULL;
        subset.actual_times_href = NULL;
        subset.actual_durations_href = NULL;
        subset.weights_href = NULL;
        subset.zero_lags_href = NULL;
        subset.cross_data_href = NULL;
        subset.auto_data_href = NULL;
        void guarded_free(gchar **str) {
            if (*str != NULL) {
                g_free(*str);
            }
            *str = NULL;
        }

        guint main_hdr_path_len = strlen(we_bdf->main_hdr->project_path);
        gboolean main_hdr_path_slash_end =
            (we_bdf->main_hdr->project_path[main_hdr_path_len - 1] == '/');
        // loop to add subsets
        while (last_index >= next_subset) {
            g_debug("adding %u", next_subset);
            // compute mid-point of integration interval
            subset.time = next_subset * subset.interval \
                + we_bdf->main_hdr->time;
            // generate project path string for next subset
            if (main_hdr_path_slash_end) {
                subset.project_path = g_strdup_printf(
                    "%s%u/", we_bdf->main_hdr->project_path, next_subset);
            }
            else {
                subset.project_path = g_strdup_printf(
                    "%s/%u/", we_bdf->main_hdr->project_path, next_subset);
            }
            // complete the subset header, then add the subset to bdf
            sdmhdr_set_subset_hrefs(we_bdf->main_hdr, &subset);
            we_bdf_add_subset(we_bdf, &subset);

            // clean up subset for next iteration
            guarded_free(&subset.project_path);
            guarded_free(&subset.flags_href);
            guarded_free(&subset.actual_times_href);
            guarded_free(&subset.actual_durations_href);
            guarded_free(&subset.weights_href);
            guarded_free(&subset.zero_lags_href);
            guarded_free(&subset.cross_data_href);
            guarded_free(&subset.auto_data_href);

            // advance subset time and index
            subset.time += subset.interval;
            ++next_subset;
        }
        g_free(subset.type);
        g_free(subset.data_struct_ref);
        rc = fsync(fileno(we_bdf->file));
        if (rc != 0) {
            g_warning("Failed to sync BDF file after adding subsets: %s",
                      g_strerror(errno));
        }
    }
}

off_t
we_bdf_add_subset(we_bdf_t *we_bdf, const sdmhdr_subset_hdr_t *header)
{
    g_return_val_if_fail(we_bdf != NULL, -1);
    g_return_val_if_fail(header != NULL, -1);
    g_return_val_if_fail(g_str_has_prefix(header->project_path,
                                          we_bdf->main_hdr->project_path), -1);
    g_return_val_if_fail(header->cross_data_href == NULL
                         || header->cross_data_type == COBDF_INT16
                         || header->cross_data_type == COBDF_INT32
#ifdef USE_ALMA_5_1_0_5_XML_WORKAROUNDS
                         || header->cross_data_type == COBDF_INT
                         || header->cross_data_type == COBDF_FLOAT
#endif
                         || header->cross_data_type == COBDF_FLOAT32, -1);

    off_t result = 0;
    subset_metadata_t *subset = NULL;

    // generate sdmDataSubsetHeader string
    gchar *header_str = NULL;
    sdmhdr_subset_hdr_write_memory(header, &header_str,
                                   WRITE_INDENTED_HEADERS);

    // convert LF chars to CRLF char sequences (done here so that MD5
    // digest computed below is accurate, although it doesn't really
    // need to be)
    crlfify(&header_str);
    gsize len_header_str = strlen(header_str);
    if (header_str[len_header_str - 1] == '\n') {
        header_str[len_header_str - CRLF_LEN] = '\0';
    }

#ifndef USE_ALMA_5_1_0_5_MIME_WORKAROUNDS
    // use ASCII representation of a digest of header string as MIME
    // boundary (this ought to vary with each integration, and is
    // reasonably long, but we won't guarantee the uniqueness property
    // of boundary values)
    gchar *bdy =
        g_compute_checksum_for_string(CHECKSUM_ALGORITHM, header_str, -1);
#else
    gchar *bdy = g_strdup(boundary2);
#endif

    flockfile(we_bdf->file);
    gssize rc = fseeko(we_bdf->file, 0, SEEK_END);
    if (rc < 0) {
        g_message("Failed to seek to end of BDF stream: %s",
                  g_strerror(errno));
        result = rc;
        goto cleanup_and_return;
    }

    // create a multipart child of 'parent'
    const gchar hdr[] =
        CRLF"--%s"CRLF
        CONTENT_TYPE": multipart/related; boundary=%s"CRLF
        CONTENT_DESCRIPTION": data and metadata subset"CRLF CRLF
        "--%s"CRLF
        CONTENT_TYPE": text/xml; charset=%s"CRLF
        CONTENT_LOCATION": %s"CRLF
        ""CRLF;
    gchar *cloc = g_build_path(G_DIR_SEPARATOR_S, header->project_path,
                               "desc.xml", NULL);
    rc = fprintf(we_bdf->file, hdr, we_bdf->main_boundary, bdy, bdy,
                 sdmhdr_encoding, cloc);
    gint last_errno = errno;
    g_free(cloc);
    if (rc < 0) {
        g_critical("Failed to write MIME headers for new data subset to BDF "
                   "stream: %s", g_strerror(last_errno));
        result = rc;
        goto cleanup_and_return;
    }
    result += rc;

    subset = g_slice_new0(subset_metadata_t);
    subset->header = sdmhdr_subset_hdr_copy(header);
    subset->header_offset = ftello(we_bdf->file);
    subset->flags_offset = -1;
    subset->actual_times_offset = -1;
    subset->actual_durations_offset = -1;
    subset->weights_offset = -1;
    subset->zero_lags_offset = -1;
    subset->cross_data_offset = -1;
    subset->auto_data_offset = -1;

    rc = fputs_unlocked(header_str, we_bdf->file);
    if (rc < 0) {
        g_critical("Failed to write data subset header to BDF stream: %s",
                   g_strerror(errno));
        result = rc;
        goto cleanup_and_return;
    }
    result += rc;

    // write empty binary components
    sdmhdr_data_struct_t *ds = we_bdf->main_hdr->data_struct;
    if (header->actual_times_href != NULL) {
        rc = write_empty_bincomp(we_bdf, bdy, header->actual_times_href,
                                 ds->actual_times->size
                                 * sizeof(cobdf_actual_times_t), -1,
                                 &(subset->actual_times_offset));
        if (rc > 0) {
            result += rc;
        }
    }
    if (rc >= 0 && header->actual_durations_href != NULL) {
        rc = write_empty_bincomp(we_bdf, bdy, header->actual_durations_href,
                                 ds->actual_durations->size
                                 * sizeof(cobdf_actual_durations_t), -1,
                                 &(subset->actual_durations_offset));
        if (rc > 0) {
            result += rc;
        }
    }
    if (rc >= 0 && header->flags_href != NULL) {
        rc = write_empty_bincomp(we_bdf, bdy, header->flags_href,
                                 ds->flags->size * sizeof(cobdf_flags_t), -1,
                                 &(subset->flags_offset));
        if (rc > 0) {
            result += rc;
        }
    }
    if (rc >= 0 && header->zero_lags_href != NULL) {
        rc = write_empty_bincomp(we_bdf, bdy, header->zero_lags_href,
                                 ds->zero_lags->size
                                 * sizeof(cobdf_zero_lags_t), -1,
                                 &(subset->zero_lags_offset));
        if (rc > 0) {
            result += rc;
        }
    }
    if (rc >= 0 && header->weights_href != NULL) {
        guint nbits = weights_nbits(ds->weights);
        gsize total_bits = ds->weights->common.size * nbits;
        gsize weights_size = total_bits / 8;
        if (total_bits % 8 != 0) {
            weights_size++;
        }
        rc = write_empty_bincomp(we_bdf, bdy, header->weights_href,
                                 weights_size, (8 - (total_bits % 8)) % 8,
                                 &(subset->weights_offset));
        if (rc > 0) {
            result += rc;
        }
    }
    if (rc >= 0 && header->auto_data_href != NULL) {
        rc = write_empty_bincomp(we_bdf, bdy, header->auto_data_href,
                                 ds->auto_data->size
                                 * sizeof(cobdf_auto_data_real_t), -1,
                                 &(subset->auto_data_offset));
        if (rc > 0) {
            result += rc;
        }
    }
    if (rc >= 0 && header->cross_data_href != NULL) {
        gsize element_size;
        switch (header->cross_data_type) {
          case COBDF_INT16:
            element_size = 2;
            break;
#ifdef USE_ALMA_5_1_0_5_XML_WORKAROUNDS
          case COBDF_INT:
          case COBDF_FLOAT:
#endif
          case COBDF_INT32:
          case COBDF_FLOAT32:
            element_size = 4;
            break;
          default:
            g_assert_not_reached();
            element_size = 0;
            break;
        }
        rc = write_empty_bincomp(we_bdf, bdy, header->cross_data_href,
                                 ds->cross_data->size * element_size, -1,
                                 &(subset->cross_data_offset));
        if (rc > 0) {
            result += rc;
        }
    }
    if (rc < 0) {
        g_warning("Failed to create space for a binary component in BDF "
                  "stream: %s", g_strerror(errno));
        result = rc;
        goto cleanup_and_return;
    }

    rc = fprintf(we_bdf->file, CRLF"--%s--", bdy);
    if (rc < 0) {
        g_critical("Failed to close MIME part for data subset in BDF "
                   "stream: %s", g_strerror(errno));
        result = rc;
        goto cleanup_and_return;
    }
    result += rc;

 cleanup_and_return:
    fflush_unlocked(we_bdf->file);
    funlockfile(we_bdf->file);
    g_free(bdy);
    if (header_str != NULL) {
        g_free(header_str);
    }
    if (result > 0) {
        g_static_rec_mutex_lock(&(we_bdf->index_mutex));
        g_hash_table_insert(we_bdf->index, g_strdup(header->project_path),
                            subset);
        g_static_rec_mutex_unlock(&(we_bdf->index_mutex));
    }
    else {
        if (subset != NULL) {
            destroy_subset_metadata(subset);
        }
    }

    return result;
}

#if 0
void
we_bdf_truncate(we_bdf_t *we_bdf, guint num_subsets)
{
    g_static_rec_mutex_lock(&(we_bdf->index_mutex));
    guint current_number = g_hash_table_size(we_bdf->index);
    if (current_number > num_subsets) {
        gint sort_by_offset(gchar **path0, gchar **path1) {
            subset_metadata_t *md0 =
                g_hash_table_lookup(we_bdf->index, *path0);
            subset_metadata_t *md1 =
                g_hash_table_lookup(we_bdf->index, *path1);
            gint result;
            if (md0->header_offset < md1->header_offset) {
                result = -1;
            }
            else if (md0->header_offset > md1->header_offset) {
                result = 1;
            }
            else {
                result = 0;
            }
            return result;
        }
        gchar **paths = we_bdf_project_paths(we_bdf);
        qsort(paths, g_strv_length(paths), sizeof(gpointer), sort_by_offset);
        subset_metadata_t *excess_subset =
            g_hash_table_lookup(we_bdf->index, *(paths + num_subsets));
        // truncate after excess_subset->header_offset (or so)
        gint rc =
            ftruncate(fileno(we_bdf->file), excess_subset->header_offset);
        if (G_LIKELY(rc == 0)) {
            // fix up MIME stuff
            // we_bdf_close(we_bdf); TODO: should we do this?
            // remove excess subsets from 'index'
            gchar **path = paths + num_subsets;
            while (*path != NULL) {
                g_hash_table_remove(we_bdf->index, *path);
                ++path;
            }
        }
        else {
            // FIXME:
        }
        g_strfreev(paths);
    }
    g_static_rec_mutex_unlock(&(we_bdf->index_mutex));
}
#endif

void
we_bdf_close(we_bdf_t *we_bdf)
{
    g_return_if_fail(we_bdf != NULL);

    gchar *end_bdy = g_strdup_printf("--%s--"CRLF, we_bdf->main_boundary);
    flockfile(we_bdf->file);
    fseeko(we_bdf->file, -3 * (strlen(we_bdf->main_boundary) + 6), SEEK_END);
    gboolean found_end = FALSE;
    gchar buf[512];
    gchar *line = fgets_unlocked(buf, sizeof(buf), we_bdf->file);
    while (line != NULL && !found_end) {
        found_end = (strcmp(end_bdy, buf) == 0);
        line = fgets_unlocked(buf, sizeof(buf), we_bdf->file);
    }
    if (!found_end) {
        fseeko(we_bdf->file, 0, SEEK_END);
        fprintf(we_bdf->file, CRLF"--%s--"CRLF, we_bdf->main_boundary);
    }
    funlockfile(we_bdf->file);
    g_free(end_bdy);
}

void
we_bdf_free(we_bdf_t *we_bdf)
{
    g_return_if_fail(we_bdf != NULL);

    if (we_bdf->index != NULL) {
        g_hash_table_destroy(we_bdf->index);
    }
    if (we_bdf->main_hdr != NULL) {
        sdmhdr_main_hdr_free(we_bdf->main_hdr);
    }
    g_static_rec_mutex_free(&(we_bdf->index_mutex));
    if (we_bdf->main_boundary != NULL) {
        g_free(we_bdf->main_boundary);
    }
    g_free(we_bdf);
}

static gssize
write_empty_bincomp(we_bdf_t *we_bdf, const gchar *bdy,
                    const gchar *bincomp_href, gsize bincomp_size,
                    gint padding, off_t *binary_offset)
{
    gsize result = -1;
    if (padding < 0) {
        const gchar hdr[] =
            CRLF"--%s"CRLF
            CONTENT_TYPE": application/octet-stream"CRLF
            CONTENT_LOCATION": %s"CRLF CRLF;
        result = fprintf(we_bdf->file, hdr, bdy, bincomp_href);
    }
    else {
        const gchar hdr[] =
            CRLF"--%s"CRLF
            CONTENT_TYPE": application/octet-stream; padding=%d"CRLF
            CONTENT_LOCATION": %s"CRLF CRLF;
        result = fprintf(we_bdf->file, hdr, bdy, padding, bincomp_href);
    }
    if (result >= 0) {
        *binary_offset = ftello(we_bdf->file);
        result = fseeko(we_bdf->file, bincomp_size, SEEK_CUR);
        if  (result >= 0) {
            result += bincomp_size;
        }
    }
    return result;
}

/**
 * read_mime_headers:
 * @file: stream
 *
 * Read MIME headers starting at current stream position of @file. It
 * is assumed that the stream is locked when this function is called.
 * Complete header values, including all parameters (with whitespace
 * compression and linebreak elimination), are associated with each
 * header.
 *
 * @returns: #GHashTable of (header, value) pairs.
 */
static GHashTable *
read_mime_headers(FILE *file)
{
    static gchar buff[512];
    GHashTable *result =
        g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
    gchar *key = NULL;
    GString *val = NULL;
    off_t off = ftello(file);
    gchar *ch = fgets_unlocked(buff, sizeof(buff), file);
    void record_hdr(void) {
        if (val == NULL) {
            val = g_string_new("");
        }
        g_hash_table_insert(result, key, val->str);
        g_string_free(val, FALSE);
        key = NULL;
        val = NULL;
    }
    while (ch != NULL && strcmp(ch, CRLF) != 0) {
        if (key != NULL && !g_ascii_isspace(*ch)) {
            // new line that starts without whitespace ends any
            // pending header
            record_hdr();
        }
        if (key == NULL) {
            // no pending header
            g_assert(val == NULL);
            // search for colon to indicate a MIME header key
            ch = strchr(buff, ':');
            if (ch != NULL) {
                // record header key
                key = g_ascii_strdown(buff, ch - buff);
                // record rest of line, without leading whitespace or
                // trailing CRLF, as header value
                ch = g_strchug(ch + 1);
                if (g_str_has_suffix(ch, CRLF)) {
                    ch[strlen(ch) - CRLF_LEN] = '\0';
                }
                val = g_string_new(ch);
            }
        }
        else {
            // pending header, so just append line to header value
            // string (w/o trailing CRLF)
            g_assert(val != NULL);
            if (g_str_has_suffix(buff, CRLF)) {
                buff[strlen(buff) - CRLF_LEN] = '\0';
            }
            g_string_append(val, buff);
        }
        // If 'ch' is NULL at this point, then no ':' was found in the
        // line and no header is pending. This condition is not
        // expected in a well-formed MIME message, so we won't scan
        // for any more MIME headers.
        if (ch == NULL) {
            gint rc = fseeko(file, off, SEEK_SET);
            if (rc < 0) {
                g_critical("Failed to seek file position: %s",
                           g_strerror(errno));
            }
        }
        else {
            off = ftello(file);
            if (off < 0) {
                g_critical("Failed to get file position: %s",
                           g_strerror(errno));
            }
            ch = fgets_unlocked(buff, sizeof(buff), file);
        }
    }
    if (key != NULL) {
        record_hdr();
    }
    return result;
}

/**
 * next_part:
 * @file: we_bdf stream
 * @bdy: MIME boundary string (w/o "--" prefix)
 *
 * Advance @file stream position to next MIME part indicated by the
 * given boundary string. The stream position after successfully
 * finding a MIME part will be at the line following the boundary
 * string.
 *
 * @returns: 0, if a MIME part was found; -1, otherwise
 */
static gint
next_part(FILE *file, const gchar *bdy)
{
    gint result = -2;
    gchar *bdy_prefix = g_strdup_printf("--%s", bdy);
    guint bdy_prefix_len = strlen(bdy_prefix);
    guint bdy_prefix_pos = 0;
    guint suffix_pos = 0;
    gchar final_end[] = "--"CRLF;
    gchar part_end[] = CRLF;

    static gchar buff[4096];
    gchar *pos = NULL;
    gchar *end = NULL;
    void replenish_buffer(void) {
        gsize num_read = fread_unlocked(buff, 1, sizeof(buff), file);
        pos = buff;
        end = &(buff[num_read]);
    }

    enum {SCAN_PREFIX = 0, SCAN_SUFFIX = 1, SCAN_PART_SUFFIX = 2,
          SCAN_FINAL_SUFFIX = 3} scan_state = SCAN_PREFIX;

    do {
        replenish_buffer();
        if (G_LIKELY(pos != end)) {
            while (G_LIKELY(result == -2 && pos < end)) {
                switch (scan_state) {
                  case SCAN_PREFIX:
                    if (G_UNLIKELY(*pos == bdy_prefix[bdy_prefix_pos])) {
                        ++bdy_prefix_pos;
                        if (bdy_prefix_pos == bdy_prefix_len) {
                            scan_state = SCAN_SUFFIX;
                            suffix_pos = 0;
                        }
                    }
                    else {
                        bdy_prefix_pos = 0;
                    }
                    break;
                  case SCAN_SUFFIX:
                    if (*pos == final_end[0]) {
                        scan_state = SCAN_FINAL_SUFFIX;
                        suffix_pos = 1;
                    }
                    else if (*pos == part_end[0]) {
                        scan_state = SCAN_PART_SUFFIX;
                        suffix_pos = 1;
#ifdef USE_ALMA_5_1_0_5_MIME_WORKAROUNDS
                        // this is needed because CRLF may be a single
                        // character in this case
                        if (suffix_pos == sizeof(part_end) - 1) {
                            result = 0;
                            gint rc =
                                fseeko(file, (off_t)pos + 1 - (off_t)end,
                                       SEEK_CUR);
                            if (rc != 0) {
                                g_critical("Failed to set BDF file position "
                                           "to point after MIME boundary: "
                                           "%s", g_strerror(errno));
                            }
                        }
#endif
                    }
                    else {
                        scan_state = SCAN_PREFIX;
                    }
                    break;
                  case SCAN_PART_SUFFIX:
                    if (*pos == part_end[suffix_pos]) {
                        ++suffix_pos;
                        if (suffix_pos == sizeof(part_end) - 1) {
                            result = 0;
                            gint rc =
                                fseeko(file, (void*)pos + 1 - (void*)end,
                                       SEEK_CUR);
                            if (rc != 0) {
                                g_critical("Failed to set BDF file position "
                                           "to point after MIME boundary: "
                                           "%s", g_strerror(errno));
                            }
                        }
                    }
                    else {
                        scan_state = SCAN_PREFIX;
                    }
                    break;
                  case SCAN_FINAL_SUFFIX:
                    if (*pos == final_end[suffix_pos]) {
                        ++suffix_pos;
                        if (suffix_pos == sizeof(final_end) - 1) {
                            result = -1;
                            gint rc =
                                fseeko(file, (void*)pos + 1 - (void*)end,
                                       SEEK_CUR);
                            if (rc != 0) {
                                g_critical("Failed to set BDF file position "
                                           "to point after MIME boundary: "
                                           "%s", g_strerror(errno));
                            }
                        }
                    }
                    else {
                        scan_state = SCAN_PREFIX;
                    }
                    break;
                  default:
                    g_assert_not_reached();
                    break;
                }
                ++pos;
            }
        }
        else {
            result = -1;
            gint rc = ferror(file);
            if (rc != 0) {
                g_critical("Read error on BDF file: %s", g_strerror(rc));
            }
        }
    } while (result == -2);
    g_free(bdy_prefix);
    return result;
}

/**
 * check_boundary:
 * @file: we_bdf stream
 * @bdy: MIME boundary (w/o "--" prefix)
 *
 * Scan a MIME boundary string at the current stream position
 * (including the leading CRLF). It is assumed that the caller has
 * locked the @file stream prior to call.
 *
 * @returns: %BDY_OK, if the given boundary string was read;
 *           %BDY_END, if the "end" (with "--" suffix) boundary
 *           string was read;
 *           %BDY_NOMATCH, if something other than a boundary string
 *           was read
 */
static gint
check_boundary(FILE *file, const gchar *bdy)
{
    gint result = BDY_NOMATCH;
    static gchar buff[4096];

    gchar *ch = fgets_unlocked(buff, sizeof(buff), file);
    if (ch != NULL) {
        if (strcmp(CRLF, ch) == 0) {
            ch = fgets_unlocked(buff, sizeof(buff), file);
        }
        else {
            ch = NULL;
            result = BDY_NOMATCH;
        }
    }
    if (ch != NULL) {
        if (g_str_has_prefix(buff, "--")
            && strncmp(buff + 2, bdy, strlen(bdy)) == 0) {
            if (g_str_has_suffix(buff, "--"CRLF)) {
                result = BDY_END;
            }
            else {
                result = BDY_OK;
            }
        }
        else {
            result = BDY_NOMATCH;
        }
    }
    return result;
}

/**
 * get_boundary:
 * @headers: MIME headers hash table, as returned from read_mime_headers()
 *
 * Parse the MIME boundary string from a set of MIME headers.
 *
 * @returns: boundary string, which caller should free using g_free()
 */
static gchar *
get_boundary(GHashTable *headers)
{
    gchar *result = NULL;
    gchar *ctype = g_hash_table_lookup(headers, "content-type");
    if (ctype != NULL) {
        const gchar boundary[] = "boundary=";
        gchar *bdy_key = strstr(ctype, boundary);
        if (bdy_key != NULL) {
            gchar *bdy_val = bdy_key + sizeof(boundary) - 1;
            if (*bdy_val == '"') {
                result = g_strdup(bdy_val + 1);
                // search for matching end quote (the following will
                // accept invalid MIME multi-part boundary strings,
                // all we're doing is unescaped quote finding)
                gchar *q = strchr(result, '"');
                while (q != NULL && *(q - 1) == '\\') {
                    q = strchr(q + 1, '"');
                }
                if (q != NULL) {
                    *q = '\0';
                }
                else {
                    g_info("Incorrect quoting of 'boundary' attribute "
                           "of 'Content-type' header.");
                    g_free(result);
                    result = NULL;
                }
            }
            else {
                result = g_strdup(bdy_val);
                // find end of boundary string by locating first
                // invalid MIME multi-part boundary string character
                gchar *q = result;
                while (g_ascii_isalnum(*q) || *q == '\'' || *q == '('
                       || *q == ')' || *q == '+' || *q == '_' || *q == ','
                       || *q == '-' || *q == '.' || *q == '/' || *q == ':'
                       || *q == '=' || *q == '?') {
                    ++q;
                }
                *q = '\0';
            }
        }
    }
    return result;
}

/**
 * slurp_blank_lines:
 * @file: we_bdf stream
 * @strictly_empty: %TRUE if only those lines with zero characters before
 * the CRLF should be considered to be blank
 * @max: maximum number of blank lines to swallow; 0 indicates no limit
 *
 * Read past sequence any blank lines starting at the current stream
 * position. It is assumed that the caller has locked the stream prior
 * to the call.
 */
static void
slurp_blank_lines(FILE *file, gboolean strictly_empty, guint max)
{
    guint num = 0;
    off_t off = ftello(file);
    if (off >= 0) {
        gchar buff[80];
        gchar *ch = fgets_unlocked(buff, sizeof(buff), file);
        while (ch != NULL) {
            buff[sizeof(buff) - 1] = '\0';
            if (!strictly_empty) {
                while (*ch == ' ' || *ch == '\t') {
                    ++ch;
                }
            }
            if (strcmp(CRLF, ch) != 0) {
                ch = NULL;
                int rc = fseeko(file, off, SEEK_SET);
                if (rc != 0) {
                    g_critical("Failed to set file position: %s",
                               g_strerror(errno));
                }
            }
            else {
                if (++num == max) {
                    ch = NULL;
                }
                else {
                    off = ftello(file);
                    if (off >= 0) {
                        ch = fgets_unlocked(buff, sizeof(buff), file);
                    }
                    else {
                        g_critical("Failed to get file position: %s",
                                   g_strerror(errno));
                    }
                }
            }
        }
    }
    else {
        g_error("Failed to get file position: %s", g_strerror(errno));
    }
}

/**
 * get_subset_metadata:
 * @we_bdf: we_bdf instance
 * @project_path: project path of desired subset
 *
 * Get the metadata for a given data subset. To avoid any chance at
 * deadlock the caller shall not have locked the we_bdf stream prior
 * to the call.
 */
static const subset_metadata_t *
get_subset_metadata(we_bdf_t *we_bdf, const gchar *project_path)
{
    g_return_val_if_fail(project_path != NULL, NULL);

    g_static_rec_mutex_lock(&(we_bdf->index_mutex));
    subset_metadata_t *result
        = g_hash_table_lookup(we_bdf->index, project_path);
    if (result == NULL) {
        // scan to end of stream for new subsets
        flockfile(we_bdf->file);
        gint rc =
            fseeko(we_bdf->file, we_bdf->highest_read_offset, SEEK_SET);
        if (rc != -1) {
            const gchar *subset_path = index_next_subset(we_bdf);
            while (subset_path != NULL) {
                we_bdf->highest_read_offset = ftello(we_bdf->file);
                if (we_bdf->highest_read_offset != -1) {
                    if (strcmp(subset_path, project_path) != 0) {
                        subset_path = index_next_subset(we_bdf);
                    }
                    else {
                        result =
                            g_hash_table_lookup(we_bdf->index, project_path);
                        g_assert(result != NULL);
                        subset_path = NULL;
                    }
                }
                else {
                    g_critical("Failed to get file position for BDF "
                               "stream: %s", g_strerror(errno));
                    subset_path = NULL;
                }
            }
        }
        else {
            g_critical("Failed to set file position for BDF stream: %s",
                       g_strerror(errno));
        }
        funlockfile(we_bdf->file);
    }
    g_static_rec_mutex_unlock(&(we_bdf->index_mutex));
    return result;
}

static const gchar *
index_next_subset(we_bdf_t *we_bdf)
{
    gchar *result = NULL;
    g_static_rec_mutex_lock(&(we_bdf->index_mutex));
    flockfile(we_bdf->file);
    if (next_part(we_bdf->file, we_bdf->main_boundary) == 0) {
        GHashTable *mime_hdrs = read_mime_headers(we_bdf->file);
        if (mime_hdrs != NULL) {
            gchar *subset_bdy = get_boundary(mime_hdrs);
            if (subset_bdy != NULL) {
                if (next_part(we_bdf->file, subset_bdy) == 0) {
                    GHashTable *subset_mime_hdrs =
                        read_mime_headers(we_bdf->file);
                    if (subset_mime_hdrs != NULL) {
                        slurp_blank_lines(we_bdf->file, FALSE, 0);
                        off_t subset_offset = ftello(we_bdf->file);
                        sdmhdr_subset_hdr_t *subset =
                            sdmhdr_subset_hdr_read_stream(we_bdf->file);
                        if (subset != NULL) {
                            result = g_strdup(subset->project_path);
                            subset_metadata_t *metadata =
                                alloc_subset_metadata(we_bdf, subset_offset,
                                                      subset_bdy, subset);
                            g_warn_if_fail(metadata != NULL);
                            g_hash_table_insert(we_bdf->index, result,
                                                metadata);
                        }
                        g_hash_table_destroy(subset_mime_hdrs);
                    }
                }
                g_free(subset_bdy);
            }
            g_hash_table_destroy(mime_hdrs);
        }
    }
    funlockfile(we_bdf->file);
    g_static_rec_mutex_unlock(&(we_bdf->index_mutex));
    return result;
}

static guint
num_cross_primitives(sdmhdr_xpp_t xpp, CobdfBincomp bincomp)
{
    guint result;

    switch (xpp) {
      case SDMHDR_XPP_RR:
      case SDMHDR_XPP_LL:
        if (bincomp == COBDF_CROSS_DATA_BINCOMP) {
            result = 2; // C(omplex) = 2 primitives
        }
        else {
            result = 1; // R(eal) = 1 primitive
        }
        break;
      case SDMHDR_XPP_RR_LL:
        if (bincomp == COBDF_CROSS_DATA_BINCOMP) {
            result = 4; // C, C
        }
        else {
            result = 2; // R, R
        }
        break;
      case SDMHDR_XPP_RR_RL_LR_LL:
        if (bincomp == COBDF_CROSS_DATA_BINCOMP) {
            result = 8; // C, C, C, C
        }
        else {
            result = 4; // R, R, R, R
        }
        break;
      case SDMHDR_XPP_INVALID:
        result = 0;
        break;
      default:
        g_assert_not_reached();
        result = 0;
        break;
    }
    return result;
}

static guint
num_auto_primitives(sdmhdr_app_t app, CobdfBincomp bincomp)
{
    guint result;

    switch (app) {
      case SDMHDR_APP_RR:
      case SDMHDR_APP_LL:
        result = 1;     // R
        break;
      case SDMHDR_APP_RR_LL:
        result = 2;     // R, R
        break;
      case SDMHDR_APP_RR_RL_LL:
        if (bincomp == COBDF_AUTO_DATA_BINCOMP) {
            result = 4; // R, C, R
        }
        else if (bincomp == COBDF_ZERO_LAGS_BINCOMP) {
            result = 2; // R, <n/a>, R
        }
        else {
            result = 3; // R, R, R
        }
        break;
      case SDMHDR_APP_INVALID:
        result = 0;
        break;
      default:
        g_assert_not_reached();
        result = 0;
        break;
    }
    return result;
}

static gchar *
spp_nodes(const gboolean *axes, const sdmhdr_main_hdr_t *hdr,
          CobdfBincomp bincomp, CobdfAxisName bal_or_ant,
          const sdmhdr_sw_t *spectral_window)
{
    gchar *result;

    if (axes[COBDF_POLARIZATION_AXIS]) {
        guint num_nodes;
        switch (bal_or_ant) {
          case COBDF_BASELINE_AXIS:
            num_nodes = num_cross_primitives(
                spectral_window->cross_pol_products, bincomp);
            break;
          case COBDF_ANTENNA_AXIS:
            num_nodes =
                num_auto_primitives(spectral_window->sd_pol_products, bincomp);
            break;
          default:
            g_assert_not_reached();
            break;
        }
        result = g_strdup_printf("{%u}", num_nodes);
    }
    else {
        result = NULL;
    }
    return result;
}

/**
 * node_multiplier:
 * @bincomp: binary component type
 *
 * The number of primitive data elements needed for an element of the
 * given binary component type, assuming no differentiation on
 * polarization product. This is needed whenever there is no POL axis
 * in the binary. Generally we can allow even
 * %COBDF_CROSS_DATA_BINCOMP and %COBDF_AUTO_DATA_BINCOMP binaries
 * without a POL axis, even though that shouldn't happen in practice,
 * if we simply return the greatest number of primitive data elements
 * among all POL axis values.
 *
 * @returns: greatest number of primitive data elements, across all
 * possible polarization products, needed for an element of @bincomp
 */
static guint
node_multiplier(CobdfBincomp bincomp)
{
    guint result;
    switch (bincomp) {
      case COBDF_CROSS_DATA_BINCOMP:
      case COBDF_AUTO_DATA_BINCOMP:
        result = 2; // Complex = 2 primitives
        break;
      default:
        result = 1; // Real = 1 primitive
        break;
    }
    return result;
}

static gchar *
apc_nodes(const gboolean *axes, const sdmhdr_main_hdr_t *hdr,
          CobdfBincomp bincomp, CobdfAxisName bal_or_ant,
          const sdmhdr_sw_t *spectral_window)
{
    gchar *result;
    gchar *nodes = spp_nodes(axes, hdr, bincomp, bal_or_ant, spectral_window);
    if (axes[COBDF_SPECTRAL_POINT_AXIS]) {
        if (nodes != NULL) {
            result = g_strdup_printf(
                "{%u,[%s]}", spectral_window->num_spectral_point, nodes);
            g_free(nodes);
        }
        else {
            result = g_strdup_printf("{%u}",
                                     node_multiplier(bincomp)
                                     * spectral_window->num_spectral_point);
        }
    }
    else {
        result = nodes;
    }
    return result;
}

static gchar *
bin_nodes(const gboolean *axes, const sdmhdr_main_hdr_t *hdr,
          CobdfBincomp bincomp, CobdfAxisName bal_or_ant,
          const sdmhdr_sw_t *spectral_window)
{
    gchar *result;
    guint num_apc = 0;
    while (num_apc < G_N_ELEMENTS(hdr->data_struct->apc)
           && hdr->data_struct->apc[num_apc] != COBDF_AP_INVALID) {
        ++num_apc;
    }

    gchar *nodes = apc_nodes(axes, hdr, bincomp, bal_or_ant, spectral_window);
    if (axes[COBDF_ATM_PHASE_CORRECTION_AXIS]) {
        if (nodes != NULL) {
            result = g_strdup_printf("{%u,[%s]}", num_apc, nodes);
            g_free(nodes);
        }
        else {
            result = g_strdup_printf("{%u}",
                                     node_multiplier(bincomp) * num_apc);
        }
    }
    else {
        result = nodes;
    }
    return result;
}

static gchar *
sw_nodes(const gboolean *axes, const sdmhdr_main_hdr_t *hdr,
         CobdfBincomp bincomp, CobdfAxisName bal_or_ant,
         const sdmhdr_bb_t *baseband, guint num_sw)
{
    gchar *result;
    GString *spec = g_string_new(NULL);
    if (axes[COBDF_BIN_AXIS]) {
        guint sw;
        for (sw = 0; sw < num_sw; ++sw) {
            const sdmhdr_sw_t *spectral_window = &(baseband->sw[sw]);
            gchar *nodes =
                bin_nodes(axes, hdr, bincomp, bal_or_ant, spectral_window);
            if (nodes != NULL) {
                g_string_append_printf(
                    spec, "{%u,[%s]},", spectral_window->num_bin, nodes);
                g_free(nodes);
            }
            else {
                g_string_append_printf(spec, "{%u},",
                                       node_multiplier(bincomp)
                                       * spectral_window->num_bin);
            }
        }
    }
    else {
        guint sw;
        for (sw = 0; sw < num_sw; ++sw) {
            const sdmhdr_sw_t *spectral_window = &(baseband->sw[sw]);
            gchar *nodes =
                bin_nodes(axes, hdr, bincomp, bal_or_ant, spectral_window);
            if (nodes != NULL) {
                g_string_append_printf(spec, "%s,", nodes);
                g_free(nodes);
            }
            else {
                // in this case, 'nodes' for all values of 'sw' should
                // be NULL as well (since that's determined by
                // 'axes'), which means that 'spec' should always be
                // of zero length.
                g_assert(spec->len == 0);
            }
        }
    }


    if (spec->len != 0) {
        g_string_truncate(spec, spec->len - 1);
        result = g_string_free(spec, FALSE);
    }
    else {
        result = NULL;
    }
    return result;
}

static gchar *
baseband_nodes(const gboolean *axes, const sdmhdr_main_hdr_t *hdr,
               CobdfBincomp bincomp, CobdfAxisName bal_or_ant,
               guint num_basebands)
{
    gchar *result;
    GString *spec = g_string_new(NULL);
    if (axes[COBDF_SPECTRAL_WINDOW_AXIS]) {
        guint bb;
        for (bb = 0; bb < num_basebands; ++bb) {
            const sdmhdr_bb_t *baseband = hdr->data_struct->basebands[bb];
            gchar *nodes = sw_nodes(axes, hdr, bincomp, bal_or_ant,
                                    baseband, baseband->num_sw);
            if (nodes != NULL) {
                g_string_append_printf(spec, "{%u,[%s]},", baseband->num_sw,
                                       nodes);
                g_free(nodes);
            }
            else {
                g_string_append_printf(spec, "{%u},",
                                       node_multiplier(bincomp)
                                       * baseband->num_sw);
            }
        }
    }
    else {
        guint bb;
        for (bb = 0; bb < num_basebands; ++bb) {
            const sdmhdr_bb_t *real_baseband = hdr->data_struct->basebands[bb];
            sdmhdr_bb_t *bb_max = g_malloc(SIZEOF_SDMHDR_BB(1));
            bb_max->sw[0].num_bin = 0;
            bb_max->sw[0].num_spectral_point = 0;
            bb_max->sw[0].cross_pol_products = SDMHDR_XPP_INVALID;
            bb_max->sw[0].sd_pol_products = SDMHDR_APP_INVALID;
            maximal_spectral_window(real_baseband, &(bb_max->sw[0]), bincomp);
            gchar *nodes =
                sw_nodes(axes, hdr, bincomp, bal_or_ant, bb_max, 1);
            if (nodes != NULL) {
                g_string_append_printf(spec, "%s,", nodes);
                g_free(nodes);
            }
            else {
                // in this case, 'nodes' for all values of 'bb' should
                // be NULL as well (since that's determined by
                // 'axes'), which means that 'spec' should always be
                // of zero length.
                g_assert(spec->len == 0);
            }
            g_free(bb_max);
        }
    }

    if (spec->len != 0) {
        g_string_truncate(spec, spec->len - 1);
        result = g_string_free(spec, FALSE);
    }
    else {
        result = NULL;
    }
    return result;
}

static gchar *
balant_nodes(const gboolean *axes, const sdmhdr_main_hdr_t *hdr,
             CobdfBincomp bincomp, CobdfAxisName bal_or_ant)
{
    gchar *result;
    guint num_bb = num_basebands(hdr->data_struct->basebands);
    g_return_val_if_fail(num_bb > 0, NULL);

    if (axes[COBDF_BASEBAND_AXIS]) {
        gchar *nodes =
            baseband_nodes(axes, hdr, bincomp, bal_or_ant, num_bb);
        if (nodes != NULL) {
            result = g_strdup_printf("{%u,[%s]}", num_bb, nodes);
            g_free(nodes);
        }
        else {
            result = g_strdup_printf("{%u}",
                                     node_multiplier(bincomp) * num_bb);
        }
    }
    else {
        sdmhdr_data_struct_t *ds = hdr->data_struct;
        sdmhdr_bb_t **real_basebands = ds->basebands;
        sdmhdr_bb_t *max_bb[1];
        max_bb[0] =
            maximal_baseband((const sdmhdr_bb_t **)real_basebands, bincomp);
        ds->basebands = max_bb;
        result = baseband_nodes(axes, hdr, bincomp, bal_or_ant, 1);
        g_free(max_bb[0]);
        ds->basebands = real_basebands;
    }
    return result;
}

static sdmhdr_bb_t *
maximal_baseband(const sdmhdr_bb_t **basebands, CobdfBincomp bincomp)
{
    g_return_val_if_fail(basebands != NULL, NULL);

    guint max_num_sw = 0;
    const sdmhdr_bb_t **bbs = basebands;
    while (*bbs != NULL) {
        if ((*bbs)->num_sw > max_num_sw) {
            max_num_sw = (*bbs)->num_sw;
        }
        ++bbs;
    }
    sdmhdr_bb_t *result = g_malloc0(SIZEOF_SDMHDR_BB(max_num_sw));
    result->num_sw = max_num_sw;
    sdmhdr_sw_t *max_sw = result->sw;
    guint i;
    for (i = 0; i < max_num_sw; ++i) {
        max_sw[i].cross_pol_products = SDMHDR_XPP_INVALID;
        max_sw[i].sd_pol_products = SDMHDR_APP_INVALID;
    }
    bbs = basebands;
    while (*bbs != NULL) {
        for (i = 0; i < (*bbs)->num_sw; ++i) {
            const sdmhdr_sw_t *sw = &((*bbs)->sw[i]);
            if (sw->num_bin > max_sw[i].num_bin) {
                max_sw[i].num_bin = sw->num_bin;
            }
            if (sw->num_spectral_point > max_sw[i].num_spectral_point) {
                max_sw[i].num_spectral_point = sw->num_spectral_point;
            }
            if (num_cross_primitives(sw->cross_pol_products, bincomp)
                > num_cross_primitives(max_sw[i].cross_pol_products,
                                       bincomp)) {
                max_sw[i].cross_pol_products = sw->cross_pol_products;
            }
            if (num_auto_primitives(sw->sd_pol_products, bincomp)
                > num_auto_primitives(max_sw[i].sd_pol_products, bincomp)) {
                max_sw[i].sd_pol_products = sw->sd_pol_products;
            }
        }
        ++bbs;
    }
    return result;
}

static void
maximal_spectral_window(const sdmhdr_bb_t *baseband, sdmhdr_sw_t *max_sw,
                        CobdfBincomp bincomp)
{
    g_return_if_fail(baseband != NULL);
    g_return_if_fail(max_sw != NULL);

    guint i;
    for (i = 0; i < baseband->num_sw; ++i) {
        const sdmhdr_sw_t *sw = &(baseband->sw[i]);
        if (sw->num_bin > max_sw->num_bin) {
            max_sw->num_bin = sw->num_bin;
        }
        if (sw->num_spectral_point > max_sw->num_spectral_point) {
            max_sw->num_spectral_point = sw->num_spectral_point;
        }
        if (num_cross_primitives(sw->cross_pol_products, bincomp)
            > num_cross_primitives(max_sw->cross_pol_products, bincomp)) {
            max_sw->cross_pol_products = sw->cross_pol_products;
        }
        if (num_auto_primitives(sw->sd_pol_products, bincomp)
            > num_auto_primitives(max_sw->sd_pol_products, bincomp)) {
            max_sw->sd_pol_products = sw->sd_pol_products;
        }
    }
}

/**
 * alloc_subset_metadata:
 * @we_bdf: we_bdf_t instance
 * @offset: stream offset to start of sdmDataSubsetHeader
 * @bdy: MIME boundary string in data subset
 * @subset: subset header
 *
 * Allocate and initialize a #subset_metadata_t instance based on the
 * binary parts following the current stream position. It is assumed
 * that we_bdf stream has been locked by the caller.
 */
static subset_metadata_t *
alloc_subset_metadata(const we_bdf_t *we_bdf, off_t offset,
                      const gchar *bdy, sdmhdr_subset_hdr_t *subset)
{
    g_return_val_if_fail(we_bdf != NULL, NULL);
    g_return_val_if_fail(we_bdf->main_hdr != NULL, NULL);
    g_return_val_if_fail(we_bdf->main_hdr->data_struct != NULL, NULL);
    g_return_val_if_fail(subset != NULL, NULL);
    g_return_val_if_fail(bdy != NULL, NULL);

    subset_metadata_t *result = g_slice_new0(subset_metadata_t);
    result->header = subset;
    result->header_offset = offset;

    // calculate the sizes of the binary parts
    gsize actual_times_size = 0;
    gsize actual_durations_size = 0;
    gsize zero_lags_size = 0;
    gsize flags_size = 0;
    gsize cross_data_size = 0;
    gsize auto_data_size = 0;
    gsize weights_size = 0;
    sdmhdr_data_struct_t *ds = we_bdf->main_hdr->data_struct;
    if (subset->actual_times_href != NULL && ds->actual_times != NULL) {
        actual_times_size =
            ds->actual_times->size * sizeof(cobdf_actual_times_t);
    }
    if (subset->actual_durations_href != NULL
        && ds->actual_durations != NULL) {
        actual_durations_size = ds->actual_durations->size
            * sizeof(cobdf_actual_durations_t);
    }
    if (subset->zero_lags_href != NULL && ds->zero_lags != NULL) {
        zero_lags_size = ds->zero_lags->size * sizeof(cobdf_zero_lags_t);
    }
    if (subset->flags_href != NULL && ds->flags != NULL) {
        flags_size = ds->flags->size * sizeof(cobdf_flags_t);
    }
    if (subset->cross_data_href != NULL && ds->cross_data != NULL) {
        guint primsize;
        switch (subset->cross_data_type) {
          case COBDF_INT16:
            primsize = 2;
            break;
#ifdef USE_ALMA_5_1_0_5_XML_WORKAROUNDS
          case COBDF_INT:
          case COBDF_FLOAT:
#endif
          case COBDF_INT32:
          case COBDF_FLOAT32:
            primsize = 4;
            break;
          default:
            // cross_data_type could have value COBDF_INT64
            // because of its type, although that value is not
            // supported for crossData; size will remain 0
            primsize = 0;
            break;
        }
        cross_data_size = ds->cross_data->size * primsize;
    }
    if (subset->auto_data_href != NULL && ds->auto_data != NULL) {
        auto_data_size =
            ds->auto_data->size * sizeof(cobdf_auto_data_real_t);
    }
    if (subset->weights_href != NULL && ds->weights != NULL) {
        guint nbits = weights_nbits(ds->weights);
        gsize total_bits = ds->weights->common.size * nbits;
        weights_size = total_bits / 8;
        if (total_bits % 8 != 0) {
            weights_size++;
        }
    }

    // now scan for the offsets of the binary parts
    result->actual_times_offset = -1;
    result->actual_durations_offset = -1;
    result->zero_lags_offset = -1;
    result->flags_offset = -1;
    result->cross_data_offset = -1;
    result->auto_data_offset = -1;
    result->weights_offset = -1;

    while (next_part(we_bdf->file, bdy) == 0) {
        GHashTable *mime_hdrs = read_mime_headers(we_bdf->file);
        const gchar *cloc = g_hash_table_lookup(mime_hdrs, "content-location");
        if (cloc != NULL) {
            off_t *offset = NULL;
            gsize size = 0;
            if (subset->actual_times_href != NULL
                && strcmp(cloc, subset->actual_times_href) == 0) {
                offset = &(result->actual_times_offset);
                size = actual_times_size;
            }
            else if (subset->actual_durations_href != NULL
                     && strcmp(cloc, subset->actual_durations_href) == 0) {
                offset = &(result->actual_durations_offset);
                size = actual_durations_size;
            }
            else if (subset->zero_lags_href != NULL
                     && strcmp(cloc, subset->zero_lags_href) == 0) {
                offset = &(result->zero_lags_offset);
                size = zero_lags_size;
            }
            else if (subset->flags_href != NULL
                     && strcmp(cloc, subset->flags_href) == 0) {
                offset = &(result->flags_offset);
                size = flags_size;
            }
            else if (subset->cross_data_href != NULL
                     && strcmp(cloc, subset->cross_data_href) == 0) {
                offset = &(result->cross_data_offset);
                size = cross_data_size;
            }
            else if (subset->auto_data_href != NULL
                     && strcmp(cloc, subset->auto_data_href) == 0) {
                offset = &(result->auto_data_offset);
                size = auto_data_size;
            }
            else if (subset->weights_href != NULL
                     && strcmp(cloc, subset->weights_href) == 0) {
                offset = &(result->weights_offset);
                size = weights_size;
            }
            if (offset != NULL) {
                *offset = ftello(we_bdf->file);
                if (*offset == -1) {
                    g_critical("Failed to get position of BDF stream: %s",
                               g_strerror(errno));
                }
                gint rc = fseeko(we_bdf->file, size, SEEK_CUR);
                if (rc != 0) {
                    g_critical("Failed to set position of BDF stream "
                               "to expected end of binary component: %s",
                               g_strerror(errno));
                }
                // read next line to ensure that we're at a MIME
                // boundary
                off_t pos = ftello(we_bdf->file);
                if (pos != -1) {
                    rc = check_boundary(we_bdf->file, bdy);
                    if (rc != BDY_OK && rc != BDY_END) {
                        g_info("Expected MIME boundary string not found "
                               "at end of binary component %s (offset: %ld, "
                               "size: %zu, end: %ld",
                               cloc, *offset, size, *offset + size);
                    }
                    rc = fseeko(we_bdf->file, pos + CRLF_LEN, SEEK_SET);
                    if (rc != 0) {
                        g_critical("Failed to reset position of BDF stream "
                                   "after reading ahead for MIME "
                                   "boundary: %s", g_strerror(errno));
                    }
                }
                else {
                    g_critical("Failed to get position of BDF stream: %s",
                               g_strerror(errno));
                }
            }
            else {
                g_message("Skipping MIME part with 'content-location: %s'",
                          cloc);
            }
        }
        else {
            g_message("Failed to find 'content-location' header in part "
                      "of BDF stream");
        }
        g_hash_table_destroy(mime_hdrs);
    }
    return result;
}

static void
destroy_subset_metadata(gpointer data)
{
    subset_metadata_t *metadata = data;
    if (metadata != NULL) {
        if (metadata->header != NULL) {
            sdmhdr_subset_hdr_free(metadata->header);
        }
        g_slice_free(subset_metadata_t, metadata);
    }
}

static guint
weights_nbits(sdmhdr_weights_t *weights)
{
    g_return_val_if_fail(weights != NULL, 0);
    g_return_val_if_fail(weights->num_weights > 0, 0);

    guint result = 0;
    guint num_weights = weights->num_weights;
    while (num_weights > 0) {
        ++result;
        num_weights >>= 1;
    }
    if (weights->num_weights == (1 << (result - 1))) {
        --result;
    }
    return result;
}

static void
crlfify(gchar **str)
{
#ifndef USE_ALMA_5_1_0_5_MIME_WORKAROUNDS
    GString *crlf_str = g_string_sized_new(strlen(*str) + 40);
    gchar *ch = *str;
    while (*ch != '\0') {
        if (*ch == '\n' && (ch == *str || *(ch - 1) != '\r')) {
            g_string_append(crlf_str, CRLF);
        }
        else {
            g_string_append_c(crlf_str, *ch);
        }
        ++ch;
    }
    g_free(*str);
    *str = g_string_free(crlf_str, FALSE);
#endif
}
