#ifndef WCBE_H_
#define WCBE_H_

#include <string.h>
#include <complex.h>
#include <glib.h>
#include <sys/socket.h>
#include <gst/gst.h>

#include <wcbe_lag_frame.h>
#include <wcbe_config.h>
#include <wcbe_util.h>

G_BEGIN_DECLS

// TODO: the following definition should come from "config.h"
#define TERM_VARRAY(type, arr) type arr[]

#define WCBE_WIDAR_FRAME_MEDIA_TYPE "application/x.widar.frame"

#define WCBE_FRAME_MEDIA_TYPE "application/x.wcbe.frame"

#define WCBE_VECTOR_MEDIA_TYPE "application/x.wcbe.vector"
#define WCBE_VECTOR_LENGTH_PROPERTY "length"
#define WCBE_VECTOR_LENGTH_PROPERTY_TYPE "(guint)"

#define WCBE_ARRAY_MEDIA_TYPE "application/x.wcbe.array"
#define WCBE_ARRAY_NUM_ROWS_PROPERTY "rows"
#define WCBE_ARRAY_NUM_ROWS_PROPERTY_TYPE "(guint)"
#define WCBE_ARRAY_NUM_COLS_PROPERTY "cols"
#define WCBE_ARRAY_NUM_COLS_PROPERTY_TYPE "(guint)"
#define WCBE_ARRAY_ROW_LENGTH_PROPERTY "rowlen"
#define WCBE_ARRAY_ROW_LENGTH_PROPERTY_TYPE "(guint)"

#define WCBE_ALIGNMENT_PROPERTY "alignment"
#define WCBE_ALIGNMENT_PROPERTY_TYPE "(guint)"
#define WCBE_PARTIAL_PRODUCT_PROPERTY "partprod"
#define WCBE_PARTIAL_PRODUCT_PROPERTY_TYPE "(guint)"
#define WCBE_MM_PARTIAL_PRODUCT 0
#define WCBE_ML_PARTIAL_PRODUCT 1
#define WCBE_LM_PARTIAL_PRODUCT 2
#define WCBE_LL_PARTIAL_PRODUCT 3
#define WCBE_ELEMENT_TYPE_PROPERTY "type"
#define WCBE_ELEMENT_TYPE_PROPERTY_TYPE "(string)"
#define WCBE_PRODUCT_ID_PROPERTY "product"
#define WCBE_PRODUCT_ID_PROPERTY_TYPE "(guint)"
#define WCBE_SUBBAND_INDEX_PROPERTY "sb"
#define WCBE_SUBBAND_INDEX_PROPERTY_TYPE "(guint)"
#define WCBE_FRAME_SIZE_PROPERTY "framesz" // num complex lags per frame
#define WCBE_FRAME_SIZE_PROPERTY_TYPE "(guint)"
#define WCBE_NUM_BITS_PROPERTY "cbits"
#define WCBE_NUM_BITS_PROPERTY_TYPE "(guint)"
#define WCBE_NUM_BINS_PROPERTY "bins"
#define WCBE_NUM_BINS_PROPERTY_TYPE "(guint)"
#define WCBE_EPOCH_PROPERTY "epoch"
#define WCBE_EPOCH_PROPERTY_TYPE "(guint)"
#define WCBE_SECONDS_PROPERTY "sec"
#define WCBE_SECONDS_PROPERTY_TYPE "(guint)"
#define WCBE_CYCLES_PROPERTY "cyc"
#define WCBE_CYCLES_PROPERTY_TYPE "(guint)"
#define WCBE_STATION_IDS_PROPERTY "stns"
#define WCBE_STATION_IDS_PROPERTY_TYPE "(guint)"
#define WCBE_BASEBAND_IDS_PROPERTY "bbs"
#define WCBE_BASEBAND_IDS_PROPERTY_TYPE "(guint)"

#define WIDAR_CLOCK_RATE 128000000

#define TIMEVALUE(ts) ((ts)->seconds * (guint64)WIDAR_CLOCK_RATE + \
                       (ts)->clock_cycles)
// The following hard codes the WIDAR clock rate, but I want to avoid
// floating point arithmetic, and integer arithmetic overflow.
#if WIDAR_CLOCK_RATE == 128000000
#  define WIDAR_TIME_TO_NS(t) ((((t) + 64) / 128) * 1000)
#else
# error "Unsupported WIDAR_CLOCK_RATE value; please change WIDAR_TIME_TO_NS macro"
#endif

#define MAX_LAGS_PER_FRAME 128
#define MAX_LAGSET_PRODUCT_ID_LEN 80

#define MAX_LAGSET_DELAY 3

#define WCBE_EVENT_ACTIVATE_CONFIG \
  GST_EVENT_MAKE_TYPE(75, GST_EVENT_TYPE_DOWNSTREAM \
                          | GST_EVENT_TYPE_SERIALIZED)
#define WCBE_EVENT_DEACTIVATE_CONFIG \
  GST_EVENT_MAKE_TYPE(76, GST_EVENT_TYPE_DOWNSTREAM \
                          | GST_EVENT_TYPE_SERIALIZED)
#define WCBE_EVENT_NEW_SUBSCAN \
  GST_EVENT_MAKE_TYPE(77, GST_EVENT_TYPE_DOWNSTREAM)
#define WCBE_EVENT_CANCEL_SUBSCAN \
  GST_EVENT_MAKE_TYPE(78, GST_EVENT_TYPE_DOWNSTREAM)

typedef struct wcbe_stream_object {
    enum {WCBE_STREAM_LAGSET, WCBE_STREAM_EVENT, WCBE_STREAM_MSG} type;
} wcbe_stream_object_t;

typedef struct wcbe_stream_event {
    wcbe_stream_object_t object;
    GstEventType event;
    guint structure_len;
    gchar structure[0];
} wcbe_stream_event_t;

typedef struct wcbe_stream_lagset {
    wcbe_stream_object_t object;
    gchar product_id[MAX_LAGSET_PRODUCT_ID_LEN];
    wcbe_vect_t lagset;
} wcbe_stream_lagset_t;

typedef struct wcbe_stream_msg {
    wcbe_stream_object_t object;
    guint structure_len;
    gchar structure[0];
} wcbe_stream_msg_t;

#define ELEMENT_LIST_MSG_NAME "element-list"
#define PROPERTY_LIST_MSG_NAME "property-list"
#define GET_PROPERTY_MSG_NAME "get-property"
#define SET_PROPERTY_MSG_NAME "set-property"
#define PAD_LIST_MSG_NAME "pad-list"
#define PAD_CAPS_MSG_NAME "pad-caps"
#define PAD_PEER_MSG_NAME "pad-peer"

static inline void wcbe_stream_lagset_init_iovec(
    wcbe_stream_lagset_t *object, wcbe_vect_t *lagset, gsize lagset_size,
    struct iovec iovec[2])
{
    g_assert(offsetof(wcbe_stream_lagset_t, object) == 0);
    iovec[0].iov_base = object;
    iovec[0].iov_len = offsetof(wcbe_stream_lagset_t, lagset);
    iovec[1].iov_base = lagset;
    iovec[1].iov_len = lagset_size;
}

typedef struct lag_spectra_t {
    guint num_lags;
    guint num_bins;
    guint64 *timestamps;
    float complex *lags;
    float *weights;
    TERM_VARRAY(guint8, data);
} lag_spectra_t;
#define SIZEOF_LAG_SPECTRA_T(nlag, nbin) \
  (sizeof(lag_spectra_t) + (nbin) * \
   ((nlag) * (sizeof(float complex) + sizeof(float)) + sizeof(guint64)))

static inline void lag_spectra_init(lag_spectra_t *spectra, guint num_lags,
                                    guint num_bins)
{
    memset(spectra, 0, SIZEOF_LAG_SPECTRA_T(num_lags, num_bins));
    spectra->num_bins = num_bins;
    spectra->num_lags = num_lags;
    spectra->timestamps = (guint64 *)spectra->data;
    spectra->lags = (float complex *)&(spectra->timestamps[num_bins]);
    spectra->weights = (float *)&(spectra->lags[num_bins * num_lags]);
}

typedef struct activation {
    lf_widar_time_t time;
    wcbe_config_t *config;
    gboolean in_use;
    gboolean deactivated;
    gboolean now;
} activation_t;


G_END_DECLS

#endif /* WCBE_H_ */
