#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <wcbe_accumulate.h>
#include <wcbe.h>
#include <user_marshal.h>

#include <glib-object.h>
#include <string.h>
#include <errno.h>

GST_DEBUG_CATEGORY_STATIC(wcbe_accumulate_debug);
#define GST_CAT_DEFAULT wcbe_accumulate_debug

static GQuark element_type_int32;
static GQuark element_type_uint32;
static GQuark element_type_int64;
static GQuark element_type_uint64;
static GQuark element_type_float;

typedef struct input_product {
    GQuark id;
    GstPad *pad;
} input_product_t;

typedef struct output_product {
    GstPad *pad;
    GQuark type;
    GstBuffer *acc;
    guint count;
    guint timeouts_remaining;
} output_product_t;
#define MAX_TIMEOUTS 3

/* Filter signals and args */
enum
{
    SIGNAL_GET_BE_INTEGRATION,
    LAST_SIGNAL
};
static guint wcbe_accumulate_signals[LAST_SIGNAL] = {0,};

enum
{
    ARG_0,
    ARG_INTERVAL,
    ARG_BUFFERS_IN,
    ARG_BUFFERS_OUT
};
// ARG_INTERVAL *property* is in microseconds
#define NSEC_PER_USEC 1000
#define DEFAULT_INTERVAL_NS (1000000uL * NSEC_PER_USEC)

#define SINK_PADS_BASE_NAME "sink"
#define SRC_PADS_BASE_NAME "src"
static const gchar sink_pads_base_name[] = SINK_PADS_BASE_NAME;
static const gchar src_pads_base_name[] = SRC_PADS_BASE_NAME;

#define ACCUMULATE_STATIC_CAPS                    \
     WCBE_VECTOR_MEDIA_TYPE                       \
     ","WCBE_ELEMENT_TYPE_PROPERTY"="             \
        WCBE_ELEMENT_TYPE_PROPERTY_TYPE           \
        "{gfloat,guint32,gint32,guint64,gint64}"  \
  ";"WCBE_ARRAY_MEDIA_TYPE                        \
     ","WCBE_ELEMENT_TYPE_PROPERTY"="             \
        WCBE_ELEMENT_TYPE_PROPERTY_TYPE           \
        "{gfloat,guint32,gint32,guint64,gint64}"

static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE(
    SINK_PADS_BASE_NAME"%d",
    GST_PAD_SINK,
    GST_PAD_REQUEST,
    GST_STATIC_CAPS(ACCUMULATE_STATIC_CAPS));

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE(
    SRC_PADS_BASE_NAME"%d",
    GST_PAD_SRC,
    GST_PAD_SOMETIMES,
    GST_STATIC_CAPS(ACCUMULATE_STATIC_CAPS));

GST_BOILERPLATE(WcbeAccumulate, wcbe_accumulate, GstElement,
                GST_TYPE_ELEMENT);

static void wcbe_accumulate_finalize(GObject *object);
static void wcbe_accumulate_get_property(
    GObject *object, guint prop_id, GValue *value, GParamSpec *pspec);
static void wcbe_accumulate_set_property(
    GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec);
static GstStateChangeReturn wcbe_accumulate_change_state(
    GstElement *element, GstStateChange transition);
static GstPad *wcbe_accumulate_request_new_pad(
    GstElement *element, GstPadTemplate *templ, const gchar* name);
static void wcbe_accumulate_release_pad(GstElement *element, GstPad *pad);
static GstFlowReturn wcbe_accumulate_chain(GstPad *pad, GstBuffer *buffer);
static gboolean wcbe_accumulate_event(GstPad *pad, GstEvent *event);
static GList *wcbe_accumulate_intlink(GstPad *pad);
//static gboolean wcbe_accumulate_setcaps(GstPad *pad, GstCaps *caps);
static GstBuffer *alloc_new_output_buffer(
    input_product_t *ip, GstBuffer *input_buffer, output_product_t *op,
    guint *rowlen);
static void accumulate_vals(GstBuffer *acc_buffer, GstBuffer *in_buffer,
                            GQuark in_type);
static guint input_product_hash(gconstpointer key);
static gboolean input_product_equal(gconstpointer a, gconstpointer b);
static void output_product_free(gpointer value);
static gpointer sweep_fn(gpointer data);
static void flush_output_buffers(WcbeAccumulate *accumulate);

static void
wcbe_accumulate_base_init(gpointer gclass)
{
    GST_DEBUG_CATEGORY_INIT(wcbe_accumulate_debug,
                            "wcbe_accumulate", 0,
                            "wcbe_accumulate debug");
    static GstElementDetails element_details = {
        "WIDAR vector/array accumulator",
        "Filter/WIDAR",
        "Accumulate vectors or arrays over some time interval",
        "Martin Pokorny <mpokorny@nrao.edu>"
    };
    GstElementClass *element_class = GST_ELEMENT_CLASS(gclass);

    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&sink_factory));
    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&src_factory));
    gst_element_class_set_details(element_class, &element_details);

    element_type_int32 = g_quark_from_string("gint32");
    element_type_uint32 = g_quark_from_string("guint32");
    element_type_int64 = g_quark_from_string("gint64");
    element_type_uint64 = g_quark_from_string("guint64");
    element_type_float = g_quark_from_string("gfloat");
}

/* initialize the plugin's class */
static void
wcbe_accumulate_class_init(WcbeAccumulateClass *klass)
{
    GObjectClass *gobject_class  = (GObjectClass *)klass;
    GstElementClass *elementclass = GST_ELEMENT_CLASS(klass);

    gobject_class->set_property =
        GST_DEBUG_FUNCPTR(wcbe_accumulate_set_property);
    gobject_class->get_property =
        GST_DEBUG_FUNCPTR(wcbe_accumulate_get_property);
    gobject_class->finalize =
        GST_DEBUG_FUNCPTR(wcbe_accumulate_finalize);

    g_object_class_install_property(
        gobject_class, ARG_INTERVAL,
        g_param_spec_uint("interval", "Accumulation interval",
                          "Time period over which to accumulate samples, "
                          "in microseconds.",
                          10, G_MAXUINT,
                          DEFAULT_INTERVAL_NS / NSEC_PER_USEC,
                          G_PARAM_READABLE));

    g_object_class_install_property(
        gobject_class, ARG_BUFFERS_IN,
        g_param_spec_uint("num-in", "Number input buffers",
                          "Total number of buffers received",
                          0, G_MAXUINT,
                          0, G_PARAM_READABLE));

    g_object_class_install_property(
        gobject_class, ARG_BUFFERS_OUT,
        g_param_spec_uint("num-out", "Number output buffers",
                          "Total number of buffers sent",
                          0, G_MAXUINT,
                          0, G_PARAM_READABLE));

    elementclass->request_new_pad =
        GST_DEBUG_FUNCPTR(wcbe_accumulate_request_new_pad);
    elementclass->release_pad =
        GST_DEBUG_FUNCPTR(wcbe_accumulate_release_pad);
    elementclass->change_state =
        GST_DEBUG_FUNCPTR(wcbe_accumulate_change_state);

    wcbe_accumulate_signals[SIGNAL_GET_BE_INTEGRATION] =
        g_signal_new("get-be-integration", G_TYPE_FROM_CLASS(klass),
                     G_SIGNAL_RUN_LAST, 0, NULL, NULL,
                     g_cclosure_user_marshal_UINT__VOID,
                     G_TYPE_UINT, 0);
}

static void
wcbe_accumulate_init(WcbeAccumulate *accumulate, WcbeAccumulateClass *klass)
{
    accumulate->src_pads = g_ptr_array_new();
    accumulate->sink_pads = g_ptr_array_new();
    accumulate->products = g_hash_table_new_full(
        input_product_hash, input_product_equal, g_free, output_product_free);
    g_static_mutex_init(&(accumulate->mtx));
    accumulate->interval = DEFAULT_INTERVAL_NS;
    accumulate->t0 = 0;
    GError *err = NULL;
    accumulate->sweeper = (GThread *)1;
    accumulate->sweeper = g_thread_create(sweep_fn, accumulate, TRUE, &err);
    if (err != NULL) {
        g_critical("Failed to spawn WcbeAccumulate 'sweeper' thread: %s",
                   err->message);
    }
}

static void
wcbe_accumulate_finalize(GObject *object)
{
    WcbeAccumulate *accumulate = WCBE_ACCUMULATE(object);

    GThread *sweeper = accumulate->sweeper;
    if (sweeper != NULL) {
        accumulate->sweeper = NULL;
        g_thread_join(sweeper);
    }

    g_hash_table_destroy(accumulate->products);

    guint i;
    for (i = 0; i < accumulate->src_pads->len; ++i) {
        GstPad *pad = g_ptr_array_index(accumulate->src_pads, i);
        gst_object_unref(GST_OBJECT(pad));
    }
    g_ptr_array_free(accumulate->src_pads, TRUE);

    for (i = 0; i < accumulate->sink_pads->len; ++i) {
        GstPad *pad = g_ptr_array_index(accumulate->sink_pads, i);
        gst_object_unref(pad);
    }
    g_ptr_array_free(accumulate->sink_pads, TRUE);
    g_static_mutex_free(&(accumulate->mtx));
}

static void
wcbe_accumulate_get_property(GObject *object, guint prop_id, GValue *value,
                             GParamSpec *pspec)
{
    WcbeAccumulate *accumulate = WCBE_ACCUMULATE(object);
    switch (prop_id) {
      case ARG_INTERVAL:
        g_value_set_uint(value, accumulate->interval / NSEC_PER_USEC);
        break;
      case ARG_BUFFERS_IN:
        g_value_set_uint(value,
                         g_atomic_int_get((gint *)&(accumulate->num_in)));
        break;
      case ARG_BUFFERS_OUT:
        g_value_set_uint(value,
                         g_atomic_int_get((gint *)&(accumulate->num_out)));
        break;
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static void
wcbe_accumulate_set_property(GObject *object, guint prop_id,
                             const GValue *value, GParamSpec *pspec)
{
    switch (prop_id) {
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static GstStateChangeReturn
wcbe_accumulate_change_state(GstElement *element, GstStateChange transition)
{
    WcbeAccumulate *acc = WCBE_ACCUMULATE(element);
    switch (transition) {
      case GST_STATE_CHANGE_NULL_TO_READY: {
          guint interval = DEFAULT_INTERVAL_NS / NSEC_PER_USEC;
          g_signal_emit(
              acc, wcbe_accumulate_signals[SIGNAL_GET_BE_INTEGRATION],
              0, &interval);
          acc->interval = interval * NSEC_PER_USEC;
          break;
      }
      default:
        break;
    }
    return GST_ELEMENT_CLASS(parent_class)->change_state(element, transition);
}

static GstFlowReturn
wcbe_accumulate_chain(GstPad *pad, GstBuffer *buffer)
{
    GstFlowReturn result = GST_FLOW_OK;
    WcbeAccumulate *accumulate = WCBE_ACCUMULATE(GST_OBJECT_PARENT(pad));

    g_atomic_int_inc((gint *)&(accumulate->num_in));

    // create a hash table key for the data in the buffer on the given
    // pad
    GstStructure *s = gst_caps_get_structure(GST_BUFFER_CAPS(buffer), 0);
    input_product_t ip = {.pad = pad};
    gboolean rc =
        gst_structure_get_uint(s, WCBE_PRODUCT_ID_PROPERTY, &(ip.id));
    const gchar *iptype_str =
        gst_structure_get_string(s, WCBE_ELEMENT_TYPE_PROPERTY);
    if (!rc || iptype_str == NULL) {
        g_debug("Buffer product id or element type missing");
        return GST_FLOW_ERROR;
    }
    GQuark iptype = g_quark_try_string(iptype_str);

    // find corresponding output_product_t instance, create a new one
    // if not found
    g_static_mutex_lock(&(accumulate->mtx));
    output_product_t *op = g_hash_table_lookup(accumulate->products, &ip);
    if (op == NULL) {
        // need dynamically allocated input_product_t
        input_product_t *ip2 = g_memdup(&ip, sizeof(ip));
        // create matching output_product_t instance
        op = g_new0(output_product_t, 1);
        // find matching src pad
        guint i;
        for (i = 0; op->pad == NULL && i < accumulate->sink_pads->len; ++i) {
            if (g_ptr_array_index(accumulate->sink_pads, i) == pad) {
                op->pad = g_ptr_array_index(accumulate->src_pads, i);
            }
        }
        // get data type of input, choose data type for output
        if (iptype == element_type_int32) {
            op->type = element_type_int64;
        }
        else if (iptype == element_type_uint32) {
            op->type = element_type_uint64;
        }
        else if (iptype == element_type_float) {
            op->type = element_type_float;
        }
        else {
            g_debug("Invalid element type");
            g_free(op);
            op = NULL;
            result = GST_FLOW_ERROR;
        }
        g_hash_table_insert(accumulate->products, ip2, op);
    }
    GstBuffer *op_buffer = NULL;
    if (op != NULL) {
        // update output_product_t instance, primarily so that the sweeper
        // thread doesn't interfere
        op_buffer = op->acc;
        op->acc = NULL;
    }
    g_static_mutex_unlock(&(accumulate->mtx));
    if (op == NULL) {
        return result;
    }

    // Push out last accumulation buffer if timestamp of current
    // buffer indicates accumulation time has been met.
    if (op_buffer != NULL
        && (GST_BUFFER_TIMESTAMP(buffer) - GST_BUFFER_TIMESTAMP(op_buffer) >=
            accumulate->interval)) {
        gst_pad_set_caps(op->pad, GST_BUFFER_CAPS(op_buffer));
        result = gst_pad_push(op->pad, op_buffer);
        op_buffer = NULL;
        g_atomic_int_inc((gint *)&(accumulate->num_out));
        op->count = 0;
        if (result != GST_FLOW_OK) {
            return result;
        }
    }

    // get a buffer for accumulation if needed
    gboolean do_accumulate = TRUE;
    if (op_buffer == NULL) {
        guint rowlen = 0;
        if (iptype != op->type) {
            op_buffer = alloc_new_output_buffer(&ip, buffer, op, &rowlen);
            op->count = 0;
        }
        else {
            op_buffer = buffer;
            gst_buffer_ref(op_buffer);
            do_accumulate = FALSE;
            op->count = 1;
        }
        GstCaps *op_caps = gst_caps_copy(GST_BUFFER_CAPS(buffer));
        gst_caps_set_simple(
            op_caps,
            WCBE_ELEMENT_TYPE_PROPERTY, G_TYPE_STRING,
            g_quark_to_string(op->type),
            NULL);
        if (rowlen != 0) {
            gst_caps_set_simple(
                op_caps,
                WCBE_ARRAY_ROW_LENGTH_PROPERTY, G_TYPE_UINT, rowlen,
                NULL);
        }
        gst_buffer_set_caps(op_buffer, op_caps);
        gst_caps_unref(op_caps);
        // set buffer timestamp to beginning of the integration period
        GST_BUFFER_TIMESTAMP(op_buffer) =
            ((GST_BUFFER_TIMESTAMP(buffer) - accumulate->t0)
             / accumulate->interval) * accumulate->interval + accumulate->t0;
    }

    // do accumulation
    if (do_accumulate) {
        accumulate_vals(op_buffer, buffer, iptype);
        op->count++;
    }
    gst_buffer_unref(buffer);

    // update the output_product_t instance
    g_static_mutex_lock(&(accumulate->mtx));
    op->acc = op_buffer;
    op->timeouts_remaining = MAX_TIMEOUTS;
    g_static_mutex_unlock(&(accumulate->mtx));

    return GST_FLOW_OK;
}

static gboolean
wcbe_accumulate_event(GstPad *pad, GstEvent *event)
{
    gboolean result = FALSE;

    WcbeAccumulate *accumulate = WCBE_ACCUMULATE(GST_OBJECT_PARENT(pad));
    switch (GST_EVENT_TYPE(event)) {
      case GST_EVENT_NEWSEGMENT: {
          guint64 start;
          gst_event_parse_new_segment(event, NULL, NULL, NULL,
                                      (gint64 *)&start, NULL, NULL);
          g_static_mutex_lock(&(accumulate->mtx));
          if (start != accumulate->t0) {
              // flush *all* output buffers since t0 must change
              flush_output_buffers(accumulate);
              accumulate->t0 = start;
          }
          g_static_mutex_unlock(&(accumulate->mtx));
          result = gst_pad_event_default(pad, event);
          g_atomic_int_set((gint *)&(accumulate->num_in), 0);
          g_atomic_int_set((gint *)&(accumulate->num_out), 0);
          break;
      }
      case GST_EVENT_FLUSH_START:
      case GST_EVENT_EOS:
        g_static_mutex_lock(&(accumulate->mtx));
        flush_output_buffers(accumulate);
        g_static_mutex_unlock(&(accumulate->mtx));
        break;
      default:
        result = gst_pad_event_default(pad, event);
        break;
    }
    return result;
}

static GList *
wcbe_accumulate_intlink(GstPad *pad)
{
    WcbeAccumulate *accumulate = WCBE_ACCUMULATE(GST_OBJECT_PARENT(pad));

    GstPad *otherpad = NULL;
    g_static_mutex_lock(&(accumulate->mtx));
    guint i;
    for (i = 0; otherpad == NULL && i < accumulate->sink_pads->len; ++i) {
        if (g_ptr_array_index(accumulate->sink_pads, i) == pad) {
            otherpad = g_ptr_array_index(accumulate->src_pads, i);
        }
        else if (g_ptr_array_index(accumulate->src_pads, i) == pad) {
            otherpad = g_ptr_array_index(accumulate->sink_pads, i);
        }
    }
    g_static_mutex_unlock(&(accumulate->mtx));
    GList *result = NULL;
    if (otherpad != NULL) {
        result = g_list_append(NULL, otherpad);
    }
    return result;
}

static GstPad *
wcbe_accumulate_request_new_pad(GstElement *element, GstPadTemplate *templ,
                                const gchar* name)
{
    GstPad *result = NULL;
    WcbeAccumulate *accumulate = WCBE_ACCUMULATE(element);
    gchar *padname = NULL;

    // get pad name suffix
    gint pad_index = -1;
    if (name != NULL) {
        if (strncmp(name, sink_pads_base_name,
                    sizeof(sink_pads_base_name) - 1) == 0) {
            guint64 idx = g_ascii_strtoull(
                name + sizeof(sink_pads_base_name) - 1, NULL, 0);
            if (idx != G_MAXUINT64) {
                pad_index = idx;
            }
        }
        padname = g_strdup(name);
    }
    else {
        pad_index = 0;
        while ((pad_index < accumulate->sink_pads->len)
               && (g_ptr_array_index(accumulate->sink_pads, pad_index)
                   != NULL)) {
            ++pad_index;
        }
        padname = g_strdup_printf(sink_factory.name_template, pad_index);
    }

    // check that pad_index is valid and unused, then add the pad
    g_static_mutex_lock(&(accumulate->mtx));
    if (pad_index >= accumulate->sink_pads->len
        || g_ptr_array_index(accumulate->sink_pads, pad_index) == NULL) {
        GstState state;
        gst_element_get_state(GST_ELEMENT(accumulate), &state, NULL, 0);
        gboolean active = (state >= GST_STATE_PAUSED);
        // create sink pad
        GstPad *sinkpad = gst_pad_new_from_template(templ, padname);
        //gst_pad_set_setcaps_function(sinkpad, wcbe_accumulate_setcaps);
        gst_pad_set_chain_function(sinkpad, wcbe_accumulate_chain);
        gst_pad_set_event_function(sinkpad, wcbe_accumulate_event);

        // create corresponding source pad
        gchar *src_name =
            g_strdup_printf(src_factory.name_template, pad_index);
        GstPad *srcpad =
            gst_pad_new_from_static_template(&src_factory, src_name);
        g_free(src_name);
        gst_pad_use_fixed_caps(srcpad);

        // set active state of both pads
        if (gst_pad_set_active(sinkpad, active)
            && gst_pad_set_active(srcpad, active)) {
            // add both pads to element, and keep track in element's
            // sink_pads and src_pads arrays
            gst_element_add_pad(element, sinkpad);
            gst_element_add_pad(element, srcpad);
            if (accumulate->sink_pads->len <= pad_index) {
                g_ptr_array_set_size(accumulate->sink_pads, pad_index + 1);
            }
            g_ptr_array_index(accumulate->sink_pads, pad_index) = sinkpad;
            if (accumulate->src_pads->len <= pad_index) {
                g_ptr_array_set_size(accumulate->src_pads, pad_index + 1);
            }
            g_ptr_array_index(accumulate->src_pads, pad_index) = srcpad;
            gst_pad_set_internal_link_function(sinkpad,
                                               wcbe_accumulate_intlink);
            gst_pad_set_internal_link_function(srcpad,
                                               wcbe_accumulate_intlink);
            result = sinkpad;
        }
        else {
            gst_object_unref(GST_OBJECT(sinkpad));
            gst_object_unref(GST_OBJECT(srcpad));
        }
    }
    g_static_mutex_unlock(&(accumulate->mtx));
    g_free(padname);
    return result;
}

static void
wcbe_accumulate_release_pad(GstElement *element, GstPad *pad)
{
    WcbeAccumulate *accumulate = WCBE_ACCUMULATE(element);

    guint i;
    g_static_mutex_lock(&(accumulate->mtx));
    for (i = 0; pad != NULL && i < accumulate->sink_pads->len; ++i) {
        if (pad == g_ptr_array_index(accumulate->sink_pads, i)) {
            gst_element_remove_pad(element, pad);
            g_ptr_array_index(accumulate->sink_pads, i) = NULL;
            gboolean remove_pad_entry(gpointer key,
                                      gpointer value G_GNUC_UNUSED,
                                      gpointer user_data G_GNUC_UNUSED) {
                input_product_t *ip = key;
                return ((ip->pad == pad) ? TRUE : FALSE);
            }
            g_hash_table_foreach_remove(accumulate->products,
                                        remove_pad_entry, NULL);
            pad = NULL;
            GstPad *srcpad = g_ptr_array_index(accumulate->src_pads, i);
            gst_element_remove_pad(element, srcpad);
            g_ptr_array_index(accumulate->src_pads, i) = NULL;
        }
    }
    g_static_mutex_unlock(&(accumulate->mtx));
}

static GstBuffer *
alloc_new_output_buffer(input_product_t *ip, GstBuffer *input_buffer,
                        output_product_t *op, guint *rowlen)
{
    g_return_val_if_fail(ip != NULL, NULL);
    g_return_val_if_fail(op != NULL, NULL);
    g_return_val_if_fail(op->type == element_type_int64
                         || op->type == element_type_uint64
                         || op->type == element_type_float, NULL);

    GstBuffer *result = NULL;
    *rowlen = 0;
    gsize opsize = 0;
    if (op->type == element_type_int64) {
        opsize = sizeof(gint64);
    }
    else if (op->type == element_type_uint64) {
        opsize = sizeof(guint64);
    }
    else if (op->type == element_type_float) {
        opsize = sizeof(gfloat);
    }
    if (opsize == 0) {
        return NULL;
    }
    GstStructure *s = gst_caps_get_structure(GST_BUFFER_CAPS(input_buffer), 0);
    gsize total_size = 0;
    guint alignment = 0;
    gst_structure_get_uint(s, WCBE_ALIGNMENT_PROPERTY, &alignment);
    if (gst_structure_has_name(s, WCBE_VECTOR_MEDIA_TYPE)) {
        guint length;
        gst_structure_get_uint(s, WCBE_VECTOR_LENGTH_PROPERTY, &length);
        total_size = length * opsize;
    }
    else {
        // gst_structure_has_name(s, WCBE_ARRAY_MEDIA_TYPE)
        guint rows;
        guint cols;
        gst_structure_get_uint(s, WCBE_ARRAY_NUM_ROWS_PROPERTY, &rows);
        gst_structure_get_uint(s, WCBE_ARRAY_NUM_COLS_PROPERTY, &cols);
        guint rowalign;
        if (alignment != 0 && opsize > alignment) {
            if (G_UNLIKELY(opsize % alignment != 0)) {
                g_warning("Array element size (%zu) is incommensurate with "
                          "row alignment (%u)", opsize, alignment);
            }
            rowalign = opsize;
        }
        else {
            rowalign = alignment;
        }
        gsize rowsize;
        if (rowalign != 0 && ((cols * opsize) % rowalign) != 0) {
            rowsize = ((cols * opsize) / rowalign + 1) * rowalign;
        }
        else {
            rowsize = cols * opsize;
        }
        *rowlen = rowsize / opsize;
        total_size = rowsize * rows;
    }
    void *mallocdata = NULL;
    if (alignment != 0) {
        gint rc = posix_memalign(&mallocdata, alignment, total_size);
        if (rc == 0) {
            memset(mallocdata, 0, total_size);
        }
        else {
            g_critical("Failed to allocate memory for "
                       "accumulation vector: %s", g_strerror(errno));
        }
    }
    else {
        mallocdata = g_malloc0(total_size);
    }
    if (mallocdata != NULL) {
        result = gst_buffer_new();
        GST_BUFFER_SIZE(result) = total_size;
        GST_BUFFER_MALLOCDATA(result) = mallocdata;
        GST_BUFFER_DATA(result) = GST_BUFFER_MALLOCDATA(result);
    }
    return result;
}

static
void accumulate_vals(GstBuffer *acc_buffer, GstBuffer *in_buffer,
                     GQuark in_type)
{
    g_return_if_fail(in_type == element_type_int32
                     || in_type == element_type_uint32
                     || in_type == element_type_float);

    GstStructure *s = gst_caps_get_structure(GST_BUFFER_CAPS(in_buffer), 0);

    if (in_type == element_type_int32 || in_type == element_type_uint32) {
        //  we cheat a little here and treat both signed and
        //  unsigned cases with the same code, which should be OK
        //  since we're only doing addition on unsigned or
        //  twos-complement integers
        gint32 *iv = (gint32 *)GST_BUFFER_DATA(in_buffer);
        gint64 *accv = (gint64 *)GST_BUFFER_DATA(acc_buffer);
        if (gst_structure_has_name(s, WCBE_VECTOR_MEDIA_TYPE)){
            guint n;
            gst_structure_get_uint(s, WCBE_VECTOR_LENGTH_PROPERTY, &n);
            while (n-- > 0) {
                *accv++ += *iv++;
            }
        }
        else {
            guint rows;
            guint cols;
            gst_structure_get_uint(s, WCBE_ARRAY_NUM_ROWS_PROPERTY, &rows);
            gst_structure_get_uint(s, WCBE_ARRAY_NUM_COLS_PROPERTY, &cols);
            guint rowlen = cols;
            gst_structure_get_uint(s, WCBE_ARRAY_ROW_LENGTH_PROPERTY, &rowlen);
            guint excess = rowlen - cols;
            while (rows-- > 0) {
                guint n = cols;
                while (n-- > 0) {
                    *accv++ += *iv++;
                }
                accv += excess;
                iv += excess;
            }
        }
    }
    else if (in_type == element_type_float) {
        gfloat *iv = (gfloat *)GST_BUFFER_DATA(in_buffer);
        gfloat *accv = (gfloat *)GST_BUFFER_DATA(acc_buffer);
        if (gst_structure_has_name(s, WCBE_VECTOR_MEDIA_TYPE)){
            guint n;
            gst_structure_get_uint(s, WCBE_VECTOR_LENGTH_PROPERTY, &n);
            while (n-- > 0) {
                *accv++ += *iv++;
            }
        }
        else {
            guint rows;
            guint cols;
            gst_structure_get_uint(s, WCBE_ARRAY_NUM_ROWS_PROPERTY, &rows);
            gst_structure_get_uint(s, WCBE_ARRAY_NUM_COLS_PROPERTY, &cols);
            guint rowlen = cols;
            gst_structure_get_uint(s, WCBE_ARRAY_ROW_LENGTH_PROPERTY, &rowlen);
            guint excess = rowlen - cols;
            while (rows-- > 0) {
                guint n = cols;
                while (n-- > 0) {
                    *accv++ += *iv++;
                }
                accv += excess;
                iv += excess;
            }
        }
    }
}

static guint
input_product_hash(gconstpointer key)
{
    const input_product_t *ip = key;
    return g_int_hash(&(ip->id)) ^ g_direct_hash(ip->pad);
}

static gboolean
input_product_equal(gconstpointer a, gconstpointer b)
{
    const input_product_t *ia = a;
    const input_product_t *ib = b;
    return ((ia->id == ib->id) && (ia->pad == ib->pad));
}

static void
output_product_free(gpointer value)
{
    output_product_t *op = value;
    if (op->acc != NULL) {
        gst_buffer_unref(op->acc);
    }
    g_free(op);
}

static gpointer
sweep_fn(gpointer data)
{
    WcbeAccumulate *accumulate = WCBE_ACCUMULATE(data);

    GPtrArray *old_buffers = g_ptr_array_new();
    void get_old_buffer(gpointer key G_GNUC_UNUSED, gpointer value,
                        gpointer user_data G_GNUC_UNUSED) {
        output_product_t *op = value;
        if (op->acc != NULL && --op->timeouts_remaining == 0) {
            g_ptr_array_add(old_buffers, op->acc);
            op->acc = NULL; // steal the element's reference
            g_ptr_array_add(old_buffers, op->pad);
            gst_object_ref(GST_OBJECT(op->pad));
        }
    }
    while (accumulate->sweeper != NULL) {
        g_static_mutex_lock(&(accumulate->mtx));
        guint64 sleep = accumulate->interval;
        g_static_mutex_unlock(&(accumulate->mtx));
        g_usleep(sleep / NSEC_PER_USEC);
        g_static_mutex_lock(&(accumulate->mtx));
        g_hash_table_foreach(accumulate->products, get_old_buffer, NULL);
        g_static_mutex_unlock(&(accumulate->mtx));
        guint i = 0;
        while (i < old_buffers->len) {
            GstBuffer *buffer = g_ptr_array_index(old_buffers, i++);
            GstPad *pad = g_ptr_array_index(old_buffers, i++);
            gst_pad_set_caps(pad, GST_BUFFER_CAPS(buffer));
            g_debug("pushing timed out buffer");
            gst_pad_push(pad, buffer);
            gst_object_unref(GST_OBJECT(pad));
            g_atomic_int_inc((gint *)&(accumulate->num_out));
        }
        old_buffers->len = 0;
    }
    g_ptr_array_free(old_buffers, TRUE);
    return NULL;
}

static void
flush_output_buffers(WcbeAccumulate *accumulate)
{
    /*
     * NOTE: accumulate->mtx lock should be held by the caller of this
     * function!
     */

    /* In flushing the output buffers, we have to ensure that the
     * synchronization of buffers across src pads is correct. For
     * example, we want to avoid pushing all buffers on one pad and
     * then all buffers on another pad, since this will destroy
     * synchronization downstream. What we do is to go through all of
     * the product ids and push up to one buffer for each id on each
     * of the pads before moving on to another id.
     */
    GSList *buffers = NULL;
    gint cmp_ids(gconstpointer a, gconstpointer b) {
        const input_product_t *ipa = ((const GSList *)a)->data;
        const input_product_t *ipb = b;
        return (gint)ipa->id - (gint)ipb->id;
    }
    void get_buffer(gpointer key, gpointer value,
                    gpointer user_data G_GNUC_UNUSED) {
        input_product_t *ip = key;
        output_product_t *op = value;
        if (op->acc != NULL) {
            GSList *plist = g_slist_find_custom(buffers, ip, cmp_ids);
            if (plist != NULL) {
                g_assert(plist->data != NULL);
                plist = g_slist_append(plist->data, ip);
            }
            else {
                plist = g_slist_append(NULL, ip);
                buffers = g_slist_append(buffers, plist);
            }
        }
    }

    /* Get a list of pending buffers for each product id. */
    g_hash_table_foreach(accumulate->products, get_buffer, NULL);

    /* For each product id in the list, if the number of buffers is
     * equal to the number of src pads, push each of the buffers onto
     * its associated src pad. If the number of buffers is not equal
     * to the number of src pads, we don't push them out, and instead
     * simply unref them. (NOTE: we don't expect the condition in
     * which the number of buffers is neither zero nor equal to the
     * number of source pads, since the upstream element should be
     * keeping synchronization anyway, especially in flush
     * conditions.)
     */
    while (buffers != NULL) {
        GSList *first_plist = buffers->data;
        if (g_slist_length(first_plist) == accumulate->src_pads->len) {
            GSList *plist = first_plist;
            while (plist != NULL) {
                input_product_t *ip = plist->data;
                output_product_t *op =
                    g_hash_table_lookup(accumulate->products, ip);
                gst_pad_set_caps(op->pad, GST_BUFFER_CAPS(op->acc));
                gst_pad_push(op->pad, op->acc);
                op->acc = NULL;
                g_atomic_int_inc((gint *)&(accumulate->num_out));
                plist = g_slist_next(plist);
            }
        }
        else {
            GSList *plist = first_plist;
            while (plist != NULL) {
                input_product_t *ip = plist->data;
                output_product_t *op =
                    g_hash_table_lookup(accumulate->products, ip);
                gst_buffer_unref(op->acc);
                op->acc = NULL;
                plist = g_slist_next(plist);
            }
        }
        buffers = g_slist_remove(buffers, first_plist);
        g_slist_free(first_plist);
    }
}
