#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/epoll.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>

#include <wcbe_bdf_visdata.h>
#include <we_bdf.h>
#include <wcbe.h>
#include <user_marshal.h>
#include <wcbe_util.h>
#include <wcbe_config.h>
#include <cobdf.h>
#include <tree.h>

#ifdef MMAP_BDF_BINARIES
# error "wcbe_bdf_visdata module does not support mmap-ed BDF binaries"
#endif

#define NSEC_PER_SEC 1000000000u

typedef struct subscan_t {
    wcbe_bdf_t *wbdf;
    GStaticRecMutex lock; // none of these locks for subscans in the
                          // subscans list shall be held when
                          // acquiring the subscans list mutex (OK to
                          // acquire while holding subscan list mutex)
    gchar *reservfile;
    GList *integrations;
    guint64 time0;
    gchar *uid;
    gboolean in_use;
    guint first_intgn_number;
} subscan_t;

typedef struct bintree_t {
    tree_t *tree;
    gint elements_remaining;
} bintree_t;

typedef struct integration_t {
    bintree_t cross_data;
    bintree_t auto_data;
    guint intgn_number;
    GHashTable *sw_buffers;
    gboolean finished;
} integration_t;

// NB: For now, we're assuming that the axis lists for the
// CROSS_DATA and AUTO_DATA binary components are FIXED!
// TODO: change this
#define BAL_AXIS_INDEX 0
#define BAB_AXIS_INDEX 1
#define SPW_AXIS_INDEX 2
#define NUM_AXES 3

typedef struct product_map_t {
    guint index[NUM_AXES + 1];
    gboolean is_cross;
    gboolean is_reversed;
    guint num_channels;
    guint num_polprods;
    guint polprod_position;
    // the following are necessary so that even when some data fail
    // to arrive, we can write zeros correctly
    gboolean is_complex[4];
} product_map_t;

typedef struct vis_buffer_t {
    product_map_t *map;
    GstBuffer *buffer;
    guint num_rows;
    guint num_cols;
    guint rowlen;
} vis_buffer_t;

typedef struct sw_buffer_t {
    vis_buffer_t vis[4];
    gint last_idx;
    guint num_vis;
} sw_buffer_t;

// for balant_index we define an error value
#define BALANT_NONE G_MAXINT

typedef enum polarization_t {
    RIGHT_POLARIZATION, LEFT_POLARIZATION, INVALID_POLARIZATION
} polarization_t;

struct _wcbe_bdf_t {
    we_bdf_t *bdf;
    gchar *path;
    gchar *iobuff;
    GStaticRecMutex mtx;
    guint refcount;
    struct flock lock;
};

// helper structure used in writing out a spectral window's data using
// all polarizations
typedef struct {
    // how many float values to write for each channel
    guint elements_per_index;
    // a pointer to the data for the polarization
    gfloat *array;
    // increment to move from start of one row to the next
    guint rowlen;
    // increment to next data element
    gint col_increment;
    // whether sense of correlation in lag set is reversed relative to
    // value to be written to BDF
    gboolean is_reversed;
} poln_buffer_metadata_t;

#define UID_VALID_CHARS G_CSET_A_2_Z G_CSET_a_2_z G_CSET_DIGITS "."
#define wcbe_bdf_lock(wbdf) g_static_rec_mutex_lock(&(wbdf->mtx))
#define wcbe_bdf_unlock(wbdf) g_static_rec_mutex_unlock(&(wbdf->mtx))

GST_DEBUG_CATEGORY_STATIC(wcbe_bdf_visdata_debug);
#define GST_CAT_DEFAULT wcbe_bdf_visdata_debug

#define FLOW_DROPPED (GST_FLOW_CUSTOM_ERROR - 1)

/* Filter signals and args */
enum
{
    SIGNAL_GET_CONFIG,
    SIGNAL_GET_PIPELINE_NAME,
    SIGNAL_GET_WCBE_NAME,
    SIGNAL_GET_BE_INTEGRATION,
    SIGNAL_GET_BDFDIR,
    SIGNAL_REGISTER_TMPFILE,
    SIGNAL_UNREGISTER_TMPFILE,
    LAST_SIGNAL
};
static guint wcbe_bdf_visdata_signals[LAST_SIGNAL] = {0,};

enum
{
    ARG_0,
    ARG_BDFDIR,
    ARG_SUBSCAN_LIST,
    ARG_NUM_WRITTEN,
    ARG_NUM_DROPPED
};

#define NSEC_PER_USEC 1000

#define VIS_STATIC_CAPS                                \
     WCBE_VECTOR_MEDIA_TYPE                            \
     ","WCBE_ELEMENT_TYPE_PROPERTY"="                  \
        WCBE_ELEMENT_TYPE_PROPERTY_TYPE "gfloat"       \
  ";"WCBE_ARRAY_MEDIA_TYPE                             \
    ","WCBE_ELEMENT_TYPE_PROPERTY"="                   \
       WCBE_ELEMENT_TYPE_PROPERTY_TYPE "gfloat"

static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE(
    "sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS(VIS_STATIC_CAPS));

GST_BOILERPLATE(WcbeBdfVisdata, wcbe_bdf_visdata, GstBaseSink,
                GST_TYPE_BASE_SINK);

static void wcbe_bdf_visdata_set_property(
    GObject * object, guint prop_id, const GValue * value, GParamSpec * pspec);
static void wcbe_bdf_visdata_get_property(
    GObject * object, guint prop_id, GValue * value, GParamSpec * pspec);
static void wcbe_bdf_visdata_finalize(GObject *object);
static gboolean wcbe_bdf_visdata_event(GstBaseSink *sink, GstEvent *event);
static GstFlowReturn wcbe_bdf_visdata_render(GstBaseSink *sink,
                                             GstBuffer *buf);
static gboolean wcbe_bdf_visdata_start(GstBaseSink *sink);
static gboolean wcbe_bdf_visdata_stop(GstBaseSink *sink);
static guint bl_bb_sw_hash(gconstpointer key);
static gboolean bl_bb_sw_equal(gconstpointer a, gconstpointer b);
static void sw_buffer_free(gpointer data);
static void end_all_subscans(WcbeBdfVisdata *vis);
static void end_subscan(WcbeBdfVisdata *vis, GList *slink);
static void end_integration(integration_t *intgn);
static gboolean start_integration(WcbeBdfVisdata *vis, subscan_t *subscan,
                                  guint integration_number);
static gboolean start_subscan(WcbeBdfVisdata *vis, const gchar *uid,
                              guint64 time0);
static void cancel_subscan(WcbeBdfVisdata *vis, const gchar *uid);
static gint compare_integrations(gconstpointer a, gconstpointer b);
static gint compare_subscans(gconstpointer a, gconstpointer b);
static void map_products(WcbeBdfVisdata *vis);
static void bb_sw_sb_map(WcbeBdfVisdata *vis, const GArray *stations,
                         const gint *balant_map,
                         guint bb_slot, const wcbe_baseband_config_t *bb,
                         guint sw_slot,
                         const wcbe_spectral_window_config_t *sw,
                         guint sbidx_slot, guint subband_index);
static gint balant_index(const GArray *stations, const gint *balant_map,
                         guint station0, guint station1);
static gboolean bbs_in_baseband(const wcbe_baseband_config_t *bb,
                                guint bb0, guint bb1,
                                polarization_t *pol0, polarization_t *pol1);
static subscan_t *find_subscan(WcbeBdfVisdata *vis, guint64 time0);
static integration_t *find_integration(
    WcbeBdfVisdata *vis, subscan_t *subscan, guint64 time0);
static GstFlowReturn handle_visibilities(
    WcbeBdfVisdata *vis, GstBuffer *buf, const GstStructure *s,
    const subscan_t *subscan, integration_t *integration);
static GstFlowReturn write_sw(integration_t *integration,
                              sw_buffer_t *sw_buffer);
static wcbe_bdf_t *wcbe_bdf_visdata_ref_bdf(WcbeBdfVisdata *vis,
                                            const gchar *uid);
static void wcbe_bdf_visdata_unref_bdf(WcbeBdfVisdata *vis, const gchar *uid);
static gboolean wcbe_bdf_visdata_lock_bdf_tail(WcbeBdfVisdata *vis,
                                               const gchar *uid);
static void wcbe_bdf_visdata_unlock_bdf_tail(WcbeBdfVisdata *vis,
                                             const gchar *uid);
static wcbe_bdf_t *open_bdf(WcbeBdfVisdata *vis, const gchar *uid);
static void wcbe_bdf_destroy(gpointer data);
static sw_buffer_t *integration_add_buffer(
    integration_t *integration, GQuark id, product_map_t *map,
    GstBuffer *buff);

static void
wcbe_bdf_visdata_base_init(gpointer gclass)
{
    GST_DEBUG_CATEGORY_INIT(wcbe_bdf_visdata_debug,
                            "wcbe_bdf_visdata", 0,
                            "wcbe_bdf_visdata debug");
    static GstElementDetails element_details = {
        "WIDAR BDF visibilities data writer",
        "Sink/WIDAR",
        "Write WIDAR visibility data to SDM BDF files",
        "Martin Pokorny <mpokorny@nrao.edu>"
    };
    GstElementClass *element_class = GST_ELEMENT_CLASS(gclass);

    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&sink_factory));
    gst_element_class_set_details(element_class, &element_details);
}

/* initialize the plugin's class */
static void
wcbe_bdf_visdata_class_init(WcbeBdfVisdataClass *klass)
{
    GObjectClass *gobject_class  = (GObjectClass *)klass;
    GstBaseSinkClass *gstbasesink_class = GST_BASE_SINK_CLASS(klass);

    gobject_class->set_property =
        GST_DEBUG_FUNCPTR(wcbe_bdf_visdata_set_property);
    gobject_class->get_property =
        GST_DEBUG_FUNCPTR(wcbe_bdf_visdata_get_property);
    gobject_class->finalize =
        GST_DEBUG_FUNCPTR(wcbe_bdf_visdata_finalize);

    gstbasesink_class->render =
        GST_DEBUG_FUNCPTR(wcbe_bdf_visdata_render);
    gstbasesink_class->start =
        GST_DEBUG_FUNCPTR(wcbe_bdf_visdata_start);
    gstbasesink_class->stop =
        GST_DEBUG_FUNCPTR(wcbe_bdf_visdata_stop);
    gstbasesink_class->event =
        GST_DEBUG_FUNCPTR(wcbe_bdf_visdata_event);

    g_object_class_install_property(
        gobject_class, ARG_BDFDIR,
        g_param_spec_string("bdfdir", "bdf workspace dir",
                            "Root workspace directory for BDF files",
                            NULL, G_PARAM_READABLE));

    g_object_class_install_property(
        gobject_class, ARG_NUM_WRITTEN,
        g_param_spec_uint("num-written", "Number lag sets written",
                          "Number of (lag set) buffers written to BDF files",
                          0, G_MAXUINT, 0, G_PARAM_READABLE));

    g_object_class_install_property(
        gobject_class, ARG_NUM_DROPPED,
        g_param_spec_uint("num-dropped", "Number lag sets dropped",
                          "Number of (lag set) buffers not written to BDF "
                          "files",
                          0, G_MAXUINT, 0, G_PARAM_READABLE));

    g_object_class_install_property(
        gobject_class, ARG_SUBSCAN_LIST,
        g_param_spec_string("subscans", "subscan list",
                            "List of ready subscans",
                            NULL, G_PARAM_READABLE));

    wcbe_bdf_visdata_signals[SIGNAL_GET_CONFIG] =
        g_signal_new("get-config", G_TYPE_FROM_CLASS(klass),
                     G_SIGNAL_RUN_LAST, 0, NULL, NULL,
                     g_cclosure_user_marshal_POINTER__VOID,
                     G_TYPE_POINTER, 0);

    wcbe_bdf_visdata_signals[SIGNAL_GET_PIPELINE_NAME] =
        g_signal_new("get-pipeline-name", G_TYPE_FROM_CLASS(klass),
                     G_SIGNAL_RUN_LAST, 0, NULL, NULL,
                     g_cclosure_user_marshal_POINTER__VOID,
                     G_TYPE_POINTER, 0);

    wcbe_bdf_visdata_signals[SIGNAL_GET_WCBE_NAME] =
        g_signal_new("get-wcbe-name", G_TYPE_FROM_CLASS(klass),
                     G_SIGNAL_RUN_LAST, 0, NULL, NULL,
                     g_cclosure_user_marshal_POINTER__VOID,
                     G_TYPE_POINTER, 0);

    wcbe_bdf_visdata_signals[SIGNAL_GET_BE_INTEGRATION] =
        g_signal_new("get-be-integration", G_TYPE_FROM_CLASS(klass),
                     G_SIGNAL_RUN_LAST, 0, NULL, NULL,
                     g_cclosure_user_marshal_UINT__VOID,
                     G_TYPE_UINT, 0);

    wcbe_bdf_visdata_signals[SIGNAL_GET_BDFDIR] =
        g_signal_new("get-bdfdir", G_TYPE_FROM_CLASS(klass),
                     G_SIGNAL_RUN_LAST, 0, NULL, NULL,
                     g_cclosure_user_marshal_POINTER__VOID,
                     G_TYPE_POINTER, 0);

    wcbe_bdf_visdata_signals[SIGNAL_REGISTER_TMPFILE] =
        g_signal_new("register-tmpfile", G_TYPE_FROM_CLASS(klass),
                     G_SIGNAL_RUN_FIRST, 0, NULL, NULL,
                     g_cclosure_marshal_VOID__STRING,
                     G_TYPE_NONE, 1, G_TYPE_STRING);

    wcbe_bdf_visdata_signals[SIGNAL_UNREGISTER_TMPFILE] =
        g_signal_new("unregister-tmpfile", G_TYPE_FROM_CLASS(klass),
                     G_SIGNAL_RUN_FIRST, 0, NULL, NULL,
                     g_cclosure_marshal_VOID__STRING,
                     G_TYPE_NONE, 1, G_TYPE_STRING);

    g_static_mutex_init(&(klass->mtx));
    klass->bdfdir = NULL;
    klass->bdfs = g_hash_table_new_full(g_str_hash, g_str_equal, g_free,
                                        wcbe_bdf_destroy);
    klass->ref_bdf = GST_DEBUG_FUNCPTR(wcbe_bdf_visdata_ref_bdf);
    klass->unref_bdf = GST_DEBUG_FUNCPTR(wcbe_bdf_visdata_unref_bdf);
    klass->lock_bdf_tail = GST_DEBUG_FUNCPTR(wcbe_bdf_visdata_lock_bdf_tail);
    klass->unlock_bdf_tail =
        GST_DEBUG_FUNCPTR(wcbe_bdf_visdata_unlock_bdf_tail);
    klass->open_bdf = GST_DEBUG_FUNCPTR(open_bdf);
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
wcbe_bdf_visdata_init(WcbeBdfVisdata *vis, WcbeBdfVisdataClass *gclass)
{
    GstBaseSink *basesink = GST_BASE_SINK(vis);
    gst_base_sink_set_sync(basesink, FALSE);
    gst_base_sink_set_async_enabled(basesink, FALSE);
    vis->config = NULL;
    vis->subscans = NULL;
    vis->product_maps = g_hash_table_new(NULL, NULL);
    g_static_rec_mutex_init(&(vis->mtx));
    vis->pipeline_name = NULL;
    vis->wcbe_name = NULL;
    vis->num_written = 0;
    vis->num_dropped = 0;
}

static void
wcbe_bdf_visdata_finalize(GObject *object)
{
    WcbeBdfVisdata *vis = WCBE_BDF_VISDATA(object);
    end_all_subscans(vis);
    if (vis->product_maps != NULL) {
        g_hash_table_destroy(vis->product_maps);
    }
    g_static_rec_mutex_free(&(vis->mtx));
    if (vis->pipeline_name != NULL) {
        g_free(vis->pipeline_name);
    }
    if (vis->wcbe_name != NULL) {
        g_free(vis->wcbe_name);
    }
    G_OBJECT_CLASS(parent_class)->finalize(object);
}

static gboolean
wcbe_bdf_visdata_start(GstBaseSink *basesink)
{
    WcbeBdfVisdata *vis = WCBE_BDF_VISDATA(basesink);
    WcbeBdfVisdataClass *klass = WCBE_BDF_VISDATA_GET_CLASS(vis);
    if (vis->wcbe_name == NULL) {
        g_signal_emit(vis, wcbe_bdf_visdata_signals[SIGNAL_GET_WCBE_NAME], 0,
                      &(vis->wcbe_name));
        vis->wcbe_name = g_strdup(vis->wcbe_name);
    }
    if (vis->pipeline_name == NULL) {
        g_signal_emit(vis, wcbe_bdf_visdata_signals[SIGNAL_GET_PIPELINE_NAME],
                      0, &(vis->pipeline_name));
        vis->pipeline_name = g_strdup(g_path_get_basename(vis->pipeline_name));
    }
    if (vis->config == NULL) {
        g_signal_emit(vis, wcbe_bdf_visdata_signals[SIGNAL_GET_CONFIG], 0,
                      &(vis->config));
        map_products(vis);
    }
    if (klass->bdfdir == NULL) {
        g_signal_emit(vis, wcbe_bdf_visdata_signals[SIGNAL_GET_BDFDIR], 0,
                      &(klass->bdfdir));
    }
    guint intgn_us;
    g_signal_emit(vis, wcbe_bdf_visdata_signals[SIGNAL_GET_BE_INTEGRATION], 0,
                  &intgn_us);
    vis->integration_period = intgn_us * NSEC_PER_USEC;
    vis->num_written = 0;
    vis->num_dropped = 0;
    return TRUE;
}

static gboolean
wcbe_bdf_visdata_stop(GstBaseSink *sink)
{
    return TRUE;
}

static void
wcbe_bdf_visdata_set_property(GObject *object, guint prop_id,
                              const GValue *value, GParamSpec *pspec)
{
    //WcbeBdfVisdata *vis = WCBE_BDF_VISDATA(object);

    switch (prop_id) {
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static void
wcbe_bdf_visdata_get_property(GObject *object, guint prop_id,
                              GValue *value, GParamSpec *pspec)
{
    WcbeBdfVisdata *vis = WCBE_BDF_VISDATA(object);
    WcbeBdfVisdataClass *klass = WCBE_BDF_VISDATA_GET_CLASS(vis);

    switch (prop_id) {
      case ARG_BDFDIR:
        g_value_set_string(value, klass->bdfdir ? klass->bdfdir : "");
        break;
      case ARG_SUBSCAN_LIST: {
        GString *str = g_string_new("");
        g_static_rec_mutex_lock(&(vis->mtx));
        GList *subscan = vis->subscans;
        while (subscan != NULL) {
            subscan_t *ss = (subscan_t *)(subscan->data);
            const gchar *format = (str->len == 0 ? "%"G_GUINT64_FORMAT"-%s" :
                                   "\t%"G_GUINT64_FORMAT"-%s");
            g_string_append_printf(str, format, ss->time0,
                                   (ss->uid ? ss->uid : "(nil)"));
            subscan = g_list_next(subscan);
        }
        g_static_rec_mutex_unlock(&(vis->mtx));
        g_value_set_string(value, str->str);
        g_string_free(str, TRUE);
        break;
      }
      case ARG_NUM_WRITTEN:
        g_value_set_uint(value, g_atomic_int_get(&(vis->num_written)));
        break;
      case ARG_NUM_DROPPED:
        g_value_set_uint(value, g_atomic_int_get(&(vis->num_dropped)));
        break;
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static GstFlowReturn
wcbe_bdf_visdata_render(GstBaseSink *sink, GstBuffer *buf)
{
    GstFlowReturn result = GST_FLOW_OK;
    WcbeBdfVisdata *vis = WCBE_BDF_VISDATA(sink);

    GstStructure *s = gst_caps_get_structure(GST_BUFFER_CAPS(buf), 0);
    guint epoch;
    guint seconds;
    guint cycles;
    if (!gst_structure_get_uint(s, WCBE_EPOCH_PROPERTY, &epoch)
        || !gst_structure_get_uint(s, WCBE_SECONDS_PROPERTY, &seconds)
        || !gst_structure_get_uint(s, WCBE_CYCLES_PROPERTY, &cycles)) {
        g_warning("Failed to get timestamp from visibility data buffer caps");
        g_atomic_int_inc((gint *)&(vis->num_dropped));
        return result;
    }
    lf_widar_time_t ts =
        {.epoch = epoch, .seconds = seconds, .clock_cycles = cycles};
    guint64 time0 = WIDAR_TIME_TO_NS(TIMEVALUE(&ts));
    subscan_t *subscan = find_subscan(vis, time0);
    if (subscan != NULL) {
        if (subscan->uid != NULL) {
            integration_t *integration = find_integration(vis, subscan, time0);
            if (integration != NULL) {
                result =
                    handle_visibilities(vis, buf, s, subscan, integration);
                g_atomic_int_inc((gint *)&(vis->num_written));
                if (result == FLOW_DROPPED) {
                    g_atomic_int_inc((gint *)&(vis->num_dropped));
                    result = GST_FLOW_OK;
                }
                // finish off old integrations
                GList *ielt = g_list_first(subscan->integrations);
                integration_t *intgn = ((ielt != NULL) ? ielt->data : NULL);
                while ((intgn != NULL && intgn->finished)
                       || (g_list_length(subscan->integrations)
                           > MAX_LAGSET_DELAY + 1)) {
                    subscan->first_intgn_number = intgn->intgn_number + 1;
                    end_integration(intgn);
                    subscan->integrations =
                        g_list_delete_link(subscan->integrations, ielt);
                    ielt = g_list_first(subscan->integrations);
                    intgn = ((ielt != NULL) ? ielt->data : NULL);
                }
            }
            else {
                g_atomic_int_inc((gint *)&(vis->num_dropped));
            }
        }
        else {
            subscan->in_use = TRUE;
            /*g_debug("No subscan UID");*/
        }
        g_static_rec_mutex_unlock(&(subscan->lock));

        // end old subscans
        g_static_rec_mutex_lock(&(vis->mtx));
        GList *s = g_list_find(vis->subscans, subscan);
        GList *prev = g_list_previous(s);
        if (prev != NULL) {
            // end all subscan before the previous
            while (vis->subscans != prev) {
                end_subscan(vis, vis->subscans);
            }
            // end previous subscan, if timestamp is at least one
            // integration period beyond start of this subscan
            if ((time0 - subscan->time0) / vis->integration_period >= 1) {
                end_subscan(vis, prev);
                g_assert(vis->subscans == s);
            }
        }
        g_static_rec_mutex_unlock(&(vis->mtx));
    }
    else {
        /*g_debug("No subscan: dropping visibilities");*/
        g_atomic_int_inc((gint *)&(vis->num_dropped));
    }
    return result;
}

static gboolean
wcbe_bdf_visdata_event(GstBaseSink *sink, GstEvent *event)
{
    gboolean result = TRUE;
    WcbeBdfVisdata *vis = WCBE_BDF_VISDATA(sink);
    switch (GST_EVENT_TYPE(event)) {
      case GST_EVENT_EOS:
        end_all_subscans(vis);
        break;
      case WCBE_EVENT_NEW_SUBSCAN: {
        guint epoch;
        guint seconds;
        guint cycles;
        if (gst_structure_get_uint(event->structure, "epoch", &epoch)
            && gst_structure_get_uint(event->structure, "seconds", &seconds)
            && gst_structure_get_uint(event->structure, "cycles", &cycles)) {
            const gchar *uid =
                gst_structure_get_string(event->structure, "uid");
            lf_widar_time_t ts = {.epoch = epoch, .seconds = seconds,
                                  .clock_cycles = cycles};
            if (!start_subscan(vis, uid, WIDAR_TIME_TO_NS(TIMEVALUE(&ts)))) {
                g_message("Failed to start subscan '%s'", uid);
            }
        }
        else {
            g_debug("'new-subscan' event arrived without time "
                    "and/or 'uid' fields...ignored");
        }
        break;
      }
      case WCBE_EVENT_CANCEL_SUBSCAN: {
        const gchar *uid =
            gst_structure_get_string(event->structure, "uid");
        g_debug("bdf writer got cancel-subscan event for uid '%s'",
                uid ? uid : "(nil)");
        if (uid != NULL) {
            cancel_subscan(vis, uid);
        }
        break;
      }
      default:
        break;
    }
    return result;
}

static guint
bl_bb_sw_hash(gconstpointer key)
{
    const product_map_t *map = key;
    return ((map->is_cross ? 1 : -1)
            * (map->index[BAL_AXIS_INDEX]
               ^ (map->index[BAB_AXIS_INDEX] << 8)
               ^ (map->index[SPW_AXIS_INDEX] << 16)));
}

static gboolean
bl_bb_sw_equal(gconstpointer a, gconstpointer b)
{
    const product_map_t *mapa = a;
    const product_map_t *mapb = b;
    return ((mapa->is_cross == mapb->is_cross)
            && (mapa->index[BAL_AXIS_INDEX] == mapb->index[BAL_AXIS_INDEX])
            && (mapa->index[BAB_AXIS_INDEX] == mapb->index[BAB_AXIS_INDEX])
            && (mapa->index[SPW_AXIS_INDEX] == mapb->index[SPW_AXIS_INDEX]));
}

static void
sw_buffer_free(gpointer data)
{
    sw_buffer_t *sw_buffer = data;
    guint i;
    for (i = 0; i < G_N_ELEMENTS(sw_buffer->vis); ++i) {
        if (sw_buffer->vis[i].buffer != NULL) {
            gst_buffer_unref(sw_buffer->vis[i].buffer);
        }
    }
    g_slice_free(sw_buffer_t, sw_buffer);
}

static void
end_all_subscans(WcbeBdfVisdata *vis)
{
    g_static_rec_mutex_lock(&(vis->mtx));
    while (vis->subscans != NULL) {
        end_subscan(vis, vis->subscans);
    }
    g_static_rec_mutex_unlock(&(vis->mtx));
}

static void
end_subscan(WcbeBdfVisdata *vis, GList *slink)
{
    subscan_t *subscan = slink->data;
    g_debug("end_subscan %s", subscan->uid ? subscan->uid : "(nil)");
    WcbeBdfVisdataClass *klass = WCBE_BDF_VISDATA_GET_CLASS(vis);
    g_static_rec_mutex_lock(&(vis->mtx));
    vis->subscans = g_list_delete_link(vis->subscans, slink);
    g_static_rec_mutex_lock(&(subscan->lock));
    g_static_rec_mutex_unlock(&(vis->mtx));
    while (subscan->integrations != NULL) {
        integration_t *intgn = subscan->integrations->data;
        subscan->first_intgn_number = intgn->intgn_number + 1;
        end_integration(intgn);
        subscan->integrations =
            g_list_delete_link(subscan->integrations, subscan->integrations);
    }
    if (subscan->wbdf != NULL) {
        klass->unref_bdf(vis, subscan->uid);
    }
    if (subscan->reservfile != NULL) {
        gint rc = unlink(subscan->reservfile);
        if (rc == 0) {
            g_signal_emit(vis,
                          wcbe_bdf_visdata_signals[SIGNAL_UNREGISTER_TMPFILE],
                          0, subscan->reservfile);
        }
        else {
            if (errno != ENOENT) {
                g_warning("Failed to unlink BDF reservfile '%s': %s",
                          subscan->reservfile, g_strerror(errno));
            }
            else {
                g_message("BDF reservfile '%s' is unexpectedly missing when "
                          "trying to unlink", subscan->reservfile);
            }
        }
        g_free(subscan->reservfile);
    }
    if (subscan->uid != NULL) {
        g_free(subscan->uid);
    }
    g_static_rec_mutex_unlock(&(subscan->lock));
    g_static_rec_mutex_free(&(subscan->lock));
    g_free(subscan);
}

static void
end_integration(integration_t *intgn)
{
    g_debug("end_integration %u", intgn->intgn_number);
    void flush_sw(gpointer key G_GNUC_UNUSED, gpointer value,
                  gpointer user_data G_GNUC_UNUSED) {
        sw_buffer_t *sw = value;
        write_sw(intgn, sw);
    }
    g_hash_table_foreach(intgn->sw_buffers, flush_sw, NULL);
    void free_bintree(bintree_t *bintree) {
        if (bintree->tree != NULL) {
            tree_free(bintree->tree);
        }
    }
    free_bintree(&(intgn->cross_data));
    free_bintree(&(intgn->auto_data));
    g_hash_table_destroy(intgn->sw_buffers);
    g_slice_free(integration_t, intgn);
}

static gboolean
start_integration(WcbeBdfVisdata *vis, subscan_t *subscan,
                  guint integration_number)
{
    g_return_val_if_fail(subscan != NULL, FALSE);
    g_return_val_if_fail(subscan->wbdf != NULL, FALSE);
    g_return_val_if_fail(subscan->wbdf->bdf != NULL, FALSE);
    g_return_val_if_fail(subscan->wbdf->bdf->main_hdr != NULL, FALSE);
    g_return_val_if_fail(subscan->wbdf->bdf->main_hdr->project_path != NULL,
                         FALSE);

    // it's assumed that subscan->lock is held by caller, i.e., locked
    // across all threads running in this instance
    gboolean result = FALSE;
    gchar *project_path =
        g_strdup_printf("%s%u/", subscan->wbdf->bdf->main_hdr->project_path,
                        integration_number);
    sdmhdr_subset_hdr_t *subset =
        we_bdf_subset_hdr(subscan->wbdf->bdf, project_path);

    if (subset == NULL) {
        // lock across all BDF writers in process
        wcbe_bdf_lock(subscan->wbdf);
        // look for subset again in case it was created before we
        // obtained the lock
        subset = we_bdf_subset_hdr(subscan->wbdf->bdf, project_path);
        if (subset == NULL) {
            sdmhdr_subset_hdr_t template = {
                .cross_data_type = COBDF_FLOAT,
                .data_struct_ref = "sdmDataHeader",
                .interval = vis->integration_period,
            };
            switch (subscan->wbdf->bdf->main_hdr->mode) {
              case COBDF_CROSS_AND_AUTO:
                template.type = "BinaryCrossAndAutoDataFullResolution";
                break;
              case COBDF_CROSS_ONLY:
                template.type = "BinaryCrossDataFullResolution";
                break;
              case COBDF_AUTO_ONLY:
                template.type = "BinaryAutoDataFullResolution";
                break;
              default:
                g_assert_not_reached();
                break;
            }
            guint last_index = integration_number;
/*             if (vis->integration_period < NSEC_PER_SEC) { */
/*                 last_index = integration_number */
/*                     + NSEC_PER_SEC / vis->integration_period + 1; */
/*             } */
            // lock across processes writing to file
            if (wcbe_bdf_visdata_lock_bdf_tail(vis, subscan->uid)) {
                g_debug("add through %u", last_index);
                we_bdf_add_subset_seq(
                    subscan->wbdf->bdf, &template, last_index);
                wcbe_bdf_visdata_unlock_bdf_tail(vis, subscan->uid);
                subset = we_bdf_subset_hdr(subscan->wbdf->bdf, project_path);
                if (subset == NULL) {
                    g_warning("Failed to find subset %s after attempt to "
                              "allocate it", project_path);
                }
            }
        }
        wcbe_bdf_unlock(subscan->wbdf);
    }

    if (subset == NULL) {
        g_warning("Failed to find or allocate integration number %d in BDF "
                  "for uid '%s'", integration_number, subscan->uid);
    }
    else if (subset->cross_data_href == NULL
             && subset->auto_data_href == NULL) {
        g_warning("Neither CROSS_DATA nor AUTO_DATA components found in BDF "
                  "for integration");
    }
    else {
        result = TRUE;
        integration_t *integration = g_slice_new0(integration_t);
        integration->intgn_number = integration_number;
        integration->finished = FALSE;
        integration->sw_buffers = g_hash_table_new_full(
            bl_bb_sw_hash, bl_bb_sw_equal, NULL, sw_buffer_free);
        gboolean init_bintree(bintree_t *bintree, CobdfBincomp bincomp) {
            gboolean res = FALSE;
            bintree->tree =
                we_bdf_bincomp(subscan->wbdf->bdf, project_path, bincomp);
            if (bintree->tree != NULL) {
                bintree->elements_remaining = (guint)tree_count(bintree->tree);
                res = TRUE;
            }
            return res;
        }
        if (subset->cross_data_href != NULL) {
            result = init_bintree(&(integration->cross_data),
                                  COBDF_CROSS_DATA_BINCOMP);
            if (!result) {
                g_info("Failed to locate CROSS_DATA in BDF file for "
                       "(%s/%s)", subscan->wbdf->bdf->main_hdr->data_oid->href,
                       project_path);
                result = FALSE;
            }
        }
        if (result && subset->auto_data_href != NULL) {
            result = init_bintree(&(integration->auto_data),
                                  COBDF_AUTO_DATA_BINCOMP);
            if (!result) {
                g_info("Failed to locate AUTO_DATA in BDF file for (%s/%s)",
                       subscan->wbdf->bdf->main_hdr->data_oid->href,
                       project_path);
            }
        }
        if (result) {
            g_static_rec_mutex_lock(&(subscan->lock));
            // We want to ensure a sorted list of integrations, so use
            // g_slist_insert_sorted rather than g_slist_append.
            if (subscan->integrations == NULL) {
                subscan->integrations = g_list_append(NULL, integration);
            }
            else {
                subscan->integrations =
                    g_list_insert_sorted(subscan->integrations, integration,
                                         compare_integrations);
            }
            g_static_rec_mutex_unlock(&(subscan->lock));
        }
        else {
            end_integration(integration);
        }
    }
    sdmhdr_subset_hdr_free(subset);
    g_free(project_path);
    return result;
}

static gboolean
start_subscan(WcbeBdfVisdata *vis, const gchar *uid, guint64 time0)
{
    WcbeBdfVisdataClass *klass = WCBE_BDF_VISDATA_GET_CLASS(vis);

    gboolean result = FALSE;

    subscan_t *subscan = g_new0(subscan_t, 1);
    g_static_rec_mutex_init(&(subscan->lock));
    subscan->time0 = time0;
    subscan->first_intgn_number = 0;

    if (uid != NULL) {
        subscan->uid = g_strdup(uid);
        wcbe_bdf_t *wbdf = klass->open_bdf(vis, uid);
        if (wbdf != NULL) {
            subscan->wbdf = wbdf;

            // generate a UID for the BDF subscan
            gchar *canon_uid = g_strcanon(g_strdup(uid), UID_VALID_CHARS, '_');

            // first check directory in which BDF file should be located
            g_static_rec_mutex_lock(&(vis->mtx));
            gchar *subscan_dir =
                g_build_filename(klass->bdfdir, canon_uid, NULL);

            // Create a reservfile: name is combination of pipeline name,
            // process id and element name in subscan_dir
            gchar *reservfile_name =
                g_strdup_printf("%s-%s-%s-%s.lck", vis->wcbe_name,
                                vis->config->name,
                                vis->pipeline_name,
                                GST_ELEMENT_NAME(GST_ELEMENT(vis)));
            g_static_rec_mutex_unlock(&(vis->mtx));
            gchar *reservfile =
                g_build_filename(subscan_dir, reservfile_name, NULL);
            g_free(reservfile_name);
            g_free(canon_uid);
            g_free(subscan_dir);

            // reservfile is empty, so just create it and close it immediately
            // thereafter
            gint fd = open(
                reservfile, O_WRONLY | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR);
            if (G_LIKELY(fd != -1)) {
                g_signal_emit(
                    vis, wcbe_bdf_visdata_signals[SIGNAL_REGISTER_TMPFILE],
                    0, reservfile);

                gint rc = close(fd);
                if (G_UNLIKELY(rc == -1)) {
                    g_message("Failed to close BDF reservfile '%s': %s",
                              reservfile, g_strerror(errno));
                }
                result = TRUE;
            }
            else {
                g_info("Failed to open BDF reservfile '%s': %s: will not "
                       "write to BDF file.", reservfile, g_strerror(errno));
            }
            subscan->reservfile = reservfile;
        }
    }
    else {
        result = TRUE;
    }

    // now just add the subscan to the subscans list
    g_static_rec_mutex_lock(&(vis->mtx));
    // Although the subscans are very likely created in order, we want
    // to ensure a sorted subscans list, so we use
    // g_slist_insert_sorted rather than g_slist_append.
    vis->subscans =
        g_list_insert_sorted(vis->subscans, subscan, compare_subscans);
    g_static_rec_mutex_unlock(&(vis->mtx));

    return result;
}

static void
cancel_subscan(WcbeBdfVisdata *vis, const gchar *uid)
{
    g_static_rec_mutex_lock(&(vis->mtx));
    GList *ss = vis->subscans;
    while (ss != NULL) {
        subscan_t *s = ss->data;
        if (s->uid != NULL && strcmp(s->uid, uid) == 0) {
            end_subscan(vis, ss);
            break;
        }
        ss = g_list_next(ss);
    }
    g_static_rec_mutex_unlock(&(vis->mtx));
}

static gint
compare_integrations(gconstpointer a, gconstpointer b)
{
    const integration_t *ia = a;
    const integration_t *ib = b;
    return ((ia->intgn_number < ib->intgn_number) ? -1 :
            ((ia->intgn_number > ib->intgn_number) ? 1 : 0));
}

static gint
compare_subscans(gconstpointer a, gconstpointer b)
{
    const subscan_t *sa = a;
    const subscan_t *sb = b;
    return ((sa->time0 < sb->time0) ? -1 :
            ((sa->time0 > sb->time0) ? 1 : 0));
}

static void
map_products(WcbeBdfVisdata *vis)
{
    const wcbe_config_t *config = vis->config;
    if (config == NULL) {
        g_warning("Failed to get configuration: no BDF output will be "
                  "created");
        return;
    }
    if (vis->pipeline_name == NULL) {
        g_warning("Failed to set pipeline name: no BDF output will be "
                  "created");
        return;
    }

    const wcbe_pipeline_config_t *myconfig = NULL;
    const wcbe_subarray_config_t *subarray = NULL;
    guint i;
    for (i = 0; myconfig == NULL && i < config->pipelines.pipelines->len;
         ++i) {
        const wcbe_pipeline_config_t *pcfg =
            &g_array_index(config->pipelines.pipelines, wcbe_pipeline_config_t,
                           i);
        if (strcmp(pcfg->id, vis->pipeline_name) == 0) {
            myconfig = pcfg;
            subarray = &g_array_index(config->output.subarrays,
                                      wcbe_subarray_config_t,
                                      myconfig->subarray_index);
        }
    }
    if (myconfig == NULL) {
        g_warning("Failed to find configuration of pipeline '%s' in config "
                  "file: no BDF output will be created", vis->pipeline_name);
        return;
    }

    gint *balant_map =
        g_new0(gint, subarray->stations->len * subarray->stations->len);
#define IDX(a, b) ((a) * subarray->stations->len + (b))
    // Initial position value is offset by 1, since we need negative
    // values to encode reversed station order.
    gint pos = 1;
    guint j;
    for (j = 0; j < subarray->stations->len; ++j) {
        balant_map[IDX(j, j)] = j + 1; // add 1 for consistency with
                                       // cross-correlations
        for (i = 0; i < j; ++i) {
            balant_map[IDX(i, j)] = pos;
            balant_map[IDX(j, i)] = -pos;
            ++pos;
        }
    }
#undef IDX

    guint swmap_idx;
    for (swmap_idx = 0; swmap_idx < myconfig->swmaps->len; ++swmap_idx) {
        const wcbe_swmap_config_t *swmap =
            &g_array_index(myconfig->swmaps, wcbe_swmap_config_t, swmap_idx);
        guint bb_slot = swmap->baseband_index;
        const wcbe_baseband_config_t *bb = &g_array_index(
            subarray->basebands, wcbe_baseband_config_t, bb_slot);
        guint sw_slot = swmap->spectral_window_index;
        const wcbe_spectral_window_config_t *sw = &g_array_index(
            bb->spectral_windows, wcbe_spectral_window_config_t, sw_slot);
        guint sbidx_slot;
        for (sbidx_slot = 0; sbidx_slot < swmap->subband_indices->len;
             ++sbidx_slot) {
            guint sbidx =
                g_array_index(swmap->subband_indices, guint, sbidx_slot);
            bb_sw_sb_map(vis, subarray->stations, balant_map,
                         bb_slot, bb, sw_slot, sw, sbidx_slot, sbidx);
        }
    }
    g_free(balant_map);
}

static void
bb_sw_sb_map(WcbeBdfVisdata *vis, const GArray *stations,
             const gint *balant_map,
             guint bb_slot, const wcbe_baseband_config_t *bb,
             guint sw_slot, const wcbe_spectral_window_config_t *sw,
             guint sbidx_slot, guint subband_index)
{
    const wcbe_input_config_t *input = &(vis->config->input);
    product_map_t base_map;
    base_map.index[BAB_AXIS_INDEX] = bb_slot;
    base_map.index[SPW_AXIS_INDEX] = sw_slot;
    base_map.index[NUM_AXES] = COORD_VECTOR_END;
    guint num_cross_polprods;
    switch (sw->xpp) {
      case SDMHDR_XPP_RR:
      case SDMHDR_XPP_LL:
        num_cross_polprods = 1;
        break;
      case SDMHDR_XPP_RR_LL:
        num_cross_polprods = 2;
        break;
      case SDMHDR_XPP_RR_RL_LR_LL:
        num_cross_polprods = 4;
        break;
      default:
        num_cross_polprods = 0;
        break;
    }
    guint num_auto_polprods;
    switch (sw->sdpp) {
      case SDMHDR_APP_RR:
      case SDMHDR_APP_LL:
        num_auto_polprods = 1;
        break;
      case SDMHDR_APP_RR_LL:
        num_auto_polprods = 2;
        break;
      case SDMHDR_APP_RR_RL_LL:
        num_auto_polprods = 3;
        break;
      default:
        num_auto_polprods = 0;
        break;
    }

    guint i;
    for (i = 0; i < input->products->len; ++i) {
        const wcbe_product_config_t *product =
            g_ptr_array_index(input->products, i);
        polarization_t polA, polB;
        if (product->subband_index == subband_index
            && bbs_in_baseband(bb, product->streamA->baseband_id,
                               product->streamB->baseband_id,
                               &polA, &polB)) {
            gint balant_slot = balant_index(stations, balant_map,
                                            product->streamA->station_id,
                                            product->streamB->station_id);
            if (balant_slot != BALANT_NONE) {
                const wcbe_stream_config_t *strm0 = product->streamA;
                const wcbe_stream_config_t *strm1 = product->streamB;
                product_map_t *map0 = g_memdup(&base_map, sizeof(base_map));
                map0->index[BAL_AXIS_INDEX] = ABS(balant_slot) - 1;
                // TODO: shouldn't assume value for num_channels below
                map0->num_channels = product->num_lags / 2;
                map0->is_cross = (strm0->station_id != strm1->station_id);
                map0->num_polprods =
                    (map0->is_cross ? num_cross_polprods : num_auto_polprods);
                if (map0->is_cross) {
                    polarization_t *pol0 = &polA;
                    polarization_t *pol1 = &polB;
                    if (balant_slot < 0) {
                        map0->is_reversed = TRUE;
                        pol0 = &polB;
                        pol1 = &polA;
                        strm0 = product->streamB;
                        strm1 = product->streamA;
                    }
                    else {
                        map0->is_reversed = FALSE;
                    }
                    // all cross pol products are complex-valued
                    map0->is_complex[0] = TRUE;
                    map0->is_complex[1] = TRUE;
                    map0->is_complex[2] = TRUE;
                    map0->is_complex[3] = TRUE;
                    switch (sw->xpp) {
                      case SDMHDR_XPP_RR:
                      case SDMHDR_XPP_LL:
                        map0->polprod_position = 0;
                        break;
                      case SDMHDR_XPP_RR_LL:
                        if (*pol0 == RIGHT_POLARIZATION) {
                            map0->polprod_position = 0;
                        }
                        else {
                            map0->polprod_position = 1;
                        }
                        break;
                      case SDMHDR_XPP_RR_RL_LR_LL:
                        if (*pol0 == RIGHT_POLARIZATION) {
                            if (*pol1 == RIGHT_POLARIZATION) {
                                map0->polprod_position = 0;
                            }
                            else {
                                map0->polprod_position = 1;
                            }
                        }
                        else {
                            if (*pol1 == RIGHT_POLARIZATION) {
                                map0->polprod_position = 2;
                            }
                            else {
                                map0->polprod_position = 3;
                            }
                        }
                        break;
                      case SDMHDR_XPP_NONE:
                        g_free(map0);
                        map0 = NULL;
                        break;
                      default:
                        g_warn_if_reached();
                        break;
                    }
                }
                else {
                    map0->is_reversed = FALSE;
                    // most sd pol products are real-valued
                    map0->is_complex[0] = FALSE;
                    map0->is_complex[1] = FALSE;
                    map0->is_complex[2] = FALSE;
                    map0->is_complex[3] = FALSE;
                    switch (sw->sdpp) {
                      case SDMHDR_APP_RR:
                      case SDMHDR_APP_LL:
                        map0->polprod_position = 0;
                        break;
                      case SDMHDR_APP_RR_LL:
                        if (polA == RIGHT_POLARIZATION) {
                            map0->polprod_position = 0;
                        }
                        else {
                            map0->polprod_position = 1;
                        }
                        break;
                      case SDMHDR_APP_RR_RL_LL:
                        map0->is_complex[1] = TRUE;
                        if (polA != polB) {
                            map0->is_reversed = (polA == LEFT_POLARIZATION);
                            map0->polprod_position = 1;
                        }
                        else {
                            if (polA == RIGHT_POLARIZATION) {
                                map0->polprod_position = 0;
                            }
                            else {
                                map0->polprod_position = 2;
                            }
                        }
                        break;
                      case SDMHDR_APP_NONE:
                        g_free(map0);
                        map0 = NULL;
                        break;
                      default:
                        g_warn_if_reached();
                        break;
                    }
                }
                g_warn_if_fail(map0 == NULL
                               || (map0->polprod_position
                                   < map0->num_polprods));
                if (map0 != NULL) {
                    GQuark id = g_quark_from_string(product->id);
                    g_hash_table_insert(vis->product_maps,
                                        GSIZE_TO_POINTER(id), map0);
                }
            }
        }
    }
}

static gint
balant_index(const GArray *stations, const gint *balant_map,
             guint station0, guint station1)
{
    gint idx0 = -1;
    gint idx1 = -1;
    guint i;
    for (i = 0; (idx0 == -1 || idx1 == -1) && i < stations->len; ++i) {
        if (g_array_index(stations, guint, i) == station0) {
            idx0 = i;
        }
        if (g_array_index(stations, guint, i) == station1) {
            idx1 = i;
        }
    }
    gint result;
    if (idx0 != -1 && idx1 != -1) {
        result = balant_map[idx0 * stations->len + idx1];
    }
    else {
        result = BALANT_NONE;
    }
    return result;
}

static gboolean
bbs_in_baseband(const wcbe_baseband_config_t *bb, guint bb0, guint bb1,
                polarization_t *pol0, polarization_t *pol1)
{
    switch (bb->num_polarizations) {
      case 2:
        if (bb0 == bb->baseband_ids[0]) {
            *pol0 = LEFT_POLARIZATION;
        }
        else if (bb0 == bb->baseband_ids[1]) {
            *pol0 = RIGHT_POLARIZATION;
        }
        else {
            *pol0 = INVALID_POLARIZATION;
        }

        if (bb1 == bb->baseband_ids[0]) {
            *pol1 = LEFT_POLARIZATION;
        }
        else if (bb1 == bb->baseband_ids[1]) {
            *pol1 = RIGHT_POLARIZATION;
        }
        else {
            *pol1 = INVALID_POLARIZATION;
        }
        break;

      case 1:
        if ((bb0 == bb->baseband_ids[0] || bb0 == bb->baseband_ids[1])
            && (bb1 == bb->baseband_ids[0] || bb1 == bb->baseband_ids[1])) {
            // the convention is that single polarization basebands
            // are recorded as right-hand polarization
            *pol0 = RIGHT_POLARIZATION;
            *pol1 = RIGHT_POLARIZATION;
        }
        break;

      default:
        g_warn_if_reached();
        *pol0 = INVALID_POLARIZATION;
        *pol1 = INVALID_POLARIZATION;
        break;
    }
    return (*pol0 != INVALID_POLARIZATION && *pol1 != INVALID_POLARIZATION);
}

static subscan_t *
find_subscan(WcbeBdfVisdata *vis, guint64 time0)
{
    subscan_t *result = NULL;
    g_static_rec_mutex_lock(&(vis->mtx));
    GList *ss = vis->subscans;
    GList *prev = NULL;
    while (ss != NULL && time0 >= ((subscan_t *)(ss->data))->time0) {
        prev = ss;
        ss = g_list_next(ss);
    }
    ss = prev;
    if (ss != NULL) {
        result = (subscan_t *)(ss->data);
        g_static_rec_mutex_lock(&(result->lock));
    }
    g_static_rec_mutex_unlock(&(vis->mtx));
    return result;
}

static integration_t *
find_integration(WcbeBdfVisdata *vis, subscan_t *subscan, guint64 time0)
{
    // it's assumed that subscan->lock is held by caller
    integration_t *result = NULL;
    guint integration_number =
        (time0 - subscan->time0) / vis->integration_period;
    if (integration_number >= subscan->first_intgn_number) {
        integration_t dummy = {.intgn_number = integration_number};
        GList *intgn = g_list_find_custom(subscan->integrations, &dummy,
                                          compare_integrations);
        if (intgn == NULL) {
            if (start_integration(vis, subscan, integration_number)) {
                intgn = g_list_find_custom(subscan->integrations, &dummy,
                                           compare_integrations);
                subscan->in_use = TRUE;
            }
            else {
                g_info("Failed to start integration %u for time %"
                       G_GUINT64_FORMAT" in subscan %s",
                       integration_number, time0, subscan->uid);
            }
        }
        if (intgn != NULL) {
            result = (integration_t *)(intgn->data);
        }
        else {
            g_debug("Failed to find integration %u in subscan %s: "
                    "visibility buffer dropped", integration_number,
                    subscan->uid);
        }
    }
    else {
        g_debug("Late visibility data (%u < %u) (subscan %s): dropped",
                integration_number, subscan->first_intgn_number,
                subscan->uid);
    }
    return result;
}

static GstFlowReturn
handle_visibilities(WcbeBdfVisdata *vis, GstBuffer *buf,
                    const GstStructure *s, const subscan_t *subscan,
                    integration_t *integration)
{
    // product id
    GQuark id;
    if (!gst_structure_get_uint(s, WCBE_PRODUCT_ID_PROPERTY, (guint *)&id)) {
        g_debug("Visibilities buffer caps missing '"WCBE_PRODUCT_ID_PROPERTY
                "' property");
        return FLOW_DROPPED;
    }

    // product map
    product_map_t *map =
        g_hash_table_lookup(vis->product_maps, GSIZE_TO_POINTER(id));
    if (map == NULL) {
/*         g_info("Visibilities buffer does not have an output mapping: " */
/*                "%s...dropped", g_quark_to_string(id)); */
        return FLOW_DROPPED;
    }
    sw_buffer_t *sw_buffer = integration_add_buffer(integration, id, map, buf);

    if (G_UNLIKELY(sw_buffer == NULL)) {
        return FLOW_DROPPED;
    }

    GstFlowReturn result = GST_FLOW_OK;
    if (sw_buffer->num_vis == map->num_polprods) {
        result = write_sw(integration, sw_buffer);
        g_hash_table_remove(integration->sw_buffers, map);
        integration->finished =
            ((integration->cross_data.elements_remaining == 0)
             && (integration->auto_data.elements_remaining == 0));
    }
    return result;
}

static sw_buffer_t *
integration_add_buffer(integration_t *integration, GQuark product_id,
                       product_map_t *map, GstBuffer *buff)
{
    sw_buffer_t *result =
        g_hash_table_lookup(integration->sw_buffers, map);
    if (result == NULL) {
        g_return_val_if_fail(map->num_polprods > 0, NULL);
        result = g_slice_new0(sw_buffer_t);
        result->last_idx = -1;
        g_hash_table_insert(integration->sw_buffers, map, result);
    }
    g_return_val_if_fail(result->num_vis < map->num_polprods, NULL);
    vis_buffer_t *visbuf = &(result->vis[map->polprod_position]);

    if (G_UNLIKELY(visbuf->buffer != NULL)) {
        g_debug("Overwritten buffer in sw_buffer_t, id %s, position %u",
                g_quark_to_string(product_id), map->polprod_position);
        gst_buffer_unref(visbuf->buffer);
    }
    visbuf->buffer = buff;
    visbuf->map = map;
    gst_buffer_ref(buff);

    // cache the buffer shape (this could be postponed, but we do it
    // here for convenience)
    GstStructure *s = gst_caps_get_structure(GST_BUFFER_CAPS(buff), 0);
    if (gst_structure_has_name(s, WCBE_VECTOR_MEDIA_TYPE)) {
        visbuf->num_rows = 1;
        if (G_UNLIKELY(!gst_structure_get_uint(s, WCBE_VECTOR_LENGTH_PROPERTY,
                                               &(visbuf->num_cols)))) {
            g_debug("Vector visibilities buffer caps missing '"
                    WCBE_VECTOR_LENGTH_PROPERTY"' property");
            visbuf->num_cols = 1;
        }
    }
    else { // gst_structure_has_name(WCBE_ARRAY_MEDIA_TYPE)
        if (G_UNLIKELY(!gst_structure_get_uint(s, WCBE_ARRAY_NUM_ROWS_PROPERTY,
                                               &(visbuf->num_rows)))) {
            g_debug("Array visibilities buffer caps missing '"
                    WCBE_ARRAY_NUM_ROWS_PROPERTY"' property");
            visbuf->num_rows = 1;
        }
        if (G_UNLIKELY(!gst_structure_get_uint(s, WCBE_ARRAY_NUM_COLS_PROPERTY,
                                               &(visbuf->num_cols)))) {
            g_debug("Array visibilities buffer caps missing '"
                    WCBE_ARRAY_NUM_COLS_PROPERTY"' property");
            visbuf->num_cols = 1;
        }
        visbuf->rowlen = visbuf->num_cols;
        gst_structure_get_uint(s, WCBE_ARRAY_ROW_LENGTH_PROPERTY,
                               &(visbuf->rowlen));
    }

    result->num_vis++;
    result->last_idx = map->polprod_position;
    return result;
}

static GstFlowReturn
write_sw(integration_t *integration, sw_buffer_t *sw_buffer)
{
    if (sw_buffer->num_vis == 0) {
        return FLOW_DROPPED;
    }
    g_return_val_if_fail(sw_buffer->last_idx != -1, FLOW_DROPPED);
    vis_buffer_t *non_null_visbuf = &(sw_buffer->vis[sw_buffer->last_idx]);
    guint num_polprods = non_null_visbuf->map->num_polprods;

    // get the CROSS_DATA or AUTO_DATA binary as needed
    bintree_t *bintree;
    if (non_null_visbuf->map->is_cross) {
        bintree = &(integration->cross_data);
    }
    else {
        bintree = &(integration->auto_data);
    }
    g_return_val_if_fail(bintree->elements_remaining >= 0, FLOW_DROPPED);

    // the shapes of the buffers in sw_buffer should all be the same
    // (modulo the "is_complex" field)
    guint i;
    for (i = 0; i < num_polprods; ++i) {
        if (sw_buffer->vis[i].buffer != NULL
            && ((sw_buffer->vis[i].num_rows != non_null_visbuf->num_rows)
                || (sw_buffer->vis[i].num_cols
                    != non_null_visbuf->num_cols))) {
            g_warning("Shapes of buffers for polarization products "
                      "are unequal");
            return FLOW_DROPPED;
        }
    }

    // get a tree branch for the axes up to SPW_AXIS_INDEX
    branch_t *branch = tree_branch(bintree->tree, non_null_visbuf->map->index);

    // use a small array of zero values for missing data
    static gfloat zero[2] = {0.0f, 0.0f};

    // accumulate metadata for each buffer's data array
    poln_buffer_metadata_t poln_metadata[4];
    for (i = 0; i < num_polprods; ++i) {
        poln_buffer_metadata_t *p = &(poln_metadata[i]);
        if (non_null_visbuf->map->is_complex[i]) {
            p->elements_per_index = 2;
        }
        else {
            p->elements_per_index = 1;
        }
        if (G_LIKELY(sw_buffer->vis[i].buffer != NULL)) {
            p->array = (gfloat *)GST_BUFFER_DATA(sw_buffer->vis[i].buffer);
            p->rowlen = sw_buffer->vis[i].rowlen;
            p->is_reversed = sw_buffer->vis[i].map->is_reversed;
            p->col_increment = 2;
        }
        else {
            p->array = zero;
            p->rowlen = 0;
            p->col_increment = 0;
            p->is_reversed = FALSE;
        }
    }
    // num_channels doesn't vary with buffer
    guint num_channels = non_null_visbuf->map->num_channels;
    // num_bins also doesn't vary with buffer
    guint num_bins = non_null_visbuf->num_rows;

    // write out the data
    GstFlowReturn result = GST_FLOW_OK;
    FILE *f = tree_branch_stream(branch);
    flockfile(f);
    gint rc = fseeko(f, branch->data_offset, SEEK_SET);
    if (rc == 0) {
        guint bin;
        for (bin = 0; result == GST_FLOW_OK && bin < num_bins; ++bin) {
            gfloat *row_start[4] =
                {[0] = poln_metadata[0].array,
                 [1] = poln_metadata[1].array,
                 [2] = poln_metadata[2].array,
                 [3] = poln_metadata[3].array
                };
            guint spp;
            guint pol;
            poln_buffer_metadata_t *pmd;
            for (spp = 0; result == GST_FLOW_OK && spp < num_channels; ++spp) {
                pmd = &(poln_metadata[0]);
                for (pol = 0;
                     result == GST_FLOW_OK && pol < num_polprods;
                     ++pol) {
                    if (pmd->is_reversed) {
                        *(pmd->array + 1) *= -1;
                    }
                    gsize n = fwrite_unlocked(
                        pmd->array, sizeof(gfloat), pmd->elements_per_index,
                        f);
                    if (G_LIKELY(n == pmd->elements_per_index)) {
                        bintree->elements_remaining -= pmd->elements_per_index;
                        pmd->array += pmd->col_increment;
                    }
                    else {
                        g_critical("Failed to write visibility data "
                                   "element to BDF file");
                        result = GST_FLOW_ERROR;
                    }
                    ++pmd;
                }
            }
            pmd = &(poln_metadata[0]);
            for (pol = 0; pol < num_polprods; ++pol) {
                pmd->array = row_start[pol] + pmd->rowlen;
                ++pmd;
            }
        }
        fflush_unlocked(f);
    }
    else {
        g_critical("Failed to seek file position in BDF file: %s",
                   g_strerror(errno));
        result = GST_FLOW_ERROR;
    }
    funlockfile(f);
    tree_branch_free(branch);

    return result;
}

static wcbe_bdf_t *
wcbe_bdf_visdata_ref_bdf(WcbeBdfVisdata *vis, const gchar *uid)
{
    WcbeBdfVisdataClass *klass = WCBE_BDF_VISDATA_GET_CLASS(vis);
    g_static_mutex_lock(&(klass->mtx));
    wcbe_bdf_t *result = g_hash_table_lookup(klass->bdfs, uid);
    if (result != NULL) {
        result->refcount++;
    }
    g_static_mutex_unlock(&(klass->mtx));
    return result;
}

static void
wcbe_bdf_visdata_unref_bdf(WcbeBdfVisdata *vis, const gchar *uid)
{
    WcbeBdfVisdataClass *klass = WCBE_BDF_VISDATA_GET_CLASS(vis);
    g_static_mutex_lock(&(klass->mtx));
    wcbe_bdf_t *wbdf = g_hash_table_lookup(klass->bdfs, uid);
    if (wbdf != NULL) {
        if (--wbdf->refcount == 0) {
            g_hash_table_remove(klass->bdfs, uid);
        };
    }
    g_static_mutex_unlock(&(klass->mtx));
}

static gboolean
wcbe_bdf_visdata_lock_bdf_tail(WcbeBdfVisdata *vis, const gchar *uid)
{
    gboolean result = FALSE;
    WcbeBdfVisdataClass *klass = WCBE_BDF_VISDATA_GET_CLASS(vis);
    wcbe_bdf_t *wbdf = klass->ref_bdf(vis, uid);
    if (wbdf != NULL) {
        wcbe_bdf_lock(wbdf);
        flockfile(wbdf->bdf->file);
        gint rc = fflush_unlocked(wbdf->bdf->file);
        if (rc != 0) {
            g_critical("Failed to flush BDF file '%s': %s",
                       wbdf->path, g_strerror(errno));
            goto done_wbdf;
        }
        rc = fseeko(wbdf->bdf->file, SEEK_END, 0);
        if (rc != 0) {
            g_critical("Failed to seek stream position to end of BDF "
                       "file '%s': %s", wbdf->path, g_strerror(errno));
            goto done_wbdf;
        }
        off_t off = ftello(wbdf->bdf->file);
        gint fd = fileno_unlocked(wbdf->bdf->file);
        if (fd != -1) {
            wbdf->lock.l_type = F_WRLCK;
            wbdf->lock.l_whence = SEEK_SET;
            wbdf->lock.l_start = off;
            wbdf->lock.l_len = 0;
            gint rc = fcntl(fd, F_SETLKW, &(wbdf->lock));
            while (rc != 0) {
                rc = fcntl(fd, F_SETLKW, &(wbdf->lock));
            }
            result = TRUE;
        }
        else {
            g_critical("Failed to get file descriptor for BDF file '%s: %s",
                       wbdf->path, g_strerror(errno));
        }

    done_wbdf:
        if (!result) {
            funlockfile(wbdf->bdf->file);
            wcbe_bdf_unlock(wbdf);
            klass->unref_bdf(vis, uid);
        }
    }
    return result;

}

static void
wcbe_bdf_visdata_unlock_bdf_tail(WcbeBdfVisdata *vis, const gchar *uid)
{
    WcbeBdfVisdataClass *klass = WCBE_BDF_VISDATA_GET_CLASS(vis);
    // get wcbe_bdf from klass->bdfs, but don't ref it, since we
    // should still own a ref
    g_static_mutex_lock(&(klass->mtx));
    wcbe_bdf_t *wbdf = g_hash_table_lookup(klass->bdfs, uid);
    g_static_mutex_unlock(&(klass->mtx));
    if (wbdf != NULL) {
        gint fd = fileno_unlocked(wbdf->bdf->file);
        if (fd != -1) {
            wbdf->lock.l_type = F_UNLCK;
            gint rc = fcntl(fd, F_SETLK, &(wbdf->lock));
            while (rc != 0 && errno == EINTR) {
                rc = fcntl(fd, F_SETLK, &(wbdf->lock));
            }
            if (rc == -1) {
                g_critical("Failed to clear write lock on BDF file '%s': %s",
                           wbdf->path, g_strerror(errno));
            }
        }
        else {
            g_critical("Failed to get file descriptor for BDF file '%s: %s",
                       wbdf->path, g_strerror(errno));
        }
        gint rc = fflush_unlocked(wbdf->bdf->file);
        if (rc != 0) {
            g_warning("Failed to flush BDF file '%s': %s", wbdf->path,
                      g_strerror(errno));
        }
        funlockfile(wbdf->bdf->file);
        wcbe_bdf_unlock(wbdf);
        klass->unref_bdf(vis, uid);
    }
}

static wcbe_bdf_t *
open_bdf(WcbeBdfVisdata *vis, const gchar *uid)
{
    g_return_val_if_fail(uid != NULL, FALSE);

    WcbeBdfVisdataClass *klass = WCBE_BDF_VISDATA_GET_CLASS(vis);
    g_static_mutex_lock(&(klass->mtx));
    wcbe_bdf_t *result = g_hash_table_lookup(klass->bdfs, uid);
    if (result == NULL) {
        gchar *canon_uid = NULL;
        gchar *subscan_dir = NULL;
        gchar *subscan_file = NULL;
        FILE *file = NULL;
        gchar *iobuff = NULL;

        // generate a UID for the BDF subscan
        canon_uid = g_strcanon(g_strdup(uid), UID_VALID_CHARS, '_');

        // first check directory in which BDF file should be located
        g_static_rec_mutex_lock(&(vis->mtx));
        subscan_dir = g_build_filename(klass->bdfdir, canon_uid, NULL);
        g_static_rec_mutex_unlock(&(vis->mtx));

        struct stat statbuf;
        gint rc = stat(subscan_dir, &statbuf);
        if (G_UNLIKELY(rc == -1)) {
            g_info("Failed to stat directory '%s': %s: will not write to "
                   "BDF file.", subscan_dir, g_strerror(errno));
            goto done_wbdf;
        }
        if (G_UNLIKELY(!S_ISDIR(statbuf.st_mode))) {
            g_info("'%s' is not a directory: will not write to BDF file.",
                   subscan_dir);
            goto done_wbdf;
        }
        if (G_UNLIKELY(access(subscan_dir, R_OK) != 0)) {
            g_info("Unable to read directory '%s': %s: will not write to "
                   "BDF file.", subscan_dir, g_strerror(errno));
            goto done_wbdf;
        }

        // now look for read/writeable BDF file
        subscan_file = g_build_filename(subscan_dir, canon_uid, NULL);
        rc = stat(subscan_file, &statbuf);
        if (G_UNLIKELY(rc == -1)) {
            g_info("Failed to stat file '%s': %s: will not write to BDF "
                   "file.", subscan_file, g_strerror(errno));
            goto done_wbdf;
        }
        if (G_UNLIKELY(!S_ISREG(statbuf.st_mode))) {
            g_info("'%s' is not a regular file: will not write to BDF "
                   "file.", subscan_file);
            goto done_wbdf;
        }
        if (G_UNLIKELY(access(subscan_file, R_OK | W_OK) != 0)) {
            g_info("Unable to read and/or write file '%s': %s: will not "
                   "write to BDF file.", subscan_file, g_strerror(errno));
            goto done_wbdf;
        }

        // open the subscan file
        file = fopen(subscan_file, "r+");
        if (G_UNLIKELY(file == NULL)) {
            g_warning("Failed to open BDF file '%s': %s: will not write "
                      "to BDF file.", subscan_file, g_strerror(errno));
            goto done_wbdf;
        }

        // allocate a larger (?) io buffer
        struct stat st;
        rc = fstat(fileno(file), &st);
        if (G_LIKELY(rc == 0)) {
            gsize bufsize = MAX(st.st_blksize, 16 * (1 << 10));
            iobuff = g_try_malloc(bufsize);
            if (G_LIKELY(iobuff != NULL)) {
                rc = setvbuf(file, iobuff, _IOFBF, bufsize);
                if (G_UNLIKELY(rc != 0)) {
                    g_free(iobuff);
                    iobuff = NULL;
                }
                else {
                    g_debug("set IO buffer on BDF file to %u bytes",
                            (guint)bufsize);
                }
            }
        }

        // read the subscan file
        we_bdf_t *bdf = we_bdf_read(file);
        if (G_UNLIKELY(bdf == NULL)) {
            g_info("Failed to parse BDF file '%s': will not write to BDF "
                   "file.", subscan_file);
            goto done_wbdf;
        }

        result = g_new0(wcbe_bdf_t, 1);
        result->bdf = bdf;
        result->path = subscan_file;
        result->iobuff = iobuff;
        subscan_file = NULL;
        g_static_rec_mutex_init(&(result->mtx));
        g_hash_table_insert(klass->bdfs, g_strdup(uid), result);

    done_wbdf:
        if (result == NULL) {
            if (iobuff != NULL) {
                g_free(iobuff);
            }
            if (file != NULL) {
                fclose(file);
            }
        }
        if (canon_uid != NULL) {
            g_free(canon_uid);
        }
        if (subscan_dir != NULL) {
            g_free(subscan_dir);
        }
        if (subscan_file != NULL) {
            g_free(subscan_file);
        }
    }
    g_static_mutex_unlock(&(klass->mtx));

    if (result != NULL) {
        klass->ref_bdf(vis, uid);
    }

    return result;
}

static void
wcbe_bdf_destroy(gpointer data)
{
    wcbe_bdf_t *wbdf = data;
    if (wbdf->bdf != NULL) {
        if (wbdf->bdf->file != NULL) {
            fclose(wbdf->bdf->file);
        }
        we_bdf_free(wbdf->bdf);
    }
    if (wbdf->iobuff != NULL) {
        g_free(wbdf->iobuff);
    }
    if (wbdf->path != NULL) {
        g_free(wbdf->path);
    }
    g_free(wbdf);
}
