/**
 * wcbe_fdvc_normalize
 *
 * Normalization of lags by using the center DV count of the frame
 * each lag belongs to.
 */
#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <wcbe_fdvc_normalize.h>
#include <wcbe.h>

typedef enum {
    ELEMENT_TYPE_INVALID = -1, ELEMENT_TYPE_INT32 = 4,
    ELEMENT_TYPE_INT64 = 8
} element_type_t;

GST_DEBUG_CATEGORY_STATIC(wcbe_fdvc_normalize_debug);
#define GST_CAT_DEFAULT wcbe_fdvc_normalize_debug

enum
{
    ARG_0,
    ARG_CORR_BITS,
    ARG_NUM_NORM,
    ARG_NUM_DROPPED
};

#define DEFAULT_CORR_BITS WCBE_4BITS
#define LAGS_PAD_BASE_NAME "lags"
#define DATAVALID_PAD_BASE_NAME "dv"
#define NORMALIZED_LAGS_PAD_BASE_NAME "nmlags"
#define WEIGHTS_PAD_BASE_NAME "wgts"

static gchar lags_pad_base_name[] = LAGS_PAD_BASE_NAME;
static gchar datavalid_pad_base_name[] = DATAVALID_PAD_BASE_NAME;
static gchar normalized_lags_pad_base_name[] = NORMALIZED_LAGS_PAD_BASE_NAME;
static gchar weights_pad_base_name[] = WEIGHTS_PAD_BASE_NAME;

static GstStaticPadTemplate lags_sink_factory = GST_STATIC_PAD_TEMPLATE(
    LAGS_PAD_BASE_NAME"%d",
    GST_PAD_SINK,
    GST_PAD_REQUEST,
    GST_STATIC_CAPS(WCBE_ARRAY_MEDIA_TYPE
                    ","WCBE_ELEMENT_TYPE_PROPERTY"="
                       WCBE_ELEMENT_TYPE_PROPERTY_TYPE "{gint32,gint64}"));

static GstStaticPadTemplate dv_sink_factory = GST_STATIC_PAD_TEMPLATE(
    DATAVALID_PAD_BASE_NAME"%d",
    GST_PAD_SINK,
    GST_PAD_REQUEST,
    GST_STATIC_CAPS(WCBE_ARRAY_MEDIA_TYPE
                    ","WCBE_ELEMENT_TYPE_PROPERTY"="
                       WCBE_ELEMENT_TYPE_PROPERTY_TYPE"{gint32,gint64}"));

static GstStaticPadTemplate lags_src_factory = GST_STATIC_PAD_TEMPLATE(
    NORMALIZED_LAGS_PAD_BASE_NAME"%d",
    GST_PAD_SRC,
    GST_PAD_REQUEST,
    GST_STATIC_CAPS(WCBE_ARRAY_MEDIA_TYPE
                    ","WCBE_ELEMENT_TYPE_PROPERTY"="
                       WCBE_ELEMENT_TYPE_PROPERTY_TYPE"gfloat"));

static GstStaticPadTemplate wgts_src_factory = GST_STATIC_PAD_TEMPLATE(
    WEIGHTS_PAD_BASE_NAME"%d",
    GST_PAD_SRC,
    GST_PAD_REQUEST,
    GST_STATIC_CAPS(WCBE_ARRAY_MEDIA_TYPE
                    ","WCBE_ELEMENT_TYPE_PROPERTY"="
                       WCBE_ELEMENT_TYPE_PROPERTY_TYPE"{guint32,guint64}"));

GST_BOILERPLATE(WcbeFdvcNormalize, wcbe_fdvc_normalize, GstElement,
                GST_TYPE_ELEMENT);

static void wcbe_fdvc_normalize_get_property(
    GObject *object, guint prop_id, GValue *value, GParamSpec *pspec);
static void wcbe_fdvc_normalize_set_property(
    GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec);
static GstPad *wcbe_fdvc_normalize_request_new_pad(
    GstElement *element, GstPadTemplate *templ, const gchar* name);
static void wcbe_fdvc_normalize_release_pad(GstElement *element, GstPad *pad);
static GList *wcbe_fdvc_normalize_intlink(GstPad *pad);
static GstFlowReturn wcbe_fdvc_normalize_chain(GstPad *pad, GstBuffer *buffer);
static gboolean wcbe_fdvc_normalize_event(GstPad *pad, GstEvent *event);
static gboolean normalize(GstBuffer *lags, GstBuffer *dvs,
                          GstBuffer **normalized, GstBuffer **weights);

static inline gpointer
align_up(gpointer ptr, guint align)
{
    if (align == 0) {
        return ptr;
    }
    gsize addr = (gsize)ptr;
    return (gpointer)(addr + ((align - (addr % align)) % align));
}

static void
wcbe_fdvc_normalize_base_init(gpointer gclass)
{
    GST_DEBUG_CATEGORY_INIT(wcbe_fdvc_normalize_debug,
                            "wcbe_fdvc_normalize", 0,
                            "wcbe_fdvc_normalize debug");
    static GstElementDetails element_details = {
        "WIDAR data valid normalization",
        "Filter/WIDAR",
        "Perform frame-based center data valid normalization",
        "Martin Pokorny <mpokorny@nrao.edu>"
    };
    GstElementClass *element_class = GST_ELEMENT_CLASS(gclass);

    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&lags_sink_factory));
    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&dv_sink_factory));
    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&lags_src_factory));
    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&wgts_src_factory));
    gst_element_class_set_details(element_class, &element_details);
}

/* initialize the plugin's class */
static void
wcbe_fdvc_normalize_class_init(WcbeFdvcNormalizeClass *klass)
{
    GObjectClass *gobject_class  = (GObjectClass *)klass;
    GstElementClass *element_class = GST_ELEMENT_CLASS(klass);

    gobject_class->set_property =
        GST_DEBUG_FUNCPTR(wcbe_fdvc_normalize_set_property);
    gobject_class->get_property =
        GST_DEBUG_FUNCPTR(wcbe_fdvc_normalize_get_property);

    g_object_class_install_property(
        gobject_class, ARG_CORR_BITS,
        g_param_spec_enum("corr-bits", "Number of bits correlated",
                          "Number of bits in correlation products input "
                          "to the element.",
                          WCBE_CORR_BITS_ENUM_TYPE, DEFAULT_CORR_BITS,
                          G_PARAM_READWRITE | G_PARAM_CONSTRUCT));

    g_object_class_install_property(
        gobject_class, ARG_NUM_NORM,
        g_param_spec_uint("num-norm", "Number normalized",
                          "Number of lag set partial products normalized",
                          0, G_MAXUINT, 0, G_PARAM_READABLE));

    g_object_class_install_property(
        gobject_class, ARG_NUM_DROPPED,
        g_param_spec_uint("num-dropped", "Number dropped",
                          "Number of lag set partial products dropped for "
                          "failed normalization",
                          0, G_MAXUINT, 0, G_PARAM_READABLE));

    element_class->request_new_pad =
        GST_DEBUG_FUNCPTR(wcbe_fdvc_normalize_request_new_pad);
    element_class->release_pad =
        GST_DEBUG_FUNCPTR(wcbe_fdvc_normalize_release_pad);
}

static void
wcbe_fdvc_normalize_init(WcbeFdvcNormalize *norm,
                         WcbeFdvcNormalizeClass *klass)
{
    norm->nbits = DEFAULT_CORR_BITS;
    memset(norm->srcpads, 0, sizeof(norm->srcpads));
    memset(norm->sinkpads, 0, sizeof(norm->sinkpads));
    memset(norm->lags, 0, sizeof(norm->lags));
    memset(norm->dvs, 0, sizeof(norm->dvs));
    memset(norm->is_normalized, 0, sizeof(norm->is_normalized));
    norm->null_quotients = FALSE;
    norm->last_partprod = -1;
    norm->num_normalized = 0;
    norm->num_dropped = 0;
}

static void
wcbe_fdvc_normalize_get_property(GObject *object, guint prop_id, GValue *value,
                                 GParamSpec *pspec)
{
    WcbeFdvcNormalize *norm = WCBE_FDVC_NORMALIZE(object);
    switch (prop_id) {
      case ARG_CORR_BITS:
        g_value_set_enum(value, norm->nbits);
        break;
      case ARG_NUM_NORM:
        g_value_set_uint(value, g_atomic_int_get(&(norm->num_normalized)));
        break;
      case ARG_NUM_DROPPED:
        g_value_set_uint(value, g_atomic_int_get(&(norm->num_dropped)));
        break;
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static void
wcbe_fdvc_normalize_set_property(GObject *object, guint prop_id,
                                 const GValue *value, GParamSpec *pspec)
{
    WcbeFdvcNormalize *norm = WCBE_FDVC_NORMALIZE(object);
    switch (prop_id) {
      case ARG_CORR_BITS: {
        WcbeCorrBits old_nbits = norm->nbits;
        norm->nbits = g_value_get_enum(value);
        // if nbits has changed, release all of the sink pads (we
        // could avoid this if ARG_CORR_BITS were
        // G_PARAM_CONSTRUCT_ONLY, but that doesn't seem to work when
        // gstreamer creates the pipeline)
        if (old_nbits != norm->nbits) {
            guint i;
            guint multiplicity = (old_nbits == WCBE_7BITS ? 4 : 1);
            for (i = 0; i < multiplicity; ++i) {
                guint j;
                for (j = 0; j < WCBE_FDVC_NORMALIZE_NUM_PADS; ++j) {
                    if (norm->sinkpads[i][j] != NULL) {
                        gst_element_release_request_pad(
                            GST_ELEMENT(object), norm->sinkpads[i][j]);
                    }
                }
            }
        }
        break;
      }
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static GstFlowReturn
wcbe_fdvc_normalize_chain(GstPad *pad, GstBuffer *buffer)
{
    GstFlowReturn result = GST_FLOW_OK;
    WcbeFdvcNormalize *norm = WCBE_FDVC_NORMALIZE(GST_OBJECT_PARENT(pad));

    guint num_partprods = (norm->nbits == WCBE_7BITS ? 4 : 1);
    guint partprod = num_partprods;
    GstBuffer **save_buffer = NULL;
    guint i;
    for (i = 0; save_buffer == NULL && i < num_partprods; ++i) {
        guint j;
        for (j = 0; save_buffer == NULL
                 && j < WCBE_FDVC_NORMALIZE_NUM_PADS; ++j) {
            if (pad == norm->sinkpads[i][j]) {
                switch (j) {
                  case WCBE_FDVC_NORMALIZE_WEIGHTS_PAD_INDEX:
                    save_buffer = &(norm->dvs[i]);
                    partprod = i;
                    break;
                  case WCBE_FDVC_NORMALIZE_LAGS_PAD_INDEX:
                    save_buffer = &(norm->lags[i]);
                    partprod = i;
                    break;
                  default:
                    g_assert_not_reached();
                    break;
                }
            }
        }
    }
    g_return_val_if_fail(partprod < num_partprods, GST_FLOW_ERROR);
    g_return_val_if_fail(save_buffer != NULL, GST_FLOW_ERROR);

    if (norm->last_partprod == -1) {
        norm->last_partprod = partprod;
    }
    if (partprod != norm->last_partprod) {
        g_debug("Switched partial product");
        for (i = 0; i < num_partprods; ++i) {
            if (norm->lags[i] != NULL) {
                gst_buffer_unref(norm->lags[i]);
            }
            if (norm->dvs[i] != NULL) {
                gst_buffer_unref(norm->dvs[i]);
            }
        }
    }
    *save_buffer = buffer;

    // First, we compute quotients for all of the partial products,
    // then, only if all quotients were successfully obtained, do we
    // push out the quotient buffers. This provides at least a bit of
    // safety in keeping the partial product streams synchronized.
    if (norm->lags[partprod] != NULL && norm->dvs[partprod] != NULL) {
        g_warn_if_fail(!norm->is_normalized[partprod]);
        GstBuffer *lags = norm->lags[partprod];
        GstBuffer *dvs = norm->dvs[partprod];
        if (G_UNLIKELY(GST_BUFFER_TIMESTAMP(lags)
                       != GST_BUFFER_TIMESTAMP(dvs))) {
            g_debug("timestamp mismatch");
        }
        GstStructure *lags_s =
            gst_caps_get_structure(GST_BUFFER_CAPS(lags), 0);
        GstStructure *dv_s =
            gst_caps_get_structure(GST_BUFFER_CAPS(dvs), 0);
        GQuark lags_id, dv_id;
        gboolean ok =
            gst_structure_get_uint(lags_s, WCBE_PRODUCT_ID_PROPERTY,
                                   &lags_id)
            && gst_structure_get_uint(dv_s, WCBE_PRODUCT_ID_PROPERTY,
                                      &dv_id)
            && (lags_id == dv_id);
        if (G_UNLIKELY(!ok)) {
            g_critical("'lags' product id differs from 'dv' product id");
            return GST_FLOW_ERROR;
        }
        guint lags_partprod, dv_partprod;
        ok =
            gst_structure_get_uint(lags_s, WCBE_PARTIAL_PRODUCT_PROPERTY,
                                   &lags_partprod)
            && gst_structure_get_uint(dv_s, WCBE_PARTIAL_PRODUCT_PROPERTY,
                                      &dv_partprod)
            && (lags_partprod == dv_partprod);
        if (G_UNLIKELY(!ok)) {
            g_critical("'lags' partial product id differs from 'dv' "
                       "partial product id");
            return GST_FLOW_ERROR;
        }
        ok = normalize(lags, dvs, &(norm->lags[partprod]),
                       &(norm->dvs[partprod]));
        norm->is_normalized[partprod] = TRUE;
        norm->null_quotients = norm->null_quotients || !ok;
    }

    gboolean all_normalized = TRUE;
    for (i = 0; all_normalized && i < num_partprods; ++i) {
        all_normalized = norm->is_normalized[i];
    }

    if (all_normalized) {
        if (G_LIKELY(!norm->null_quotients)) {
            for (i = 0; i < num_partprods; ++i) {
                GstFlowReturn rc = GST_FLOW_ERROR;
                GstPad *nmlags_pad =
                    norm->srcpads[i][WCBE_FDVC_NORMALIZE_LAGS_PAD_INDEX];
                if (nmlags_pad != NULL) {
                    if (gst_pad_set_caps(nmlags_pad,
                                         GST_BUFFER_CAPS(norm->lags[i]))) {
                        rc = gst_pad_push(nmlags_pad, norm->lags[i]);
                    }
                    else {
                        g_debug("Failed to set pad caps on nmlags pad");
                        gst_buffer_unref(norm->lags[i]);
                    }
                }
                else {
                    gst_buffer_unref(norm->lags[i]);
                }
                if (rc != GST_FLOW_OK) {
                    result = rc;
                }
                GstPad *wgts_pad =
                    norm->srcpads[i][WCBE_FDVC_NORMALIZE_WEIGHTS_PAD_INDEX];
                if (wgts_pad != NULL) {
                    if (gst_pad_set_caps(wgts_pad,
                                         GST_BUFFER_CAPS(norm->dvs[i]))) {
                        rc = gst_pad_push(wgts_pad, norm->dvs[i]);
                    }
                    else {
                        g_debug("Failed to set pad caps on wgts pad");
                        gst_buffer_unref(norm->dvs[i]);
                    }
                }
                else {
                    gst_buffer_unref(norm->dvs[i]);
                }
                if (rc != GST_FLOW_OK) {
                    result = rc;
                }
                norm->lags[i] = NULL;
                norm->dvs[i] = NULL;
                norm->is_normalized[i] = FALSE;
                g_atomic_int_inc((gint *)&(norm->num_normalized));
            }
        }
        else {
            for (i = 0; i < num_partprods; ++i) {
                if (norm->lags[i] != NULL) {
                    gst_buffer_unref(norm->lags[i]);
                    norm->lags[i] = NULL;
                }
                if (norm->dvs[i] != NULL) {
                    gst_buffer_unref(norm->dvs[i]);
                    norm->dvs[i] = NULL;
                }
                norm->is_normalized[i] = FALSE;
                g_atomic_int_inc((gint *)&(norm->num_dropped));
            }
        }
        norm->last_partprod = -1;
        norm->null_quotients = FALSE;
    }

    return result;
}

static GstPad *
wcbe_fdvc_normalize_request_new_pad(GstElement *element, GstPadTemplate *templ,
                                    const gchar* name)
{
    GstPad *result = NULL;
    WcbeFdvcNormalize *norm = WCBE_FDVC_NORMALIZE(element);
    GstState state;
    gst_element_get_state(element, &state, NULL, 0);
    gboolean active = (state >= GST_STATE_PAUSED);

    // get pad name prefix and subprod number from 'name'
    const gchar *name_prefix = NULL;
    if (strncmp(name, lags_pad_base_name,
                sizeof(lags_pad_base_name) - 1) == 0) {
        name_prefix = lags_pad_base_name;
    }
    else if (strncmp(name, datavalid_pad_base_name,
                     sizeof(datavalid_pad_base_name) - 1) == 0) {
        name_prefix = datavalid_pad_base_name;
    }
    else if (strncmp(name, normalized_lags_pad_base_name,
                     sizeof(normalized_lags_pad_base_name) - 1) == 0) {
        name_prefix = normalized_lags_pad_base_name;
    }
    else if (strncmp(name, weights_pad_base_name,
                     sizeof(weights_pad_base_name) - 1) == 0) {
        name_prefix = weights_pad_base_name;
    }

    if (name_prefix != NULL) {
        gint subprod_index = -1;
        guint64 idx = g_ascii_strtoull(name + strlen(name_prefix), NULL, 0);
        if (idx != G_MAXUINT64) {
            subprod_index = idx;
        }

        // check that pad_index and subprod_index are valid before adding
        // pad
        GstPad **pad = NULL;
        if (subprod_index != -1
            && ((norm->nbits == WCBE_4BITS && subprod_index < 1)
                || (norm->nbits == WCBE_7BITS && subprod_index < 4))) {
            if (name_prefix == lags_pad_base_name) {
                pad = &(norm->sinkpads[subprod_index]
                        [WCBE_FDVC_NORMALIZE_LAGS_PAD_INDEX]);
            }
            else if (name_prefix == datavalid_pad_base_name) {
                pad = &(norm->sinkpads[subprod_index]
                        [WCBE_FDVC_NORMALIZE_WEIGHTS_PAD_INDEX]);
            }
            else if (name_prefix == normalized_lags_pad_base_name) {
                pad = &(norm->srcpads[subprod_index]
                        [WCBE_FDVC_NORMALIZE_LAGS_PAD_INDEX]);
            }
            else {
                g_assert(name_prefix == weights_pad_base_name);
                pad = &(norm->srcpads[subprod_index]
                        [WCBE_FDVC_NORMALIZE_WEIGHTS_PAD_INDEX]);
            }
        }
        if (pad != NULL) {
            *pad = gst_pad_new_from_template(templ, name);
            if (*pad != NULL) {
                gst_pad_set_internal_link_function(
                    *pad, wcbe_fdvc_normalize_intlink);
                if (name_prefix == lags_pad_base_name
                    || name_prefix == datavalid_pad_base_name) {
                    gst_pad_set_chain_function(
                        *pad, wcbe_fdvc_normalize_chain);
                }
                if (gst_pad_set_active(*pad, active)
                    && gst_element_add_pad(element, *pad)) {
                    result = *pad;
                    if (gst_pad_get_direction(*pad) == GST_PAD_SINK) {
                        gst_pad_set_event_function(
                            *pad, wcbe_fdvc_normalize_event);
                    }
                }
                else {
                    g_debug("Failed to add pad to WcbeFdvcNormalize instance");
                    gst_object_unref(GST_OBJECT(*pad));
                    *pad = NULL;
                }
            }
            else {
                g_debug("Failed get new pad from request pad template");
            }
        }
    }
    return result;
}

static void
wcbe_fdvc_normalize_release_pad(GstElement *element, GstPad *pad)
{
    WcbeFdvcNormalize *norm = WCBE_FDVC_NORMALIZE(element);
    GstPad **padp = NULL;
    guint i;
    for (i = 0; padp == NULL && i < 4; ++i) {
        guint j;
        for (j = 0; padp == NULL && j < WCBE_FDVC_NORMALIZE_NUM_PADS; ++j) {
            if (norm->sinkpads[i][j] == pad) {
                padp = &(norm->sinkpads[i][j]);
            }
            else if (norm->srcpads[i][j] == pad) {
                padp = &(norm->srcpads[i][j]);
            }
        }
    }
    if (padp != NULL) {
        gst_element_remove_pad(element, pad);
        *padp = NULL;
    }
}

static GList *
wcbe_fdvc_normalize_intlink(GstPad *pad)
{
    WcbeFdvcNormalize *norm = WCBE_FDVC_NORMALIZE(GST_OBJECT_PARENT(pad));

    GList *result = NULL;
    guint i;
    GstPad **mate = NULL;
    for (i = 0; mate == NULL && i < 4; ++i) {
        guint j;
        for (j = 0; mate == NULL && j < WCBE_FDVC_NORMALIZE_NUM_PADS; ++j) {
            if (norm->sinkpads[i][j] == pad) {
                mate = &(norm->srcpads[i][j]);
            }
            if (norm->srcpads[i][j] == pad) {
                mate = &(norm->sinkpads[i][j]);
            }
        }
    }
    if (*mate != NULL) {
        result = g_list_append(result, *mate);
    }
    return result;
}

static gboolean
wcbe_fdvc_normalize_event(GstPad *pad, GstEvent *event)
{
    if (GST_EVENT_TYPE(event) == GST_EVENT_NEWSEGMENT) {
        WcbeFdvcNormalize *norm = WCBE_FDVC_NORMALIZE(gst_pad_get_parent(pad));
        g_atomic_int_set(&(norm->num_normalized), 0);
        g_atomic_int_set(&(norm->num_dropped), 0);
        gst_object_unref(norm);
    }
    gboolean result = gst_pad_event_default(pad, event);
    return result;
}

static gboolean
normalize(GstBuffer *lags, GstBuffer *dvs, GstBuffer **normalized,
          GstBuffer **weights)
{
    g_return_val_if_fail(lags != NULL, FALSE);
    g_return_val_if_fail(dvs != NULL, FALSE);
    g_return_val_if_fail(normalized != NULL, FALSE);
    g_return_val_if_fail(weights != NULL, FALSE);

    GstStructure *lags_caps_sr =
        gst_caps_get_structure(GST_BUFFER_CAPS(lags), 0);
    GstStructure *dvs_caps_sr =
        gst_caps_get_structure(GST_BUFFER_CAPS(dvs), 0);

    guint lags_nrows;
    gst_structure_get_uint(lags_caps_sr, WCBE_ARRAY_NUM_ROWS_PROPERTY,
                           &lags_nrows);
    guint dvs_nrows;
    gst_structure_get_uint(dvs_caps_sr, WCBE_ARRAY_NUM_ROWS_PROPERTY,
                           &dvs_nrows);
    if (lags_nrows != dvs_nrows) {
        g_warning("Number of rows in lags and DV input buffers in"
                  "WcbeFdvcNormalize is unequal");
        return FALSE;
    }

    const gchar *lags_type_str =
        gst_structure_get_string(lags_caps_sr, WCBE_ELEMENT_TYPE_PROPERTY);
    element_type_t lags_type = ELEMENT_TYPE_INVALID;
    gsize lags_elt_size = 0;
    if (strcmp(lags_type_str, "gint32") == 0) {
        lags_type = ELEMENT_TYPE_INT32;
        lags_elt_size = sizeof(gint32);
    }
    else if (strcmp(lags_type_str, "gint64") == 0) {
        lags_type = ELEMENT_TYPE_INT64;
        lags_elt_size = sizeof(gint64);
    }

    const gchar *dvs_type_str =
        gst_structure_get_string(dvs_caps_sr, WCBE_ELEMENT_TYPE_PROPERTY);
    element_type_t dvs_type = ELEMENT_TYPE_INVALID;
    if (strcmp(dvs_type_str, "gint32") == 0) {
        dvs_type = ELEMENT_TYPE_INT32;
    }
    else if (strcmp(dvs_type_str, "gint64") == 0) {
        dvs_type = ELEMENT_TYPE_INT64;
    }

    if (lags_type == ELEMENT_TYPE_INVALID
        || dvs_type == ELEMENT_TYPE_INVALID) {
        g_warning("Invalid input element type(s) in WcbeFdvcNormalize");
        return FALSE;
    }

    // we just overwrite the data in the input lags buffer since
    // sizeof(gfloat) <= sizeof(gint32) (and sizeof(gfloat) <=
    // sizeof(gint64))
    g_assert(sizeof(gfloat) <= sizeof(gint32));
    *normalized = gst_buffer_make_writable(lags);
    // also overwrite the data in the input dvs buffer since weights
    // are of the same type as dvs
    *weights = gst_buffer_make_writable(dvs);

    // read a few properties from the input buffers needed for looping
    // over the elements
    guint framesz;
    gst_structure_get_uint(dvs_caps_sr, WCBE_FRAME_SIZE_PROPERTY, &framesz);
    // Since framesz is the number of complex lags per frame, but
    // lags_ncols will be the twice that number (NUM_COLS counts
    // number of primitive data elements), we must multiply framesz by
    // two so that it counts the number of primitive values per frame.
    framesz *= 2;

    // get number of columns and row length of input arrays
    guint lags_ncols;
    gst_structure_get_uint(lags_caps_sr, WCBE_ARRAY_NUM_COLS_PROPERTY,
                           &lags_ncols);
    guint lags_rowlen = lags_ncols;
    gst_structure_get_uint(lags_caps_sr, WCBE_ARRAY_ROW_LENGTH_PROPERTY,
                           &lags_rowlen);
    guint lags_row_excess = lags_rowlen - lags_ncols;

    guint dvs_ncols;
    gst_structure_get_uint(dvs_caps_sr, WCBE_ARRAY_NUM_COLS_PROPERTY,
                           &dvs_ncols);
    guint dvs_rowlen = dvs_ncols;
    gst_structure_get_uint(dvs_caps_sr, WCBE_ARRAY_ROW_LENGTH_PROPERTY,
                           &dvs_rowlen);
    guint dvs_row_excess = dvs_rowlen - dvs_ncols;

    // Size of wgt (output) elements equals size of dv (input)
    // elements, so layout of output wgts array is the same as the
    // input dvs array; however, each row in the output vector has half
    // the number of elements as the corresponding row in the input
    // vector.
    guint wgts_row_excess = dvs_row_excess + dvs_ncols / 2;

    // Size of nmlag (output) elements is less than or equal to lags
    // (input) elements, so the layout may be different in the two
    // arrays. What's important is that the output array pointer not
    // get ahead of the input array pointer, and that the row
    // alignment remains unchanged. What we do here, then, is to keep
    // every row starting at the same place in both input and output
    // arrays.
    guint nmlags_row_excess =
        (lags_elt_size / sizeof(gfloat)) * lags_rowlen - lags_ncols;

    // write to output buffers
#define NORM_FDVC(lagstype, dvstype, wgtstype, ftype)            \
    void norm_fdvc_##lagstype##_##dvstype##_##wgtstype(void) {   \
        lagstype *lag = (lagstype *)GST_BUFFER_DATA(lags);       \
        dvstype *dv = (dvstype *)GST_BUFFER_DATA(dvs);           \
        gfloat *nmlag = (gfloat *)GST_BUFFER_DATA(*normalized);  \
        wgtstype *wgt = (wgtstype *)GST_BUFFER_DATA(*weights);    \
        guint i;                                                 \
        for (i = 0; i < lags_nrows; ++i) {                       \
            guint j = 0;                                         \
            while (j < lags_ncols) {                             \
                guint k;                                         \
                if (*dv == 0) {                                  \
                    for (k = 0; k < framesz; ++k) {              \
                        *nmlag++ = 0.0f;                         \
                    }                                            \
                    lag += framesz;                              \
                }                                                \
                else {                                           \
                    ftype divisor = ABS(*dv);                    \
                    for (k = 0; k < framesz; ++k) {              \
                        *nmlag++ = ((ftype)*lag++) / divisor;    \
                    }                                            \
                }                                                \
                j += framesz;                                    \
                *wgt++ = ABS(*dv);                               \
                dv += 2;                                         \
            }                                                    \
           lag += lags_row_excess;                               \
           nmlag += nmlags_row_excess;                           \
           dv += dvs_row_excess;                                 \
           wgt += wgts_row_excess;                               \
        }                                                        \
    }

    NORM_FDVC(gint32, gint32, guint32, gfloat)
    NORM_FDVC(gint32, gint64, guint64, gdouble)
    NORM_FDVC(gint64, gint32, guint32, gdouble)
    NORM_FDVC(gint64, gint64, guint64, gdouble)
    if (lags_type == ELEMENT_TYPE_INT32) {
        if (dvs_type == ELEMENT_TYPE_INT32) {
            norm_fdvc_gint32_gint32_guint32();
        }
        else { // dvs_type == ELEMENT_TYPE_INT64
            norm_fdvc_gint32_gint64_guint64();
        }
    }
    else { // lags_type == ELEMENT_TYPE_INT64
        if (dvs_type == ELEMENT_TYPE_INT32) {
            norm_fdvc_gint64_gint32_guint32();
        }
        else { // dvs_type == ELEMENT_TYPE_INT64
            norm_fdvc_gint64_gint64_guint64();
        }
    }

    // set caps of output buffers
    GST_BUFFER_CAPS(*normalized) =
        gst_caps_make_writable(GST_BUFFER_CAPS(*normalized));
    gst_caps_set_simple(GST_BUFFER_CAPS(*normalized),
                        WCBE_ELEMENT_TYPE_PROPERTY, G_TYPE_STRING, "gfloat",
                        WCBE_ARRAY_ROW_LENGTH_PROPERTY, G_TYPE_UINT,
                        lags_ncols + nmlags_row_excess,
                        NULL);

    GST_BUFFER_CAPS(*weights) =
        gst_caps_make_writable(GST_BUFFER_CAPS(*weights));
    const gchar *wgtstype;
    if (dvs_type == ELEMENT_TYPE_INT32) {
        wgtstype = "guint32";
    }
    else {
        wgtstype = "guint64";
    }
    gst_caps_set_simple(
        GST_BUFFER_CAPS(*weights),
        WCBE_ELEMENT_TYPE_PROPERTY, G_TYPE_STRING, wgtstype,
        WCBE_ARRAY_NUM_COLS_PROPERTY, G_TYPE_UINT, dvs_ncols / 2,
        WCBE_ARRAY_ROW_LENGTH_PROPERTY, G_TYPE_UINT,
        dvs_ncols + dvs_row_excess,
        NULL);
    return TRUE;
}
