#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <stdio.h>
#include <errno.h>

#include <wcbe_fft.h>
#include <wcbe.h>
#include <wcbe_util.h>
#include <user_marshal.h>

#include <fftw3.h>

#define FFT_DIRECTION FFTW_BACKWARD

// Define the following when all of the channels, both positive and
// negative frequencies, should be written out. Otherwise, only the
// positive frequencies are written out.
#undef ALL_CHANNELS

GST_DEBUG_CATEGORY_STATIC(wcbe_fft_debug);
#define GST_CAT_DEFAULT wcbe_fft_debug

/* Filter signals and args */
enum
{
    /* FILL ME */
    LAST_SIGNAL
};

enum
{
    ARG_0,
    ARG_NUM_FFT
};

typedef struct plan_key_t {
    guint num_bins;
    guint vlen;
    guint rowlen;
} plan_key_t;

#define ALIGNMENT 16
#define ELEMENT_TYPE gfloat

#define FFT_STATIC_CAPS                           \
     WCBE_VECTOR_MEDIA_TYPE                       \
     ","WCBE_ELEMENT_TYPE_PROPERTY"="             \
        WCBE_ELEMENT_TYPE_PROPERTY_TYPE "gfloat"  \
     ","WCBE_ALIGNMENT_PROPERTY"="                \
        WCBE_ALIGNMENT_PROPERTY_TYPE "16"         \
  ";"WCBE_ARRAY_MEDIA_TYPE                        \
    ","WCBE_ELEMENT_TYPE_PROPERTY"="              \
       WCBE_ELEMENT_TYPE_PROPERTY_TYPE "gfloat"   \
    ","WCBE_ALIGNMENT_PROPERTY"="                 \
       WCBE_ALIGNMENT_PROPERTY_TYPE "16"

static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE(
    "sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS(FFT_STATIC_CAPS)
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE(
    "src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS(FFT_STATIC_CAPS)
    );

GST_BOILERPLATE(WcbeFft, wcbe_fft, GstBaseTransform,
                GST_TYPE_BASE_TRANSFORM);

static void wcbe_fft_finalize(GObject *gobject);
static void wcbe_fft_get_property(GObject *object, guint prop_id,
                                  GValue *value, GParamSpec *pspec);
static GstFlowReturn wcbe_fft_transform_ip(GstBaseTransform *trans,
                                           GstBuffer *buffer);
static GstCaps *wcbe_fft_transform_caps(GstBaseTransform *trans,
                                        GstPadDirection direction,
                                        GstCaps *caps);
static gboolean wcbe_fft_event(GstBaseTransform *trans, GstEvent *event);
static guint plan_key_hash(gconstpointer key);
static gboolean plan_key_equal(gconstpointer a, gconstpointer b);

static void
wcbe_fft_base_init(gpointer gclass)
{
    GST_DEBUG_CATEGORY_INIT(wcbe_fft_debug, "wcbe_fft_debug",
                            0, "wcbe_fft debug");
    static GstElementDetails element_details = {
        "WIDAR FFT",
        "Transform/WIDAR",
        "Perform 1-D complex fast Fourier transform on data streams",
        "Martin Pokorny <mpokorny@nrao.edu>"
    };

    GstElementClass *element_class = GST_ELEMENT_CLASS(gclass);

    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&sink_factory));
    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&src_factory));
    gst_element_class_set_details(element_class, &element_details);

    gint rc = fftwf_import_system_wisdom();
    if (rc == 0) {
        g_message("Failed to import FFTWF system wisdom file: "
                  "FFTWF performance will be poor.");
    }
}

/* initialize the plugin's class */
static void
wcbe_fft_class_init(WcbeFftClass * klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    GstBaseTransformClass *gstbasetrans_class =
        GST_BASE_TRANSFORM_CLASS(klass);

    gobject_class->finalize = GST_DEBUG_FUNCPTR(wcbe_fft_finalize);
    gobject_class->get_property =
        GST_DEBUG_FUNCPTR(wcbe_fft_get_property);

    g_object_class_install_property(
        gobject_class, ARG_NUM_FFT,
        g_param_spec_uint("num-fft", "Number transforms",
                          "Number of lag sets subjected to FFT",
                          0, G_MAXUINT, 0, G_PARAM_READABLE));

    gstbasetrans_class->transform_ip =
        GST_DEBUG_FUNCPTR(wcbe_fft_transform_ip);
    gstbasetrans_class->transform_caps =
        GST_DEBUG_FUNCPTR(wcbe_fft_transform_caps);
    gstbasetrans_class->event = GST_DEBUG_FUNCPTR(wcbe_fft_event);
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
wcbe_fft_init(WcbeFft *fft, WcbeFftClass *gclass)
{
    fft->fftwf_plans =
        g_hash_table_new_full(plan_key_hash, plan_key_equal, g_free,
                              (GDestroyNotify)fftwf_destroy_plan);
    fft->num_fft = 0;
}

static void
wcbe_fft_finalize(GObject *gobject)
{
    WcbeFft *fft = WCBE_FFT(gobject);
    if (fft->fftwf_plans != NULL) {
        g_hash_table_destroy(fft->fftwf_plans);
    }
}

static void
wcbe_fft_get_property(GObject *object, guint prop_id, GValue *value,
                      GParamSpec *pspec)
{
    WcbeFft *fft = WCBE_FFT(object);
    switch (prop_id) {
      case ARG_NUM_FFT:
        g_value_set_uint(value, g_atomic_int_get((gint *)&(fft->num_fft)));
        break;
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static GstFlowReturn
wcbe_fft_transform_ip(GstBaseTransform *trans, GstBuffer *buf)
{
    WcbeFft *fft = WCBE_FFT(trans);

    GstStructure *s = gst_caps_get_structure(GST_BUFFER_CAPS(buf), 0);
    plan_key_t key = {.num_bins = 0, .vlen = 0, .rowlen = 0};
    if (gst_structure_has_name(s, WCBE_VECTOR_MEDIA_TYPE)) {
        key.num_bins = 1;
        gst_structure_get_uint(s, WCBE_VECTOR_LENGTH_PROPERTY, &(key.vlen));
    }
    else {
        gst_structure_get_uint(s, WCBE_ARRAY_NUM_ROWS_PROPERTY,
                               &(key.num_bins));
        gst_structure_get_uint(s, WCBE_ARRAY_NUM_COLS_PROPERTY, &(key.vlen));
        gst_structure_get_uint(s, WCBE_ARRAY_ROW_LENGTH_PROPERTY,
                               &(key.rowlen));
    }
    if (key.num_bins == 0 || key.vlen == 0) {
        return GST_BASE_TRANSFORM_FLOW_DROPPED;
    }
    // Handling NUM_COLS and ROW_LENGTH is a bit tricky because 1)
    // they count primitive values in the array/vector, 2) the
    // transform_caps function applies the truncation needed to
    // eliminate the negative frequencies from the result (affecting
    // only NUM_ROWS), except in the case that ALL_CHANNELS has been
    // defined, and 3) the transform_caps function has been called
    // before this function. What this all means is that key.vlen at
    // this point holds the number of primitive values in the output,
    // which, when ALL_CHANNELS is not defined, equals the number of
    // complex values in the input; and when ALL_CHANNELS is defined,
    // key.vlen is twice the number of complex values in the input.
    // OTOH, key.rowlen is equal to the number of primitive values in
    // the input and output, which is twice the number of complex
    // values. We change key.rowlen to count the number of complex
    // values, since that is what's needed to create the FFTW plan.
#ifdef ALL_CHANNELS
    key.vlen /= 2;
#endif
    key.rowlen /= 2;

    fftwf_complex *array = (fftwf_complex *)GST_BUFFER_DATA(buf);
    if ((gsize)array % 16 != 0) {
        g_debug("Lags array is not aligned");
        return GST_BASE_TRANSFORM_FLOW_DROPPED;
    }
    fftwf_plan plan = g_hash_table_lookup(fft->fftwf_plans, &key);
    if (G_UNLIKELY((void *)plan == NULL)) {
        plan_key_t *k = g_memdup(&key, sizeof(plan_key_t));
        // The FFTW_ESTIMATE flag is used so that little time is spent
        // choosing a plan; therefore it pays to create system wisdom
        // prior to running this code on any given system. If the
        // planner flag is ever "upgraded" from FFTW_ESTIMATE, note
        // that the planning process would then _destroy_ the data in
        // the input array, so we'd have to use an array other than
        // that in "buf".
        g_debug("generating FFTWF plan for %u x %u (%u)",
                key.num_bins, key.vlen, key.rowlen);
        plan = fftwf_plan_many_dft(
            1, (gint *)&(key.vlen), key.num_bins,
            array, NULL, 1, key.rowlen, array, NULL, 1, key.rowlen,
            FFT_DIRECTION, FFTW_ESTIMATE);
        if (G_LIKELY(plan != NULL)) {
            g_hash_table_insert(fft->fftwf_plans, k, plan);
        }
        else {
            g_warning("Failed to generate FFTWF plan for vlen=%u, num_bin=%u",
                      key.vlen, key.num_bins);
            g_free(k);
            return GST_BASE_TRANSFORM_FLOW_DROPPED;
        }
    }
    fftwf_execute_dft(plan, array, array);
#if !defined(ALL_CHANNELS) && FFT_DIRECTION == FFTW_FORWARD
    // the frequencies we want are reversed, and in the second half,
    // so we copy and reverse them into the first half
    fftwf_complex *front = array;
    fftwf_complex *back = array + key.vlen - 1;
    while (back > front) {
        *front++ = *back--;
    }
#endif
    g_atomic_int_inc((gint *)&(fft->num_fft));
    return GST_FLOW_OK;
}

static GstCaps *
wcbe_fft_transform_caps(GstBaseTransform *trans, GstPadDirection direction,
                        GstCaps *caps)
{
    GstCaps *result = gst_caps_copy(caps);
#ifndef ALL_CHANNELS
    GstStructure *s = gst_caps_get_structure(result, 0);
    const gchar *prop;
    if (gst_structure_has_name(s, WCBE_VECTOR_MEDIA_TYPE)) {
        prop = WCBE_VECTOR_LENGTH_PROPERTY;
    }
    else {
        prop = WCBE_ARRAY_NUM_COLS_PROPERTY;
    }
    guint numcols;
    if (gst_structure_get_uint(s, prop, &numcols)) {
        switch (direction) {
          case GST_PAD_SRC:
            numcols *= 2;
            break;
          case GST_PAD_SINK:
            numcols /= 2;
            break;
          default:
            break;
        }
        gst_caps_set_simple(result, prop, G_TYPE_UINT, numcols, NULL);
    }
#endif
    return result;
}

static gboolean
wcbe_fft_event(GstBaseTransform *trans, GstEvent *event)
{
    if (GST_EVENT_TYPE(event) == GST_EVENT_NEWSEGMENT) {
        WcbeFft *fft = WCBE_FFT(trans);
        g_atomic_int_set((gint *)&(fft->num_fft), 0);
    }
    return TRUE;
}

static guint
plan_key_hash(gconstpointer key)
{
    const plan_key_t *plan = key;
    return (g_int_hash(&(plan->num_bins)) ^ g_int_hash(&(plan->vlen))
            ^ g_int_hash(&(plan->rowlen)));
}

static gboolean
plan_key_equal(gconstpointer a, gconstpointer b)
{
    const plan_key_t *pa = a;
    const plan_key_t *pb = b;
    return ((pa->num_bins == pb->num_bins) && (pa->vlen == pb->vlen)
            && (pa->rowlen == pb->rowlen));
}
