#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>

#include <wcbe_framedump.h>
#include <wcbe.h>
#include <wcbe_sort.h>
#include <wcbe_util.h>
#include <dump_frame.h>
#include <dump_bin_frame.h>

GST_DEBUG_CATEGORY_STATIC(wcbe_framedump_debug);
#define GST_CAT_DEFAULT wcbe_framedump_debug

/* Filter signals and args */
enum
{
    /* FILL ME */
    LAST_SIGNAL
};

enum
{
    ARG_0,
    ARG_FORMAT,
    ARG_DO_OUTPUT,
    ARG_MAX_FILE_SIZE,
    ARG_OUTPUT_DIR
};

#define DEFAULT_FORMAT WCBE_FRAMEDUMP_BINARY
#define DEFAULT_MAX_FILE_SIZE (25 * (1 << 20))
#define MIN_MAX_FILE_SIZE (100 * (1 << 10))
#define MAX_MAX_FILE_SIZE (1 << 30)

static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE(
    "sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS(WCBE_WIDAR_FRAME_MEDIA_TYPE)
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE(
    "src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS(WCBE_WIDAR_FRAME_MEDIA_TYPE)
    );

GST_BOILERPLATE(WcbeFramedump, wcbe_framedump, GstBaseTransform,
                GST_TYPE_BASE_TRANSFORM);

static void wcbe_framedump_finalize(GObject *object);
static void wcbe_framedump_set_property(
    GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec);
static void wcbe_framedump_get_property(
    GObject *object, guint prop_id, GValue *value, GParamSpec *pspec);
static gboolean wcbe_framedump_start(GstBaseTransform *trans);
static gboolean wcbe_framedump_stop(GstBaseTransform *trans);
static GstFlowReturn wcbe_framedump_transform_ip(GstBaseTransform *trans,
                                                 GstBuffer *buffer);
static void set_format(WcbeFramedump *framedump,
                       WcbeFramedumpFormat format, gboolean init);
static void set_do_output(WcbeFramedump *framedump, gboolean do_output);
static void set_output_dir(WcbeFramedump *framedump, const gchar *dir);

static void
wcbe_framedump_base_init(gpointer gclass)
{
    GST_DEBUG_CATEGORY_INIT(wcbe_framedump_debug, "wcbe_framedump_debug",
                            0, "wcbe_framedump debug");
    static GstElementDetails element_details = {
        "WIDAR frame dump",
        "Transform/WIDAR",
        "Dump WIDAR binary frames to files",
        "Martin Pokorny <mpokorny@nrao.edu>"
    };
    GstElementClass *element_class = GST_ELEMENT_CLASS(gclass);

    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&sink_factory));
    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&src_factory));
    gst_element_class_set_details(element_class, &element_details);
}

/* initialize the plugin's class */
static void
wcbe_framedump_class_init(WcbeFramedumpClass * klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    GstBaseTransformClass *gstbasetrans_class =
        GST_BASE_TRANSFORM_CLASS(klass);

    gobject_class->set_property =
        GST_DEBUG_FUNCPTR(wcbe_framedump_set_property);
    gobject_class->get_property =
        GST_DEBUG_FUNCPTR(wcbe_framedump_get_property);
    gobject_class->finalize =
        GST_DEBUG_FUNCPTR(wcbe_framedump_finalize);

    g_object_class_install_property(
        gobject_class, ARG_DO_OUTPUT,
        g_param_spec_boolean("do-output", "Dump frames",
                             "Produce frame dump files?",
                             FALSE, G_PARAM_READWRITE));
    g_object_class_install_property(
        gobject_class, ARG_FORMAT,
        g_param_spec_enum("format", "Format", "Output format switch",
                          WCBE_TYPE_FRAMEDUMP_FORMAT, DEFAULT_FORMAT,
                          G_PARAM_READWRITE));
    g_object_class_install_property(
        gobject_class, ARG_MAX_FILE_SIZE,
        g_param_spec_int64("max-file-size", "Max file size",
                           "Maximum dump file size",
                           MIN_MAX_FILE_SIZE, MAX_MAX_FILE_SIZE,
                           DEFAULT_MAX_FILE_SIZE, G_PARAM_READWRITE));
    g_object_class_install_property(
        gobject_class, ARG_OUTPUT_DIR,
        g_param_spec_string("output-dir", "File output directory",
                            "Directory to write BLF files into",
                            NULL, G_PARAM_READWRITE));

    gstbasetrans_class->start = GST_DEBUG_FUNCPTR(wcbe_framedump_start);
    gstbasetrans_class->stop = GST_DEBUG_FUNCPTR(wcbe_framedump_stop);
    gstbasetrans_class->transform_ip =
        GST_DEBUG_FUNCPTR(wcbe_framedump_transform_ip);
    gstbasetrans_class->passthrough_on_same_caps = TRUE;
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
wcbe_framedump_init(WcbeFramedump *framedump,
                    WcbeFramedumpClass *gclass)
{
    framedump->close_fn = NULL;
    set_format(framedump, DEFAULT_FORMAT, TRUE);
    framedump->do_output = FALSE;
    framedump->max_file_size = DEFAULT_MAX_FILE_SIZE;
    framedump->output_dir = NULL;
    g_static_mutex_init(&(framedump->output_mutex));
}

static void
wcbe_framedump_finalize(GObject *object)
{
    WcbeFramedump *framedump = WCBE_FRAMEDUMP(object);
    g_static_mutex_free(&(framedump->output_mutex));
    if (framedump->output_dir != NULL) {
        g_free(framedump->output_dir);
    }
}

static void
wcbe_framedump_set_property(GObject *object, guint prop_id,
                            const GValue *value, GParamSpec *pspec)
{
    WcbeFramedump *framedump = WCBE_FRAMEDUMP(object);

    switch (prop_id) {
      case ARG_FORMAT:
        set_format(framedump, g_value_get_enum(value), FALSE);
        break;
      case ARG_DO_OUTPUT:
        set_do_output(framedump, g_value_get_boolean(value));
        break;
      case ARG_MAX_FILE_SIZE: {
        off64_t (*current_size_fn)(void);
        void (*close_fn)(void);
        if (framedump->format == WCBE_FRAMEDUMP_ASCII) {
            current_size_fn = dump_frame_max_file_size;
            close_fn = dump_frame_close_all;
        }
        else {
            current_size_fn = dump_bin_frame_max_file_size;
            close_fn = dump_bin_frame_close_all;
        }
        off64_t new_max = g_value_get_int64(value);
        g_static_mutex_lock(&(framedump->output_mutex));
        if (current_size_fn() >= new_max) {
            close_fn();
        }
        framedump->max_file_size = new_max;
        g_static_mutex_unlock(&(framedump->output_mutex));
        break;
      }
      case ARG_OUTPUT_DIR:
        set_output_dir(framedump, g_value_get_string(value));
        break;
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static void
wcbe_framedump_get_property(GObject *object, guint prop_id,
                            GValue *value, GParamSpec *pspec)
{
    WcbeFramedump *framedump = WCBE_FRAMEDUMP(object);

    switch (prop_id) {
      case ARG_FORMAT:
        g_value_set_enum(value, framedump->format);
        break;
      case ARG_DO_OUTPUT:
        g_value_set_boolean(value, framedump->do_output);
        break;
      case ARG_MAX_FILE_SIZE:
        g_value_set_int64(value, framedump->max_file_size);
        break;
      case ARG_OUTPUT_DIR:
        if (framedump->output_dir != NULL) {
            g_value_set_string(value, framedump->output_dir);
        }
        else {
            gchar *cd = g_get_current_dir();
            g_value_set_string(value, cd);
            g_free(cd);
        }
        break;
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

/* GstBaseSink vmethod implementations */

static gboolean
wcbe_framedump_start(GstBaseTransform *trans)
{
    return TRUE;
}

static gboolean
wcbe_framedump_stop(GstBaseTransform *trans)
{
    WcbeFramedump *framedump = WCBE_FRAMEDUMP(trans);
    g_static_mutex_lock(&(framedump->output_mutex));
    dump_bin_frame_close_all();
    dump_frame_close_all();
    g_static_mutex_unlock(&(framedump->output_mutex));
    return TRUE;
}

static GstFlowReturn
wcbe_framedump_transform_ip(GstBaseTransform *trans, GstBuffer *buf)
{
    WcbeFramedump *framedump = WCBE_FRAMEDUMP(trans);
    g_static_mutex_lock(&(framedump->output_mutex));
    if (framedump->do_output) {
        lf_lag_frame_t *lf = (lf_lag_frame_t *)GST_BUFFER_DATA(buf);
        off64_t current_size = framedump->output_fn(lf->words);
        if (current_size > framedump->max_file_size) {
            framedump->close_fn();
        }
    }
    g_static_mutex_unlock(&(framedump->output_mutex));
    return GST_FLOW_OK;
}

GType
wcbe_framedump_format_get_type(void)
{
    static GType framedump_format_type = 0;
    if (framedump_format_type == 0) {
        static GEnumValue format_types[] = {
            {WCBE_FRAMEDUMP_ASCII, "WCBE_FRAMEDUMP_ASCII", "ASCII"},
            {WCBE_FRAMEDUMP_BINARY, "WCBE_FRAMEDUMP_BINARY", "BINARY"},
            {0, NULL, NULL}
        };
        framedump_format_type =
            g_enum_register_static("WcbeFramedumpFormat", format_types);
    }
    return framedump_format_type;
}

static void
set_format(WcbeFramedump *framedump, WcbeFramedumpFormat format,
           gboolean init)
{
    if (framedump->format != format) {
        g_static_mutex_lock(&(framedump->output_mutex));
        framedump->format = format;
        void (*close_fn)(void) = framedump->close_fn;
        switch (format) {
          case WCBE_FRAMEDUMP_BINARY:
            framedump->output_fn = dump_bin_frame;
            framedump->close_fn = dump_bin_frame_close_all;
            break;
          case WCBE_FRAMEDUMP_ASCII:
            framedump->output_fn = dump_frame;
            framedump->close_fn = dump_frame_close_all;
            break;
          default:
            g_assert_not_reached();
            break;
        }
        if (!init) {
            close_fn();
        }
        g_static_mutex_unlock(&(framedump->output_mutex));
    }
}

static void
set_do_output(WcbeFramedump *framedump, gboolean do_output)
{
    g_static_mutex_lock(&(framedump->output_mutex));
    if (framedump->do_output && !do_output && framedump->close_fn != NULL) {
        framedump->close_fn();
    }
    framedump->do_output = do_output;
    g_static_mutex_unlock(&(framedump->output_mutex));
}

static void
set_output_dir(WcbeFramedump *framedump, const gchar *dir)
{
    if (framedump->output_dir == NULL
        || strcmp(dir, framedump->output_dir) != 0) {
        g_static_mutex_lock(&(framedump->output_mutex));
        if (framedump->close_fn != NULL) {
            framedump->close_fn();
        }
        if (framedump->output_dir != NULL) {
            g_free(framedump->output_dir);
        }
        if (strlen(dir) != 0) {
            framedump->output_dir = g_strdup(dir);
        }
        else {
            framedump->output_dir = g_get_current_dir();
        }
        dump_frame_set_directory(framedump->output_dir);
        dump_bin_frame_set_directory(framedump->output_dir);
        g_static_mutex_unlock(&(framedump->output_mutex));
    }
}
