#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <wcbe_lagset_demux.h>
#include <wcbe.h>
#include <wcbe_enums.h>
#include <wcbe_lagset.h>

#include <string.h>

GST_DEBUG_CATEGORY_STATIC(wcbe_lagset_demux_debug);
#define GST_CAT_DEFAULT wcbe_lagset_demux_debug

enum
{
    ARG_0,
    ARG_CORR_BITS
};

#define DEFAULT_CORR_BITS WCBE_4BITS
#define LAGS_PAD_BASE_NAME "lags"
#define TIMESTAMP_PAD_BASE_NAME "ts"
#define DATAVALID_PAD_BASE_NAME "dv"

static gchar lags_pad_base_name[] = LAGS_PAD_BASE_NAME;
static gchar timestamp_pad_base_name[] = TIMESTAMP_PAD_BASE_NAME;
static gchar datavalid_pad_base_name[] = DATAVALID_PAD_BASE_NAME;

static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE(
    "sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS(WCBE_LAGSET_MEDIA_TYPE));

static GstStaticPadTemplate lags_factory = GST_STATIC_PAD_TEMPLATE(
    LAGS_PAD_BASE_NAME"%d",
    GST_PAD_SRC,
    GST_PAD_REQUEST,
    GST_STATIC_CAPS(
        WCBE_ARRAY_MEDIA_TYPE
        ","WCBE_ELEMENT_TYPE_PROPERTY"="
           WCBE_ELEMENT_TYPE_PROPERTY_TYPE "gint32"
        ","WCBE_ALIGNMENT_PROPERTY"="
           WCBE_ALIGNMENT_PROPERTY_TYPE "16"));

static GstStaticPadTemplate ts_factory = GST_STATIC_PAD_TEMPLATE(
    TIMESTAMP_PAD_BASE_NAME"%d",
    GST_PAD_SRC,
    GST_PAD_REQUEST,
    GST_STATIC_CAPS(WCBE_ARRAY_MEDIA_TYPE
                    ","WCBE_ELEMENT_TYPE_PROPERTY"="
                       WCBE_ELEMENT_TYPE_PROPERTY_TYPE "guint64"));

static GstStaticPadTemplate dv_factory = GST_STATIC_PAD_TEMPLATE(
    DATAVALID_PAD_BASE_NAME"%d",
    GST_PAD_SRC,
    GST_PAD_REQUEST,
    GST_STATIC_CAPS(WCBE_ARRAY_MEDIA_TYPE
                    ","WCBE_ELEMENT_TYPE_PROPERTY"="
                       WCBE_ELEMENT_TYPE_PROPERTY_TYPE "gint32"));

GST_BOILERPLATE(WcbeLagsetDemux, wcbe_lagset_demux, GstElement,
                GST_TYPE_ELEMENT);

static void wcbe_lagset_demux_finalize(GObject *object);
static void wcbe_lagset_demux_get_property(
    GObject *object, guint prop_id, GValue *value, GParamSpec *pspec);
static void wcbe_lagset_demux_set_property(
    GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec);
static GstPad *wcbe_lagset_demux_request_new_pad(
    GstElement *element, GstPadTemplate *templ, const gchar* name);
static void wcbe_lagset_demux_release_pad(GstElement *element, GstPad *pad);
static GstFlowReturn wcbe_lagset_demux_chain(GstPad *pad, GstBuffer *buffer);
static GstFlowReturn chain_one_subprod(
    GstBuffer *buffer, GstPad **pads, GstCaps **caps,
    const wcbe_lagset_header_t *lagset_hdr,
    const wcbe_lagset_frame_header_t *frame_hdrs, const lf_lag_t *lags);
static GstFlowReturn chain_4bit_lagset(WcbeLagsetDemux *demux,
                                       GstBuffer *buffer, GstCaps **caps);
static GstFlowReturn chain_7bit_lagset(WcbeLagsetDemux *demux,
                                       GstBuffer *buffer, GstCaps **caps);
static void free_prodcaps(gpointer data);
static GstCaps **get_prodcaps(GHashTable *prodcaps, wcbe_lagset_header_t *hdr);
static void add_lagset_hdr_props(GstCaps *caps, wcbe_lagset_header_t *hdr);

static void
wcbe_lagset_demux_base_init(gpointer gclass)
{
    GST_DEBUG_CATEGORY_INIT(wcbe_lagset_demux_debug,
                            "wcbe_lagset_demux", 0,
                            "wcbe_lagset_demux debug");
    static GstElementDetails element_details = {
        "WIDAR lagset demuxer",
        "Demuxer/WIDAR",
        "Demultiplex lagset stream into streams for lags (one per partial "
        "product), frame timestamps, "
        "frame data valid counts, and lagset metadata",
        "Martin Pokorny <mpokorny@nrao.edu>"
    };
    GstElementClass *element_class = GST_ELEMENT_CLASS(gclass);

    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&sink_factory));
    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&lags_factory));
    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&ts_factory));
    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&dv_factory));
    gst_element_class_set_details(element_class, &element_details);
}

/* initialize the plugin's class */
static void
wcbe_lagset_demux_class_init(WcbeLagsetDemuxClass *klass)
{
    GObjectClass *gobject_class  = (GObjectClass *)klass;
    GstElementClass *elementclass = GST_ELEMENT_CLASS(klass);

    gobject_class->set_property =
        GST_DEBUG_FUNCPTR(wcbe_lagset_demux_set_property);
    gobject_class->get_property =
        GST_DEBUG_FUNCPTR(wcbe_lagset_demux_get_property);
    gobject_class->finalize =
        GST_DEBUG_FUNCPTR(wcbe_lagset_demux_finalize);

    g_object_class_install_property(
        gobject_class, ARG_CORR_BITS,
        g_param_spec_enum("corr-bits", "Number of bits correlated",
                          "Number of bits in correlation products input "
                          "to the demuxer.",
                          WCBE_CORR_BITS_ENUM_TYPE, DEFAULT_CORR_BITS,
                          G_PARAM_READWRITE | G_PARAM_CONSTRUCT));

    elementclass->request_new_pad =
        GST_DEBUG_FUNCPTR(wcbe_lagset_demux_request_new_pad);
    elementclass->release_pad =
        GST_DEBUG_FUNCPTR(wcbe_lagset_demux_release_pad);
}

static void
wcbe_lagset_demux_init(WcbeLagsetDemux *demux, WcbeLagsetDemuxClass *klass)
{
    gst_element_create_all_pads(GST_ELEMENT(demux));
    demux->sinkpad = gst_element_get_static_pad(GST_ELEMENT(demux), "sink");
    gst_pad_set_bufferalloc_function(demux->sinkpad,
                                     wcbe_vect_pad_buffer_alloc);
    gst_pad_use_fixed_caps(demux->sinkpad);
    gst_pad_set_chain_function(demux->sinkpad, wcbe_lagset_demux_chain);
    GstCaps *caps = gst_caps_new_simple(WCBE_LAGSET_MEDIA_TYPE, NULL);
    gst_pad_set_caps(demux->sinkpad, caps);
    gst_caps_unref(caps);
    gst_object_unref(GST_OBJECT(demux->sinkpad));
    demux->nbits = DEFAULT_CORR_BITS;
    memset(demux->srcpads, 0, sizeof(demux->srcpads));
    demux->prodcaps = g_hash_table_new_full(
        g_direct_hash, g_direct_equal, g_free, free_prodcaps);
}

static void
wcbe_lagset_demux_finalize(GObject *object)
{
    WcbeLagsetDemux *demux = WCBE_LAGSET_DEMUX(object);
    g_hash_table_destroy(demux->prodcaps);
}

static void
wcbe_lagset_demux_get_property(GObject *object, guint prop_id, GValue *value,
                               GParamSpec *pspec)
{
    WcbeLagsetDemux *demux = WCBE_LAGSET_DEMUX(object);
    switch (prop_id) {
      case ARG_CORR_BITS:
        g_value_set_enum(value, demux->nbits);
        break;
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static void
wcbe_lagset_demux_set_property(GObject *object, guint prop_id,
                               const GValue *value, GParamSpec *pspec)
{
    WcbeLagsetDemux *demux = WCBE_LAGSET_DEMUX(object);
    switch (prop_id) {
      case ARG_CORR_BITS: {
        WcbeCorrBits old_nbits = demux->nbits;
        demux->nbits = g_value_get_enum(value);
        // if nbits has changed, release all of the srcpads (we could
        // avoid this if ARG_CORR_BITS were G_PARAM_CONSTRUCT_ONLY,
        // but that doesn't seem to work when gstreamer creates the
        // pipeline)
        if (old_nbits != demux->nbits) {
            guint i;
            for (i = 0; i < G_N_ELEMENTS(demux->srcpads); ++i) {
                guint j;
                for (j = 0; j < G_N_ELEMENTS(demux->srcpads[0]); ++j) {
                    if (demux->srcpads[i][j] != NULL) {
                        // the following calls
                        // wcbe_lagset_demux_release_pad, which will
                        // remove the pad, and set the ptr in the
                        // array to NULL
                        gst_element_release_request_pad(
                            GST_ELEMENT(object), demux->srcpads[i][j]);
                    }
                }
            }
        }
        break;
      }
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static GstFlowReturn
wcbe_lagset_demux_chain(GstPad *pad, GstBuffer *buffer)
{
    GstFlowReturn result = GST_FLOW_OK;
    WcbeLagsetDemux *demux = WCBE_LAGSET_DEMUX(GST_OBJECT_PARENT(pad));
    wcbe_vect_t *lagset = (wcbe_vect_t *)GST_BUFFER_DATA(buffer);
    wcbe_lagset_header_t *lagset_hdr = &wcbe_vect_index(
        lagset, WCBE_LAGSET_HEADER_BLOCK_INDEX, wcbe_lagset_header_t, 0);
    g_return_val_if_fail(lagset_hdr->nbits == demux->nbits, GST_FLOW_ERROR);
    GstCaps **prodcaps = get_prodcaps(demux->prodcaps, lagset_hdr);
    GstCaps **caps = prodcaps;
    while (*caps != NULL) {
        gst_caps_set_simple(
            *caps,
            WCBE_EPOCH_PROPERTY, G_TYPE_UINT, lagset_hdr->timestamp.epoch,
            WCBE_SECONDS_PROPERTY, G_TYPE_UINT, lagset_hdr->timestamp.seconds,
            WCBE_CYCLES_PROPERTY, G_TYPE_UINT,
                                  lagset_hdr->timestamp.clock_cycles,
            NULL);
        ++caps;
    }
    switch (lagset_hdr->nbits) {
      case 4:
        result = chain_4bit_lagset(demux, buffer, prodcaps);
        break;
      case 7:
        result = chain_7bit_lagset(demux, buffer, prodcaps);
        break;
      default:
        break;
    }
    gst_buffer_unref(buffer);
    return result;
}

static GstPad *
wcbe_lagset_demux_request_new_pad(GstElement *element, GstPadTemplate *templ,
                                  const gchar* name)
{
    GstPad *result = NULL;
    WcbeLagsetDemux *demux = WCBE_LAGSET_DEMUX(element);

    if (name == NULL) {
        g_critical("wcbe_lagset_demux_request_new_pad name is NULL");
    }

    // get pad name prefix and subprod number from 'name'
    gint pad_index = -1;
    gint subprod_index = -1;
    if (strncmp(name, lags_pad_base_name,
                sizeof(lags_pad_base_name) - 1) == 0) {
        pad_index = WCBE_LAGSET_DEMUX_LAGS_PAD_INDEX;
        guint64 idx = g_ascii_strtoull(
            name + sizeof(lags_pad_base_name) - 1, NULL, 0);
        if (idx != G_MAXUINT64) {
            subprod_index = idx;
        }
    }
    else if (strncmp(name, timestamp_pad_base_name,
                     sizeof(timestamp_pad_base_name) - 1) == 0) {
        pad_index = WCBE_LAGSET_DEMUX_TIMESTAMP_PAD_INDEX;
        guint64 idx = g_ascii_strtoull(
            name + sizeof(timestamp_pad_base_name) - 1, NULL, 0);
        if (idx != G_MAXUINT64) {
            subprod_index = idx;
        }
    }
    else if (strncmp(name, datavalid_pad_base_name,
                     sizeof(datavalid_pad_base_name) - 1) == 0) {
        pad_index = WCBE_LAGSET_DEMUX_DATAVALID_PAD_INDEX;
        guint64 idx = g_ascii_strtoull(
            name + sizeof(datavalid_pad_base_name) - 1, NULL, 0);
        if (idx != G_MAXUINT64) {
            subprod_index = idx;
        }
    }

    // check that pad_index and subprod_index are valid and unused,
    // then add the pad
    if (pad_index != -1 && subprod_index != -1
        && ((demux->nbits == WCBE_4BITS && subprod_index < 1)
            || (demux->nbits == WCBE_7BITS && subprod_index < 4))
        && demux->srcpads[subprod_index][pad_index] == NULL) {

        GstState state;
        gst_element_get_state(GST_ELEMENT(demux), &state, NULL, 0);
        gboolean active = (state >= GST_STATE_PAUSED);
        GstPad **pad = &(demux->srcpads[subprod_index][pad_index]);
        *pad = gst_pad_new_from_template(templ, name);
        if (*pad != NULL) {
            gst_pad_use_fixed_caps(*pad);
            GstCaps *caps = gst_pad_get_caps(*pad);
            if (caps != NULL) {
                gst_caps_set_simple(
                    caps, WCBE_NUM_BITS_PROPERTY, G_TYPE_UINT, demux->nbits,
                    NULL);
                if (gst_pad_set_caps(*pad, caps)) {
                    gst_caps_unref(caps);
                    if (gst_pad_set_active(*pad, active)) {
                        if (gst_element_add_pad(GST_ELEMENT(demux), *pad)) {
                            result = *pad;
                        }
                        else {
                            GST_DEBUG("Failed to add src pad to "
                                      "WcbeLagsetDemux instance");
                        }
                    }
                    else {
                        GST_DEBUG("Failed to set src pad active state to %d",
                                  active);
                    }
                }
                else {
                    GST_DEBUG("Failed to set WcbeLagsetDemux pad caps "
                              "with 'nbits'");
                }
            }
            else {
                GST_DEBUG("Failed to get WcbeLagsetDemux pad caps");
            }
        }
        else {
            GST_DEBUG("Failed to create new pad from static template "
                      "for WcbeLagsetDemux instance");
        }
    }
    else {
        g_debug("Invalid src pad request for WcbeLagsetDemux.");
    }
    return result;
}

static void
wcbe_lagset_demux_release_pad(GstElement *element, GstPad *pad)
{
    WcbeLagsetDemux *demux = WCBE_LAGSET_DEMUX(element);
    guint i;
    for (i = 0; pad != NULL && i < G_N_ELEMENTS(demux->srcpads); ++i) {
        guint j;
        for (j = 0; pad != NULL && j < G_N_ELEMENTS(demux->srcpads[0]); ++j) {
            if (pad == demux->srcpads[i][j]) {
                gst_element_remove_pad(element, pad);
                demux->srcpads[i][j] = NULL;
                pad = NULL;
            }
        }
    }
}

static GstFlowReturn
chain_one_subprod(GstBuffer *buffer, GstPad **pads, GstCaps **caps,
                  const wcbe_lagset_header_t *lagset_hdr,
                  const wcbe_lagset_frame_header_t *frame_hdrs,
                  const lf_lag_t *lags)
{
    GstFlowReturn result = GST_FLOW_OK;
    GstBuffer *lags_buffer = NULL;
    GstBuffer *ts_buffer = NULL;
    GstBuffer *dv_buffer = NULL;

    // 'lags' buffer
    if (pads[WCBE_LAGSET_DEMUX_LAGS_PAD_INDEX] != NULL
        && gst_pad_set_caps(pads[WCBE_LAGSET_DEMUX_LAGS_PAD_INDEX],
                            caps[WCBE_LAGSET_DEMUX_LAGS_PAD_INDEX])) {
        lags_buffer =
            gst_buffer_create_sub(buffer,
                                  (guint8 *)lags - GST_BUFFER_DATA(buffer),
                                  sizeof(*lags) * lagset_hdr->num_lags_per_bin
                                  * lagset_hdr->num_bins);
        GST_BUFFER_FLAG_UNSET(lags_buffer, GST_BUFFER_FLAG_READONLY);
        gst_buffer_copy_metadata(lags_buffer, buffer,
                                 GST_BUFFER_COPY_TIMESTAMPS);
        gst_buffer_set_caps(lags_buffer,
                            caps[WCBE_LAGSET_DEMUX_LAGS_PAD_INDEX]);
    }

    guint total_frames = lagset_hdr->num_frames_per_bin * lagset_hdr->num_bins;

    // 'ts' buffer
    if (pads[WCBE_LAGSET_DEMUX_TIMESTAMP_PAD_INDEX] != NULL
        && gst_pad_set_caps(pads[WCBE_LAGSET_DEMUX_TIMESTAMP_PAD_INDEX],
                            caps[WCBE_LAGSET_DEMUX_TIMESTAMP_PAD_INDEX])) {
        result = gst_pad_alloc_buffer(
            pads[WCBE_LAGSET_DEMUX_TIMESTAMP_PAD_INDEX], 0,
            total_frames * sizeof(guint64),
            caps[WCBE_LAGSET_DEMUX_TIMESTAMP_PAD_INDEX], &ts_buffer);
        if (G_UNLIKELY(result != GST_FLOW_OK)) {
            g_debug("failed ts buffer alloc");
            goto cleanup_and_return;
        }
        gst_buffer_copy_metadata(ts_buffer, buffer,
                                 GST_BUFFER_COPY_TIMESTAMPS);
    }

    // 'dv' buffer
    if (pads[WCBE_LAGSET_DEMUX_DATAVALID_PAD_INDEX] != NULL
        && gst_pad_set_caps(pads[WCBE_LAGSET_DEMUX_DATAVALID_PAD_INDEX],
                            caps[WCBE_LAGSET_DEMUX_DATAVALID_PAD_INDEX])) {
        result = gst_pad_alloc_buffer(
            pads[WCBE_LAGSET_DEMUX_DATAVALID_PAD_INDEX], 0,
            2 * total_frames * sizeof(gint32),
            caps[WCBE_LAGSET_DEMUX_DATAVALID_PAD_INDEX], &dv_buffer);
        if (G_UNLIKELY(result != GST_FLOW_OK)) {
            g_debug("failed dv buffer alloc");
            goto cleanup_and_return;
        }
        gst_buffer_copy_metadata(dv_buffer, buffer,
                                 GST_BUFFER_COPY_TIMESTAMPS);
    }

    // copy timestamps and dvcounts to buffers
    guint64 *ts = (ts_buffer ? (guint64 *)GST_BUFFER_DATA(ts_buffer) : NULL);
    gint32 *dv = (dv_buffer ? (gint32 *)GST_BUFFER_DATA(dv_buffer) : NULL);
    if (ts != NULL || dv != NULL) {
        const wcbe_lagset_frame_header_t *hdr = frame_hdrs;
        while (total_frames-- > 0) {
            if (ts != NULL) {
                *ts++ = TIMEVALUE(&(hdr->timestamp));
            }
            if (dv != NULL) {
                if (hdr->reverse_conjugate) {
                    *dv++ = -hdr->data_valid_center;
                    *dv++ = -hdr->data_valid_edge;
                }
                else {
                    *dv++ = hdr->data_valid_center;
                    *dv++ = hdr->data_valid_edge;
                }
            }
            ++hdr;
        }
    }

    // push buffers
    if (lags_buffer != NULL) {
        result =
            gst_pad_push(pads[WCBE_LAGSET_DEMUX_LAGS_PAD_INDEX], lags_buffer);
        lags_buffer = NULL;
        if (G_UNLIKELY(result != GST_FLOW_OK)) {
            g_debug("failed to push lags");
            goto cleanup_and_return;
        }
    }
    if (ts_buffer != NULL) {
        result = gst_pad_push(
            pads[WCBE_LAGSET_DEMUX_TIMESTAMP_PAD_INDEX], ts_buffer);
        ts_buffer = NULL;
        if (G_UNLIKELY(result != GST_FLOW_OK)) {
            g_debug("failed to push ts");
            goto cleanup_and_return;
        }
    }
    if (dv_buffer != NULL) {
        result = gst_pad_push(
            pads[WCBE_LAGSET_DEMUX_DATAVALID_PAD_INDEX], dv_buffer);
        dv_buffer = NULL;
        if (G_UNLIKELY(result != GST_FLOW_OK)) {
            g_debug("failed to push dv");
            goto cleanup_and_return;
        }
    }
    return result;

 cleanup_and_return:
    if (lags_buffer != NULL) {
        gst_buffer_unref(lags_buffer);
    }
    if (ts_buffer != NULL) {
        gst_buffer_unref(ts_buffer);
    }
    if (dv_buffer != NULL) {
        gst_buffer_unref(dv_buffer);
    }
    return result;
}

static GstFlowReturn
chain_4bit_lagset(WcbeLagsetDemux *demux, GstBuffer *buffer, GstCaps **caps)
{
    wcbe_vect_t *lagset = (wcbe_vect_t *)GST_BUFFER_DATA(buffer);
    wcbe_lagset_header_t *lagset_hdr = &wcbe_vect_index(
        lagset, WCBE_LAGSET_HEADER_BLOCK_INDEX, wcbe_lagset_header_t, 0);
    wcbe_lagset_frame_header_t *fmhdrs =
        &wcbe_vect_index(lagset, WCBE_LAGSET_FRAME_HEADERS_BLOCK_INDEX,
                         wcbe_lagset_frame_header_t, 0);
    lf_lag_t *lags =
        &wcbe_vect_index(lagset, WCBE_LAGSET_LAGS_BLOCK_INDEX, lf_lag_t, 0);
    return chain_one_subprod(buffer, demux->srcpads[0], caps, lagset_hdr,
                             fmhdrs, lags);
}

static GstFlowReturn
chain_7bit_lagset(WcbeLagsetDemux *demux, GstBuffer *buffer, GstCaps **caps)
{
    GstFlowReturn result = GST_FLOW_OK;
    wcbe_vect_t *lagset = (wcbe_vect_t *)GST_BUFFER_DATA(buffer);
    wcbe_lagset_header_t *lagset_hdr = &wcbe_vect_index(
        lagset, WCBE_LAGSET_HEADER_BLOCK_INDEX, wcbe_lagset_header_t, 0);
    wcbe_lagset_frame_header_t *fmhdrs =
        &wcbe_vect_index(lagset, WCBE_LAGSET_FRAME_HEADERS_BLOCK_INDEX,
                         wcbe_lagset_frame_header_t, 0);
    lf_lag_t *lags =
        &wcbe_vect_index(lagset, WCBE_LAGSET_LAGS_BLOCK_INDEX, lf_lag_t, 0);
    guint frames_per_subprod =
        lagset_hdr->num_frames_per_bin * lagset_hdr->num_bins;
    guint lags_per_subprod =
        lagset_hdr->num_lags_per_bin * lagset_hdr->num_bins;
    guint i;
    for (i = 0; result == GST_FLOW_OK && i < 4; ++i) {
        result = chain_one_subprod(buffer, demux->srcpads[i], caps,
                                   lagset_hdr, fmhdrs, lags);
        caps += WCBE_LAGSET_DEMUX_NUM_SRC_PADS;
        fmhdrs += frames_per_subprod;
        lags += lags_per_subprod;
    }
    return result;
}

static void
free_prodcaps(gpointer data)
{
    GstCaps **c = data;
    while (*c != NULL) {
        gst_caps_unref(*c++);
    }
    g_free(data);
}

static GstCaps **
get_prodcaps(GHashTable *prodcaps, wcbe_lagset_header_t *hdr)
{
    GstCaps **result = g_hash_table_lookup(prodcaps, GINT_TO_POINTER(hdr->id));
    if (result == NULL) {
        guint pad_multiplicity = (hdr->nbits == 7 ? 4 : 1);
        result = g_new0(GstCaps *,
                        pad_multiplicity * WCBE_LAGSET_DEMUX_NUM_SRC_PADS + 1);
        guint i;
        for (i = 0; i < pad_multiplicity; ++i) {
            // 'lags' src caps
            GstCaps *caps =
                gst_caps_copy(gst_static_pad_template_get_caps(&lags_factory));
            result[i * WCBE_LAGSET_DEMUX_NUM_SRC_PADS
                   + WCBE_LAGSET_DEMUX_LAGS_PAD_INDEX] = caps;
            gst_caps_set_simple(
                caps,
                WCBE_ARRAY_NUM_ROWS_PROPERTY, G_TYPE_UINT, hdr->num_bins,
                WCBE_ARRAY_NUM_COLS_PROPERTY, G_TYPE_UINT,
                                              2 * hdr->num_lags_per_bin,
                WCBE_ARRAY_ROW_LENGTH_PROPERTY, G_TYPE_UINT,
                                                2 * hdr->num_lags_per_bin,
                WCBE_PARTIAL_PRODUCT_PROPERTY, G_TYPE_UINT, i,
                NULL);
            add_lagset_hdr_props(caps, hdr);

            // 'ts' src caps
            caps =
                gst_caps_copy(gst_static_pad_template_get_caps(&ts_factory));
            result[i * WCBE_LAGSET_DEMUX_NUM_SRC_PADS
                   + WCBE_LAGSET_DEMUX_TIMESTAMP_PAD_INDEX] = caps;
            gst_caps_set_simple(
                caps,
                WCBE_ARRAY_NUM_ROWS_PROPERTY, G_TYPE_UINT, hdr->num_bins,
                WCBE_ARRAY_NUM_COLS_PROPERTY, G_TYPE_UINT,
                                              hdr->num_frames_per_bin,
                WCBE_ARRAY_ROW_LENGTH_PROPERTY, G_TYPE_UINT,
                                                hdr->num_frames_per_bin,
                WCBE_PARTIAL_PRODUCT_PROPERTY, G_TYPE_UINT, i,
                NULL);
            add_lagset_hdr_props(caps, hdr);

            // 'dv' src caps
            caps =
                gst_caps_copy(gst_static_pad_template_get_caps(&dv_factory));
            result[i * WCBE_LAGSET_DEMUX_NUM_SRC_PADS
                   + WCBE_LAGSET_DEMUX_DATAVALID_PAD_INDEX] = caps;
            gst_caps_set_simple(
                caps,
                WCBE_ARRAY_NUM_ROWS_PROPERTY, G_TYPE_UINT, hdr->num_bins,
                WCBE_ARRAY_NUM_COLS_PROPERTY, G_TYPE_UINT,
                                              2 * hdr->num_frames_per_bin,
                WCBE_ARRAY_ROW_LENGTH_PROPERTY, G_TYPE_UINT,
                                                2 * hdr->num_frames_per_bin,
                WCBE_PARTIAL_PRODUCT_PROPERTY, G_TYPE_UINT, i,
                NULL);
            add_lagset_hdr_props(caps, hdr);
        }
        g_hash_table_insert(prodcaps, GINT_TO_POINTER(hdr->id), result);
    }
    else {
        GstCaps **caps = result;
        while (*caps != NULL) {
            *caps = gst_caps_make_writable(*caps);
            ++caps;
        }
    }
    return result;
}

static void
add_lagset_hdr_props(GstCaps *caps, wcbe_lagset_header_t *hdr)
{
    gst_caps_set_simple(
        caps,
        WCBE_PRODUCT_ID_PROPERTY, G_TYPE_UINT, hdr->id,
        WCBE_SUBBAND_INDEX_PROPERTY, G_TYPE_UINT, hdr->sbIdx,
        WCBE_FRAME_SIZE_PROPERTY, G_TYPE_UINT,
        hdr->num_lags_per_bin / hdr->num_frames_per_bin,
        WCBE_NUM_BITS_PROPERTY, G_TYPE_UINT, hdr->nbits,
        WCBE_NUM_BINS_PROPERTY, G_TYPE_UINT, hdr->num_bins,
        NULL);
    GstStructure *s = gst_structure_copy(gst_caps_get_structure(caps, 0));

    GValue i = {0,};
    g_value_init(&i, G_TYPE_UINT);
    GValue array = {0,};
    g_value_init(&array, GST_TYPE_ARRAY);

    // station id array
    g_value_set_uint(&i, hdr->stnA);
    gst_value_array_append_value(&array, &i);
    g_value_set_uint(&i, hdr->stnB);
    gst_value_array_append_value(&array, &i);
    gst_structure_set_value(s, WCBE_STATION_IDS_PROPERTY, &array);

    // baseband id array
    g_value_reset(&array);
    g_value_set_uint(&i, hdr->bbA);
    gst_value_array_append_value(&array, &i);
    g_value_set_uint(&i, hdr->bbB);
    gst_value_array_append_value(&array, &i);
    gst_structure_set_value(s, WCBE_BASEBAND_IDS_PROPERTY, &array);

    g_value_unset(&array);
    g_value_unset(&i);

    gst_caps_remove_structure(caps, 0);
    gst_caps_append_structure(caps, s);
}
