#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <stdio.h>
#include <errno.h>
#include <time.h>

#include <wcbe_lagset_dump.h>
#include <wcbe.h>
#include <wcbe_util.h>
#include <wcbe_lagset.h>

GST_DEBUG_CATEGORY_STATIC(wcbe_lagset_dump_debug);
#define GST_CAT_DEFAULT wcbe_lagset_dump_debug

/* Filter signals and args */
enum
{
    /* FILL ME */
    LAST_SIGNAL
};

enum
{
    ARG_0,
    ARG_DO_OUTPUT,
    ARG_MAX_FILE_SIZE,
    ARG_REGEX,
    ARG_OUTPUT_DIR
};

typedef struct product_key_t {
    GQuark config;
    GQuark product;
} product_key_t;

typedef struct product_file_t {
    gchar *filename;
    FILE *file;
    gsize record_size;
    gulong num_lagsets;
} product_file_t;

#define BLS_VERSION_NUMBER 4
#define DEFAULT_MAX_FILE_SIZE (10 * (1 << 20))
#define MIN_MAX_FILE_SIZE (100 * (1 << 10))
#define MAX_MAX_FILE_SIZE (1 << 30)
#define DEFAULT_PRODUCTS_REGEX ".+"

static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE(
    "sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS(WCBE_LAGSET_MEDIA_TYPE)
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE(
    "src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS(WCBE_LAGSET_MEDIA_TYPE)
    );

GST_BOILERPLATE(WcbeLagsetDump, wcbe_lagset_dump, GstBaseTransform,
                GST_TYPE_BASE_TRANSFORM);

static void wcbe_lagset_dump_finalize(GObject *gobject);
static void wcbe_lagset_dump_set_property(
    GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec);
static void wcbe_lagset_dump_get_property(
    GObject *object, guint prop_id, GValue *value, GParamSpec *pspec);
static gboolean wcbe_lagset_dump_stop(GstBaseTransform *trans);
static GstFlowReturn wcbe_lagset_dump_transform_ip(GstBaseTransform *trans,
                                                   GstBuffer *buffer);
static gboolean wcbe_lagset_dump_src_event(GstBaseTransform *trans,
                                           GstEvent *event);
static guint product_key_hash(gconstpointer key);
static gboolean product_key_equal(gconstpointer a, gconstpointer b);
static void destroy_product_file(gpointer data);
static void output_lagset(WcbeLagsetDump *lagset_dump, wcbe_vect_t *lagset,
                          product_file_t *pfile);
static gboolean wcbe_lagset_dump_create_config(
    WcbeLagsetDump *lagset_dump, const wcbe_config_t *config,
    GstPipeline *pipeline);
static void wcbe_lagset_dump_destroy_config(
    WcbeLagsetDump *lagset_dump, const wcbe_config_t *config,
    GstPipeline *pipeline);
static void regenerate_output_products(WcbeLagsetDump *lagset_dump);
static void close_product_file(product_file_t *pfile);
static void close_all_product_files(WcbeLagsetDump *lagset_dump);
static void set_do_output(WcbeLagsetDump *lagset_dump, gboolean do_output);
static void set_output_dir(WcbeLagsetDump *lagset_dump, const gchar *dir);
static void parent_set_cb(GstObject *gstobject, GstObject *parent,
                          gpointer user_data);

#define LAGSET_SYNC_WORD 0xF83E0F83

static void
wcbe_lagset_dump_base_init(gpointer gclass)
{
    GST_DEBUG_CATEGORY_INIT(wcbe_lagset_dump_debug, "wcbe_lagset_dump_debug",
                            0, "wcbe_lagset_dump debug");
    static GstElementDetails element_details = {
        "WIDAR binary lagset dump",
        "Transform/WIDAR",
        "Dump WIDAR lag sets to files",
        "Martin Pokorny <mpokorny@nrao.edu>"
    };
    GstElementClass *element_class = GST_ELEMENT_CLASS(gclass);

    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&sink_factory));
    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&src_factory));
    gst_element_class_set_details(element_class, &element_details);
}

/* initialize the plugin's class */
static void
wcbe_lagset_dump_class_init(WcbeLagsetDumpClass * klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    GstBaseTransformClass *gstbasetrans_class =
        GST_BASE_TRANSFORM_CLASS(klass);

    gobject_class->set_property =
        GST_DEBUG_FUNCPTR(wcbe_lagset_dump_set_property);
    gobject_class->get_property =
        GST_DEBUG_FUNCPTR(wcbe_lagset_dump_get_property);
    gobject_class->finalize =
        GST_DEBUG_FUNCPTR(wcbe_lagset_dump_finalize);

    g_object_class_install_property(
        gobject_class, ARG_DO_OUTPUT,
        g_param_spec_boolean("do-output", "Dump lag sets",
                             "Produce binary lag set files?",
                             FALSE, G_PARAM_READWRITE));
    g_object_class_install_property(
        gobject_class, ARG_MAX_FILE_SIZE,
        g_param_spec_int64("max-file-size", "Max file size",
                           "Maximum dump file size",
                           MIN_MAX_FILE_SIZE, MAX_MAX_FILE_SIZE,
                           DEFAULT_MAX_FILE_SIZE, G_PARAM_READWRITE));
    g_object_class_install_property(
        gobject_class, ARG_REGEX,
        g_param_spec_string("products", "Product name regex",
                            "Matching product names to dump",
                            DEFAULT_PRODUCTS_REGEX, G_PARAM_READWRITE));
    g_object_class_install_property(
        gobject_class, ARG_OUTPUT_DIR,
        g_param_spec_string("output-dir", "File output directory",
                            "Directory to write BLS files into",
                            NULL, G_PARAM_READWRITE));

    gstbasetrans_class->stop = GST_DEBUG_FUNCPTR(wcbe_lagset_dump_stop);
    gstbasetrans_class->transform_ip =
        GST_DEBUG_FUNCPTR(wcbe_lagset_dump_transform_ip);
    gstbasetrans_class->src_event =
        GST_DEBUG_FUNCPTR(wcbe_lagset_dump_src_event);
    gstbasetrans_class->passthrough_on_same_caps = TRUE;
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
wcbe_lagset_dump_init(WcbeLagsetDump *lagset_dump,
                      WcbeLagsetDumpClass *gclass)
{
    g_signal_connect(GST_OBJECT(lagset_dump), "parent-set",
                     G_CALLBACK(parent_set_cb), NULL);
    lagset_dump->do_output = FALSE;
    lagset_dump->max_file_size = DEFAULT_MAX_FILE_SIZE;
    GError *err = NULL;
    lagset_dump->product_regex =
        g_regex_new(DEFAULT_PRODUCTS_REGEX, G_REGEX_OPTIMIZE, 0, &err);
    g_assert(err == NULL);
    g_static_mutex_init(&(lagset_dump->mutex));
    lagset_dump->output_products = g_hash_table_new_full(
        product_key_hash, product_key_equal, g_free, destroy_product_file);
    lagset_dump->configs = g_ptr_array_new();
    lagset_dump->output_dir = NULL;
}

static void
wcbe_lagset_dump_finalize(GObject *gobject)
{
    WcbeLagsetDump *lagset_dump = WCBE_LAGSET_DUMP(gobject);
    g_static_mutex_free(&(lagset_dump->mutex));
    if (lagset_dump->product_regex != NULL) {
        g_regex_unref(lagset_dump->product_regex);
    }
    g_hash_table_destroy(lagset_dump->output_products);
    g_ptr_array_free(lagset_dump->configs, TRUE);
    if (lagset_dump->output_dir != NULL) {
        g_free(lagset_dump->output_dir);
    }
}

static void
wcbe_lagset_dump_set_property(GObject *object, guint prop_id,
                              const GValue *value, GParamSpec *pspec)
{
    WcbeLagsetDump *lagset_dump = WCBE_LAGSET_DUMP(object);

    switch (prop_id) {
      case ARG_DO_OUTPUT:
        set_do_output(lagset_dump, g_value_get_boolean(value));
        break;
      case ARG_MAX_FILE_SIZE: {
        off64_t new_max = g_value_get_int64(value);
        g_static_mutex_lock(&(lagset_dump->mutex));
        lagset_dump->max_file_size = new_max;
        g_static_mutex_unlock(&(lagset_dump->mutex));
        break;
      }
      case ARG_REGEX: {
        const gchar *str = g_value_get_string(value);
        GError *err = NULL;
        GRegex *new_regex = g_regex_new(str, G_REGEX_OPTIMIZE, 0, &err);
        if (err == NULL) {
            g_static_mutex_lock(&(lagset_dump->mutex));
            if (lagset_dump->product_regex != NULL) {
                g_regex_unref(lagset_dump->product_regex);
            }
            lagset_dump->product_regex = new_regex;
            g_hash_table_remove_all(lagset_dump->output_products);
            regenerate_output_products(lagset_dump);
            g_static_mutex_unlock(&(lagset_dump->mutex));
        }
        else {
            g_message("Failed to compile regular expression '%s' "
                      "for lagset product names: %s",
                      str, err->message);
        }
        break;
      }
      case ARG_OUTPUT_DIR:
        set_output_dir(lagset_dump, g_value_get_string(value));
        break;
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static void
wcbe_lagset_dump_get_property(GObject *object, guint prop_id,
                              GValue *value, GParamSpec *pspec)
{
    WcbeLagsetDump *lagset_dump = WCBE_LAGSET_DUMP(object);

    switch (prop_id) {
      case ARG_DO_OUTPUT:
        g_value_set_boolean(value, lagset_dump->do_output);
        break;
      case ARG_MAX_FILE_SIZE:
        g_value_set_int64(value, lagset_dump->max_file_size);
        break;
      case ARG_REGEX:
        if (lagset_dump->product_regex != NULL) {
            g_value_set_string(
                value, g_regex_get_pattern(lagset_dump->product_regex));
        }
        else {
            g_value_set_string(value, "");
        }
        break;
      case ARG_OUTPUT_DIR:
        if (lagset_dump->output_dir != NULL) {
            g_value_set_string(value, lagset_dump->output_dir);
        }
        else {
            gchar *cd = g_get_current_dir();
            g_value_set_string(value, cd);
            g_free(cd);
        }
        break;
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static gboolean
wcbe_lagset_dump_stop(GstBaseTransform *trans)
{
    WcbeLagsetDump *lagset_dump = WCBE_LAGSET_DUMP(trans);
    close_all_product_files(lagset_dump);
    return TRUE;
}

static GstFlowReturn
wcbe_lagset_dump_transform_ip(GstBaseTransform *trans, GstBuffer *buf)
{
    WcbeLagsetDump *lagset_dump = WCBE_LAGSET_DUMP(trans);

    if (g_atomic_int_get(&(lagset_dump->do_output))
        && g_static_mutex_trylock(&(lagset_dump->mutex))) {
        wcbe_vect_t *lagset = (wcbe_vect_t *)GST_BUFFER_DATA(buf);
        wcbe_lagset_header_t *lagset_hdr = &wcbe_vect_index(
            lagset, WCBE_LAGSET_HEADER_BLOCK_INDEX, wcbe_lagset_header_t, 0);
        product_key_t key =
            {.config = lagset_hdr->config, .product = lagset_hdr->id};
        product_file_t *pfile =
            g_hash_table_lookup(lagset_dump->output_products, &key);
        if (pfile != NULL) {
            output_lagset(lagset_dump, lagset, pfile);
        }
        g_static_mutex_unlock(&(lagset_dump->mutex));
    }
    return GST_FLOW_OK;
}

static gboolean
wcbe_lagset_dump_src_event(GstBaseTransform *trans, GstEvent *event)
{
    WcbeLagsetDump *lagset_dump = WCBE_LAGSET_DUMP(trans);

    switch (GST_EVENT_TYPE(event)) {
      case WCBE_EVENT_DEACTIVATE_CONFIG: {
        const GstStructure *structure = gst_event_get_structure(event);
        gpointer config;
        if (gst_structure_get_pointer(structure, "configuration", &config)) {
            GQuark cfg = g_quark_from_string(((wcbe_config_t *)config)->name);
            void close_config_pfile(gpointer key, gpointer value,
                                    gpointer user_data G_GNUC_UNUSED) {
                product_key_t *pkey = key;
                if (cfg == pkey->config) {
                    product_file_t *pfile = value;
                    close_product_file(pfile);
                }
            }
            g_static_mutex_lock(&(lagset_dump->mutex));
            g_hash_table_foreach(lagset_dump->output_products,
                                 close_config_pfile, NULL);
            g_static_mutex_unlock(&(lagset_dump->mutex));
        }
        else {
            g_debug("Failed to read 'configuration' field from "
                    "WCBE_EVENT_DEACTIVATE_CONFIG event structure");
        }
        break;
      }
      default:
        break;
    }
    return gst_pad_push_event(GST_BASE_TRANSFORM_SINK_PAD(trans), event);
}

static guint
product_key_hash(gconstpointer key)
{
    const product_key_t *pkey = key;
#if SIZEOF_GUINT < SIZEOF_GUINT64
    return pkey->config ^ pkey->product;
#else
    return (((guint64)(pkey->config)) << 32) | pkey->product;
#endif
}

static gboolean
product_key_equal(gconstpointer a, gconstpointer b)
{
    const product_key_t *key_a = a;
    const product_key_t *key_b = b;
    return ((key_a->config == key_b->config)
            && (key_a->product == key_b->product));
}

static void
destroy_product_file(gpointer data)
{
    product_file_t *pfile = data;
    close_product_file(pfile);
    g_free(pfile);
}

static void
output_lagset(WcbeLagsetDump *lagset_dump, wcbe_vect_t *lagset,
              product_file_t *pfile)
{
    g_return_if_fail(lagset != NULL);
    g_return_if_fail(pfile != NULL);

    const wcbe_lagset_header_t *lagset_hdr = &wcbe_vect_index(
        lagset, WCBE_LAGSET_HEADER_BLOCK_INDEX, wcbe_lagset_header_t, 0);
    const wcbe_lagset_frame_header_t *frame_header =
        &wcbe_vect_index(lagset, WCBE_LAGSET_FRAME_HEADERS_BLOCK_INDEX,
                         wcbe_lagset_frame_header_t, 0);

    if (pfile->file == NULL) {
        // format time parts of BLS file name
        struct tm obs_tm;
        time_t obstime = lagset_hdr->timestamp.seconds;
        gmtime_r(&obstime, &obs_tm);
        struct tm now_tm;
        time_t now = time(NULL);
        localtime_r(&now, &now_tm);
        static const gchar tm_tmpl[] = "YYYY-DDD-HH-MM-SS";
        static const gchar tm_format[] = "%Y-%j-%H-%M-%S";
        gchar obsstr[sizeof(tm_tmpl)];
        gchar nowstr[sizeof(tm_tmpl)];
        strftime(obsstr, sizeof(obsstr), tm_format, &obs_tm);
        strftime(nowstr, sizeof(nowstr), tm_format, &now_tm);
        // create BLS file name
        gchar *name =
            g_strdup_printf("%s-A%03u%02u%01uB%03u%02u%01u-%s.bls",
                            obsstr,
                            lagset_hdr->stnA, HEADER_SBID_A(frame_header),
                            lagset_hdr->bbA,
                            lagset_hdr->stnB, HEADER_SBID_B(frame_header),
                            lagset_hdr->bbB,
                            nowstr);
        // open the new file
        if (lagset_dump->output_dir != NULL) {
            pfile->filename =
                g_build_filename(lagset_dump->output_dir, name, NULL);
            g_free(name);
        }
        else {
            pfile->filename = name;
        }
        pfile->num_lagsets = 0;
        pfile->file = fopen(pfile->filename, "w");
        if (pfile->file == NULL) {
            g_warning("Failed to open new binary lag set file '%s': %s",
                      pfile->filename, g_strerror(errno));
            goto error_return;
        }
        // write file header
        static const struct bls_header {
            gchar id[3];
            guint8 version;
            guint32 num_lagsets;
        } bls_header = {{'B', 'L', 'S'}, BLS_VERSION_NUMBER, 0};
        if (fwrite(&bls_header, sizeof(bls_header), 1, pfile->file) != 1) {
            goto error_return;
        }
    }
#ifndef G_DISABLE_ASSERT
    off_t start = ftello(pfile->file);
#endif

    // write lag set
    gboolean write_word(guint32 w) {
        return (fwrite(&w, sizeof(guint32), 1, pfile->file) == 1);
    }
    guint nbits_multiplier = (lagset_hdr->nbits == WCBE_7BITS ? 4 : 1);
    guint num_frames = lagset_hdr->num_frames_per_bin * lagset_hdr->num_bins
        * nbits_multiplier;

    // Write lag set header. Need a non-empty frame header for
    // station, baseband and subband ids (all frames may not have
    // arrived).
    guint i = 0;
    while (i < num_frames && frame_header->key[1] == 0) {
        ++frame_header;
    }
    if (i == num_frames) {
        frame_header =
            &wcbe_vect_index(lagset, WCBE_LAGSET_FRAME_HEADERS_BLOCK_INDEX,
                             wcbe_lagset_frame_header_t, 0);
    }
    if (!write_word(LAGSET_SYNC_WORD)
        || !write_word(frame_header->key[1])
        || !write_word((lagset_hdr->nbits << 29)
                       | (0x1FFFFFFF & num_frames))) {
        goto error_return;
    }

    // Write frame headers
    frame_header =
        &wcbe_vect_index(lagset, WCBE_LAGSET_FRAME_HEADERS_BLOCK_INDEX,
                         wcbe_lagset_frame_header_t, 0);
    for (i = 0; i < num_frames; ++i) {
        if (!write_word((frame_header->key[0] & 0xFFFFFFF8)
                        | (frame_header->reverse_conjugate ? 1 : 0))
            || !write_word(frame_header->key[2])
            || !write_word(frame_header->timestamp.seconds)
            || !write_word(frame_header->timestamp.clock_cycles
                           | (frame_header->timestamp.epoch << 29))
            || !write_word(frame_header->data_valid_center)
            || !write_word(frame_header->data_valid_edge)
            || !write_word(frame_header->key[3])
            || !write_word(frame_header->checksum)) {
            goto error_return;
        }
        ++frame_header;
    }
    // Write num_bins and num_lags
    write_word(lagset_hdr->num_bins);
    write_word(lagset_hdr->num_lags_per_bin);

    // Write lags
    gsize num_clags32 = lagset_hdr->num_lags_per_bin * lagset_hdr->num_bins
        * nbits_multiplier;
    lf_lag_t *lags =
        &wcbe_vect_index(lagset, WCBE_LAGSET_LAGS_BLOCK_INDEX, lf_lag_t, 0);
    fwrite(lags, sizeof(lf_lag_t), num_clags32, pfile->file);

    // Close file if it's reached the max allowed size.
    pfile->num_lagsets++;
    g_assert(ftello(pfile->file) - start == pfile->record_size);
    if (ftello(pfile->file) > lagset_dump->max_file_size
        || pfile->num_lagsets == G_MAXUINT32) {
        close_product_file(pfile);
    }
    return;

 error_return:
    g_warning("Failed to write part of a binary lag set record to file '%s': "
              "%s", pfile->filename, g_strerror(errno));
    close_product_file(pfile);
}

static gboolean
wcbe_lagset_dump_create_config(WcbeLagsetDump *lagset_dump,
                               const wcbe_config_t *config,
                               GstPipeline *pipeline G_GNUC_UNUSED)
{
    g_static_mutex_lock(&(lagset_dump->mutex));
    g_ptr_array_add(lagset_dump->configs, (wcbe_config_t *)config);
    regenerate_output_products(lagset_dump);
    g_static_mutex_unlock(&(lagset_dump->mutex));
    return TRUE;
}

static void
wcbe_lagset_dump_destroy_config(WcbeLagsetDump *lagset_dump,
                                const wcbe_config_t *config,
                                GstPipeline *pipeline G_GNUC_UNUSED)
{
    g_static_mutex_lock(&(lagset_dump->mutex));
    g_ptr_array_remove_fast(lagset_dump->configs, (wcbe_config_t *)config);
    regenerate_output_products(lagset_dump);
    g_static_mutex_unlock(&(lagset_dump->mutex));
}

static void
regenerate_output_products(WcbeLagsetDump *lagset_dump)
{
    // Remove products for configs no longer in lagset_dump->configs
    // array.
    gboolean rm_config_product(gpointer key, gpointer value G_GNUC_UNUSED,
                               gpointer user_data G_GNUC_UNUSED) {
        product_key_t *pkey = key;
        gboolean match = FALSE;
        guint i;
        for (i = 0; !match && i < lagset_dump->configs->len; ++i) {
            wcbe_config_t *config = g_ptr_array_index(lagset_dump->configs, i);
            match = (g_quark_from_string(config->name) == pkey->config);
        }
        return !match;
    }
    g_hash_table_foreach_remove(lagset_dump->output_products,
                                rm_config_product, NULL);

    // Add new products (only way to do that is go through all
    // configs, and test whether their products are in the
    // output_products hash table).
    if (lagset_dump->product_regex != NULL) {
        GString *product_name = g_string_sized_new(40);
        void scan_products(gpointer data, gpointer user_data G_GNUC_UNUSED) {
            wcbe_config_t *config = data;
            g_string_assign(product_name, config->name);
            g_string_append_c(product_name, '/');
            gsize config_len = product_name->len;
            void add_product(gpointer data, gpointer user_data G_GNUC_UNUSED) {
                wcbe_product_config_t *product = data;
                g_string_truncate(product_name, config_len);
                g_string_append(product_name, product->id);
                if (g_regex_match(lagset_dump->product_regex,
                                  product_name->str, 0, NULL)) {
                    product_key_t *key = g_new(product_key_t, 1);
                    key->config = g_quark_from_string(config->name);
                    key->product = g_quark_from_string(product->id);
                    if (g_hash_table_lookup(lagset_dump->output_products, key)
                        == NULL) {
                        product_file_t *val = g_new0(product_file_t, 1);
                        // The following calculation of num_frames works for
                        // both 4 and 7 bit correlation products on the
                        // assumption that the number of frames in each of the
                        // four components in the 7 bit case don't vary.
                        guint num_frames = 0;
                        wcbe_product4_config_t *prd4 =
                            (wcbe_product4_config_t *)product;
                        guint i;
                        for (i = 0; i < prd4->lag_chains->len; ++i) {
                            wcbe_lag_chain_config_t *chain =
                                &g_array_index(prd4->lag_chains,
                                               wcbe_lag_chain_config_t, i);
                            num_frames += chain->cccN - chain->ccc0 + 1;
                        }
                        val->record_size = sizeof(guint32)
                            * (5 + (product->nbits == 7 ? 8 : 2)
                               * (4 * num_frames
                                  + product->num_bins * product->num_lags));
                        g_hash_table_insert(
                            lagset_dump->output_products, key, val);
                    }
                    else {
                        g_free(key);
                    }
                }
            }
            if (config->input.products != NULL) {
                g_ptr_array_foreach(config->input.products, add_product, NULL);
            }
        }
        g_ptr_array_foreach(lagset_dump->configs, scan_products, NULL);
        g_string_free(product_name, TRUE);
    }
}

static void
close_product_file(product_file_t *pfile)
{
    if (pfile->file != NULL) {
        gint rc = fseek(pfile->file, 4, SEEK_SET);
        if (rc == 0) {
            if (fwrite(&(pfile->num_lagsets), sizeof(guint32), 1, pfile->file)
                != 1) {
                g_warning("Failed to write number of lag sets to BLS file "
                          "header: %s", g_strerror(errno));
            }
        }
        else {
            g_warning("Failed to move BLS file position to write number of "
                      "lag sets in file: %s", g_strerror(errno));
        }
        rc = fclose(pfile->file);
        if (rc != 0) {
            g_warning("Failed to close binary lag set file '%s': %s",
                      pfile->filename, g_strerror(errno));
        }
        pfile->file = NULL;
        g_free(pfile->filename);
        pfile->filename = NULL;
    }
}

static void
close_all_product_files(WcbeLagsetDump *lagset_dump)
{
    void close_file(gpointer key G_GNUC_UNUSED, gpointer value,
                    gpointer user_data G_GNUC_UNUSED) {
        product_file_t *pfile = value;
        close_product_file(pfile);
    }
    g_static_mutex_lock(&(lagset_dump->mutex));
    g_hash_table_foreach(lagset_dump->output_products, close_file, NULL);
    g_static_mutex_unlock(&(lagset_dump->mutex));
}

static void
set_do_output(WcbeLagsetDump *lagset_dump, gboolean do_output)
{
    gboolean old_do_output = lagset_dump->do_output;
    g_atomic_int_set(&(lagset_dump->do_output), do_output);
    if (old_do_output && !do_output) {
        close_all_product_files(lagset_dump);
    }
}

static void
set_output_dir(WcbeLagsetDump *lagset_dump, const gchar *dir)
{
    g_return_if_fail(dir != NULL);

    if ((lagset_dump->output_dir == NULL && strlen(dir) != 0)
        || strcmp(dir, lagset_dump->output_dir) != 0) {
        gboolean do_output = lagset_dump->do_output;
        set_do_output(lagset_dump, FALSE);
        if (lagset_dump->output_dir != NULL) {
            g_free(lagset_dump->output_dir);
        }
        if (strlen(dir) != 0) {
            lagset_dump->output_dir = g_strdup(dir);
        }
        else {
            lagset_dump->output_dir = NULL;
        }
        set_do_output(lagset_dump, do_output);
    }
}

static void
parent_set_cb(GstObject *gstobject, GstObject *parent,
              gpointer user_data G_GNUC_UNUSED)
{
    g_debug("lagset_dump connecting signals");
    if (object_has_signal(G_OBJECT(parent), "create-configuration")) {
        g_signal_connect_swapped(parent, "create-configuration",
                                 G_CALLBACK(wcbe_lagset_dump_create_config),
                                 gstobject);
    }
    if (object_has_signal(G_OBJECT(parent), "destroy-configuration")) {
        g_signal_connect_swapped(parent, "destroy-configuration",
                                 G_CALLBACK(wcbe_lagset_dump_destroy_config),
                                 gstobject);
    }
}
