#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/epoll.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>

#include <wcbe_lagset_sink.h>
#include <wcbe.h>
#include <user_marshal.h>
#include <wcbe_util.h>
#include <wcbe_config.h>
#include <wcbe_lagset.h>
#include <tree.h>

typedef struct pipelinesock {
    gchar *name;
    struct sockaddr_un addr;
} pipelinesock_t;

typedef struct wcbe_lagset_sink_config {
    const wcbe_config_t *wcfg;
    gchar *sockdir;
    GArray *socks;
    GHashTable *destinations;
    gboolean flushing;
    gboolean is_active;
    WcbeLagsetSink *sink;
} wcbe_lagset_sink_config_t;

GST_DEBUG_CATEGORY_STATIC(wcbe_lagset_sink_debug);
#define GST_CAT_DEFAULT wcbe_lagset_sink_debug

/* Filter signals and args */
enum
{
    SIGNAL_CREATE_CFG,
    SIGNAL_ACTIVATE_CFG,
    SIGNAL_DEACTIVATE_CFG,
    SIGNAL_DESTROY_CFG,
    SIGNAL_EOS_PIPELINE,
    LAST_SIGNAL
};

enum
{
    ARG_0,
    ARG_CONFIGS,
    ARG_SOCKDIR,
    ARG_ORPHANED_LAGSETS
};

static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE(
    "sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS(WCBE_LAGSET_MEDIA_TYPE));

static guint wcbe_lagset_sink_signals[LAST_SIGNAL] = {0};

GST_BOILERPLATE(WcbeLagsetSink, wcbe_lagset_sink, GstBaseSink,
                GST_TYPE_BASE_SINK);

static void wcbe_lagset_sink_set_property(
    GObject * object, guint prop_id, const GValue * value, GParamSpec * pspec);
static void wcbe_lagset_sink_get_property(
    GObject * object, guint prop_id, GValue * value, GParamSpec * pspec);
static void wcbe_lagset_sink_finalize(GObject *object);
static gboolean wcbe_lagset_sink_event(GstBaseSink *sink, GstEvent *event);
static GstFlowReturn wcbe_lagset_sink_render(
    GstBaseSink *sink, GstBuffer *buf);
static gboolean wcbe_lagset_sink_start(GstBaseSink *sink);
static gboolean wcbe_lagset_sink_stop(GstBaseSink *sink);
static GstFlowReturn wcbe_lagset_sink_buffer_alloc(
    GstBaseSink *sink, guint64 offset, guint size, GstCaps *caps,
    GstBuffer **buf);
static void handle_new_subscan(WcbeLagsetSink *sink, const gchar *config,
                               const gchar *uid, guint epoch, guint seconds,
                               guint cycles, gpointer data);
static void handle_cancel_subscan(WcbeLagsetSink *sink, const gchar *config,
                                  const gchar *uid, gpointer data);
static void handle_activate_config(WcbeLagsetSink *sink,
                                   const wcbe_config_t *config,
                                   guint64 time0);
static void handle_deactivate_config(WcbeLagsetSink *sink,
                                     const wcbe_config_t *config);
static void deactivate_config(wcbe_lagset_sink_config_t *cfg);

static void send_stream_object(gint sock, struct msghdr *msghdr);
static void forward_event(wcbe_lagset_sink_config_t *cfg,
                          GstEventType event_type, GstStructure *s);
static wcbe_lagset_sink_config_t *create_config(
    const gchar *sockdir, const wcbe_config_t *config);
static void pipeline_config(wcbe_lagset_sink_config_t *cfg,
                            const wcbe_config_t *config,
                            wcbe_pipeline_config_t *pipeline, branch_t *pids);
static tree_t *map_products(const wcbe_config_t *config, branch_t **branch);
static void fill_product_tree(const wcbe_config_t *config, branch_t *tree);
static wcbe_subarray_config_t *find_product_subarray(
    const wcbe_config_t *config, wcbe_product_config_t *product,
    guint coords[]);
static wcbe_baseband_config_t *find_product_baseband(
    wcbe_subarray_config_t *subarray, wcbe_product_config_t *product,
    guint coords[]);
static gboolean find_product_spectral_window(
    const wcbe_config_t *config, wcbe_baseband_config_t *baseband,
    wcbe_product_config_t *product, guint coords[]);
static void destroy_config(wcbe_lagset_sink_config_t *cfg);
static void destroy_config_silently(wcbe_lagset_sink_config_t *cfg);
static gint open_sock(WcbeLagsetSink *sink);
static void close_sock(WcbeLagsetSink *sink);
static gboolean wcbe_lagset_sink_create_config(
    WcbeLagsetSink *sink, const wcbe_config_t *config, GstPipeline *pipeline);
static void wcbe_lagset_sink_destroy_config(
    WcbeLagsetSink *sink, const wcbe_config_t *config, GstPipeline *pipeline);
static void parent_set_cb(GstObject *gstobject, GstObject *parent,
                          gpointer user_data);

#define SUBARRAY_IDX 0
#define BASEBAND_IDX 1
#define SPECTRAL_WINDOW_IDX 2
#define SUBBAND_IDX 3

static inline wcbe_lagset_sink_config_t *
get_cfg_data(WcbeLagsetSink *sink, const wcbe_config_t *config)
{
    WcbeLagsetSinkClass *klass =
        WCBE_LAGSET_SINK_CLASS(GST_ELEMENT_GET_CLASS(sink));
    return g_dataset_id_get_data(config, klass->cfg_data);
}

static void
wcbe_lagset_sink_base_init(gpointer gclass)
{
    GST_DEBUG_CATEGORY_INIT(wcbe_lagset_sink_debug,
                            "wcbe_lagset_sink", 0,
                            "wcbe_lagset_sink debug");
    static GstElementDetails element_details = {
        "WIDAR lagset sink",
        "Sink/WIDAR",
        "Write WIDAR lag sets to Unix domain sockets",
        "Martin Pokorny <mpokorny@nrao.edu>"
    };
    GstElementClass *element_class = GST_ELEMENT_CLASS(gclass);

    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&sink_factory));
    gst_element_class_set_details(element_class, &element_details);
}

/* initialize the plugin's class */
static void
wcbe_lagset_sink_class_init(WcbeLagsetSinkClass *klass)
{
    GObjectClass *gobject_class  = (GObjectClass *)klass;
    GstBaseSinkClass *gstbasesink_class = (GstBaseSinkClass *)klass;

    gobject_class->set_property =
        GST_DEBUG_FUNCPTR(wcbe_lagset_sink_set_property);
    gobject_class->get_property =
        GST_DEBUG_FUNCPTR(wcbe_lagset_sink_get_property);
    gobject_class->finalize =
        GST_DEBUG_FUNCPTR(wcbe_lagset_sink_finalize);

    gstbasesink_class->render =
        GST_DEBUG_FUNCPTR(wcbe_lagset_sink_render);
    gstbasesink_class->start =
        GST_DEBUG_FUNCPTR(wcbe_lagset_sink_start);
    gstbasesink_class->stop =
        GST_DEBUG_FUNCPTR(wcbe_lagset_sink_stop);
    gstbasesink_class->event =
        GST_DEBUG_FUNCPTR(wcbe_lagset_sink_event);
    gstbasesink_class->buffer_alloc =
        GST_DEBUG_FUNCPTR(wcbe_lagset_sink_buffer_alloc);

    g_object_class_install_property(
        gobject_class, ARG_SOCKDIR,
        g_param_spec_string("sockdir", "socket dir",
                            "Directory for lagset sockets",
                            NULL, G_PARAM_READWRITE));

    g_object_class_install_property(
        gobject_class, ARG_CONFIGS,
        g_param_spec_string("configs", "available configurations",
                            "List of available configurations",
                            NULL, G_PARAM_READABLE));

    g_object_class_install_property(
        gobject_class, ARG_ORPHANED_LAGSETS,
        g_param_spec_uint64("orphaned-lagsets",
                            "Number of orphaned lag sets",
                            "Number of lag sets received without "
                            "associated pipeline",
                            0, G_MAXUINT64,0, G_PARAM_READABLE));

    wcbe_lagset_sink_signals[SIGNAL_CREATE_CFG] =
        g_signal_new("config-created", G_TYPE_FROM_CLASS(klass),
                     G_SIGNAL_RUN_FIRST,
                     0, NULL, NULL,
                     g_cclosure_user_marshal_VOID__STRING_BOOLEAN,
                     G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_BOOLEAN);

    wcbe_lagset_sink_signals[SIGNAL_ACTIVATE_CFG] =
        g_signal_new("config-activated", G_TYPE_FROM_CLASS(klass),
                     G_SIGNAL_RUN_FIRST,
                     0, NULL, NULL,
                     g_cclosure_user_marshal_VOID__STRING_BOOLEAN,
                     G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_BOOLEAN);

    wcbe_lagset_sink_signals[SIGNAL_DEACTIVATE_CFG] =
        g_signal_new("config-deactivated", G_TYPE_FROM_CLASS(klass),
                     G_SIGNAL_RUN_FIRST,
                     0, NULL, NULL,
                     g_cclosure_marshal_VOID__STRING,
                     G_TYPE_NONE, 1, G_TYPE_STRING);

    wcbe_lagset_sink_signals[SIGNAL_DESTROY_CFG] =
        g_signal_new("config-destroyed", G_TYPE_FROM_CLASS(klass),
                     G_SIGNAL_RUN_FIRST,
                     0, NULL, NULL,
                     g_cclosure_user_marshal_VOID__STRING_BOOLEAN,
                     G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_BOOLEAN);

    wcbe_lagset_sink_signals[SIGNAL_EOS_PIPELINE] =
        g_signal_new("pipeline-eos", G_TYPE_FROM_CLASS(klass),
                     G_SIGNAL_RUN_FIRST,
                     0, NULL, NULL,
                     g_cclosure_marshal_VOID__STRING,
                     G_TYPE_NONE, 1, G_TYPE_STRING);

    klass->cfg_data = g_quark_from_string("wcbe_lagset_sink_cfg_data");
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
wcbe_lagset_sink_init(WcbeLagsetSink *lsink, WcbeLagsetSinkClass *gclass)
{
    GstBaseSink *basesink = GST_BASE_SINK(lsink);
    g_signal_connect(GST_OBJECT(lsink), "parent-set",
                     G_CALLBACK(parent_set_cb), NULL);
    gst_base_sink_set_sync(basesink, FALSE);
    gst_base_sink_set_async_enabled(basesink, FALSE);
    lsink->sockdir = NULL;
    lsink->sockfd = -1;
    lsink->configs = g_hash_table_new(g_direct_hash, g_direct_equal);
    g_static_mutex_init(&(lsink->configs_mutex));
    lsink->num_orphaned_lagsets = 0;
}

static void
wcbe_lagset_sink_finalize(GObject *object)
{
    WcbeLagsetSink *lsink = WCBE_LAGSET_SINK(object);

    void deactivate(gpointer key, gpointer value, gpointer user_data) {
        wcbe_lagset_sink_config_t *cfg = value;
        destroy_config(cfg);
    }
    g_static_mutex_lock(&(lsink->configs_mutex));
    g_hash_table_foreach(lsink->configs, deactivate, NULL);
    g_hash_table_destroy(lsink->configs);
    g_static_mutex_unlock(&(lsink->configs_mutex));
    g_static_mutex_free(&(lsink->configs_mutex));

    close_sock(lsink);
    if (lsink->sockdir != NULL) {
        g_free(lsink->sockdir);
    }

    G_OBJECT_CLASS(parent_class)->finalize(object);
}

static gboolean
wcbe_lagset_sink_start(GstBaseSink *sink)
{
    WcbeLagsetSink *lsink = WCBE_LAGSET_SINK(sink);
    lsink->num_orphaned_lagsets = 0;
    return TRUE;
}

static gboolean
wcbe_lagset_sink_stop(GstBaseSink *sink)
{
    return TRUE;
}

static GstFlowReturn
wcbe_lagset_sink_buffer_alloc(GstBaseSink *sink, guint64 offset, guint size,
                              GstCaps *caps, GstBuffer **buf)
{
    return wcbe_vect_pad_buffer_alloc(GST_BASE_SINK_PAD(sink), offset, size,
                                      caps, buf);
}

static void
wcbe_lagset_sink_set_property(GObject *object, guint prop_id,
                              const GValue *value, GParamSpec *pspec)
{
    WcbeLagsetSink *sink = WCBE_LAGSET_SINK(object);

    switch (prop_id) {
      case ARG_SOCKDIR: {
        const gchar *sockdir = g_value_get_string(value);
        if (sink->sockdir == NULL || strcmp(sink->sockdir, sockdir) != 0) {
            if (sink->sockdir != NULL) {
                // Can't transfer configurations (sockets) to new
                // directory, so just destroy them.
                void dconf(gpointer key, gpointer value,
                           gpointer user_data) {
                    wcbe_lagset_sink_config_t *cfg = value;
                    destroy_config(cfg);
                }
                g_static_mutex_lock(&(sink->configs_mutex));
                g_hash_table_foreach(sink->configs, dconf, NULL);
                g_hash_table_remove_all(sink->configs);
                g_static_mutex_unlock(&(sink->configs_mutex));
                close_sock(sink);
                g_free(sink->sockdir);
            }
            sink->sockdir = g_strdup(sockdir);
            gint rc = open_sock(sink);
            if (rc != 0) {
                g_info("Failed to open socket in directory '%s'", sockdir);
                g_free(sink->sockdir);
                sink->sockdir = NULL;
            }
        }
        break;
      }
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static void
wcbe_lagset_sink_get_property(GObject *object, guint prop_id,
                              GValue *value, GParamSpec *pspec)
{
    WcbeLagsetSink *sink = WCBE_LAGSET_SINK(object);

    switch (prop_id) {
      case ARG_SOCKDIR:
        g_value_set_string(value, sink->sockdir ? sink->sockdir : "");
        break;
      case ARG_CONFIGS: {
        GString *configs = g_string_new("");
        void get_name(gpointer key, gpointer value, gpointer user_data) {
            wcbe_lagset_sink_config_t *cfg = value;
            const gchar *fmt = ((configs->len == 0) ? "%s" : "\t%s");
            g_string_append_printf(configs, fmt, cfg->wcfg->name);
        }
        g_static_mutex_lock(&(sink->configs_mutex));
        g_hash_table_foreach(sink->configs, get_name, NULL);
        g_static_mutex_unlock(&(sink->configs_mutex));
        g_value_set_string(value, configs->str);
        g_string_free(configs, TRUE);
        break;
      }
      case ARG_ORPHANED_LAGSETS:
#if SIZEOF_UINT64_T == SIZEOF_UNSIGNED
        g_value_set_uint64(value,
                           (guint64)g_atomic_int_get(
                               &(sink->num_orphaned_lagsets)));
#else
        // The following is exposed to a race condition w.r.t. the
        // pipeline thread updating the num_orphaned_lagsets value ---
        // we'll live with it, a mutex lock seems like overkill.
        g_value_set_uint64(value, sink->num_orphaned_lagsets);
#endif
        break;
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

/* GstBaseSink vmethod implementations */

/* render function
 * this function does the actual processing
 */

static GstFlowReturn
wcbe_lagset_sink_render(GstBaseSink *sink, GstBuffer *buf)
{
    WcbeLagsetSink *lsink = WCBE_LAGSET_SINK(sink);
    wcbe_stream_lagset_t stream_lagset;
    stream_lagset.object.type = WCBE_STREAM_LAGSET;
    wcbe_vect_t *lagset = (wcbe_vect_t *)GST_BUFFER_DATA(buf);
    wcbe_lagset_header_t *lagset_hdr = &wcbe_vect_index(
        lagset, WCBE_LAGSET_HEADER_BLOCK_INDEX, wcbe_lagset_header_t, 0);
    g_static_mutex_lock(&(lsink->configs_mutex));
    wcbe_lagset_sink_config_t *cfg = g_hash_table_lookup(
        lsink->configs, GINT_TO_POINTER(lagset_hdr->config));
    g_static_mutex_unlock(&(lsink->configs_mutex));
    if (G_LIKELY(cfg != NULL)) {
        pipelinesock_t *pipelinesock =
            g_hash_table_lookup(cfg->destinations,
                                GINT_TO_POINTER(lagset_hdr->id));
        if (G_LIKELY(pipelinesock != NULL)) {
            // Use iovec for message---should involve less data copying.
            struct iovec iovec[2];
            wcbe_stream_lagset_init_iovec(
                &stream_lagset, lagset, GST_BUFFER_SIZE(buf), iovec);
            g_strlcpy(stream_lagset.product_id,
                      g_quark_to_string(lagset_hdr->id),
                      sizeof(stream_lagset.product_id));
            struct msghdr msghdr = {
                .msg_name = &(pipelinesock->addr),
                .msg_namelen = sizeof(pipelinesock->addr),
                .msg_control = NULL, .msg_controllen = 0,
                .msg_iov = iovec,
                .msg_iovlen = G_N_ELEMENTS(iovec)};
            send_stream_object(lsink->sockfd, &msghdr);
        }
        else {
            lsink->num_orphaned_lagsets++;
        }
    }
    else {
        g_info("Lagset with unexpected 'config' received...and dropped");
        lsink->num_orphaned_lagsets++;
    }
    return GST_FLOW_OK;
}

static gboolean
wcbe_lagset_sink_event(GstBaseSink *sink, GstEvent *event)
{
    gboolean result;
    WcbeLagsetSink *lsink = WCBE_LAGSET_SINK(sink);
    switch (GST_EVENT_TYPE(event)) {
      case GST_EVENT_EOS:
      case GST_EVENT_FLUSH_STOP:
      case GST_EVENT_FLUSH_START: {
        void fwd(gpointer key, gpointer value, gpointer user_data) {
            wcbe_lagset_sink_config_t *cfg = value;
            forward_event(cfg, GST_EVENT_TYPE(event), NULL);
        }
        g_static_mutex_lock(&(lsink->configs_mutex));
        g_hash_table_foreach(lsink->configs, fwd, NULL);
        g_static_mutex_unlock(&(lsink->configs_mutex));
        result = TRUE;
        break;
      }
      case WCBE_EVENT_ACTIVATE_CONFIG: {
        g_assert(gst_structure_has_name(event->structure, "activate-config"));
        g_assert(gst_structure_has_field(event->structure, "configuration"));
        g_assert(gst_structure_has_field(event->structure, "time0"));
        gpointer configuration = NULL;
        guint64 ts;
        gst_structure_get_pointer(event->structure, "configuration",
                                  &configuration);
        gst_structure_get_uint64(event->structure, "time0", &ts);
        handle_activate_config(lsink, configuration, ts);
        result = TRUE;
        break;
      }
      case WCBE_EVENT_DEACTIVATE_CONFIG: {
        g_assert(gst_structure_has_name(event->structure,
                                        "deactivate-config"));
        g_assert(gst_structure_has_field(event->structure, "configuration"));
        gpointer configuration = NULL;
        gst_structure_get_pointer(event->structure, "configuration",
                                  &configuration);
        handle_deactivate_config(lsink, configuration);
        result = TRUE;
        break;
      }
      default:
        result = TRUE;
    }
    return result;
}

static void
send_stream_object(gint sock, struct msghdr *msghdr)
{
    wcbe_stream_object_t *stream_object = msghdr->msg_iov[0].iov_base;
    gsize msg_size(void) {
        gsize result = 0;
        guint i;
        for (i = 0; i < msghdr->msg_iovlen; ++i) {
            result += msghdr->msg_iov[i].iov_len;
        }
        return result;
    }
    gsize expected_size;
    switch (stream_object->type) {
      case WCBE_STREAM_LAGSET: {
        wcbe_vect_t *lagset = msghdr->msg_iov[msghdr->msg_iovlen - 1].iov_base;
        expected_size = offsetof(wcbe_stream_lagset_t, lagset)
            + lagset->total_size;
        break;
      }
      case WCBE_STREAM_EVENT: {
        wcbe_stream_event_t *event = (wcbe_stream_event_t *)stream_object;
        expected_size = sizeof(wcbe_stream_event_t) + event->structure_len;
        break;
      }
      case WCBE_STREAM_MSG: {
        wcbe_stream_msg_t *msg = (wcbe_stream_msg_t *)stream_object;
        expected_size = sizeof(wcbe_stream_msg_t) + msg->structure_len;
        break;
      }
      default:
        g_assert_not_reached();
        break;
    }
    g_assert(expected_size == msg_size());
    errno = 0;
    gssize bytes_written = sendmsg(sock, msghdr, 0);
    if (G_UNLIKELY(bytes_written != expected_size)) {
        const gchar *errstr;
        if (errno != 0) {
            errstr = g_strerror(errno);
        }
        else {
            errstr = "'errno' not set";
        }
        g_warning("Failed to write stream object to socket '%s': %s",
		  ((struct sockaddr_un *)(msghdr->msg_name))->sun_path,
		  errstr);
    }
}

static void
forward_event(wcbe_lagset_sink_config_t *cfg, GstEventType event_type,
              GstStructure *s)
{
    if (cfg->sink == NULL) {
        return;
    }

    gchar *s_str = ((s != NULL) ? gst_structure_to_string(s) : g_strdup(""));
    wcbe_stream_event_t stream_event;
    stream_event.object.type = WCBE_STREAM_EVENT;
    stream_event.event = event_type;
    stream_event.structure_len = strlen(s_str) + 1;
    struct iovec iovec[] =
        {{.iov_base = &stream_event, .iov_len = sizeof(stream_event)},
         {.iov_base = s_str, .iov_len = stream_event.structure_len}};
    struct msghdr msghdr = {.msg_name = NULL, .msg_namelen = 0,
                            .msg_control = NULL, .msg_controllen = 0,
                            .msg_iov = iovec, .msg_iovlen = 2};
    guint i;
    for (i = 0; i < cfg->socks->len; ++i) {
        pipelinesock_t *pipelinesock =
            &g_array_index(cfg->socks, pipelinesock_t, i);
        msghdr.msg_name = &(pipelinesock->addr);
        msghdr.msg_namelen = sizeof(pipelinesock->addr);
        if (event_type == GST_EVENT_EOS) {
            g_signal_emit(cfg->sink,
                          wcbe_lagset_sink_signals[SIGNAL_EOS_PIPELINE],
                          0, pipelinesock->name);
        }
        send_stream_object(cfg->sink->sockfd, &msghdr);
    }
    g_free(s_str);
}

static wcbe_lagset_sink_config_t *
create_config(const gchar *sockdir, const wcbe_config_t *config)
{
    wcbe_lagset_sink_config_t *result =
        g_slice_new0(wcbe_lagset_sink_config_t);
    result->wcfg = config;
    result->flushing = FALSE;
    result->is_active = FALSE;
    result->socks = g_array_new(FALSE, FALSE, sizeof(pipelinesock_t));
    result->destinations = g_hash_table_new(g_direct_hash, g_direct_equal);
    result->sockdir = g_build_filename(sockdir, config->name, NULL);
    branch_t *pids0 = NULL;
    tree_t *pids = map_products(config, &pids0);
    if (pids != NULL) {
        if (config->pipelines.pipelines != NULL) {
            guint i;
            for (i = 0; i < config->pipelines.pipelines->len; ++i) {
                wcbe_pipeline_config_t *pipeline = &g_array_index(
                    config->pipelines.pipelines, wcbe_pipeline_config_t, i);
                pipeline_config(result, config, pipeline, pids0);
            }
            gboolean free_plist(guint *coords G_GNUC_UNUSED, gpointer element,
                                gpointer user_data G_GNUC_UNUSED) {
                GArray **plist = element;
                if (*plist != NULL) {
                    g_array_free(*plist, TRUE);
                }
                return FALSE;
            }
            tree_branch_foreach(pids0, free_plist, NULL);
            tree_branch_free(pids0);
            tree_free(pids);
        }
    }
    else {
        g_info("Failed to find any pipelines in configuration '%s'",
               config->name);
    }
    return result;
}

static void
pipeline_config(wcbe_lagset_sink_config_t *cfg, const wcbe_config_t *config,
                wcbe_pipeline_config_t *pipeline, branch_t *pids)
{
    pipelinesock_t sck;
    sck.addr.sun_family = AF_UNIX;
    gint rc = g_snprintf(sck.addr.sun_path, sizeof(sck.addr.sun_path), "%s/%s",
                         cfg->sockdir, pipeline->id);
    g_assert(rc < sizeof(sck.addr.sun_path));
    gint cfg_offset =
        strrchr(cfg->sockdir, G_DIR_SEPARATOR) + 1 - cfg->sockdir;
    g_array_append_val(cfg->socks, sck);
    pipelinesock_t *pipelinesock =
        &g_array_index(cfg->socks, pipelinesock_t, cfg->socks->len - 1);
    pipelinesock->name = &(pipelinesock->addr.sun_path[cfg_offset]);

    // Use "pids" to find ids for products that get mapped to this
    // pipeline.

    // loop through "swmaps"
    gpointer elements = tree_branch_memory(pids);
    guint i;
    for (i = 0; i < pipeline->swmaps->len; ++i) {
        wcbe_swmap_config_t *swmap =
            &g_array_index(pipeline->swmaps, wcbe_swmap_config_t, i);
        // "pids" is indexed by (subarray, baseband, spectral
        // window, sub-band); first three are available as simple
        // attributes of swmap
        guint pids_idx[5] = {0, 0, 0, 0, COORD_VECTOR_END};
        pids_idx[0] = pipeline->subarray_index;
        pids_idx[1] = swmap->baseband_index;
        pids_idx[2] = swmap->spectral_window_index;
        guint j;
        for (j = 0; j < swmap->subband_indices->len; ++j) {
            pids_idx[3] = j;
            // with index to "pids", look up array of products
            GArray **products =
                elements + tree_branch_index_offset(pids, pids_idx);
            if (*products != NULL) {
                // Save mapping of product ids to socket. Expect multiple
                // products if for no other reason than mapping is on the
                // basis of subarray, not baseline.
                guint k;
                for (k = 0; k < (*products)->len; ++k) {
                    gint product_id = g_array_index((*products), gint, k);
                    g_hash_table_insert(cfg->destinations,
                                        GINT_TO_POINTER(product_id),
                                        pipelinesock);
                }
            }
        }
    }
}

/**
 * map_products:
 * @config: configuration
 * @branch: pointer to branch_t pointer
 *
 * Create a map of (subarray index, baseband (pair) index, spectral
 * window index, subband index) to a list of product ids based on a
 * backend configuration. The procedure to produce this map (actually
 * an n-ary tree) uses all components of a configuration, that is, the
 * input, output and pipeline components. The idea is to take each
 * product, and determine which subarray, baseband, etc. it gets
 * mapped to, and then enter the product at the appropriate index of
 * the map. Note that the created map is essentially the inverse of
 * the mapping from each product to subarray, baseband, etc. as
 * determined by the configuration.
 *
 * Schematically, the procedure of going from product to (subarray,
 * baseband, spectral window, subband) index is the following.
 * <ul>
 *  <li>product is mapped to
 *      (left stream id, right stream id, subband index)</li>
 *  <li>stream ids are mapped to station ids and baseband ids</li>
 *  <li>station ids are mapped to subarray indexes</li>
 *  <li>baseband ids are mapped to baseband indexes and range of
 *      spectral window indexes</li>
 *  <li>narrow list of spectral window indexes by scanning the
 *      pipeline "swmaps" for matching subarray, baseband and subband
 *      indexes</li>
 * </ul>
 *
 * 'branch' pointer will be set to branch that is created by this
 * function for the entire tree
 */
static tree_t *
map_products(const wcbe_config_t *config, branch_t **branch)
{
    g_return_val_if_fail(config != NULL, NULL);

    tree_t *result;

    if (config->output.subarrays == NULL) {
        return NULL;
    }

    // First, we allocate memory for the tree that holds arrays of
    // product ids. To do that, we need to generate a tree
    // "specification" string.
    const gchar *fmt(guint i) {
        return ((i == 0) ? "{%u,[" : ",{%u,[");
    }
    const gchar *close_str = "]}";
    GString *spec = g_string_new("");
    g_string_printf(spec, fmt(0), config->output.subarrays->len);
    guint i;
    for (i = 0; i < config->output.subarrays->len; ++i) {
        wcbe_subarray_config_t *sa = &g_array_index(
            config->output.subarrays, wcbe_subarray_config_t, i);
        g_string_append_printf(spec, fmt(i), sa->basebands->len);
        guint j;
        for (j = 0; j < sa->basebands->len; ++j) {
            wcbe_baseband_config_t *bb =
                &g_array_index(sa->basebands, wcbe_baseband_config_t, j);
            g_string_append_printf(spec, fmt(j), bb->spectral_windows->len);
            guint k;
            for (k = 0; k < bb->spectral_windows->len; ++k) {
                wcbe_spectral_window_config_t *sw = &g_array_index(
                    bb->spectral_windows, wcbe_spectral_window_config_t, k);
                g_string_append_printf(spec,
                                       ((k == 0) ? "{%u}" : ",{%u}"),
                                       sw->num_subbands);
            }
            g_string_append(spec, close_str);
        }
        g_string_append(spec, close_str);
    }
    g_string_append(spec, close_str);
    g_debug("Product map tree spec: %s", spec->str);

    // Create the tree and allocate memory for it
    result = tree_new(spec->str, sizeof(gpointer));
    if (result != NULL) {
        gint rc = tree_alloc(result);
        if (rc == 0) {
            *branch = tree_branch_root(result);
            // initialize tree by zeroing its elements
            tree_branch_clear(*branch);
            fill_product_tree(config, *branch);
        }
        else {
            g_warning("Failed to allocate product tree map memory");
        }
    }
    else {
        g_warning("Failed to create product map tree for %s", spec->str);
    }
    g_string_free(spec, TRUE);
    return result;
}

static void
fill_product_tree(const wcbe_config_t *config, branch_t *tree)
{
    if (config->input.products == NULL) {
        return;
    }

    guint coords[5] = {0, 0, 0, 0, COORD_VECTOR_END};
    gpointer elements = tree_branch_memory(tree);
    guint i;
    for (i = 0; i < config->input.products->len; ++i) {
        wcbe_product_config_t *product =
            g_ptr_array_index(config->input.products, i);
        coords[SUBBAND_IDX] = product->subband_index;
        wcbe_subarray_config_t *subarray =
            find_product_subarray(config, product, coords);
        if (subarray != NULL) {
            wcbe_baseband_config_t *baseband =
                find_product_baseband(subarray, product, coords);
            if (baseband != NULL) {
                g_assert(baseband->spectral_windows->len > 0);
                gboolean rc = find_product_spectral_window(
                    config, baseband, product, coords);
                if (rc) {
                    g_debug("Filling %s in tree at %d %d %d %d",
                            product->id, coords[SUBARRAY_IDX],
                            coords[BASEBAND_IDX], coords[SPECTRAL_WINDOW_IDX],
                            coords[SUBBAND_IDX]);
                    GArray **prd_array =
                        elements + tree_branch_index_offset(tree, coords);
                    if (*prd_array == NULL) {
                        *prd_array = g_array_new(FALSE, FALSE, sizeof(gint));
                    }
                    gint id = g_quark_from_string(product->id);
                    g_array_append_val(*prd_array, id);
                }
                else {
                    g_debug("Unfilled product %s: no matching "
                            "spectral window", product->id);
                }
            }
            else {
                g_debug("Unfilled product %s: no matching baseband",
                        product->id);
            }
        }
        else {
            g_debug("Unfilled product %s: no matching subarray",
                    product->id);
        }
    }
}

static wcbe_subarray_config_t *
find_product_subarray(const wcbe_config_t *config,
                      wcbe_product_config_t *product, guint coords[])
{
    wcbe_subarray_config_t *result = NULL;
    if (config->output.subarrays == NULL) {
        return NULL;
    }
    guint stnL = product->streamA->station_id;
    guint stnR = product->streamB->station_id;
    guint i;
    for (i = 0; result == NULL && i < config->output.subarrays->len; ++i) {
        wcbe_subarray_config_t *sa = &g_array_index(
            config->output.subarrays, wcbe_subarray_config_t, i);
        gboolean has_station(guint stn) {
            gboolean result = FALSE;
            guint i;
            for (i = 0; !result && i < sa->stations->len; ++i) {
                result = (g_array_index(sa->stations, guint, i) == stn);
            }
            return result;
        }
        if (has_station(stnL) && has_station(stnR)) {
            coords[SUBARRAY_IDX] = i;
            result = sa;
        }
    }
    return result;
}

static wcbe_baseband_config_t *
find_product_baseband(wcbe_subarray_config_t *subarray,
                      wcbe_product_config_t *product, guint coords[])
{
    wcbe_baseband_config_t *result = NULL;
    guint bbA = product->streamA->baseband_id;
    guint bbB = product->streamB->baseband_id;
    guint i;
    for (i = 0; result == NULL && i < subarray->basebands->len; ++i) {
        wcbe_baseband_config_t *baseband =
            &g_array_index(subarray->basebands, wcbe_baseband_config_t, i);
        if (baseband->num_polarizations == 2) {
            if ((bbA == baseband->baseband_ids[0]
                 && bbB == baseband->baseband_ids[1])
                || (bbB == baseband->baseband_ids[0]
                    && bbA == baseband->baseband_ids[1])) {
                coords[BASEBAND_IDX] = i;
                result = baseband;
            }
        }
        else {
            if ((bbA == baseband->baseband_ids[0]
                 || bbA == baseband->baseband_ids[1])
                && (bbB == baseband->baseband_ids[0]
                    || bbB == baseband->baseband_ids[1])) {
                coords[BASEBAND_IDX] = i;
                result = baseband;
            }
        }
    }
    return result;
}

static gboolean
find_product_spectral_window(const wcbe_config_t *config,
                             wcbe_baseband_config_t *baseband,
                             wcbe_product_config_t *product, guint coords[])
{
    gboolean result = FALSE;
    if (config->pipelines.pipelines == NULL) {
        return TRUE;
    }
    guint sw_idx_max = baseband->spectral_windows->len - 1;
    guint i;
    for (i = 0; i < config->pipelines.pipelines->len; ++i) {
        wcbe_pipeline_config_t *pipeline = &g_array_index(
            config->pipelines.pipelines, wcbe_pipeline_config_t, i);
        guint j;
        for (j = 0; j < pipeline->swmaps->len; ++j) {
            wcbe_swmap_config_t *swmap =
                &g_array_index(pipeline->swmaps, wcbe_swmap_config_t, j);
            if (pipeline->subarray_index == coords[SUBARRAY_IDX]
                && swmap->baseband_index == coords[BASEBAND_IDX]
                && swmap->spectral_window_index <= sw_idx_max) {
                guint k;
                for (k = 0; !result && k < swmap->subband_indices->len; ++k) {
                    if (g_array_index(swmap->subband_indices, guint, k)
                        == coords[SUBBAND_IDX]) {
                        coords[SPECTRAL_WINDOW_IDX] =
                            swmap->spectral_window_index;
                        coords[SUBBAND_IDX] = k;
                        result = TRUE;
                    }
                }
            }
        }
    }
    return result;
}

static void
destroy_config(wcbe_lagset_sink_config_t *cfg)
{
    gchar *name = g_strdup(cfg->wcfg->name);
    gboolean was_active = cfg->is_active;
    destroy_config_silently(cfg);
    if (was_active) {
        g_signal_emit(cfg->sink,
                      wcbe_lagset_sink_signals[SIGNAL_DEACTIVATE_CFG],
                      0, name, TRUE);
    }
    g_free(name);
}

static void
destroy_config_silently(wcbe_lagset_sink_config_t *cfg)
{
    deactivate_config(cfg);

    if (cfg->sockdir != NULL) {
        g_free(cfg->sockdir);
    }
    if (cfg->destinations != NULL) {
        g_hash_table_destroy(cfg->destinations);
    }
    if (cfg->socks != NULL) {
        g_array_free(cfg->socks, TRUE);
    }
    g_slice_free(wcbe_lagset_sink_config_t, cfg);
    WcbeLagsetSinkClass *klass =
        WCBE_LAGSET_SINK_CLASS(GST_ELEMENT_GET_CLASS(cfg->sink));
    g_dataset_id_remove_data(cfg->wcfg, klass->cfg_data);
}

static gint
open_sock(WcbeLagsetSink *sink)
{
    gint result = -1;
    sink->sockfd = socket(PF_UNIX, SOCK_DGRAM, 0);
    if (sink->sockfd != -1) {
        struct sockaddr_un addr = {.sun_family = AF_UNIX};
        gchar *path =
            g_build_filename(sink->sockdir, "lagsets", NULL);
        gint rc = g_strlcpy(addr.sun_path, path, sizeof(addr.sun_path));
        if (rc < sizeof(addr.sun_path)) {
            g_assert(rc > 0);
            rc = bind(sink->sockfd, (struct sockaddr *)&addr, sizeof(addr));
            if (rc != -1) {
                rc = fcntl(sink->sockfd, F_SETFL, O_NONBLOCK);
                if (rc == 0) {
                    result = 0;
                }
                else {
                    g_error("Failed to set socket to non-blocking mode: %s",
			    g_strerror(errno));
                }
            }
            else {
                g_warning("Failed to bind socket to address '%s': %s",
			  addr.sun_path, g_strerror(errno));
            }
        }
        else {
            g_warning("Socket name too long (%d >= %zd)",
		      rc, sizeof(addr.sun_path));
        }
        g_free(path);
    }
    return result;
}

static void
close_sock(WcbeLagsetSink *sink)
{
    gint rc;
    if (sink->sockfd != -1) {
        rc = close(sink->sockfd);
        if (rc != 0) {
            g_warning("Failed to close socket: %s", g_strerror(errno));
        }
        sink->sockfd = -1;
    }
    if (sink->sockdir != NULL) {
        gchar *path =
            g_build_filename(sink->sockdir, "lagsets", NULL);
        rc = unlink(path);
        if (rc != 0 && errno != ENOENT) {
            g_warning("Failed to unlink socket directory name '%s': %s",
		      path, g_strerror(errno));
        }
        g_free(path);
    }
}

static gboolean
wcbe_lagset_sink_create_config(WcbeLagsetSink *sink,
                               const wcbe_config_t *config,
                               GstPipeline *pipeline G_GNUC_UNUSED)
{
    g_return_val_if_fail(sink != NULL, FALSE);
    g_return_val_if_fail(config != NULL, FALSE);

    gboolean result = FALSE;
    if (sink->sockdir != NULL) {
        wcbe_lagset_sink_config_t *cfg = create_config(sink->sockdir, config);
        if (cfg != NULL) {
            cfg->sink = sink;
            WcbeLagsetSinkClass *klass =
                WCBE_LAGSET_SINK_CLASS(GST_ELEMENT_GET_CLASS(sink));
            g_dataset_id_set_data(config, klass->cfg_data, cfg);
            g_static_mutex_lock(&(sink->configs_mutex));
            g_hash_table_insert(sink->configs,
                                GINT_TO_POINTER(
                                    g_quark_from_string(config->name)),
                                cfg);
            g_static_mutex_unlock(&(sink->configs_mutex));
            result = TRUE;
        }
    }
    else {
        g_info("Configuration '%s' failed: 'sockdir' value is NULL",
               config->name);
    }
    return result;
}

static void
wcbe_lagset_sink_destroy_config(WcbeLagsetSink *sink,
                                const wcbe_config_t *config,
                                GstPipeline *pipeline G_GNUC_UNUSED)
{
    wcbe_lagset_sink_config_t *cfg = get_cfg_data(sink, config);
    if (cfg != NULL) {
        destroy_config(cfg);
        GstState state;
        gst_element_get_state(GST_ELEMENT(sink), &state, NULL, 0);
        if (state != GST_STATE_PLAYING) {
            forward_event(cfg, GST_EVENT_EOS, NULL);
        }
        g_static_mutex_lock(&(sink->configs_mutex));
        g_hash_table_remove(
            sink->configs, GINT_TO_POINTER(g_quark_from_string(config->name)));
        g_static_mutex_unlock(&(sink->configs_mutex));
    }
}

static void
handle_new_subscan(WcbeLagsetSink *sink, const gchar *configname,
                   const gchar *uid, guint epoch, guint seconds, guint cycles,
                   gpointer data G_GNUC_UNUSED)
{
    GQuark configid = g_quark_from_string(configname);
    wcbe_lagset_sink_config_t *cfg =
        g_hash_table_lookup(sink->configs, GINT_TO_POINTER(configid));
    GstStructure *s =
        gst_structure_new("new-subscan",
                          "epoch", G_TYPE_UINT, epoch,
                          "seconds", G_TYPE_UINT, seconds,
                          "cycles", G_TYPE_UINT, cycles,
                          NULL);
    if (uid != NULL && strlen(uid) != 0) {
        gst_structure_set(s, "uid", G_TYPE_STRING, uid, NULL);
    }
    forward_event(cfg, WCBE_EVENT_NEW_SUBSCAN, s);
    gst_structure_free(s);
}

static void
handle_cancel_subscan(WcbeLagsetSink *sink, const gchar *configname,
                      const gchar *uid, gpointer data G_GNUC_UNUSED)
{
    GQuark configid = g_quark_from_string(configname);
    wcbe_lagset_sink_config_t *cfg =
        g_hash_table_lookup(sink->configs, GINT_TO_POINTER(configid));
    GstStructure *s =
        gst_structure_new("cancel-subscan",
                          "uid", G_TYPE_STRING, uid,
                          NULL);
    forward_event(cfg, WCBE_EVENT_CANCEL_SUBSCAN, s);
    gst_structure_free(s);
}

static void
handle_activate_config(WcbeLagsetSink *sink, const wcbe_config_t *config,
                       guint64 time0)
{
    wcbe_lagset_sink_config_t *cfg = get_cfg_data(sink, config);
    if (!cfg->is_active) {
        cfg->is_active = TRUE;
	    if (cfg->flushing) {
	        forward_event(cfg, GST_EVENT_FLUSH_STOP, NULL);
             cfg->flushing = FALSE;
	    }
         GstEvent *event =
             gst_event_new_new_segment(FALSE, 1.0, GST_FORMAT_BUFFERS, time0,
                                       -1, time0);
         GstStructure *s = gst_structure_copy(gst_event_get_structure(event));
         gst_event_unref(event);
	    forward_event(cfg, GST_EVENT_NEWSEGMENT, s);
         gst_structure_free(s);
    }
    g_signal_emit(sink, wcbe_lagset_sink_signals[SIGNAL_ACTIVATE_CFG],
                  0, config->name, TRUE);
}

static void
deactivate_config(wcbe_lagset_sink_config_t *cfg)
{
    if (cfg->is_active) {
        cfg->is_active = FALSE;
        if (!cfg->flushing) {
            forward_event(cfg, GST_EVENT_FLUSH_START, NULL);
            cfg->flushing = TRUE;
        }
    }
}

static void
handle_deactivate_config(WcbeLagsetSink *sink, const wcbe_config_t *config)
{
    wcbe_lagset_sink_config_t *cfg = get_cfg_data(sink, config);
    if (cfg->is_active) {
        deactivate_config(cfg);
        g_signal_emit(sink, wcbe_lagset_sink_signals[SIGNAL_DEACTIVATE_CFG],
                      0, config->name, TRUE);
    }
}

static void
parent_set_cb(GstObject *gstobject, GstObject *parent,
              gpointer user_data G_GNUC_UNUSED)
{
    if (object_has_signal(G_OBJECT(parent), "create-configuration")) {
        g_signal_connect_swapped(parent, "create-configuration",
                                 G_CALLBACK(wcbe_lagset_sink_create_config),
                                 gstobject);
    }
    if (object_has_signal(G_OBJECT(parent), "destroy-configuration")) {
        g_signal_connect_swapped(parent, "destroy-configuration",
                                 G_CALLBACK(wcbe_lagset_sink_destroy_config),
                                 gstobject);
    }
    if (object_has_signal(G_OBJECT(parent), "new-subscan")) {
        g_signal_connect_swapped(parent, "new-subscan",
                                 G_CALLBACK(handle_new_subscan),
                                 gstobject);
    }
    if (object_has_signal(G_OBJECT(parent), "cancel-subscan")) {
        g_signal_connect_swapped(parent, "cancel-subscan",
                                 G_CALLBACK(handle_cancel_subscan),
                                 gstobject);
    }
}
