#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/epoll.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>

#include <wcbe.h>
#include <wcbe_lagset_src.h>
#include <wcbe_util.h>
#include <wcbe_lagset.h>

typedef enum {STREAM_MSG, EOF_MSG, ERROR_MSG, UNLOCK_MSG} queue_msg_type_t;
typedef struct queue_msg_t {
    queue_msg_type_t type;
    wcbe_stream_object_t object;
    union {
        GstBuffer* buffer;
        GstMessage *message;
        GstEvent *event;
    };
} queue_msg_t;

// The control socket stuff initially was copied from GStreamer's
// 'gstfdsrc' module.
//
//  * Copyright (C) 1999,2000 Erik Walthinsen <omega@cse.ogi.edu>
//  *                    2000 Wim Taymans <wtay@chello.be>
//  *                    2005 Philippe Khalaf <burger@speedy.org>
//
// Licensed under the LGPL v2.

/* the select call is also performed on the control sockets, that way
 * we can send special commands to unblock the select call */
#define CONTROL_STOP            'S'     /* stop the select call */
#define CONTROL_SOCKETS(src)   src->control_sock
#define WRITE_SOCKET(src)      src->control_sock[1]
#define READ_SOCKET(src)       src->control_sock[0]

#define SEND_COMMAND(src, command)          \
G_STMT_START {                              \
  unsigned char c; c = command;             \
  write (WRITE_SOCKET(src), &c, 1);         \
} G_STMT_END

#define READ_COMMAND(src, command, res)        \
G_STMT_START {                                 \
  res = read(READ_SOCKET(src), &command, 1);   \
} G_STMT_END

GST_DEBUG_CATEGORY_STATIC(wcbe_lagset_src_debug);
#define GST_CAT_DEFAULT wcbe_lagset_src_debug

/* Filter signals and args */
enum
{
    /* FILL ME */
    LAST_SIGNAL
};

enum
{
    ARG_0,
    ARG_INPUT,
    ARG_NUM_LAGSETS,
    ARG_LATEST_TS
};

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE(
    "src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS(WCBE_LAGSET_MEDIA_TYPE)
    );

GST_BOILERPLATE(WcbeLagsetSrc, wcbe_lagset_src, GstPushSrc,
                GST_TYPE_PUSH_SRC);

static void wcbe_lagset_src_finalize(GObject *object);
static void wcbe_lagset_src_set_property(
    GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec);
static void wcbe_lagset_src_get_property(
    GObject *object, guint prop_id, GValue *value, GParamSpec *pspec);
static gboolean wcbe_lagset_src_start(GstBaseSrc *src);
static gboolean wcbe_lagset_src_stop(GstBaseSrc *src);
static GstFlowReturn wcbe_lagset_src_create(GstPushSrc *src, GstBuffer **buf);
static gboolean wcbe_lagset_src_unlock(GstBaseSrc *src);
static gboolean wcbe_lagset_src_unlock_stop(GstBaseSrc *src);
static gpointer stream_input_fun(gpointer data);
static gint open_sock(const gchar *path);
static void free_queue_msg(gpointer data);
static queue_msg_type_t source_input(WcbeLagsetSrc *src);
static queue_msg_t *handle_stream_msg(WcbeLagsetSrc *src);

static void
wcbe_lagset_src_base_init(gpointer gclass)
{
    GST_DEBUG_CATEGORY_INIT(wcbe_lagset_src_debug,
                            "wcbe_lagset_src", 0,
                            "wcbe_lagset_src debug");
    static GstElementDetails element_details = {
        "WIDAR lagset source",
        "Source/WIDAR",
        "WIDAR lagset source",
        "Martin Pokorny <mpokorny@nrao.edu>"
    };
    GstElementClass *element_class = GST_ELEMENT_CLASS(gclass);

    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&src_factory));
    gst_element_class_set_details(element_class, &element_details);
}

/* initialize the plugin's class */
static void
wcbe_lagset_src_class_init(WcbeLagsetSrcClass * klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    GstBaseSrcClass *gstbasesrc_class = GST_BASE_SRC_CLASS(klass);
    GstPushSrcClass *gstpushsrc_class = GST_PUSH_SRC_CLASS(klass);

    gobject_class->set_property =
        GST_DEBUG_FUNCPTR(wcbe_lagset_src_set_property);
    gobject_class->get_property =
        GST_DEBUG_FUNCPTR(wcbe_lagset_src_get_property);
    gobject_class->finalize =
        GST_DEBUG_FUNCPTR(wcbe_lagset_src_finalize);

    g_object_class_install_property(
        gobject_class, ARG_INPUT,
        g_param_spec_string("input", "Lag set socket",
                            "Lag set source socket",
                            NULL, G_PARAM_READWRITE));

    g_object_class_install_property(
        gobject_class, ARG_NUM_LAGSETS,
        g_param_spec_uint64("num-lagsets",
                            "Number of lag sets sent",
                            "Get number of lag sets sent",
                            0, G_MAXUINT64,0, G_PARAM_READABLE));

    g_object_class_install_property(
        gobject_class, ARG_LATEST_TS,
        g_param_spec_string("latest-ts", "Latest lag set timestamp",
                            "Latest timestamp of received lag sets",
                            NULL, G_PARAM_READABLE));

    gstbasesrc_class->start = GST_DEBUG_FUNCPTR(wcbe_lagset_src_start);
    gstbasesrc_class->stop = GST_DEBUG_FUNCPTR(wcbe_lagset_src_stop);
    gstbasesrc_class->unlock = GST_DEBUG_FUNCPTR(wcbe_lagset_src_unlock);
    gstbasesrc_class->unlock_stop =
        GST_DEBUG_FUNCPTR(wcbe_lagset_src_unlock_stop);

    gstpushsrc_class->create = GST_DEBUG_FUNCPTR(wcbe_lagset_src_create);
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
wcbe_lagset_src_init(WcbeLagsetSrc *src, WcbeLagsetSrcClass *gclass)
{
    src->sockname = NULL;
    src->sockfd = -1;
    READ_SOCKET(src) = -1;
    WRITE_SOCKET(src) = -1;
    src->epollfd = -1;
    src->caps = NULL;
    src->num_lagsets = 0;
    src->flushing = FALSE;
    gst_base_src_set_live(GST_BASE_SRC(src), TRUE);
    src->msg_queue = g_async_queue_new_full(free_queue_msg);
    src->msg_thread = NULL;
    src->latest_epoch = 0;
    src->latest_seconds = 0;
    src->latest_cycles = 0;
    g_static_mutex_init(&(src->latest_mtx));
}

static void
wcbe_lagset_src_finalize(GObject *object)
{
    WcbeLagsetSrc *src = WCBE_LAGSET_SRC(object);

    g_async_queue_unref(src->msg_queue);
    if (src->sockname != NULL) {
        g_free(src->sockname);
    }
    if (src->sockfd != -1) {
        close(src->sockfd);
    }
    if (src->epollfd != -1) {
        close(src->epollfd);
    }
    if (READ_SOCKET(src) > 0) {
        close(READ_SOCKET(src));
    }
    if (WRITE_SOCKET(src) > 0) {
        close(WRITE_SOCKET(src));
    }
    g_static_mutex_free(&(src->latest_mtx));
    G_OBJECT_CLASS(parent_class)->finalize(object);
}

static void
wcbe_lagset_src_set_property(GObject *object, guint prop_id,
                             const GValue *value, GParamSpec *pspec)
{
    WcbeLagsetSrc *src = WCBE_LAGSET_SRC(object);

    switch (prop_id) {
      case ARG_INPUT:
        src->sockname = g_strdup(g_value_get_string(value));
        break;
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static void
wcbe_lagset_src_get_property(GObject *object, guint prop_id,
                             GValue *value, GParamSpec *pspec)
{
    WcbeLagsetSrc *src = WCBE_LAGSET_SRC(object);

    switch (prop_id) {
      case ARG_INPUT:
        g_value_set_string(value, src->sockname);
        break;
      case ARG_NUM_LAGSETS:
#if SIZEOF_UINT64_T == SIZEOF_UNSIGNED
        g_value_set_uint64(value,
                           (guint64)g_atomic_int_get(&(src->num_lagsets)));
#else
        // The following is exposed to a race condition w.r.t. the
        // pipeline thread updating the num_frames value --- we'll
        // live with it, a mutex lock seems like overkill.
        g_value_set_uint64(value, src->num_lagsets);
#endif
        break;
      case ARG_LATEST_TS: {
        g_static_mutex_lock(&(src->latest_mtx));
        guint epoch = src->latest_epoch;
        guint seconds = src->latest_seconds;
        guint cycles = src->latest_cycles;
        g_static_mutex_unlock(&(src->latest_mtx));
        gchar *ts = g_strdup_printf("%u:%u:%u", epoch, seconds, cycles);
        g_value_set_string(value, ts);
        g_free(ts);
        break;
      }
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

/* GstBaseSrc vmethod implementations */

static gboolean
wcbe_lagset_src_start(GstBaseSrc *gstbasesrc)
{
    gboolean result = TRUE;
    WcbeLagsetSrc *src = WCBE_LAGSET_SRC(gstbasesrc);

    src->sockfd = open_sock(src->sockname);
    if (src->sockfd != -1) {
        // Open the control socket pair
        gint control_sock[2];
        gint rc = socketpair(PF_UNIX, SOCK_STREAM, 0, control_sock);
        if (rc == 0) {
            READ_SOCKET(src) = control_sock[0];
            WRITE_SOCKET(src) = control_sock[1];
            fcntl(READ_SOCKET(src), F_SETFL, O_NONBLOCK);
            fcntl(WRITE_SOCKET(src), F_SETFL, O_NONBLOCK);
        }
        else {
            g_warning("Failed to create control socket pair: %s",
                      g_strerror(errno));
            result = FALSE;
        }
    }
    else {
        g_warning("Failed to open input socket '%s': %s", src->sockname,
                  g_strerror(errno));
        result = FALSE;
    }

    if (result) {
        GError *err = NULL;
        src->msg_thread = g_thread_create(stream_input_fun, src, TRUE, &err);
        if (src->msg_thread == NULL) {
            g_critical("Failed to spawn frame thread: %s", err->message);
            result = FALSE;
        }
    }

    if (result) {
        src->num_lagsets = 0;
        src->caps = gst_pad_get_caps(GST_BASE_SRC_PAD(gstbasesrc));
    }

    return result;
}

static gboolean
wcbe_lagset_src_stop(GstBaseSrc *gstbasesrc)
{
    WcbeLagsetSrc *src = WCBE_LAGSET_SRC(gstbasesrc);
    gint rc;

    if (src->msg_thread != NULL) {
        GThread *msg_thread = src->msg_thread;
        g_atomic_pointer_set(&(src->msg_thread), NULL);
        wcbe_lagset_src_unlock(gstbasesrc);
        g_thread_join(msg_thread);
    }

    if (src->sockfd > 0) {
        rc = close(src->sockfd);
        if (rc != 0) {
            g_warning("Failed to close lagset source socket: %s",
                      g_strerror(errno));
        }
        rc = unlink(src->sockname);
        if (rc != 0 && errno != ENOENT) {
            g_warning("Failed to unlink lagset socket directory entry: %s",
                      g_strerror(errno));
        }
    }
    src->sockfd = -1;

    if (READ_SOCKET(src) > 0) {
        close(READ_SOCKET(src));
        READ_SOCKET(src) = 0;
    }

    if (WRITE_SOCKET(src) > 0) {
        close(WRITE_SOCKET(src));
        WRITE_SOCKET(src) = 0;
    }

    if (src->caps != NULL) {
        gst_caps_unref(src->caps);
        src->caps = NULL;
    }

    return TRUE;
}

static gboolean
wcbe_lagset_src_unlock(GstBaseSrc *basesrc)
{
    WcbeLagsetSrc *src = WCBE_LAGSET_SRC(basesrc);
    SEND_COMMAND(src, CONTROL_STOP);
    return TRUE;
}

static gboolean
wcbe_lagset_src_unlock_stop(GstBaseSrc *basesrc)
{
    WcbeLagsetSrc *src = WCBE_LAGSET_SRC(basesrc);

    /* read all stop commands */
    int res = 0;
    while (res >= 0) {
        gchar command;
        READ_COMMAND(src, command, res);
    }
    g_async_queue_lock(src->msg_queue);
    guint len = g_async_queue_length_unlocked(src->msg_queue);
    while (len > 0) {
        queue_msg_t *msg = g_async_queue_pop_unlocked(src->msg_queue);
        if (msg->type != UNLOCK_MSG) {
            g_async_queue_push_unlocked(src->msg_queue, msg);
        }
        len--;
    }
    g_async_queue_unlock(src->msg_queue);
    return TRUE;
}

static GstFlowReturn
wcbe_lagset_src_create(GstPushSrc *pushsrc, GstBuffer **outbuf)
{
    GstFlowReturn result = GST_FLOW_OK;
    WcbeLagsetSrc *src = WCBE_LAGSET_SRC(pushsrc);

    // Read a frame or message
    queue_msg_t *msg = g_async_queue_pop(src->msg_queue);
    switch (msg->type) {
      case STREAM_MSG:
        // Check read result
        if (msg->buffer != NULL) {
            result = GST_FLOW_OK;
            *outbuf = msg->buffer;
            GST_BUFFER_OFFSET(*outbuf) = src->num_lagsets;
#if SIZEOF_UINT64_T == SIZEOF_UNSIGNED
            g_atomic_int_inc(&(src->num_lagsets));
#else
            src->num_lagsets++;
#endif
            wcbe_vect_t *lagset = (wcbe_vect_t *)GST_BUFFER_DATA(*outbuf);
            wcbe_lagset_header_t *lagset_hdr =
                &wcbe_vect_index(lagset, WCBE_LAGSET_HEADER_BLOCK_INDEX,
                                 wcbe_lagset_header_t, 0);
            guint epoch = lagset_hdr->timestamp.epoch;
            guint seconds = lagset_hdr->timestamp.seconds;
            guint cycles = lagset_hdr->timestamp.clock_cycles;
            if ((epoch > src->latest_epoch)
                || ((epoch == src->latest_epoch)
                    && ((seconds > src->latest_seconds)
                        || ((seconds == src->latest_seconds)
                            && (cycles > src->latest_cycles))))) {
                g_static_mutex_lock(&(src->latest_mtx));
                src->latest_epoch = epoch;
                src->latest_seconds = seconds;
                src->latest_cycles = cycles;
                g_static_mutex_unlock(&(src->latest_mtx));
            }
            *outbuf = msg->buffer;
        }
        else {
            result = GST_FLOW_UNEXPECTED;
        }
        msg->buffer = NULL;
        break;
      case ERROR_MSG:
        result = GST_FLOW_ERROR;
        break;
      case UNLOCK_MSG:
        result = GST_FLOW_WRONG_STATE;
        break;
      case EOF_MSG:
        result = GST_FLOW_UNEXPECTED;
        break;
      default:
        g_assert_not_reached();
        break;
    }
    free_queue_msg(msg);
    return result;
}

static gint
open_sock(const gchar *path)
{
    gint result = -1;
    gint sockfd = socket(PF_UNIX, SOCK_DGRAM, 0);
    if (sockfd != -1) {
        struct sockaddr_un addr = {.sun_family = AF_UNIX};
        gint rc = g_strlcpy(addr.sun_path, path, sizeof(addr.sun_path));
        if (rc < sizeof(addr.sun_path)) {
            g_assert(rc > 0);
            rc = bind(sockfd, (struct sockaddr *)&addr, sizeof(addr));
            if (rc != -1) {
                rc = fcntl(sockfd, F_SETFL, O_NONBLOCK);
                if (rc != -1) {
                    result = sockfd;
                }
                else {
                    GST_ERROR("Failed to set socket to non-blocking "
                              "read mode: %s", g_strerror(errno));
                }
            }
            else {
                GST_WARNING("Failed to bind Unix domain socket: %s",
                            g_strerror(errno));
            }
        }
        else {
            GST_WARNING("Unix domain socket path name too long");
        }
        if (result == -1) {
            close(sockfd);
        }
    }
    else {
        GST_ERROR("Failed to open Unix domain socket: %s",
                  g_strerror(errno));
    }
    return result;
}

static gpointer
stream_input_fun(gpointer data)
{
    WcbeLagsetSrc *src = WCBE_LAGSET_SRC(data);

    while (g_atomic_pointer_get(&(src->msg_thread))) {
        queue_msg_t *msg = NULL;
        switch (source_input(src)) {
          case STREAM_MSG:
            msg = handle_stream_msg(src);
            if (msg->object.type == WCBE_STREAM_EVENT) {
                switch (GST_EVENT_TYPE(msg->event)) {
                  case GST_EVENT_EOS:
                    if (src->flushing) {
                        gst_pad_push_event(
                            GST_BASE_SRC_PAD(GST_BASE_SRC(src)),
                            gst_event_new_flush_stop());
                        src->flushing = FALSE;
                    }
                    break;
                  case GST_EVENT_FLUSH_START:
                    src->flushing = TRUE;
                    break;
                  case GST_EVENT_FLUSH_STOP:
                    src->flushing = FALSE;
                    break;
                  case GST_EVENT_NEWSEGMENT:
                    src->num_lagsets = 0;
                    break;
                  default:
                    break;
                }
                gst_pad_push_event(GST_BASE_SRC_PAD(GST_BASE_SRC(src)),
                                   msg->event);
                msg->event = NULL;
                free_queue_msg(msg);
                msg = NULL;
            }
            else if (msg->object.type == WCBE_STREAM_MSG) {
                GstBus *bus = gst_element_get_bus(GST_ELEMENT(src));
                gst_bus_post(bus, msg->message);
                gst_object_unref(GST_OBJECT(bus));
                msg->message = NULL;
                free_queue_msg(msg);
                msg = NULL;
            }
            break;
          case ERROR_MSG:
            msg = g_slice_new(queue_msg_t);
            msg->type = ERROR_MSG;
            break;
          case UNLOCK_MSG:
            msg = g_slice_new(queue_msg_t);
            msg->type = UNLOCK_MSG;
            break;
          default:
            g_assert_not_reached();
            break;
        }
        if (msg != NULL) {
            g_async_queue_push(src->msg_queue, msg);
        }
    }
    return NULL;
}

static void
free_queue_msg(gpointer data)
{
    queue_msg_t *msg = data;
    if (msg->type == STREAM_MSG) {
        switch (msg->object.type) {
          case WCBE_STREAM_LAGSET:
            if (msg->buffer != NULL) {
                gst_buffer_unref(msg->buffer);
            }
            break;
          case WCBE_STREAM_EVENT:
            if (msg->event != NULL) {
                gst_event_unref(msg->event);
            }
            break;
          case WCBE_STREAM_MSG:
            if (msg->message != NULL) {
                gst_message_unref(msg->message);
            }
            break;
          default:
            g_assert_not_reached();
            break;
        }
    }
    g_slice_free(queue_msg_t, msg);
}

static queue_msg_type_t
source_input(WcbeLagsetSrc *src)
{
    queue_msg_type_t result;

    // 'select' on channel fd and control_sock to find which has data
    // to read.
    fd_set readfds;
    gint rc;
    FD_ZERO(&readfds);
    FD_SET(src->sockfd, &readfds);
    FD_SET(READ_SOCKET(src), &readfds);
    do {
        rc = select(FD_SETSIZE, &readfds, NULL, NULL, NULL);
    } while (rc == -1 && errno == EINTR);
    if (rc == -1) {
        g_debug("'select' failed: %s", g_strerror(errno));
        result = ERROR_MSG;
    }
    // Quit, if control socket has data.
    else if (FD_ISSET(READ_SOCKET(src), &readfds)) {
        /* read all commands */
        int res = 0;
        while (res >= 0) {
            gchar command;
            READ_COMMAND(src, command, res);
        }
        result = UNLOCK_MSG;
    }
    else {
        result = STREAM_MSG;
    }
    return result;
}

static queue_msg_t *
handle_stream_msg(WcbeLagsetSrc *src)
{
    queue_msg_t *result = NULL;
    gchar hdr[MAX(sizeof(wcbe_stream_event_t),
                  sizeof(wcbe_stream_lagset_t) + wcbe_vect_build_size(NULL))];
    struct iovec iovec[2] = {{.iov_base = hdr, .iov_len = sizeof(hdr)}};
    struct msghdr msghdr = {
        .msg_name = NULL, .msg_namelen = 0,
        .msg_control = NULL, .msg_controllen = 0,
        .msg_iov = iovec, .msg_iovlen = 1};
    errno = 0;
    gssize numbytes = recvmsg(src->sockfd, &msghdr, MSG_PEEK | MSG_TRUNC);
    if (G_LIKELY(numbytes > 0)) {
        wcbe_stream_object_t *stream_object = (wcbe_stream_object_t *)hdr;
        result = g_slice_new0(queue_msg_t);
        result->type = STREAM_MSG;
        result->object.type = stream_object->type;
        switch (stream_object->type) {
          case WCBE_STREAM_LAGSET: {
            wcbe_stream_lagset_t *stream_lagset =
                (wcbe_stream_lagset_t *)stream_object;
            wcbe_vect_t *lagset = &(stream_lagset->lagset);
            gsize lagset_size = lagset->total_size;
            result->buffer = gst_buffer_new();
            if (G_LIKELY(result->buffer != NULL)) {
                void *mem;
                gint rc = posix_memalign(&mem, wcbe_vect_alignment(),
                                         lagset_size);
                if (rc == 0) {
                    GST_BUFFER_MALLOCDATA(result->buffer) = mem;
                    GST_BUFFER_DATA(result->buffer) =
                        GST_BUFFER_MALLOCDATA(result->buffer);
                    GST_BUFFER_SIZE(result->buffer) = lagset_size;
                    gst_buffer_set_caps(result->buffer, src->caps);
                    wcbe_vect_check_alignment(
                        GST_BUFFER_DATA(result->buffer), TRUE);
                    msghdr.msg_iovlen = G_N_ELEMENTS(iovec);
                    lagset = (wcbe_vect_t *)GST_BUFFER_DATA(result->buffer);
                    wcbe_stream_lagset_init_iovec(
                        stream_lagset, lagset, lagset_size, iovec);
                    gssize expected_size =
                        offsetof(wcbe_stream_lagset_t, lagset) + lagset_size;
                    errno = 0;
                    numbytes = recvmsg(src->sockfd, &msghdr, 0);
                    if (G_LIKELY(numbytes == expected_size)) {
                        wcbe_lagset_header_t *lagset_hdr = &wcbe_vect_index(
                            lagset, WCBE_LAGSET_HEADER_BLOCK_INDEX,
                            wcbe_lagset_header_t, 0);
                        GST_BUFFER_TIMESTAMP(result->buffer) =
                            WIDAR_TIME_TO_NS(
                                TIMEVALUE(&(lagset_hdr->timestamp)));
                        // Rewrite lagset 'id' quark, since it was
                        // generated in another process.
                        lagset_hdr->id =
                            g_quark_from_string(stream_lagset->product_id);
                    }
                    else {
                        g_debug("Failed to read lagset: %s",
                                g_strerror(errno));
                        free_queue_msg(result);
                        result = NULL;
                    }
                }
                else {
                    g_warning("Failed to allocate memory for lagset: %s",
                              g_strerror(errno));
                }
            }
            else {
                GST_DEBUG("Failed to allocate buffer on source pad");
            }
            break;
          }
          case WCBE_STREAM_EVENT: {
            errno = 0;
            wcbe_stream_event_t *stream_event =
                (wcbe_stream_event_t *)stream_object;
            gsize event_size =
                sizeof(wcbe_stream_event_t) + stream_event->structure_len;
            stream_event = g_malloc(event_size);
            iovec[0].iov_base = stream_event;
            iovec[0].iov_len = event_size;
            errno = 0;
            numbytes = recvmsg(src->sockfd, &msghdr, 0);
            if (G_LIKELY(numbytes == event_size)) {
                switch (stream_event->event) {
                  case GST_EVENT_EOS:
                    result->event = gst_event_new_eos();
                    break;
                  case GST_EVENT_FLUSH_START:
                    result->event = gst_event_new_flush_start();
                    break;
                  case GST_EVENT_FLUSH_STOP:
                    result->event = gst_event_new_flush_stop();
                    break;
                  case GST_EVENT_NEWSEGMENT: {
                    GstStructure *s = gst_structure_from_string(
                        stream_event->structure, NULL);
                    result->event =
                        gst_event_new_custom(GST_EVENT_NEWSEGMENT, s);
                    g_warn_if_fail(result->event != NULL);
                    break;
                  }
                  case WCBE_EVENT_NEW_SUBSCAN: {
                    g_debug("got new-subscan event: %s",
                            stream_event->structure);
                    GstStructure *s = gst_structure_from_string(
                        stream_event->structure, NULL);
                    result->event =
                        gst_event_new_custom(WCBE_EVENT_NEW_SUBSCAN, s);
                    break;
                  }
                  case WCBE_EVENT_CANCEL_SUBSCAN: {
                    GstStructure *s = gst_structure_from_string(
                        stream_event->structure, NULL);
                    result->event =
                        gst_event_new_custom(WCBE_EVENT_CANCEL_SUBSCAN, s);
                    break;
                  }
                  default:
                    GST_DEBUG("Unexpected event on input: %d",
                              (gint)stream_event->event);
                    free_queue_msg(result);
                    result = NULL;
                    break;
                }
            }
            else {
                g_info("Failed to read expected number of bytes in "
                       "received stream event");
                free_queue_msg(result);
                result = NULL;
            }
            g_free(stream_event);
            break;
          }
          case WCBE_STREAM_MSG: {
            wcbe_stream_msg_t *stream_msg = (wcbe_stream_msg_t *)stream_object;
            gsize msg_size =
                sizeof(wcbe_stream_msg_t) + stream_msg->structure_len;
            stream_msg = g_malloc(msg_size);
            iovec[0].iov_base = stream_msg;
            iovec[0].iov_len = msg_size;
            errno = 0;
            numbytes = recvmsg(src->sockfd, &msghdr, 0);
            if (G_LIKELY(numbytes == msg_size)) {
                GstStructure *structure =
                    gst_structure_new(STREAM_MSG_NAME,
                                      "msg", G_TYPE_STRING,
                                      stream_msg->structure,
                                      NULL);
                if (G_LIKELY(structure != NULL)) {
                    result->message = gst_message_new_application(
                        GST_OBJECT(src), structure);
                }
                else {
                    g_debug("Failed to parse serialized structure in "
                            "stream message");
                    free_queue_msg(result);
                    result = NULL;
                }
            }
            else {
                g_info("Failed to read expected number of bytes in "
                       "received stream message");
                free_queue_msg(result);
                result = NULL;
            }
            g_free(stream_msg);
            break;
          }
          default:
            GST_DEBUG("Received unexpected object type: %d",
                      stream_object->type);
            free_queue_msg(result);
            result = NULL;
            break;
        }
    }
    return result;
}
