/**
 * wcbe_lc7
 *
 * Linear combination of the parts of seven-bit correlation products
 * (lags only, weights are dropped for now).
 */
#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <wcbe_lc7.h>
#include <wcbe.h>

typedef enum {
    ELEMENT_TYPE_INVALID = -1, ELEMENT_TYPE_INT32 = 4,
    ELEMENT_TYPE_INT64 = 8
} element_type_t;

GST_DEBUG_CATEGORY_STATIC(wcbe_lc7_debug);
#define GST_CAT_DEFAULT wcbe_lc7_debug

enum
{
    ARG_0,
    ARG_NUM_SENT,
    ARG_NUM_DROPPED
};

#define DEFAULT_CORR_BITS WCBE_4BITS
#define PARTIAL_LAGS_PAD_BASE_NAME "plags"
#define PARTIAL_WEIGHTS_PAD_BASE_NAME "pwgts"
#define COMBINED_LAGS_PAD_NAME "lags"

static gchar partial_lags_pad_base_name[] = PARTIAL_LAGS_PAD_BASE_NAME;
static gchar partial_weights_pad_base_name[] G_GNUC_UNUSED =
    PARTIAL_WEIGHTS_PAD_BASE_NAME;

static GstStaticPadTemplate partial_lags_sink_factory =
    GST_STATIC_PAD_TEMPLATE(
        PARTIAL_LAGS_PAD_BASE_NAME,
        GST_PAD_SINK,
        GST_PAD_ALWAYS,
        GST_STATIC_CAPS(WCBE_ARRAY_MEDIA_TYPE
                        ","WCBE_ELEMENT_TYPE_PROPERTY"="
                        WCBE_ELEMENT_TYPE_PROPERTY_TYPE "gfloat"));

static GstStaticPadTemplate partial_wgts_sink_factory =
    GST_STATIC_PAD_TEMPLATE(
        PARTIAL_WEIGHTS_PAD_BASE_NAME,
        GST_PAD_SINK,
        GST_PAD_ALWAYS,
        GST_STATIC_CAPS(WCBE_ARRAY_MEDIA_TYPE
                        ","WCBE_ELEMENT_TYPE_PROPERTY"="
                        WCBE_ELEMENT_TYPE_PROPERTY_TYPE"{guint32,guint64}"));

static GstStaticPadTemplate combined_lags_src_factory =
    GST_STATIC_PAD_TEMPLATE(
        COMBINED_LAGS_PAD_NAME,
        GST_PAD_SRC,
        GST_PAD_ALWAYS,
        GST_STATIC_CAPS(WCBE_ARRAY_MEDIA_TYPE
                        ","WCBE_ELEMENT_TYPE_PROPERTY"="
                        WCBE_ELEMENT_TYPE_PROPERTY_TYPE"gfloat"));

GST_BOILERPLATE(WcbeLc7, wcbe_lc7, GstElement, GST_TYPE_ELEMENT);

static void wcbe_lc7_get_property(
    GObject *object, guint prop_id, GValue *value, GParamSpec *pspec);
static GstFlowReturn wcbe_lc7_chain(GstPad *pad, GstBuffer *buffer);
static gboolean wcbe_lc7_event(GstPad *pad, GstEvent *event);
static void combine(WcbeLc7 *lc7);
static GList *wcbe_lc7_intlink(GstPad *pad);

#define NUM_PARTPRODS 4

static void
wcbe_lc7_base_init(gpointer gclass)
{
    GST_DEBUG_CATEGORY_INIT(wcbe_lc7_debug, "wcbe_lc7", 0, "wcbe_lc7 debug");
    static GstElementDetails element_details = {
        "WIDAR 7-bit product linear combinator",
        "Filter/WIDAR",
        "Perform linear combination of 7-bit partial products",
        "Martin Pokorny <mpokorny@nrao.edu>"
    };
    GstElementClass *element_class = GST_ELEMENT_CLASS(gclass);

    gst_element_class_add_pad_template(
        element_class,
        gst_static_pad_template_get(&partial_lags_sink_factory));
    gst_element_class_add_pad_template(
        element_class,
        gst_static_pad_template_get(&partial_wgts_sink_factory));
    gst_element_class_add_pad_template(
        element_class,
        gst_static_pad_template_get(&combined_lags_src_factory));
    gst_element_class_set_details(element_class, &element_details);
}

/* initialize the plugin's class */
static void
wcbe_lc7_class_init(WcbeLc7Class *klass)
{
    GObjectClass *gobject_class  = (GObjectClass *)klass;

    gobject_class->get_property =
        GST_DEBUG_FUNCPTR(wcbe_lc7_get_property);

    g_object_class_install_property(
        gobject_class, ARG_NUM_SENT,
        g_param_spec_uint("num-sent", "Number sent",
                          "Number of complete lag set products created "
                          "and pushed",
                          0, G_MAXUINT, 0, G_PARAM_READABLE));

    g_object_class_install_property(
        gobject_class, ARG_NUM_DROPPED,
        g_param_spec_uint("num-dropped", "Number dropped",
                          "Number of complete lag set products dropped",
                          0, G_MAXUINT, 0, G_PARAM_READABLE));
}

static void
wcbe_lc7_init(WcbeLc7 *lc7, WcbeLc7Class *klass)
{
    memset(lc7->lags, 0, sizeof(lc7->lags));
    memset(lc7->wgts, 0, sizeof(lc7->wgts));
    lc7->last_partprod = -1;
    lc7->num_dropped = 0;
    lc7->num_sent = 0;
    GstElement *gstelement = GST_ELEMENT(lc7);
    gst_element_create_all_pads(gstelement);
    guint i;
    for (i = 0; i < G_N_ELEMENTS(lc7->lags); ++i) {
        gchar name[sizeof(partial_lags_pad_base_name) + 2];
        g_snprintf(name, sizeof(name), PARTIAL_LAGS_PAD_BASE_NAME"%1u", i);
        GstPad *pad = gst_element_get_static_pad(gstelement, name);
        lc7->sinkpads[i][WCBE_LC7_LAGS_PAD_INDEX] = pad;
        if (pad != NULL) {
            gst_pad_set_chain_function(pad, wcbe_lc7_chain);
            gst_pad_set_internal_link_function(pad, wcbe_lc7_intlink);
            gst_pad_set_event_function(pad, wcbe_lc7_event);
        }
        g_object_unref(pad);
        g_snprintf(name, sizeof(name), PARTIAL_WEIGHTS_PAD_BASE_NAME"%1u", i);
        pad = gst_element_get_static_pad(gstelement, name);
        lc7->sinkpads[i][WCBE_LC7_WEIGHTS_PAD_INDEX] = pad;
        if (pad != NULL) {
            gst_pad_set_chain_function(pad, wcbe_lc7_chain);
            gst_pad_set_internal_link_function(pad, wcbe_lc7_intlink);
            gst_pad_set_event_function(pad, wcbe_lc7_event);
        }
        g_object_unref(pad);
    }
    lc7->srcpad =
        gst_element_get_static_pad(gstelement, COMBINED_LAGS_PAD_NAME);
    gst_pad_set_internal_link_function(lc7->srcpad, wcbe_lc7_intlink);
    gst_pad_set_event_function(lc7->srcpad, wcbe_lc7_event);
    g_object_unref(lc7->srcpad);
}

static void
wcbe_lc7_get_property(GObject *object, guint prop_id, GValue *value,
                      GParamSpec *pspec)
{
    WcbeLc7 *lc7 = WCBE_LC7(object);
    switch (prop_id) {
      case ARG_NUM_SENT:
        g_value_set_uint(value, g_atomic_int_get((gint *)&(lc7->num_sent)));
        break;
      case ARG_NUM_DROPPED:
        g_value_set_uint(value, g_atomic_int_get((gint *)&(lc7->num_dropped)));
        break;
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static GstFlowReturn
wcbe_lc7_chain(GstPad *pad, GstBuffer *buffer)
{
    WcbeLc7 *lc7 = WCBE_LC7(GST_OBJECT_PARENT(pad));

    guint partprod = NUM_PARTPRODS;
    GstBuffer **save_buffer = NULL;
    guint i;
    for (i = 0; save_buffer == NULL && i < NUM_PARTPRODS; ++i) {
        guint j;
        for (j = 0; save_buffer == NULL && j < WCBE_LC7_NUM_PADS; ++j) {
            if (pad == lc7->sinkpads[i][j]) {
                GstStructure *caps_sr =
                    gst_caps_get_structure(GST_BUFFER_CAPS(buffer), 0);
                if (G_LIKELY(gst_structure_get_uint(
                                 caps_sr,
                                 WCBE_PARTIAL_PRODUCT_PROPERTY,
                                 &partprod))) {
                    switch (j) {
                      case WCBE_LC7_WEIGHTS_PAD_INDEX:
                        save_buffer = &(lc7->wgts[partprod]);
                        break;
                      case WCBE_LC7_LAGS_PAD_INDEX:
                        save_buffer = &(lc7->lags[partprod]);
                        break;
                      default:
                        g_assert_not_reached();
                        break;
                    }
                }
            }
        }
    }
    g_return_val_if_fail(partprod < NUM_PARTPRODS, GST_FLOW_ERROR);
    g_return_val_if_fail(save_buffer != NULL, GST_FLOW_ERROR);

    if (lc7->last_partprod == -1) {
        lc7->last_partprod = partprod;
    }
    if (partprod != lc7->last_partprod) {
        g_debug("Switched partial product");
        for (i = 0; i < NUM_PARTPRODS; ++i) {
            if (lc7->lags[i] != NULL) {
                gst_buffer_unref(lc7->lags[i]);
                lc7->lags[i] = NULL;
            }
            if (lc7->wgts[i] != NULL) {
                gst_buffer_unref(lc7->wgts[i]);
                lc7->wgts[i] = 0;
            }
        }
    }
    *save_buffer = buffer;

    gboolean all_products = TRUE;
    for (i = 0; all_products && i < NUM_PARTPRODS; ++i) {
        all_products = (lc7->wgts[i] != NULL && lc7->lags[i] != NULL);
    }

    if (all_products) {
        combine(lc7);
        GstBuffer *lags_buffer = lc7->lags[WCBE_MM_PARTIAL_PRODUCT];
        GstFlowReturn rc = GST_FLOW_ERROR;
        if (gst_pad_set_caps(lc7->srcpad, GST_BUFFER_CAPS(lags_buffer))) {
            rc = gst_pad_push(lc7->srcpad, lags_buffer);
        }
        else {
            g_debug("Failed to set pad caps on src pad");
        }
        if (G_LIKELY(rc == GST_FLOW_OK)) {
            g_atomic_int_inc((gint *)&(lc7->num_sent));
        }
        else {
            g_atomic_int_inc((gint *)&(lc7->num_dropped));
        }
        for (i = 0; i < NUM_PARTPRODS; ++i) {
            gst_buffer_unref(lc7->lags[i]);
            lc7->lags[i] = NULL;
            gst_buffer_unref(lc7->wgts[i]);
            lc7->wgts[i] = NULL;
        }
        lc7->last_partprod = -1;
    }

    return GST_FLOW_OK;
}

static gboolean
wcbe_lc7_event(GstPad *pad, GstEvent *event)
{
    if (GST_EVENT_TYPE(event) == GST_EVENT_NEWSEGMENT) {
        WcbeLc7 *lc7 = WCBE_LC7(gst_pad_get_parent(pad));
        g_atomic_int_set((gint *)&(lc7->num_sent), 0);
        g_atomic_int_set((gint *)&(lc7->num_dropped), 0);
        gst_object_unref(lc7);
    }
    gboolean result = gst_pad_event_default(pad, event);
    return result;
}

static void
combine(WcbeLc7 *lc7)
{
    // This function overwrites the data in the input
    // lags[WCBE_MM_PARTIAL_PRODUCT] buffer with the linear
    // combination of lags[0..3].

    GstStructure *lags_caps_sr =
        gst_caps_get_structure(GST_BUFFER_CAPS(lc7->lags[0]), 0);

    guint lags_nrows;
    gst_structure_get_uint(lags_caps_sr, WCBE_ARRAY_NUM_ROWS_PROPERTY,
                           &lags_nrows);

    // read a few properties from the input buffers needed for looping
    // over the elements
    // get number of columns and row length of input arrays
    guint lags_ncols;
    gst_structure_get_uint(lags_caps_sr, WCBE_ARRAY_NUM_COLS_PROPERTY,
                           &lags_ncols);
    guint lags_rowlen = lags_ncols;
    gst_structure_get_uint(lags_caps_sr, WCBE_ARRAY_ROW_LENGTH_PROPERTY,
                           &lags_rowlen);
    guint lags_row_excess = lags_rowlen - lags_ncols;

    // result goes into lc7->lags[WCBE_MM_PARTIAL_PRODUCT] buffer
    gfloat *lagMM =
        (gfloat *)GST_BUFFER_DATA(lc7->lags[WCBE_MM_PARTIAL_PRODUCT]);
    gfloat *lagML =
        (gfloat *)GST_BUFFER_DATA(lc7->lags[WCBE_ML_PARTIAL_PRODUCT]);
    gfloat *lagLM =
        (gfloat *)GST_BUFFER_DATA(lc7->lags[WCBE_LM_PARTIAL_PRODUCT]);
    gfloat *lagLL =
        (gfloat *)GST_BUFFER_DATA(lc7->lags[WCBE_LL_PARTIAL_PRODUCT]);

    guint i;
    for (i = 0; i < lags_nrows; ++i) {
        guint j;
        for (j = 0; j < lags_ncols; ++j) {
            *lagMM = 64.0 * *lagMM + (8.0 * (*lagML + *lagLM) + *lagLL);
            ++lagMM;
            ++lagML;
            ++lagLM;
            ++lagLL;
        }
        lagMM += lags_row_excess;
        lagML += lags_row_excess;
        lagLM += lags_row_excess;
        lagLL += lags_row_excess;
    }
}

static GList *
wcbe_lc7_intlink(GstPad *pad)
{
    WcbeLc7 *lc7 = WCBE_LC7(GST_OBJECT_PARENT(pad));
    GList *result = NULL;
    if (pad == lc7->srcpad) {
        guint i;
        for (i = 0; i < 4; ++i) {
            result = g_list_append(result,
                                   lc7->sinkpads[i][WCBE_LC7_LAGS_PAD_INDEX]);
        }
    }
    else {
        guint i;
        for (i = 0; result == NULL && i < 4; ++i) {
            if (pad == lc7->sinkpads[i][WCBE_LC7_LAGS_PAD_INDEX]) {
                result = g_list_append(result, lc7->srcpad);
            }
        }
    }
    return result;
}
