#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>

#include <wcbe_sort.h>
#include <wcbe.h>
#include <user_marshal.h>
#include <wcbe_util.h>
#include <wcbe_config.h>
#include <wcbe_lagset.h>

#if GLIB_CHECK_VERSION(2, 14, 0)
// g_slice_dup macro is broken: it produces warnings under gcc 4.1 on
// a 64-bit platform. We define a corrected version for use here.
#undef g_slice_dup
#define g_slice_dup(type, mem)                                  \
  (1 ? g_slice_copy (sizeof (type), (mem)) : (type*) GINT_TO_POINTER ((type*) 0 == (mem)))
#else
// define our own g_slice_dup
#define g_slice_dup(type, mem) ({\
    type *_ret = g_slice_new(type); \
    memcpy(_ret, mem, sizeof(type)); \
    _ret;
})
#endif

// Resolution of DUMPTRIG signals, in number of WIDAR clock cycles.
#define DUMPTRIG_RESOLUTION 2

#define MAX_TIMEOUTS MAX_LAGSET_DELAY

// define the following macros to facilitate adding debug statements
// to mutex operations
#define MUTEX_LOCK(mtx) do { \
    g_mutex_lock(mtx); \
} while (0)
#define MUTEX_UNLOCK(mtx) do { \
    g_mutex_unlock(mtx); \
} while (0)
#define REC_MUTEX_LOCK(mtx) do { \
    g_static_rec_mutex_lock(&(mtx)); \
} while (0)
#define REC_MUTEX_UNLOCK(mtx) do { \
    g_static_rec_mutex_unlock(&(mtx)); \
} while (0)

#define LAG_FRAME_KEY_FTYPE(lf) (LF_BITFIELD_EXTRACT((lf)->key[0], 3, 0))

// Hash keys depend on ftype, ccc, bbid[xy], sbid[xy], sid[xy],
// cchip_id, recirc_blk_y, recirc_blk_x, and board_id only; thus, we
// apply a mask to the keys to retain only these fields.
static const lf_data_frame_word_t lta_hash_mask[] = {
    0x00001FFF, 0xFFFFFFFF, 0x0000FFFF, 0xFFFF0000
};
static const lf_data_frame_word_t cc_hash_mask[] = {
    0x000001FF, 0xFFFFFFFF, 0x0000FFFF, 0x00000000
};

typedef struct product_queues {
    GSList *queues;
    GHashTable *destinations;
} product_queues_t;

typedef struct wcbe_sort_config {
    product_queues_t pqs_template;
    const wcbe_config_t *wcfg;
    GQuark config_q;
    WcbeSort *sort;
} wcbe_sort_config_t;

typedef struct wcbe_sort_activation {
    product_queues_t pqs;
    wcbe_sort_config_t *config;
    activation_t *actvtn;
    WcbeSort *sort;
    gboolean timed_out;
} wcbe_sort_activation_t;

typedef struct output_queue {
    GQuark product_id;
    lf_widar_time_t last_ts;
    GQueue *queue;
    GMutex *qmutex;
    GTimer *timer;
    gdouble timeout;
    gint timeouts_remaining;
    guint64 hw_dump;
    guint num_frames_per_bin; // frames per lagset (w/o nbits multiplier)
    guint num_lags_per_bin;   // lags per lagset (w/o nbits multiplier)
    guint num_bins;           // bins (multiple of lagsets)
    guint nbits;              // number of correlation bits
    guint sbIdx;              // (renumbered) sub-band index
    guint stnA;               // (reordered) station A id
    guint stnB;               // (reordered) station B id
    guint bbA;                // (reordered) baseband A id
    guint bbB;                // (reordered) baseband B id
} output_queue_t;

typedef struct queue_element {
    GstBuffer *buffer;
    guint unfilled_frames;
    guint refcount;
} queue_element_t;

typedef struct frame_dest {
    output_queue_t *oqueue;
    guint header_idx;
    guint lags_idx;
    gboolean reverse_conjugate;
} frame_dest_t;

typedef struct stream_t {
    guint baseband;
    guint subband;
    guint station;
} stream_t;

GST_DEBUG_CATEGORY_STATIC(wcbe_sort_debug);
#define GST_CAT_DEFAULT wcbe_sort_debug

/* Filter signals and args */
enum
{
    SIGNAL_CREATE_CFG,
    SIGNAL_ACTIVATE_CFG,
    SIGNAL_DEACTIVATE_CFG,
    SIGNAL_DESTROY_CFG,
    LAST_SIGNAL
};

enum
{
    ARG_0,
    ARG_ACTIVE_CONFIGS,
    ARG_ORPHANED_FRAMES,
    ARG_NUM_LAGSETS
};


static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE(
    "sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS(WCBE_FRAME_MEDIA_TYPE)
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE(
    "src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS(WCBE_LAGSET_MEDIA_TYPE)
    );

static guint wcbe_sort_signals[LAST_SIGNAL] = {0};

GST_BOILERPLATE(WcbeSort, wcbe_sort, GstElement, GST_TYPE_ELEMENT);

static void wcbe_sort_set_property(
    GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec);
static void wcbe_sort_get_property(
    GObject *object, guint prop_id, GValue *value, GParamSpec *pspec);
static void wcbe_sort_finalize(GObject *object);
static GstFlowReturn wcbe_sort_chain(GstPad *pad, GstBuffer *buf);
static gboolean wcbe_sort_event(GstPad *pad, GstEvent *event);
static GstStateChangeReturn wcbe_sort_change_state(
    GstElement *element, GstStateChange transition);

static gint cmp_timestamps(const lf_widar_time_t *ts0,
                           const lf_widar_time_t *ts1,
                           const lf_widar_time_t *t0,
                           guint64 block);
static const lf_widar_time_t *queue_element_time(const queue_element_t *qe);
static GstFlowReturn push_queue_element(WcbeSort *sort, output_queue_t *oq,
                                        queue_element_t *qe);
static void queue_element_ref(queue_element_t *qe);
static void queue_element_unref(queue_element_t *qe);
static gpointer timeout_fun(gpointer data);
static gboolean timeout_queues(WcbeSort *sort, wcbe_sort_activation_t *act);
static guint64 set_timeouts(GList *activations);
static guint frame_hash(gconstpointer key);
static gboolean frame_equal(gconstpointer a, gconstpointer b);
static void destroy_config(WcbeSort *sort, wcbe_sort_config_t *cfg);
static void destroy_config_silently(WcbeSort *sort, wcbe_sort_config_t *cfg);
static wcbe_sort_config_t *create_config(WcbeSort *sort,
                                         const wcbe_config_t *config);
static void product_config(product_queues_t *pq,
                           const wcbe_product_config_t *product,
                           const wcbe_config_t *config);
static gint cmp_segments(gconstpointer a, gconstpointer b);
static void flush_all_queues(WcbeSort *sort, product_queues_t *pq);
static void free_lf_lag_frame(gpointer data);
static void free_frame_dest(gpointer data);
static void free_product_queues(product_queues_t *pq);
static void remove_activation(WcbeSort *sort,
                              wcbe_sort_activation_t *activation);
static void signal_deactivation(WcbeSort *sort,
                                const activation_t *activation);
static wcbe_sort_activation_t *find_activation(WcbeSort *sort,
                                               const lf_lag_frame_t *lf);
static void list_reverse_foreach(GList *list, GFunc func, gpointer user_data);
static gboolean wcbe_sort_create_config(
    WcbeSort *sort, const wcbe_config_t *config, GstPipeline *pipeline);
static void wcbe_sort_destroy_config(
    WcbeSort *sort, const wcbe_config_t *config, GstPipeline *pipeline);
static gboolean wcbe_sort_add_activation(
    WcbeSort *sort, activation_t *activation, GstPipeline *pipeline);
static void wcbe_sort_remove_activation(
    WcbeSort *sort, activation_t *activation, GstPipeline *pipeline);
static void parent_set_cb(GstObject *gstobject, GstObject *parent,
                          gpointer user_data);

static inline wcbe_sort_config_t *
get_cfg_data(WcbeSort *sort, const wcbe_config_t *config)
{
    WcbeSortClass *klass = WCBE_SORT_CLASS(GST_ELEMENT_GET_CLASS(sort));
    return g_dataset_id_get_data(config, klass->cfg_data);
}

static void
wcbe_sort_base_init(gpointer gclass)
{
    GST_DEBUG_CATEGORY_INIT(wcbe_sort_debug, "wcbe_sort_debug",
                            0, "wcbe_sort debug");
    static GstElementDetails element_details = {
        "WIDAR lag set sorter",
        "Sorter/WIDAR",
        "Assemble and temporally sort WIDAR lag sets",
        "Martin Pokorny <mpokorny@nrao.edu>"
    };
    GstElementClass *element_class = GST_ELEMENT_CLASS(gclass);

    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&src_factory));
    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&sink_factory));
    gst_element_class_set_details(element_class, &element_details);
}

/* initialize the plugin's class */
static void
wcbe_sort_class_init(WcbeSortClass *klass)
{
    GObjectClass *gobject_class;
    GstElementClass *gstelement_class;

    gobject_class = (GObjectClass *)klass;
    gstelement_class = (GstElementClass *)klass;

    gobject_class->set_property =
        GST_DEBUG_FUNCPTR(wcbe_sort_set_property);
    gobject_class->get_property =
        GST_DEBUG_FUNCPTR(wcbe_sort_get_property);
    gobject_class->finalize =
        GST_DEBUG_FUNCPTR(wcbe_sort_finalize);

    gstelement_class->change_state =
        GST_DEBUG_FUNCPTR(wcbe_sort_change_state);

    g_object_class_install_property(
        gobject_class, ARG_ACTIVE_CONFIGS,
        g_param_spec_string("active-configs", "active configurations",
                            "List of currently active configurations",
                            NULL, G_PARAM_READABLE));

    g_object_class_install_property(
        gobject_class, ARG_ORPHANED_FRAMES,
        g_param_spec_uint64("orphaned-frames",
                            "Number of orphaned lag frames",
                            "Get number of frames received without associated "
                            "lag set in active configuration",
                            0, G_MAXUINT64,0, G_PARAM_READABLE));

    g_object_class_install_property(
        gobject_class, ARG_NUM_LAGSETS,
        g_param_spec_uint64("num-lagsets",
                            "Number of lag sets",
                            "Get number of sent lag sets.",
                            0, G_MAXUINT64,0, G_PARAM_READABLE));

    wcbe_sort_signals[SIGNAL_CREATE_CFG] =
        g_signal_new("config-created", G_TYPE_FROM_CLASS(klass),
                     G_SIGNAL_RUN_FIRST,
                     0, NULL, NULL,
                     g_cclosure_user_marshal_VOID__STRING_BOOLEAN,
                     G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_BOOLEAN);

    wcbe_sort_signals[SIGNAL_ACTIVATE_CFG] =
        g_signal_new("config-activated", G_TYPE_FROM_CLASS(klass),
                     G_SIGNAL_RUN_FIRST,
                     0, NULL, NULL,
                     g_cclosure_user_marshal_VOID__STRING_BOOLEAN,
                     G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_BOOLEAN);

    wcbe_sort_signals[SIGNAL_DEACTIVATE_CFG] =
        g_signal_new("config-deactivated", G_TYPE_FROM_CLASS(klass),
                     G_SIGNAL_RUN_FIRST,
                     0, NULL, NULL,
                     g_cclosure_marshal_VOID__POINTER,
                     G_TYPE_NONE, 1, G_TYPE_POINTER);

    wcbe_sort_signals[SIGNAL_DESTROY_CFG] =
        g_signal_new("config-destroyed", G_TYPE_FROM_CLASS(klass),
                     G_SIGNAL_RUN_FIRST,
                     0, NULL, NULL,
                     g_cclosure_user_marshal_VOID__STRING_BOOLEAN,
                     G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_BOOLEAN);
    klass->cfg_data = g_quark_from_string("wcbe_sort_cfg_data");
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
wcbe_sort_init(WcbeSort *sort,
               WcbeSortClass *gclass)
{
    GstElementClass *klass = GST_ELEMENT_GET_CLASS(sort);

    g_signal_connect(GST_OBJECT(sort), "parent-set",
                     G_CALLBACK(parent_set_cb), NULL);

    sort->sinkpad = gst_pad_new_from_template(
        gst_element_class_get_pad_template(klass, "sink"), "sink");

    sort->srcpad = gst_pad_new_from_template(
        gst_element_class_get_pad_template(klass, "src"), "src");
    gst_pad_use_fixed_caps(sort->srcpad);
    gst_pad_set_caps(sort->srcpad,
                     gst_static_pad_template_get_caps(&src_factory));

    gst_element_add_pad(GST_ELEMENT(sort), sort->sinkpad);
    gst_element_add_pad(GST_ELEMENT(sort), sort->srcpad);
    gst_pad_set_chain_function(sort->sinkpad, wcbe_sort_chain);
    gst_pad_set_event_function(sort->sinkpad, wcbe_sort_event);

    sort->dummy_buffer = gst_buffer_new();
    GST_BUFFER_MALLOCDATA(sort->dummy_buffer) =
        (void *)wcbe_vect_alloc(1, 1, sizeof(wcbe_lagset_header_t));
    GST_BUFFER_DATA(sort->dummy_buffer) =
        GST_BUFFER_MALLOCDATA(sort->dummy_buffer);
    sort->changed_active_configs_mutex = g_mutex_new();
    sort->changed_active_configs = g_cond_new();
    sort->active_configs_changed = FALSE;
    sort->activations = NULL;
    sort->active_configs = g_array_new(FALSE, FALSE, sizeof(GQuark));
    g_static_rec_mutex_init(&(sort->activations_mutex));
    GError *err = NULL;
    sort->timeout_thread = g_thread_create(timeout_fun, sort, TRUE, &err);
    if (sort->timeout_thread == NULL) {
        GST_ERROR("Failed to spawn timeout thread: %s", err->message);
    }
}

static void
wcbe_sort_finalize(GObject *object)
{
    WcbeSort *sort = WCBE_SORT(object);

    if (sort->timeout_thread != NULL) {
        GThread *timeout_thread = sort->timeout_thread;
        g_atomic_pointer_set(&(sort->timeout_thread), NULL);
        MUTEX_LOCK(sort->changed_active_configs_mutex);
        g_cond_signal(sort->changed_active_configs);
        MUTEX_UNLOCK(sort->changed_active_configs_mutex);
        g_thread_join(timeout_thread);
    }

    // now we may ignore locking, since timeout thread is gone
    void deactivate(gpointer data, gpointer user_data) {
        wcbe_sort_activation_t *activation = data;
        signal_deactivation(sort, activation->actvtn);
    }
    g_list_foreach(sort->activations, deactivate, NULL);
    g_list_free(sort->activations);
    g_array_free(sort->active_configs, TRUE);
    g_static_rec_mutex_free(&(sort->activations_mutex));

    if (sort->changed_active_configs_mutex != NULL) {
        g_mutex_free(sort->changed_active_configs_mutex);
    }
    if (sort->changed_active_configs != NULL) {
        g_cond_free(sort->changed_active_configs);
    }
    if (sort->dummy_buffer != NULL) {
        gst_buffer_unref(sort->dummy_buffer);
    }
    G_OBJECT_CLASS(parent_class)->finalize(object);
}

static void
wcbe_sort_set_property(GObject *object, guint prop_id,
                       const GValue *value, GParamSpec *pspec)
{
    switch (prop_id) {
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static void
wcbe_sort_get_property(GObject *object, guint prop_id,
                       GValue *value, GParamSpec *pspec)
{
    WcbeSort *sort = WCBE_SORT(object);

    switch (prop_id) {
      case ARG_ACTIVE_CONFIGS: {
        // Note that the same configuration name can appear multiple
        // times in the returned list. I think that's good, since
        // there is valuable information in the repetitions.
        GString *configs = g_string_new("");
        void active_config_name(gpointer data, gpointer user_data) {
            const wcbe_sort_activation_t *act = data;
            if (act->actvtn->in_use) {
                const gchar *fmt = ((configs->len == 0) ? "%s" : "\t%s");
                g_string_append_printf(configs, fmt, act->config->wcfg->name);
            }
        }
        REC_MUTEX_LOCK(sort->activations_mutex);
        list_reverse_foreach(g_list_last(sort->activations),
                             active_config_name, NULL);
        REC_MUTEX_UNLOCK(sort->activations_mutex);
        g_value_set_string(value, configs->str);
        g_string_free(configs, TRUE);
        break;
      }
      case ARG_ORPHANED_FRAMES:
#if SIZEOF_UINT64_T == SIZEOF_UNSIGNED
        g_value_set_uint64(value,
                           (guint64)g_atomic_int_get(
                               &(sort->num_orphaned_frames)));
#else
        // The following is exposed to a race condition w.r.t. the
        // pipeline thread updating the num_orphaned_frames value ---
        // we'll live with it, a mutex lock seems like overkill.
        g_value_set_uint64(value, sort->num_orphaned_frames);
#endif
        break;
      case ARG_NUM_LAGSETS:
#if SIZEOF_UINT64_T == SIZEOF_UNSIGNED
        g_value_set_uint64(value,
                           (guint64)g_atomic_int_get(
                               &(sort->num_lagsets)));
#else
        // NB: above comment applies here too
        g_value_set_uint64(value, sort->num_lagsets);
#endif
        break;
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

/* GstElement vmethod implementations */

/* chain function
 * this function does the actual processing
 */

static inline guint
index_w_bin(guint base_idx, WcbeCorrBits nbits, guint size_per_subprod_per_bin,
            guint bin, guint num_bins)
{
    guint result;
    switch (nbits) {
      case WCBE_4BITS:
        result = base_idx + size_per_subprod_per_bin * bin;
        break;
      case WCBE_7BITS:
        result = (base_idx % size_per_subprod_per_bin)
            + size_per_subprod_per_bin
            * (bin + (base_idx / size_per_subprod_per_bin) * num_bins);
        break;
      default:
        g_assert_not_reached();
        result = 0;
        break;
    }
    return result;
}

static const lf_widar_time_t *
queue_element_time(const queue_element_t *qe)
{
    const wcbe_lagset_header_t *ls = &wcbe_vect_index(
        (wcbe_vect_t *)GST_BUFFER_DATA(qe->buffer),
        WCBE_LAGSET_HEADER_BLOCK_INDEX, wcbe_lagset_header_t, 0);
    return &(ls->timestamp);
}

static GstFlowReturn
wcbe_sort_chain(GstPad *pad, GstBuffer *buf)
{
    GstFlowReturn result = GST_FLOW_OK;
    WcbeSort *sort = WCBE_SORT(gst_pad_get_parent(pad));

    lf_lag_frame_t *lf = (lf_lag_frame_t *)GST_BUFFER_DATA(buf);
    g_assert(lf != NULL);
    wcbe_sort_activation_t *sort_activation = find_activation(sort, lf);
    if (G_LIKELY(sort_activation != NULL)) {
        frame_dest_t *dest = NULL;
        if (G_LIKELY(sort_activation->pqs.destinations != NULL)) {
            dest = g_hash_table_lookup(sort_activation->pqs.destinations, lf);
        }
        if (G_LIKELY(dest != NULL)) {
            output_queue_t *oq = dest->oqueue;
            g_assert(oq->queue != NULL);
            g_atomic_int_set(&(oq->timeouts_remaining), MAX_TIMEOUTS);
            queue_element_t dummy_qe = {.buffer = sort->dummy_buffer};
            wcbe_lagset_header_t *dummy_ls_hdr = &wcbe_vect_index(
                (wcbe_vect_t *)GST_BUFFER_DATA(dummy_qe.buffer), 0,
                wcbe_lagset_header_t, 0);
            memcpy(&(dummy_ls_hdr->timestamp), &(lf->timestamp),
                   sizeof(dummy_ls_hdr->timestamp));
            gint cmp(gconstpointer a, gconstpointer b) {
                const queue_element_t *qea = a;
                const queue_element_t *qeb = b;
                return cmp_timestamps(
                    queue_element_time(qea), queue_element_time(qeb),
                    &(sort_activation->actvtn->time), oq->hw_dump);
            }
            gint cmpd(gconstpointer a, gconstpointer b,
                      gpointer user_data G_GNUC_UNUSED) {
                return -cmp(a, b);
            }
            MUTEX_LOCK(oq->qmutex);
            GList *ls = g_queue_find_custom(oq->queue, &dummy_qe, cmp);
            queue_element_t *lagset_qe = NULL;
            wcbe_vect_t *lagset = NULL;
            if (ls != NULL) {
                lagset_qe = ls->data;
                queue_element_ref(lagset_qe);
                MUTEX_UNLOCK(oq->qmutex);
                g_assert(lagset_qe->buffer != NULL);
                lagset = (wcbe_vect_t *)GST_BUFFER_DATA(lagset_qe->buffer);
            }
            else if (cmp_timestamps(&(lf->timestamp), &(oq->last_ts),
                                    &(sort_activation->actvtn->time),
                                    oq->hw_dump) > 0) {
                MUTEX_UNLOCK(oq->qmutex);
                lagset_qe = g_slice_new(queue_element_t);
                lagset_qe->unfilled_frames = oq->num_frames_per_bin
                    * oq->num_bins * (oq->nbits == WCBE_7BITS ? 4 : 1);
                lagset_qe->refcount = 1;
                lagset_qe->buffer = gst_buffer_new();
                if (G_LIKELY(lagset_qe->buffer != NULL)) {
                    lagset = wcbe_lagset_new(oq->num_frames_per_bin,
                                             oq->num_lags_per_bin,
                                             oq->num_bins,
                                             oq->nbits, TRUE);
                    GST_BUFFER_MALLOCDATA(lagset_qe->buffer) = (void *)lagset;
                    GST_BUFFER_DATA(lagset_qe->buffer) =
                        GST_BUFFER_MALLOCDATA(lagset_qe->buffer);
                    GST_BUFFER_SIZE(lagset_qe->buffer) = lagset->total_size;
                    gst_buffer_set_caps(lagset_qe->buffer,
                                        GST_PAD_CAPS(sort->srcpad));
                    wcbe_vect_check_alignment(lagset, TRUE);
                    g_assert(GST_BUFFER_SIZE(lagset_qe->buffer)
                             == lagset->total_size);
                    wcbe_lagset_header_t *lagset_hdr = &wcbe_vect_index(
                        (wcbe_vect_t *)GST_BUFFER_DATA(lagset_qe->buffer),
                        WCBE_LAGSET_HEADER_BLOCK_INDEX, wcbe_lagset_header_t,
                        0);
                    lagset_hdr->config = sort_activation->config->config_q;
                    lagset_hdr->id = oq->product_id;
                    lagset_hdr->sbIdx = oq->sbIdx;
                    lagset_hdr->stnA = oq->stnA;
                    lagset_hdr->stnB = oq->stnB;
                    lagset_hdr->bbA = oq->bbA;
                    lagset_hdr->bbB = oq->bbB;
                    memcpy(&(lagset_hdr->timestamp),
                           &(dummy_ls_hdr->timestamp),
                           sizeof(dummy_ls_hdr->timestamp));
                    GST_BUFFER_TIMESTAMP(lagset_qe->buffer) =
                        WIDAR_TIME_TO_NS(TIMEVALUE(&(lagset_hdr->timestamp)));
                    queue_element_ref(lagset_qe); // on behalf of queue
                    MUTEX_LOCK(oq->qmutex);
                    g_queue_insert_sorted(oq->queue, lagset_qe, cmpd, NULL);
                    MUTEX_UNLOCK(oq->qmutex);
                }
                else {
                    GST_INFO("Failed to alloc buffer on source pad");
                }
            }
            if (G_LIKELY(lagset != NULL)) {
                wcbe_lagset_header_t *lagset_hdr = &wcbe_vect_index(
                    (wcbe_vect_t *)GST_BUFFER_DATA(lagset_qe->buffer),
                    WCBE_LAGSET_HEADER_BLOCK_INDEX, wcbe_lagset_header_t,
                    0);
                // The 'mod lagset_hdr->num_bins' below allows us to map
                // multiple LTA bins to a single, logical phase bin.
                // Whether this is useful beyond the case in which the
                // number of logical phase bins is one, I'm not sure.
                // Also note that the "bank" bit of the lta_bin value
                // is ignored.
                guint bin;
                guint num_lags;
                if (G_LIKELY(LAG_FRAME_KEY_FTYPE(lf)
                             == LF_LTA_DATA_FRAME_FTYPE)) {
                    num_lags = LF_LTA_FRAME_NUM_LAGS(lf);
                    bin = LF_LTA_FRAME_PHASE_BIN(lf);
                }
                else {
                    num_lags = LF_CC_FRAME_NUM_CLAGS(lf);
                    bin = LF_CC_FRAME_LTA_BIN(lf);
                }
                bin = bin % lagset_hdr->num_bins;
                guint header_idx =
                    index_w_bin(dest->header_idx, lagset_hdr->nbits,
                                lagset_hdr->num_frames_per_bin, bin,
                                lagset_hdr->num_bins);
                wcbe_lagset_frame_header_t *header = &wcbe_vect_index(
                    lagset, WCBE_LAGSET_FRAME_HEADERS_BLOCK_INDEX,
                    wcbe_lagset_frame_header_t, header_idx);
                if (!header->filled) {
                    header->filled = TRUE;
                    guint lags_idx =
                        index_w_bin(dest->lags_idx, lagset_hdr->nbits,
                                    lagset_hdr->num_lags_per_bin, bin,
                                    lagset_hdr->num_bins);
                    lf_lag_t *lags = &wcbe_vect_index(
                        lagset, WCBE_LAGSET_LAGS_BLOCK_INDEX, lf_lag_t,
                        lags_idx);
                    memcpy(header->key, lf->key,
                           LF_KEY_LENGTH * sizeof(lf_data_frame_word_t));
                    memcpy(&(header->timestamp), &(lf->timestamp),
                           sizeof(header->timestamp));
                    header->data_valid_center =
                        LF_LAG_FRAME_DVCOUNT_CENTER(lf);
                    header->data_valid_edge = LF_LAG_FRAME_DVCOUNT_EDGE(lf);
                    header->reverse_conjugate = dest->reverse_conjugate;
                    if (G_LIKELY(LAG_FRAME_KEY_FTYPE(lf)
                                 == LF_LTA_DATA_FRAME_FTYPE)) {
                        header->checksum = LF_LTA_FRAME_CHECKSUM(lf);
                    }
                    else {
                        header->data_bias = LF_CC_FRAME_DATA_BIAS(lf);
                        header->checksum = LF_CC_FRAME_CHECKSUM(lf);
                    }
                    if (dest->reverse_conjugate) {
                        header->key[1] =
                            (((header->key[1] >> 16) & 0xFFFF)
                             | ((header->key[1] << 16) & 0xFFFF0000));
                        header->key[2] =
                            ((header->key[2] & 0xFFFF0000)
                             | (((header->key[2] >> 8) & 0xFF)
                                | ((header->key[2] & 0xFF) << 8)));
                        lf_lag_t *lag_src = (lf_lag_t *)&LF_LAG_FRAME_LAG0(lf);
                        lf_lag_t *lag_dest = &(lags[num_lags - 1]);
                        guint i;
                        for (i = 0; i < num_lags; ++i) {
                            // "~" operator is a GCC extension for
                            // complex conjugation
                            *lag_dest-- = ~(*lag_src++);
                        }
                    }
                    else {
                        memcpy(lags, &LF_LAG_FRAME_LAG0(lf),
                               num_lags * sizeof(lf_lag_t));
                    }
                    lagset_qe->unfilled_frames--;
                }
                else {
                    g_info("Received frame mapped to previously "
                           "filled frame in lagset.");
                }
            }
            else {
                g_info("Failed to find lagset for received frame (%s)",
                       g_quark_to_string(oq->product_id));
            }
            if (G_LIKELY(lagset_qe != NULL)) {
                queue_element_unref(lagset_qe);

                // Output lag sets if appropriate
                MUTEX_LOCK(oq->qmutex);
                queue_element_t *qe = g_queue_peek_tail(oq->queue);
                while ((qe != NULL && qe->unfilled_frames == 0)
                       || (g_queue_get_length(oq->queue) > MAX_LAGSET_DELAY)) {
                    guint len = g_queue_get_length(oq->queue);
                    if (len > 3) {
                        void pts(gpointer data, gpointer user_data) {
                            queue_element_t *e = data;
                            const lf_widar_time_t *t = queue_element_time(e);
                            g_debug("%u:%u:%u, %u", t->epoch, t->seconds,
                                    t->clock_cycles, e->unfilled_frames);
                        }
                        g_queue_foreach(oq->queue, pts, NULL);
                    }
                    qe = g_queue_pop_tail(oq->queue);
                    push_queue_element(sort, oq, qe);
                    qe = g_queue_peek_tail(oq->queue);
                }
                MUTEX_UNLOCK(oq->qmutex);
            }
        }
        else {
            sort->num_orphaned_frames++;
        }
    }
    else {
        sort->num_orphaned_frames++;
    }

    gst_buffer_unref(buf);
    gst_object_unref(GST_OBJECT(sort));

    return result;
}

static gboolean
wcbe_sort_event(GstPad *pad, GstEvent *event)
{
    gboolean result;
    WcbeSort *sort = WCBE_SORT(gst_pad_get_parent(pad));
    switch (GST_EVENT_TYPE(event)) {
      case GST_EVENT_EOS:
      case GST_EVENT_FLUSH_START: {
        // Output all remaining lag sets. Flush queues in order from
        // oldest to newest activation times.
        REC_MUTEX_LOCK(sort->activations_mutex);
        void flush(gpointer data, gpointer user_data) {
            wcbe_sort_activation_t *act = data;
            flush_all_queues(sort, &(act->pqs));
        }
        list_reverse_foreach(g_list_last(sort->activations), flush, NULL);
        REC_MUTEX_UNLOCK(sort->activations_mutex);
        result = gst_pad_push_event(sort->srcpad, event);
        break;
      }
      default:
        result = gst_pad_event_default(pad, event);
        break;
    }
    gst_object_unref(GST_OBJECT(sort));
    return result;
}


static GstStateChangeReturn
wcbe_sort_change_state(GstElement *element, GstStateChange transition)
{
    GstStateChangeReturn result = GST_STATE_CHANGE_SUCCESS;
    WcbeSort *sort = WCBE_SORT(element);

    switch (transition) {
      case GST_STATE_CHANGE_READY_TO_PAUSED:
        result = GST_STATE_CHANGE_NO_PREROLL;
        break;
      default:
        break;
    }

    GstStateChangeReturn ret =
        GST_ELEMENT_CLASS(parent_class)->change_state(element, transition);
    if (ret == GST_STATE_CHANGE_FAILURE) {
        result = GST_STATE_CHANGE_FAILURE;
    }
    else {
        switch (transition) {
          case GST_STATE_CHANGE_READY_TO_PAUSED:
            // don't change 'result'
            break;
          case GST_STATE_CHANGE_PAUSED_TO_READY: {
            // Output all remaining lag sets. Flush queues in order from
            // oldest to newest activation times.
            REC_MUTEX_LOCK(sort->activations_mutex);
            void flush(gpointer data, gpointer user_data) {
                wcbe_sort_activation_t *act = data;
                flush_all_queues(sort, &(act->pqs));
            }
            list_reverse_foreach(g_list_last(sort->activations), flush, NULL);
            REC_MUTEX_UNLOCK(sort->activations_mutex);
            result = ret;
            break;
          }
          default:
            result = ret;
            break;
        }
    }

    return result;
}

static gint
cmp_timestamps(const lf_widar_time_t *ts0, const lf_widar_time_t *ts1,
               const lf_widar_time_t *t0, guint64 block)
{
    g_return_val_if_fail(ts0 != NULL, 0);
    g_return_val_if_fail(ts1 != NULL, 0);

    gint result;
    if (G_LIKELY(ts0->epoch == ts1->epoch)) {
        guint64 tb0 = TIMEVALUE(ts0);
        guint64 tb1 = TIMEVALUE(ts1);
        if (t0 != NULL) {
            tb0 -= TIMEVALUE(t0);
            tb1 -= TIMEVALUE(t0);
        }
        if (block > 1) {
            tb0 /= block;
            tb1 /= block;
        }
        if (tb0 < tb1) {
            result = -1;
        }
        else if (tb0 > tb1) {
            result = 1;
        }
        else {
            result = 0;
        }
    }
    else {
        result = ts0->epoch - ts1->epoch;
    }
    return result;
}

static GstFlowReturn
push_queue_element(WcbeSort *sort, output_queue_t *oq, queue_element_t *qe)
{
    g_return_val_if_fail(sort != NULL, GST_FLOW_ERROR);
    g_return_val_if_fail(qe != NULL, GST_FLOW_ERROR);

    GstFlowReturn result;
    const lf_widar_time_t *ts = queue_element_time(qe);
    if (TIMEVALUE(ts) > TIMEVALUE(&(oq->last_ts))) {
        oq->last_ts = *ts;
    }
    if (qe->unfilled_frames > 0) {
        g_debug("Pushing lagset element with %u unfilled frames (%s)",
                qe->unfilled_frames, g_quark_to_string(oq->product_id));
    }
    result = gst_pad_push(sort->srcpad, qe->buffer);
    if (result == GST_FLOW_OK) {
        sort->num_lagsets++;
    }
    else {
        GST_INFO("Failed to push buffer");
    }
    qe->buffer = NULL; // do this since gst_pad_push steals the ref
    if (qe->refcount != 1) {
        g_warning("queue element refcount != 1");
    }
    queue_element_unref(qe);
    return GST_FLOW_OK;
}

static void
queue_element_ref(queue_element_t *qe)
{
    g_return_if_fail(qe != NULL);
    g_atomic_int_inc((gint *)&(qe->refcount));
}

static void
queue_element_unref(queue_element_t *qe)
{
    g_return_if_fail(qe != NULL);
    if (g_atomic_int_dec_and_test((gint *)&(qe->refcount))) {
        if (qe->buffer != NULL) {
            gst_buffer_unref(qe->buffer);
        }
        g_slice_free(queue_element_t, qe);
    }
}

static gpointer
timeout_fun(gpointer data)
{
    WcbeSort *sort = data;
    gulong timeout = 0;

    while (g_atomic_pointer_get(&(sort->timeout_thread)) != NULL) {
        if (timeout > 0) {
            // sleep for one timeout interval
            g_usleep(timeout);
            REC_MUTEX_LOCK(sort->activations_mutex);
            // check for newly timed out activations
            GList *act = g_list_last(sort->activations);
            guint num_in_use = 0;
            while (act != NULL) {
                GList *prev_act = g_list_previous(act);
                wcbe_sort_activation_t *activation = act->data;
                if (activation->actvtn->in_use) {
                    ++num_in_use;
                    if (activation->pqs.queues != NULL) {
                        if (!activation->timed_out
                            && timeout_queues(sort, activation)) {
                            activation->timed_out = TRUE;
                            g_debug("activation for '%s' timed out",
                                    activation->actvtn->config->name);
                        }
                    }
                    else {
                        // handle activations without any queues
                        // specially: they time out when the following
                        // activation is in use
                        if (prev_act != NULL
                            && (((wcbe_sort_activation_t *)(prev_act->data))
                                ->actvtn->in_use)) {
                            activation->timed_out = TRUE;
                            g_debug("activation for '%s' timed out",
                                    activation->actvtn->config->name);
                        }
                    }
                }
                act = prev_act;
            }
            // remove timed out activations, but only if there is at
            // least one other activation in use
            act = g_list_last(sort->activations);
            while (num_in_use > 1 && act != NULL) {
                GList *prev_act = g_list_previous(act);
                wcbe_sort_activation_t *activation = act->data;
                if (activation->timed_out) {
                    g_assert(activation->actvtn->in_use);
                    --num_in_use;
                    remove_activation(sort, activation);
                }
                act = prev_act;
            }
            REC_MUTEX_UNLOCK(sort->activations_mutex);
        }
        MUTEX_LOCK(sort->changed_active_configs_mutex);

        // if timeout is zero, just wait until the active
        // configurations list has changed
        if (timeout == 0) {
            while (!sort->active_configs_changed
                   && (g_atomic_pointer_get(&(sort->timeout_thread))
                       != NULL)) {
                g_cond_wait(sort->changed_active_configs,
                            sort->changed_active_configs_mutex);
            }
        }

        // if active configurations list has changed, recompute the
        // timeout interval used by this function
        if (sort->active_configs_changed) {
            sort->active_configs_changed = FALSE;
            MUTEX_UNLOCK(sort->changed_active_configs_mutex);
            guint64 dump;
            REC_MUTEX_LOCK(sort->activations_mutex);
            if (sort->activations != NULL) {
                dump = set_timeouts(sort->activations);
            }
            else {
                dump = 0;
            }
            REC_MUTEX_UNLOCK(sort->activations_mutex);
            timeout = dump / (WIDAR_CLOCK_RATE / G_USEC_PER_SEC);
            if (timeout != 0) {
                timeout = CLAMP(timeout, 50000, 500000);
            }
            else {
                g_info("Unable to set non-zero timeout value "
                       "in dump event timeout generator");
            }
        }
        else {
            MUTEX_UNLOCK(sort->changed_active_configs_mutex);
        }
    }
    return NULL;
}

/**
 * timeout_queues
 *
 * Returns: %TRUE, iff all queues in the activation have timed out
 */
static gboolean
timeout_queues(WcbeSort *sort, wcbe_sort_activation_t *act)
{
    gboolean result = TRUE;

    GSList *oql = act->pqs.queues;
    while (oql != NULL) {
        output_queue_t *oq = oql->data;
        g_assert(oq->queue != NULL && oq->timer != NULL && oq->qmutex != NULL);
        if (oq->timeout > 0
            && g_timer_elapsed(oq->timer, NULL) >= oq->timeout) {
            g_timer_start(oq->timer);
            if (g_atomic_int_dec_and_test(&(oq->timeouts_remaining))) {
                MUTEX_LOCK(oq->qmutex);
                // Check "timeouts_remaining" once more in case something
                // entered the queue between our first test and mutex
                // acquisition.
                if (g_atomic_int_get(&(oq->timeouts_remaining)) == 0) {
                    queue_element_t *qe = g_queue_pop_tail(oq->queue);
                    if (qe != NULL) {
                        push_queue_element(sort, oq, qe);
                    }
                    g_atomic_int_compare_and_exchange(
                        &(oq->timeouts_remaining), 0, 1);
                }
                else {
                    result = FALSE;
                }
                MUTEX_UNLOCK(oq->qmutex);
            }
            else {
                result = FALSE;
            }
        }
        else {
            result = FALSE;
        }
        oql = g_slist_next(oql);
    }
    return result;
}

/** set_timeouts
 *
 * Return the minimum of the hw_dump values of all the queues. It
 * might make some sense to return the GCD of the hw_dump values to
 * minimize the rate at which the timeout_fun loop runs; however,
 * since that loop imposes both upper and lower limits on its rate, I
 * have instead opted for this simpler solution.
 */
static guint64
set_timeouts(GList *activations)
{
    guint64 min = G_MAXUINT64;
    gboolean have_zero = FALSE;
    void foreach_queue(gpointer data, gpointer reserved) {
        output_queue_t *oq = data;
        // Ignore zero hw_dump values for now, but note their
        // presence.
        if (oq->hw_dump > 0) {
            if (oq->hw_dump < min) {
                min = oq->hw_dump;
            }
        }
        else {
            have_zero = TRUE;
        }
        if (oq->timeout == 0) {
            oq->timeout = oq->hw_dump / (double)WIDAR_CLOCK_RATE;
            g_atomic_int_set(&(oq->timeouts_remaining), MAX_TIMEOUTS);
            g_timer_start(oq->timer);
        }
    }
    void foreach_activation(gpointer data, gpointer reserved) {
        wcbe_sort_activation_t *act = data;
        g_slist_foreach(act->pqs.queues, foreach_queue, NULL);
    }
    g_list_foreach(activations, foreach_activation, NULL);

    // If all hw_dump values are zero, min will have its initial
    // value, so zero is the correct operational min value.
    if (min == G_MAXUINT64 && have_zero) {
        min = 0;
    }
    return min;
}

static guint
frame_hash(gconstpointer key)
{
    const lf_lag_frame_t *lf = key;
    const lf_data_frame_word_t *mask;

    if (G_LIKELY(LAG_FRAME_KEY_FTYPE(lf) == LF_LTA_DATA_FRAME_FTYPE)) {
        mask = lta_hash_mask;
    }
    else {
        mask = cc_hash_mask;
    }
    return ((lf->key[0] & mask[0]) ^ (lf->key[1] & mask[1])
            ^ (lf->key[2] & mask[2]) ^ (lf->key[3] & mask[3]));
}

static gboolean
frame_equal(gconstpointer a, gconstpointer b)
{
    const lf_lag_frame_t *lfa = a;
    const lf_lag_frame_t *lfb = b;
    gboolean result;
    if (G_LIKELY(LAG_FRAME_KEY_FTYPE(lfa) == LAG_FRAME_KEY_FTYPE(lfb))) {
        const lf_data_frame_word_t *mask;
        if (G_LIKELY(LAG_FRAME_KEY_FTYPE(lfa) == LF_LTA_DATA_FRAME_FTYPE)) {
            mask = lta_hash_mask;
        }
        else {
            mask = cc_hash_mask;
        }
        result = ((((lfa->key[0] & mask[0]) == (lfb->key[0] & mask[0]))
                   && ((lfa->key[1] & mask[1]) == (lfb->key[1] & mask[1]))
                   && ((lfa->key[2] & mask[2]) == (lfb->key[2] & mask[2]))
                   && ((lfa->key[3] & mask[3]) == (lfb->key[3] & mask[3])))
                  ? TRUE : FALSE);
    }
    else {
        result = FALSE;
    }
    return result;
}

static void
destroy_config(WcbeSort *sort, wcbe_sort_config_t *cfg)
{
    gchar *name = g_strdup(cfg->wcfg->name);
    destroy_config_silently(sort, cfg);
    g_signal_emit(cfg->sort, wcbe_sort_signals[SIGNAL_DESTROY_CFG],
                  0, name, TRUE);
    g_free(name);
}

static void
destroy_config_silently(WcbeSort *sort, wcbe_sort_config_t *cfg)
{
    REC_MUTEX_LOCK(sort->activations_mutex);
    GList *act = g_list_last(sort->activations);
    while (act != NULL) {
        GList *prev = g_list_previous(act);
        wcbe_sort_activation_t *a = act->data;
        if (a->config == cfg) {
            remove_activation(sort, a);
        }
        act = prev;
    }
    REC_MUTEX_UNLOCK(sort->activations_mutex);
    free_product_queues(&(cfg->pqs_template));
    g_slice_free(wcbe_sort_config_t, cfg);
}

static wcbe_sort_config_t *
create_config(WcbeSort *sort, const wcbe_config_t *config)
{
    wcbe_sort_config_t *result = g_slice_new0(wcbe_sort_config_t);
    result->wcfg = config;
    result->config_q = g_quark_from_string(config->name);
    GPtrArray *products = config->input.products;
    if (products != NULL) {
        // Note: don't need to do anything when removing hash
        // table values except to free the frame_dest_t
        // memory; freeing the output_queue_t element within
        // the frame_dest_t is handled elsewhere (config
        // 'queues').
        result->pqs_template.destinations = g_hash_table_new_full(
            frame_hash, frame_equal, free_lf_lag_frame, free_frame_dest);
        guint i;
        for (i = 0; i < products->len; ++i) {
            wcbe_product_config_t *product = g_ptr_array_index(products, i);
            product_config(&(result->pqs_template), product, config);
        }
    }
    return result;
}

static void
product_config(product_queues_t *pq, const wcbe_product_config_t *product,
               const wcbe_config_t *config)
{
    if (product->streamA == NULL || product->streamB == NULL) {
        g_message("Product '%s' has unresolved stream ids: not configured",
                  product->id);
	return;
    }

    output_queue_t *oq = g_slice_new0(output_queue_t);
    // oq->queue = NULL;
    // oq->qmutex = NULL;
    // oq->timer = NULL;
    pq->queues = g_slist_prepend(pq->queues, oq);
    oq->last_ts = (lf_widar_time_t){0, 0, 0};
    oq->product_id = g_quark_from_string(product->id);
    oq->timeouts_remaining = MAX_TIMEOUTS + 1;
    oq->num_bins = product->num_bins;
    oq->num_lags_per_bin = product->num_lags;
    oq->sbIdx = product->subband_index;

    guint lags_per_subprod_per_frame;
    if (oq->num_lags_per_bin != MAX_LAGS_PER_FRAME / 2) {
        lags_per_subprod_per_frame = MAX_LAGS_PER_FRAME;
    }
    else {
        lags_per_subprod_per_frame = MAX_LAGS_PER_FRAME / 2;
    }
    oq->num_frames_per_bin = oq->num_lags_per_bin / lags_per_subprod_per_frame;

    // find dump time of frames from correlator for this product
    guint64 hw_dump = (((guint64)product->integration.min_hw
                        * product->integration.hw
                        * product->integration.lta * WIDAR_CLOCK_RATE)
                       / G_USEC_PER_SEC);
    g_debug("Product '%s' dump period: %"G_GUINT64_FORMAT,
            product->id, hw_dump);

    // gather all lag chains into a single array
    GPtrArray *lag_chains;
    // 'product4' and 'product7' elements need to be handled slight
    // differently because the 'subproductXX' children of 'product7'
    // contain ordering information.
    oq->nbits = product->nbits;
    if (oq->nbits == WCBE_4BITS) {
        wcbe_product4_config_t *p4 = (wcbe_product4_config_t *)product;
        lag_chains = g_ptr_array_sized_new(p4->lag_chains->len);
        guint i;
        for (i = 0; i < p4->lag_chains->len; ++i) {
            g_ptr_array_add(lag_chains,
                            &g_array_index(p4->lag_chains,
                                           wcbe_lag_chain_config_t, i));
        }
    }
    else /* oq->nbits == WCBE_7BITS */ {
        wcbe_product7_config_t *p7 = (wcbe_product7_config_t *)product;
        lag_chains = g_ptr_array_sized_new(p7->lag_chains_ll->len
            + p7->lag_chains_lm->len + p7->lag_chains_ml->len
            + p7->lag_chains_mm->len);
        guint i;
        for (i = 0; i < p7->lag_chains_ll->len; ++i) {
            g_ptr_array_add(lag_chains,
                            &g_array_index(p7->lag_chains_ll,
                                           wcbe_lag_chain_config_t, i));
        }
        for (i = 0; i < p7->lag_chains_lm->len; ++i) {
            g_ptr_array_add(lag_chains,
                            &g_array_index(p7->lag_chains_lm,
                                           wcbe_lag_chain_config_t, i));
        }
        for (i = 0; i < p7->lag_chains_ml->len; ++i) {
            g_ptr_array_add(lag_chains,
                            &g_array_index(p7->lag_chains_ml,
                                          wcbe_lag_chain_config_t, i));
        }
        for (i = 0; i < p7->lag_chains_mm->len; ++i) {
            g_ptr_array_add(lag_chains,
                            &g_array_index(p7->lag_chains_mm,
                                           wcbe_lag_chain_config_t, i));
        }
    }
    g_ptr_array_sort(lag_chains, cmp_segments);

    // Get total number of recirculation blocks
    guint lc_recirc = 0;
    guint total_recirc = 0;
    guint i;
    for (i = 0; i < lag_chains->len; ++i) {
        wcbe_lag_chain_config_t *lc_segment =
            g_ptr_array_index(lag_chains, i);
        total_recirc += lc_segment->recirc;
        // The following checks for an unlikely possibility. I think
        // we might be able work around this restriction, but it seems
        // unlikely to be necessary.
        if (i > 0 && lc_segment->recirc != lc_recirc) {
            g_warning("'lagChainSegment's with differing 'rec' "
                      "attributes within one product are unsupported");
        }
        lc_recirc = lc_segment->recirc;
    }
    // TODO: The following isn't correct in the case of burst dumping
    // from the correlator chips. In that case, the rate at which
    // frames arrive at the backend depends on the rate at which they
    // can be transmitted over the network, not the (short-term) rate
    // at which they are created. This gets difficult since the rate
    // at which frames from the given product arrive at the backend
    // depends on the total data bandwidth used by the baseline board
    // as a whole.
    oq->hw_dump = oq->num_bins * lc_recirc * hw_dump;

    // Create mappings for frames in this product. This is the heart
    // of the algorithm for ordering the lags as they arrive. Note
    // that the mapping ignores the bin index; that is taken care of
    // in the wcbe_sort_chain function to account for the case in
    // which no binning occurs and to support mapping of LTA bins to
    // logical phase bins. OTOH, ordering of 7-bit product components,
    // recirculation and lag chain segments are all handled by this
    // mapping.
    gboolean oq_ready = FALSE;
    guint header_idx = 0;
    guint lags_idx = 0;
#define STREAM_A 0
#define STREAM_B 1
    guint recirc[] = {[STREAM_B] = 0, [STREAM_A] = total_recirc - 1};
    guint recirc_x_stream, recirc_y_stream;
    guint lc;
    for (lc = 0; lc < lag_chains->len; ++lc) {
        wcbe_lag_chain_config_t *lc_segment =
            g_ptr_array_index(lag_chains, lc);
        guint nblocks = lc_segment->recirc;
        lf_data_frame_word_t chip_id =
            ((lc_segment->ccY << 3) | lc_segment->ccX);
        gboolean y_is_b;
        wcbe_stream_config_t *streamX, *streamY;
        if (lc_segment->inpX.stream == 'A') {
            g_warn_if_fail(lc_segment->inpY.stream == 'B');
            recirc_x_stream = STREAM_A;
            streamX = product->streamA;
            recirc_y_stream = STREAM_B;
            streamY = product->streamB;
            y_is_b = TRUE;
        }
        else /* lc_segment->inpX.stream == 'B' */ {
            g_warn_if_fail(lc_segment->inpX.stream == 'B');
            g_warn_if_fail(lc_segment->inpY.stream == 'A');
            recirc_x_stream = STREAM_B;
            streamX = product->streamB;
            recirc_y_stream = STREAM_A;
            streamY = product->streamA;
            y_is_b = FALSE;
        }
        if (!oq_ready) {
            oq->stnA = streamX->station_id;
            oq->stnB = streamY->station_id;
            oq->bbA = streamX->baseband_id;
            oq->bbB = streamY->baseband_id;
            oq_ready = TRUE;
        }
        guint start_y = recirc[recirc_y_stream];
        guint blk;
        for (blk = 0; blk < nblocks; ++blk) {
            gint ccc = (y_is_b ? lc_segment->cccN : lc_segment->ccc0);
            while ((gint)lc_segment->ccc0 <= ccc
                   && ccc <= (gint)lc_segment->cccN) {
                frame_dest_t *dest = g_slice_new(frame_dest_t);
                dest->oqueue = oq;
                dest->header_idx = header_idx;
                dest->lags_idx = lags_idx;
                dest->reverse_conjugate = !y_is_b;
                header_idx++;
                lags_idx += lags_per_subprod_per_frame;
                lf_lag_frame_t *lf = g_slice_new(lf_lag_frame_t);
                lf_lta_key_value(
                    start_y, nblocks,
                    chip_id,
                    ccc,
                    recirc[recirc_x_stream], recirc[recirc_y_stream],
                    streamX->baseband_id, streamY->baseband_id,
                    streamX->subband_id, streamY->subband_id,
                    streamX->station_id, streamY->station_id,
                    lc_segment->board_id,
                    0,    // lta_bin
                    lf->key);
                g_hash_table_insert(pq->destinations, lf, dest);
                if (y_is_b) {
                    --ccc;
                }
                else {
                    ++ccc;
                }
            }
            g_debug("cid: %u, nblk: %u, sty: %u, rx: %u, ry: %u, ra: %u, "
                    "rb: %u, conj: %d",
                    chip_id, nblocks, start_y, recirc[recirc_x_stream],
                    recirc[recirc_y_stream], recirc[STREAM_A],
                    recirc[STREAM_B], !y_is_b);
            --recirc[STREAM_A];
            ++recirc[STREAM_B];
        }
    }
#undef STREAM_A
#undef STREAM_B
    guint corr_factor = ((oq->nbits == WCBE_4BITS) ? 1 : 4);
    if (lags_idx != oq->num_lags_per_bin * corr_factor) {
        g_warning("Number of lags computed while ordering lag frames in "
                  "product '%s' (%u) disagrees with number given by "
                  "configuration (%u)", product->id, lags_idx,
                  oq->num_lags_per_bin * corr_factor);
    }
    g_ptr_array_free(lag_chains, TRUE);
}

static gint
cmp_segments(gconstpointer a, gconstpointer b)
{
    const wcbe_lag_chain_config_t * const *lca = a;
    const wcbe_lag_chain_config_t * const *lcb = b;
    gint result;
    result = (*lcb)->subprod - (*lca)->subprod; // note reverse order
    if (result == 0) {
#define BLOCK(lc) ((gint)((*(lc))->inpX.stream == 'B' ? \
                          (*(lc))->inpX.lag_block : ((*lc))->inpY.lag_block))
        result = BLOCK(lca) - BLOCK(lcb);
#undef BLOCK
    }
    return result;
}

static void
flush_all_queues(WcbeSort *sort, product_queues_t *pq)
{
    g_return_if_fail(pq->queues == NULL
                     || ((output_queue_t *)(pq->queues->data))->queue != NULL);

    void flush_queue(gpointer data, gpointer user_data) {
        g_return_if_fail(data != NULL);
        output_queue_t *oq = data;
        MUTEX_LOCK(oq->qmutex);
        queue_element_t *qe = g_queue_pop_tail(oq->queue);
        while (qe != NULL) {
            push_queue_element(sort, oq, qe);
            qe = g_queue_pop_tail(oq->queue);
        }
        MUTEX_UNLOCK(oq->qmutex);
    }
    g_slist_foreach(pq->queues, flush_queue, NULL);
}

static gboolean
wcbe_sort_create_config(WcbeSort *sort, const wcbe_config_t *config,
                        GstPipeline *pipeline G_GNUC_UNUSED)
{
    g_return_val_if_fail(sort != NULL, FALSE);
    g_return_val_if_fail(config != NULL, FALSE);

    gboolean result = FALSE;
    wcbe_sort_config_t *cfg = create_config(sort, config);
    if (cfg != NULL) {
        cfg->sort = sort;
        WcbeSortClass *klass = WCBE_SORT_CLASS(GST_ELEMENT_GET_CLASS(sort));
        g_dataset_id_set_data(config, klass->cfg_data, cfg);
        result = TRUE;
    }
    else {
        g_info("Configuration '%s' failed", config->name);
    }
    return result;
}

static void
wcbe_sort_destroy_config(WcbeSort *sort, const wcbe_config_t *config,
                         GstPipeline *pipeline G_GNUC_UNUSED)
{
    wcbe_sort_config_t *cfg = get_cfg_data(sort, config);
    destroy_config(sort, cfg);
    WcbeSortClass *klass = WCBE_SORT_CLASS(GST_ELEMENT_GET_CLASS(sort));
    g_dataset_id_remove_data(config, klass->cfg_data);
}

static void
free_lf_lag_frame(gpointer data)
{
    g_slice_free(lf_lag_frame_t, data);
}

static void
free_frame_dest(gpointer data)
{
    g_slice_free(frame_dest_t, data);
}

static gboolean
wcbe_sort_add_activation(WcbeSort *sort, activation_t *activation,
                         GstPipeline *pipeline G_GNUC_UNUSED)
{
    gboolean result = TRUE;

    wcbe_sort_activation_t *act = g_slice_new0(wcbe_sort_activation_t);
    act->actvtn = activation;

    // compute time for a "now" activation
    REC_MUTEX_LOCK(sort->activations_mutex);
    if (activation->now) {
        GList *last = g_list_last(sort->activations);
        wcbe_sort_activation_t *earliest =
            ((last != NULL) ? (wcbe_sort_activation_t *)(last->data) : NULL);
        while (last != NULL
               && (!earliest->actvtn->in_use
                   || earliest->actvtn->deactivated)) {
            last = g_list_previous(last);
            earliest = ((last != NULL)
                        ? (wcbe_sort_activation_t *)(last->data) : NULL);
        }
        if (earliest == NULL) {
            activation->time.epoch = 0;
            activation->time.seconds = 0;
            activation->time.clock_cycles = 0;
        }
        else {
            activation->time.epoch = earliest->actvtn->time.epoch;
            activation->time.seconds = earliest->actvtn->time.seconds + 1;
            activation->time.clock_cycles = 0;
        }
        g_debug("set now activation time to %u:%u:%u", activation->time.epoch,
                activation->time.seconds, activation->time.clock_cycles);
    }

    // check that this isn't a duplicate
    gint cmp_times(gconstpointer a, gconstpointer b) {
        const wcbe_sort_activation_t *act_a = a;
        const wcbe_sort_activation_t *act_b = b;
        // order here corresponds to greatest to least timestamp
        return cmp_timestamps(
            &(act_b->actvtn->time), &(act_a->actvtn->time), NULL, 0);
    }
    if (g_list_find_custom(sort->activations, act, cmp_times) != NULL) {
        result = FALSE;
    }
    REC_MUTEX_UNLOCK(sort->activations_mutex);
    if (!result) {
        g_debug("rejected duplicate activation time");
        g_slice_free(wcbe_sort_activation_t, act);
        return FALSE;
    }

    // create wcbe_sort_activation_t instance for this activation
    act->sort = sort;
    act->config = get_cfg_data(sort, activation->config);
    act->timed_out = FALSE;
    act->pqs.destinations = g_hash_table_new_full(frame_hash, frame_equal,
                                                  free_lf_lag_frame,
                                                  free_frame_dest);
    act->pqs.queues = NULL;

    // Copy 'destinations' hash table from template in associated
    // wcbe_sort_config_t instance. Note that we cannot simply copy
    // output_queue_t instances, since they may be shared by several
    // frame destinations; instead we create a mapping from
    // output_queue_t instances in the template to newly created
    // output_queue_t instances, and apply that to the copied
    // destinations.
    GHashTable *oqs = g_hash_table_new(NULL, NULL);
    void cpydest(gpointer key, gpointer value, gpointer user_data) {
        lf_lag_frame_t *lf = key;
        frame_dest_t *dest = value;
        lf_lag_frame_t *new_key = g_slice_dup(lf_lag_frame_t, lf);
        frame_dest_t *new_dest= g_slice_dup(frame_dest_t, dest);
        output_queue_t *oq = g_hash_table_lookup(oqs, dest->oqueue);
        if (oq == NULL) {
            oq = g_slice_dup(output_queue_t, dest->oqueue);
            oq->queue = g_queue_new();
            oq->qmutex = g_mutex_new();
            oq->timer = g_timer_new();
            oq->last_ts = activation->time;
            g_hash_table_insert(oqs, dest->oqueue, oq);
            act->pqs.queues = g_slist_prepend(act->pqs.queues, oq);
            // TODO: check that there is no memory leak or double free
            // of oq
        }
        new_dest->oqueue = oq;
        g_hash_table_insert(act->pqs.destinations, new_key, new_dest);
    }
    if (act->config->pqs_template.destinations != NULL) {
        g_hash_table_foreach(
            act->config->pqs_template.destinations, cpydest, NULL);
    }
    g_hash_table_destroy(oqs);

    // insert wcbe_sort_activation_t instance into list of activations
    // (in order from greatest to least activation times)
    REC_MUTEX_LOCK(sort->activations_mutex);
    g_debug("adding activation, time %u %u %u", act->actvtn->time.epoch,
            act->actvtn->time.seconds, act->actvtn->time.clock_cycles);
    sort->activations =
        g_list_insert_sorted(sort->activations, act, cmp_times);
    REC_MUTEX_UNLOCK(sort->activations_mutex);

    return TRUE;
}

static void
free_product_queues(product_queues_t *pq)
{
    while (pq->queues != NULL) {
        void free_elt(gpointer data, gpointer user_data) {
            queue_element_t *qe = data;
            queue_element_unref(qe);
        }
        output_queue_t *oq = pq->queues->data;
        pq->queues = g_slist_remove(pq->queues, oq);
        if (oq->queue != NULL) {
            g_assert(oq->qmutex != NULL);
            g_assert(oq->timer != NULL);
            g_queue_foreach(oq->queue, free_elt, NULL);
            g_queue_free(oq->queue);
            g_mutex_free(oq->qmutex);
            g_timer_destroy(oq->timer);
        }
        else {
            g_assert(oq->qmutex == NULL);
            g_assert(oq->timer == NULL);
        }
        g_slice_free(output_queue_t, oq);
    }
    if (pq->destinations != NULL) {
        g_hash_table_destroy(pq->destinations);
    }
}

static void
wcbe_sort_remove_activation(WcbeSort *sort, activation_t *activation,
                            GstPipeline *pipeline G_GNUC_UNUSED)
{
    wcbe_sort_activation_t act = {.actvtn = activation};
    gint cmp(gconstpointer a, gconstpointer b) {
        const wcbe_sort_activation_t *act_a = a;
        const wcbe_sort_activation_t *act_b = b;
        return ((act_a->actvtn == act_b->actvtn) ? 0 : 1);
    }
    REC_MUTEX_LOCK(sort->activations_mutex);
    GList *actl = g_list_find_custom(sort->activations, &act, cmp);
    wcbe_sort_activation_t *old_act = (wcbe_sort_activation_t *)(actl->data);
    remove_activation(sort, old_act);
    REC_MUTEX_UNLOCK(sort->activations_mutex);
}

static void
remove_activation(WcbeSort *sort, wcbe_sort_activation_t *activation)
{
    gboolean config_in_use = FALSE;
    g_debug("remove_activation for '%s'", activation->actvtn->config->name);
    REC_MUTEX_LOCK(sort->activations_mutex);
    // removing the activation from the list is easy
    sort->activations = g_list_remove(sort->activations, activation);
    // now we need to find whether the associated configuration is no
    // longer active
    GList *act = sort->activations;
    while (act != NULL && !config_in_use) {
        wcbe_sort_activation_t *a = act->data;
        config_in_use = ((a->config->config_q == activation->config->config_q)
                         && a->actvtn->in_use);
        act = g_list_next(act);
    }
    if (!config_in_use) {
        // remove the configuration from the active_configs array
        guint i;
        for (i = 0; i < sort->active_configs->len; ++i) {
            if (activation->config->config_q ==
                g_array_index(sort->active_configs, GQuark, i)) {
                g_array_remove_index_fast(sort->active_configs, i);
                break;
            }
        }
        REC_MUTEX_UNLOCK(sort->activations_mutex);
        // signal change in active configurations
        MUTEX_LOCK(sort->changed_active_configs_mutex);
        sort->active_configs_changed = TRUE;
        g_cond_signal(sort->changed_active_configs);
        MUTEX_UNLOCK(sort->changed_active_configs_mutex);
        flush_all_queues(sort, &(activation->pqs));
        // inject WCBE_EVENT_DEACTIVATE_CONFIG event into pipeline
        GstEvent *deactivate_event = gst_event_new_custom(
            WCBE_EVENT_DEACTIVATE_CONFIG,
            gst_structure_new("deactivate-config",
                              "configuration", G_TYPE_POINTER,
                              activation->config->wcfg,
                              NULL));
        if (!gst_element_send_event(GST_ELEMENT(sort), deactivate_event)) {
            GST_INFO("Failed to send 'deactivate-config' event");
        }
    }
    else {
        REC_MUTEX_UNLOCK(sort->activations_mutex);
    }

    free_product_queues(&(activation->pqs));
    signal_deactivation(sort, activation->actvtn);
    g_slice_free(wcbe_sort_activation_t, activation);
}

static void
signal_deactivation(WcbeSort *sort, const activation_t *activation)
{
    g_signal_emit(sort, wcbe_sort_signals[SIGNAL_DEACTIVATE_CFG],
                  0, activation);
}

static wcbe_sort_activation_t *
find_activation(WcbeSort *sort, const lf_lag_frame_t *lf)
{
    wcbe_sort_activation_t *result = NULL;
    gboolean new_configs = FALSE;

    gboolean is_active_config(wcbe_sort_activation_t *a) {
        // check whether the associated configuration is in the
        // active_configs array
        gboolean result = FALSE;
        guint i;
        for (i = 0; !result && i < sort->active_configs->len; ++i) {
            result = (a->config->config_q
                      == g_array_index(sort->active_configs, GQuark, i));
        }
        return result;
    }
    REC_MUTEX_LOCK(sort->activations_mutex);
    // find the latest activation that precedes the lag frame time stamp
    GList *act = sort->activations;
    while (act != NULL
           && (cmp_timestamps(&(lf->timestamp),
                              &(((wcbe_sort_activation_t *)(act->data))
                                ->actvtn->time), NULL, 0)
               <= 0)) {
        act = g_list_next(act);
    }
    if (act != NULL) {
        result = (wcbe_sort_activation_t *)(act->data);
        // mark this and all prior activations as in use
        while (act != NULL) {
            wcbe_sort_activation_t *activation = act->data;
            if (!activation->actvtn->in_use) {
                g_debug("find_activation mark in_use '%s'",
                        activation->actvtn->config->name);
                activation->actvtn->in_use = TRUE;
                if (!is_active_config(activation)) {
                    // add associated configuration to the
                    // active_configs array
                    new_configs = TRUE;
                    g_array_append_val(sort->active_configs,
                                       activation->config->config_q);
                    // inject WCBE_EVENT_ACTIVATE_CONFIG event into pipeline
                    g_signal_emit(sort, wcbe_sort_signals[SIGNAL_ACTIVATE_CFG],
                                  0, activation->config->wcfg->name, TRUE);
                    GstEvent *activate_event = gst_event_new_custom(
                        WCBE_EVENT_ACTIVATE_CONFIG,
                        gst_structure_new(
                            "activate-config",
                            "configuration", G_TYPE_POINTER,
                            activation->config->wcfg,
                            "time0", G_TYPE_UINT64,
                            TIMEVALUE(&(activation->actvtn->time)),
                            NULL));
                    if (!gst_element_send_event(GST_ELEMENT(sort),
                                                activate_event)) {
                        GST_INFO("Failed to send 'activate-config' event");
                    }
                }
            }
            act = g_list_next(act);
        }
    }
    REC_MUTEX_UNLOCK(sort->activations_mutex);
    if (new_configs) {
        // notify a change of active configuration
        MUTEX_LOCK(sort->changed_active_configs_mutex);
        sort->active_configs_changed = TRUE;
        g_cond_signal(sort->changed_active_configs);
        MUTEX_UNLOCK(sort->changed_active_configs_mutex);
    }
    return result;
}

static void
list_reverse_foreach(GList *list, GFunc func, gpointer user_data)
{
    while (list != NULL) {
        GList *prev = g_list_previous(list);
        (*func)(list->data, user_data);
        list = prev;
    }
}

static void
parent_set_cb(GstObject *gstobject, GstObject *parent,
              gpointer user_data G_GNUC_UNUSED)
{
    if (object_has_signal(G_OBJECT(parent), "create-configuration")) {
        g_signal_connect_swapped(parent, "create-configuration",
                                 G_CALLBACK(wcbe_sort_create_config),
                                 gstobject);
    }
    if (object_has_signal(G_OBJECT(parent), "destroy-configuration")) {
        g_signal_connect_swapped(parent, "destroy-configuration",
                                 G_CALLBACK(wcbe_sort_destroy_config),
                                 gstobject);
    }
    if (object_has_signal(G_OBJECT(parent), "add-activation")) {
        g_signal_connect_swapped(parent, "add-activation",
                                 G_CALLBACK(wcbe_sort_add_activation),
                                 gstobject);
    }
    if (object_has_signal(G_OBJECT(parent), "remove-activation")) {
        g_signal_connect_swapped(parent, "remove-activation",
                                 G_CALLBACK(wcbe_sort_remove_activation),
                                 gstobject);
    }
}
