#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netdb.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>

#include <wcbe_src.h>
#include <wcbe.h>
#include <wcbe_lag_frame.h>
#include <wcbe_util.h>

typedef enum {FRAME_MSG, EOF_MSG, ERROR_MSG, UNLOCK_MSG} queue_msg_type_t;
typedef struct queue_msg_t {
    queue_msg_type_t type;
    GstBuffer* frame_buffer;
} queue_msg_t;

// The control socket stuff initially was copied from GStreamer's
// 'gstfdsrc' module.
//
//  * Copyright (C) 1999,2000 Erik Walthinsen <omega@cse.ogi.edu>
//  *                    2000 Wim Taymans <wtay@chello.be>
//  *                    2005 Philippe Khalaf <burger@speedy.org>
//
// Licensed under the LGPL v2.

/* the select call is also performed on the control sockets, that way
 * we can send special commands to unblock the select call */
#define CONTROL_UNLOCK            'S'     /* stop the select call */
#define CONTROL_SOCKETS(src)   src->control_sock
#define WRITE_SOCKET(src)      src->control_sock[1]
#define READ_SOCKET(src)       src->control_sock[0]

#define SEND_COMMAND(src, command)          \
G_STMT_START {                              \
  unsigned char c; c = command;             \
  write (WRITE_SOCKET(src), &c, 1);         \
} G_STMT_END

#define READ_COMMAND(src, command, res)        \
G_STMT_START {                                 \
  res = read(READ_SOCKET(src), &command, 1);   \
} G_STMT_END

GST_DEBUG_CATEGORY_STATIC(wcbe_src_debug);
#define GST_CAT_DEFAULT wcbe_src_debug

/* Filter signals and args */
enum
{
    /* FILL ME */
    LAST_SIGNAL
};

enum
{
    ARG_0,
    ARG_URI,
    ARG_NUM_FRAMES,
    ARG_FPS,
    ARG_TIMESTAMP_BASE,
    ARG_FRAME_RATE,
    ARG_NUM_DROPPED_FRAMES
};

#define DEFAULT_FPS 100
#define EWMA_DECAY (4.0 / 5.0)
#define MAX_QUEUE_SIZE 10000

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE(
    "src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS(WCBE_WIDAR_FRAME_MEDIA_TYPE));

static void wcbe_src_uri_handler_init(gpointer g_iface, gpointer iface_data);

static void
_do_init(GType wcbe_src_type)
{
    static const GInterfaceInfo urihandler_info = {
        wcbe_src_uri_handler_init,
        NULL,
        NULL
    };
    g_type_add_interface_static(wcbe_src_type, GST_TYPE_URI_HANDLER,
                                &urihandler_info);
    GST_DEBUG_CATEGORY_INIT(wcbe_src_debug, "wcbe_src_debug", 0,
                            "wcbe_src debug");
}

GST_BOILERPLATE_FULL(WcbeSrc, wcbe_src, GstPushSrc, GST_TYPE_PUSH_SRC,
                     _do_init);

static void wcbe_src_finalize(GObject *object);
static void wcbe_src_set_property(GObject *object, guint prop_id,
                                  const GValue *value, GParamSpec *pspec);
static void wcbe_src_get_property(GObject *object, guint prop_id,
                                  GValue *value, GParamSpec *pspec);

static gboolean wcbe_src_start(GstBaseSrc *src);
static gboolean wcbe_src_stop(GstBaseSrc *src);
static void wcbe_src_get_times(GstBaseSrc *src, GstBuffer *buffer,
                               GstClockTime *start, GstClockTime *end);
static GstFlowReturn wcbe_src_create(GstPushSrc *src, GstBuffer **buf);
static gboolean wcbe_src_unlock(GstBaseSrc *src);
static gboolean wcbe_src_unlock_stop(GstBaseSrc *src);

static GstURIType wcbe_src_uri_get_type(void);
static gchar **wcbe_src_uri_get_protocols(void);
static const gchar *wcbe_src_uri_get_uri(GstURIHandler *handler);
static gboolean wcbe_src_uri_set_uri(GstURIHandler *handler, const gchar *uri);
static gpointer rate_fun(gpointer data);
static gpointer msg_fun(gpointer data);
static void queue_file_frames(WcbeSrc *src);
static void queue_widar_frames(WcbeSrc *src);
static void free_queue_msg(gpointer data);
static queue_msg_type_t source_input(WcbeSrc *src);

static void
wcbe_src_base_init(gpointer gclass)
{
    static GstElementDetails element_details = {
        "WIDAR frame source",
        "Source",
        "WIDAR frame source",
        "Martin Pokorny <mpokorny@nrao.edu>"
    };
    GstElementClass *element_class = GST_ELEMENT_CLASS(gclass);

    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get (&src_factory));
    gst_element_class_set_details(element_class, &element_details);
}

/* initialize the plugin's class */
static void
wcbe_src_class_init(WcbeSrcClass * klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    GstBaseSrcClass *gstbasesrc_class = GST_BASE_SRC_CLASS(klass);
    GstPushSrcClass *gstpushsrc_class = GST_PUSH_SRC_CLASS(klass);

    gobject_class->set_property =
        GST_DEBUG_FUNCPTR(wcbe_src_set_property);
    gobject_class->get_property =
        GST_DEBUG_FUNCPTR(wcbe_src_get_property);
    gobject_class->finalize =
        GST_DEBUG_FUNCPTR(wcbe_src_finalize);

    g_object_class_install_property(
        gobject_class, ARG_TIMESTAMP_BASE,
        g_param_spec_string("timestamp-base", "Regenerated timestamp base",
                            "Base time for regenerated file-based "
                            "lag frame timestamps",
                            NULL, G_PARAM_READWRITE));

    g_object_class_install_property(
        gobject_class, ARG_URI,
        g_param_spec_string("uri", "WIDAR frame source URI",
                            "Set frame source",
                            NULL, G_PARAM_READWRITE));

    g_object_class_install_property(
        gobject_class, ARG_NUM_FRAMES,
        g_param_spec_uint("num-frames",
                          "Number of WIDAR data frames sent",
                          "Get number of frames sent",
                          0, G_MAXUINT, 0, G_PARAM_READABLE));

    g_object_class_install_property(
        gobject_class, ARG_FPS,
        g_param_spec_uint("fps",
                          "Playback frames per second",
                          "Set playback rate of file input",
                          0, G_MAXUINT,
                          DEFAULT_FPS, G_PARAM_READWRITE));

    g_object_class_install_property(
        gobject_class, ARG_FRAME_RATE,
        g_param_spec_float("actual-rate",
                           "Measured frames per second",
                           "Rate of frames sent into pipeline, in frames "
                           "per second",
                           0, G_MAXFLOAT,
                           0, G_PARAM_READABLE));

    g_object_class_install_property(
        gobject_class, ARG_NUM_DROPPED_FRAMES,
        g_param_spec_uint("dropped-frames",
                          "Number of dropped WIDAR data frames",
                          "Get number of frames dropped internally",
                          0, G_MAXUINT, 0, G_PARAM_READABLE));

    gstbasesrc_class->start = GST_DEBUG_FUNCPTR(wcbe_src_start);
    gstbasesrc_class->stop = GST_DEBUG_FUNCPTR(wcbe_src_stop);
    gstbasesrc_class->get_times = GST_DEBUG_FUNCPTR(wcbe_src_get_times);
    gstbasesrc_class->unlock = GST_DEBUG_FUNCPTR(wcbe_src_unlock);
    gstbasesrc_class->unlock_stop = GST_DEBUG_FUNCPTR(wcbe_src_unlock_stop);
    gstpushsrc_class->create = GST_DEBUG_FUNCPTR(wcbe_src_create);
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
wcbe_src_init(WcbeSrc *src, WcbeSrcClass *gclass)
{
    GstBaseSrc *gstbasesrc = GST_BASE_SRC(src);

    gst_base_src_set_format(gstbasesrc, GST_FORMAT_BUFFERS);
    src->uri = NULL;
    src->fps = DEFAULT_FPS;
    src->channel = NULL;
    src->ts0 = NULL;
    src->ts0_epoch = 0;
    src->ts0_seconds = 0;
    src->ts0_cycles = 0;
    src->smoothed_rate = 0.0f;
    g_static_mutex_init(&(src->ts0_mutex));
    g_static_mutex_init(&(src->rate_mutex));
    GError *err = NULL;
    src->rate_thread = g_thread_create(rate_fun, src, TRUE, &err);
    if (src->rate_thread == NULL) {
        g_critical("Failed to spawn rate thread: %s", err->message);
    }
    READ_SOCKET(src) = 0;
    WRITE_SOCKET(src) = 0;
    src->caps = NULL;
    src->num_frames = 0;
    src->num_dropped_frames = 0;
    src->msg_queue = g_async_queue_new_full(free_queue_msg);
    src->msg_thread = NULL;
}

static void
wcbe_src_finalize(GObject *object)
{
    WcbeSrc *src = WCBE_SRC(object);
    g_async_queue_unref(src->msg_queue);
    if (READ_SOCKET(src) > 0) {
        close(READ_SOCKET(src));
    }
    if (WRITE_SOCKET(src) > 0) {
        close(WRITE_SOCKET(src));
    }
    g_static_mutex_free(&(src->ts0_mutex));
    if (src->rate_thread != NULL) {
        GThread *rate_thread = src->rate_thread;
        g_atomic_pointer_set(&(src->rate_thread), NULL);
        g_thread_join(rate_thread);
    }
    g_static_mutex_free(&(src->rate_mutex));
    if (src->ts0 != NULL) {
        g_free(src->ts0);
    }
    G_OBJECT_CLASS(parent_class)->finalize(object);
}

static void
wcbe_src_set_property(GObject *object, guint prop_id,
                      const GValue *value, GParamSpec *pspec)
{
    WcbeSrc *src = WCBE_SRC(object);

    switch (prop_id) {
      case ARG_URI:
        gst_uri_handler_set_uri(GST_URI_HANDLER(object),
                                g_value_get_string(value));
        break;
      case ARG_FPS:
        g_atomic_int_set((gint *)&(src->fps), g_value_get_uint(value));
        break;
      case ARG_TIMESTAMP_BASE: {
        const gchar *ts0 = g_value_get_string(value);
        if (ts0[0] != '\0') {
            gchar **ts0_parts = g_strsplit(ts0, ":", 0);
            if (g_strv_length(ts0_parts) == 3) {
                char *endptr = NULL;
                errno = 0;
                gulong ts0_epoch = strtoul(ts0_parts[0], &endptr, 0);
                if (errno != 0 || *endptr != '\0') {
                    goto bad_ts0;
                }
                gulong ts0_seconds = strtoul(ts0_parts[1], &endptr, 0);
                if (errno != 0 || *endptr != '\0') {
                    goto bad_ts0;
                }
                gulong ts0_cycles = strtoul(ts0_parts[2], &endptr, 0);
                if (errno != 0 || *endptr != '\0') {
                    goto bad_ts0;
                }
                if (ts0_epoch >= (1uL << 3)
#if SIZEOF_UNSIGNED_LONG > 4
                    || ts0_seconds >= (1uL << 32)
#endif
                    || ts0_cycles >= WIDAR_CLOCK_RATE) {
                    goto bad_ts0;
                }
                g_static_mutex_lock(&(src->ts0_mutex));
                src->ts0_epoch = (guint)ts0_epoch;
                src->ts0_seconds = (guint)ts0_seconds;
                src->ts0_cycles = (guint)ts0_cycles;
                if (src->ts0 != NULL) {
                    g_free(src->ts0);
                }
                src->ts0 = g_strdup(ts0);
                g_static_mutex_unlock(&(src->ts0_mutex));
                goto finish_timestamp_base;
            }
        bad_ts0:
            g_info("Failed to set timestamp-base: invalid string '%s'", ts0);
        finish_timestamp_base:
            g_strfreev(ts0_parts);
        }
        else {
            g_static_mutex_lock(&(src->ts0_mutex));
            if (src->ts0 != NULL) {
                g_free(src->ts0);
                src->ts0 = NULL;
            }
            g_static_mutex_unlock(&(src->ts0_mutex));
        }
        break;
      }
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

static void
wcbe_src_get_property(GObject *object, guint prop_id,
                      GValue *value, GParamSpec *pspec)
{
    WcbeSrc *src = WCBE_SRC(object);

    switch (prop_id) {
      case ARG_URI:
        g_value_set_string(value,
                           gst_uri_handler_get_uri(GST_URI_HANDLER(object)));
        break;
      case ARG_NUM_FRAMES:
        g_value_set_uint(value, (guint)g_atomic_int_get(&(src->num_frames)));
        break;
      case ARG_FPS:
        g_value_set_uint(value, src->fps);
        break;
      case ARG_TIMESTAMP_BASE: {
        gchar *ts0;
        if (src->ts0 != NULL) {
            ts0 = src->ts0;
        }
        else {
            ts0 = "";
        }
        g_value_set_string(value, ts0);
        break;
      }
      case ARG_FRAME_RATE:
        g_static_mutex_lock(&(src->rate_mutex));
        g_value_set_float(value, src->smoothed_rate);
        g_static_mutex_unlock(&(src->rate_mutex));
        break;
      case ARG_NUM_DROPPED_FRAMES:
        g_value_set_uint(value,
                         (guint)g_atomic_int_get(&(src->num_dropped_frames)));
        break;
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

/* GstBaseSrc vmethod implementations */

static gboolean
wcbe_src_start(GstBaseSrc *gstbasesrc)
{
    gboolean result = TRUE;
    WcbeSrc *src = WCBE_SRC(gstbasesrc);

    g_assert(src->channel == NULL);

    if (src->uri == NULL) {
        return FALSE;
    }

    // Open the control socket pair
    gint control_sock[2];
    int rc = socketpair (PF_UNIX, SOCK_STREAM, 0, control_sock);
    if (rc == 0) {
        READ_SOCKET(src) = control_sock[0];
        WRITE_SOCKET(src) = control_sock[1];
        fcntl(READ_SOCKET(src), F_SETFL, O_NONBLOCK);
        fcntl(WRITE_SOCKET(src), F_SETFL, O_NONBLOCK);
    }
    else {
        g_warning("Failed to create control socket pair");
        result = FALSE;
    }

    if (result) {
        GError *err = NULL;
        src->msg_thread = g_thread_create(msg_fun, src, TRUE, &err);
        if (src->msg_thread == NULL) {
            g_critical("Failed to spawn frame thread: %s", err->message);
            result = FALSE;
        }
    }

    if (result) {
        src->caps = gst_caps_new_simple(WCBE_WIDAR_FRAME_MEDIA_TYPE, NULL);
        g_atomic_int_set((gint *)&(src->num_frames), 0);
        g_atomic_int_set((gint *)&(src->num_dropped_frames), 0);
        gst_base_src_set_live(gstbasesrc, TRUE);
    }

    return result;
}

static gboolean
wcbe_src_stop(GstBaseSrc *gstbasesrc)
{
    WcbeSrc *src = WCBE_SRC(gstbasesrc);

    if (src->msg_thread != NULL) {
        GThread *msg_thread = src->msg_thread;
        g_atomic_pointer_set(&(src->msg_thread), NULL);
        wcbe_src_unlock(gstbasesrc);
        g_thread_join(msg_thread);
    }

    if (READ_SOCKET(src) > 0) {
        close(READ_SOCKET(src));
        READ_SOCKET(src) = 0;
    }
    if (WRITE_SOCKET(src) > 0) {
        close(WRITE_SOCKET(src));
        WRITE_SOCKET(src) = 0;
    }
    if (src->caps != NULL) {
        gst_caps_unref(src->caps);
        src->caps = NULL;
    }
    return TRUE;
}

static gboolean
wcbe_src_unlock(GstBaseSrc *basesrc)
{
    WcbeSrc *src = WCBE_SRC(basesrc);
    SEND_COMMAND(src, CONTROL_UNLOCK);
    return TRUE;
}

static gboolean
wcbe_src_unlock_stop(GstBaseSrc *basesrc)
{
    WcbeSrc *src = WCBE_SRC(basesrc);

    /* read all stop commands */
    int res = 0;
    while (res >= 0) {
        gchar command;
        READ_COMMAND(src, command, res);
    }
    g_async_queue_lock(src->msg_queue);
    guint len = g_async_queue_length_unlocked(src->msg_queue);
    while (len > 0) {
        queue_msg_t *msg = g_async_queue_pop_unlocked(src->msg_queue);
        if (msg->type != UNLOCK_MSG) {
            g_async_queue_push_unlocked(src->msg_queue, msg);
        }
        else {
            free_queue_msg(msg);
        }
        len--;
    }
    g_async_queue_unlock(src->msg_queue);
    return TRUE;
}

static void
wcbe_src_get_times(GstBaseSrc *basesrc, GstBuffer *buffer,
                   GstClockTime *start, GstClockTime *end)
{
    WcbeSrc *src = WCBE_SRC(basesrc);
    if (src->uri[0] == 'f') {
        guint fps = (guint)g_atomic_int_get((gint *)&(src->fps));
        *start = ((src->num_frames - 1) * GST_SECOND) / fps;
        *end = (src->num_frames * GST_SECOND) / fps;
    }
}

/* GstPushSrc vmethod implementations */

static GstFlowReturn
wcbe_src_create(GstPushSrc *pushsrc, GstBuffer **outbuf)
{
    GstFlowReturn result;
    WcbeSrc *src = WCBE_SRC(pushsrc);

    // Read a frame or message
    queue_msg_t *msg = g_async_queue_pop(src->msg_queue);
    switch (msg->type) {
      case FRAME_MSG:
        // Check read result
        if (msg->frame_buffer != NULL) {
            result = GST_FLOW_OK;
            *outbuf = msg->frame_buffer;
            GST_BUFFER_TIMESTAMP(*outbuf) = GST_CLOCK_TIME_NONE;
            GST_BUFFER_OFFSET(*outbuf) = src->num_frames;
            g_static_mutex_lock(&(src->ts0_mutex));
            if (src->ts0 != NULL) {
                guint fps = (guint)g_atomic_int_get((gint *)&(src->fps));
                guint32 seconds = src->num_frames / fps + src->ts0_seconds;
                guint32 cycles =
                    (src->num_frames % fps) * (WIDAR_CLOCK_RATE / fps)
                    + src->ts0_cycles;
                if (cycles >= WIDAR_CLOCK_RATE) {
                    seconds += 1;
                    cycles -= WIDAR_CLOCK_RATE;
                }
                lf_data_frame_word_t *df =
                    (lf_data_frame_word_t *)GST_BUFFER_DATA(*outbuf);
                df[4] = g_htonl(seconds);
                df[5] = g_htonl((src->ts0_epoch << 29) | cycles);
                guint num_words =
                    ((LF_DATA_FRAME_FTYPE(df) == LF_LTA_DATA_FRAME_FTYPE) ?
                     LF_DATA_FRAME_NUM_WORDS(LF_LTA_DATA_FRAME_NUM_LAGS(df)) :
                     LF_DATA_FRAME_NUM_WORDS(LF_CC_DATA_FRAME_NUM_LAGS(df)))
                    - 1;
                lf_data_frame_word_t checksum = 0;
                guint i;
                for (i = 0; i < num_words; ++i) {
                    checksum += g_ntohl(*df++);
                }
                *df = g_htonl(checksum);
            }
            g_static_mutex_unlock(&(src->ts0_mutex));
            g_atomic_int_inc((gint *)&(src->num_frames));
        }
        else {
            result = GST_FLOW_UNEXPECTED;
        }
        msg->frame_buffer = NULL;
        break;
      case ERROR_MSG:
        result = GST_FLOW_ERROR;
        break;
      case UNLOCK_MSG:
        result = GST_FLOW_WRONG_STATE;
        break;
      case EOF_MSG:
        result = GST_FLOW_UNEXPECTED;
        break;
      default:
        g_assert_not_reached();
        break;
    }
    free_queue_msg(msg);
    return result;
}

static void
wcbe_src_uri_handler_init(gpointer g_iface, gpointer iface_data)
{
    GstURIHandlerInterface *iface = (GstURIHandlerInterface *)g_iface;

    iface->get_type = GST_DEBUG_FUNCPTR(wcbe_src_uri_get_type);
    iface->get_protocols = GST_DEBUG_FUNCPTR(wcbe_src_uri_get_protocols);
    iface->get_uri = GST_DEBUG_FUNCPTR(wcbe_src_uri_get_uri);
    iface->set_uri = GST_DEBUG_FUNCPTR(wcbe_src_uri_set_uri);
}

static GstURIType
wcbe_src_uri_get_type(void)
{
    return GST_URI_SRC;
}

static gchar **
wcbe_src_uri_get_protocols(void)
{
    static gchar *result[] = {"widar", "file", NULL};
    return result;
}

static const gchar *
wcbe_src_uri_get_uri(GstURIHandler *handler)
{
    WcbeSrc *src = WCBE_SRC(handler);
    return src->uri;
}

static gboolean
wcbe_src_uri_set_uri(GstURIHandler *handler, const gchar *uri)
{
    gboolean result = FALSE;
    gboolean wrong_state = FALSE;
    WcbeSrc *src = WCBE_SRC(handler);

    GST_STATE_LOCK(src);
    {
        GstState state = GST_STATE(src);
        if (state != GST_STATE_READY && state != GST_STATE_NULL) {
            wrong_state = TRUE;
        }
    }
    GST_STATE_UNLOCK(src);

    if (!wrong_state) {
        if (src->uri != NULL) {
            g_free(src->uri);
        }
        src->uri = NULL;
        gchar *protocol = gst_uri_get_protocol(uri);
        if (strcmp(protocol, "widar") == 0) {
            gchar *location = gst_uri_get_location(uri);
            gchar *colon = strchr(location, ':');
            if (colon != NULL) {
                gulong port = strtoul(&colon[1], NULL, 10);
                if (port != 0) {
                    *colon = '\0';
                    struct hostent *hostent = gethostbyname(location);
                    if (hostent != NULL && hostent->h_addrtype == AF_INET) {
                        // uri seems to be a hostname or IP address
                        result = TRUE;
                        src->uri = g_strdup(uri);
                    }
                }
            }
            g_free(location);
        }
        else if (strcmp(protocol, "file") == 0) {
            gchar *location = gst_uri_get_location(uri);
            struct stat statbuf;
            int rc = stat(location, &statbuf);
            if (rc != -1 && (statbuf.st_mode & S_IRUSR)) {
                // uri seems to be a filename
                result = TRUE;
                src->uri = g_strdup(uri);
            }
            g_free(location);
        }
        g_free(protocol);
    }
    return result;
}

static gpointer
rate_fun(gpointer data)
{
    WcbeSrc *src = WCBE_SRC(data);
    guint prev_num_frames = 0;
    GTimer *timer = g_timer_new();
    while (g_atomic_pointer_get(&(src->rate_thread)) != NULL) {
        guint num_frames = g_atomic_int_get(&(src->num_frames));
        gdouble elapsed = g_timer_elapsed(timer, NULL);
        if (elapsed > 0.0) {
            g_timer_start(timer);
            guint frame_diff = num_frames - prev_num_frames;
            if (num_frames < prev_num_frames) {
                frame_diff += G_MAXUINT;
            }
            prev_num_frames = num_frames;
            gdouble rate = frame_diff / elapsed;
            gdouble wgt = exp(-EWMA_DECAY * elapsed);
            gfloat smoothed_rate =
                wgt * src->smoothed_rate + (1.0 - wgt) * rate;
            g_static_mutex_lock(&(src->rate_mutex));
            src->smoothed_rate = smoothed_rate;
            g_static_mutex_unlock(&(src->rate_mutex));
        }
        g_usleep(1 * G_USEC_PER_SEC);
    }
    g_timer_destroy(timer);
    return NULL;
}

static gpointer
msg_fun(gpointer data)
{
    WcbeSrc *src = WCBE_SRC(data);

    // Create a GIOChannel from which to read frames.
    gchar *location = gst_uri_get_location(src->uri);
    if (src->uri[0] == 'w') { // 'widar' URI protocol
        gchar *colon = strchr(location, ':');
        if (colon != NULL) {
            gulong port = strtoul(&colon[1], NULL, 10);
            if (port != 0) {
                *colon = '\0';
                struct hostent *hostent = gethostbyname(location);
                if (hostent != NULL && hostent->h_addrtype == AF_INET) {
                    gint sock = socket(PF_INET, SOCK_DGRAM, 0);
                    if (sock != -1) {
                        gsize bufsize = 8 << 20;
                        int rc = setsockopt(sock, SOL_SOCKET, SO_RCVBUF, &bufsize,
                                            sizeof(bufsize));
                        if (rc != 0) {
                            g_warning("Failed to set socket receive buffer size: %s",
                                      strerror(errno));
                        }
                        struct sockaddr_in sockaddr =
                            {.sin_family = AF_INET,
                             .sin_port = htons((short)port),
                             .sin_addr = *(struct in_addr*)hostent->h_addr
                            };
                        rc = bind(sock, (struct sockaddr *)&sockaddr,
                                  sizeof(sockaddr));
                        if (rc != -1) {
                            src->channel = g_io_channel_unix_new(sock);
                            if (src->channel == NULL) {
                                g_warning("Failed to create GIOChannel "
                                          "on socket");
                            }
                        }
                        else {
                            g_message("Failed to bind listening socket: %s",
                                      g_strerror(errno));
                        }
                        if (src->channel == NULL) {
                            close(sock);
                        }
                    }
                    else {
                        g_message("Failed to open listening socket");
                    }
                }
                else {
                    g_info("'gethostbyname(%s)' call failed", location);
                }
            }
            else {
                g_info("Failed to parse URI port number");
            }
        }
        else {
            g_info("widar URI has no port number");
        }
    }

    else if (src->uri[0] == 'f') { // 'file' URI protocol
        gint fd = open(location, O_RDONLY);
        if (fd != -1) {
            src->channel = g_io_channel_unix_new(fd);
            if (src->channel == NULL) {
                g_warning("Failed to create GIOChannel on socket");
            }
        }
        else {
            g_info("Failed to open input file '%s': %s", location,
                   g_strerror(errno));
        }
    }
    g_free(location);

    // Initialize the GIOChannel
    if (src->channel != NULL) {
        g_io_channel_set_close_on_unref(src->channel, TRUE);
        GError *err = NULL;
        if ((g_io_channel_set_encoding(src->channel, NULL, &err)
             == G_IO_STATUS_NORMAL)
            && (g_io_channel_set_flags(src->channel, G_IO_FLAG_NONBLOCK, &err)
                == G_IO_STATUS_NORMAL)) {
            g_io_channel_set_buffered(src->channel, FALSE);
        }
        else {
            g_warning("Failed to initialize GIOChannel");
            err = NULL;
            g_io_channel_shutdown(src->channel, FALSE, &err);
            g_io_channel_unref(src->channel);
            src->channel = NULL;
        }
    }

    // Read frames from source and push them onto src->msg_queue
    if (src->channel != NULL) {
        if (src->uri[0] == 'w') {
            queue_widar_frames(src);
        }
        else {
            queue_file_frames(src);
        }
    }

    if (src->channel != NULL) {
        GError *err = NULL;
        g_io_channel_shutdown(src->channel, FALSE, &err);
        g_io_channel_unref(src->channel);
        src->channel = NULL;
    }

    return NULL;
}

static void
queue_file_frames(WcbeSrc *src)
{
    GstState state;
    GError *err;
    GstBuffer *buf = NULL;
    lf_lag_frame_t lf G_GNUC_UNUSED;

    while (g_atomic_pointer_get(&(src->msg_thread))) {
        switch (source_input(src)) {
          case FRAME_MSG:
            // Check element state
            gst_element_get_state(GST_ELEMENT(src), &state, NULL,
                                  10 * GST_MSECOND);
            // wait for playing state
            if (state == GST_STATE_PLAYING) {
                if (buf == NULL) {
                    // Allocate buffer
                    buf = gst_buffer_try_new_and_alloc(sizeof(lf_lag_frame_t));
                    if (buf == NULL) {
                        g_usleep(G_USEC_PER_SEC * 1000);
                    }
                }
                if (buf != NULL) {
                    gst_buffer_set_caps(buf, src->caps);
                    // Read a frame from file
                    gsize bytes_read = 0;
                    GIOStatus st;
                    err = NULL;
                    st = g_io_channel_read_chars(
                        src->channel, (gchar *)GST_BUFFER_DATA(buf),
                        sizeof(lf.words), &bytes_read, &err);

                    switch (st) {
                      case G_IO_STATUS_NORMAL:
                        if (bytes_read > 0) {
                            while (g_async_queue_length(src->msg_queue) >=
                                   (3 * MAX_QUEUE_SIZE) / 4) {
                                g_usleep(
                                    ((MAX_QUEUE_SIZE / 2) /
                                     (guint)g_atomic_int_get(
                                         (gint *)&(src->fps)))
                                    * G_USEC_PER_SEC);
                            }
                            //GST_BUFFER_SIZE(buf) = bytes_read;
                            queue_msg_t *msg = g_slice_new(queue_msg_t);
                            msg->type = FRAME_MSG;
                            msg->frame_buffer = buf;
                            g_async_queue_push(src->msg_queue, msg);
                            buf = NULL; // will trigger a new buffer alloc
                        }
                        break;
                      case G_IO_STATUS_ERROR: {
                          g_message("Read error on frame source: %s",
                                    err->message);
                          queue_msg_t *msg = g_slice_new(queue_msg_t);
                          msg->type = ERROR_MSG;
                          g_async_queue_push(src->msg_queue, msg);
                          break;
                      }
                      case G_IO_STATUS_AGAIN:
                        g_usleep(G_USEC_PER_SEC / 1000);
                        break;
                      case G_IO_STATUS_EOF: {
                          g_atomic_pointer_set(&(src->msg_thread), NULL);
                          queue_msg_t *msg = g_slice_new(queue_msg_t);
                          msg->type = EOF_MSG;
                          g_async_queue_push(src->msg_queue, msg);
                          break;
                      }
                      default:
                        break;
                    }
                }
            }
            break;
          case ERROR_MSG: {
            queue_msg_t *msg = g_slice_new(queue_msg_t);
            msg->type = ERROR_MSG;
            g_async_queue_push(src->msg_queue, msg);
            break;
          }
          case UNLOCK_MSG: {
            queue_msg_t *msg = g_slice_new(queue_msg_t);
            msg->type = UNLOCK_MSG;
            g_async_queue_push(src->msg_queue, msg);
            break;
          }
          default:
            g_assert_not_reached();
            break;
        }
    }
    if (buf != NULL) {
        gst_buffer_unref(buf);
    }
}

static void
queue_widar_frames(WcbeSrc *src)
{
    GstState state;
    GError *err;
    lf_lag_frame_t lf G_GNUC_UNUSED;
    GstBuffer *throwaway = gst_buffer_new_and_alloc(sizeof(lf_lag_frame_t));
    GstBuffer *buf = NULL;

    while (g_atomic_pointer_get(&(src->msg_thread))) {
        if (buf == NULL || buf == throwaway) {
            // Allocate buffer
            buf = gst_buffer_try_new_and_alloc(sizeof(lf_lag_frame_t));
            if (buf == NULL) {
                buf = throwaway;
            }
        }
        gst_buffer_set_caps(buf, src->caps);
        switch (source_input(src)) {
          case FRAME_MSG: {
            // Read a frame from socket
            gsize bytes_read = 0;
            GIOStatus st;
            err = NULL;
            st = g_io_channel_read_chars(src->channel,
                                         (gchar *)GST_BUFFER_DATA(buf),
                                         sizeof(lf.words), &bytes_read, &err);
            // Check element state
            gst_element_get_state(GST_ELEMENT(src), &state, NULL, 0);
            switch (st) {
              case G_IO_STATUS_NORMAL:
                // Push buffer onto queue in playing state; otherwise
                // don't do anything, and we'll reuse the buffer for
                // the next frame.
                if (state == GST_STATE_PLAYING && bytes_read > 0
                    && buf != throwaway) {
                    //GST_BUFFER_SIZE(buf) = bytes_read;
                    g_async_queue_lock(src->msg_queue);
                    if (g_async_queue_length_unlocked(src->msg_queue)
                        < MAX_QUEUE_SIZE) {
                        queue_msg_t *msg = g_slice_new(queue_msg_t);
                        msg->type = FRAME_MSG;
                        msg->frame_buffer = buf;
                        g_async_queue_push_unlocked(src->msg_queue, msg);
                        buf = NULL; // will trigger a new buffer alloc
                    }
                    else {
                        g_atomic_int_inc((gint *)&(src->num_dropped_frames));
                    }
                    g_async_queue_unlock(src->msg_queue);
                }
                break;
              case G_IO_STATUS_ERROR: {
                  g_message("Read error on frame source: %s", err->message);
                  queue_msg_t *msg = g_slice_new(queue_msg_t);
                  msg->type = ERROR_MSG;
                  g_async_queue_push(src->msg_queue, msg);
                  break;
              }
              case G_IO_STATUS_AGAIN:
                g_usleep(G_USEC_PER_SEC / 1000);
                break;
              case G_IO_STATUS_EOF: {
                g_atomic_pointer_set(&(src->msg_thread), NULL);
                queue_msg_t *msg = g_slice_new(queue_msg_t);
                msg->type = EOF_MSG;
                g_async_queue_push(src->msg_queue, msg);
                break;
              }
              default:
                break;
            }
            break;
          }
          case ERROR_MSG: {
            queue_msg_t *msg = g_slice_new(queue_msg_t);
            msg->type = ERROR_MSG;
            g_async_queue_push(src->msg_queue, msg);
            break;
          }
          case UNLOCK_MSG: {
            queue_msg_t *msg = g_slice_new(queue_msg_t);
            msg->type = UNLOCK_MSG;
            g_async_queue_push(src->msg_queue, msg);
            break;
          }
          default:
            g_assert_not_reached();
            break;
        }
    }

    if (buf != NULL && buf != throwaway) {
        gst_buffer_unref(buf);
    }
    gst_buffer_unref(throwaway);
}

static void
free_queue_msg(gpointer data)
{
    queue_msg_t *msg = data;
    if (msg->type == FRAME_MSG && msg->frame_buffer != NULL) {
        gst_buffer_unref(msg->frame_buffer);
    }
    g_slice_free(queue_msg_t, msg);
}

static queue_msg_type_t
source_input(WcbeSrc *src)
{
    queue_msg_type_t result;

    // 'select' on channel fd and control_sock to find which has data
    // to read.
    fd_set readfds;
    gint rc;
    FD_ZERO(&readfds);
    FD_SET(g_io_channel_unix_get_fd(src->channel), &readfds);
    FD_SET(READ_SOCKET(src), &readfds);
    do {
        rc = select(FD_SETSIZE, &readfds, NULL, NULL, NULL);
    } while (rc == -1 && errno == EINTR);
    if (rc == -1) {
        g_debug("'select' failed: %s", g_strerror(errno));
        result = ERROR_MSG;
    }
    // Quit, if control socket has data.
    else if (FD_ISSET(READ_SOCKET(src), &readfds)) {
        /* read all commands */
        int res = 0;
        while (res >= 0) {
            gchar command;
            READ_COMMAND(src, command, res);
        }
        result = UNLOCK_MSG;
    }
    else {
        result = FRAME_MSG;
    }
    return result;
}
