#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <stdio.h>
#include <errno.h>

#include <wcbe_vect_rotate.h>
#include <wcbe.h>
#include <wcbe_util.h>

GST_DEBUG_CATEGORY_STATIC(wcbe_vect_rotate_debug);
#define GST_CAT_DEFAULT wcbe_vect_rotate_debug

/* Filter signals and args */
enum
{
    /* FILL ME */
    LAST_SIGNAL
};

enum
{
    ARG_0
};

#define VECT_ROTATE_STATIC_CAPS                   \
     WCBE_VECTOR_MEDIA_TYPE                       \
     ","WCBE_ELEMENT_TYPE_PROPERTY"="             \
        WCBE_ELEMENT_TYPE_PROPERTY_TYPE           \
        "{gfloat,guint32,gint32,guint64,gint64}"  \
  ";"WCBE_ARRAY_MEDIA_TYPE                        \
    ","WCBE_ELEMENT_TYPE_PROPERTY"="              \
       WCBE_ELEMENT_TYPE_PROPERTY_TYPE            \
       "{gfloat,guint32,gint32,guint64,gint64}"

static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE(
    "sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS(VECT_ROTATE_STATIC_CAPS)
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE(
    "src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS(VECT_ROTATE_STATIC_CAPS)
    );

GST_BOILERPLATE(WcbeVectRotate, wcbe_vect_rotate, GstBaseTransform,
                GST_TYPE_BASE_TRANSFORM);

static GstFlowReturn wcbe_vect_rotate_transform_ip(GstBaseTransform *trans,
                                                   GstBuffer *buffer);
static GstCaps *wcbe_vect_rotate_transform_caps(GstBaseTransform *trans,
                                                GstPadDirection direction,
                                                GstCaps *caps);
static void
wcbe_vect_rotate_base_init(gpointer gclass)
{
    GST_DEBUG_CATEGORY_INIT(wcbe_vect_rotate_debug, "wcbe_vect_rotate_debug",
                            0, "wcbe_vect_rotate debug");
    static GstElementDetails element_details = {
        "WIDAR vector rotation",
        "Transform/WIDAR",
        "Rotate elements in a vector by half the vector's length",
        "Martin Pokorny <mpokorny@nrao.edu>"
    };

    GstElementClass *element_class = GST_ELEMENT_CLASS(gclass);

    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&sink_factory));
    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&src_factory));
    gst_element_class_set_details(element_class, &element_details);
}

/* initialize the plugin's class */
static void
wcbe_vect_rotate_class_init(WcbeVectRotateClass * klass)
{
    GstBaseTransformClass *gstbasetrans_class =
        GST_BASE_TRANSFORM_CLASS(klass);

    gstbasetrans_class->transform_ip =
        GST_DEBUG_FUNCPTR(wcbe_vect_rotate_transform_ip);
    gstbasetrans_class->transform_caps =
        GST_DEBUG_FUNCPTR(wcbe_vect_rotate_transform_caps);
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
wcbe_vect_rotate_init(WcbeVectRotate *vr G_GNUC_UNUSED,
                      WcbeVectRotateClass *gclass G_GNUC_UNUSED)
{
    return;
}

static GstFlowReturn
wcbe_vect_rotate_transform_ip(GstBaseTransform *trans, GstBuffer *buf)
{
    GstStructure *s = gst_caps_get_structure(GST_BUFFER_CAPS(buf), 0);
    guint vlen = 0;
    guint numv = 0;
    guint rowlen = 0;
    if (gst_structure_has_name(s, WCBE_VECTOR_MEDIA_TYPE)) {
        numv = 1;
        gst_structure_get_uint(s, WCBE_VECTOR_LENGTH_PROPERTY, &vlen);
    }
    else {
        gst_structure_get_uint(s, WCBE_ARRAY_NUM_ROWS_PROPERTY, &numv);
        gst_structure_get_uint(s, WCBE_ARRAY_NUM_COLS_PROPERTY, &vlen);
        rowlen = vlen;
        gst_structure_get_uint(s, WCBE_ARRAY_ROW_LENGTH_PROPERTY, &rowlen);
    }
    const gchar *etype =
        gst_structure_get_string(s, WCBE_ELEMENT_TYPE_PROPERTY);
    if (G_UNLIKELY(numv == 0 || vlen == 0 || etype == NULL)) {
        return GST_BASE_TRANSFORM_FLOW_DROPPED;
    }
    guint halflen = vlen / 2;

#define ROTATE_FN(type)                                \
    void rotate_##type(void) {                         \
        type *array = (type *)GST_BUFFER_DATA(buf);    \
        guint i;                                       \
        for (i = 0; i < numv; ++i) {                   \
            type *front = array;                       \
            type *back = &(array[halflen]);            \
            guint j;                                   \
            for (j = 0; j < halflen; ++j) {            \
                type tmp = *front;                     \
                *front++ = *back;                      \
                *back++ = tmp;                         \
            }                                          \
            array += rowlen;                           \
        }                                              \
    }

    ROTATE_FN(gfloat)
    ROTATE_FN(guint32)
    ROTATE_FN(guint64)

    if (strcmp(etype, "gfloat") == 0) {
        rotate_gfloat();
    }
    else if ((strcmp(etype, "guint32") == 0)
             || (strcmp(etype, "gint32") == 0)) {
        rotate_guint32();
    }
    else if ((strcmp(etype, "guint64") == 0)
             || (strcmp(etype, "gint64") == 0)) {
        rotate_guint64();
    }
    else {
        g_warning("WcbeVectRotate received unsupported element type '%s'",
                  etype);
        return GST_BASE_TRANSFORM_FLOW_DROPPED;
    }
    return GST_FLOW_OK;
}

static GstCaps *
wcbe_vect_rotate_transform_caps(GstBaseTransform *trans,
                                GstPadDirection direction,
                                GstCaps *caps)
{
    return gst_caps_copy(caps);
}
