class Enum (type):
    """Enumeration meta-class.

    To make an enumeration type (class), the class neads to have two
    class variables:
    1) __metaclass__ = Enum
    2) __enums set to a dictionary mapping enumeration strings
    to integers.

    For example,

    class Foo (object):
        __metaclass__ = Enum
        __enums = {'foo': 0, 'bar': 1}

    implements an enumeration type with the values 'foo' and 'bar',
    mapped to 0 and 1, respectively.

    Enumeration values are immutable, and can be created in various
    ways, for example:
    Foo.foo
    Foo('foo')
    Foo(0)

    Enumeration values are ordered, and can be compared. They can also
    be converted to strings or integers as needed.

    The only caveat is to avoid assigning values to any class
    variables (something I've not found a way to prohibit). Doing so
    will only cause problems (new variables are created, but the
    enumeration values remain immutable)!
    """

    def __new__(klass, name, bases, dict):
        enums_name = '_%s__enums' % name
        assert dict.has_key(enums_name)
        enums = dict[enums_name]
        dict['__slots__'] = ['_val']
        return type.__new__(klass, name, bases, dict)

    def __init__(klass, name, bases, dict):
        super(Enum, klass).__init__(name, bases, dict)
        enums_name = '_%s__enums' % name
        assert dict.has_key(enums_name)
        enums = dict[enums_name]
        renums = {}
        for (k, v) in enums.items():
            setattr(klass, k, EnumValue(v))
            renums[v] = k
        def __init__(self, x):
            if x in enums.keys():
                value = enums[x]
            elif x in enums.values():
                value = x
            else:
                raise ValueError
            self._val = value
            return
        setattr(klass, '__init__', __init__)
        setattr(klass, '_%s__renums' % name, renums)
        setattr(klass, '__repr__',
                lambda (self): '%s.%s' % (name, renums[self._val]))
        setattr(klass, '__str__',
                lambda (self): '%s.%s' % (name, renums[self._val]))
        setattr(klass, 'strs', staticmethod(lambda : enums.keys()))
        setattr(klass, 'vals', staticmethod(lambda : renums.keys()))
        setattr(klass, 'nick', lambda (self): '%s' % renums[self._val])
        def lt(self, other):
            if isinstance(other, klass):
                result = (self._val < other._val)
            else:
                result = NotImplemented
            return result
        setattr(klass, '__lt__', lt)
        def le(self, other):
            if isinstance(other, klass):
                result = (self._val <= other._val)
            else:
                result = NotImplemented
            return result
        setattr(klass, '__le__', le)
        def eq(self, other):
            if isinstance(other, klass):
                result = (self._val == other._val)
            else:
                result = NotImplemented
            return result
        setattr(klass, '__eq__', eq)
        def ne(self, other):
            if isinstance(other, klass):
                result = (self._val != other._val)
            else:
                result = NotImplemented
            return result
        setattr(klass, '__ne__', ne)
        def gt(self, other):
            if isinstance(other, klass):
                result = (self._val > other._val)
            else:
                result = NotImplemented
            return result
        setattr(klass, '__gt__', gt)
        def ge(self, other):
            if isinstance(other, klass):
                result = (self._val >= other._val)
            else:
                result = NotImplemented
            return result
        setattr(klass, '__ge__', ge)
        setattr(klass, '__int__', lambda (self): self._val)

class EnumValue (object):
    """Enumeration value object for slots in an Enumeration class."""

    def __init__(self, val):
        self._val = val
        return

    def __get__(self, instance, owner):
        if instance is None:
            return owner(self._val)
        else:
            raise AttributeError
