#! /usr/bin/env python
import os
import os.path
import stat
import errno
import fuse
from fuse import Fuse
import sys
from time import time, sleep

from threading import Thread
from ThreadSafeDict import Dict
from PipelineMC import *

fuse.fuse_python_api = (0, 2)

accmode = os.O_RDONLY | os.O_WRONLY | os.O_RDWR

class CommFailure (Exception):
    pass

class Stat(fuse.Stat):
    def __init__(self):
        self.st_mode = 0
        self.st_ino = 0
        self.st_dev = 0
        self.st_nlink = 0
        self.st_uid = 0
        self.st_gid = 0
        self.st_size = 0
        self.st_atime = 0
        self.st_mtime = 0
        self.st_ctime = 0
        self.st_blocks = 0
        self.st_blksize = 512
        self.st_rdev = 0
        return

class BackendDentry (object):

    dir_mode = stat.S_IFDIR | stat.S_IRWXU | \
        stat.S_IRWXG | stat.S_ISGID | stat.S_IROTH | stat.S_IXOTH
    file_mode = stat.S_IFREG | stat.S_IRUSR | stat.S_IWUSR | \
        stat.S_IRGRP | stat.S_IWGRP | stat.S_IROTH
    dir_access = ~0 & accmode
    file_access = ~(os.O_RDONLY | os.O_WRONLY | os.O_RDWR) & accmode
    xattr_prefix = 'user.edu.nrao.widar.'
    ENOATTR = errno.ENODATA

    def __init__(self, parent, name):
        self.client = parent.client
        self.parent = parent
        self.name = name
        self.root_node = parent.root_node
        if self.parent.be_path is not None:
            if self.name is not None:
                self.be_path = os.path.join(self.parent.be_path, self.name)
            else:
                self.be_path = self.parent.be_path
        else:
            self.be_path = self.name
        if self.be_path not in self.root_node.times:
            st = Stat()
            self.root_node.times[self.be_path] = st
            t = int(time())
            st.st_mtime = t
            st.st_atime = t
            st.st_ctime = t
        self.is_dir = False
        self.is_file = False
        return

    def __children(self):
        return {}
    children = property(lambda(self): self.__children())

    def mark_times(self, *fields):
        t = int(time())
        times = self.root_node.times[self.be_path]
        for f in fields:
            setattr(times, 'st_%s' % f, t)
        return

    def get_node(self, path):
        result = None
        if path == '' or path == '/':
            result = self
        else:
            children = self.children
            sep = path.find('/')
            if sep != -1:
                first = path[:sep]
                rest = path[sep + 1:]
            else:
                first = path
                rest = ''
            if first in children.keys():
                result = children[first].get_node(rest)
        return result

class BackendDir (BackendDentry):

    mode = BackendDentry.dir_mode
    access = BackendDentry.dir_access

    def __init__(self, parent, name):
        super(BackendDir, self).__init__(parent, name)
        self.is_dir = True
        return

    def __attr(self):
        st = Stat()
        st.st_mode = self.mode
        st.st_nlink = 2 + self.num_child_dirs()
        return st
    attr = property(lambda(self): self.__attr())

    def num_child_dirs(self):
        result = 0
        for ch in self.children.values():
            if ch.is_dir:
                result += 1
        return result

class BackendFile (BackendDentry):

    def __init__(self, parent, name, readable=True, writable=False):
        super(BackendFile, self).__init__(parent, name)
        self.is_file = True
        self.buf = None
        self.is_readable = readable
        self.is_writable = writable
        self.is_written = False
        self.mode = stat.S_IFREG
        self.access = 0
        if self.is_readable:
            self.mode |= stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH
            self.access |= os.O_RDONLY
        if self.is_writable:
            self.mode |= stat.S_IWUSR | stat.S_IWGRP
            self.access |= os.O_WRONLY
        if self.is_readable and self.is_writable:
            self.access |= os.O_RDWR
        self.access = ~self.access & accmode
        return

    def __attr(self):
        st = Stat()
        st.st_mode = self.mode
        st.st_nlink = 1
        return st
    attr = property(lambda(self): self.__attr())

    def open(self, flags):
        result = 0
        if flags & os.O_WRONLY:
            self.buf = ''
        else:
            self.buf = self.query()
        return 0

    def create(self):
        self.buf = ''
        return 0

    def release(self):
        if self.buf is not None and self.is_writable and self.is_written:
            self.command(self.buf)
        return 0

    def truncate(self, length):
        if self.buf is not None and self.is_writable:
            buflen = len(self.buf)
            if length < buflen:
                self.buf = self.buf[:length]
            else:
                self.buf = self.buf + ((length - buflen) * '\0')
        return 0

    def read(self, size, offset):
        if (self.buf is None or self.buf == -errno.EINVAL
            or not self.is_readable):
            result = -errno.EINVAL
        else:
            buflen = len(self.buf)
            if offset < buflen:
                result = self.buf[offset:min(offset + size, buflen)]
            else:
                result = None
        return result

    def write(self, buf, offset):
        if (self.buf is None or not self.is_writable):
            result = -errno.EINVAL
        else:
            self.is_written = True
            buflen = len(self.buf)
            if offset == buflen:
                self.buf += buf
            elif offset > buflen:
                self.buf = self.buf + ((offset - buflen) * '\0') + buf
            else:
                prefix = self.buf[:offset] + buf
                if len(prefix) >= buflen:
                    self.buf = prefix
                else:
                    self.buf = prefix + self.buf[len(prefix):]
            result = len(buf)
        return result

class AliasedClient:
    def __init__(self, client, alias):
        self.client = client
        self.alias = alias
        return

    def send_msg(self, msg):
        result = self.client.send_msg(self.alias, msg)
        if result == -1:
            raise CommFailure
        return result

    def recv_msg(self):
        result = self.client.recv_msg(self.alias)
        if result is None:
            raise CommFailure
        return result

    def mc_transaction(self, msg):
        result = self.client.mc_transaction(self.alias, msg)
        if result is None:
            raise CommFailure
        return result

class Backend (BackendDir):

    def __init__(self, client, alias):
        self.client = AliasedClient(client, alias)
        self.be_path = ''
        self.new_config = None
        self.new_activation = None
        self.times = Dict()
        self.root_node = self
        self.times[''] = Stat()
        self.mark_times('ctime', 'mtime')
        self.open_files = Dict()
        return

    def __children(self):
        input_stage = PipelineClient.input_stage
        result = {'configurations': Configurations(self),
                  'activations': Activations(self),
                  'flow' : FlowControl(self),
                  'src' : SrcControl(self),
                  'name' : RemoteName(self),
                  'auxdir' : Auxdir(self),
                  'bdfdir' : Bdfdir(self),
                  'new_subscan' : NewSubscan(self),
                  'cancel_subscan': CancelSubscan(self),
                  '00README' : BackendReadme(self)}
        src = SrcControl(self).query()
        if src.startswith('file://'):
            result['rate'] = RateControl(self)
            result['ts0'] = Ts0Control(self)
        active = [a.config for a in Activations(self).children.values() if
                  (isinstance(a, Activation) and a.state == 'in-use')]
        if len(active) == 1:
            result['active'] = ActiveLink(self, active[0])
        return result
    children = property(lambda(self): self.__children())

    def relative_path(path):
        while len(path) > 0 and path[0] == '/':
            path = path[1:]
        return path
    relative_path = staticmethod(relative_path)

    def get_listing(self, path):
        path = Backend.relative_path(path)
        node = self.get_node(path)
        node.mark_times('atime')
        return node.children.keys()

    def getattr(self, path):
        path = Backend.relative_path(path)
        node = self.get_node(path)
        if node is not None:
            result = node.attr
            st = self.times[node.be_path]
            result.st_atime = st.st_atime
            result.st_mtime = st.st_mtime
            result.st_ctime = st.st_ctime
        else:
            result = -errno.ENOENT
        return result

    def getxattr(self, path, name, size):
        path = Backend.relative_path(path)
        node = self.get_node(path)
        if node is not None:
            try:
                name = name[len(self.xattr_prefix):]
                val =  node.xattr[name]
            except:
                result = -self.ENOATTR
            else:
                if size == 0:
                    result = len(val)
                else:
                    result = val
        else:
            result = -errno.ENOENT
        return result

    def listxattr(self, path, size):
        path = Backend.relative_path(path)
        node = self.get_node(path)
        if node is not None:
            keys = [self.xattr_prefix + k for k in node.xattr.keys()]
            if size == 0:
                result = len(''.join(keys)) + len(keys)
            else:
                result = keys
        else:
            result = -errno.ENONENT
        return result

    def create(self, path, mode):
        path = Backend.relative_path(path)
        (parent, tail) = os.path.split(path)
        node = self.get_node(parent)
        if node is not None:
            (result, name) = node.create(tail)
            if result == 0:
                new_path = os.path.join(parent, name)
                new_node = self.get_node(new_path)
                if new_node is not None:
                    new_node.mark_times('mtime', 'ctime')
                    new_node.create()
                    self.open_files[new_path] = new_node
        else:
            result = -errno.EACCES
        return result

    def open(self, path, flags):
        path = Backend.relative_path(path)
        node = self.get_node(path)
        if node is None:
            result = -errno.ENOENT
        else:
            if flags & node.access != 0:
                result = -errno.EACCES
            else:
                result = node.open(flags)
                if result == 0:
                    self.open_files[path] = node
        return result

    def truncate(self, path, len):
        path = Backend.relative_path(path)
        if path in self.open_files:
            node = self.open_files[path]
        else:
            node = self.get_node(path)
        if node is None:
            result = -errno.ENOENT
        else:
            result = node.truncate(len)
        return result

    def release(self, path, flags):
        path = Backend.relative_path(path)
        if path in self.open_files:
            node = self.open_files[path]
        else:
            node = self.get_node(path)
        if node is None:
            result = -errno.ENOENT
        else:
            if flags & node.access != 0:
                result = -errno.EACCES
            else:
                result = node.release()
                if path in self.open_files:
                    del self.open_files[path]
        return result

    def read(self, path, size, offset):
        path = Backend.relative_path(path)
        try:
            if path in self.open_files:
                node = self.open_files[path]
            else:
                node = self.get_node(path)
            result = node.read(size, offset)
            if result is not None and len(result) > 0:
                node.mark_times('atime')
        except:
            result = -errno.EINVAL
        return result

    def write(self, path, buf, offset):
        path = Backend.relative_path(path)
        try:
            if path in self.open_files:
                node = self.open_files[path]
            else:
                node = self.get_node(path)
            result = node.write(buf, offset)
            if result > 0:
                node.mark_times('mtime')
        except:
            result = -errno.EINVAL
        return result

    def readlink(self, path):
        path = Backend.relative_path(path)
        try:
            node = self.get_node(path)
            result = node.readlink()
            if len(result) > 0:
                node.mark_times('atime')
            return result
        except:
            return -errno.ENOENT

    def rmdir(self, path):
        path = Backend.relative_path(path)
        node = self.get_node(path)
        if node.is_dir:
            result = node.rmdir()
            if result:
                self.clean_dict(self.times, path)
                self.clean_dict(self.open_files, path)
        else:
            result = -errno.ENOTDIR
        return result

    def clean_dict(self, dict, path):
        dict.acquire()
        removed = [n for n in dict.keys() if n.startswith(path)]
        for r in removed:
            del dict[r]
        dict.release()
        return

class BackendReadme (BackendFile):

    readme = """\
WCBE instance directory
-----------------------

This directory represents a running WCBE instance. Directory entries
are described below.

Directories

  * activations: Contains configuration activations.

  * configurations: Contains detailed information on configurations,
    including pipeline elements and their properties.

Files

  Note that when writing to files in this directory, one should always
  overwrite the contents of the existing file. In particular, do not
  attempt to delete the existing file, and recreate a new file of the
  same name in its place.

  * auxdir: Contains name of directory used by instance for auxiliary
    data outputs (i.e., BLF and BLS files). To change the directory,
    simply write to the file; for example, from a bash shell:

    $ echo '/path/to/auxiliary/data/files' > auxdir

    N.B.: the path is a path on the node running the instance.

  * bdfdir: Contains name of directory in which the master
    node-created BDF directories are located. To change the directory,
    simply write to the file; for example, from a bash shell:

    $ echo '/path/to/bdf/directory' > bdfdir

    N.B.: the path is a path on the node running the instance.

    This value is used only when spawning a lag set pipeline instance;
    in particular, changes to this value will not effect running lag
    set pipeline instances.

  * flow: Contains the "go" vs. "stop" state of the pipeline. Values
    that can be written to the file are "on" and "off" (or,
    respectively "1" and "0"). For example, from a bash shell, to stop
    the flow of lag frames through the pipeline:

    $ echo off > flow

  * name: Contains the instance name.

  * new_subscan: Writing a string to this file will send the instance
    a "new subscan" signal. To signal a new subscan with uid UID under
    configuration CONF at time EPOCH:SECOND:CYCLES, the string to
    write has the format EPOCH:SECONDS:CYCLES-CONF-UID. For example,
    from a bash shell, to signal a new subscan in configuration "c2"
    with uid "uid://evla/cbe/X42" to be effective at time 0:1234:0:

    $ echo 0:1234:0-c2-uid://evla/cbe/X42 > new_subscan

    The UID may be omitted, or replaced by the string "null", to
    indicate a "non-writing" subscan; meaning that data in the subscan
    shall not be written to a BDF file.

    Note that the names of configurations are the names of
    sub-directories in the 'confgurations' directory.

  * cancel_subscan: Writing a string to this file will send the
    instance a "cancel subscan" signal. To signal the cancellation of
    the subscan with uid UID under configuration CONF, the string to
    write has the format CONF-UID. For example, from a bash shell, to
    signal the cancellation in configuration "c2" of the subscan with
    the uid "uid://evla/cbe/X42":

    $ echo c2-uid://evla/cbe/X42 > cancel_subscan

    Note that the names of configurations are the names of
    sub-directories in the 'confgurations' directory.

  * src: Contains a URI for the lag frame source of the pipeline.
    There are two options for the URI scheme: "file" and "widar".

    The "file" scheme takes the usual format to identify a file on the
    host running the instance. Files that may be used as a source of
    lag frames are in the BLF (binary lag file) format. When a file is
    used as the lag frame source for the instance several other files
    appear in the present directory.

    * rate: Contains the rate at which to read lag frames from the
      file, in frames per second. For example, from a bash shell, to
      set the "playback" rate to 1000 frames per second:

      $ echo 1000 > rate

    * ts0: Contains the initial timestamp to use for regenerated
      timestamps of the frames read from the file. The purpose of this
      setting is to allow the playback of a BLF file at frame rates
      other than the rate at which frames were produced when recorded.
      For example, from the bash shell, to set the initial timestamp
      to epoch=1, seconds=1234, clock cycles=64000000:

      $ echo 1:1234:64000000 > ts0

    The "widar" scheme identifies a network interface and port on the
    instance host from which to receive UDP packets carrying lag
    frames. Note that the network interface and port are not those of
    the transmitting socket. The network interface can be expressed
    either as an IP address (in "dotted quad" format) or a hostname
    resolvable to an IP address on the host. For example, from a bash
    shell, to set the lag frame source for the instance to the IP
    address associated with the hostname "cbe-node-in-01" and port
    12000:

    $ echo widar://cbe-node-in-01:12000 > src

Symbolic links

  The configuration in which a WCBE instance is currently processing
  lag frames is known as the "active" configuration. When the instance
  has an active configuration, a symbolic link named 'active' will
  appear in this directory that links to the sub-directory under the
  'configurations' directory for the active configuration.

"""

    def __init__(self, parent):
        super(BackendReadme, self).__init__(parent, None)
        return

    def query(self):
        return self.readme

class ActiveLink (BackendFile):

    def __init__(self, parent, active):
        super(ActiveLink, self).__init__(parent, None)
        self.mode = (self.mode & ~stat.S_IFREG) | stat.S_IFLNK
        self.linkval = os.path.join('configurations', active)
        return

    def __attr(self):
        st = Stat()
        st.st_mode = self.mode
        st.st_size = len(self.linkval)
        return st
    attr = property(lambda(self): self.__attr())

    def readlink(self):
        return self.linkval

class Configurations (BackendDir):

    def __init__(self, parent):
        super(Configurations, self).__init__(parent, None)
        return

    def __children(self):
        msg = ConfigListMsg(intent=MsgIntent.query)
        try:
            configs = self.client.mc_transaction(msg).configs
        except DisconnectedError:
            raise
        except CommFailure:
            raise
        except:
            configs = None
        result = {}
        if configs is not None:
            state = {}
            def add_state(cfgs, st):
                for c in cfgs:
                    state[c] = st
            add_state(configs.ready, 'ready')
            add_state(configs.dying, 'dying')
            for (cfg, st) in state.items():
                result[cfg] = Config(self, cfg, st)
        if self.parent.new_config is not None:
            result[self.parent.new_config] = \
                ConfigNew(self, self.parent.new_config)
        result['00README'] = ConfigurationsReadme(self)
        return result
    children = property(lambda(self): self.__children())

    def create(self, config):
        assert config not in self.children.keys()
        if self.parent.new_config is None:
            self.parent.new_config = config
            result = (0, config)
        else:
            result = (-errno.EACCES, None)
        return result

class ConfigurationsReadme (BackendReadme):

    readme = """\
Configurations directory
------------------------

This directory contains one sub-directory for each configuration in
the WCBE instance.

A configuration may be created on the WCBE instance by creating a file
in this directory. The file may have any name, which becomes the name
of the configuration, and the file contents must be a valid WCBE
configuration document. For example, assuming the existence of a WCBE
configuration document named "/path/to/config.xml" (on the system on
which the pipelinefs filesystem is mounted), from a bash shell, to
create a configuration named "c000":

$ cat /path/to/config.xml > c000

If the configuration is successfully created on the WCBE instance, the
(target) file will be replaced with a sub-directory of the same name.
If the creation fails, the file is simply deleted.

A configuration may be detroyed on the WCBE instance by removing the
sub-directory (which will not be empty). For example, to destroy the
configuration created above:

$ rmdir c000

When the configuration has been destroyed on the WCBE instance, its
sub-directory will no longer appear in this directory. Note that there
may be some delay between the command to destroy a configuration and
its actual destruction. The chief cause of this delay will be that the
configuraion is active, and thus will not be destroyed until an
activation for a different configuration has taken over. One may read
the file named "state" in the configuration's sub-directory to
determine whether its destruction has been scheduled.

"""

class ConfigActions (BackendFile):

    def __init__(self, parent, actions):
        super(ConfigActions, self).__init__(parent, None)
        self.actions = actions
        return

    def query(self):
        return '\n'.join(self.actions) + '\n'

class ConfigNew (BackendFile):

    def __init__(self, parent, name):
        super(ConfigNew, self).__init__(parent, None, readable=False,
                                        writable=True)
        self.config_name = name
        self.desc = ''
        return

    def command(self, value):
        msg = ConfigCreateMsg(name=self.config_name, desc=value,
                              intent=MsgIntent.command)
        try:
            self.client.mc_transaction(msg)
            result = len(value)
        except DisconnectedError:
            raise
        except:
            result = -errno.EINVAL
        self.parent.parent.new_config = None
        return result

class Config (BackendDir):

    def __init__(self, parent, name, state):
        super(Config, self).__init__(parent, name)
        self.state = state
        return

    def __children(self):
        result = {'state' : ConfigState(self, self.state),
                  'pipelines' : PipelineDir(self),
                  'activations' : ConfigActivations(self),
                  'specification' : ConfigSpecification(self),
                  'integration' : ConfigIntegration(self),
                  '00README' : ConfigReadme(self)}
        return result
    children = property(lambda(self): self.__children())

    def rmdir(self):
        msg = ConfigDestroyMsg(name=self.name, intent=MsgIntent.command)
        try:
            self.client.mc_transaction(msg)
            result = 0
        except DisconnectedError:
            raise
        except CommFailure:
            raise
        except:
            result = -errno.EINVAL # TODO: correct error value?
        return result

class ConfigReadme (BackendReadme):

    readme = """\
Configuration directory
-----------------------

This directory contains information about the configuration.

* activations: Contains a sorted list of activation times for the
  configuration. Activation times are in WIDAR format
  (epoch:seconds:cycles), and each time is prefixed by a character
  that indicates that activation's state: the '+' character
  corresponds to the 'ready' state; a '-' character, the 'dying'
  state, and a '!' character, the 'in-use' state. See the "00README"
  file in the associated activation directory (under
  ../../activations) for further information on these states.

* specification: Contains the WCBE configuration document as used by
  the WCBE instance. Note that the document may be slightly altered
  from the document used to create the configuration.

* state: Contains a description of the configuration's state on the
  WCBE instance. Two values are possible: 'ready' to indicate that the
  configuration is active or can be activated, and 'dying' to indicate
  that the configuration is scheduled for destruction (and hence
  cannot be activated again). A configuration may be in the dying
  state after a command for its destruction has been received, and yet
  before a different configuration has been activated.

"""

class PipelineDir (BackendDir):

    def __init__(self, parent):
        super(PipelineDir, self).__init__(parent, None)
        return

    def __children(self):
        msg = PipelineListMsg(config=self.be_path, intent=MsgIntent.query)
        try:
            pipelines = self.client.mc_transaction(msg).pipelines
        except DisconnectedError:
            raise
        except CommFailure:
            raise
        except:
            pipelines = []
        result = {}
        for p in pipelines:
            result[p] = Pipeline(self, p)
        result['00README'] = PipelineDirReadme(self)
        return result
    children = property(lambda(self): self.__children())

class PipelineDirReadme (BackendReadme):

    readme = """\
Pipelines directory
-------------------

This directory contains a sub-directory for every lagset processing
stage pipeline in the configuration. Additionally, it may contain a
sub-directory for the input state pipeline, when the associated
configuration is active.

The sub-directories for lagset processing stage pipelines are named
according to the 'id' attributes of the pipeline configuration
elements in the associated WCBE configuration document. Since these
attributes are optional in the configuration document, the names may
have been auto-generated by the WCBE instance. The file
"../specification" will contain the configuration document with the
'id' attributes filled in (where necessary), so an association can be
made between the configuration document and the pipeline names in all
cases.

The input stage pipeline always appears under the name "INPUT_STAGE".

"""

class ConfigState (BackendFile):

    def __init__(self, parent, state):
        super(ConfigState, self).__init__(parent, None)
        self.state = state
        return

    def query(self):
        return self.state + '\n'

class ConfigSpecification (BackendFile):

    def __init__(self, parent):
        super(ConfigSpecification, self).__init__(parent, None)
        return

    def query(self):
        msg = ConfigDocMsg(name=self.parent.name)
        try:
            doc = self.client.mc_transaction(msg).doc
        except DisconnectedError:
            raise
        except CommFailure:
            raise
        except:
            if doc[-1] != '\n':
                doc += '\n'
        return doc

class ConfigIntegration (BackendFile):

    def __init__(self, parent):
        super(ConfigIntegration, self).__init__(parent, None)
        return

    def __xattr(self):
        return {'gtype': 'guint'}
    xattr = property(lambda (self): self.__xattr())

    def query(self):
        msg = ConfigIntegrationMsg(config=self.parent.name)
        try:
            result = str(self.client.mc_transaction(msg).usec) + '\n'
        except DisconnectedError:
            raise
        except CommFailure:
            raise
        except:
            result = '\n'
        return result

class ConfigActivations (BackendFile):

    def __init__(self, parent):
        super(ConfigActivations, self).__init__(parent, None)
        return

    def query(self):
        msg = ActivationListMsg()
        try:
            activations = self.client.mc_transaction(msg).activations
        except DisconnectedError:
            raise
        except CommFailure:
            raise
        except:
            activations = ActivationList()
        config = self.parent.name
        def format(acts, st):
            result = []
            for a in acts:
                parts = a.split(':')
                if parts[0] == config:
                    result.append('%1d:%010d:%010d%c' %
                                  (int(parts[1]), int(parts[2]), int(parts[3]),
                                   st))
            return result
        acts = format(activations.ready, '+') + \
               format(activations.in_use, '!') + \
               format(activations.dying, '-')
        acts.sort()
        def rewrite(acc, a):
            return acc + ('%c %s\n' % (a[-1:], a[:-1]))
        return reduce(rewrite, acts, '')

class Activations (BackendDir):

    def __init__(self, parent):
        super(Activations, self).__init__(parent, None)
        return

    def __children(self):
        msg = ActivationListMsg()
        result = {}
        try:
            activations = self.client.mc_transaction(msg).activations
        except DisconnectedError:
            raise
        except CommFailure:
            raise
        except:
            activations = ActivationList()
        def add_activations(acts, state):
            for a in acts:
                parts = a.split(':')
                config = parts[0]
                epoch = int(parts[1])
                seconds = int(parts[2])
                cycles = int(parts[3])
                result[Activations.complete_name(
                    config, epoch, seconds, cycles)] = \
                    Activation(self, config=config, epoch=epoch,
                               seconds=seconds, cycles=cycles,
                               state=state)
            return
        add_activations(activations.ready, 'ready')
        add_activations(activations.in_use, 'in-use')
        add_activations(activations.dying, 'dying')
        if self.parent.new_activation is not None:
            if self.parent.new_activation.startswith('now'):
                actparts = self.parent.new_activation.split('-')
                if len(actparts) == 2 and actparts[0] == 'now':
                    result[self.parent.new_activation] = \
                        ActivationNew(self, config=actparts[1].strip(),
                                      now=True)
            else:
                actparts = self.parent.new_activation.split(':')
                if len(actparts) == 3:
                    actparts[2:] = actparts[2].split('-')
                    if len(actparts) == 4:
                        result[self.parent.new_activation] = \
                            ActivationNew(self, config=actparts[3].strip(),
                                          epoch=int(actparts[0]),
                                          seconds=int(actparts[1]),
                                          cycles=int(actparts[2]))
        result['00README'] = ActivationsReadme(self)
        return result
    children = property(lambda(self): self.__children())

    def complete_name(config, epoch, seconds, cycles):
        return '%1d:%010d:%010d-%s' % \
               (int(epoch), int(seconds), int(cycles), config.strip())
    complete_name = staticmethod(complete_name)

    def create(self, activation):
        assert activation not in self.children.keys()
        if self.parent.new_activation is None:
            self.parent.new_activation = activation
            result = (0, self.parent.new_activation)
        else:
            result = (-errno.EACCES, None)
        return result

class ActivationsReadme (BackendReadme):

    readme = """\
Configuration activations directory
-----------------------------------

This directory contains one sub-directory for each current
configuration activation in the WCBE instance. The names of
sub-directories are determined by the activation time and the name of
the activated configuration, joined by a hyphen character. Note that
the times are in WIDAR format, that is, EPOCH:SECONDS:CYCLES. For
example, an activation with the time of epoch=0, seconds=1234,
cycles=0 for the configuration named "c0" would appear in this
directory as a sub-directory with the name
"0:0000001234:0000000000-c0".

To create an activation on the WCBE instance, one needs only to create
a file and write to it a single character. The file name determines
the activation that is sent to the WCBE instance, in the same way that
the activation determines the sub-directory name as described above.
However, whereas the sub-directory name is strictly formatted with
respect to the numbers of digits in the three time components, one may
use as any number of digits in those components of the file name. For
example, from a bash shell, to create the activation in the previous
paragraph:

$ echo 1 > 0:1234:0-c0

Note that the directory will not be created, and the initial file will
be deleted, if the activation could not be created on the WCBE
instance. In particular, the configuration associated with an
activation must be in existence on the WCBE instance before the
activation can be created.

One other form of the file name is accepted, the so-called "now
activation". A "now activation" command creates an activation on the
WCBE instance with an activation time slightly beyond (by about one
second) the current activation's activation time. Note that a "now
activation\'s" activation time therefore may be in the past. This
newly created activation will cause the named configuration to become
active immediately after its creation on the WCBE instance. After the
"now activation" is created, it will appear in this directory as any
other activation. As an example, from a bash shell, to create a "now
activation" for the configuration named "c1":

$ echo 1 > now-c1

An activation may be destroyed in the WCBE instance by removing its
sub-directory. Note that the directory will not be empty when removing
it. For example, from a bash shell, to destroy the activation
associated with the configuration "c0" created in the example above:

$ rmdir 0\:0000001234\:0000000000-c0

(The backslashes are needed in the bash shell to escape the ':'
characters in the file name.)

"""

class Activation (BackendDir):

    def __init__(self, parent, config, epoch, seconds, cycles, state):
        super(Activation, self).__init__(parent, None)
        self.config = config
        self.epoch = epoch
        self.seconds = seconds
        self.cycles = cycles
        self.state = state
        return

    def __children(self):
        result = {'config': ActivationConfig(self, self.config),
                  'time': ActivationTime(self, self.epoch,
                                         self.seconds, self.cycles),
                  'state': ActivationState(self, self.state),
                  '00README': ActivationReadme(self)}
        return result
    children = property(lambda(self): self.__children())

    def rmdir(self):
        msg = RemoveActivationMsg(config=self.config, epoch=self.epoch,
                                  seconds=self.seconds, cycles=self.cycles)
        try:
            self.client.mc_transaction(msg)
            result = 0
        except DisconnectedError:
            raise
        except:
            result = -errno.EINVAL # TODO: correct error value?
        return result

class ActivationReadme (BackendReadme):

    readme = """\
Activation directory
--------------------

This directory contains two files and a symbolic link that describe
the activation.

* config: A symlink that points to the associated configuration in the
  configurations directory (i.e., ../../configurations).

* state: Contains the state of the activation, one of 'ready',
  'in-use', or 'dying'. 'ready' indicates that the activation exists,
  but its activation time has not been reached; 'in-use' indicates
  that the activation time has been reached and the associated
  configuration is in effect; and 'dying' indicates that a more recent
  activation has taken over, but the activation's resources have not
  yet been completely freed.

* time: Contains the activation time as a string of three
  space-separated numbers equal to the epoch, seconds and clock cycles
  of the time, in that order.

"""

class ActivationTime (BackendFile):

    def __init__(self, parent, epoch, seconds, cycles):
        super(ActivationTime, self).__init__(parent, None)
        self.val = '%d %d %d\n' % (epoch, seconds, cycles)
        return

    def query(self):
        return self.val

class ActivationState (BackendFile):

    def __init__(self, parent, state):
        super(ActivationState, self).__init__(parent, None)
        self.val = state
        return

    def query(self):
        return self.val + '\n'

class ActivationConfig (BackendFile):

    def __init__(self, parent, config):
        super(ActivationConfig, self).__init__(parent, None)
        self.mode = (self.mode & ~stat.S_IFREG) | stat.S_IFLNK
        self.linkval = os.path.join('..', '..', 'configurations', config)
        return

    def __attr(self):
        st = Stat()
        st.st_mode = self.mode
        st.st_size = len(self.linkval)
        return st
    attr = property(lambda(self): self.__attr())

    def readlink(self):
        return self.linkval

class ActivationNew (BackendFile):

    def __init__(self, parent, config, epoch=0, seconds=0, cycles=0,
                 now=False):
        super(ActivationNew, self).__init__(parent, None, readable=False,
                                            writable=True)
        self.config = config
        self.epoch = epoch
        self.seconds = seconds
        self.cycles = cycles
        self.now = now
        return

    def command(self, value):
        if self.now:
            msg = AddActivationMsg(config=self.config, now=True)
        else:
            msg = AddActivationMsg(config=self.config, epoch=self.epoch,
                                   seconds=self.seconds, cycles=self.cycles)
        try:
            self.client.mc_transaction(msg)
            result = len(value)
        except:
            result = -errno.EINVAL
        self.parent.parent.new_activation = None
        return result

class Pipeline (BackendDir):

    def __init__(self, parent, name):
        super(Pipeline, self).__init__(parent, name)
        return

    def __children(self):
        msg = ElementListMsg(pipeline=self.be_path, intent=MsgIntent.query)
        try:
            elements = self.client.mc_transaction(msg).elements
        except DisconnectedError:
            raise
        except CommFailure:
            raise
        except:
            elements = []
        result = {}
        for e in elements:
            result[e] = Element(self, e)
        result['00README'] = PipelineReadme(self)
        return result
    children = property(lambda(self): self.__children())

class PipelineReadme (BackendReadme):

    readme = """\
Pipeline directory
------------------

This directory contains a sub-directory for each (GStreamer) element
in the pipeline. In the case of a lagset processing stage pipeline,
the sub-directories are named according to the names of the elements
as given in the configuration document. In the case of the input stage
pipeline, the names are fixed by the input stage pipeline program.

"""

class Element (BackendDir):

    def __init__(self, parent, name):
        super(Element, self).__init__(parent, name)
        return

    def __children(self):
        result = {'properties': PropertyDir(self),
                  'pads': PadDir(self),
                  '00README': ElementReadme(self)}
        return result
    children = property(lambda(self):self.__children())

class ElementReadme (BackendReadme):

    readme = """\
Element directory
-----------------

This directory contains information for the (GStreamer) element.

* pads: Contains information on the element's pads.

* properties: Contains information on the element's properties.

"""

class PropertyDir (BackendDir):

    def __init__(self, parent):
        super(PropertyDir, self).__init__(parent, None)
        return

    def __children(self):
        msg = PropertyListMsg(element=self.be_path, intent=MsgIntent.query)
        try:
            properties = self.client.mc_transaction(msg).properties
        except DisconnectedError:
            raise
        except CommFailure:
            raise
        except:
            properties = []
        result = {}
        for p in properties:
            langle = p.find('<')
            assert langle != -1
            rangle = p.rfind('>')
            assert rangle != -1
            name = p[:langle]
            comma = p[langle:].find(',') + langle
            assert comma != -1
            valtype = p[langle + 1:comma]
            rw =  p[comma + 1:rangle]
            result[name] = Property(self, name, valtype, readable=('r' in rw),
                                    writable=('w' in rw))
        result['00README'] = PropertyDirReadme(self)
        return result
    children = property(lambda(self): self.__children())

class PropertyDirReadme (BackendReadme):

    readme = """\
Element properties directory
----------------------------

This directory contains information on the element's properties. Each
element property appears as a single file. File read/write permissions
are set in accordance with each property's read- and write-ability.

The data type of a property can be inspected by reading the
"user.edu.nrao.widar.gtype" extended file attribute. These data types
are expressed as GLib data types or custom enumeration and flag types.
To read the extended file attribute value from a bash shell:

$ getfattr --name user.edu.nrao.widar.gtype [PATH]

or, perhaps shorter:

$ getfattr --dump [PATH]

"""

class PadDir (BackendDir):

    def __init__(self, parent):
        super(PadDir, self).__init__(parent, None)
        return

    def __children(self):
        msg = PadListMsg(element=self.be_path, intent=MsgIntent.query)
        try:
            pads = self.client.mc_transaction(msg).pads
        except DisconnectedError:
            raise
        except CommFailure:
            raise
        except:
            pads = []
        result = {}
        for p in pads:
            result[p] = Pad(self, p)
        result['00README'] = PadDirReadme(self)
        return result
    children = property(lambda(self): self.__children())

class PadDirReadme (BackendReadme):

    readme = """\
Element pads directory
----------------------

This directory contains a sub-directory for every pad of the element.

"""

class Pad (BackendDir):

    def __init__(self, parent, name):
        super(Pad, self).__init__(parent, name)
        return

    def __children(self):
        result = {'peer': PadPeer(self),
                  'caps': PadCaps(self),
                  '00README': PadReadme(self)}
        return result
    children = property(lambda(self): self.__children())

class PadReadme (BackendReadme):

    readme = """\
Pad directory
-------------

This directory contains information on the state of an element's pad.

* caps: The capabilities of the pad.

* peer: A symbolic link to the directory representing the linked pad.

"""

class PadPeer (BackendFile):

    def __init__(self, parent):
        super(PadPeer, self).__init__(parent, None)
        self.mode = (self.mode & ~stat.S_IFREG) | stat.S_IFLNK
        return

    def __attr(self):
        st = Stat()
        st.st_mode = self.mode
        st.st_size = len(self.readlink())
        return st
    attr = property(lambda(self): self.__attr())

    def readlink(self):
        msg = PadPeerMsg(pad=self.be_path, intent=MsgIntent.query)
        try:
            result = self.client.mc_transaction(msg).peer
            (element, pad) = result.split('/')
            result = os.path.join('..', '..', '..', element, 'pads', pad)
        except DisconnectedError:
            raise
        except CommFailure:
            raise
        except:
            result = ''
        return result

class PadCaps (BackendFile):

    def __init__(self, parent):
        super(PadCaps, self).__init__(parent, None)
        return

    def query(self):
        msg = PadCapsMsg(pad=self.be_path, intent=MsgIntent.query)
        try:
            result = self.client.mc_transaction(msg).caps
        except DisconnectedError:
            raise
        except CommFailure:
            raise
        except:
            result = ''
        return result + '\n'

class FlowControl (BackendFile):

    def __init__(self, parent):
        super(FlowControl, self).__init__(parent, None, readable=True,
                                          writable=True)
        return

    def query(self):
        msg = FlowMsg(intent=MsgIntent.query)
        try:
            result = self.client.mc_transaction(msg).state.nick()
        except DisconnectedError:
            raise
        except CommFailure:
            raise
        except:
            result = ''
        return result + '\n'

    def command(self, value):
        try:
            msg = FlowMsg(intent=MsgIntent.command, state=Flow(value.strip()))
            self.client.mc_transaction(msg)
            result = len(value)
        except DisconnectedError:
            raise
        except:
            result = -errno.EINVAL
        return result

class SrcControl (BackendFile):

    def __init__(self, parent):
        super(SrcControl, self).__init__(parent, None, readable=True,
                                         writable=True)
        return

    def query(self):
        msg = SrcMsg(intent=MsgIntent.query)
        try:
            result = self.client.mc_transaction(msg).name
        except DisconnectedError:
            raise
        except CommFailure:
            raise
        except:
            result = ''
        return result + '\n'

    def command(self, value):
        try:
            msg = SrcMsg(intent=MsgIntent.command, name=value.strip())
            self.client.mc_transaction(msg)
            result = len(value)
        except DisconnectedError:
            raise
        except:
            result = -errno.EINVAL
        return result

class RateControl (BackendFile):

    def __init__(self, parent):
        super(RateControl, self).__init__(parent, None, readable=True,
                                          writable=True)
        return

    def query(self):
        msg = RateMsg(intent=MsgIntent.query)
        try:
            result = str(self.client.mc_transaction(msg).rate)
        except DisconnectedError:
            raise
        except CommFailure:
            raise
        except:
            result = ''
        return result + '\n'

    def command(self, value):
        try:
            msg = RateMsg(intent=MsgIntent.command, rate=int(value.strip()))
            self.client.mc_transaction(msg)
            result = len(value)
        except DisconnectedError:
            raise
        except:
            result = -errno.EINVAL
        return result

class Ts0Control (BackendFile):

    def __init__(self, parent):
        super(Ts0Control, self).__init__(parent, None, readable=True,
                                         writable=True)
        return

    def query(self):
        msg = Ts0Msg(intent=MsgIntent.query)
        try:
            result = self.client.mc_transaction(msg).ts0
        except DisconnectedError:
            raise
        except CommFailure:
            raise
        except:
            result = ''
        return result + '\n'

    def command(self, value):
        try:
            msg = Ts0Msg(intent=MsgIntent.command, ts0=value.strip())
            self.client.mc_transaction(msg)
            result = len(value)
        except DisconnectedError:
            raise
        except:
            result = -errno.EINVAL
        return result

class RemoteName (BackendFile):

    def __init__(self, parent):
        super(RemoteName, self).__init__(parent, None)
        return

    def query(self):
        msg = NameMsg(intent=MsgIntent.query)
        try:
            result = self.client.mc_transaction(msg).name
        except DisconnectedError:
            raise
        except CommFailure:
            raise
        except:
            result = ''
        return result + '\n'

class Auxdir (BackendFile):

    def __init__(self, parent):
        super(Auxdir, self).__init__(parent, None, readable=True,
                                     writable=True)
        return

    def query(self):
        msg = AuxOutputDirMsg(intent=MsgIntent.query)
        try:
            result = self.client.mc_transaction(msg).name
        except DisconnectedError:
            raise
        except CommFailure:
            raise
        except:
            result = ''
        return result + '\n'

    def command(self, value):
        try:
            msg = AuxOutputDirMsg(intent=MsgIntent.command, name=value.strip())
            self.client.mc_transaction(msg)
            result = len(value)
        except DisconnectedError:
            raise
        except:
            result = -errno.EINVAL
        return result

class NewSubscan (BackendFile):

    def __init__(self, parent):
        super(NewSubscan, self).__init__(parent, None, readable=False,
                                         writable=True)
        return

    def command(self, value):
        parts = value.split('-')
        uid = None
        if len(parts) == 2:
            uid = 'null'
        elif len(parts) == 3:
            uid = parts[2].strip()
        if uid is not None:
            config = parts[1].strip()
            tparts = parts[0].split(':')
            if len(tparts) == 3:
                try:
                    msg = NewSubscanMsg(config=config, uid=uid,
                                        epoch=int(tparts[0]),
                                        seconds=int(tparts[1]),
                                        cycles=int(tparts[2]))
                    self.client.mc_transaction(msg)
                    result = len(value)
                except DisconnectedError:
                    raise
                except:
                    result = -errno.EINVAL
        else:
            result = 0
        return result

class CancelSubscan (BackendFile):

    def __init__(self, parent):
        super(CancelSubscan, self).__init__(parent, None, readable=False,
                                            writable=True)
        return

    def command(self, value):
        parts = value.split('-')
        if len(parts) == 2:
            config = parts[0].strip()
            uid = parts[1].strip()
            try:
                msg = CancelSubscanMsg(config=config, uid=uid)
                self.client.mc_transaction(msg)
                result = len(value)
            except DisconnectedError:
                raise
            except:
                result = -errno.EINVAL
        else:
            result = 0
        return result

class Bdfdir (BackendFile):

    def __init__(self, parent):
        super(Bdfdir, self).__init__(parent, None, readable=True,
                                     writable=True)
        return

    def query(self):
        msg = BdfOutputDirMsg(intent=MsgIntent.query)
        try:
            result = self.client.mc_transaction(msg).name
        except DisconnectedError:
            raise
        except CommFailure:
            raise
        except:
            result = ''
        return result + '\n'

    def command(self, value):
        try:
            msg = BdfOutputDirMsg(intent=MsgIntent.command, name=value.strip())
            self.client.mc_transaction(msg)
            result = len(value)
        except DisconnectedError:
            raise
        except:
            result = -errno.EINVAL
        return result

class Property (BackendFile):

    def __init__(self, parent, name, type, readable, writable):
        super(Property, self).__init__(parent, name, readable=readable,
                                       writable=writable)
        self.type = type
        return

    def __xattr(self):
        return {'gtype': self.type}
    xattr = property(lambda (self): self.__xattr())

    def query(self):
        msg = PropertyMsg(prop=self.be_path, valtype=self.type,
                          intent=MsgIntent.query)
        try:
            result = str(self.client.mc_transaction(msg).value)
        except DisconnectedError:
            raise
        except CommFailure:
            raise
        except:
            result = ''
        return result + '\n'

    def command(self, value):
        try:
            msg = PropertyMsg(prop=self.be_path, valtype=self.type,
                              value=value.strip(), intent=MsgIntent.command)
            self.client.mc_transaction(msg)
            result = len(value)
        except DisconnectedError:
            raise
        except:
            result = -errno.EINVAL
        return result

class CacheTimeout (Thread):

    def __init__(self, cache, timeout):
        super(CacheTimeout, self).__init__()
        self.cache = cache
        self.timeout = timeout
        return

    def run(self):
        while 1:
            t = time()
            old = [(k, v) for (k, v) in self.cache.items()
                   if t > v[1] + self.timeout]
            for (k, v) in old:
                self.cache.acquire()
                try:
                    if t > self.cache[k][1] + self.timeout:
                        del self.cache[k]
                except KeyError:
                    pass
                self.cache.release()
            sleep(self.timeout)
        return

class PipelinefsReadme:

    readme =  """\
pipelinefs root directory
-------------------------

A pipelinefs filesystem is mounted on this directory; it is used to
monitor and control WCBE backend instances on nodes in the backend
cluster from the cluster's master node.

Running WCBE instances appear in this directory as sub-directories.

A WCBE instance on a (remote) node is identified by a host name
HOSTNAME, a port number PORT (for M&C communication with the
instance), and, optionally, an instance name NAME. A WCBE instance may
be connected to or newly spawned by creating a file in this directory,
and writing a string of the form 'HOSTNAME:PORT:NAME' or
'HOSTNAME:PORT' to that file. When communication with the instance has
been established, the file in this directory will become a
sub-directory (with the same name as the file); if spawning and/or
communication is unsuccessful, the file is automatically deleted.

For example, from a bash shell one may spawn a WCBE instance named
'wcbe01' on the host 'cbe-node-01', communicating on port 9000,
appearing as a sub-directory (of this directory) named '01' with the
following command:

$ echo 'cbe-node-01:9000:wcbe01' > 01

To shut down an instance, remove the directory for that instance
(which will not be empty when doing this). For example, to shut down
the instance spawned by the previous example:

$ rmdir 01

Finally, note that when the filesystem is unmounted, a file named
".state" will be written to the mount point directory. This file will
contain information on the WCBE instances that were connected at the
time the filesystem was unmounted. When the filesystem is remounted,
an attempt will be made to reconnect to those instances recorded in
the ".state" file in the mount point directory. In no way is the
presence or absence of this file essential to the proper startup of
the filesystem.

"""
    def __init__(self):
        self.access = os.O_RDONLY
        self.st = Stat()
        self.st.st_mode = \
            stat.S_IFREG | stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH
        self.st.st_size = len(self.readme)
        self.st.st_nlink = 1
        t = int(time())
        self.st.st_atime = t
        self.st.st_mtime = t
        self.st.st_ctime = t
        return

    def touch_atime(self):
        self.st.st_atime = int(time())
        return

    def read(self, size, offset):
        self.st.st_atime = int(time())
        if offset == 0:
            result = self.readme
        else:
            result = None
        return result

    def open(self, flags):
        if flags & self.access != 0:
            return -errno.EACCES
        return 0

    def getattr(self):
        return self.st

class PipelineFS (Fuse):

    def __init__(self, *args, **kw):
        super(PipelineFS, self).__init__(*args, **kw)
        self.pipeline_state = []
        self.client = None
        self.trees = {}
        self.readme = PipelinefsReadme()
        return

    def main(self, *args, **kw):
        statefile_path = os.path.abspath(
            os.path.join(self.fuse_args.mountpoint, '.state'))
        try:
            statefile = open(statefile_path, 'r')
            for line in statefile:
                p = line.split()
                if len(p) == 2:
                    self.pipeline_state.append(tuple(p))
            statefile.close()
        except:
            pass
        Fuse.main(self, *args, **kw)
        statefile = open(statefile_path, 'w')
        for t in self.trees:
            (be, addr) = self.trees[t]
            statefile.write('%s\t%s\n' % (addr, t))
        statefile.close()
        del self.trees
        if self.client is not None:
            del self.client
        return

    def fsinit(self):
        self.cache_timeout = 1
        self.attrs = Dict()
        tmout = CacheTimeout(self.attrs, self.cache_timeout + 1)
        tmout.setDaemon(1)
        tmout.start()
        self.xattrs = Dict()
        tmout = CacheTimeout(self.xattrs, self.cache_timeout + 1)
        tmout.setDaemon(1)
        tmout.start()
        self.client = PipelineClient()
        self.new_trees = []
        self.times = Dict()
        self.times[''] = Stat()
        self.mark_times('', 'ctime', 'mtime')
        for (addr, alias) in self.pipeline_state:
            self.new_trees.append(alias)
            self.times[alias] = Stat()
            rc = self.link_pipeline(alias, addr, False)
        return

    def unlink_pipeline(self, alias):
        self.client.remove_pipeline(alias)
        del self.trees[alias]
        try:
            del self.times[alias]
        except KeyError:
            pass
        return

    def link_pipeline(self, alias, addr, allow_new=True):
        self.new_trees.remove(alias)
        del self.times[alias]
        name_parts = addr.split(':')
        rc = False
        if len(name_parts) == 1:
            rc = self.client.add_pipeline(alias=alias,
                                          name=name_parts[0],
                                          allow_new=allow_new)
        elif len(name_parts) == 2:
            rc = self.client.add_pipeline(alias=alias,
                                          host=name_parts[0],
                                          port=int(name_parts[1]),
                                          allow_new=allow_new)
        elif len(name_parts) == 3:
            rc = self.client.add_pipeline(alias=alias,
                                          host=name_parts[0],
                                          port=int(name_parts[1]),
                                          name=name_parts[2],
                                          allow_new=allow_new)
        if rc:
            self.trees[alias] = (Backend(alias=alias, client=self.client),
                                 addr)
            result = 0
        else:
            result = -errno.ENXIO # TODO: return a better error code?
        return result

    def mark_times(self, path, *fields):
        t = int(time())
        try:
            times = self.times[path]
        except KeyError:
            times = Stat()
            self.times[path] = times
        for f in fields:
            setattr(times, 'st_%s' % f, t)
        return

    def split_pipeline_path(self, path):
        if path[0] == '/':
            path = path[1:]
        slash = path.find('/')
        if slash != -1:
            result = (path[:slash], path[slash:])
        else:
            result = (path, '')
        return result

    def getattr(self, path):
        if path == '/00README':
            result = self.readme.getattr()
        else:
            try:
                (attr, lasttm) = self.attrs[path]
            except KeyError:
                (attr, lasttm) = (None, None)
            if (attr is None) or (time() > lasttm + self.cache_timeout):
                (alias, path) = self.split_pipeline_path(path)
                try:
                    result = self.trees[alias][0].getattr(path)
                    self.attrs[path] = (result, time())
                except KeyError:
                    if path == '':
                        st = Stat()
                        if alias == '':
                            st.st_mode = stat.S_IFDIR | stat.S_IRWXU | \
                                stat.S_IRWXG | stat.S_ISGID | \
                                stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH
                            st.st_nlink = 2 + len(self.trees.keys())
                        elif alias in self.new_trees:
                            st.st_mode = stat.S_IFREG | stat.S_IRUSR | \
                                         stat.S_IWUSR | stat.S_IRGRP | \
                                         stat.S_IROTH
                            st.st_nlink = 1
                        else:
                            st = -errno.ENOENT
                        if st != -errno.ENOENT:
                            st.st_atime = self.times[alias].st_atime
                            st.st_mtime = self.times[alias].st_mtime
                            st.st_ctime = self.times[alias].st_ctime
                        result = st
                    else:
                        result = None
                    self.attrs[path] = (result, time())
                except DisconnectedError:
                    self.unlink_pipeline(alias)
                    result = None
                    self.attrs[path] = (None, time())
                except CommFailure:
                    result = -errno.ENOENT
                    self.attrs[path] = (None, time())
            else:
                result = attr
        return result

    def readdir(self, path, offset):
        result = None
        (alias, path) = self.split_pipeline_path(path)
        if alias != '':
            try:
                tree = self.trees[alias][0]
            except KeyError:
                pass
            else:
                try:
                    listing = tree.get_listing(path)
                except DisconnectedError:
                    self.unlink_pipeline(alias)
                    listing = []
                except CommFailure:
                    listing = []
                    yield -errno.ENOENT
                for e in listing:
                    yield fuse.Direntry(e)
        else:
            for t in self.trees:
                self.trees[t][0].mark_times('atime')
                yield fuse.Direntry(t)
            for t in self.new_trees:
                self.mark_times(t, 'atime')
                yield fuse.Direntry(t)
                self.readme.touch_atime()
            yield fuse.Direntry('00README')
        return

    def open(self, path, flags):
        if path == '/00README':
            result = self.readme.open(flags)
        else:
            (alias, path) = self.split_pipeline_path(path)
            try:
                result = self.trees[alias][0].open(path, flags)
            except KeyError:
                if path == '' and alias in self.new_trees:
                    result = 0
                else:
                    result = -errno.ENOENT
            except DisconnectedError:
                self.unlink_pipeline(alias)
                result = -errno.EFAULT
            except CommFailure:
                result = -errno.ENOENT
        return result

    def release(self, path, flags):
        if path == '/00README':
            result = 0
        else:
            (alias, path) = self.split_pipeline_path(path)
            try:
                result = self.trees[alias][0].release(path, flags)
            except KeyError:
                if path == '' and alias in self.new_trees:
                    result = 0
                else:
                    result = -errno.ENOENT
            except DisconnectedError:
                self.unlink_pipeline(alias)
                result = -errno.EFAULT
            except CommFailure:
                result = -errno.ENOENT
        return result

    def create(self, path, mode, info):
        (alias, path) = self.split_pipeline_path(path)
        try:
            result = self.trees[alias][0].create(path, mode)
        except KeyError:
            if path == '':
                self.mark_times(alias, 'ctime', 'mtime')
                self.new_trees.append(alias)
                result = 0
            else:
                result = -errno.EFAULT
        except DisconnectedError:
            self.unlink_pipeline(alias)
            result = -errno.EFAULT
        except CommFailure:
            result = -errno.EFAULT
        return result

    def mknod(self, path, mode, dev):
        if mode & stat.S_IFREG:
            result = self.create(path, mode, '')
        else:
            result = None
        return result

    def read(self, path, size, offset):
        if path == '/00README':
            result = self.readme.read(size, offset)
        else:
            (alias, path) = self.split_pipeline_path(path)
            try:
                result = self.trees[alias][0].read(path, size, offset)
            except KeyError:
                if path == '' and alias in self.new_trees:
                    self.mark_times(alias, 'atime')
                    result = ''
                else:
                    result = -errno.EINVAL
            except DisconnectedError:
                self.unlink_pipeline(alias)
                result = -errno.EFAULT
            except CommFailure:
                result = -errno.EFAULT
        return result

    def write(self, path, buf, offset):
        (alias, path) = self.split_pipeline_path(path)
        try:
            result = self.trees[alias][0].write(path, buf, offset)
        except KeyError:
            if path == '' and alias in self.new_trees:
                result = self.link_pipeline(alias, buf.strip())
                if result == 0:
                    result = len(buf)
            else:
                result = -errno.EINVAL
        except DisconnectedError:
            self.unlink_pipeline(alias)
            result = -errno.EFAULT
        except CommFailure:
            result = -errno.EFAULT
        return result

    def truncate(self, path, len):
        (alias, path) = self.split_pipeline_path(path)
        try:
            result = self.trees[alias][0].truncate(path, len)
        except KeyError:
            if path == '' and alias in self.new_trees:
                result = self.link_pipeline(alias, buf.strip())
            else:
                result = -errno.ENOENT
        except DisconnectedError:
            self.unlink_pipeline(alias)
            result = -errno.EFAULT
        except CommFailure:
            result = -errno.EFAULT
        return result

    def rmdir(self, path):
        (alias, path) = self.split_pipeline_path(path)
        try:
            tree = self.trees[alias][0]
        except KeyError:
            result = -errno.EFAULT
        else:
            if path != '' and path != '/':
                try:
                    result = self.trees[alias][0].rmdir(path)
                except DisconnectedError:
                    self.unlink_pipeline(alias)
                    result = -errno.EFAULT
                except CommFailure:
                    result = -errno.EFAULT
            else:
                self.unlink_pipeline(alias)
                result = 0
        return result

    def readlink(self, path):
        (alias, path) = self.split_pipeline_path(path)
        try:
            result = self.trees[alias][0].readlink(path)
        except KeyError:
            result = None
        except DisconnectedError:
            self.unlink_pipeline(alias)
            result = -errno.EFAULT
        except CommFailure:
            result = -errno.EFAULT
        return result

    def getxattr(self, path, name, size):
        try:
            (xattr, lasttm) = self.xattrs[path]
        except KeyError:
            (xattr, lasttm) = (None, None)
        if (xattr is None) or (time() > lasttm + self.cache_timeout):
            (alias, path) = self.split_pipeline_path(path)
            try:
                result = self.trees[alias][0].getxattr(path, name, size)
                self.xattrs[path] = (result, time())
            except KeyError:
                result = None
                self.xattrs[path] = (result, time())
            except DisconnectedError:
                self.unlink_pipeline(alias)
                result = -errno.EFAULT
            except CommFailure:
                result = -errno.EFAULT
        else:
            result = xattr
        return result

    def listxattr(self, path, size):
        (alias, path) = self.split_pipeline_path(path)
        try:
            result = self.trees[alias][0].listxattr(path, size)
        except KeyError:
            result = None
        except DisconnectedError:
            self.unlink_pipeline(alias)
            result = -errno.EFAULT
        except CommFailure:
            result = -errno.EFAULT
        return result

def main():

    usage = "WCBE filesystem" + Fuse.fusage

    server = PipelineFS(usage=usage)
    server.parse(errex=1)
    optdict = server.fuse_args.optdict
    if 'uid' not in optdict:
        optdict['uid'] = os.geteuid()
    if 'gid' not in optdict:
        optdict['gid'] = os.getegid()
    server.fuse_args.optlist.add('direct_io')
    server.fuse_args.optlist.add('nonempty')
    server.fuse_args.optlist.add('allow_other')
    server.fuse_args.optlist.add('noatime')

    #if server.fuse_args.mount_expected():
    #    os.chdir(server.root)

    server.main()

if __name__ == '__main__':
    main()
