# -*- mode: python -*-
import socket
import stat
import os
import struct
import threading
import select
import errno
from tempfile import mkstemp
from time import time, sleep
from ThreadSafeDict import Dict

from Enum import Enum

class DumpFormat (object):
    __metaclass__ = Enum
    __enums = {'ascii':0, 'binary':1, 'none':2}

class Flow (object):
    __metaclass__ = Enum
    __enums = {'off':0, 'on':1}

class WcbeCorrBits (object):
    __metaclass__ = Enum
    __enums = {'four': 4, 'seven': 7}

class MsgType (object):
    __metaclass__ = Enum
    __enums = {'config_create':0, 'add_activation': 1,
               'remove_activation': 2, 'config_destroy':3,
               'config_list':4, 'activation_list': 5, 'src':6, 'ts0':7,
               'framedump':8, 'flow':9, 'rate':10, 'name':11, 'quit':12,
               'pipeline_list':13, 'element_list':14, 'property_list':15,
               'property':16, 'pad_list':17, 'pad_caps':18, 'pad_peer':19,
               'config_doc':20, 'aux_output_dir':21, 'release':22,
               'new_subscan':23, 'bdf_output_dir':24, 'config_integration':25,
               'cancel_subscan':26}

class MsgDirection (object):
    __metaclass__ = Enum
    __enums = {'request':0, 'response':1}

class MsgIntent (object):
    __metaclass__ = Enum
    __enums = {'command':0, 'query':1}

def str_trunc(str):
    null = str.find('\0')
    if null == -1:
        result = str
    else:
        result = str[:null]
    return result

def recvfrom(sock, fmt):
    sz = struct.calcsize(fmt)
    result = ''
    addr = None
    while sz > 0:
        (msg, addr) = sock.recvfrom(sz, socket.MSG_WAITALL)
        result += msg
        sz -= len(msg)
    return (result, addr)

class CbeMsg (object):

    format_str = ''
    type = ''
    base_fmt = 'iiiI'

    def __init__(self, direction, intent, id):
        self.fields = {}
        self.fields['intent'] = intent
        self.fields['direction'] = direction
        self.id = id
        return

    intent = property(lambda (self):self.fields['intent'])
    direction = property(lambda (self):self.fields['direction'])

    def msg_classes():
        return (ConfigCreateMsg, AddActivationMsg, RemoveActivationMsg,
                ConfigDestroyMsg, ConfigListMsg, ActivationListMsg, SrcMsg,
                Ts0Msg, FramedumpMsg, FlowMsg, RateMsg, NameMsg, QuitMsg,
                PipelineListMsg, ElementListMsg, PropertyListMsg, PropertyMsg,
                PadListMsg, PadCapsMsg, PadPeerMsg, ConfigDocMsg,
                AuxOutputDirMsg, ReleaseMsg, NewSubscanMsg, BdfOutputDirMsg,
                ConfigIntegrationMsg, CancelSubscanMsg)
    msg_classes = staticmethod(msg_classes)

    def class_lookup(code):
        klass = [k for k in CbeMsg.msg_classes() if int(k.type) == code]
        if len(klass) != 1:
            raise IndexError, code
        return klass[0]
    class_lookup = staticmethod(class_lookup)

    def msg_type(str):
        (typ, dir, intent, id) = \
              struct.unpack(CbeMsg.base_fmt,
                            str[:struct.calcsize(CbeMsg.base_fmt)])
        return CbeMsg.class_lookup(typ)
    msg_type = staticmethod(msg_type)

    def pack(self):
        packargs = self.pack_args()
        if len(packargs) == 2:
            (fmt, args) = packargs
            result = struct.pack(CbeMsg.base_fmt + fmt, self.type,
                                 self.direction, self.intent, self.id, *args)
        else:
            # We allow for the possibility that a call to pack_args
            # specifies the alignment for the call to struct.pack.
            # This seems somewhat brittle, but ought to work in those
            # instances where we need to put an arbitrary type at the
            # end of the message (i.e, only PropertyMsg for now).
            (fmt, args, align) = packargs
            result = struct.pack(align + CbeMsg.base_fmt + fmt, self.type,
                                 self.direction, self.intent, self.id, *args)
        return result

    def unpack(str):
        size = len(str) - struct.calcsize(CbeMsg.base_fmt)
        (type, dir, intent, id, msg) = \
               struct.unpack((CbeMsg.base_fmt + '%ds') % size, str)
        klass = CbeMsg.class_lookup(type)
        fields = (dir, intent, id) + klass.unpack(msg)
        return klass(*klass.convert_fields(*fields))
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id):
        return (MsgDirection(direction), MsgIntent(intent), id)
    convert_fields = staticmethod(convert_fields)

    def recvfrom(cls, sock):
        fmt = CbeMsg.base_fmt + cls.format_str
        return recvfrom(sock, fmt)
    recvfrom = classmethod(recvfrom)

class StrMsg (CbeMsg):

    format_str = "256s"

    def __init__(self, intent, id=0, name='', direction=MsgDirection.request):
        super(StrMsg, self).__init__(direction, intent, id)
        self.fields['name'] = str_trunc(name)
        return

    name = property(lambda (self):self.fields['name'])

    def pack_args(self):
        return (self.format_str, (self.name,))

    def unpack(str):
        return struct.unpack(StrMsg.format_str, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, name):
        return (MsgIntent(intent), id, name, MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class NameMsg (StrMsg):
    type = MsgType.name

class SrcMsg (StrMsg):
    type = MsgType.src

class ConfigCreateMsg (CbeMsg):

    type = MsgType.config_create
    format_str = '256sI%ds'

    def __init__(self, name, desc, intent=MsgIntent.command, id=0,
                 direction=MsgDirection.request):
        super(ConfigCreateMsg, self).__init__(direction, intent, id)
        self.fields['name'] = str_trunc(name)
        self.fields['desc'] = str_trunc(desc)
        return

    name = property(lambda (self):self.fields['name'])
    desc = property(lambda (self):self.fields['desc'])

    def pack_args(self):
        desc_len = len(self.desc) + 1
        return (self.format_str % desc_len, (self.name, desc_len, self.desc))

    def unpack(str):
        size = len(str) - struct.calcsize(ConfigCreateMsg.format_str[:-3])
        return struct.unpack(ConfigCreateMsg.format_str % size, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, name, desc_len, desc):
        return (name, desc, MsgIntent(intent), id, MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

    def recvfrom(cls, sock):
        fmt = CbeMsg.base_fmt + cls.format_str[:5]
        msg = sock.recv(struct.calcsize(fmt),
                        socket.MSG_PEEK | socket.MSG_WAITALL)
        f = struct.unpack(fmt, msg)
        fmt = CbeMsg.base_fmt + (cls.format_str % f[-1])
        return recvfrom(sock, fmt)
    recvfrom = classmethod(recvfrom)

class ActivationMsg (CbeMsg):

    format_str = None
    __format_prefix = '256si'
    __format_suffix = 'Bxxx'

    def __init__(self, config, epoch=0, seconds=0, cycles=0,
                 now=False, intent=MsgIntent.command, id=0,
                 direction=MsgDirection.request):
        super(ActivationMsg, self).__init__(direction, intent, id)
        self.fields['config'] = str_trunc(config)
        self.fields['epoch'] = epoch
        self.fields['seconds'] = seconds
        self.fields['cycles'] = cycles
        self.fields['now'] = now
        return

    def __compute_fmtstr():
        if struct.calcsize('Q') == 4:
            ifmt = 'Q'
        elif struct.calcsize('L') == 4:
            ifmt = 'L'
        elif struct.calcsize('I') == 4:
            ifmt = 'I'
        elif struct.calcsize('H') == 4:
            ifmt = 'H'
        else:
            raise Exception('Failed to determine struct format for 32 bit '
                            'integer')
        ActivationMsg.format_str = ActivationMsg.__format_prefix + ifmt + \
                                   ifmt + ActivationMsg.__format_suffix
        return ActivationMsg.format_str
    __compute_fmtstr = staticmethod(__compute_fmtstr)

    config = property(lambda (self):self.fields['config'])
    epoch = property(lambda (self):self.fields['epoch'])
    seconds = property(lambda (self):self.fields['seconds'])
    cycles = property(lambda (self):self.fields['cycles'])
    now = property(lambda (self):self.fields['now'])

    def pack_args(self):
        if self.format_str is None:
            self.format_str = ActivationMsg.__compute_fmtstr()
        return (self.format_str,
                (self.config, self.now, self.seconds, self.cycles, self.epoch))

    def unpack(str):
        if ActivationMsg.format_str is None:
            ActivationMsg.__compute_fmtstr()
        return struct.unpack(AddActivationMsg.format_str, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, config, now,
                       seconds, cycles, epoch):
        return (config, epoch, seconds, cycles, now, MsgIntent(intent), id,
                MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class AddActivationMsg (ActivationMsg):
    type = MsgType.add_activation

class RemoveActivationMsg (ActivationMsg):
    type = MsgType.remove_activation

class ConfigDestroyMsg (StrMsg):
    type = MsgType.config_destroy

class ConfigList:

    def __init__(self):
        self.ready = ()
        self.dying = ()
        return

    def encode(self):
        list = []
        for c in self.ready:
            list.append('(%s)' % c)
        for c in self.dying:
            list.append('[%s]'% c)
        if len(list) > 0:
            result = reduce(lambda x, y: '%s\t%s' % (x, y), list)
        else:
            result = ""
        return result

    def decode(self, configs_len, configs_str):
        configs = []
        if configs_len > 0:
            configs = configs_str.split('\t')
            configs[-1] = str_trunc(configs[-1])
        def filter(chars):
            return [c[1:-1] for c in configs if c[0] in chars]
        self.ready = tuple(filter('('))
        self.dying = tuple(filter('['))
        return

class ConfigListMsg (CbeMsg):

    format_str = 'I%ds'
    type = MsgType.config_list

    def __init__(self, configs=ConfigList(),
                 intent=MsgIntent.query, id=0, direction=MsgDirection.request):
        super(ConfigListMsg, self).__init__(direction, intent, id)
        self.fields['configs'] = configs
        return

    configs = property(lambda (self):self.fields['configs'])

    def pack_args(self):
        result = self.fields['configs'].encode()
        rlen = len(result) + 1
        if rlen != 1:
            result = (self.format_str % rlen, (rlen, result))
        else:
            result = (self.format_str[:-3], (0,))
        return result

    def unpack(str):
        size = len(str) - struct.calcsize(ConfigListMsg.format_str[:-3])
        return struct.unpack(ConfigListMsg.format_str % size, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, configs_len, configs_str):
        configs = ConfigList()
        configs.decode(configs_len, configs_str)
        return (configs, MsgIntent(intent), id, MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

    def recvfrom(cls, sock):
        fmt = CbeMsg.base_fmt + cls.format_str[:1]
        msg = sock.recv(struct.calcsize(fmt),
                        socket.MSG_PEEK | socket.MSG_WAITALL)
        f = struct.unpack(fmt, msg)
        fmt = CbeMsg.base_fmt + (cls.format_str % f[-1])
        return recvfrom(sock, fmt)
    recvfrom = classmethod(recvfrom)

class ActivationList:

    def __init__(self):
        self.ready = ()
        self.in_use = ()
        self.dying = ()
        return

    def encode(self):
        list = []
        for c in self.ready:
            list.append('(%s)' % c)
        for c in self.in_use:
            list.append('<%s>' % c)
        for c in self.dying:
            list.append('[%s]' % c)
        if len(list) > 0:
            result = reduce(lambda x, y: '%s\t%s' % (x, y), list)
        else:
            result = ""
        return result

    def decode(self, configs_len, configs_str):
        configs = []
        if configs_len > 0:
            configs = configs_str.split('\t')
            configs[-1] = str_trunc(configs[-1])
        def filter(chars):
            return [c[1:-1] for c in configs if c[0] in chars]
        self.ready = tuple(filter('('))
        self.in_use = tuple(filter('<'))
        self.dying = tuple(filter('['))
        return

class ActivationListMsg (CbeMsg):

    format_str = 'I%ds'
    type = MsgType.activation_list

    def __init__(self, activations=ActivationList(),
                 intent=MsgIntent.query, id=0, direction=MsgDirection.request):
        super(ActivationListMsg, self).__init__(direction, intent, id)
        self.fields['activations'] = activations
        return

    activations = property(lambda (self):self.fields['activations'])

    def pack_args(self):
        result = self.fields['activations'].encode()
        rlen = len(result) + 1
        if rlen != 1:
            result = (self.format_str % rlen, (rlen, result))
        else:
            result = (self.format_str[:-3], (0,))
        return result

    def unpack(str):
        size = len(str) - struct.calcsize(ActivationListMsg.format_str[:-3])
        return struct.unpack(ActivationListMsg.format_str % size, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, acts_len, acts_str):
        activations = ActivationList()
        activations.decode(acts_len, acts_str)
        return (activations, MsgIntent(intent), id, MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

    def recvfrom(cls, sock):
        fmt = CbeMsg.base_fmt + cls.format_str[:1]
        msg = sock.recv(struct.calcsize(fmt),
                        socket.MSG_PEEK | socket.MSG_WAITALL)
        f = struct.unpack(fmt, msg)
        fmt = CbeMsg.base_fmt + (cls.format_str % f[-1])
        return recvfrom(sock, fmt)
    recvfrom = classmethod(recvfrom)

class Ts0Msg (CbeMsg):

    format_str = "256s"
    type = MsgType.ts0

    def __init__(self, intent, id=0, ts0='', direction=MsgDirection.request):
        super(Ts0Msg, self).__init__(direction, intent, id)
        self.fields['ts0'] = str_trunc(ts0)
        return

    ts0 = property(lambda (self):self.fields['ts0'])

    def pack_args(self):
        return (self.format_str, (self.ts0,))

    def unpack(str):
        return struct.unpack(Ts0Msg.format_str, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, ts0):
        return (MsgIntent(intent), id, ts0, MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class FramedumpMsg (CbeMsg):

    format_str = "i"
    type = MsgType.framedump

    def __init__(self, intent, id=0, format=DumpFormat.none,
                 direction=MsgDirection.request):
        super(FramedumpMsg, self).__init__(direction, intent, id)
        self.fields['format'] = format
        return

    format = property(lambda (self):self.fields['format'])

    def pack_args(self):
        return (self.format_str, (self.format,))

    def unpack(str):
        return struct.unpack(FramedumpMsg.format_str, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, format):
        return (MsgIntent(intent), id, DumpFormat(format),
                MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class FlowMsg (CbeMsg):

    format_str = "i"
    type = MsgType.flow

    def __init__(self, intent, id=0, state=Flow.off,
                 direction=MsgDirection.request):
        super(FlowMsg, self).__init__(direction, intent, id)
        self.fields['state'] = state
        return

    state = property(lambda (self):self.fields['state'])

    def pack_args(self):
        return (self.format_str, (self.state,))

    def unpack(str):
        return struct.unpack(FlowMsg.format_str, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, state):
        return (MsgIntent(intent), id, Flow(state), MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class RateMsg (CbeMsg):

    format_str = "I"
    type = MsgType.rate

    def __init__(self, intent, id=0, rate=0,
                 direction=MsgDirection.request):
        super(RateMsg, self).__init__(direction, intent, id)
        self.fields['rate'] = rate
        return

    rate = property(lambda (self):self.fields['rate'])

    def pack_args(self):
        return (self.format_str, (self.rate,))

    def unpack(str):
        return struct.unpack(RateMsg.format_str, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, rate):
        return (MsgIntent(intent), id, rate, MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class QuitMsg (CbeMsg):

    format_str = "I"
    type = MsgType.quit

    def __init__(self, id=0, intent=MsgIntent.command,
                 direction=MsgDirection.request):
        super(QuitMsg, self).__init__(direction, intent, id)
        return

    def pack_args(self):
        return (self.format_str, (0,))

    def unpack(str):
        return struct.unpack(QuitMsg.format_str, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, reserved):
        return (id, MsgIntent(intent), MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class ReleaseMsg (CbeMsg):

    format_str = "I"
    type = MsgType.release

    def __init__(self, id=0, intent=MsgIntent.command,
                 direction=MsgDirection.request):
        super(ReleaseMsg, self).__init__(direction, intent, id)
        return

    def pack_args(self):
        return (self.format_str, (0,))

    def unpack(str):
        return struct.unpack(ReleaseMsg.format_str, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, reserved):
        return (id, MsgIntent(intent), MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class StringListMsg (CbeMsg):

    format_str = '256sI%ds'

    def __init__(self, intent=MsgIntent.query, id=0,
                 direction=MsgDirection.request):
        super(StringListMsg, self).__init__(direction, intent, id)
        return

    def pack_args(self, name_field, list_field):
        if len(self.fields[list_field]) > 0:
            joined = reduce(lambda x, y: '%s\t%s' % (x, y),
                            self.fields[list_field])
        else:
            joined = ""
        jlen = len(joined) + 1
        if jlen != 1:
            result = (self.format_str % jlen,
                      (self.fields[name_field], jlen, joined))
        else:
            result = (self.format_str[:-3], (self.fields[name_field], 0))
        return result

    def unpack(str):
        size = len(str) - struct.calcsize(StringListMsg.format_str[:-3])
        return struct.unpack(StringListMsg.format_str % size, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, name, list_len, list_str):
        list = []
        if list_len > 0:
            list = list_str.split('\t')
            list[-1] = str_trunc(list[-1])
        return (name, tuple(list), MsgIntent(intent), id,
                MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

    def recvfrom(cls, sock):
        fmt = CbeMsg.base_fmt + cls.format_str[:5]
        msg = sock.recv(struct.calcsize(fmt),
                        socket.MSG_PEEK | socket.MSG_WAITALL)
        f = struct.unpack(fmt, msg)
        fmt = CbeMsg.base_fmt + (cls.format_str % f[-1])
        return recvfrom(sock, fmt)
    recvfrom = classmethod(recvfrom)

class PipelineListMsg (StringListMsg):

    type = MsgType.pipeline_list

    def __init__(self, config='', pipelines=(), intent=MsgIntent.query, id=0,
                 direction=MsgDirection.request):
        super(PipelineListMsg, self).__init__(intent, id, direction)
        self.fields['config'] = str_trunc(config)
        self.fields['pipelines'] = pipelines
        return

    config = property(lambda(self):self.fields['config'])
    pipelines = property(lambda (self):self.fields['pipelines'])

    def pack_args(self):
        return super(PipelineListMsg, self).pack_args('config', 'pipelines')

class ElementListMsg (StringListMsg):

    type = MsgType.element_list

    def __init__(self, pipeline='', elements=(), intent=MsgIntent.query, id=0,
                 direction=MsgDirection.request):
        super(ElementListMsg, self).__init__(intent, id, direction)
        self.fields['pipeline'] = str_trunc(pipeline)
        self.fields['elements'] = elements
        return

    pipeline = property(lambda(self):self.fields['pipeline'])
    elements = property(lambda (self):self.fields['elements'])

    def pack_args(self):
        return super(ElementListMsg, self).pack_args('pipeline', 'elements')

class PropertyListMsg (StringListMsg):

    type = MsgType.property_list

    def __init__(self, element='', properties=(), intent=MsgIntent.query, id=0,
                 direction=MsgDirection.request):
        super(PropertyListMsg, self).__init__(intent, id, direction)
        self.fields['element'] = str_trunc(element)
        self.fields['properties'] = properties
        return

    element= property(lambda(self):self.fields['element'])
    properties = property(lambda (self):self.fields['properties'])

    def pack_args(self):
        return super(PropertyListMsg, self).pack_args('element', 'properties')

class PadListMsg (StringListMsg):

    type = MsgType.pad_list

    def __init__(self, element='', pads=(), intent=MsgIntent.query, id=0,
                 direction=MsgDirection.request):
        super(PadListMsg, self).__init__(intent, id, direction)
        self.fields['element'] = str_trunc(element)
        self.fields['pads'] = pads
        return

    element= property(lambda(self):self.fields['element'])
    pads = property(lambda (self):self.fields['pads'])

    def pack_args(self):
        return super(PadListMsg, self).pack_args('element', 'pads')

class PropertyMsg (CbeMsg):

    type = MsgType.property

    def __init__(self, prop='', value=None, valtype='',
                 intent=MsgIntent.query, id=0, direction=MsgDirection.request):
        super(PropertyMsg, self).__init__(direction, intent, id)
        self.fields['prop'] = str_trunc(prop)
        self.valtype = valtype
        format_str = '256s256sI'
        if value is None:
            self.fields['value'] = None
            self.format_str = format_str
            self.value_len = 0
        else:
            if valtype == 'gchar':
                self.fields['value'] = ord(str(value)[0])
                self.value_len = struct.calcsize('b')
                self.format_str = format_str + "b"
            elif valtype == 'guchar':
                self.fields['value'] = ord(str(value)[0])
                self.value_len = struct.calcsize('B')
                self.format_str = format_str + "B"
            elif valtype == 'gboolean':
                self.fields['value'] = int(value)
                self.value_len = struct.calcsize('i')
                self.format_str = format_str + 'i'
            elif valtype == 'gint':
                self.fields['value'] = int(value)
                self.value_len = struct.calcsize('i')
                self.format_str = format_str + 'i'
            elif valtype == 'guint':
                self.fields['value'] = long(value)
                self.value_len = struct.calcsize('I')
                self.format_str = format_str + 'I'
            elif valtype == 'glong':
                self.fields['value'] = int(value)
                self.value_len = struct.calcsize('l')
                self.format_str = format_str + 'l'
            elif valtype == 'gulong':
                self.fields['value'] = long(value)
                self.value_len = struct.calcsize('L')
                self.format_str = format_str + 'L'
            elif valtype == 'gint64':
                self.fields['value'] = long(value)
                self.value_len = struct.calcsize('q')
                self.format_str = format_str + 'q'
            elif valtype == 'guint64':
                self.fields['value'] = long(value)
                self.value_len = struct.calcsize('Q')
                self.format_str = format_str + 'Q'
            elif valtype == 'gfloat':
                self.fields['value'] = float(value)
                self.value_len = struct.calcsize('f')
                self.format_str = format_str + 'f'
            elif valtype == 'gdouble':
                self.fields['value'] = float(value)
                self.value_len = struct.calcsize('d')
                self.format_str = format_str + 'd'
            elif valtype == 'WcbeFramedumpFormat':
                self.fields['value'] = int(DumpFormat(value))
                self.value_len = struct.calcsize('i')
                self.format_str = format_str + 'i'
            elif valtype == 'WcbeCorrBits':
                self.fields['value'] = int(WcbeCorrBits(value))
                self.value_len = struct.calcsize('i')
                self.format_str = format_str + 'i'
            else: # either valtype == 'gchararray' or something else
                self.fields['value'] = str(value)
                self.value_len = len(self.fields['value']) + 1
                self.format_str = format_str + ("%ds" % self.value_len)
        return

    prop = property(lambda(self):self.fields['prop'])
    value = property(lambda(self):self.fields['value'])

    def pack_args(self):
        if self.value is None:
            result = (self.format_str, (self.prop, '', 0))
        else:
            result = (self.format_str,
                      (self.prop, self.valtype, self.value_len, self.value),
                      '=')
        return result

    def unpack(str):
        size = len(str) - struct.calcsize('256s256sI')
        return struct.unpack('256s256sI%ds' % size, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, property, valtype,
                       value_len, value_str):
        valtype = str_trunc(valtype)
        if valtype == 'gchararray':
            value = value_str
        elif valtype == 'gchar':
            (value,) = struct.unpack('b', value_str)
        elif valtype == 'guchar':
            (value,) = struct.unpack('B', value_str)
        elif valtype == 'gboolean':
            (bool,) = struct.unpack('i', value_str)
            if bool == 0:
                value = False
            else:
                value = True
        elif valtype == 'gint':
            (value,) = struct.unpack('i', value_str)
        elif valtype == 'guint':
            (value,) = struct.unpack('I', value_str)
        elif valtype == 'glong':
            (value,) = struct.unpack('l', value_str)
        elif valtype == 'gulong':
            (value,) = struct.unpack('L', value_str)
        elif valtype == 'gint64':
            (value,) = struct.unpack('q', value_str)
        elif valtype == 'guint64':
            (value,) = struct.unpack('Q', value_str)
        elif valtype == 'gfloat':
            (value,) = struct.unpack('f', value_str)
        elif valtype == 'gdouble':
            (value,) = struct.unpack('d', value_str)
        elif valtype == 'WcbeFramedumpFormat':
            (enum,) = struct.unpack('i', value_str)
            if enum == 0:
                value = 'ascii'
            elif enum == 1:
                value = 'binary'
            else:
                value = 'none'
        elif valtype == 'WcbeCorrBits':
            (enum,) = struct.unpack('i', value_str)
            try:
                value = WcbeCorrBits(enum).nick()
            except ValueError:
                value = enum
        else:
            value = value_str # FIXME: convert enums and flags
        return (property, value, valtype, MsgIntent(intent), id,
                MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

    def recvfrom(cls, sock):
        pfmt = '256s256sI'
        fmt = CbeMsg.base_fmt + pfmt
        msg = sock.recv(struct.calcsize(fmt),
                        socket.MSG_PEEK | socket.MSG_WAITALL)
        f = struct.unpack(fmt, msg)
        fmt = CbeMsg.base_fmt + ((pfmt + '%ds') % f[-1])
        return recvfrom(sock, fmt)
    recvfrom = classmethod(recvfrom)

class PadCapsMsg (CbeMsg):

    format_str = "256s256s"
    type = MsgType.pad_caps

    def __init__(self, pad='', caps='', intent=MsgIntent.query, id=0,
                 direction=MsgDirection.request):
        super(PadCapsMsg, self).__init__(direction, intent, id)
        self.fields['pad'] = str_trunc(pad)
        self.fields['caps'] = str_trunc(caps)
        return

    pad = property(lambda (self):self.fields['pad'])
    caps = property(lambda (self):self.fields['caps'])

    def pack_args(self):
        return (self.format_str, (self.pad, self.caps))

    def unpack(str):
        return struct.unpack(PadCapsMsg.format_str, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, pad, caps):
        return (pad, caps, MsgIntent(intent), id, MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class PadPeerMsg (CbeMsg):

    format_str = "256s256s"
    type = MsgType.pad_peer

    def __init__(self, pad='', peer='', intent=MsgIntent.query, id=0,
                 direction=MsgDirection.request):
        super(PadPeerMsg, self).__init__(direction, intent, id)
        self.fields['pad'] = str_trunc(pad)
        self.fields['peer'] = str_trunc(peer)
        return

    pad = property(lambda (self):self.fields['pad'])
    peer = property(lambda (self):self.fields['peer'])

    def pack_args(self):
        return (self.format_str, (self.pad, self.peer))

    def unpack(str):
        return struct.unpack(PadPeerMsg.format_str, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, pad, peer):
        return (pad, peer, MsgIntent(intent), id, MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class ConfigDocMsg (CbeMsg):

    format_str = "256sI%ds"
    type = MsgType.config_doc

    def __init__(self, name='', doc='', intent=MsgIntent.query, id=0,
                 direction=MsgDirection.request):
        super(ConfigDocMsg, self).__init__(direction, intent, id)
        self.fields['name'] = str_trunc(name)
        self.fields['doc'] = str_trunc(doc)
        return

    name = property(lambda (self):self.fields['name'])
    doc = property(lambda (self):self.fields['doc'])

    def pack_args(self):
        doc_len = len(self.doc) + 1
        result = (self.format_str % doc_len, (self.name, doc_len, self.doc))
        return result

    def unpack(str):
        size = len(str) - struct.calcsize(ConfigDocMsg.format_str[:-3])
        return struct.unpack(ConfigDocMsg.format_str % size, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, name, doc_len, doc_str):
        return (name, doc_str, MsgIntent(intent), id, MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

    def recvfrom(cls, sock):
        fmt = CbeMsg.base_fmt + cls.format_str[:5]
        msg = sock.recv(struct.calcsize(fmt),
                        socket.MSG_PEEK | socket.MSG_WAITALL)
        f = struct.unpack(fmt, msg)
        fmt = CbeMsg.base_fmt + (cls.format_str % f[-1])
        return recvfrom(sock, fmt)
    recvfrom = classmethod(recvfrom)

class AuxOutputDirMsg (StrMsg):
    type = MsgType.aux_output_dir

class BdfOutputDirMsg (StrMsg):
    type = MsgType.bdf_output_dir

class NewSubscanMsg (CbeMsg):

    type = MsgType.new_subscan
    format_str = None
    __format_prefix = '256s256s'
    __format_suffix = 'Bxxx'

    def __init__(self, config, uid, epoch, cycles, seconds,
                 intent=MsgIntent.command, id=0,
                 direction=MsgDirection.request):
        super(NewSubscanMsg, self).__init__(direction, intent, id)
        self.fields['config'] = str_trunc(config)
        self.fields['uid'] = str_trunc(uid)
        self.fields['epoch'] = epoch
        self.fields['seconds'] = seconds
        self.fields['cycles'] = cycles
        return

    def __compute_fmtstr():
        if struct.calcsize('Q') == 4:
            ifmt = 'Q'
        elif struct.calcsize('L') == 4:
            ifmt = 'L'
        elif struct.calcsize('I') == 4:
            ifmt = 'I'
        elif struct.calcsize('H') == 4:
            ifmt = 'H'
        else:
            raise Exception('Failed to determine struct format for 32 bit '
                            'integer')
        NewSubscanMsg.format_str = NewSubscanMsg.__format_prefix + \
                                   ifmt + ifmt + NewSubscanMsg.__format_suffix
        return NewSubscanMsg.format_str
    __compute_fmtstr = staticmethod(__compute_fmtstr)

    config = property(lambda (self):self.fields['config'])
    uid = property(lambda (self):self.fields['uid'])
    epoch = property(lambda (self):self.fields['epoch'])
    seconds = property(lambda (self):self.fields['seconds'])
    cycles = property(lambda (self):self.fields['cycles'])

    def pack_args(self):
        if self.format_str is None:
            self.format_str = NewSubscanMsg.__compute_fmtstr()
        if self.uid is not None:
            uid = self.uid
        else:
            uid = ''
        return (self.format_str, (self.config, uid, self.seconds,
                                  self.cycles, self.epoch))

    def unpack(str):
        if NewSubscanMsg.format_str is None:
            NewSubscanMsg.__compute_fmtstr()
        return struct.unpack(NewSubscanMsg.format_str, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, config, uid, seconds, cycles,
                       epoch):
        return (config, uid, epoch, seconds, cycles,
                MsgIntent(intent), id, MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class CancelSubscanMsg (CbeMsg):

    type = MsgType.cancel_subscan
    format_str = '256s256s'

    def __init__(self, config, uid, intent=MsgIntent.command, id=0,
                 direction=MsgDirection.request):
        super(CancelSubscanMsg, self).__init__(direction, intent, id)
        self.fields['config'] = str_trunc(config)
        self.fields['uid'] = str_trunc(uid)
        return

    config = property(lambda (self):self.fields['config'])
    uid = property(lambda (self):self.fields['uid'])

    def pack_args(self):
        return (self.format_str, (self.config, self.uid))

    def unpack(str):
        return struct.unpack(CancelSubscanMsg.format_str, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, config, uid):
        return (config, uid, MsgIntent(intent), id, MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class ConfigIntegrationMsg (CbeMsg):

    type = MsgType.config_integration
    format_str = '256sI'

    def __init__(self, config, usec=0, intent=MsgIntent.query, id=0,
                 direction=MsgDirection.request):
        super(ConfigIntegrationMsg, self).__init__(direction, intent, id)
        self.fields['config'] = str_trunc(config)
        self.fields['usec'] = usec
        return

    config = property(lambda (self):self.fields['config'])
    usec = property(lambda (self):self.fields['usec'])

    def pack_args(self):
        return (self.format_str, (self.config, self.usec))

    def unpack(str):
        return struct.unpack(ConfigIntegrationMsg.format_str, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, config, usec):
        return (config, usec, MsgIntent(intent), id, MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

__next_socket_id = 0

def get_socket_id():
    global __next_socket_id
    result = __next_socket_id
    __next_socket_id += 1
    return result

def unix_socketpair(type=socket.SOCK_DGRAM):
    "This function can be replaced by socket.socketpair in Python 2.4."
    if 'socketpair' in dir(socket):
        (sock0, sock1) = socket.socketpair(socket.AF_UNIX, socket.SOCK_DGRAM)
    else:
        sock0 = socket.socket(socket.AF_UNIX, type)
        name0 = struct.pack('b107s', 0, '/PipelineMC/%d/%d' %
                            (os.getpid(), get_socket_id()))
        sock0.bind(name0)
        sock1 = socket.socket(socket.AF_UNIX, type)
        name1 = struct.pack('b107s', 0, '/PipelineMC/%d/%d' % \
                            (os.getpid(), get_socket_id()))
        sock1.bind(name1)
        sock0.connect(name1)
        sock1.connect(name0)
    return (sock0, sock1)

class MsgFIFO (object):

    def __init__(self, sock=None):
        self.cond = threading.Condition()
        self.fifo = []
        self.sock = sock
        return

    def __del__(self):
        if self.sock is not None:
            self.sock.shutdown(socket.SHUT_RDWR)
        return

    def push(self, msg):
        self.cond.acquire()
        self.fifo.append(msg)
        if self.sock is not None:
            self.sock.send('a')
        self.cond.notify()
        self.cond.release()
        return

    def pop(self, timeout=None):
        self.cond.acquire()
        if len(self.fifo) == 0:
            self.cond.wait(timeout)
        if len(self.fifo) > 0:
            result = self.fifo.pop(0)
        else:
            result = None
        self.cond.release()
        return result

    def len(self):
        return len(self.fifo)

class ControlSockType (object):
    __metaclass__ = Enum
    __enums = {'unix': 0, 'udp': 1, 'sctp': 2, 'tcp': 3}

class SocketMonitor (threading.Thread):

    def __init__(self, name, inet=None, unix=None, pop_timeout=None):
        super(SocketMonitor, self).__init__(name=name)
        self.setDaemon(1)
        self.in_fifos = {}
        self.out_fifos = {}
        self.out_fifo_socks = {}
        self.out_fifo_notice = unix_socketpair()
        self.pop_timeout = pop_timeout
        self.in_lock = threading.Lock() # guards self.in_fifos
        self.out_lock = threading.Lock() # guards self.out_fifos
        self.inet = inet
        self.unix = unix
        self.addr = None
        self.in_shutdown = []
        return

    def sock(self):
        if self.inet is not None:
            result = self.inet
        else:
            result = self.unix
        return result

    def shutdown(self, addr):
        self.in_lock.acquire()
        this_shutdown = addr
        self.in_shutdown.append(addr)
        self.in_lock.release()
        outlen = 1
        while outlen > 0:
            sleep(0.1)
            self.out_lock.acquire()
            outlen = reduce(lambda a, f: a + f.len(),
                            [fifo for (addr, (fifo, socks))
                             in self.out_fifos.items()
                             if addr == this_shutdown], 0)
            self.out_lock.release()
        self.in_lock.acquire()
        del self.in_shutdown[self.in_shutdown.index(this_shutdown)]
        self.in_lock.release()

        self.out_lock.acquire()
        fd = self.out_fifos[addr][1][1].fileno()
        del self.out_fifos[addr]
        del self.out_fifo_socks[fd]
        self.out_fifo_notice[0].send('a')
        self.out_lock.release()
        return

    def __push_msg(self, addr, msg):
        self.in_lock.acquire()
        if addr not in self.in_shutdown:
            try:
                fifo = self.in_fifos[addr]
            except KeyError:
                fifo = MsgFIFO()
                self.in_fifos[addr] = fifo
            result = fifo.push(msg)
        else:
            result = None
        self.in_lock.release()
        return result

    def pop_msg(self, addr, timeout=None):
        self.in_lock.acquire()
        try:
            fifo = self.in_fifos[addr]
        except KeyError:
            fifo = None
        self.in_lock.release()
        if fifo is not None:
            if timeout is None:
                timeout = self.pop_timeout
            result = fifo.pop(timeout)
        else:
            result = None
        return result

    def push_msg(self, msg, addr=0):
        self.out_lock.acquire()
        try:
            fifo = self.out_fifos[addr][0]
        except KeyError:
            socks = unix_socketpair()
            fifo = MsgFIFO(socks[0])
            self.out_fifos[addr] = (fifo, socks)
            self.out_fifo_socks[socks[1].fileno()] = (socks[1], addr, fifo)
            self.out_fifo_notice[0].send('a')
        self.out_lock.release()
        return fifo.push(msg)

    def run(self):
        inet = None
        unix = None
        hdrsize = struct.calcsize(CbeMsg.base_fmt)
        poll = select.poll()
        if self.inet is not None:
            ctlsock = self.inet
        if self.unix is not None:
            ctlsock = self.unix
        ctlsockfd = ctlsock.fileno()
        poll.register(ctlsockfd, select.POLLIN | select.POLLHUP)
        poll.register(self.out_fifo_notice[1], select.POLLIN)
        out_fifo_notice_fd = self.out_fifo_notice[1].fileno()
        fifo_fds = set([])
        finished = False
        send = None
        while not finished:
            try:
                events = poll.poll(1000)
            except select.error, e:
                events = []
            for (fd, event) in events:
                if fd == out_fifo_notice_fd:
                    self.out_fifo_notice[1].recv(1)
                    self.out_lock.acquire()
                    current_fifo_fds = set(self.out_fifo_socks.keys())
                    self.out_lock.release()
                    old_fifo_fds = fifo_fds - current_fifo_fds
                    for s in old_fifo_fds:
                        poll.unregister(s)
                    new_fifo_fds = current_fifo_fds - fifo_fds
                    for s in new_fifo_fds:
                        poll.register(s, select.POLLIN)
                    fifo_fds = current_fifo_fds
                elif fd in fifo_fds:
                    self.out_lock.acquire()
                    (sck, addr, fifo) = self.out_fifo_socks[fd]
                    self.out_lock.release()
                    sck.recv(1)
                    if send is None:
                        msg = fifo.pop(0.0)
                        try:
                            if addr == 0:
                                sent = ctlsock.send(msg)
                            else:
                                sent = ctlsock.sendto(msg, addr)
                        except socket.error:
                            sent = 0
                            msg = ''
                        except socket.timeout:
                            sent = 0
                            msg = ''
                        if sent != len(msg):
                            poll.unregister(ctlsockfd)
                            poll.register(ctlsockfd,
                                          select.POLLOUT | select.POLLHUP)
                            send = (msg[sent:], addr)
                elif (event & select.POLLOUT) != 0:
                    assert send is not None
                    (msg, addr) = send
                    try:
                        if addr == 0:
                            sent = ctlsock.send(msg)
                        else:
                            sent = ctlsock.sendto(msg, addr)
                    except socket.error:
                        sent = 0
                        msg = ''
                    except socket.timeout:
                        sent = 0
                        msg = ''
                    if sent == len(msg):
                        poll.unregister(ctlsockfd)
                        poll.register(ctlsockfd,
                                      select.POLLIN | select.POLLHUP)
                        send = None
                    else:
                        send = (msg[sent:], addr)
                elif (event & select.POLLIN) != 0:
                    try:
                        (msg, addr) = \
                            ctlsock.recvfrom(hdrsize, socket.MSG_PEEK)
                    except socket.error:
                        msg = None
                    except socket.timeout:
                        msg = None
                    if msg is not None:
                        if len(msg) == hdrsize:
                            poll.unregister(ctlsockfd)
                            typ = CbeMsg.msg_type(msg)
                            (msg, addr) = typ.recvfrom(ctlsock)
                            if self.addr is not None:
                                addr = self.addr
                                self.__push_msg(addr, msg)
                            poll.register(ctlsockfd,
                                          select.POLLIN | select.POLLHUP)
                        else:
                            # On pipeline shutdown, we sometimes get
                            # messages of zero length...along with
                            # POLLREMOVE event bit, but that isn't a
                            # constant defined within the select
                            # module.
                            finished = True
                elif (event & (select.POLLHUP | select.POLLERR)) != 0:
                    finished = True
        return

class DisconnectedError (Exception):

    def __init__(self):
        return

class PipelineInstance:

    def __init__(self, family, addr=None, name=None, monitor=None, timeout=1.0,
                 socktype=ControlSockType.tcp):
        if monitor is not None:
            self.monitor = monitor
        else:
            if socktype == ControlSockType.tcp:
                self.tcpsk = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self.tcpsk.settimeout(timeout)
                name = 'mon%s:%s' % (addr[0], addr[1])
                self.monitor = SocketMonitor(name=name,
                                             inet=self.tcpsk, unix=None,
                                             pop_timeout=timeout)
            else:
                raise "PipelineInstance needs 'monitor' when socket type " \
                    "is not TCP."
        self.family = family
        self.addr = addr
        self.name = name
        self.__msgid = 0
        self.lock = threading.RLock()
        self.num_aliases = 0
        self.socktype = socktype
        self.connected = False
        return

    def __del__(self):
        if self.connected and self.socktype == ControlSockType.tcp:
            if self.socktype == ControlSockType.tcp:
                self.monitor.shutdown(0)
            else:
                self.monitor.shutdown(self.addr)
            self.tcpsk.shutdown(socket.SHUT_RDWR)
        return

    def __next_msgid(self):
        result = self.__msgid
        self.__msgid = (self.__msgid + 1) % 65536
        return result

    msgid = property(lambda (self):self.__next_msgid())

    def connect(self):
        if not self.connected and (self.addr is not None):
            if self.socktype == ControlSockType.tcp:
                rc = self.monitor.inet.connect_ex(self.addr)
                if rc == 0:
                    self.connected = True
                elif rc == errno.EINPROGRESS:
                    p = select.poll()
                    sfd = self.monitor.inet.fileno()
                    p.register(sfd, select.POLLOUT)
                    events = p.poll(10000)
                    for (fd, event) in events:
                        if fd == sfd and event & select.POLLOUT != 0:
                            rc = self.monitor.inet.getsockopt(
                                socket.SOL_SOCKET, socket.SO_ERROR)
                            if rc == 0:
                                self.connected = True
                                break
            else:
                self.connected = True
            if self.connected:
                if not self.monitor.isAlive():
                    self.monitor.addr = self.addr
                    self.monitor.start()
        return self.connected

    def acquire(self):
        return self.lock.acquire()

    def release(self):
        return self.lock.release()

    def send_msg(self, msg):
        result = -1
        if self.connect() and self.monitor.isAlive():
            msg.id = self.msgid
            if self.socktype != ControlSockType.tcp:
                self.monitor.push_msg(msg.pack(), self.addr)
            else:
                self.monitor.push_msg(msg.pack())
            result = msg.id
        else:
            raise DisconnectedError()
        return result

    def recv_msg(self):
        msg = None
        if self.addr is not None:
            msg = self.monitor.pop_msg(self.addr)
            if msg is not None:
                msg = CbeMsg.unpack(msg)
        return msg

class PipelineClient (object):

    input_stage = "INPUT_STAGE"
    timeout = 2.5

    def __init__(self, sockdir='/tmp',
                 socktype=ControlSockType.tcp,
                 netif=None):
        super(PipelineClient, self).__init__()
        self.pipelines = Dict()
        self.sockdir = sockdir
        self.socktype = socktype
        self.monitor0 = None
        if socktype == ControlSockType.unix:
            self.usock = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
            # To get a reasonable name for our socket, I think that
            # the use of mktemp followed by bind would be sufficient;
            # but mktemp is deprecated, so we use mkstemp instead.
            # Unfortunately, mkstemp also produces an open file, which
            # is useless and must be closed and unlinked immediately
            # after opening. So there is a race condition here, but I
            # don't think that it's harmful.
            (tempfd, usockname) = mkstemp(dir=self.sockdir)
            os.close(tempfd)
            os.unlink(usockname)
            # the following raises socket.error if self.mysockname is in use
            self.usock.bind(usockname)
        else:
            self.usock = None
        if socktype == ControlSockType.udp:
            isock =  socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        elif socktype == ControlSockType.sctp:
            if hasattr(socket, 'IPPROTO_SCTP'):
                proto = socket.IPPROTO_SCTP
            else:
                # The socket module seems to do OK if we just use for
                # the protocol argument of socket.socket the integer
                # value of IPPROTO_SCTP, which is 132 on all Linux
                # systems that I've worked on, but I suppose it could
                # be inappropriate for some.
                proto = 132
            isock = socket.socket(socket.AF_INET, socket.SOCK_SEQPACKET,
                                  proto)
            if netif is not None:
                isock.bind((netif, 0))
        else:
            isock = None
        if (isock is not None) or (self.usock is not None):
            self.monitor0 = SocketMonitor(name='monitor0', inet=isock,
                                          unix=self.usock,
                                          pop_timeout=self.timeout)
        return

    def __del__(self):
        for p in self.pipelines.keys():
            self.remove_pipeline(p, False)
        if self.usock is not None:
            os.unlink(self.usock.getsockname())
        sleep(1.0)
        return

    def probe_addr(self, alias):
        try:
            probe = self.mc_transaction(alias=alias,
                                        msg=NameMsg(intent=MsgIntent.query))
        except DisconnectedError:
            probe = None
        if probe is not None:
            result = probe.name
        else:
            result = None
        return result

    def get_instance(self, name, host, port):
        result = None
        family = None
        if host is None and port is None:
            family = socket.AF_UNIX
            addr = os.path.join(self.sockdir, name, 'ctlsock')
        elif host is not None and port is not None:
            family = socket.AF_INET
            addr = (socket.gethostbyname(host), port)
        if family is not None:
            for p in self.pipelines.values():
                if p.family == family and p.addr == addr:
                    if name is None or p.name == name:
                        result = p
            if result is None:
                if self.socktype != ControlSockType.tcp:
                    monitor = self.monitor0
                else:
                    monitor = None
                result = PipelineInstance(monitor=monitor, family=family,
                                          addr=addr, name=name,
                                          timeout=self.timeout)
            if result is not None:
                result.acquire()
        return result

    def has_pipeline(self, alias):
        return alias in self.pipelines

    def add_pipeline(self, alias=None, name=None, host=None,
                     port=None, allow_new=True):
        result = False
        if alias is not None:
            # first create or find current PipelineInstance
            self.pipelines.acquire()
            pline = self.get_instance(name=name, host=host, port=port)
            if pline is not None:
                do_probe = False
                # pline was created, add it to self.pipelines
                # atomically to prevent a race condition with
                # another thread trying the same
                if alias not in self.pipelines:
                    self.pipelines[alias] = pline
                    self.pipelines.release()
                    pline.num_aliases += 1
                    # send a NameMsg probe
                    probe = self.probe_addr(alias)
                    if probe is not None:
                        if name is not None:
                            if name == probe:
                                result = True
                            else:
                                print 'Pipeline instance exists at %s, ' \
                                    'but the reported name is %s, ' \
                                    'not %s' % \
                                    (str(pline.addr), probe, name)
                        else:
                            if pline.name is None:
                                pline.name = probe
                            result = True
                    elif allow_new:
                        # probe failed, spawn new pipeline if allowed
                        args = ['wcbe']
                        if name is not None:
                            args.extend(['--name', name])
                        if pline.family == socket.AF_UNIX:
                            args.extend(['--sockdir', self.sockdir])
                            os.spawnvp(os.P_NOWAIT, 'wcbe', args)
                            do_probe = True
                        else: # pline.family == socket.AF_INET
                            args.extend(['--ctladdr',
                                         '%s:%d' % (pline.addr[0],
                                                    pline.addr[1])])
                            rc = os.system('ssh -x %s %s&' %
                                           (pline.addr[0], ' '.join(args)))
                            if rc == 0:
                                do_probe = True
                else:
                    # Alias already exists, verify that method
                    # arguments agree with pipeline instance
                    # parameters.
                    instance = self.pipelines[alias]
                    self.pipelines.release()
                    # Acquire the lock on the "current" pipeline
                    # instance to block in case another thread is in
                    # the process of spawning it.
                    instance.acquire()
                    try:
                        if instance.family == socket.AF_UNIX:
                            result = (host is None and port is None and
                                      os.path.join(self.sockdir, name,
                                                   'ctlsock') == instance.addr)
                        else: # pline.family = socket.AF_INET
                            result = ((socket.gethostbyname(host), port) ==
                                      instance.addr and
                                      (name is None or instance.name == name))
                    except:
                        pass
                    instance.release()

                if do_probe:
                    # send name probes to determine that pipeline
                    # is alive
                    t0 = time()
                    while time() - t0 < 15 and not result:
                        sleep(0.5)
                        probe = self.probe_addr(alias)
                        if probe is not None:
                            pline.name = probe
                            result = True

                # atomically remove PipelineInstance from
                # self.pipelines in case pipeline didn't start
                self.pipelines.acquire()
                if not result and alias in self.pipelines:
                    del self.pipelines[alias]
                self.pipelines.release()

                pline.release()

            else:
                self.pipelines.release()
        return result

    def remove_pipeline(self, alias, allow_quit=True):
        try:
            pline = self.pipelines[alias]
        except KeyError:
            pline = None
        if pline is not None:
            pline.acquire()
            pline.num_aliases -= 1
            if pline.num_aliases == 0:
                try:
                    if allow_quit:
                        self.mc_transaction(alias, QuitMsg())
                    else:
                        self.mc_transaction(alias, ReleaseMsg())
                except DisconnectedError:
                    pass
            self.pipelines.acquire()
            del self.pipelines[alias]
            self.pipelines.release()
            pline.release()
        return

    def send_msg(self, alias, msg):
        return self.pipelines[alias].send_msg(msg)

    def recv_msg(self, alias):
        return self.pipelines[alias].recv_msg()

    def mc_transaction(self, alias, msg):
        result = None
        if msg.direction == MsgDirection.request:
            pline = self.pipelines[alias]
            pline.acquire()
            try:
                msg_id = self.send_msg(alias, msg)
                if msg_id != -1:
                    if msg.intent == MsgIntent.query:
                        t0 = time()
                        result_id = msg_id - 1
                        while time() - t0 < self.timeout \
                                  and result_id != msg_id:
                            result = self.recv_msg(alias)
                            if result is not None:
                                result_id = result.id
                            else:
                                sleep(0.01)
                    else:
                        result = 0
            finally:
                pline.release()
        return result

    def connected(self, alias):
        pline = self.pipelines[alias]
        return pline.connected and pline.monitor.isActive()
