import os
import os.path
import stat
from glob import glob
from fnmatch import fnmatch
from pwd import getpwnam
from grp import getgrnam

class WcbeError (Exception):
    """Base class for wcbe exceptions.
    """
    pass

def set_environment(filename, env=os.environ):
    """Read default environment variables from 'filename' and set
    corresponding effective environment variables in 'env' if they do
    not already exist there . Note that quoted values in 'filename'
    are not properly parsed by this function.
    """
    for line in open(filename, 'r'):
        line = line.strip()
        if line.startswith('DEFAULT_WCBE'):
            line = line[len('DEFAULT_'):].split('#')[0]
            assign = line.split('=', 2)
            if len(assign) == 2:
                name = assign[0].strip()
                if name not in env:
                    env[name] = assign[1].split()[0]
    return

def check_pipelinefs():
    if 'WCBE_PIPELINEFS' in os.environ:
        for line in open('/proc/mounts', 'r'):
            fields = line.split()
            if fields[1] == os.environ['WCBE_PIPELINEFS'] \
               and (fields[0] == 'pipelinefs' or fields[0] == 'fuse'):
                uid = getpwnam(os.environ['WCBE_USER']).pw_uid
                grp = getgrnam(os.environ['WCBE_GROUP']).gr_gid
                statinfo = os.stat(os.environ['WCBE_PIPELINEFS'])
                if uid == statinfo[stat.ST_UID] \
                   and grp == statinfo[stat.ST_GID] \
                   and (statinfo[stat.ST_MODE]
                        & (stat.S_IRWXU | stat.S_IRWXG)) \
                        == (stat.S_IRWXU | stat.S_IRWXG) \
                   and (statinfo[stat.ST_MODE]
                        & (stat.S_IROTH | stat.S_IXOTH)) \
                        == (stat.S_IROTH | stat.S_IXOTH):
                    return True
    return False

def check_bdfdir():
    if 'WCBE_BDFDIR' in os.environ:
        private = os.path.join(os.environ['WCBE_BDFDIR'], '.private')
        if os.path.isdir(private):
            uid = getpwnam(os.environ['WCBE_USER']).pw_uid
            gid = getgrnam(os.environ['WCBE_GROUP']).gr_gid
            bdf_stat = os.stat(os.environ['WCBE_PIPELINEFS'])
            private_stat = os.stat(private)
            if gid == bdf_stat[stat.ST_GID] \
               and uid == private_stat[stat.ST_UID] \
               and (bdf_stat[stat.ST_MODE] & stat.S_IRWXG) == stat.S_IRWXG \
               and (private_stat[stat.ST_MODE] & stat.S_IRWXU) == stat.S_IRWXU:
                return True
    return False

def get_lfxset(env=os.environ):
    """Read the 'lfx' file for the list of lag frame interface
    hostnames, and generate the matched set of pipelinefs directory
    names together with a name glob pattern for instance numbers.
    """
    result = set([])
    num_id_parts = int(env['WCBE_NODE_ID_PARTS'])
    for line in open(env['WCBE_LFX_FILE'], 'r'):
        line = line.strip()
        hash = line.find('#')
        if hash != -1:
            line = line[:hash]
        if line.startswith(env['WCBE_NODE_PREFIX']):
            lfx = line[len(env['WCBE_NODE_PREFIX']):]
            num_parts = lfx.count('-') + 1
            if num_parts == num_id_parts + 1:
                result.add(lfx + '-*')
            else:
                print >> stderr, \
                      ('Name "%s" in %s does not follow wcbetool naming '\
                       'conventions for a lag frame interface hostname.' \
                       % (line, env['WCBE_LFX_FILE']))
    return result

def name_glob(patterns, names, free_proc_id, env=os.environ):
    """Match top-level pipelinefs directory 'patterns' to 'names'
    using shell wildcard-style matching. Matching is done on the
    dash-separated components of the patterns and names.
    'free_proc_ids' should be True or False depending on whether the
    matching ignores the process id part of the names.
    """
    num_id_parts = int(env['WCBE_NODE_ID_PARTS'])
    split_patterns = []
    for p in patterns:
        s = p.split('-')
        if len(s) < num_id_parts + 2:
            s.extend((num_id_parts + 2 - len(s)) * ['*'])
        if free_proc_id:
            if s[num_id_parts + 1] == '*':
                s[num_id_parts + 1] = '0'
        split_patterns.append(s)
    result = set([])
    for name in names:
        name_parts = name.split('-')
        if len(name_parts) == num_id_parts + 2:
            for p in split_patterns:
                if free_proc_id:
                    if all([fnmatch(*a) for a
                            in zip(name_parts[:-1], p[:-1])]):
                        result.add('-'.join(name_parts[:-1] + [p[-1]]))
                else:
                    if all([fnmatch(*a) for a in zip(name_parts, p)]):
                        result.add(name)
    return result

def pipelinefs_glob(patterns, lfxset, env=os.environ):
    paths = [p for p in glob(os.path.join(env['WCBE_PIPELINEFS'], '*'))]
    dirs = name_glob(patterns, [os.path.basename(p) for p in paths], False)
    dirs = name_glob(lfxset, dirs, False)
    dirs = [os.path.join(env['WCBE_PIPELINEFS'], d) for d in dirs]
    result = []
    for d in dirs:
        try:
            statinfo = os.stat(d)
            if stat.S_ISDIR(statinfo[stat.ST_MODE]):
                result.append(d)
        except OSError:
            pass
    return result

def read_pipeline_file(name):
    """Read a single string from the file 'name' (intended for
    pipelinefs files).
    """
    f = open(name, 'r')
    result = f.readline()
    f.close()
    return result.strip()

def instance_id(path, num_id_parts=-1):
    """Convert the path to a wcbe instance in pipelinefs to the
    hostname for the M&C interface of that instance, the hostname
    suffix, the nic id, and the process index.
    """
    pipelinedir = os.path.basename(path)
    parts = pipelinedir.split('-')
    if num_id_parts == -1:
        num_id_parts = int(os.environ['WCBE_NODE_ID_PARTS'])
    id = '-'.join(parts[:num_id_parts])
    hostname = os.environ['WCBE_NODE_PREFIX'] + id
    return (pipelinedir, hostname, id, int(parts[num_id_parts]),
            int(parts[num_id_parts + 1]))
